#include <dsassert.h>

#include <dslib.h>
#include <dsdebug.h>

#include <dsbufferpool.h>


static void dsBufferPoolZeroContext(DSBufferPoolContext ctx) {
    ds_assert_return(ctx);

    ctx->first = NULL;
    ctx->last = NULL;
    ctx->lock = NULL;
    ctx->allocated = NULL;
    ctx->create = NULL;
    ctx->restore = NULL;
    ctx->resize = NULL;
    ctx->replace = NULL;
    ctx->destroy = NULL;
}


DSBufferPoolContext dsBufferPoolCreateContext(DSMallocContext mctx) {
    DSBufferPoolContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "dsBufferPoolCreateContext");    

    ctx = (DSBufferPoolContext)dsMalloc(mctx,sizeof(struct DSBufferPoolContextT));
    if (ctx) {
	ctx->malloc=mctx;
	dsBufferPoolZeroContext(ctx);
    }
    return ctx;
}

int dsBufferPoolInitContext(DSBufferPoolContext ctx, unsigned long first_bin, unsigned char bins) {
    int err;
    unsigned char i;
    DSBufferPoolBuffer *ptr1, *ptr2;
    unsigned long *ptr3;

    dsDebugServer(DS_DEBUG_NOTICE, "dsBufferPoolInitContext");    

    ds_assert_param(ctx);
    ds_assert_param(bins);
    ds_assert_param(first_bin>8);

    ptr1 = (DSBufferPoolBuffer*)dsMalloc(ctx->malloc, bins*sizeof(DSBufferPoolBuffer));
    ptr2 = (DSBufferPoolBuffer*)dsMalloc(ctx->malloc, bins*sizeof(DSBufferPoolBuffer));
    ptr3 = (unsigned long*)dsMalloc(ctx->malloc, (bins+1)*sizeof(unsigned long));
    
    if ((!ptr1)||(!ptr2)||(!ptr3)) {
	if (ptr1) dsFree(ctx->malloc, ptr1);
	if (ptr2) dsFree(ctx->malloc, ptr2);
	if (ptr3) dsFree(ctx->malloc, ptr3);
	return DS_ERR_MEM;
    }
    
    for (i=0;i<bins;i++) {
	ptr1[i]=NULL;
	ptr2[i]=NULL;
	ptr3[i]=0;
    }
    ptr3[bins]=0;

    ctx->lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	dsBufferPoolFreeContext(ctx);
	return DS_ERR_MEM;
    }
    err = dsLockInitContext(ctx->lock);
    if (err) {
	dsBufferPoolFreeContext(ctx);
	return err;
    }

    ctx->bins = bins;

    ctx->first = ptr1;
    ctx->last = ptr2;
    ctx->allocated = ptr3;
    
    for (i=0;first_bin;first_bin>>=1) i++;
    ctx->lsr = i;
    
    return 0;
}

void dsBufferPoolFreeContext(DSBufferPoolContext ctx) {
    unsigned char i;
    DSBufferPoolBuffer ptr1, ptr2;

    dsDebugServer(DS_DEBUG_NOTICE, "dsBufferPoolFreeContext");    

    ds_assert_return(ctx);
    
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    
    // check allocated zeroes
    if (ctx->first) {
	for (i=0;i<ctx->bins;i++) {
	    for (ptr1=ctx->first[i];ptr1;ptr1=ptr2) {
		ptr2 = ptr1->next;
		if (ctx->destroy) ctx->destroy(ctx, ptr1);
		else dsFree(ctx->malloc, ptr1);
	    }
	}
	dsFree(ctx->malloc, ctx->first);
    }

    if (ctx->last) dsFree(ctx->malloc, ctx->last);

    if (ctx->allocated) {
	for (i=0;i<=ctx->bins;i++) {
	    if (ctx->allocated[i]) {
		if (i==ctx->bins) 
		    dsDebugServer(DS_DEBUG_ERROR, "dsBufferPoolFreeContext: We lost somethere %lu huge buffers", ctx->allocated[i]);
		else
		    dsDebugServer(DS_DEBUG_ERROR, "dsBufferPoolFreeContext: We lost somethere %lu buffers of size %lu", ctx->allocated[i], 1<<(ctx->lsr+i/*-1*/));
	    }
	}
	dsFree(ctx->malloc, ctx->allocated);
    }

    dsBufferPoolZeroContext(ctx);    
}

void dsBufferPoolDestroyContext(DSBufferPoolContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "dsBufferPoolDestroyContext");    
    dsBufferPoolFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

int dsBufferPoolSetFunctions(DSBufferPoolContext ctx, DSBufferPoolCreateBufferFunction create, DSBufferPoolRestoreBufferFunction restore, DSBufferPoolResizeBufferFunction resize, DSBufferPoolReplaceBufferFunction replace, DSBufferPoolDestroyBufferFunction destroy) {
    ctx->create = create;
    ctx->restore = restore;
    ctx->resize = resize;
    ctx->replace = replace;
    ctx->destroy = destroy;
    return 0;
}

int dsBufferPoolSetDefaultSize(DSBufferPoolContext ctx, unsigned long default_size) {
    unsigned long int i;
    unsigned char j;
    
//    default_size=65537;
    for (j=0,i=(default_size>>ctx->lsr);i;i>>=1) j++;
    i = (1<<(j+ctx->lsr-1));
    if (i<default_size) ctx->default_size = i<<1;
    else {
	ctx->default_size=i;
	j--;
    }

    if (j<ctx->bins) ctx->default_bin = j;
    else {
	ctx->default_size = 0;
	return DS_ERR_INVALID_ARGUMENT;
    }
    
    return 0;
}



DSBufferPoolBuffer dsBufferPoolGetBuffer(DSBufferPoolContext ctx, unsigned long size) {
    DSBufferPoolBuffer buffer;
    unsigned long int i;
    unsigned char j;
    
    ds_assert_null(ctx);
    
    if (size) {
        for (j=0,i=size>>ctx->lsr;i;i>>=1) j++;
	i = (1<<(j+ctx->lsr-1));
	if (i<size) {
	    size=i<<1;
	    j++;
	} else size=i;

	if (j>=ctx->bins) {
	    dsDebugServer(DS_DEBUG_NOTICE, "dsBufferPoolGetBuffer: Allocating new BigBuffer: %lu", size);

	    if (ctx->create) buffer = ctx->create(ctx, size);
	    else {
		buffer = dsMalloc(ctx->malloc, size + sizeof(struct DSBufferPoolBufferT));
		if (buffer) buffer->buffer_size = size;
	    }
	    if (!buffer) return buffer;

	    dsLockWrite(ctx->lock);
	    ctx->allocated[ctx->bins]++;
	    dsLockReturn(ctx->lock);
	    return buffer;
	}
    } else {
	size = ctx->default_size;
	if (size) j = ctx->default_bin;
	else ds_assert_null(size);
    }
    
    dsLockWrite(ctx->lock);
    buffer = ctx->first[j];
    if (buffer) {
	dsDebugServer(DS_DEBUG_DEBUG, "dsBufferPoolGetBuffer: GetBuffer: %lu", size);
	if (buffer->next) {
	    ctx->first[j]=buffer->next;
	} else {
	    ctx->first[j]=NULL;
	    ctx->last[j]=NULL;
	}
	i=0;
    } else {
	dsDebugServer(DS_DEBUG_NOTICE, "dsBufferPoolGetBuffer: Allocating new buffer: %lu", size);
	if (ctx->create) buffer = ctx->create(ctx, size);
	else {
	    buffer = dsMalloc(ctx->malloc, size + sizeof(struct DSBufferPoolBufferT));
	    if (buffer) buffer->buffer_size = size;
	}
	i=1;
    }
    if (buffer) ctx->allocated[j]++;
    dsLockReturn(ctx->lock);
    
    if (!i) {
	if (ctx->restore) ctx->restore(ctx,buffer);
    }

    return buffer;
}

void dsBufferPoolReturnBuffer(DSBufferPoolContext ctx, DSBufferPoolBuffer buffer) {
    unsigned long size;
    unsigned long int i;
    unsigned char j;

    ds_assert_return(ctx);
    ds_assert_return(buffer);
    
    size = buffer->buffer_size;
    
    if (size == ctx->default_size) j = ctx->default_bin;
    else {
	for (j=0, i=(size>>ctx->lsr);i;i>>=1) j++;

	if (j>=ctx->bins) {
	    ds_assert_return(ctx->allocated[ctx->bins]>0);

	    dsDebugServer(DS_DEBUG_NOTICE, "dsBufferPoolReturnBuffer: Destroying BigBuffer: %lu", size);

	    if (ctx->destroy) ctx->destroy(ctx, buffer);
	    else dsFree(ctx->malloc, buffer);

	    dsLockWrite(ctx->lock);
	    ctx->allocated[ctx->bins]--;
	    dsLockReturn(ctx->lock);
	    return;
	}
    }

    dsDebugServer(DS_DEBUG_DEBUG, "dsBufferPoolReturnBuffer: ReturnBuffer: %lu", size);


    dsLockWrite(ctx->lock);
    ds_assert_info(ctx->allocated[j]>0);
    
    if (ctx->allocated[j]>0) {
	if (ctx->last[j]) {
	    ctx->last[j]->next = buffer;
	    ctx->last[j] = buffer;
	} else {
	    ctx->first[j] = buffer;
	    ctx->last[j] = buffer;
	}    
	buffer->next = NULL;

	ctx->allocated[j]--;
    }
    
    dsLockReturn(ctx->lock);

}


int dsBufferPoolResizeBuffer(DSBufferPoolContext ctx, DSBufferPoolBuffer *buffer, unsigned long size) {
    int err;
    void *buff;
    unsigned long old_size;
    unsigned long int i;
    unsigned char j, old_j;

    ds_assert_param(ctx);
    ds_assert_param(buffer);

    old_size = (*buffer)->buffer_size;
    for (old_j=0, i=(old_size>>ctx->lsr);i;i>>=1) old_j++;


    for (j=0,i=size>>ctx->lsr;i;i>>=1) j++;
    i = (1<<(j+ctx->lsr-1));
    if (i<size) {
	size=i<<1;
	j++;
    } else size=i;

    dsDebugServer(DS_DEBUG_DEBUG, "dsBufferPoolGetBuffer: Resizing Buffer: %p. Old Size: %lu, New Size: %lu", buffer, (*buffer)->buffer_size, size);

    if (ctx->resize) {
	err = ctx->resize(ctx,buffer,size);
	if (err) return err;
    } else {
	buff = dsRealloc(ctx->malloc, *(void**)buffer, size + sizeof(struct DSBufferPoolBufferT));
	if (!buff) return DS_ERR_MEM;
	((DSBufferPoolBuffer)buff)->buffer_size = size;
	*(void**)buffer=buff;
    }
    
    if (j>ctx->bins) {
	j = ctx->bins;
	if (old_j>ctx->bins) old_j = ctx->bins;
    }

    dsLockWrite(ctx->lock);
    ds_assert_param(ctx->allocated[old_j]>0);
    ctx->allocated[old_j]--;
    ctx->allocated[j]++;
    dsLockReturn(ctx->lock);

    return 0;
}

int dsBufferPoolReplaceBuffer(DSBufferPoolContext ctx, DSBufferPoolBuffer *buffer, unsigned long size) {
    int err;
    void *buff;
    unsigned long old_size;
    unsigned long int i;
    unsigned char j, old_j;

    ds_assert_param(ctx);
    ds_assert_param(buffer);

    old_size = (*buffer)->buffer_size;
    for (old_j=0, i=(old_size>>ctx->lsr);i;i>>=1) old_j++;

    for (j=0,i=size>>ctx->lsr;i;i>>=1) j++;
    i = (1<<(j+ctx->lsr-1));

    if (i<size) {
	size=i<<1;
	j++;
    } else size=i;

    dsDebugServer(DS_DEBUG_DEBUG, "dsBufferPoolGetBuffer: Replacing Buffer: %p. Old Size: %lu, New Size: %lu", buffer, (*buffer)->buffer_size, size);

    if (ctx->replace) {
	err = ctx->replace(ctx, buffer, size);
	if (err) return err;
    } else {
	buff = (char*)dsMalloc(ctx->malloc, size + sizeof(struct DSBufferPoolBufferT));
	if (!buff) return DS_ERR_MEM;
	((DSBufferPoolBuffer)buff)->buffer_size = size;
	dsFree(ctx->malloc, *(void**)buffer);
	*(void**)buffer = buff;
    }

    if (j>ctx->bins) {
	j = ctx->bins;
	if (old_j>ctx->bins) old_j = ctx->bins;
    }
    dsLockWrite(ctx->lock);
    ds_assert_info(ctx->allocated[old_j]>0);
    ctx->allocated[old_j]--;
    ctx->allocated[j]++;
    dsLockReturn(ctx->lock);
    return 0;
}


void dsBufferPoolMantain(DSBufferPoolContext ctx) {
    // ToDo: Free unnececary buffers
}
