#define _DS_DATA_BUFFER_C
#include <string.h>

#include <dsassert.h>
#include <dslib.h>
#include <dsdebug.h>

#include <dsbinaryconfiguration.h>
#include <dsdatabuffer.h>
#include <dsbufferpoolsimplebuffer.h>

#define HEADER_SIZE (sizeof(void*)+sizeof(unsigned long))
#define ALIGN(pos,alignment) \
    if (pos&(alignment-1)) pos = (pos&~(alignment-1))+alignment;

#define DS_DATA_BUFFER_APPROXIMATION_INTERVAL 20
#define DS_DATA_BUFFER_APPROXIMATION_INTERVAL_05 10
#define DS_DATA_BUFFER_APPROXIMATION_INTERVAL_15 30
#define DS_DATA_BUFFER_APPROXIMATION_RISE 2

DSAllocator dsAllocatorDataBuffer() {
    return &ds_allocator_databuffer;
}

DSDataBufferOptions dsDataBufferGetDefaultOptions() {
    int err;
    DSDataBufferOptions options;

    options = (DSDataBufferOptions)malloc(sizeof(struct DSDataBufferOptionsT));
    if (!options) return options;
    
    err = dsDataBufferInitDefaultOptions(options);
    if (err) {
	dsDataBufferDestroyOptions(options);
	return NULL;
    }
    
    return options;
}

int dsDataBufferInitDefaultOptions(DSDataBufferOptions options) {
    ds_assert_param(options);
        
    if (!options) return DS_ERR_INVALID_ARGUMENT;
    
    memcpy(options, &ds_databuffer_options, sizeof(struct DSDataBufferOptionsT));
    return 0;
}

void dsDataBufferDestroyOptions(DSDataBufferOptions options) {
    if (options) free(options);
}

int dsDataBufferOptionsSetSize(DSDataBufferOptions options, unsigned long minimal_size, unsigned long maximal_size) {
    ds_assert_param(options);
    
    options->minimal_size = minimal_size;
    options->maximal_size = maximal_size;
    return 0;
}

int dsDataBufferOptionsSetBlockSize(DSDataBufferOptions options, unsigned long block_size) {
    ds_assert_param(options);

    options->block_size = block_size;
        
    return 0;
}

int dsDataBufferOptionsSetAllocationSize(DSDataBufferOptions options, unsigned long fixed_allocation_size) {
    ds_assert_param(options);
    
    options->fixed_allocation_size = fixed_allocation_size;
    
    return 0;
}

int dsDataBufferOptionsSetBufferPool(DSDataBufferOptions options, DSBufferPoolContext pool) {
    ds_assert_param(options);
    
    if (dsBufferPoolCheckSimpleBufferFunctions(pool)) return DS_ERR_DATABUFFER_OPTIONS_INVALID_BUFFERPOOL;
    options->pool = pool;
    
    return 0;
}


void dsDataBufferZeroContext(DSDataBufferContext ctx) {
    ds_assert_return(ctx);
    ctx->first_chunk = NULL;
    ctx->buffers = NULL;
    ctx->current = NULL;
    ctx->pool = NULL;
    
    ctx->n = 0;
    ctx->max1 = 0;
    ctx->max2 = 0;
    ctx->save = 0;
    ctx->maxfill = 0;
    ctx->allocation_size = 0;
    ctx->preallocated = 0;
}

DSDataBufferContext dsDataBufferCreateContext(DSMallocContext mctx) {
    DSDataBufferContext ctx;

    ctx = (DSDataBufferContext)dsMalloc(mctx,sizeof(struct DSDataBufferContextT));
    if (ctx) {
	ctx->base.malloc=mctx;
	/*ctx->base.allocator=&ds_allocator_databuffer;*/
	memcpy(&ctx->base.allocator, &ds_allocator_databuffer, sizeof(struct DSAllocatorT));
	dsDataBufferZeroContext(ctx);
    }
    return ctx;
}

int dsDataBufferInitContext(DSDataBufferContext ctx, DSDataBufferOptions options) {
    unsigned long i, size;
    DSBufferContext buffer;
    
    ds_assert_param(ctx);

    if (!options) return DS_ERR_DATABUFFER_OPTIONS;
    if (!options->pool) return DS_ERR_DATABUFFER_OPTIONS_MISSING_BUFFERPOOL;

    dsDataBufferZeroContext(ctx);

    ctx->pool = options->pool;

    size = options->minimal_size;
    for (i=1;i<size;i<<=1);
    ctx->minimal_size = i;

    size = options->maximal_size;
    if (size) {
	for (;i<size;i<<=1);
	ctx->maximal_size = i;
    } else ctx->maximal_size = 0;
    
    if (options->block_size) {
        size = options->block_size;
	for (i=1;i<size;i<<=1);
	ctx->block_size=i;
    } else ctx->block_size=1>>(options->pool->lsr-2); /* lsr>3 */
    
    if (options->fixed_allocation_size) { 
/*	if (options->fixed_allocation_size<ctx->block_size) ctx->fixed_allocation_size=ctx->block_size;
	else {
    	    size = options->fixed_allocation_size;
	    for (i=1;i<size;i<<=1);
	    ctx->fixed_allocation_size=i;
	}*/
	ctx->fixed_allocation_size = options->fixed_allocation_size;
    }
    else ctx->fixed_allocation_size = 0;
    
    /* ctx->fixed_allocation_size = 0; For benchmarking DS */

    buffer = (DSBufferContext)dsBufferPoolGetBuffer(options->pool, ctx->minimal_size /* 0 is OK */);
    if (!buffer) return DS_ERR_BUFFERPOOL_GET;

    ctx->current = buffer;
    ctx->buffers = buffer;
    
    return 0;
}

void dsDataBufferFreeContext(DSDataBufferContext ctx) {
    DSBufferContext buffer, nbuffer;

    ds_assert_return(ctx);

    for (buffer=ctx->buffers;buffer;buffer=nbuffer) {
	nbuffer = buffer->next;
	dsBufferPoolReturnBuffer(ctx->pool, (DSBufferPoolBuffer)buffer);
    }
    dsDataBufferZeroContext(ctx);    
}

void dsDataBufferDestroyContext(DSDataBufferContext ctx) {
    dsDataBufferFreeContext(ctx);
    dsFree(ctx->base.malloc,ctx);
}

DSBufferContext dsDataBufferClear(DSDataBufferContext ctx, DSBufferContext buffer) {
    DSBufferContext nbuffer;
    unsigned char destroy = 0;
    
    if ((ctx->buffers == buffer)&&(buffer->buffer_start == buffer->buffer_end)) destroy = 1;
    
    /* DS. Really destroy everything up to last buffer? */
    for (nbuffer=buffer->next;nbuffer;nbuffer=buffer->next) {
	if (destroy) dsBufferPoolReturnBuffer(ctx->pool, (DSBufferPoolBuffer)buffer);
	buffer = nbuffer;
    }
    if (destroy) ctx->buffers = buffer;
    return buffer;
}

int dsDataBufferStart(DSDataBufferContext ctx) {
    ds_assert_param(ctx);
    
    if (ctx->fixed_allocation_size) {
	if (ctx->preallocated > ctx->fixed_allocation_size) return 0;
	ctx->preallocated = dsBufferReserve(ctx->current, ctx->maximal_size, ctx->fixed_allocation_size);
    } else {    
	if (ctx->n<DS_DATA_BUFFER_APPROXIMATION_INTERVAL) {
    	    if (ctx->preallocated > ctx->max1) return 0;
	    if (ctx->max1>ctx->block_size) ctx->preallocated = dsBufferReserve(ctx->current, ctx->maximal_size, ctx->max1);
	    else ctx->preallocated = dsBufferReserve(ctx->current, ctx->maximal_size, ctx->block_size);
	} else {
	    if (ctx->preallocated > ctx->max2) return 0;
	    if (ctx->max2>ctx->block_size) ctx->preallocated = dsBufferReserve(ctx->current, ctx->maximal_size, ctx->max2);
	    else ctx->preallocated = dsBufferReserve(ctx->current, ctx->maximal_size, ctx->block_size);
	}
    }

    if (ctx->current->next) ctx->current=dsDataBufferClear(ctx, ctx->current);
//    for (buffer=ctx->current;buffer->next;buffer=buffer->next);
//    ctx->current=buffer;
    
    return 0;
}


unsigned long dsDataBufferReserve(DSDataBufferContext ctx, DSAllocation notrequired, unsigned long size) {
//    DSBufferContext buffer;
    
    dsDebugServer(DS_DEBUG_WARNING, "Reserve %p: %lu %lu\n", ctx, size, ctx->preallocated);
    if (ctx->preallocated < size) {
	if (size>ctx->block_size) ctx->preallocated = dsBufferReserve(ctx->current, ctx->maximal_size, size);
	else ctx->preallocated = dsBufferReserve(ctx->current, ctx->maximal_size, ctx->block_size);

	if (ctx->current->next) ctx->current=dsDataBufferClear(ctx, ctx->current);
//	for (buffer=ctx->current;buffer->next;buffer=buffer->next);
//	ctx->current=buffer;
    } 
    dsDebugServer(DS_DEBUG_WARNING, "Reserve %p: %lu %lu.\n", ctx, size, ctx->preallocated);
    return ctx->preallocated;
}

void dsDataBufferTrim(DSDataBufferContext ctx, DSAllocation notrequired, unsigned long size) {
    dsBufferTrim(ctx->current, size);
}


void *dsDataBufferAllocate(DSDataBufferContext ctx, DSAllocation notrequired, unsigned long size, unsigned char alignment) {
    DSBufferContext buffer;
    unsigned long preallocate;

    ds_assert_null(ctx);
    buffer = ctx->current;
    ds_assert_null(buffer);

    preallocate = size + ds_binary_configuration->align_allocation + HEADER_SIZE;
    if (preallocate>ctx->preallocated) {
	if (preallocate>ctx->block_size) ctx->preallocated = dsBufferReserve(buffer, ctx->maximal_size, preallocate);
	else ctx->preallocated = dsBufferReserve(buffer, ctx->maximal_size, ctx->block_size);
	if (!ctx->preallocated) return NULL;
	
	if (buffer->next) {
	    buffer=dsDataBufferClear(ctx, buffer);
	    ctx->current=buffer;
	}
    }

    return dsBufferFastAllocate(buffer, size, alignment, &ctx->preallocated);
}

void dsDataBufferFree(DSDataBufferContext ctx, DSAllocation buffer) {
    ds_assert_return(ctx);
    ds_assert_return(buffer);
    
    ctx->buffers = dsBufferFree(ctx->buffers, (void*)buffer);
}

void *dsDataBufferAcceptChunk(DSDataBufferContext ctx, DSAllocation notrequired, unsigned long *req_size) {
    void *buffer;
    unsigned long size;

    ds_assert_null(ctx);
    
    if (ctx->fixed_allocation_size) {
	buffer=dsBufferChunkAccept(ctx->current, req_size);
    } else {
	if (!req_size) req_size = &size;
	buffer = (void*)dsBufferChunkAccept(ctx->current, req_size);
	if (buffer) ctx->allocation_size += *req_size;
    }

    if (!ctx->first_chunk) {
	ctx->first_chunk = ctx->current;
	ctx->allocation = (DSAllocation)buffer;
    }
    
    return buffer;
}

void dsDataBufferCancelChunk(DSDataBufferContext ctx, DSAllocation notrequired) {
    ds_assert_return(ctx);
    dsBufferChunkCancel(ctx->current);
    /* Ignoring rise of preallocated space */
}

void *dsDataBufferSimpleChunk(DSDataBufferContext ctx, DSAllocation notrequired, unsigned long size) {
    DSBufferContext buffer;
    unsigned long preallocate;

    ds_assert_null(ctx);
    
    buffer = ctx->current;
    
    preallocate = size + ds_binary_configuration->align_allocation + HEADER_SIZE;
    if (preallocate>ctx->preallocated) {
	if (preallocate>ctx->block_size) ctx->preallocated = dsBufferReserve(buffer, ctx->maximal_size, preallocate);
	else ctx->preallocated = dsBufferReserve(buffer, ctx->maximal_size, ctx->block_size);
	if (!ctx->preallocated) return NULL;	
        if (buffer->next) {
	    buffer=dsDataBufferClear(ctx, ctx->current);
	    ctx->current = buffer;
	}
/*	for (;buffer->next;buffer=buffer->next);
	ctx->current=buffer;*/
    }
    
    if (!ctx->fixed_allocation_size) ctx->allocation_size += size;
    
    return dsBufferChunkSimple(buffer, size, &ctx->preallocated);
}

DSAllocation dsDataBufferSimpleAllocate(DSDataBufferContext ctx, unsigned long size) {
    DSBufferContext buffer;
    void *buff;
    unsigned long used_space;
    unsigned long preallocate;
    
    ds_assert_null(ctx);
    ds_assert_null(!ctx->first_chunk);

    buffer = ctx->current;

    preallocate = size + ds_binary_configuration->align_allocation + HEADER_SIZE;
//    dsDebugServer(DS_DEBUG_WARNING, "SimpleAllocate %p: %lu %lu %lu (%u %u)", ctx, size, preallocate, ctx->preallocated, ds_binary_configuration->align_allocation, HEADER_SIZE);
    if (preallocate>ctx->preallocated) {
	if (preallocate>ctx->block_size) ctx->preallocated = dsBufferReserve(buffer, ctx->maximal_size,  preallocate);
	else ctx->preallocated = dsBufferReserve(buffer, ctx->maximal_size, ctx->block_size);
	if (!ctx->preallocated) return NULL;	
        if (buffer->next) {
	    buffer=dsDataBufferClear(ctx, ctx->current);
	    ctx->current = buffer;
	}
//	dsDebugServer(DS_DEBUG_WARNING, "SimpleAllocate %p: %lu %lu %lu.", ctx, size, preallocate, ctx->preallocated);
/*
	for (;buffer->next;buffer=buffer->next);
	ctx->current=buffer;*/
    }

    if (ctx->buffers == buffer) {
	buff=dsBufferSimpleAllocate(buffer, size, &ctx->preallocated, &used_space);
	if (used_space > ctx->maxfill) ctx->maxfill = used_space;
    } else {
	buff=dsBufferSimpleAllocate(buffer, size, &ctx->preallocated, NULL);
	ctx->maxfill = buffer->buffer_size;
    }

/*
    In case of single call to this function instead of common way, this data can be
    considerate non-standard, and therefore shouldn't be used in approximation.
    
    if (!ctx->fixed_allocation_size) {
	if ((++(ctx->n))==DS_DATA_BUFFER_APPROXIMATION_INTERVAL) ctx->max1 = 0;
	else if (ctx->n==DS_DATA_BUFFER_APPROXIMATION_INTERVAL_15) {
	    ctx->max2 = 0;
	    ctx->n = DS_DATA_BUFFER_APPROXIMATION_INTERVAL_05;
	}
        
	used_space = ctx->save;

	if ((size>DS_DATA_BUFFER_APPROXIMATION_RISE*ctx->max1)&&(size>DS_DATA_BUFFER_APPROXIMATION_RISE*ctx->max2)) {
	    if (used_space) {
		if (size>ctx->max1) ctx->max1 = size;
	        if (size>ctx->max2) ctx->max2 = size;
		
		if (used_space>ctx->max1) ctx->max1 = used_space;
	        if (used_space>ctx->max2) ctx->max2 = used_space;
		
		ctx->save = 0;
	    } else ctx->save = size;
	} else {
	    if (size>ctx->max1) ctx->max1=size;
	    if (size>ctx->max2) ctx->max2=size;

	    if (used_space) {
		if (used_space<DS_DATA_BUFFER_APPROXIMATION_RISE*size) {
		    if (used_space>ctx->max1) ctx->max1 = used_space;
		    if (used_space>ctx->max2) ctx->max2 = used_space;
	        }
		ctx->save=0;
	    }
	}
    }
*/

/*
     Preparing for next round.
    ToDo: Find better place? (Manager?)
    
    Commented just for same case as above, - in case of good place to allocate memory
    is found, - enable.
    dsDataBufferStart(ctx);
*/

    return (DSAllocation)buff;
}

DSAllocation dsDataBufferAccept(DSDataBufferContext ctx, DSAllocation notrequired) {
    unsigned long size;
    unsigned long used_space;


    if (ctx->buffers == ctx->current) {
	if (dsBufferAccept(ctx->current, &used_space)) return NULL;
	if (used_space > ctx->maxfill) ctx->maxfill = used_space;
    } else {
	if (dsBufferAccept(ctx->current, NULL)) return NULL;
	ctx->maxfill = ctx->current->buffer_size;
    }
    
    if (!ctx->fixed_allocation_size) {
	if ((++(ctx->n))==DS_DATA_BUFFER_APPROXIMATION_INTERVAL) ctx->max1 = 0;
	else if (ctx->n==DS_DATA_BUFFER_APPROXIMATION_INTERVAL_15) {
	    ctx->max2 = 0;
	    ctx->n = DS_DATA_BUFFER_APPROXIMATION_INTERVAL_05;
	}
        
	size = ctx->allocation_size;
	used_space = ctx->save;

	if ((size>DS_DATA_BUFFER_APPROXIMATION_RISE*ctx->max1)&&(size>DS_DATA_BUFFER_APPROXIMATION_RISE*ctx->max2)) {
	    if (used_space) {
		if (size>ctx->max1) ctx->max1 = size;
	        if (size>ctx->max2) ctx->max2 = size;
		
		if (used_space>ctx->max1) ctx->max1 = used_space;
	        if (used_space>ctx->max2) ctx->max2 = used_space;
		
		ctx->save = 0;
	    } else ctx->save = size;
	} else {
	    if (size>ctx->max1) ctx->max1=size;
	    if (size>ctx->max2) ctx->max2=size;

	    if (used_space) {
		if (used_space<DS_DATA_BUFFER_APPROXIMATION_RISE*size) {
		    if (used_space>ctx->max1) ctx->max1 = used_space;
		    if (used_space>ctx->max2) ctx->max2 = used_space;
	        }
		ctx->save=0;
	    }
	}
	ctx->allocation_size = 0;
    }

    ctx->first_chunk = NULL;
    
    /* Preparing for next round.
    ToDo: Find better place? (Manager?) */
    dsDataBufferStart(ctx);
    
    return ctx->allocation;
}

void dsDataBufferCancel(DSDataBufferContext ctx, DSAllocation notrequired) {
    DSBufferContext new_first, first, chunk;
    
    ds_assert_return(ctx);
    
    first=ctx->first_chunk;
    new_first=dsBufferCancel(first);
    if (new_first!=first) {
	if (ctx->buffers==first) ctx->buffers=new_first;
	else {
	    for (chunk=ctx->buffers;chunk->next!=first;chunk=chunk->next);
	    chunk->next = new_first;
	}
    }
    
    ctx->first_chunk = NULL;
    ctx->preallocated = 0;

    /* Preparing for next round.
    ToDo: Find better place? (Manager?) */
    dsDataBufferStart(ctx);
}


int dsDataBufferCall(DSDataBufferContext ctx, DSAllocation allocation, DSSendFunction func, void *attr, unsigned char align) {
    int err;
    char *ptr, *next;
    unsigned long size;
    
    for (ptr=(void*)allocation;ptr;ptr=next) {
	size = *((unsigned long*)(ptr-HEADER_SIZE));
	next = *((char**)(ptr-sizeof(void*)));
	if ((next)&&(align)) ALIGN(size, ds_binary_configuration->align_allocation);
	err=func(attr,ptr,size);
	if (err) return err;
    }
    return 0;
}

int dsDataBufferSend(DSAllocation allocation, DSSendFunction func, void *attr) {
    int err;
    char *ptr, *next;
    unsigned long size;

    for (ptr=(void*)allocation;ptr;ptr=next) {
	size = *((unsigned long*)(ptr-HEADER_SIZE));
	next = *((char**)(ptr-sizeof(void*)));
	if (next) ALIGN(size, ds_binary_configuration->align_allocation);
	err=func(attr,ptr,size);
	if (err) return err;
    }
    return 0;
}

unsigned long dsDataBufferSizeof(DSDataBufferContext ctx, DSAllocation allocation, unsigned char align) {
    char *ptr = (char*)allocation;
    unsigned long res=0,size;
    
    while (ptr) {
	size = *((unsigned long*)(ptr-HEADER_SIZE));
	ptr = *((char**)(ptr-sizeof(void*)));
	if ((ptr)&&(align)) ALIGN(size, ds_binary_configuration->align_allocation);
	res+=size;
    }
    return res;
}

void* dsDataBufferGet(DSDataBufferContext ctx, DSAllocation allocation, unsigned long *size) {
    ds_assert_null(allocation);
    if (size) *size = *((unsigned long*)((char*)allocation-HEADER_SIZE));
    return (void*)allocation;
}

void* dsDataBufferGetNext(DSDataBufferContext ctx, DSAllocation allocation, void *current, unsigned long *size) {
    char *ptr;
    ds_assert_null(current);
    
    ptr=*((char**)((char*)current-sizeof(void*)));
    if ((ptr)&&(size)) *size=*((unsigned long*)(ptr-HEADER_SIZE));
    return (void*)ptr;
}

void dsDataBufferMantain(DSDataBufferContext ctx) {
    /* ToDo: resize buffers over maximal_size, and just for long time lowly filled
    maximal_size(soft), minimal_size */
    puts("dsDataBufferMantain");
}
