#include <dslib.h>

#include <dsmodule.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <dsbinaryformatnative.h>
#include <dsbinaryformatopcxml.h>

#include <dsdataconverter.h>


DSDataConverterContext dsDataConverterCreateContext(DSMallocContext mctx) {
    return dsNewModule(DataConverter, mctx);
}

int dsDataConverterInitContext(DSDataConverterContext ctx, size_t representations) {
    ds_assert_param(ctx);

    if (!representations) representations = 2;
    
    ctx->representation = (DSRepresentationInstance*)dsMalloc(DS_MODULE(ctx)->malloc, representations * sizeof(DSRepresentationInstance));
    if (!ctx->representation) return DS_ERR_MEM;
    
    ctx->allocated_representations = representations;

    dsModuleMarkInitialized(ctx);

    return 0;
}

void dsDataConverterFreeContext(DSDataConverterContext ctx) {
    DSDataChunkForm i;
    
    if (ctx->representation) {
	for (i=0;i<ctx->supported_representations;i++) {
	    if (ctx->representation[i]) {
		dsRepresentationDestroyInstance(ctx->representation[i]);
	    }
	}
	dsFree(DS_MODULE(ctx)->malloc, ctx->representation);
    }

    dsZeroModule(DataConverter, ctx);
}

void dsDataConverterDestroyContext(DSDataConverterContext ctx) {
    dsDestroyModule(DataConverter, ctx);
}


int dsDataConverterRegisterRepresentationInstance(DSDataConverterContext ctx, DSRepresentationInstance base, DSRepresentationInstance rep) {
    DSDataChunkForm repid;
    DSRepresentationInstance *representation;

    ds_module_assert(ctx);
    
    if (ctx->supported_representations == ctx->allocated_representations) {
	representation = (DSRepresentationInstance*)dsRealloc(DS_MODULE(ctx)->malloc, ctx->representation, 2 * ctx->allocated_representations * sizeof(DSRepresentationInstance));
	if (!representation) return DS_ERR_MEM;	
	
	ctx->representation = representation;
	ctx->allocated_representations *= 2;	
    } else representation = ctx->representation;
    
    repid = ctx->supported_representations++;

    rep->id = repid;
    rep->base = base;
    representation[repid] = rep;
    
    return 0;
}

DSRepresentationInstance dsDataConverterGetRepresentation(DSDataConverterContext ctx, DSDataChunkForm repid) {
    ds_module_assert_default(ctx, NULL);
    ds_assert_default(repid < ctx->supported_representations, NULL);
    
    return ctx->representation[repid];
}

DSRepresentationInstance dsDataConverterRegisterRepresentation(DSDataConverterContext ctx, DSRepresentationInstance base, DSRepresentation rep, DSRepresentationOptions options) {
    int err;
    DSRepresentationInstance instance;

    ds_module_assert_default(ctx, NULL);
    ds_assert_default(rep, NULL);

    instance = dsRepresentationCreateInstance(rep, DS_MODULE(ctx)->malloc, options);
    if (instance) {
	err = dsDataConverterRegisterRepresentationInstance(ctx, base, instance);
	if (err) dsRepresentationDestroyInstance(instance);
	else return instance;
    }
    
    return NULL;    
}

DSRepresentationInstance dsDataConverterRegisterBinaryRepresentation(DSDataConverterContext ctx, DSRepresentationInstance base, DSBinaryFormat format, DSBinaryType type, DSAllocatorContext allocator) {
    int err;
    DSRepresentationInstance instance;
    DSRepresentationBinaryOptionsS options;

    ds_module_assert_default(ctx, NULL);
    ds_assert_default(format, NULL);

    dsRepresentationBinaryInitDefaultOptions(&options, format, type);
    dsRepresentationBinaryOptionsSetAllocator(&options, allocator);

    return dsDataConverterRegisterRepresentation(ctx, base, dsRepresentationBinary(), (DSRepresentationOptions)&options);
}

DSRepresentationInstance dsDataConverterRegisterNativeRepresentation(DSDataConverterContext ctx, DSBinaryType type, DSAllocatorContext allocator) {
    ds_module_assert_default(ctx, NULL);

    return dsDataConverterRegisterBinaryRepresentation(ctx, NULL, dsBinaryFormatNative(), type, allocator);
}

DSRepresentationInstance dsDataConverterRegisterOPCXMLRepresentation(DSDataConverterContext ctx, DSRepresentationInstance base, DSBinaryType type) {
    ds_module_assert_default(ctx, NULL);
    
    return dsDataConverterRegisterBinaryRepresentation(ctx, base, dsBinaryFormatOPCXML(), type, NULL);
}


DSDataConverterContext dsDataConverterCreateOPCXML(DSMallocContext mctx, DSBinaryType type) {
    int err;
    
    DSDataConverterContext ctx;
    DSRepresentationInstance native;
    
    ctx = dsDataConverterCreateContext(mctx);
    if (!ctx) return NULL;
    
    err = dsDataConverterInitContext(ctx, 2);
    if (!err) {
	dsDataConverterDestroyContext(ctx);
	return NULL;
    }

    native = dsDataConverterRegisterNativeRepresentation(ctx, type, NULL);
    if (!native) {
	dsDataConverterDestroyContext(ctx);
	return NULL;
    }
    
    if (!dsDataConverterRegisterOPCXMLRepresentation(ctx, native, type)) {
	dsDataConverterDestroyContext(ctx);
	return NULL;
    }
    
    return ctx;    
}


/* Reduced paths optimization */
int dsDataConverterConvert(DSDataConverterContext ctx, DSRepresentationInstance rep, DSDataChunkContext chunk) {
    int err;
    DSRepresentationInstance base;
    DSDataConst data, basedata;
    DSDataChunkForm i;

    ds_assert_param(ctx);
    ds_assert_param(chunk);
    
    data = dsRepresentationGet(rep, chunk);
    if (data) return 0;
    
    base = rep->base;
    if (base) {
	basedata = dsRepresentationGet(base, chunk);
        if (!basedata) {
	    err = dsDataConverterConvert(ctx, base, chunk);
	    if (err) return err;
	}

	dsDebugServer(DS_DEBUG_DEBUG, "dsDataConverter: Convert: Rep%u reading from Rep%u", rep->id, base->id);
	err = dsRepresentationReadBase(rep, chunk);
	if (err) return err;
    } else {
	    // We do not need Root (base-less representation)
	for (i=chunk->supported_forms;i>0;i--) {
	    basedata = dsDataChunkGet(chunk, (DSDataChunkForm)i);
	    if (basedata) {
		base = ctx->representation[i];
	    
		dsDebugServer(DS_DEBUG_DEBUG, "dsDataConverter: Convert: Rep%u write to Rep%u",base->id, base->base->id);
		err = dsRepresentationWriteBase(base, chunk);

		    // My Base representation should have smaller ID.
		if (err) {
		    if ((err == DS_ERR_NOTSUPPORTED)||(err == DS_ERR_NOTIMPLEMENTED)) continue;
		    return err;
		} else {
		    if (base->base == rep) break;
		    /*We can't do that, since some path could be incomplete,
		    however it should not be
		    i = 1 + base->base->id;*/
		}
	    }
	}
	
	if (i<=0) return DS_ERR_EMPTY;
    }

    return 0;
}

int dsDataConverterConvertData(DSDataConverterContext ctx, DSRepresentationInstance to, DSData dst, DSRepresentationInstance from, DSDataConst src) {
    int err;
    
    DSDataChunkContext chunk;
    DSDataConst data;
    
    chunk = dsDataChunkCreateContext(DS_MODULE(ctx)->malloc);
    if (!chunk) return DS_ERR_MEM;
    
    
    err = dsDataChunkInitContext(chunk, NULL, ctx->allocated_representations);
    if (err) {
	dsDataChunkDestroyContext(chunk);
	return err;
    }

    err = dsDataChunkSet(chunk, from->id, src);

    if (!err) {
	err = dsDataConverterConvert(ctx, to, chunk);
	if (!err) {
	    data = dsDataChunkGet(chunk, to->id);
	    if (data) dsDataCopy(dst, src);
	}
    }

    dsDataChunkSet(chunk, from->id, NULL);
    dsDataChunkSet(chunk, to->id, NULL);
    dsDataChunkDestroyContext(chunk);
    
    return err;
}

int dsDataConverterConvertBinaryData(DSDataConverterContext ctx, DSRepresentationInstance to, DSData dst, DSRepresentationInstance from, DSDataConst src, DSBinaryType type) {

}
