#include <dslib.h>

#include <dsassert.h>
#include <dsdebug.h>
#include <dsallocatorsimple.h>

#include <dsbinaryformatnative.h>
#include <dsbinaryformatopcxml.h>

#include <dsrepresentationbinary.h>

static const struct DSRepresentationBinaryOptionsT ds_representation_binary_options = {
    { },
    NULL, 	/* allocator */
    NULL, 	/* format */
    NULL 	/* type */
};

DSRepresentationBinaryOptions dsRepresentationBinaryGetDefaultOptions(DSBinaryFormat format, DSBinaryType type) {
    int err;
    DSRepresentationBinaryOptions options;

    options = (DSRepresentationBinaryOptions)malloc(sizeof(struct DSRepresentationBinaryOptionsT));
    if (!options) return options;
    
    err = dsRepresentationBinaryInitDefaultOptions(options, format, type);
    if (err) {
	dsRepresentationBinaryDestroyOptions(options);
	return NULL;
    }
    return options;
}

int dsRepresentationBinaryInitDefaultOptions(DSRepresentationBinaryOptions options, DSBinaryFormat format, DSBinaryType type) {
    ds_assert_param(options);

    memcpy(options, &ds_representation_binary_options, sizeof(struct DSRepresentationBinaryOptionsT));
    dsRepresentationInitDefaultOptions((DSRepresentationOptions)options);
    options->format = format;
    options->type = type;
    
    return 0;
}

void dsRepresentationBinaryDestroyOptions(DSRepresentationBinaryOptions options) {
    if (options) free(options);
}

int dsRepresentationBinaryOptionsSetAllocator(DSRepresentationBinaryOptions options, DSAllocatorContext allocator) {
    ds_assert_param(options);
    
    options->allocator = allocator;
    return 0;
}

int dsRepresentationBinaryOptionsSetFormatOptions(DSRepresentationBinaryOptions options, DSBinaryFormatInstanceOptions format_options) {
    ds_assert_param(options);
    
    options->format_options = format_options;
    return 0;
}

static void dsRepresentationBinaryDestroyContext(DSMallocContext mctx, DSRepresentationBinaryContext ctx) {
    ds_assert(!ctx->grabed_allocator);

    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    if (ctx->allocator_lock) dsLockDestroyContext(ctx->allocator_lock);
    if (ctx->format) dsBinaryFormatDestroyInstance(ctx->format);
    if (ctx->allocator) dsAllocatorDestroyContext(ctx->allocator);
    dsFree(mctx, ctx);
}

static DSRepresentationInstanceData dsRepresentationBinaryCreateContext(DSMallocContext mctx, DSRepresentationBinaryOptions options) {
    int err = 0;
    
    DSRepresentationBinaryContext ctx;

    ds_assert_null(options);
    ds_assert_null(options->format);

    ctx = dsNewClass(mctx, DSRepresentationBinaryContext);
    if (!ctx) return NULL;

    if (options->allocator) ctx->allocator = options->allocator;
    else if (options->format->allocator != DS_BINARY_FORMAT_ALLOCATOR_TYPE_ABSENT) {
	ctx->allocator = (DSAllocatorContext)dsAllocatorSimpleCreateContext(mctx);
	if (ctx->allocator) err = dsAllocatorSimpleInitContext((DSAllocatorSimpleContext)ctx->allocator, NULL);
	else err = DS_ERR_MEM;
	
	if (err) {
	    dsRepresentationBinaryDestroyContext(mctx, ctx);
	    return NULL;
	}
    } else ctx->allocator = NULL;
    
    ctx->format = dsBinaryFormatCreateInstance(options->format, mctx, options->format_options);
    if (!ctx->format) {
	dsRepresentationBinaryDestroyContext(mctx, ctx);
	return NULL;
    }
    
    ctx->type = options->type;
    
    ctx->lock = dsLockCreateContext(mctx);
    if ((!ctx->lock)) {
	dsRepresentationBinaryDestroyContext(mctx, ctx);
	return NULL;
    }

    err = dsLockInitContext(ctx->lock);
    if (err) {
	dsRepresentationBinaryDestroyContext(mctx, ctx);
	return NULL;
    }

    if (ctx->allocator) {
	ctx->allocator_lock = dsLockCreateContext(mctx);
	if (!ctx->allocator_lock) {
	    dsRepresentationBinaryDestroyContext(mctx, ctx);
	    return NULL;
	}
	
	err = dsLockInitContext(ctx->allocator_lock);
	if (err) {
	    dsRepresentationBinaryDestroyContext(mctx, ctx);
	    return NULL;
	}
    }

    return (DSRepresentationInstanceData)ctx;
}

int dsRepresentationBinaryMantainAllocator(DSRepresentationInstance instance) {
    DSRepresentationBinaryContext ctx;

    ds_assert_param(instance);
    
    ctx = (DSRepresentationBinaryContext)instance->data;
    if (ctx->allocator) dsAllocatorMantain(ctx->allocator);
    return 0;
}

int dsRepresentationBinaryPushData(DSRepresentationInstance instance, DSDataChunkContext chunk, DSDataConst data) {
    int err = 0;
    DSData chunk_data;

    DSRepresentationBinaryContext ctx;

    ds_assert_param(instance);
    ds_assert_param(chunk);
    ds_assert_param(data);

    ctx = (DSRepresentationBinaryContext)instance->data;
    ds_assert_param(ctx->type);

    chunk_data = dsDataChunkGetData(chunk, instance->id);
    ds_assert_err(chunk_data, DS_ERR_INTERNAL);
    if (chunk_data->data) return DS_ERR_INITIALIZED;

    dsDataChunkLock(chunk);
    if (ctx->allocator_lock) dsLockWrite(ctx->allocator_lock);
    err = dsBinaryFormatImport(ctx->format, ctx->type, ctx->allocator, chunk_data, data);
    if (ctx->allocator_lock) dsLockReturn(ctx->allocator_lock);
    dsDataChunkUnlock(chunk);

    return err;
}

int dsRepresentationBinaryReadBase(DSRepresentationInstance instance, DSDataChunkContext chunk) {
    int err;

    DSRepresentationBinaryContext ctx;
    DSDataConst basedata;
    DSData data;

    ds_assert_param(instance);
    ds_assert_param(instance->base);
    ds_assert_param(chunk);

    ctx = (DSRepresentationBinaryContext)instance->data;
    ds_assert_param(ctx->type);

    data = dsDataChunkGetData(chunk, instance->id);
    ds_assert_err(data, DS_ERR_INTERNAL);
    if (data->data) return DS_ERR_INITIALIZED;

    basedata = dsDataChunkGet(chunk, instance->base->id);
    if (!basedata) return DS_ERR_EMPTY;
	
    dsDataChunkLock(chunk);
    if (ctx->allocator_lock) dsLockWrite(ctx->allocator_lock);
    err = dsBinaryFormatReadFromBase(ctx->format, ctx->type, ctx->allocator, data, basedata);
    if (ctx->allocator_lock) dsLockReturn(ctx->allocator_lock);
    dsDataChunkUnlock(chunk);
    
    return err;
}

int dsRepresentationBinaryWriteBase(DSRepresentationInstance instance, DSDataChunkContext chunk) {
    int err;

    DSAllocatorContext allocator;
    DSRepresentationBinaryContext ctx;
    DSData basedata;
    DSDataConst data;

    ds_assert_param(instance);
    ds_assert_param(instance->base);
    ds_assert_param(chunk);

    ctx = (DSRepresentationBinaryContext)instance->data;
    ds_assert_param(ctx->type);

    data = dsDataChunkGet(chunk, instance->id);
    if (!data) return DS_ERR_EMPTY;

    basedata = dsDataChunkGetData(chunk, instance->base->id);
    ds_assert_err(basedata, DS_ERR_INTERNAL);
    if (basedata->data) return DS_ERR_INITIALIZED;
    
    allocator = ((DSRepresentationBinaryContext)instance->base->data)->allocator;
	
    dsDataChunkLock(chunk);
    if (ctx->allocator_lock) dsLockWrite(ctx->allocator_lock);
    err = dsBinaryFormatWriteToBase(ctx->format, ctx->type, allocator, data, basedata);
    if (ctx->allocator_lock) dsLockReturn(ctx->allocator_lock);
    dsDataChunkUnlock(chunk);
    
    return err;
}

static int dsRepresentationBinarySetupHandler(DSRepresentationInstance instance, DSDataHandler handler);

static DSRepresentationS ds_representation_binary = {
    (DSRepresentationCreateFunction)dsRepresentationBinaryCreateContext,
    (DSRepresentationDestroyFunction)dsRepresentationBinaryDestroyContext,
    dsRepresentationBinarySetupHandler,
    dsRepresentationBinaryMantainAllocator,
    dsRepresentationBinaryPushData,
    dsRepresentationBinaryReadBase,
    dsRepresentationBinaryWriteBase
};

DSRepresentation dsRepresentationBinary() {
    return &ds_representation_binary;
}

DSAllocatorContext dsRepresentationBinaryGetAllocator(DSRepresentationInstance instance) {
    DSRepresentationBinaryContext ctx;
    
    ds_assert_null(instance);
    ds_assert_null(instance->representation == &ds_representation_binary);

    ctx = (DSRepresentationBinaryContext)instance->data;
    if (ctx->allocator) {
	if (ctx->allocator_lock) dsLockWrite(ctx->allocator_lock);
	ctx->grabed_allocator = 1;
	return ctx->allocator;
    }
    return NULL;
}

int dsRepresentationBinaryReturnAllocator(DSRepresentationInstance instance, DSAllocatorContext allocator) {
    DSRepresentationBinaryContext ctx;
    
    ds_assert_param(instance);
    ds_assert_param(instance->representation == &ds_representation_binary);
    ds_assert_err(allocator == ctx->allocator, DS_ERR_ALLOCATOR_INVALID);

    ctx = (DSRepresentationBinaryContext)instance->data;

    if (ctx->grabed_allocator>0) {
	ctx->grabed_allocator = 0;
        if (ctx->allocator_lock) dsLockReturn(ctx->allocator_lock);
	return 0;
    }
    
    return DS_ERR_ALLOCATOR_NOTOWNER;
}

DSRepresentationInstance dsRepresentationBinaryCreateNativeInstance(DSMallocContext mctx, DSBinaryType type, DSAllocatorContext allocator) {
    DSRepresentationBinaryOptionsS options;

    dsRepresentationBinaryInitDefaultOptions(&options, dsBinaryFormatNative(), type);
    if (allocator) dsRepresentationBinaryOptionsSetAllocator(&options, allocator);
    
    return dsRepresentationCreateInstance(&ds_representation_binary, mctx, (DSRepresentationOptions)&options);
}

DSRepresentationInstance dsRepresentationBinaryCreateOPCXMLInstance(DSMallocContext mctx, DSBinaryType type) {
    DSRepresentationBinaryOptionsS options;

    dsRepresentationBinaryInitDefaultOptions(&options, dsBinaryFormatOPCXML(), type);
    
    return dsRepresentationCreateInstance(&ds_representation_binary, mctx, (DSRepresentationOptions)&options);
}

int dsRepresentationBinaryPushAllocation(DSRepresentationInstance ctx, DSDataChunkContext chunk, DSAllocation allocation) {
    int err;
    DSDataS data;
    
    ds_assert_param(ctx);
    ds_assert_param(chunk);
    ds_assert_param(allocation);
    ds_assert_param(ctx->representation == &ds_representation_binary);
    
    err = dsDataSetAllocation(&data, ((DSRepresentationBinaryContext)ctx->data)->allocator, allocation, DS_DATA_SIZE_UNKNOWN);
    if (err) return err;
    
    return dsRepresentationPushData(ctx, chunk, &data);
}

static int dsRepresentationDataDestroy(DSData data) {
    int err;
    DSRepresentationInstance instance;
    DSRepresentationBinaryContext ctx;
    
    ds_assert_param(data);

    instance = (DSRepresentationInstance)data->handler->ctx;
    if (!instance) instance = (DSRepresentationInstance)data->owner;
    ds_assert_param(instance);
    ds_assert_param(instance->representation == &ds_representation_binary);

    ctx = (DSRepresentationBinaryContext)instance->data;
    ds_assert_param(ctx);
    
    err = dsBinaryFormatDestruct(ctx->format, ctx->type, data);
    if (err) return err;

    return 0;    
}

static int dsRepresentationDataCopy(DSDataCopier copier, DSData dst, DSDataConst src) {
    DSRepresentationInstance instance;
    DSRepresentationBinaryContext ctx;
    
    ds_assert_param(src);
    ds_assert_param(dst);
    ds_assert_param(copier);
    
    if (copier->type != DS_DATA_COPIER_TYPE_ALLOCATOR) return DS_ERR_NOTSUPPORTED;

    instance = (DSRepresentationInstance)src->handler->ctx;
    if (!instance) instance = (DSRepresentationInstance)src->owner;
    ds_assert_param(instance);
    ds_assert_param(instance->representation == &ds_representation_binary);

    ctx = (DSRepresentationBinaryContext)instance->data;
    ds_assert_param(ctx);
    
    return dsBinaryFormatCopy(ctx->format, ctx->type, (DSAllocatorContext)copier->ctx, dst, src);
}

static int dsRepresentationDataPipe(DSDataConst data, DSPipeFunction func, void *attr) {
    DSRepresentationInstance instance;
    DSRepresentationBinaryContext ctx;
    
    ds_assert_param(data);

    instance = (DSRepresentationInstance)data->handler->ctx;
    if (!instance) instance = (DSRepresentationInstance)data->owner;
    ds_assert_param(instance);
    ds_assert_param(instance->representation == &ds_representation_binary);

    ctx = (DSRepresentationBinaryContext)instance->data;
    ds_assert_param(ctx);
    
    return dsBinaryFormatPipe(ctx->format, ctx->type, data, func, attr);
}

static int dsRepresentationBinarySetupHandler(DSRepresentationInstance instance, DSDataHandler handler) {
    DSRepresentationBinaryContext ctx;

    ds_assert_param(instance);
    
    ctx = (DSRepresentationBinaryContext)instance->data;

    dsDataHandlerConfigureDefaultMethods(handler);
    dsDataHandlerSetContext(handler, instance);

    if (ctx->format->format->DestructData)
	dsDataHandlerSetDestructorFunction(handler, dsRepresentationDataDestroy);
    if (ctx->format->format->CopyData)
	dsDataHandlerSetCopyFunction(handler, dsRepresentationDataCopy);
    if (ctx->format->format->PipeData)
	dsDataHandlerSetPipeFunction(handler, dsRepresentationDataPipe);

    return 0;
}
