#include <dslib.h>
#include <dsdata.h>

#include <dsassert.h>
#include <dsdebug.h>
#include <dsxml.h>

#include <dsdatahandler.h>

int dsDataSetupDefaultCopier(DSDataCopier copier) {
    ds_assert_param(copier);

    copier->type = DS_DATA_COPIER_TYPE_ANY;

    return 0;
}

int dsDataSetupAllocatorCopier(DSDataCopier copier, DSAllocatorContext allocator) {
    ds_assert_param(copier);
    ds_assert_param(allocator);
        
    copier->type = DS_DATA_COPIER_TYPE_ALLOCATOR;
    copier->ctx = (void*)allocator;

    return 0;
}

int dsDataSetupMallocCopier(DSDataCopier copier, DSMallocContext malloc) {
    ds_assert_param(copier);
        
    copier->type = DS_DATA_COPIER_TYPE_MALLOC;
    copier->ctx = (void*)malloc;

    return 0;
}

DSDataHandler dsDataHandlerCreate(DSMallocContext mctx) {
    return (DSDataHandler)dsMalloc(mctx, sizeof(DSDataHandlerS));
}

void dsDataHandlerDestroy(DSMallocContext mctx, DSDataHandler handler) {
    ds_assert_return(handler);
    dsFree(mctx, handler);
}

int dsDataHandlerConfigureDefaultMethods(DSDataHandler handler) {
    ds_assert_param(handler);
    
    memset(handler, 0, sizeof(DSDataHandlerS));
    return 0;
}

int dsDataHandlerCopy(DSDataHandler dst, DSDataHandler src) {
    ds_assert_param(dst);
    ds_assert_param(src);
    
    memcpy(dst, src, sizeof(DSDataHandlerS));

    return 0;
}

int dsDataHandlerSetType(DSDataHandler handler, DSDataHandlerType type) {
    ds_assert_param(handler);

    handler->type = type;
    return 0;
}

int dsDataHandlerSetUserType(DSDataHandler handler, DSDataHandlerUserType type) {
    ds_assert_param(handler);

    handler->user_type = type;
    return 0;
}

int dsDataHandlerSetContext(DSDataHandler handler, void *ctx) {
    ds_assert_param(handler);
    
    handler->ctx = ctx;
    return 0;
}

int dsDataHandlerSetCopyFunction(DSDataHandler handler, DSDataCopyFunction func) {
    ds_assert_param(handler);
    
    handler->copy = func;
    return 0;
}

int dsDataHandlerSetPipeFunction(DSDataHandler handler, DSDataPipeFunction func) {
    ds_assert_param(handler);
    
    handler->pipe = func;
    return 0;
}

int dsDataHandlerSetDestructorFunction(DSDataHandler handler, DSDataDestructorFunction func) {
    ds_assert_param(handler);
    
    handler->destructor = func;
    return 0;
}


int dsDataDestroyContent(DSData data) {
    int err;
    ds_assert_param(data);
    
    if (!data->data) return 0;    
    if ((data->handler)&&(data->handler->destructor)&&(data->handler->destructor != dsDataDestroyContent)) {
	err = data->handler->destructor(data);
	if (!err) data->data = NULL;
	return err;
    }

    switch (data->type) {
        case DS_DATA_TYPE_POINTER:
	    dsFree((DSMallocContext)data->owner, (void*)data->data);
	break;
	case DS_DATA_TYPE_DATABUFFER:
	case DS_DATA_TYPE_ALLOCATION:
	    ds_assert_param(data->owner);
	    dsAllocatorFree((DSAllocatorContext)data->owner, (DSAllocation)data->data);
	break;
	case DS_DATA_TYPE_XMLNODEPTR:
	    xmlFreeNode((xmlNodePtr)data->data);
	break;
	case DS_DATA_TYPE_REFERENCE:
	break;
	case DS_DATA_TYPE_CUSTOM:
	    dsDebugServer(DS_DEBUG_ERROR, "dsData(%p): Data destruction failed (the custom data does not provide destroy handler)", data);
	    return DS_ERR_NOTSUPPORTED;
	case DS_DATA_TYPE_INCOMPLETE:
	    dsDebugServer(DS_DEBUG_ERROR, "dsData(%p): Data destruction failed (the data is incomplete)", data);
	    return DS_ERR_INCOMPLETE;
	default:
	    dsDebugServer(DS_DEBUG_ERROR, "dsData(%p): Data destruction failed (unknown data type)", data);
	    return DS_ERR_NOTSUPPORTED;
    }
    
    data->data = NULL;
    
    return 0;
}

int dsDataCopyContent(DSDataCopier copier, DSData dst, DSDataConst src) {
    void *data;

    DSDataCopierS tmp;
    
    ds_assert_param(dst);
    ds_assert_param(src);

    if (!src->data) {
	dst->data = NULL;
	return 0;    
    }
    
    if (!copier) {
	dsDataSetupDefaultCopier(&tmp);
	copier = &tmp;
    }
    
    if ((src->handler)&&(src->handler->copy)&&(src->handler->copy != dsDataCopyContent)) return src->handler->copy(copier, dst, src);

    switch (src->type) {
        case DS_DATA_TYPE_POINTER:
	    if (src->size) {
		switch (copier->type) {
		    case DS_DATA_COPIER_TYPE_MALLOC:
			data = dsMalloc((DSMallocContext)copier->ctx, src->size);
			if (data) {
			    dsDataSetPointer(dst, (DSMallocContext)copier->ctx, data, src->size);
			}
		    break;
		    case DS_DATA_COPIER_TYPE_ALLOCATOR:
			ds_assert_param(copier->ctx);
			data = dsAllocatorSimpleAllocate((DSAllocatorContext)copier->ctx, src->size);
			if (data) {
			    dsDataSetAllocation(dst, (DSAllocatorContext)copier->ctx, data, src->size);
			}
		    break;
		    case DS_DATA_COPIER_TYPE_ANY:			
			data = dsMalloc(NULL, src->size);
			if (data) {
			    dsDataSetPointer(dst, NULL, data, src->size);
			}
		    break;
		    default:
			dsDebugServer(DS_DEBUG_INFO, "dsData(%p): CopyContent is failed (the copier type \"%i\" is not supported)", src, (int)copier->type);
			return DS_ERR_NOTSUPPORTED;
		}
		if (data) {
		    memcpy(data, src->data, src->size);
//		    dsDataSetPointer(dst, dstowner, data, src->size);
//		    dsDataSetHandler(dst, src->handler);
		} else {
		    dsDebugServer(DS_DEBUG_NOTICE, "dsData(%p): CopyContent is failed (DataCopier (type: %i, %p) have failed to allocate %lu bytes of memory)", src, copier->type, copier->ctx, (unsigned long)src->size);
		    return DS_ERR_MEM;
		}
	    } else {
		dsDebugServer(DS_DEBUG_INFO, "dsData(%p): CopyContent is failed (the size of the data is not known)", src);
		return DS_ERR_EMPTY;
	    }    
	break;
	case DS_DATA_TYPE_DATABUFFER:
	case DS_DATA_TYPE_ALLOCATION:
	    dsDebugServer(DS_DEBUG_INFO, "dsData(%p): CopyContent is failed (Allocations are only supported by appropriate handlers)", src);
	    return DS_ERR_NOTSUPPORTED;
	break;
	case DS_DATA_TYPE_XMLNODEPTR:
	    dsDebugServer(DS_DEBUG_INFO, "dsData(%p): CopyContent is failed (XmlNodes are not supported at the moment)", src);
	    return DS_ERR_NOTSUPPORTED;
	break;
	case DS_DATA_TYPE_REFERENCE:
	    return dsDataCopyContent(copier, dst, (DSDataConst)src->data);
	case DS_DATA_TYPE_CUSTOM:
	    dsDebugServer(DS_DEBUG_ERROR, "dsData(%p): CopyContent is failed (the custom data does not provide copy handler)", src);
	    return DS_ERR_NOTSUPPORTED;
	case DS_DATA_TYPE_INCOMPLETE:
	    dsDebugServer(DS_DEBUG_ERROR, "dsData(%p): CopyContent is failed (the data is incomplete)", src);
	    return DS_ERR_INCOMPLETE;
	default:
	    dsDebugServer(DS_DEBUG_ERROR, "dsData(%p): CopyContent is failed (unknown data type)", src);
	    return DS_ERR_NOTSUPPORTED;
    }	
    if (src->user_type) dsDataSetUserType(dst, src->user_type);

    return 0;
}


int dsDataPipeContent(DSDataConst data, DSPipeFunction func, void *attr) {
    int err;
    
    ds_assert_param(data);
    ds_assert_param(func);
    
    if (!data->data) return 0;    
    if ((data->handler)&&(data->handler->pipe)&&(data->handler->pipe != dsDataPipeContent)) return data->handler->pipe(data,func,attr);

    switch (data->type) {
        case DS_DATA_TYPE_POINTER:
	    if (data->size) {
		err = func(attr, data->data, data->size);
		if (err) return err;
	    } else {
		dsDebugServer(DS_DEBUG_INFO, "dsData(%p): PipeContent is failed (the size of the data is not known)", data);
		return DS_ERR_EMPTY;
	    }    
	break;
	case DS_DATA_TYPE_DATABUFFER:
	case DS_DATA_TYPE_ALLOCATION:
	    ds_assert_param(data->owner);
	    err = dsAllocatorCall(((DSAllocatorContext)data->owner), ((DSAllocation)data->data), func, attr, 1);
	    if (err) return err;
	break;
	case DS_DATA_TYPE_XMLNODEPTR:
	    dsDebugServer(DS_DEBUG_INFO, "dsData(%p): PipeContent is failed (XmlNodes are not supported at the moment)", data);
	    return DS_ERR_NOTSUPPORTED;
	break;
	case DS_DATA_TYPE_REFERENCE:
	    return dsDataPipeContent((DSDataConst)data->data, func, attr);
	case DS_DATA_TYPE_CUSTOM:
	    dsDebugServer(DS_DEBUG_ERROR, "dsData(%p): PipeContent is failed (the custom data does not provide copy handler)", data);
	    return DS_ERR_NOTSUPPORTED;
	case DS_DATA_TYPE_INCOMPLETE:
	    dsDebugServer(DS_DEBUG_ERROR, "dsData(%p): PipeContent is failed (the data is incomplete)", data);
	    return DS_ERR_INCOMPLETE;
	default:
	    dsDebugServer(DS_DEBUG_ERROR, "dsData(%p): PipeContent is failed (unknown data type)", data);
	    return DS_ERR_NOTSUPPORTED;
    }	
    return 0;
}


int dsDataDestroyContentDummy(DSData data) {
    ds_assert_param(data);

    return 0;
}

static DSDataHandlerS DS_DATA_HANDLER_STATIC_DATA_S = {
    NULL,
    DS_DATA_HANDLER_TYPE_STANDARD,
    0,
    NULL,				/* copy */
    NULL,				/* pipe */
    dsDataDestroyContentDummy
};

DSDataHandler DS_DATA_HANDLER_STATIC_DATA = &DS_DATA_HANDLER_STATIC_DATA_S;
