#include <dsassert.h>

#include <dslib.h>
#include <dsdebug.h>
#include <dserror.h>
#include <dschunkpool.h>

void dsChunkPoolZeroContext(DSChunkPoolContext ctx) {
    ds_assert_return(ctx);
    ctx->first_chunk = NULL;
    ctx->last_chunk = NULL;
    ctx->lock = NULL;
    ctx->defhandler = NULL;
}

DSChunkPoolContext dsChunkPoolCreateContext(DSMallocContext mctx) {
    DSChunkPoolContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "dsChunkPoolCreateContext");    

    ctx = (DSChunkPoolContext)dsMalloc(mctx,sizeof(struct DSChunkPoolContextT));
    if (ctx) {
	ctx->malloc=mctx;
	dsChunkPoolZeroContext(ctx);
    }
    return ctx;
}

int dsChunkPoolInitContext(DSChunkPoolContext ctx, unsigned char bins) {
    int err;
    unsigned char i;
    DSDataChunkContext *ptr1, *ptr2;

    dsDebugServer(DS_DEBUG_NOTICE, "dsChunkPoolInitContext");    

    ds_assert_param(ctx);
    ds_assert_param(bins);

    ptr1 = (DSDataChunkContext*)dsMalloc(ctx->malloc, bins*sizeof(DSDataChunkContext));
    ptr2 = (DSDataChunkContext*)dsMalloc(ctx->malloc, bins*sizeof(DSDataChunkContext));
    if ((!ptr1)||(!ptr2)) {
	if (ptr1) dsFree(ctx->malloc, ptr1);
	if (ptr2) dsFree(ctx->malloc, ptr2);
	return DS_ERR_MEM;
    }
    
    for (i=0;i<bins;i++) {
	ptr1[i]=NULL;
	ptr2[i]=NULL;
    }

    ctx->lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	dsChunkPoolFreeContext(ctx);
	return DS_ERR_MEM;
    }
    err = dsLockInitContext(ctx->lock);
    if (err) {
	dsChunkPoolFreeContext(ctx);
	return err;
    }

    ctx->bins = bins;
    ctx->first_chunk = ptr1;
    ctx->last_chunk = ptr2;
    
    return 0;
}

void dsChunkPoolFreeContext(DSChunkPoolContext ctx) {
    unsigned char i;
    DSDataChunkContext ptr1, ptr2;

    dsDebugServer(DS_DEBUG_NOTICE, "dsChunkPoolFreeContext");    

    ds_assert_return(ctx);
    
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    
    if (ctx->first_chunk) {
	for (i=0;i<ctx->bins;i++) {
	    for (ptr1=ctx->first_chunk[i];ptr1;ptr1=ptr2) {
		ptr2 = ptr1->next;
		dsDataChunkDestroyContext(ptr1);
	    }
	}
	dsFree(ctx->malloc, ctx->first_chunk);
    }
    if (ctx->last_chunk) dsFree(ctx->malloc, ctx->last_chunk);

    dsChunkPoolZeroContext(ctx);    
}

void dsChunkPoolDestroyContext(DSChunkPoolContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "dsChunkPoolDestroyContext");    
    dsChunkPoolFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}


int dsChunkPoolSetDefaultDataHandler(DSChunkPoolContext ctx, DSDataHandler defhandler) {
    ds_assert_param(ctx);
    ctx->defhandler = defhandler;
    return 0;
}

DSDataChunkContext dsChunkPoolGetChunk(DSChunkPoolContext ctx, DSDataHandler defhandler, unsigned int forms) {
    int err;
    DSDataChunkContext chunk;
    unsigned short int i;
    unsigned char j;
    
    ds_assert_null(ctx);
    ds_assert_null(forms);

    for (j=0,i=(forms>>1);i;i>>=1) j++;
    if (forms != (1<<j)) j++;
    i = (1<<j);
    
    if (j>=ctx->bins) {
	dsDebugServer(DS_DEBUG_WARNING, "dsChunkPoolGetChunk: You should increase number of bins.");
	j=ctx->bins-1;
    }

    
    dsLockWrite(ctx->lock);
    
    chunk = ctx->first_chunk[j];
    if (chunk) {
	if (chunk->next) {
	    ctx->first_chunk[j]=chunk->next;
	} else {
	    ctx->first_chunk[j]=NULL;
	    ctx->last_chunk[j]=NULL;
	}
	dsDataChunkRestoreContext(chunk, defhandler?defhandler:ctx->defhandler);
    } else {
	chunk = dsDataChunkCreateContext(ctx->malloc);
	if (chunk) {
	    err = dsDataChunkInitContext(chunk, defhandler?defhandler:ctx->defhandler, i);
	    if (err) {
		dsDebugServer(DS_DEBUG_ERROR, "dsDataChunkInitContext: %lx (%s)", err, dsError(err, NULL));
		dsDataChunkDestroyContext(chunk);
		chunk = NULL;
	    }
	}
    }
    dsLockReturn(ctx->lock);

    return chunk;
}

void dsChunkPoolReturnChunk(DSChunkPoolContext ctx, DSDataChunkContext chunk) {
    unsigned short int i;
    unsigned char j;

    ds_assert_return(ctx);
    ds_assert_return(chunk);

    dsDataChunkFreeContext(chunk);
    
    for (j=0,i=(chunk->supported_forms>>1);i;i>>=1) j++;
    if (chunk->supported_forms != (1<<j)) j++;
    
    if (j>=ctx->bins) {
	dsDebugServer(DS_DEBUG_WARNING, "dsChunkPoolReturnChunk: You should increase number of bins.");
	j=ctx->bins-1;
    }

    dsLockWrite(ctx->lock);
    if (ctx->last_chunk[j]) {
	ctx->last_chunk[j]->next = chunk;
	ctx->last_chunk[j] = chunk;
    } else {
	ctx->first_chunk[j] = chunk;
	ctx->last_chunk[j] = chunk;
    }
    chunk->next = NULL;    
    dsLockReturn(ctx->lock);
}

