#include <dsassert.h>

#include <dslib.h>
#include <dsdatachunk.h>

DSDataChunkContext dsDataChunkCreateContext(DSMallocContext mctx) {
    int err;
    DSLockContext lock;
    DSConditionContext cond;
    DSDataChunkContext ctx;
    
    lock = dsLockCreateContext(mctx);
    if (!lock) return NULL;
    err = dsLockInitContext(lock);
    if (err) {
	dsLockDestroyContext(lock);
	return NULL;
    }
    
    cond = dsConditionCreateContext(mctx);
    if (!cond) {
	dsLockDestroyContext(lock);
	return NULL;
    }
    err = dsConditionInitContext(cond);
    if (err) {
	dsLockDestroyContext(lock);
	dsConditionDestroyContext(cond);
	return NULL;
    }

    ctx = (DSDataChunkContext)dsMalloc(mctx,sizeof(struct DSDataChunkContextT));
    if (ctx) {
	ctx->malloc = mctx;
	ctx->lock = lock;
	ctx->cond = cond;
	ctx->data = NULL;
	ctx->supported_forms = 0;
    } else {
	dsLockDestroyContext(lock);
	dsConditionDestroyContext(cond);
    }
    return ctx;
}

int dsDataChunkInitContext(DSDataChunkContext ctx, DSDataHandler defhandler, unsigned long supported_forms) {
    long i;
    DSData data;

    ds_assert_param(ctx);
    ds_assert_param(supported_forms);
    
    if ((!ctx->data)||(ctx->supported_forms!=supported_forms)) {
	data = (DSData)dsRealloc(ctx->malloc, ctx->data, supported_forms*sizeof(DSDataS));
	if (!data) return DS_ERR_MEM;
	ctx->data = data;
	ctx->supported_forms=supported_forms;
    }
    return dsDataChunkRestoreContext(ctx, defhandler);
}

int dsDataChunkRestoreContext(DSDataChunkContext ctx, DSDataHandler defhandler) {
    DSData data;
    long i;

    ctx->next = NULL;
    ctx->referers = 0;
    ctx->delayed_referers = 0;
    ctx->defhandler = defhandler;

    data = ctx->data;

    for (i=ctx->supported_forms-1;i>=0;i--) dsDataZero(data+i);

    return 0;
}


void dsDataChunkFreeContext(DSDataChunkContext ctx) {
    DSData data;
    int i;
    
    ds_assert_return(ctx);

    data = ctx->data;
    for (i=ctx->supported_forms-1;i>=0;i--)
	dsDataDestroyContent(data+i);
}

void dsDataChunkDestroyContext(DSDataChunkContext ctx) {
    dsDataChunkFreeContext(ctx);
    if (ctx->data) dsFree(ctx->malloc, ctx->data);
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    if (ctx->cond) dsConditionDestroyContext(ctx->cond);
    dsFree(ctx->malloc,ctx);
}

int dsDataChunkSet(DSDataChunkContext ctx, DSDataChunkForm format, DSDataConst data) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->data);
    ds_assert_param(format<ctx->supported_forms);

    dsDataCopy(ctx->data + format, data);
    if (!data->handler) dsDataSetHandler(ctx->data + format, data->handler);
    return 0;
}


DSDataConst dsDataChunkGet(DSDataChunkContext ctx, DSDataChunkForm format) {
    DSData data;
    
    ds_assert_null(ctx);
    ds_assert_null(ctx->data);
    
    if ((format<ctx->supported_forms)&&(ctx->data[format].data)) {
	data = ctx->data + format;
	if (!data->handler) data->handler = ctx->defhandler;
	return (DSDataConst)data;
    }
    return NULL;
}

DSData dsDataChunkGetData(DSDataChunkContext ctx, DSDataChunkForm format) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->data);
    
    if (format<ctx->supported_forms)
	return ctx->data+format;

    return NULL;
}

DSDataChunkForm dsDataChunkFindData(DSDataChunkContext ctx) {
    DSData data;
    unsigned short i, representations;

    ds_assert_err(ctx, (unsigned short)-1);
    
    data = ctx->data;
    ds_assert_err(data, (unsigned short)-1);
    
    representations = ctx->supported_forms;
    
    for (i=0;i<representations;i++)
	if (!dsDataIsZero(data+i)) return i;
    return (unsigned short)-1;
}


int dsDataChunkSetInfo(DSDataChunkContext ctx, DSDataConst info) {
    DSDataCopierS copier;
    
    ds_assert_param(ctx);
    
    dsDataSetupMallocCopier(&copier, ctx->malloc);
    return dsDataCopyContent(&copier, &ctx->info, info);
}


DSDataConst dsDataChunkGetInfo(DSDataChunkContext ctx) {
    ds_assert_null(ctx);
    
    return &ctx->info;
}

DSDataChunkContext dsDataChunkGetChunkByInfo(DSDataConst ctx) {
    ds_assert_null(ctx);
    
    return (DSDataChunkContext)ctx;
}


unsigned long dsDataChunkRef(DSDataChunkContext ctx) {
    unsigned long res;
    dsLockWrite(((DSDataChunkContext)ctx)->lock);
    res=++(((DSDataChunkContext)ctx)->referers);
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    dsConditionBroadcast(ctx->cond);
    return res;
}

unsigned long dsDataChunkUnref(DSDataChunkContext ctx) {
    unsigned long res;
    dsLockWrite(((DSDataChunkContext)ctx)->lock);
    res=--(((DSDataChunkContext)ctx)->referers);
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    dsConditionBroadcast(ctx->cond);
    return res;
}

unsigned long dsDataChunkGetRefs(DSDataChunkContext ctx) {
    unsigned long res;
    dsLockRead(((DSDataChunkContext)ctx)->lock);
    res=((DSDataChunkContext)ctx)->referers;
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    return res;
}

unsigned long dsDataChunkRefByInfo(DSDataConst ctx) {
    unsigned long res;
    dsLockWrite(((DSDataChunkContext)ctx)->lock);
    res=++(((DSDataChunkContext)ctx)->referers);
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    dsConditionBroadcast(((DSDataChunkContext)ctx)->cond);
    return res;
}

unsigned long dsDataChunkUnrefByInfo(DSDataConst ctx) {
    unsigned long res;
    dsLockWrite(((DSDataChunkContext)ctx)->lock);
    res=--(((DSDataChunkContext)ctx)->referers);
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    dsConditionBroadcast(((DSDataChunkContext)ctx)->cond);
    return res;
}

unsigned long dsDataChunkGetRefsByInfo(DSDataConst ctx) {
    unsigned long res;
    dsLockRead(((DSDataChunkContext)ctx)->lock);
    res=((DSDataChunkContext)ctx)->referers;
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    return res;
}

unsigned long dsDataChunkDelayedRef(DSDataChunkContext ctx) {
    unsigned long res;
    dsLockWrite(((DSDataChunkContext)ctx)->lock);
    res=++(((DSDataChunkContext)ctx)->delayed_referers);
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    dsConditionBroadcast(ctx->cond);
    return res;
}

unsigned long dsDataChunkRefDelayed(DSDataChunkContext ctx) {
    unsigned long res;
    dsLockWrite(((DSDataChunkContext)ctx)->lock);
    (((DSDataChunkContext)ctx)->delayed_referers)--;
    res=++(((DSDataChunkContext)ctx)->referers);
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    dsConditionBroadcast(ctx->cond);
    return res;
}

unsigned long dsDataChunkDelayedUnref(DSDataChunkContext ctx) {
    unsigned long res;
    dsLockWrite(((DSDataChunkContext)ctx)->lock);
    res=--(((DSDataChunkContext)ctx)->delayed_referers);
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    dsConditionBroadcast(ctx->cond);
    return res;
}

unsigned long dsDataChunkGetDelayedRefs(DSDataChunkContext ctx) {
    unsigned long res;
    dsLockRead(((DSDataChunkContext)ctx)->lock);
    res=((DSDataChunkContext)ctx)->delayed_referers;
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    return res;
}

unsigned long dsDataChunkDelayedRefByInfo(DSDataConst ctx) {
    unsigned long res;
    dsLockWrite(((DSDataChunkContext)ctx)->lock);
    res=++(((DSDataChunkContext)ctx)->delayed_referers);
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    dsConditionBroadcast(((DSDataChunkContext)ctx)->cond);
    return res;
}

unsigned long dsDataChunkRefDelayedByInfo(DSDataConst ctx) {
    unsigned long res;
    dsLockWrite(((DSDataChunkContext)ctx)->lock);
    (((DSDataChunkContext)ctx)->delayed_referers)--;
    res=++(((DSDataChunkContext)ctx)->referers);
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    dsConditionBroadcast(((DSDataChunkContext)ctx)->cond);
    return res;
}

unsigned long dsDataChunkDelayedUnrefByInfo(DSDataConst ctx) {
    unsigned long res;
    dsLockWrite(((DSDataChunkContext)ctx)->lock);
    res=--(((DSDataChunkContext)ctx)->delayed_referers);
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    dsConditionBroadcast(((DSDataChunkContext)ctx)->cond);
    return res;
}

unsigned long dsDataChunkGetDelayedRefsByInfo(DSDataConst ctx) {
    unsigned long res;
    dsLockRead(((DSDataChunkContext)ctx)->lock);
    res=((DSDataChunkContext)ctx)->delayed_referers;
    dsLockReturn(((DSDataChunkContext)ctx)->lock);
    return res;
}


void dsDataChunkLock(DSDataChunkContext ctx) {
    ds_assert_return(ctx);
    dsLockWrite(ctx->lock);
}

void dsDataChunkUnlock(DSDataChunkContext ctx) {
    ds_assert_return(ctx);
    dsLockReturn(ctx->lock);
}

int dsDataChunkWaitUnref(DSDataChunkContext ctx, unsigned char delayed) {
    ds_assert_param(ctx);
    if (delayed) {
	dsConditionPreLock(ctx->cond);
	while ((ctx->referers)||(ctx->delayed_referers)) dsConditionWaitLocked(ctx->cond);
	dsConditionUnLock(ctx->cond);
    } else {
	dsConditionPreLock(ctx->cond);
	while (ctx->referers) dsConditionWaitLocked(ctx->cond);
	dsConditionUnLock(ctx->cond);
    }
    return 0;
}

void dsDataChunkRun(DSDataChunkContext ctx, DSDataChunkRunFunction func, void *attr) {
    long i;

    ds_assert_return(ctx);

    for (i=ctx->supported_forms-1;i>=0;i--) 
	if (ctx->data[i].data) func(attr, ctx->data+i);
}
