#include <stdio.h>
#include <dsassert.h>

#include <dslib.h>
#include <dsmalloc.h>
#include <dsdebug.h>
#include <dsdoublehash.h>

#include <dsholderbucket.h>
#include <dsholder.h>

static DSHolderContext holder_ctx = NULL;

static void dsHolderZeroContext(DSHolderContext ctx) {
    ds_assert_return(ctx);
    ctx->lock=NULL;
    ctx->cond=NULL;
    ctx->double_hash=NULL;
    ctx->max_size=0;
    dsDurationZeroDuration(&ctx->timeout);
}

DSHolderContext dsHolderCreateContext(DSMallocContext mctx) {
    DSHolderContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "dsHolderCreateContext");    

    ctx = (DSHolderContext)dsMalloc(mctx, sizeof(struct DSHolderContextT));
    if (ctx) {
	ctx->malloc = mctx;
	dsHolderZeroContext(ctx);
    }
    return ctx;
}

// ToDo: Hard Quota, Soft Quota
int dsHolderInitContext(DSHolderContext ctx, unsigned long size) {
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "dsHolderInitContext");    

    ds_assert_param(ctx);
    
    dsHolderZeroContext(ctx);
    
    ctx->double_hash = dsDoubleHashCreateContext(ctx->malloc);
    if (!ctx->double_hash) return DS_ERR_MEM;

    if (dsDoubleHashInitContext(ctx->double_hash, size, size+1, 1, NULL)) {
	dsHolderFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    ctx->lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	dsHolderFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    err = dsLockInitContext(ctx->lock);
    if (err) {
	dsHolderFreeContext(ctx);
	return err;
    }
    
    ctx->cond = dsConditionCreateContext(ctx->malloc);
    if (!ctx->cond) {
	dsHolderFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    err = dsConditionInitContext(ctx->cond);
    if (err) {
	dsHolderFreeContext(ctx);
	return err;
    }
    
    return 0;
}

void dsHolderSetMaxSize(DSHolderContext ctx, unsigned long max_size) {
    ds_assert_return(ctx);
    
    ctx->max_size = max_size;
    dsDoubleHashSetMaximumSize(ctx->double_hash, max_size);
}

void dsHolderSetTimeout(DSHolderContext ctx, DSDuration timeout) {
    ds_assert_return(ctx);
    
    dsDurationCopy(&ctx->timeout, timeout);
}

void dsHolderFreeContext(DSHolderContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "dsHolderFreeContext");    

    ds_assert_return(ctx);

    if (ctx->cond) dsConditionDestroyContext(ctx->cond);
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    if (ctx->double_hash) dsDoubleHashDestroyContext(ctx->double_hash);
    dsHolderZeroContext(ctx);
}

void dsHolderDestroyContext(DSHolderContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "dsHolderDestroyContext");    

    ds_assert_return(ctx);
    dsHolderFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

void dsHolderMantain(DSHolderContext ctx) {
    unsigned long cleaned;
    
    dsDebugServer(DS_DEBUG_DEBUG, "dsHolderMantain");
    
    ds_assert_return(ctx);
    
    dsLockWrite(ctx->lock);
    cleaned = dsDoubleHashClean(ctx->double_hash, (DSDoubleHashCleanFunction)dsHolderBucketMantain, NULL);
    if (cleaned) dsConditionBroadcast(ctx->cond);
        
    // ToDo: reduce size if overloaded
    dsLockReturn(ctx->lock);
}


unsigned long dsHolderPut(DSHolderContext ctx, DSHolderBucketContext bucket) {
    unsigned long pos;

    ds_assert_err(ctx, (unsigned long)-1);
    ds_assert_err(bucket, (unsigned long)-1);

    dsLockWrite(ctx->lock);
    pos=dsDoubleHashInsert(ctx->double_hash, bucket, PTR2INT(bucket));
    if (pos==(unsigned long)-2) dsConditionPreLock(ctx->cond);
    dsLockReturn(ctx->lock);
    
    while (pos==(unsigned long)-2) {
	if (!dsDurationIsZero(&ctx->timeout)) {
	    pos=dsConditionTimedWait(ctx->cond, &ctx->timeout);
	    if (pos) {
		dsDebugServer(DS_DEBUG_ERROR, "dsHolderPut: Timed Out, bucket: %p", bucket);
		return (unsigned long)-1;
	    }
	} else dsConditionWait(ctx->cond);
	
	dsLockWrite(ctx->lock);
	pos=dsDoubleHashInsert(ctx->double_hash, bucket, PTR2INT(bucket));
	if (pos==(unsigned long)-2) dsConditionPreLock(ctx->cond);
	dsLockReturn(ctx->lock);
    }
    
    dsDebugServer(DS_DEBUG_DEBUG, "dsHolderPut: %lu, bucket: %p", pos, bucket);    

    return pos;
}


DSHolderBucketContext dsHolderGet(DSHolderContext ctx, unsigned long pos, unsigned long magic, unsigned long secret) {
    DSHolderBucketContext bucket;

    ds_assert_null(ctx);
    
    dsLockWrite(ctx->lock);

    bucket=(DSHolderBucketContext)dsDoubleHashGetLink(ctx->double_hash, pos);
    if (!dsHolderBucketCheckMagicNumber(bucket, magic)) {
	dsDebugServer(DS_DEBUG_WARNING, "dsHolderGet: %lu, bucket: %p. Invalid Magic Number!", pos, bucket);
	dsLockReturn(ctx->lock);
	return NULL;
    }
    if (!dsHolderBucketCheckSecretNumber(bucket, secret)) {
	dsDebugServer(DS_DEBUG_WARNING, "dsHolderGet: %lu, bucket: %p. Invalid Secret Number!", pos, bucket);
	dsLockReturn(ctx->lock);
	return NULL;
    }

    dsHolderBucketLock(bucket);
    dsDoubleHashGet(ctx->double_hash, pos);    
    dsLockReturn(ctx->lock);
    dsHolderBucketUnlock(bucket);
    
    dsConditionSend(ctx->cond);
    
    dsDebugServer(DS_DEBUG_DEBUG, "dsHolderGet: %lu, bucket: %p", pos, bucket);    

    return bucket;
}

const void* dsHolderGetLink(DSHolderContext ctx, unsigned long pos, unsigned long magic, unsigned long secret) {
    DSHolderBucketContext bucket;

    ds_assert_null(ctx);
    
    dsLockWrite(ctx->lock);
    bucket=(DSHolderBucketContext)dsDoubleHashGetLink(ctx->double_hash, pos);   
    if (!dsHolderBucketCheckMagicNumber(bucket, magic)) {
	dsDebugServer(DS_DEBUG_WARNING, "dsHolderGetLink: %lu, bucket: %p. Invalid Magic Number!", pos, bucket);
	dsLockReturn(ctx->lock);
	return NULL;
    }
    if (!dsHolderBucketCheckSecretNumber(bucket, secret)) {
	dsDebugServer(DS_DEBUG_WARNING, "dsHolderGetLink: %lu, bucket: %p. Invalid Secret Number!", pos, bucket);
	dsLockReturn(ctx->lock);
	return NULL;
    }
    dsHolderBucketLock(bucket);
    dsLockReturn(ctx->lock);
    
    return bucket->data;
}

void dsHolderReturnLink(DSHolderContext ctx, unsigned long pos) {
    DSHolderBucketContext bucket;
    
    ds_assert_return(ctx);
    
    bucket=(DSHolderBucketContext)dsDoubleHashGetLink(ctx->double_hash, pos);
    ds_assert_return(bucket);
    
    dsHolderBucketRestartTimer(bucket);
    dsHolderBucketUnlock(bucket);
}
