#include <stdio.h>

#include <dslib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <dsdatamanager.h>

static const struct DSDataManagerOptionsT ds_datamanager_options = {
    0,								/* Flags */
    NULL,							/* Chunk Pool */
    1,								/* Buffer Min */
    0								/* Buffer Max */
};

DSDataManagerOptions dsDataManagerGetDefaultOptions() {
    int err;
    DSDataManagerOptions options;

    options = (DSDataManagerOptions)dsMalloc(NULL, sizeof(struct DSDataManagerOptionsT));
    if (!options) return options;
    
    err = dsDataManagerInitDefaultOptions(options);
    if (err) {
	dsDataManagerDestroyOptions(options);
	return NULL;
    }
    return options;
}

int dsDataManagerInitDefaultOptions(DSDataManagerOptions options) {
    ds_assert_param(options);
    memcpy(options, &ds_datamanager_options, sizeof(DSDataManagerOptionsS));
    return 0;
}

int dsDataManagerOptionsSetCacheSize(DSDataManagerOptions options, unsigned long min, unsigned long max) {
    ds_assert_param(options);
    options->cache_min = min;
    options->cache_max = max;
    return 0;
}

int dsDataManagerOptionsSetChunkPool(DSDataManagerOptions options, DSChunkPoolContext chunk_pool) {
    ds_assert_param(options);
    options->chunk_pool = chunk_pool;
    return 0;
}

int dsDataManagerOptionsSetMissingChunkPool(DSDataManagerOptions options, DSChunkPoolContext chunk_pool) {
    ds_assert_param(options);
    if (!options->chunk_pool) {
	options->chunk_pool = chunk_pool;
    }
    return 0;
}

int dsDataManagerOptionsSetFlags(DSDataManagerOptions options, unsigned long flags) {
    ds_assert_param(options);

    options->flags = flags;
    return 0;
}

int dsDataManagerOptionsSetSequentialMode(DSDataManagerOptions options) {
    ds_assert_param(options);
    
    options->flags |= DS_DATA_MANAGER_FLAG_SEQUENTIAL_SOURCE;
    return 0;
}


void dsDataManagerDestroyOptions(DSDataManagerOptions options) {
    if (options) free(options);
}

void dsDataManagerFreeContext(DSDataManagerContext ctx) {
    DSDataChunkContext chunk, nchunk;

    ds_assert_return(ctx);
    
    for (chunk=ctx->chunks;chunk;chunk=nchunk) {
        nchunk = chunk->next;
	dsDataChunkWaitUnref(chunk, 0);
        dsDataChunkDestroyContext(chunk);
    }

    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    if (ctx->chunk_lock) dsLockDestroyContext(ctx->chunk_lock);

    memset(ctx, 0, sizeof(DSDataManagerContextS));
}

int dsDataManagerInitContext(DSDataManagerContext ctx, DSDataManagerOptions options) {
    int err;
    
    ctx->lock = dsLockCreateContext(DS_MODULE(ctx)->malloc);
    ctx->chunk_lock = dsLockCreateContext(DS_MODULE(ctx)->malloc);

    if ((!ctx->lock)||(!ctx->chunk_lock)) {
	dsDataManagerFreeContext(ctx);
	return DS_ERR_LOCK_CREATE;
    }

    err = dsLockInitContext(ctx->lock);
    if (!err) err = dsLockInitContext(ctx->chunk_lock);

    if (err) {
	dsDataManagerFreeContext(ctx);
	return err;
    }

    ctx->chunk_pool = options->chunk_pool;
    
    if (options->cache_min) ctx->cache_min = options->cache_min;
    else ctx->cache_min = 1;
    ctx->cache_max = options->cache_max;

    ctx->flags = options->flags;

    ctx->refs = 1;
    
    return 0;
}

int dsDataManagerRef(DSDataManagerContext ctx) {
    ds_assert_param(ctx);
    
    dsLockWrite(ctx->lock);
    ctx->refs++;
    dsLockReturn(ctx->lock);
    return 0;
}

int dsDataManagerUnref(DSDataManagerContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->manager.Destructor);

    if (!ctx->lock) {
	ctx->manager.Destructor(ctx);
	return 0;
    }
    
    dsLockWrite(ctx->lock);
    if ((--ctx->refs)==0) {
	dsLockReturn(ctx->lock);
	ctx->manager.Destructor(ctx);
	return 0;
    }
    dsLockReturn(ctx->lock);
    return 0;
}

int dsDataManagerPushPointer(DSDataManagerContext ctx, DSDataChunkForm repid, DSDataConst info, void *buffer, size_t size) {
    int err;
    DSDataS data;
    
    ds_assert_param(ctx);
    
    if (!ctx->manager.PushData) return DS_ERR_NOTSUPPORTED;
    
    err = dsDataSetPointer(&data, NULL, buffer, size);
    if (err) return err;
    
    return ctx->manager.PushData(ctx, repid, info, &data);
}

int dsDataManagerPushAllocation(DSDataManagerContext ctx, DSDataChunkForm repid,  DSDataConst info, DSAllocatorContext allocator, DSAllocation allocation) {
    int err;
    DSDataS data;

    ds_assert_param(ctx);
    if (!ctx->manager.PushData) return DS_ERR_NOTSUPPORTED;
    
    err = dsDataSetAllocation(&data, allocator, allocation, DS_DATA_SIZE_UNKNOWN);
    if (err) return err;
    
    return ctx->manager.PushData(ctx, repid, info, &data);
}

static inline_function int dsDataManagerClearInternal(DSDataManagerContext ctx) {
    DSChunkPoolContext chunk_pool;
    DSDataChunkContext chunk, nchunk;
    unsigned long buffer_min, buffer_max;
    unsigned long ref;
    
	/* After chunk_lock locked number of referers can decrease, but not increase! */

    chunk_pool = ctx->chunk_pool;
    chunk = ctx->chunks;

    buffer_min = ctx->cache_min;
    buffer_max = ctx->cache_max;

    nchunk = chunk;


    ref = dsDataChunkGetRefs(chunk);
    if (buffer_max) {
        for (;((ctx->nchunks>buffer_max)&&(!ref));chunk=nchunk) {
            nchunk = chunk->next;
	    ctx->chunks=nchunk;
	    ctx->nchunks--;
	    dsChunkPoolReturnChunk(chunk_pool, chunk);
	    ref = dsDataChunkGetRefs(nchunk);
	}
    }

    for (;((!ref)&&(ctx->nchunks>buffer_min));chunk=nchunk) {
	nchunk=chunk->next;
	ctx->chunks=nchunk;
	ctx->nchunks--;
	dsChunkPoolReturnChunk(chunk_pool, chunk);
	ref = dsDataChunkGetRefs(nchunk);
	if (!ref) ref = dsDataChunkGetDelayedRefs(nchunk);
    }

    
    return 0;
}

int dsDataManagerClear(DSDataManagerContext ctx) {
    int err;
    
    ds_assert_param(ctx);
    
    dsLockWrite(ctx->chunk_lock);
    err = dsDataManagerClearInternal(ctx);
    dsLockReturn(ctx->chunk_lock);

    return err;
}

int dsDataManagerRegisterChunk(DSDataManagerContext ctx, DSDataChunkContext chunk) {
    ds_assert_param(ctx);
    ds_assert_param(chunk);

    dsLockWrite(ctx->chunk_lock);
    ctx->nchunks++;
    if (ctx->current) {
	/* What to do with OOB items, with time earlier than current one?
	For Fast items can be assumed only one sequential data source? For items
	with multiple or/and non-sequential data sources OPCDataContext should be used
	if (dsTimeCmp(ctx->current->info.time, info->time)>0) {
	}
	*/
	ctx->current->next = chunk;
	ctx->current=chunk;

	if ((ctx->nchunks>ctx->cache_min)&&(!dsDataChunkGetRefs(ctx->chunks))) 
	    dsDataManagerClearInternal(ctx);
    } else {
	ctx->chunks = chunk;
	ctx->current = chunk;
    }
    dsLockReturn(ctx->chunk_lock);
    
    return 0;
}

int dsDataManagerReturnData(DSDataManagerContext ctx, DSDataConst info) {
    ds_assert_param(ctx&&info);

    if (ctx->manager.ReturnData) ctx->manager.ReturnData(ctx, info);

    dsLockWrite(ctx->chunk_lock);
    if ((!dsDataChunkUnrefByInfo(info))&&(ctx->chunks == dsDataChunkGetChunkByInfo(info)))
	dsDataManagerClearInternal(ctx);
    dsLockReturn(ctx->chunk_lock);
    
    dsDataManagerUnref(ctx);

    return 0;
}

int dsDataManagerSetAcquireFunction(DSDataManagerContext ctx, DSDataManagerAcquireFunction acquire) {
    ds_assert_param(ctx);
    
    ctx->manager.Acquire = acquire;
    return 0;
}
