#include <dslib.h>
#include <dsdebug.h>
#include <dsassert.h>

#include <dsbinaryformat.h>
#include <dsbinaryformatnative.h>
#include <dsdatachunk.h>
#include <dsdatamanagerfast.h>
#include <dsdataconverter.h>


DSDataManagerFastOptions dsDataManagerFastGetDefaultOptions() {
    int err;
    DSDataManagerFastOptions options;

    options = (DSDataManagerFastOptions)malloc(sizeof(struct DSDataManagerFastOptionsT));
    if (!options) return options;
    
    err = dsDataManagerFastInitDefaultOptions(options);
    if (err) {
	dsDataManagerFastDestroyOptions(options);
	return NULL;
    }
    return options;
}

int dsDataManagerFastInitDefaultOptions(DSDataManagerFastOptions options) {
    ds_assert_param(options);
    dsDataManagerInitDefaultOptions(&options->manager);
    dsDataBufferInitDefaultOptions(&options->allocator);
    return 0;
}

void dsDataManagerFastDestroyOptions(DSDataManagerFastOptions options) {
    if (options) free(options);
}

int dsDataManagerFastOptionsSetBufferSize(DSDataManagerFastOptions options, unsigned long min, unsigned long max) {
    ds_assert_param(options);
    options->allocator.minimal_size = min;
    options->allocator.maximal_size = max;
    return 0;
}

int dsDataManagerFastOptionsSetBufferPool(DSDataManagerFastOptions options, void *pool) {
    ds_assert_param(options);
    options->allocator.pool = pool;
    return 0;
}

int dsDataManagerFastOptionsSetMissingBufferPool(DSDataManagerFastOptions options, void *pool) {
    ds_assert_param(options);
    if (!options->allocator.pool) options->allocator.pool = pool;
    return 0;
}

int dsDataManagerFastOptionsSetFixedAllocationSize(DSDataManagerFastOptions options, size_t size) {
    ds_assert_param(options);
    if (size) {
	options->allocator.fixed_allocation_size = size;
    }
    return 0;
}


int dsDataManagerFastDataDestroy(DSData data) {
    DSDataManagerFastContext ctx;
    
    ds_assert_param(data);

    ctx = (DSDataManagerFastContext)data->handler->ctx;
    if (!ctx) ctx = (DSDataManagerFastContext)data->owner;
    
    ds_assert_param(ctx);
    
    return dsBinaryFormatDestruct(ctx->format, ctx->type, data);
}

int dsDataManagerFastDataCopy(DSDataCopier copier, DSData dst, DSDataConst src) {
    DSDataManagerFastContext ctx;
    
    ds_assert_param(src);
    ds_assert_param(dst);
    ds_assert_param(copier);
    
    if (copier->type != DS_DATA_COPIER_TYPE_ALLOCATOR) return DS_ERR_NOTSUPPORTED;

    ctx = (DSDataManagerFastContext)src->handler->ctx;
    if (!ctx) ctx = (DSDataManagerFastContext)src->owner;
    ds_assert_param(ctx);
    
    return dsBinaryFormatCopy(ctx->format, ctx->type, (DSAllocatorContext)copier->ctx, dst, src);
}

int dsDataManagerFastDataPipe(DSDataConst data, DSPipeFunction func, void *attr) {
    DSDataManagerFastContext ctx;
    
    ds_assert_param(data);

    ctx = (DSDataManagerFastContext)data->handler->ctx;
    if (!ctx) ctx = (DSDataManagerFastContext)data->owner;
    ds_assert_param(ctx);
    
    return dsBinaryFormatPipe(ctx->format, ctx->type, data, func, attr);
}

int dsDataManagerFastDataUserize(DSData user_data, DSDataConst data, DSDataManagerUserizeOptions options) {
    int err;
    DSDataConverterContext converter;
    DSDataManagerFastContext ctx;
    
    ds_assert_param(user_data);
    ds_assert_param(data);
    
    ds_assert_param(data->type == DS_DATA_TYPE_ALLOCATION);

    if ((!options)||(!options->representation)||(strcmp(options->representation, DS_DATA_MANAGER_FORM_OPCXML)))
	return DS_ERR_NOTAPPLICABLE;
    
    ctx = (DSDataManagerFastContext)data->handler->ctx;
    if (!ctx) ctx = (DSDataManagerFastContext)data->owner;
    ds_assert_param(ctx);
    
    converter = dsDataConverterCreateOPCXML(NULL, ctx->type);
    if (!converter) return dsErrorModule(DS_ERR_CREATE, DS_MODULE_DataConverter);

    err = dsDataConverterConvertData(converter, 
	dsDataConverterGetRepresentation(converter, DS_DATA_CONVERTER_OPCXML_FORM_OPCXML), user_data, 
	dsDataConverterGetRepresentation(converter, DS_DATA_CONVERTER_OPCXML_FORM_NATIVE), data
    );
    
    dsDataConverterDestroyContext(converter);
    
    return err;
}

static void dsDataManagerFastFreeContext(DSDataManagerFastContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "dsDataManagerFast(%p): FreeContext", ctx);    

    ds_assert_return(ctx);
    ds_assert_info(!ctx->grabed_allocator);

    dsDataManagerFreeContext((DSDataManagerContext)ctx);

	// All chunks should be freed before deallocating allocator
    if (ctx->format) dsBinaryFormatDestroyInstance(ctx->format);
    if (ctx->allocator) dsDataBufferDestroyContext(ctx->allocator);
    if (ctx->allocator_lock) dsLockDestroyContext(ctx->allocator_lock);
    
    memset(ctx, 0, sizeof(DSDataManagerFastContextS));
}

static void dsDataManagerFastDestroyContext(DSDataManagerFastContext ctx) {
    dsDestroyModule(DataManagerFast, ctx);
}

static const DSDataManagerS ds_datamanager_fast = {
    (DSDestroyContextFunction)dsDataManagerFastDestroyContext,
    NULL,							/* Acquire (Costumized by user) */
    (DSDataManagerPushDataFunction)dsDataManagerFastPushData,
    (DSDataManagerGetAllocatorFunction)dsDataManagerFastGetAllocator,
    (DSDataManagerReturnAllocatorFunction)dsDataManagerFastReturnAllocator,
    (DSDataManagerGetDataFunction)dsDataManagerFastGetData,
    NULL,							/* ReturnData */
    (DSDataManagerMantainFunction)dsDataManagerFastMantain,
    (DSDataManagerGetFormByNameFunction)dsDataManagerFastGetFormByName
};

DSDataManagerFastContext dsDataManagerFastCreateContext(DSMallocContext mctx) {
    DSDataManagerFastContext ctx;
    
    ctx = dsNewModule(DataManagerFast, mctx);
    if (ctx) {
	memcpy(&((DSDataManagerContext)ctx)->manager, &ds_datamanager_fast, sizeof(DSDataManagerS));
    }

    return ctx;
}

int dsDataManagerFastInitContext(DSDataManagerFastContext ctx, DSBinaryType type, DSDataManagerFastOptions options) {
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "dsDataManagerFast(%p): InitContext (Type: %p)", ctx, type);

    ds_assert_param(ctx);
    ds_assert_param(type);
	
	// DS: Create Local One, if not passed
    ds_assert_param(DS_DATA_MANAGER_OPTIONS(options)->chunk_pool);
    
    err = dsDataManagerInitContext((DSDataManagerContext)ctx, (DSDataManagerOptions)options);
    if (err) {
	dsDataManagerFastFreeContext(ctx);
	return err;
    }

    ctx->type = type;

    ctx->format = dsBinaryFormatCreateInstance(dsBinaryFormatNative(), DS_MODULE(ctx)->malloc, NULL);
    if (!ctx->format) {
	dsDataManagerFastFreeContext(ctx);
	return DS_ERR_FORMAT_NATIVE_CREATE_INSTANCE;
    }    

    ctx->allocator = dsDataBufferCreateContext(DS_MODULE(ctx)->malloc);
    if (!ctx->allocator) {
	dsDataManagerFastFreeContext(ctx);
	return DS_ERR_DATABUFFER_CREATE;
    }
    
    err = dsDataBufferInitContext(ctx->allocator, &options->allocator);
    if (err) {
	dsDataManagerFastFreeContext(ctx);
	return err;
    }

    ctx->allocator_lock = dsLockCreateContext(DS_MODULE(ctx)->malloc);
    if (!ctx->allocator_lock) {
	dsDataManagerFastFreeContext(ctx);
	return DS_ERR_LOCK_CREATE;
    }

    err = dsLockInitContext(ctx->allocator_lock);
    if (err) {
	dsDataManagerFastFreeContext(ctx);
	return err;
    }

    memset(&ctx->handler, 0, sizeof(DSDataManagerDataHandler));
    dsDataHandlerConfigureDefaultMethods((DSDataHandler)&ctx->handler);
    dsDataHandlerSetContext(DS_DATA_HANDLER(&ctx->handler), ctx);
    dsDataHandlerSetType(DS_DATA_HANDLER(&ctx->handler), DS_DATA_HANDLER_TYPE_DATAMANAGER);
    dsDataHandlerSetUserType(DS_DATA_HANDLER(&ctx->handler), DS_DATA_MANAGER_DATAHANDLER_TYPE_MANAGERFAST);

    if (ctx->format->format->DestructData)
	dsDataHandlerSetDestructorFunction(DS_DATA_HANDLER(&ctx->handler), dsDataManagerFastDataDestroy);
    if (ctx->format->format->CopyData)
	dsDataHandlerSetCopyFunction(DS_DATA_HANDLER(&ctx->handler), dsDataManagerFastDataCopy);
    if (ctx->format->format->PipeData)
	dsDataHandlerSetPipeFunction(DS_DATA_HANDLER(&ctx->handler), dsDataManagerFastDataPipe);
    
    dsModuleMarkInitialized(ctx);

    return 0;
}

int dsDataManagerFastSetAcquireFunction(DSDataManagerFastContext ctx, DSDataManagerAcquireFunction acquire) {
    ds_assert_param(ctx);
    
    DS_DATA_MANAGER(ctx)->manager.Acquire = acquire;
    return 0;
}

int dsDataManagerFastSetUserizeFunction(DSDataManagerFastContext ctx, DSDataManagerGetUserDataFunction get, void *userizer) {
    ds_assert_param(ctx);
    
    ctx->handler.userizer = userizer;
    ctx->handler.GetUserData = get;
    return 0;
}

int dsDataManagerFastPushData(DSDataManagerFastContext ctx, DSDataChunkForm repid, DSDataConst info, DSDataConst data) {
    int err;
    DSDataChunkContext chunk;
    DSData chunkdata;

    dsDebugServer(DS_DEBUG_DEBUG, "dsDataManagerFast(%p): Push", ctx);

    ds_assert_param(ctx);
    ds_assert_param(data);
    ds_assert_param(!repid);

    if (ctx->grabed_allocator) return DS_ERR_BUSY;

    chunk = dsChunkPoolGetChunk(DS_DATA_MANAGER(ctx)->chunk_pool, (DSDataHandler)&ctx->handler, 1);
    if (!chunk) return DS_ERR_MEM;

    err = dsDataChunkSetInfo(chunk, info);

    if (!err) {    
	dsLockWrite(ctx->allocator_lock);
	chunkdata = dsDataChunkGetData(chunk, DS_DATA_CHUNK_FORM_1);
	err = dsBinaryFormatImport(ctx->format, ctx->type, (DSAllocatorContext)ctx->allocator, chunkdata, data);
	if (!err) err = dsDataManagerRegisterChunk(DS_DATA_MANAGER(ctx), chunk);
	dsLockReturn(ctx->allocator_lock);
    }
    
    if (err) dsChunkPoolReturnChunk(DS_DATA_MANAGER(ctx)->chunk_pool, chunk);
    
    return err;
}

int dsDataManagerFastGetData(DSDataManagerFastContext ctx, DSDataChunkForm repid, DSDataConst *info, DSDataConst *data) {
    DSDataChunkContext current;

    ds_assert_complexctx(ctx, DS_DATA_MANAGER(ctx)->current);
    ds_assert_param(!repid);

    dsLockRead(DS_DATA_MANAGER(ctx)->chunk_lock);
    current =  DS_DATA_MANAGER(ctx)->current;
    dsDataChunkRef(current);
    dsLockReturn(DS_DATA_MANAGER(ctx)->chunk_lock);
    
    if (data) *data=dsDataChunkGet(current, 0);
    if (info) *info=dsDataChunkGetInfo(current);
    
    dsDataManagerRef((DSDataManagerContext)ctx);
    
    return 0;
}


DSAllocatorContext dsDataManagerFastGetAllocator(DSDataManagerFastContext ctx, DSDataChunkForm repid) {
    ds_assert_null(ctx);
    ds_assert_null(!repid);
    ds_assert_null(!ctx->grabed_allocator);
    
    dsLockWrite(ctx->allocator_lock);
    ctx->grabed_allocator = 1;
    return (DSAllocatorContext)ctx->allocator;
}

int dsDataManagerFastReturnAllocator(DSDataManagerFastContext ctx, DSDataChunkForm repid, DSAllocatorContext allocator) {
    ds_assert_param(ctx);
    ds_assert_param(!repid);
    ds_assert_param(ctx->grabed_allocator);
    
    ctx->grabed_allocator = 0;
    dsLockReturn(ctx->allocator_lock);
    
    return 0;
}

int dsDataManagerFastMantain(DSDataManagerFastContext ctx) {
    ds_assert_return(ctx);
    if (ctx->allocator) dsDataBufferMantain(ctx->allocator);
    return 0;
}

DSDataChunkForm dsDataManagerFastGetFormByName(DSDataManagerContext ctx, const char *title) {
    ds_assert_param(ctx);
    
    if (!strcmp(title, DS_DATA_MANAGER_FORM_NATIVE)) return 0;
    return DS_DATA_CHUNK_FORM_INVALID;
}

