#include <dslib.h>

#include <dsassert.h>
#include <dserror.h>
#include <dsdebug.h>

#include <dsbinaryformatnative.h>
#include <dsbinaryformatopcxml.h>
#include <dsrepresentationbinary.h>
#include <dsallocatorsimple.h>

#include <dsdatamanagerrep.h>

#define DS_DATA_MANAGER_REP_HASHTABLE_SIZE 16
#define DS_DATA_MANAGER_REP_ALLOCATE_PRESENTATIONS 2


DSDataManagerRepOptions dsDataManagerRepGetDefaultOptions() {
    int err;
    DSDataManagerRepOptions options;

    options = (DSDataManagerRepOptions)malloc(sizeof(struct DSDataManagerRepOptionsT));
    if (!options) return options;
    
    err = dsDataManagerRepInitDefaultOptions(options);
    if (err) {
	dsDataManagerRepDestroyOptions(options);
	return NULL;
    }
    return options;
}

int dsDataManagerRepInitDefaultOptions(DSDataManagerRepOptions options) {
    ds_assert_param(options);
    dsDataManagerInitDefaultOptions(&options->manager);
    dsDataBufferInitDefaultOptions(&options->allocator);
    return 0;
}

void dsDataManagerRepDestroyOptions(DSDataManagerRepOptions options) {
    if (options) free(options);
}

int dsDataManagerRepOptionsSetBufferSize(DSDataManagerRepOptions options, unsigned long min, unsigned long max) {
    ds_assert_param(options);
    options->allocator.minimal_size = min;
    options->allocator.maximal_size = max;
    return 0;
}

int dsDataManagerRepOptionsSetBufferPool(DSDataManagerRepOptions options, DSBufferPoolContext buffer_pool) {
    ds_assert_param(options);
    options->allocator.pool = buffer_pool;
    return 0;
}

int dsDataManagerRepOptionsSetMissingBufferPool(DSDataManagerRepOptions options, DSBufferPoolContext buffer_pool) {
    ds_assert_param(options);
    if (!options->allocator.pool) options->allocator.pool = buffer_pool;
    return 0;
}

int dsDataManagerRepOptionsSetFixedAllocationSize(DSDataManagerRepOptions options, size_t size) {
    ds_assert_param(options);
    if (size) {
	options->allocator.fixed_allocation_size = size;
    }
    return 0;
}

static void dsDataManagerRepFreeContext(DSDataManagerRepContext ctx) {
    ds_assert_return(ctx);

    dsDataManagerFreeContext((DSDataManagerContext)ctx);

	// All chunks should be freed before deallocating allocator
    if (ctx->converter) dsDestroyModule(DataConverter, ctx->converter);
    if (ctx->rephash) dsHashTableDestroy(ctx->rephash);    
    if (ctx->representation) dsFree(DS_MODULE(ctx)->malloc, ctx->representation);
    if (ctx->representation_lock) dsLockDestroyContext(ctx->representation_lock);
    if (ctx->writer_lock) dsLockDestroyContext(ctx->writer_lock);
    
    dsZeroModule(DataManagerRep, ctx);

    dsModuleDebug(ctx, DS_DEBUG_DEBUG, "FreeContext");    
}

static void dsDataManagerRepDestroyContext(DSDataManagerRepContext ctx) {
    dsDestroyModule(DataManagerRep, ctx);
}

static const DSDataManagerS ds_datamanager_rep = {
    (DSDestroyContextFunction)dsDataManagerRepDestroyContext,
    NULL,							/* Acquire (Costumized by user) */
    (DSDataManagerPushDataFunction)dsDataManagerRepPushData,
    (DSDataManagerGetAllocatorFunction)dsDataManagerRepGetAllocator,
    (DSDataManagerReturnAllocatorFunction)dsDataManagerRepReturnAllocator,
    (DSDataManagerGetDataFunction)dsDataManagerRepGetData,
    NULL,							/* ReturnData */
    (DSDataManagerMantainFunction)dsDataManagerRepMantain,
    (DSDataManagerGetFormByNameFunction)dsDataManagerRepGetFormByName
};

DSDataManagerRepContext dsDataManagerRepCreateContext(DSMallocContext mctx) {
    DSDataManagerRepContext ctx;

    ctx = dsNewModule(DataManagerRep, mctx);
    if (ctx) {
	memcpy(&((DSDataManagerContext)ctx)->manager, &ds_datamanager_rep, sizeof(DSDataManagerS));
    }

    return ctx;
}

int dsDataManagerRepInitContext(DSDataManagerRepContext ctx, DSDataManagerRepOptions options) {
    int err;
    unsigned short int i;

    dsModuleDebug(ctx, DS_DEBUG_DEBUG, "InitContext");

    ds_assert_param(ctx);
    ds_assert_param(options);

    err = dsDataManagerInitContext((DSDataManagerContext)ctx, (DSDataManagerOptions)options);
    if (err) {
	dsDataManagerRepFreeContext(ctx);
	return err;
    }

    memcpy(&ctx->databuffer_options, &options->allocator, sizeof(struct DSDataBufferOptionsT));

    ctx->representation_lock = dsLockCreateContext(DS_MODULE(ctx)->malloc);
    ctx->writer_lock = dsLockCreateContext(DS_MODULE(ctx)->malloc);
    if ((!ctx->representation_lock)||(!ctx->writer_lock)) {
	dsDataManagerRepFreeContext(ctx);
	return dsModuleError(ctx, DS_ERR_MEM);
    }
    err = dsLockInitContext(ctx->representation_lock);
    if (err) {
	dsDataManagerRepFreeContext(ctx);
	return err;
    }
    err = dsLockInitContext(ctx->writer_lock);
    if (err) {
	dsDataManagerRepFreeContext(ctx);
	return err;
    }
    
    ctx->converter = dsModuleNewModule(ctx, DataConverter);
    if (!ctx->converter) {
	dsDataManagerRepFreeContext(ctx);
	return dsModuleError(ctx, DS_ERR_MEM);
    }
    
    err = dsDataConverterInitContext(ctx->converter, DS_DATA_MANAGER_REP_ALLOCATE_PRESENTATIONS);
    if (err) {
	dsDataManagerRepFreeContext(ctx);
	return err;
    }
    
    ctx->allocated_forms = DS_DATA_MANAGER_REP_ALLOCATE_PRESENTATIONS;

    ctx->representation = (DSDataManagerRepRepresentationData)dsMalloc(DS_MODULE(ctx)->malloc, ctx->allocated_forms * sizeof(DSDataManagerRepRepresentationDataS));
    ctx->rephash = dsHashTableInit(DS_MODULE(ctx)->malloc, DS_DATA_MANAGER_REP_HASHTABLE_SIZE, DS_HASH_TYPE_STATICPOINTER, NULL);
    if ((!ctx->representation)||(!ctx->rephash)) {
	dsDataManagerRepFreeContext(ctx);
	return dsModuleError(ctx, DS_ERR_MEM);
    }
/*
    dsHashTableSetDestructor(ctx->rephash, (HashTableDestructorFunction)dsDataManagerRepDestroyRepresentationData);
*/
    ctx->supported_forms = 0;
    
    dsModuleMarkInitialized(ctx);

    return 0;
}

DSDataManagerRepContext dsDataManagerRepCreateOPCXML(DSMallocContext mctx, DSBinaryType type, DSDataManagerRepOptions options) {
    int err;
    
    DSDataChunkForm native;
    DSDataManagerRepContext ctx;
    
    ctx = dsDataManagerRepCreateContext(mctx);
    if (!ctx) return NULL;
    
    err = dsDataManagerRepInitContext(ctx, options);
    if (err) {
	dsDataManagerRepDestroyContext(ctx);
	return NULL;
    }
    
    native = dsDataManagerRepRegisterNativeRepresentation(ctx, type, NULL);
    if (native == DS_DATA_CHUNK_FORM_INVALID) {
	dsDataManagerRepDestroyContext(ctx);
	return NULL;
    }
    
    if (dsDataManagerRepRegisterOPCXMLRepresentation(ctx, type) == DS_DATA_CHUNK_FORM_INVALID) {
	dsDataManagerRepDestroyContext(ctx);
	return NULL;
    }
    
    return ctx;
}


DSDataChunkForm dsDataManagerRepRegisterRepresentation(DSDataManagerRepContext ctx, DSDataChunkForm base, const char *title, DSRepresentation representation, DSRepresentationOptions options) {
    int err;

    DSDataChunkForm repid;
    DSRepresentationInstance instance;

    DSDataManagerRepRepresentationData representations;
    DSDataManagerRepRepresentationData repdata, basedata;
    
    ds_module_assert_default(ctx, DS_DATA_CHUNK_FORM_INVALID);
    ds_assert_default(representation, DS_DATA_CHUNK_FORM_INVALID);
    ds_assert_default(title, DS_DATA_CHUNK_FORM_INVALID);
    
    if (ctx->supported_forms) {
	if (base >= ctx->supported_forms) {
	    dsModuleDebug(ctx, DS_DEBUG_ERROR, "RegisterRepresentation(%s). Non-existing base representation (%i) is specified", title, (int)base);
	    return DS_DATA_CHUNK_FORM_INVALID;
	}
    } else {
	if (base != DS_DATA_CHUNK_FORM_MISSING) {
	    dsModuleDebug(ctx, DS_DEBUG_ERROR, "RegisterRepresentation(%s). The root representation is already registered (the base representation should be specified)", title);
	    return DS_DATA_CHUNK_FORM_INVALID;
	}
    }

    if (dsDataManagerRepGetFormByName(ctx, title) != DS_DATA_CHUNK_FORM_MISSING) {
	dsModuleDebug(ctx, DS_DEBUG_ERROR, "RegisterRepresentation(%s). The representation is already registered", title);
	return DS_DATA_CHUNK_FORM_INVALID;
    }

    dsLockWrite(ctx->representation_lock);
    if (ctx->supported_forms == ctx->allocated_forms) {
	representations = (DSDataManagerRepRepresentationData)dsRealloc(DS_MODULE(ctx)->malloc, ctx->representation, 2 * ctx->allocated_forms * sizeof(DSDataManagerRepRepresentationDataS));
	if (!representations) {
	    dsLockReturn(ctx->representation_lock);
	    dsModuleDebug(ctx, DS_DEBUG_CRITICAL, "RegisterRepresentation(%s). Memory allocation is failed", title);
	    return DS_DATA_CHUNK_FORM_INVALID;
	}
	ctx->representation = representations;
	ctx->allocated_forms *= 2;
    }

    repid = ctx->supported_forms;
    
    repdata = ctx->representation + repid;
    memset(&repdata->handler, 0, sizeof(DSDataManagerDataHandler));
    dsDataHandlerSetType((DSDataHandler)&repdata->handler, DS_DATA_HANDLER_TYPE_DATAMANAGER);
    dsDataHandlerSetType((DSDataHandler)&repdata->handler, DS_DATA_MANAGER_DATAHANDLER_TYPE_REPRESENTATION);
    dsRepresentationOptionsSetDataHandler(options, (DSDataHandler)&repdata->handler);

    instance = dsRepresentationCreateInstance(representation, DS_MODULE(ctx)->malloc, options);
    if (!instance) {
	dsModuleDebug(ctx, DS_DEBUG_ERROR, "RegisterRepresentation(%s). Creation of Representation Instance is failed", title);
	return DS_DATA_CHUNK_FORM_INVALID;
    }

    repdata->title = title;
    repdata->formid = repid;
    repdata->reference = instance;
    repdata->flags = 0;

    if (base == DS_DATA_CHUNK_FORM_MISSING) basedata = NULL;
    else basedata = ctx->representation + base;
    
    err = dsHashTableAdd(ctx->rephash, repdata, 1);
    if (err) {
	dsRepresentationDestroyInstance(instance);
	dsLockReturn(ctx->representation_lock);
	dsModuleDebug(ctx, DS_DEBUG_ERROR, "RegisterRepresentation(%s). Registration of Representation in the HashTable is failed with error: %X(%s)", title, err, dsError(err, NULL));
	return DS_DATA_CHUNK_FORM_INVALID;
    }

    err = dsDataConverterRegisterRepresentationInstance(ctx->converter, basedata?(basedata->reference):NULL, instance);
    if (err) {
	dsHashTableDelete(ctx->rephash, (void*)repdata);
	dsLockReturn(ctx->representation_lock);
	dsRepresentationDestroyInstance(instance);
	dsModuleDebug(ctx, DS_DEBUG_ERROR, "RegisterRepresentation(%s). DataConverter is not accepting representation, error: %X(%s)", title, err, dsError(err, NULL));
	return DS_DATA_CHUNK_FORM_INVALID;
    }

    dsLockReturn(ctx->representation_lock);

    ctx->supported_forms++;
    dsModuleDebug(ctx, DS_DEBUG_DEBUG, "RegisterRepresentation(%s): Registered as %u.", title, repid);

    return repid;
}

DSDataChunkForm dsDataManagerRepRegisterBinaryRepresentation(DSDataManagerRepContext ctx, DSDataChunkForm base, const char *title, DSBinaryFormat format, DSBinaryType type, DSAllocatorContext allocator) {
    int err;
    DSRepresentationBinaryOptionsS options;

    DSDataChunkForm repid;
    
    ds_module_assert_default(ctx, DS_DATA_CHUNK_FORM_INVALID);
    ds_assert_default(format, DS_DATA_CHUNK_FORM_INVALID);
    ds_assert_default(type, DS_DATA_CHUNK_FORM_INVALID);
    ds_assert_default(type, DS_DATA_CHUNK_FORM_INVALID);
    
    dsRepresentationBinaryInitDefaultOptions(&options, format, type);

    if (!allocator) {
      switch(format->allocator) {
        case DS_BINARY_FORMAT_ALLOCATOR_TYPE_FULL:
	    if ((ctx->databuffer_options.pool)&&(DS_DATA_MANAGER(ctx)->flags&DS_DATA_MANAGER_FLAG_SEQUENTIAL_SOURCE)) {
		dsDataBufferOptionsSetAllocationSize(&ctx->databuffer_options, type->size);

		allocator = (DSAllocatorContext)dsDataBufferCreateContext(DS_MODULE(ctx)->malloc);
		if (!allocator) {
		    dsModuleDebug(ctx, DS_DEBUG_CRITICAL, "RegisterBinaryRepresentation(%s).  Memory allocation for DataBuffer(Complete Allocator) is failed", title);
		    return DS_ERR_DATABUFFER_CREATE;
		}
	    
		err = dsDataBufferInitContext((DSDataBufferContext)allocator, &ctx->databuffer_options);
		if (err) {
		    dsDataBufferDestroyContext((DSDataBufferContext)allocator);
		    dsModuleDebug(ctx, DS_DEBUG_ERROR, "RegisterBinaryRepresentation(%s).  DataBuffer (Complete Allocator) initialization failed with error: %X(%s)", title, err, dsError(err, NULL));
		    return DS_DATA_CHUNK_FORM_INVALID;
		}
		break;
	    }
	case DS_BINARY_FORMAT_ALLOCATOR_TYPE_MINIMAL:
	    allocator = (DSAllocatorContext)dsAllocatorSimpleCreateContext(DS_MODULE(ctx)->malloc);
	    if (!allocator) {
		dsModuleDebug(ctx, DS_DEBUG_CRITICAL, "RegisterBinaryRepresentation(%s).  Memory allocation for AllocatorSimple (Simple Allocator) is failed", title);
		return DS_ERR_ALLOCATOR_CREATE;
	    }
	    
	    err = dsAllocatorSimpleInitContext((DSAllocatorSimpleContext)allocator, NULL);
	    if (err) {
		dsModuleDebug(ctx, DS_DEBUG_ERROR, "RegisterBinaryRepresentation(%s). AllocatorSimple (Simple Allocator) initialization failed with error: %X(%s)", title, err, dsError(err, NULL));
		dsAllocatorSimpleDestroyContext((DSAllocatorSimpleContext)allocator);
		return DS_DATA_CHUNK_FORM_INVALID;
	    }
	break;
	case DS_BINARY_FORMAT_ALLOCATOR_TYPE_ABSENT:
	break;
	default:
	    dsModuleDebug(ctx, DS_DEBUG_ERROR, "RegisterBinaryRepresentation(%s). Unsupported allocator type (%i) is required by format", title, (int)format->allocator);
	    return DS_DATA_CHUNK_FORM_INVALID;
      }
    }
    dsRepresentationBinaryOptionsSetAllocator(&options, allocator);

    repid = dsDataManagerRepRegisterRepresentation(ctx, base, title, dsRepresentationBinary(), (DSRepresentationOptions)&options);
    if (repid != DS_DATA_CHUNK_FORM_INVALID) {
	ctx->representation[repid].flags |= DS_DATA_MANAGER_REP_REPRESENTATION_BINARY;
    }

    return repid;
}

DSDataChunkForm dsDataManagerRepRegisterNativeRepresentation(DSDataManagerRepContext ctx, DSBinaryType type, DSAllocatorContext allocator) {
    DSDataChunkForm repid;

    ds_module_assert_default(ctx, DS_DATA_CHUNK_FORM_INVALID);
    ds_assert_default(type, DS_DATA_CHUNK_FORM_INVALID);
    ds_assert_default(!ctx->supported_forms, DS_DATA_CHUNK_FORM_INVALID);

    repid = dsDataManagerRepRegisterBinaryRepresentation(ctx, DS_DATA_CHUNK_FORM_MISSING, DS_DATA_MANAGER_FORM_NATIVE, dsBinaryFormatNative(), type, allocator);

    if (repid != DS_DATA_CHUNK_FORM_INVALID) {
	ctx->representation[repid].flags |= DS_DATA_MANAGER_REP_REPRESENTATION_NATIVE;
    }
    
    return repid;
}

DSDataChunkForm dsDataManagerRepRegisterOPCXMLRepresentation(DSDataManagerRepContext ctx, DSBinaryType type) {
    DSDataChunkForm base;
    
    ds_module_assert_default(ctx, DS_DATA_CHUNK_FORM_INVALID);
    ds_assert_default(type, DS_DATA_CHUNK_FORM_INVALID);
    
    if (ctx->supported_forms) {
	base = dsDataManagerRepGetFormByName(ctx, DS_DATA_MANAGER_FORM_NATIVE);
	ds_assert_default(base != DS_DATA_CHUNK_FORM_INVALID, DS_DATA_CHUNK_FORM_INVALID);
    } else base = DS_DATA_CHUNK_FORM_MISSING;
    
    return dsDataManagerRepRegisterBinaryRepresentation(ctx, base, DS_DATA_MANAGER_FORM_OPCXML, dsBinaryFormatOPCXML(), type, NULL);
}

DSAllocatorContext dsDataManagerRepGetAllocator(DSDataManagerRepContext ctx, DSDataChunkForm repid) {
    DSRepresentationInstance rctx;
    DSAllocatorContext allocator;

    ds_module_assert_default(ctx, NULL);
    ds_assert_null(repid<ctx->supported_forms);

    if ((ctx->representation[repid].flags|DS_DATA_MANAGER_REP_REPRESENTATION_BINARY)==0) {
	dsModuleDebug(ctx, DS_DEBUG_WARNING, "GetAllocator request for non-binary representation");
	return NULL;
    }
	
    dsLockRead(ctx->representation_lock);
    rctx = ctx->representation[repid].reference;
    dsLockReturn(ctx->representation_lock);
        
    if (!rctx) return NULL;
    dsLockWrite(ctx->writer_lock);

	/* Checks if the allocator is binary are inside */
    allocator = dsRepresentationBinaryGetAllocator(rctx);
    if (allocator) dsDataManagerRef(DS_DATA_MANAGER(ctx));
    else dsLockReturn(ctx->writer_lock);
    
    return allocator;
}

int dsDataManagerRepReturnAllocator(DSDataManagerRepContext ctx, DSDataChunkForm repid, DSAllocatorContext allocator) {
    int err;
    DSRepresentationInstance rctx;

    ds_module_assert_return(ctx);
    ds_assert_return(repid<ctx->supported_forms);

    dsLockRead(ctx->representation_lock);
    rctx = ctx->representation[repid].reference;
    dsLockReturn(ctx->representation_lock);

    err = dsRepresentationBinaryReturnAllocator(rctx, allocator);
    if (err) return err;
    
    dsDataManagerUnref(DS_DATA_MANAGER(ctx));
    dsLockReturn(ctx->writer_lock);

    return 0;
}

int dsDataManagerRepPushData(DSDataManagerRepContext ctx, DSDataChunkForm repid, DSDataConst info, DSDataConst data) {
    int err;
    
    DSRepresentationInstance rctx;

    DSDataChunkContext chunk;

    ds_module_assert(ctx);
    ds_assert_param(data);
    ds_assert_param(repid<ctx->supported_forms);

    dsLockRead(ctx->representation_lock);
    rctx = ctx->representation[repid].reference;
    dsLockReturn(ctx->representation_lock);
    
    if (!rctx) return dsModuleError(ctx, DS_ERR_DISABLED_ITEM);

    chunk = dsChunkPoolGetChunk(DS_DATA_MANAGER(ctx)->chunk_pool, NULL, ctx->supported_forms);
    if (!chunk) return DS_ERR_MEM;
    
    err = dsDataChunkSetInfo(chunk, info);

    if (!err) {
	dsLockWrite(ctx->writer_lock);
	err = dsRepresentationPushData(rctx, chunk, data);
	if (!err) err = dsDataManagerRegisterChunk(DS_DATA_MANAGER(ctx), chunk);
	    // ToDo: prelock datachunk for all subscriptions may need it (delayed_ref)
	dsLockReturn(ctx->writer_lock);
    }
    
    if (err) dsChunkPoolReturnChunk(DS_DATA_MANAGER(ctx)->chunk_pool, chunk);

    return err;
}

int dsDataManagerRepGetData(DSDataManagerRepContext ctx, DSDataChunkForm repid, DSDataConst *info, DSDataConst *data) {
    int err;
    DSRepresentationInstance rctx;

    DSDataChunkContext current;
    
    ds_module_assert(ctx);
    ds_assert_param(repid<ctx->supported_forms);
    
    if (!DS_DATA_MANAGER(ctx)->current) {
	dsModuleDebug(ctx, DS_DEBUG_NOTICE, "Empty. There is no data pushed still");
	return dsModuleError(ctx, DS_ERR_EMPTY);
    }

    dsLockRead(DS_DATA_MANAGER(ctx)->chunk_lock);
    current =  DS_DATA_MANAGER(ctx)->current;
    dsDataChunkRef(current);
    dsLockReturn(DS_DATA_MANAGER(ctx)->chunk_lock);

    dsLockRead(ctx->representation_lock);
    rctx = ctx->representation[repid].reference;

    err = dsDataConverterConvert(ctx->converter, rctx, current);
    dsLockReturn(ctx->representation_lock);
    
    if (err) {
	dsModuleDebug(ctx, DS_DEBUG_NOTICE, "Data conversion is failed with error: %X(%s)", err, dsError(err, NULL));
	dsDataChunkUnref(current);
	return err;
    }
    
    if (data) *data=dsRepresentationGet(rctx, current);
    if (info) *info=dsDataChunkGetInfo(current);

/*
    {
	char *buf;
	size_t size;
	dsXmlDumpMemory((*data)->data, &buf, &size);
	puts(buf);
	free(buf);
    }
*/

    dsDataManagerRef((DSDataManagerContext)ctx);

    return 0;
}


int dsDataManagerRepMantain(DSDataManagerRepContext ctx) {
    int err;
    
    DSDataChunkForm i;
    DSDataManagerRepRepresentationData representation;

    ds_module_assert(ctx);
    
    dsLockRead(ctx->representation_lock);
    representation = ctx->representation;
    for (i=0; i<ctx->supported_forms; i++) {
	if (representation[i].reference) {
	    err = dsRepresentationMantain(representation[i].reference);
	    if (err) {
		dsModuleDebug(ctx, DS_DEBUG_ERROR, "Representation \"%s\" maintenance is failed with error: %X(%s)", representation[i].title, err, dsError(err, NULL));
	    }
	}
    }
    dsLockReturn(ctx->representation_lock);
    return 0;
}

DSDataChunkForm dsDataManagerRepGetFormByName(DSDataManagerRepContext ctx, const char *repname) {
    DSDataManagerRepRepresentationData rep;

    ds_module_assert_default(ctx, DS_DATA_CHUNK_FORM_INVALID);    
    ds_assert_default(repname, DS_DATA_CHUNK_FORM_INVALID);
    
    dsLockRead(ctx->representation_lock);
    rep = (DSDataManagerRepRepresentationData)dsHashTableFind(ctx->rephash, (void*)repname);
    dsLockReturn(ctx->representation_lock);
    
    if (!rep) return DS_DATA_CHUNK_FORM_INVALID;

    return rep->formid;
}

int dsDataManagerRepSetUserizeFunction(DSDataManagerRepContext ctx, DSDataChunkForm repid, DSDataManagerGetUserDataFunction get, void *userizer) {
    ds_module_assert(ctx);    
    ds_assert_param(repid<ctx->supported_forms);

    ctx->representation[repid].handler.userizer = userizer;
    ctx->representation[repid].handler.GetUserData = get;

    return 0;
}
