#define _DS_INCLUDE_XML_HEADERS

#include <stdio.h>
#include <dsassert.h>

#include <libxml/schemasInternals.h>

#include <dslib.h>
#include <dsdebug.h>

#include <dsopcbinarydictionary.h>
#include <dsopcbinary.h>

void dsOPCBinaryZeroContext(DSOPCBinaryContext ctx) {
    ds_assert_return(ctx);
    
    ctx->lock = NULL;
    ctx->dictdir = NULL;
    ctx->schema_rep = NULL;
}

DSOPCBinaryContext dsOPCBinaryCreateContext(DSMallocContext mctx) {
    DSOPCBinaryContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatDescriptionOPCBinaryCreateContext");

    ctx = (DSOPCBinaryContext)dsMalloc(mctx,sizeof(struct DSOPCBinaryContextT));
    if (ctx) {
	ctx->malloc=mctx;
	dsOPCBinaryZeroContext(ctx);
    }
    return ctx;
}

int dsOPCBinaryInitContext(DSOPCBinaryContext ctx, const char *schema, const char *dictdir) {
    int err;
    xmlSchemaParserCtxtPtr schema_parser;

    ds_assert_param(ctx);
    
    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatDescriptionOPCBinaryInitContext: %s", schema);

    dsOPCBinaryZeroContext(ctx);

    if (dsExchangeString(ctx->malloc, &ctx->dictdir, dictdir)) return DS_ERR_MEM;
    ctx->dictdir_chars = strlen(ctx->dictdir);

    schema_parser=xmlSchemaNewParserCtxt(schema);
    if (!schema_parser) {
	dsOPCBinaryFreeContext(ctx);
	return DS_ERR_XML_SCHEMA_CREATE;
    }
    ctx->schema_rep=xmlSchemaParse(schema_parser);
    xmlSchemaFreeParserCtxt(schema_parser);
    if (!ctx->schema_rep) {
	dsOPCBinaryFreeContext(ctx);
	return DS_ERR_XML_SCHEMA_CREATE;
    }
    
    ctx->lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	dsOPCBinaryFreeContext(ctx);
	return DS_ERR_MEM;
    }
    err = dsLockInitContext(ctx->lock);
    if (err) {
	dsOPCBinaryFreeContext(ctx);
	return err;
    }


//    xmlSchemaSetValidErrors(ctx->schema_ctx,NULL,NULL,NULL);

    return 0;        
}

void dsOPCBinaryFreeContext(DSOPCBinaryContext ctx) {
    ds_assert_return(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatDescriptionOPCBinaryFreeContext");

    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    if (ctx->dictdir) dsExchangeString(ctx->malloc, &ctx->dictdir, NULL);
    if (ctx->schema_rep) xmlSchemaFree(ctx->schema_rep);
    
    dsOPCBinaryZeroContext(ctx);
}

void dsOPCBinaryDestroyContext(DSOPCBinaryContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatDescriptionOPCBinaryDestroyContext");
    dsOPCBinaryFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

int dsOPCBinaryValidateXML(DSOPCBinaryContext ctx, xmlDocPtr doc) {
    return 0;
}

int dsOPCBinaryValidateDictionary(DSOPCBinaryContext ctx, DSOPCBinaryDictionaryContext dctx) {
    dsLockWrite(ctx->lock);
    dsLockReturn(ctx->lock);
    return 0;
}

DSOPCBinaryDictionaryContext dsOPCBinaryCreateDictionaryFromMemory(DSOPCBinaryContext ctx, const char *xml, unsigned long length) {
    int err;
    xmlDocPtr doc;
    xmlSchemaValidCtxtPtr schema_ctx;
    DSOPCBinaryDictionaryContext dctx;
    
    ds_assert_null(ctx);
    ds_assert_null(xml);
    ds_assert_null(length>0);
    ds_assert_null(ctx->schema_rep);

    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatDescriptionOPCBinaryCreateDictionaryFromMemory");

    if (!length) length=xmlStrlen(xml);
    doc = xmlReadMemory(xml,length,"xml",NULL,0);
    if (!doc) return NULL;
    
    dsLockWrite(ctx->lock);
    schema_ctx = xmlSchemaNewValidCtxt(ctx->schema_rep);
    dsLockReturn(ctx->lock);

    if (!schema_ctx) {
	xmlFreeDoc(doc);
	return NULL;
    }

    xmlSchemaSetValidErrors(schema_ctx,NULL,NULL,NULL);
    err = xmlSchemaValidateDoc(schema_ctx,doc);
    xmlSchemaFreeValidCtxt(schema_ctx);
    
    if (err) {
	xmlFreeDoc(doc);
	return NULL;
    }

    dctx=dsOPCBinaryDictionaryCreateContext(ctx->malloc);
    if (!dctx) return dctx;
    
    err = dsOPCBinaryDictionaryInitContext(dctx, doc, ctx->schema_rep->targetNamespace);
    if (err) {
	xmlFreeDoc(doc);
	dsOPCBinaryDictionaryDestroyContext(dctx);
	return NULL;
    }

    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatDescriptionOPCBinaryCreateDictionary: Ok. Created %p", dctx);

    return dctx;    
}

DSOPCBinaryDictionaryContext dsOPCBinaryCreateDictionary(DSOPCBinaryContext ctx, const char *dictionary) {
    int err;
    char *stmp;
    xmlDocPtr doc;
    xmlSchemaValidCtxtPtr schema_ctx;
    DSOPCBinaryDictionaryContext dctx;
    
    ds_assert_null(ctx);
    ds_assert_null(dictionary);
    ds_assert_null(ctx->schema_rep);

    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatDescriptionOPCBinaryCreateDictionary: %s", dictionary);

    if (dictionary[0]=='/') doc = xmlReadFile(dictionary,NULL,0);
    else {
	stmp = (char*)dsMalloca(ctx->dictdir_chars+strlen(dictionary)+2);
	if (!stmp) return NULL;
	sprintf(stmp,"%s/%s",ctx->dictdir,dictionary);
	doc = xmlReadFile(stmp,NULL,0);
	dsFreea(stmp);
    }
    if (!doc) return NULL;

    
    dsLockWrite(ctx->lock);
    schema_ctx = xmlSchemaNewValidCtxt(ctx->schema_rep);
    dsLockReturn(ctx->lock);

    if (!schema_ctx) {
	xmlFreeDoc(doc);
	return NULL;
    }

    xmlSchemaSetValidErrors(schema_ctx,NULL,NULL,NULL);
    err = xmlSchemaValidateDoc(schema_ctx,doc);
    xmlSchemaFreeValidCtxt(schema_ctx);
    
    if (err) {
	xmlFreeDoc(doc);
	return NULL;
    }

    dctx=dsOPCBinaryDictionaryCreateContext(ctx->malloc);
    if (!dctx) return dctx;
    
    err = dsOPCBinaryDictionaryInitContext(dctx, doc, ctx->schema_rep->targetNamespace);
    if (err) {
	xmlFreeDoc(doc);
	dsOPCBinaryDictionaryDestroyContext(dctx);
	return NULL;
    }

    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatDescriptionOPCBinaryCreateDictionary: Ok. Created %p", dctx);

    return dctx;    
}
