#define _DS_INCLUDE_XML_HEADERS
#define _DS_OPCBINARY_DICTIONARY_C
#include <dslib.h>
#include <dsassert.h>

#include <dsdebug.h>
#include <dsxml.h>

#include <dsbinaryconfiguration.h>
#include <dsbinarytype.h>

#include <dsopcbinarydictionary.h>

void dsOPCBinaryDictionaryZeroContext(DSOPCBinaryDictionaryContext ctx) {
    ds_assert_return(ctx);

    ctx->xpath_ctx = NULL;
    ctx->xml_doc = NULL;
    ctx->lock = NULL;
    ctx->types = NULL;
    ctx->default_string_encoding = NULL;
    ctx->default_float_format = NULL;
}

DSOPCBinaryDictionaryContext dsOPCBinaryDictionaryCreateContext(DSMallocContext mctx) {
    DSOPCBinaryDictionaryContext ctx;

    ctx = (DSOPCBinaryDictionaryContext)dsMalloc(mctx,sizeof(struct DSOPCBinaryDictionaryContextT));
    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatDescriptionOPCBinaryDictionary %p: CreateContext", ctx);

    if (ctx) {
	ctx->malloc=mctx;
	dsOPCBinaryDictionaryZeroContext(ctx);
    }
    return ctx;
}

int dsOPCBinaryDictionaryInitContext(DSOPCBinaryDictionaryContext ctx, xmlDocPtr doc, const char *opcbinary_namespace) {
    int err;
    xmlNodePtr node;
    xmlAttrPtr attr;
    
    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatDescriptionOPCBinaryDictionary %p: InitContext", ctx);

    ds_assert_param(ctx);
    ds_assert_param(doc);

    dsOPCBinaryDictionaryZeroContext(ctx);

    ctx->lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	dsOPCBinaryDictionaryFreeContext(ctx);
	return DS_ERR_MEM;
    }
    err = dsLockInitContext(ctx->lock);
    if (err) {
	dsOPCBinaryDictionaryFreeContext(ctx);
	return err;
    }
    
    ctx->types = dsEnumerationCreateContext(ctx->malloc);
    if (!ctx->types) {
	dsOPCBinaryDictionaryFreeContext(ctx);
	return DS_ERR_MEM;
    }

    err = dsEnumerationInitContext(ctx->types, (DSEnumerationBucket)&ds_opcbinary_types_enumeration);
    if (err) {
	dsOPCBinaryDictionaryFreeContext(ctx);
	return err;
    }
	
    ctx->xml_doc = doc;

    ctx->xpath_ctx = xmlXPathNewContext(doc);
    if (!ctx->xpath_ctx) {
	dsOPCBinaryDictionaryFreeContext(ctx);
	return DS_ERR_XML_XPATH_CREATE;
    }

    if (xmlXPathRegisterNs(ctx->xpath_ctx, "OPCBinary", opcbinary_namespace)) {
	dsOPCBinaryDictionaryFreeContext(ctx);
	return DS_ERR_XML_XPATH_REGISTERNAMESPACE;
    }

    ctx->xsins = xmlSearchNsByHref(doc, xmlDocGetRootElement(doc), DS_XML_XSI_NAMESPACE);
    if (!ctx->xsins) {
	dsOPCBinaryDictionaryFreeContext(ctx);
	return DS_ERR_NOTFOUND;
    }

    node = xmlDocGetRootElement(doc);
    attr=xmlHasProp(node, "DefaultBigEndian");
    if ((attr)&&((!xmlStrcmp(attr->children->content,"false"))||(!xmlStrcmp(attr->children->content,"0"))))
	ctx->default_big_endian=0;
    else ctx->default_big_endian=1;
    attr=xmlHasProp(node, "DefaultCharWidth");
    if (attr) ctx->default_char_width = atoi(attr->children->content);
    else ctx->default_char_width = 2;
    attr=xmlHasProp(node, "DefaultStringEncoding");
    if (attr) dsExchangeString(ctx->malloc, &ctx->default_string_encoding, attr->children->content);
    else dsExchangeString(ctx->malloc, &ctx->default_string_encoding, "UCS-2");
    attr=xmlHasProp(node, "DefaultFloatFormat");
    if (attr) dsExchangeString(ctx->malloc, &ctx->default_float_format, attr->children->content);
    else dsExchangeString(ctx->malloc, &ctx->default_float_format, "IEEE-754");

    return 0;        
}

void dsOPCBinaryDictionaryFreeContext(DSOPCBinaryDictionaryContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatDescriptionOPCBinaryDictionary %p: FreeContext", ctx);

    ds_assert_return(ctx);

	// ToDo: Wait for lock
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    if (ctx->default_string_encoding) dsExchangeString(ctx->malloc,&ctx->default_string_encoding, NULL);
    if (ctx->default_float_format) dsExchangeString(ctx->malloc, &ctx->default_float_format, NULL);
    if (ctx->types) dsEnumerationDestroyContext(ctx->types);
    if (ctx->xpath_ctx) xmlXPathFreeContext(ctx->xpath_ctx);
    if (ctx->xml_doc) xmlFreeDoc(ctx->xml_doc);
    dsOPCBinaryDictionaryZeroContext(ctx);
}

void dsOPCBinaryDictionaryDestroyContext(DSOPCBinaryDictionaryContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatDescriptionOPCBinaryDictionary %p: DestroyContext", ctx);
    dsOPCBinaryDictionaryFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}


#define XPATH_LOCALIZATION_EXPRESSION "//OPCBinary:TypeDescription[@TypeID]"

int dsOPCBinaryDictionaryLocalise(DSOPCBinaryDictionaryContext ctx) {
    int err;
    DSOPCBinaryTypes field_type;
    xmlAttrPtr attr;
    xmlXPathObjectPtr obj;
    xmlNodePtr pnode, node;
    xmlNodeSetPtr node_set;
    unsigned long nnodes;
    const char *stmp;
    unsigned long i;
    
    ds_assert_param(ctx);
    
    puts("Localising");
    
    obj = xmlXPathEvalExpression(XPATH_LOCALIZATION_EXPRESSION, ctx->xpath_ctx);
    if (!obj) return 0;
    node_set = obj->nodesetval;
    if (!node_set) return 0;
    nnodes = node_set->nodeNr;
    
    for (i=0;i<nnodes;i++) {
	pnode = node_set->nodeTab[i];
        
	printf("Node: %s\n", dsXmlGetPropString(pnode, "TypeID", NULL));

	if (ds_binary_configuration->endianess)
	    xmlSetProp(pnode, "DefaultBigEndian", "true");
	else
	    xmlSetProp(pnode, "DefaultBigEndian", "false");
	
	
	for (node=pnode->children;node;node=node->next) {
	    attr=xmlHasNsProp(node, "type", ctx->xsins->href);
	    if (!attr) continue;
	    printf(" Type: %s\n", attr->children->content);

	    field_type = dsEnumerationGetID(ctx->types, attr->children->content);
	    switch(field_type) {
		case DS_OPCBINARY_INTEGER:
		break;
		case DS_OPCBINARY_FLOATINGPOINT:
		    stmp=dsXmlGetPropString(node, "Format", NULL);
		    if (stmp) break;

		    if (!xmlStrcmp(stmp,"float")) {
		        if (ds_binary_configuration->align_float>1)
			    dsXmlSetPropInt(node, "Alignment", ds_binary_configuration->align_float);
			else {
			    attr = xmlHasProp(node, "Alignment");
			    if (attr) xmlRemoveProp(attr);
			}
			if (ds_binary_configuration->float_format==FLOAT_FORMAT_IEEE) {
			    if (ds_binary_configuration->float_endianess == ds_binary_configuration->endianess)
				xmlSetProp(node, "FloatFormat", "IEEE-754");
			    else if (ds_binary_configuration->float_endianess==ENDIANESS_LITTLE_ENDIAN)
				xmlSetProp(node, "FloatFormat", "IEEE-754-LE");
			    else if (ds_binary_configuration->float_endianess==ENDIANESS_BIG_ENDIAN)
				xmlSetProp(node, "FloatFormat", "IEEE-754-BE");
			    else if (ds_binary_configuration->float_endianess==ENDIANESS_MIXED_ENDIAN)
				xmlSetProp(node, "FloatFormat", "IEEE-754-ME");
    			    else 
				return DS_ERR_NOTSUPPORTED;
			} else {
			    return DS_ERR_NOTSUPPORTED;
			}
			dsXmlSetPropInt(node, "Length", ds_binary_configuration->float_size);
			break;
		    }

		    if (!xmlStrcmp(stmp,"double")) {
		        if (ds_binary_configuration->align_float>1)
			    dsXmlSetPropInt(node, "Alignment", ds_binary_configuration->align_double);
			else {
			    attr = xmlHasProp(node, "Alignment");
			    if (attr) xmlRemoveProp(attr);
			}
			if (ds_binary_configuration->float_format==FLOAT_FORMAT_IEEE) {
			    if (ds_binary_configuration->float_endianess == ds_binary_configuration->endianess)
				xmlSetProp(node, "FloatFormat", "IEEE-754");
			    else if (ds_binary_configuration->float_endianess==ENDIANESS_LITTLE_ENDIAN)
				xmlSetProp(node, "FloatFormat", "IEEE-754-LE");
			    else if (ds_binary_configuration->float_endianess==ENDIANESS_BIG_ENDIAN)
				xmlSetProp(node, "FloatFormat", "IEEE-754-BE");
			    else if (ds_binary_configuration->float_endianess==ENDIANESS_MIXED_ENDIAN)
				xmlSetProp(node, "FloatFormat", "IEEE-754-ME");
    			    else 
				return DS_ERR_NOTSUPPORTED;
			} else {
			    return DS_ERR_NOTSUPPORTED;
			}
			dsXmlSetPropInt(node, "Length", ds_binary_configuration->double_size);
			break;
		    }
		break;
		case DS_OPCBINARY_CHARSTRING:
		break;
		case DS_OPCBINARY_TYPEREFERENCE:
		break;
		case DS_OPCBINARY_INT8:
		    if (ds_binary_configuration->align_int8>1)
			dsXmlSetPropInt(node, "Alignment", ds_binary_configuration->align_int8);
		    else {
			attr = xmlHasProp(node, "Alignment");
			if (attr) xmlRemoveProp(attr);
		    }
		break;
		case DS_OPCBINARY_INT16:
		    if (ds_binary_configuration->align_int16>1)
			dsXmlSetPropInt(node, "Alignment", ds_binary_configuration->align_int16);
		    else {
			attr = xmlHasProp(node, "Alignment");
			if (attr) xmlRemoveProp(attr);
		    }
		break;
		case DS_OPCBINARY_INT32:
		    if (ds_binary_configuration->align_int32>1)
			dsXmlSetPropInt(node, "Alignment", ds_binary_configuration->align_int32);
		    else {
			attr = xmlHasProp(node, "Alignment");
			if (attr) xmlRemoveProp(attr);
		    }
		break;
		case DS_OPCBINARY_INT64:
		    if (ds_binary_configuration->align_int64>1)
			dsXmlSetPropInt(node, "Alignment", ds_binary_configuration->align_int64);
		    else {
			attr = xmlHasProp(node, "Alignment");
			if (attr) xmlRemoveProp(attr);
		    }
		break;
		case DS_OPCBINARY_UINT8:
		    if (ds_binary_configuration->align_int8>1)
			dsXmlSetPropInt(node, "Alignment", ds_binary_configuration->align_int8);
		    else {
			attr = xmlHasProp(node, "Alignment");
			if (attr) xmlRemoveProp(attr);
		    }
		break;
		case DS_OPCBINARY_UINT16:
		    if (ds_binary_configuration->align_int16>1)
			dsXmlSetPropInt(node, "Alignment", ds_binary_configuration->align_int16);
		    else {
			attr = xmlHasProp(node, "Alignment");
			if (attr) xmlRemoveProp(attr);
		    }
		break;
		case DS_OPCBINARY_UINT32:
		    if (ds_binary_configuration->align_int32>1)
			dsXmlSetPropInt(node, "Alignment", ds_binary_configuration->align_int32);
		    else {
			attr = xmlHasProp(node, "Alignment");
			if (attr) xmlRemoveProp(attr);
		    }
		break;
		case DS_OPCBINARY_UINT64:
		    if (ds_binary_configuration->align_int64>1)
			dsXmlSetPropInt(node, "Alignment", ds_binary_configuration->align_int64);
		    else {
			attr = xmlHasProp(node, "Alignment");
			if (attr) xmlRemoveProp(attr);
		    }
		break;
		case DS_OPCBINARY_SINGLE:
		    if (ds_binary_configuration->align_float>1)
			dsXmlSetPropInt(node, "Alignment", ds_binary_configuration->align_float);
		    else {
			attr = xmlHasProp(node, "Alignment");
			if (attr) xmlRemoveProp(attr);
		    }
		    if (ds_binary_configuration->float_format==FLOAT_FORMAT_IEEE) {
			if (ds_binary_configuration->float_endianess == ds_binary_configuration->endianess) {
			    if (ds_binary_configuration->float_size != 4) {
				xmlNodeSetName(node, "FloatingPoint");
				xmlSetProp(node, "FloatFormat", "IEEE-754");
				xmlSetProp(node, "Format", "float");
				dsXmlSetPropInt(node, "Length", ds_binary_configuration->float_size);
			    }
			} else {
			    xmlNodeSetName(node, "FloatingPoint");
			    if (ds_binary_configuration->float_endianess==ENDIANESS_LITTLE_ENDIAN)
				xmlSetProp(node, "FloatFormat", "IEEE-754-LE");
			    else if (ds_binary_configuration->float_endianess==ENDIANESS_BIG_ENDIAN)
				xmlSetProp(node, "FloatFormat", "IEEE-754-BE");
			    else if (ds_binary_configuration->float_endianess==ENDIANESS_MIXED_ENDIAN)
				xmlSetProp(node, "FloatFormat", "IEEE-754-ME");
			    else 
				return DS_ERR_NOTSUPPORTED;
			    xmlSetProp(node, "Format", "float");
			    dsXmlSetPropInt(node, "Length", ds_binary_configuration->float_size);
			}
		    } else {
			return DS_ERR_NOTSUPPORTED;
		    }
		break;
		case DS_OPCBINARY_DOUBLE:
		    if (ds_binary_configuration->align_double>1)
			dsXmlSetPropInt(node, "Alignment", ds_binary_configuration->align_double);
		    else {
			attr = xmlHasProp(node, "Alignment");
			if (attr) xmlRemoveProp(attr);
		    }
		    if (ds_binary_configuration->float_format==FLOAT_FORMAT_IEEE) {
			if (ds_binary_configuration->float_endianess == ds_binary_configuration->endianess) {
			    if (ds_binary_configuration->double_size != 8) {
				xmlNodeSetName(node, "FloatingPoint");
				xmlSetProp(node, "FloatFormat", "IEEE-754");
				xmlSetProp(node, "Format", "double");
				dsXmlSetPropInt(node, "Length", ds_binary_configuration->double_size);
			    }
			} else {
			    xmlNodeSetName(node, "FloatingPoint");
			    if (ds_binary_configuration->float_endianess==ENDIANESS_LITTLE_ENDIAN)
				xmlSetProp(node, "FloatFormat", "IEEE-754-LE");
			    else if (ds_binary_configuration->float_endianess==ENDIANESS_BIG_ENDIAN)
				xmlSetProp(node, "FloatFormat", "IEEE-754-BE");
			    else if (ds_binary_configuration->float_endianess==ENDIANESS_MIXED_ENDIAN)
				xmlSetProp(node, "FloatFormat", "IEEE-754-ME");
			    else 
				return DS_ERR_NOTSUPPORTED;
			    xmlSetProp(node, "Format", "double");
			    dsXmlSetPropInt(node, "Length", ds_binary_configuration->double_size);
			}
		    } else {
			return DS_ERR_NOTSUPPORTED;
		    }
		break;
		case DS_OPCBINARY_ASCII:
		    err=1;
		break;
		case DS_OPCBINARY_UNICODE:
		    err=1;
		break;
		default:
		    err=1;
		break;
	    
	    }
	}	
    }
    
    xmlXPathFreeObject(obj);
    return 0;    
}


#define XPATH_EXPRESSION "//OPCBinary:TypeDescription[@TypeID=\"%s\"]"

inline_function static void process_node_basics(DSBinaryType ctype, xmlNodePtr node) {
    xmlAttrPtr attr;

    attr=xmlHasProp(node, "ElementCount");
    if (attr) dsBinaryTypeSetElementCount(ctype, atol(attr->children->content));
//    attr=xmlHasProp(node, "ElementCountRef");
//    if (attr) dsBinaryTypeSetElementCount(ctype, ref);
    attr=xmlHasProp(node, "FieldTerminator");
    if (attr) dsBinaryTypeSetTerminator(ctype, attr->children->content);
    
    dsBinaryTypeSetAlignment(ctype, dsXmlGetPropInt(node, "Alignment", 1));
}

DSBinaryType dsOPCBinaryDictionaryCreateType(DSOPCBinaryDictionaryContext ctx, const char *type) {
    int err = 0;
    char *xpath;
    unsigned long nnodes;
    DSMallocContext mctx;
    xmlNodePtr pnode, node;
    xmlNodeSetPtr node_set;
    xmlXPathObjectPtr obj;
    xmlAttrPtr attr;
    DSBinaryType ptype;
    DSBinaryType ctype;
    DSOPCBinaryTypes field_type;
    unsigned char big_endian;
    unsigned char char_width;
    char *string_encoding;
    char *float_format;
    
    unsigned long size;
    
    ds_assert_null(ctx);
    ds_assert_null(ctx->xpath_ctx);
    ds_assert_null(type);
    
    dsDebugServer(DS_DEBUG_NOTICE, "dsBinaryFormatDescriptionOPCBinaryDictionaryCreateType: %s", type);
    xpath = (char*)malloc((sizeof(XPATH_EXPRESSION)+strlen(type))*sizeof(char));
    if (!xpath) return NULL;
    sprintf(xpath, XPATH_EXPRESSION, type);
    
    dsLockWrite(ctx->lock);
    obj = xmlXPathEvalExpression(xpath,ctx->xpath_ctx);
    dsLockReturn(ctx->lock);
    
    if (!obj) return NULL;
    node_set = obj->nodesetval;
    if (!node_set) {
	xmlXPathFreeObject(obj);
	return NULL;
    }
    nnodes = node_set->nodeNr;
    if (nnodes!=1) {
        xmlXPathFreeObject(obj);
	return NULL;
    }
    pnode = node_set->nodeTab[0];
    if (pnode->ns) printf("Node: %s:%s\n",pnode->ns->href,pnode->name);
    else printf("Node: %s\n",pnode->name);
    
    mctx = ctx->malloc;
    attr=xmlHasProp(pnode, "DefaultBigEndian");
    if (attr) {
	if ((!xmlStrcmp(attr->children->content,"false"))||(!xmlStrcmp(attr->children->content,"0")))
	    big_endian = 0;
	else
	    big_endian = 1;
    } else big_endian=ctx->default_big_endian;
    attr=xmlHasProp(pnode, "DefaultCharWidth");
    if (attr) char_width = atoi(attr->children->content);
    else char_width=ctx->default_char_width;
    attr=xmlHasProp(pnode, "DefaultStringEncoding");
    if (attr) string_encoding = attr->children->content;
    else string_encoding = ctx->default_string_encoding;
    attr=xmlHasProp(pnode, "DefaultFloatFormat");
    if (attr) float_format = attr->children->content;
    else float_format = ctx->default_float_format;

    for (nnodes=0,node=pnode->children;(node)&&(nnodes<2);node=node->next)
	if ((node->type==XML_ELEMENT_NODE)&&(!xmlStrcmp(node->name, "Field"))) nnodes++;

    if (node) {
	ptype = (DSBinaryType)dsBinaryTypeComplexCreate(ctx->malloc);
	if (!ptype) {
	    xmlXPathFreeObject(obj);
	    return NULL;
	}
	dsBinaryTypeSetMaster(ptype);
    } else ptype = NULL;

    size=0;
    for (node=pnode->children;node;node=node->next) {
	if ((node->type==XML_ELEMENT_NODE)&&(!xmlStrcmp(node->name, "Field"))) {
	    attr=xmlHasNsProp(node, "type", ctx->xsins->href);
	    if ((err)||(!attr)) {
		xmlXPathFreeObject(obj);
		break;
	    }
	    field_type = dsEnumerationGetID(ctx->types, attr->children->content);
	    printf("New Node: %s(%u)\n",node->name,field_type);
	    
	    ctype = NULL;
	    switch(field_type) {
		case DS_OPCBINARY_INTEGER:
		    err=1;
		break;
		case DS_OPCBINARY_FLOATINGPOINT:
		    err=1;
		break;
		case DS_OPCBINARY_CHARSTRING:
		    err=1;
		break;
		case DS_OPCBINARY_TYPEREFERENCE:
		    err=1;
		break;
		case DS_OPCBINARY_INT8:
		    ctype = (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 1, 1, big_endian);
		    if (ctype) {
			process_node_basics(ctype, node);
			if (ptype) dsBinaryTypePush(ptype,ctype);
			else dsBinaryTypeSetMaster(ctype);
		    } else err=2;
		break;
		case DS_OPCBINARY_INT16:
		    ctype = (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 1, 2, big_endian);
		    if (ctype) {
			process_node_basics(ctype, node);
			if (ptype) dsBinaryTypePush(ptype,ctype);
			else dsBinaryTypeSetMaster(ctype);
		    } else err=2;
		break;
		case DS_OPCBINARY_INT32:
		    ctype = (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 1, 4, big_endian);
		    if (ctype) {
			process_node_basics(ctype, node);
			if (ptype) dsBinaryTypePush(ptype,ctype);
			else dsBinaryTypeSetMaster(ctype);
		    } else err=2;
		break;
		case DS_OPCBINARY_INT64:
		    ctype = (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 1, 8, big_endian);
		    if (ctype) {
			process_node_basics(ctype, node);
			if (ptype) dsBinaryTypePush(ptype,ctype);
			else dsBinaryTypeSetMaster(ctype);
		    } else err=2;
		break;
		case DS_OPCBINARY_UINT8:
		    ctype = (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 0, 1, big_endian);
		    if (ctype) {
			process_node_basics(ctype, node);
			if (ptype) dsBinaryTypePush(ptype,ctype);
			else dsBinaryTypeSetMaster(ctype);
		    } else err=2;
		break;
		case DS_OPCBINARY_UINT16:
		    ctype = (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 0, 2, big_endian);
		    if (ctype) {
			process_node_basics(ctype, node);
			if (ptype) dsBinaryTypePush(ptype,ctype);
			else dsBinaryTypeSetMaster(ctype);
		    } else err=2;
		break;
		case DS_OPCBINARY_UINT32:
		    ctype = (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 0, 4, big_endian);
		    if (ctype) {
			process_node_basics(ctype, node);
			if (ptype) dsBinaryTypePush(ptype,ctype);
			else dsBinaryTypeSetMaster(ctype);
		    } else err=2;
		break;
		case DS_OPCBINARY_UINT64:
		    ctype = (DSBinaryType)dsBinaryTypeIntegerCreate(mctx, 0, 8, big_endian);
		    if (ctype) {
			process_node_basics(ctype, node);
			if (ptype) dsBinaryTypePush(ptype,ctype);
			else dsBinaryTypeSetMaster(ctype);
		    } else err=2;
		break;
		case DS_OPCBINARY_SINGLE:
		    ctype = (DSBinaryType)dsBinaryTypeFloatCreate(mctx, DS_FLOAT_IEEE, 4, big_endian);
		    if (ctype) {
			process_node_basics(ctype, node);
			if (ptype) dsBinaryTypePush(ptype,ctype);
			else dsBinaryTypeSetMaster(ctype);
		    } else err=2;
		break;
		case DS_OPCBINARY_DOUBLE:
		    ctype = (DSBinaryType)dsBinaryTypeFloatCreate(mctx, DS_FLOAT_IEEE, 8, big_endian);
		    if (ctype) {
			process_node_basics(ctype, node);
			if (ptype) dsBinaryTypePush(ptype,ctype);
			else dsBinaryTypeSetMaster(ctype);
		    } else err=2;
		break;
		case DS_OPCBINARY_ASCII:
		    err=1;
		break;
		case DS_OPCBINARY_UNICODE:
		    err=1;
		break;
		default:
		    err=1;
		break;
	    }
	    if ((ptype)&&(ctype)&&(size!=(unsigned long)-1)) {
		printf("SubSize: %u\n",ctype->size);
		printf("Alignment: %u\n",ctype->align);
		if (ctype->size) {
    		    if ((ctype->align>1)&&(size&(ctype->align-1))) 
			size=(size&~(ctype->align-1))+ctype->align;
		    size+=ctype->size;
		}
		else size=(unsigned long)-1;
	    }
	}
    }

    xmlXPathFreeObject(obj);
    
    if (err) {
	if ((ptype)||(ctype)) dsBinaryTypeDestroy(ptype?ptype:ctype);
	return NULL;
    }

    if (ptype) {
	if (size) ptype->size=size;
	printf("UltimateSize: %lu\n",ptype->size);
    }
    
    return ptype?ptype:ctype;
}
