#include <dslib.h>
#include <string.h>

#define _DS_DEBUG_C

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif /* HAVE_ERRNO_H */
#include <stdio.h>
#include <stdarg.h>

#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif /* HAVE_SYS_SOCKET_H */
#ifdef HAVE_SYS_UN_H
# include <sys/un.h>
#endif /* HAVE_SYS_UN_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */


#include <dsassert.h>
#include <dslock.h>
#include <dsdebug.h>
#include <dsbasethread.h>

#include <dserror.h>
#include <dsmodule.h>

#define MAX_DEBUG_PATH_NAME 255
#define _DEBUG_SERVER  (debug_file?debug_file:stderr)
#define _DEBUG_OPC  (debug_file?debug_file:stderr)

static DSDebugMethod debug_mode = DS_DEBUG_METHOD_STDERR;

static DSDebugPriority debug_level = DS_DEBUG_WARNING;
static DSLockContext debug_lock = NULL;

static char debug_filename[MAX_DEBUG_PATH_NAME + 1] = {0};
static FILE *debug_file = 0;
static unsigned int options_fixed = 0;
static unsigned char logtime = 1;
static unsigned char loglevel = 0;

#ifndef _WIN32
static FILE *debug_connection = NULL;
#endif /* _WIN32 */

int dsDebugInit() {
    int err;
    
    debug_lock = dsLockCreateContext(dsMallocGetDefault());
    if (!debug_lock) return DS_ERR_LOCK_CREATE;
    
    err = dsLockInitContext(debug_lock);
    if (err) {
	dsLockDestroyContext(debug_lock);
	debug_lock = NULL;
	return err;
    }
    
    return 0;
}

int dsDebugSetLevel(DSDebugPriority priority) {
    debug_level = priority;
    return 0;
}

int dsDebugCheckLevel(DSDebugPriority priority) {
    if (debug_level<priority) return 0;
    return 1;
}

int dsDebugGetLevel() {
    return debug_level;
}

int dsDebugSetMethod(DSDebugMethod method, const void *attr) {
    FILE *f;

# ifndef _WIN32
    if (debug_connection) {
	fclose(debug_connection);
	debug_connection = NULL;
    }
# endif /* _WIN32 */

    debug_mode = DS_DEBUG_METHOD_STDERR;
    debug_filename[0] = 0;
        
    if (method == DS_DEBUG_METHOD_FILE) {
	if (!attr) return DS_ERR_INVALID_ARGUMENT;
	
	if ((!strcasecmp((char*)attr,"stdout"))||(!strcasecmp((char*)attr,"stderr"))) method = DS_DEBUG_METHOD_STDERR;
# ifndef _WIN32
	if (!strncasecmp((char*)attr,"socket://",9))   {
	    method = DS_DEBUG_METHOD_SOCKET;
	    attr=(void*)((char*)attr+9);
	}
# endif /* _WIN32 */
	else if (strstr((char*)attr, "://")) return DS_ERR_NOTSUPPORTED;
    }

# ifndef _WIN32
    if (method == DS_DEBUG_METHOD_FILE) {
	if (!attr) return DS_ERR_INVALID_ARGUMENT;

	if (!strncasecmp((char*)attr,"socket://",9)) {
	    attr=(void*)((char*)attr+9);
	}
	else if (strstr((char*)attr, "://")) return DS_ERR_NOTSUPPORTED;
    }
# endif /* _WIN32 */

    switch (method) {
	case DS_DEBUG_METHOD_STDERR:
	    debug_filename[0]=0;
	    debug_file = NULL;
	break;
	case DS_DEBUG_METHOD_FILE:
	    ds_assert_param(attr);
	    f = fopen(attr, "a+");
	    if (f) fclose(f);
	    else return DS_ERR_INVALID_FILE;
	    strncpy(debug_filename, (char*)attr, MAX_DEBUG_PATH_NAME);
	    debug_filename[MAX_DEBUG_PATH_NAME] = 0;
	break;
# ifndef _WIN32
	case DS_DEBUG_METHOD_SOCKET:
	    strncpy(debug_filename, (char*)attr, MAX_DEBUG_PATH_NAME);
	    debug_filename[MAX_DEBUG_PATH_NAME] = 0;
	    if (!(options_fixed&DS_DEBUG_LOG_TIME)) logtime = 1;
	    if (!(options_fixed&DS_DEBUG_LOG_LEVEL)) loglevel = 1;
	break;
# endif /* _WIN32 */
	default:
	    return DS_ERR_NOTSUPPORTED;
    }

    debug_mode = method;

    return 0;
}

int dsDebugSetOptions(unsigned int options) {
    options_fixed = (unsigned int)-1;
    logtime = (options&DS_DEBUG_LOG_TIME)?1:0;
    loglevel = (options&DS_DEBUG_LOG_LEVEL)?1:0;
    return 0;
}

int dsDebugSetLogTimeOption(unsigned char flag) {
    options_fixed|=DS_DEBUG_LOG_TIME;
    logtime = (flag)?1:0;
    
    return 0;
}

int dsDebugSetLogLevelOption(unsigned char flag) {
    options_fixed|=DS_DEBUG_LOG_LEVEL;
    loglevel = (flag)?1:0;
    
    return 0;
}

/*
static DSDebugPriority debug_limit = DS_DEBUG_DEBUG + 1;
int dsDebugLimitLevel(DSDebugPriority priority) {
    debug_limit = priority + 1;
    return 0;
}

DSDebugPriority dsDebugGetLimit() {
    return debug_limit - 1;
}
*/


# ifndef _WIN32
static int debug_socket = 0;

static void dsDebugCloseSocket() {
    if (debug_connection) {
	fclose(debug_connection);
	debug_connection = NULL;
        debug_socket = 0;
    }
}

static FILE *dsDebugOpenSocket(const char *filename) {
    int flags; 
    int sock;
    struct sockaddr_un mysock;
    FILE *res;
    int sndbufsize = 512 * 1024;
    struct timeval tv = {0, 100};
    fd_set readfs, writefs;
    

    if (debug_connection) {
	if (!fflush(debug_connection)) {
		// Checking connection is still alive
	    if (send(debug_socket," ",1,0)>0) return debug_connection;
	}

	dsDebugCloseSocket();
    }
    
    sock = socket(PF_UNIX, SOCK_STREAM, 0);
    if (!sock) return NULL;
    
    mysock.sun_family=AF_UNIX;
    strncpy(mysock.sun_path, filename, sizeof(mysock.sun_path)-1);
    
    flags=fcntl(sock,F_GETFL,0);
    if ((flags<0)||(fcntl(sock,F_SETFL,flags|O_NONBLOCK)<0)) {
	close(sock);
	return NULL;
    }

    if (connect(sock,(struct sockaddr*)&mysock,sizeof(mysock))) {
	FD_ZERO(&readfs);
	FD_SET(sock, &readfs);
	if ((select(sock+1, &readfs, NULL, NULL, &tv)<=0)||(connect(sock,(struct sockaddr*)&mysock,sizeof(mysock)))) {
	    close(sock);
	    return NULL;
	}
    }
    
//    setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &sndbufsize, sizeof(sndbufsize));

    res = fdopen(sock, "w");
    if (res) {
	debug_connection = res;
	debug_socket = sock;
    } else close(sock);
    
    return res;
}
# endif /* _WIN32 */

void dsDebugDestroy() {
# ifndef _WIN32
    if (debug_connection) dsDebugCloseSocket();
# endif /* _WIN32 */
    if (debug_lock) {
	dsLockDestroyContext(debug_lock);
	debug_lock = NULL;
    }
}


# ifdef _WIN32 
#  define _DEBUG_LOCK_OUTPUT_ \
    if (debug_lock) dsLockWrite(debug_lock); \
    if (debug_filename[0]) { \
	debug_file = fopen(debug_filename, "a+"); \
    }
#  define _DEBUG_UNLOCK_OUTPUT_ \
    if (debug_file) { fclose(debug_file); debug_file = NULL; } \
    if (debug_lock) dsLockReturn(debug_lock);
# else /* _WIN32 */
#  define _DEBUG_LOCK_OUTPUT_ \
    if (debug_lock) dsLockWrite(debug_lock); \
    if (debug_filename[0]) { \
        if (debug_mode == DS_DEBUG_METHOD_SOCKET) {\
	    debug_file = dsDebugOpenSocket(debug_filename); \
	    if (debug_file) fprintf(debug_file, "<DS_DEBUG>"); \
	} else \
	    debug_file = fopen(debug_filename, "a+"); \
    }
#  define _DEBUG_UNLOCK_OUTPUT_ \
    if (debug_file) { \
	if (debug_mode == DS_DEBUG_METHOD_SOCKET) { \
	    fprintf(debug_file, "</DS_DEBUG>"); \
	    if (fflush(debug_file)<0) dsDebugCloseSocket(); /* DS: We can loose event */ \
	} else fclose(debug_file); \
	debug_file = NULL; \
    } \
    if (debug_lock) dsLockReturn(debug_lock);
# endif /* _WIN32 */



static void dsDebugServerLogTime(DSDebugPriority priority) {
    DSTimeS curtime;
    char datestr[64];
    
    if (logtime) {
	dsTimeCurrentTime(&curtime);
	dsTimeGetDateTime(&curtime, datestr);
	fprintf(_DEBUG_SERVER, "%s ", datestr); 
    }
    
    if (loglevel) {
	fprintf(_DEBUG_SERVER, "%s ", ds_debug_priority[priority]); 
    }
}

static void dsDebugRequestLogTime(DSDebugPriority priority) {
    DSTimeS curtime;
    char datestr[64];

    if (logtime) {    
	dsTimeCurrentTime(&curtime);
	dsTimeGetDateTime(&curtime, datestr);
	fprintf(_DEBUG_OPC, "%s ", datestr); 
    }
    if (loglevel) {
	fprintf(_DEBUG_SERVER, "%s ", ds_debug_priority[priority]); 
    }
}


void dsDebugServerAssert(const char *format, va_list ap) {
    _DEBUG_LOCK_OUTPUT_
    dsDebugServerLogTime(DS_DEBUG_NOTHING);
    vfprintf(_DEBUG_SERVER, format, ap);
    _DEBUG_UNLOCK_OUTPUT_
}

enum DSDebugErrorTypeT {
    DS_DEBUG_ERROR_DS = 0,
    DS_DEBUG_ERROR_SYSTEM
};
typedef enum DSDebugErrorTypeT DSDebugErrorType;


static void dsDebugPrintError(FILE *output, int err, DSDebugErrorType type, const char *source, const char *message) {
    if (type == DS_DEBUG_ERROR_DS) {
	if (!source) source = dsErrorSource(err);
	if (!message) message = dsError(err, NULL);
    }

    if (source) {
	if (message) fprintf(output, "Error: 0x%X(Source: %s, Message: %s)", err, source, message);
	else fprintf(output, "Error: 0x%X(Source: %s)", err, source);
    } else {
	if (message) fprintf(output, "Error: 0x%X(%s)", err, message);
	else fprintf(output, "Error: 0x%X", err);
    }
}

void dsDebugModule(DSDebugPriority priority, const char *module, const void *ptr, const char *format, ...) {
#ifdef _DEBUG_SERVER
    va_list ap;

    if (debug_level<priority) return;

    va_start(ap, format);
    _DEBUG_LOCK_OUTPUT_
    dsDebugServerLogTime(priority);
    if (module) fprintf(_DEBUG_SERVER, "%s(%p): ", module, ptr);
    vfprintf(_DEBUG_SERVER,format,ap);
    fprintf(_DEBUG_SERVER,"\n");
    _DEBUG_UNLOCK_OUTPUT_
    va_end(ap);
#endif /* _DEBUG_SERVER */
}

void dsDebugModuleError(DSDebugPriority priority, const char *module, const void *ptr, int err, const char *format, ...) {
#ifdef _DEBUG_SERVER
    va_list ap;

    if (debug_level<priority) return;

    va_start(ap, format);
    _DEBUG_LOCK_OUTPUT_
    dsDebugServerLogTime(priority);
    if (module) fprintf(_DEBUG_SERVER, "%s(%p): ", module, ptr);
    vfprintf(_DEBUG_SERVER,format,ap);
    dsDebugPrintError(_DEBUG_SERVER, err, DS_DEBUG_ERROR_DS, NULL, NULL);
    fprintf(_DEBUG_SERVER,"\n");
    _DEBUG_UNLOCK_OUTPUT_
    va_end(ap);
#endif /* _DEBUG_SERVER */
}

void dsDebugServer(DSDebugPriority priority, const char *format, ...) {
#ifdef _DEBUG_SERVER
    va_list ap;

    if (debug_level<priority) return;

    va_start(ap, format);
    _DEBUG_LOCK_OUTPUT_
    dsDebugServerLogTime(priority);
    vfprintf(_DEBUG_SERVER,format,ap);
    fprintf(_DEBUG_SERVER,"\n");
    _DEBUG_UNLOCK_OUTPUT_
    va_end(ap);
#endif /* _DEBUG_SERVER */
}

void dsDebugServerError(DSDebugPriority priority, int err, const char *format, ...) {
#ifdef _DEBUG_SERVER
    va_list ap;

    if (debug_level<priority) return;

    va_start(ap, format);
    _DEBUG_LOCK_OUTPUT_
    dsDebugServerLogTime(priority);
    vfprintf(_DEBUG_SERVER,format,ap);
    dsDebugPrintError(_DEBUG_SERVER, err, DS_DEBUG_ERROR_DS, NULL, NULL);
    fprintf(_DEBUG_SERVER,"\n");
    _DEBUG_UNLOCK_OUTPUT_
    va_end(ap);
#endif /* _DEBUG_SERVER */
}

void dsDebugRequest(DSDebugPriority priority, const char *format, ...) {
#ifdef _DEBUG_OPC
    char stmp[DS_BASE_THREAD_ID_SIZE];
    va_list ap;

    if (debug_level<priority) return;

    va_start(ap, format);
    _DEBUG_LOCK_OUTPUT_
    dsDebugRequestLogTime(priority);
    fprintf(_DEBUG_OPC,"REQUEST %s:", dsThreadGetID(stmp));
    vfprintf(_DEBUG_OPC,format,ap);
    fprintf(_DEBUG_OPC,"\n");
    _DEBUG_UNLOCK_OUTPUT_
    va_end(ap);
#endif /* _DEBUG_OPC */
}

void dsDebugRequestError(DSDebugPriority priority, int err, const char *format, ...) {
#ifdef _DEBUG_OPC
    char stmp[DS_BASE_THREAD_ID_SIZE];
    va_list ap;
    
    if (debug_level<priority) return;

    va_start(ap, format);
    _DEBUG_LOCK_OUTPUT_
    dsDebugRequestLogTime(priority);
    fprintf(_DEBUG_OPC,"REQUEST %s:", dsThreadGetID(stmp));
    vfprintf(_DEBUG_OPC,format,ap);
    dsDebugPrintError(_DEBUG_SERVER, err, DS_DEBUG_ERROR_DS, NULL, NULL);
    fprintf(_DEBUG_OPC,"\n");
    _DEBUG_UNLOCK_OUTPUT_
    va_end(ap);
#endif /* _DEBUG_OPC */
}

void dsDebugServerSystemError(DSDebugPriority priority, int err, const char *prefix, ...) { 
#ifdef _WIN32
    char *lpMsgBuf = NULL;
    DWORD dw = err?err:GetLastError(); 
# ifndef PHARLAP
    size_t pos;
# endif
#endif /* _WIN32 */
    va_list ap;

    if (debug_level<priority) return;

    va_start(ap, prefix);
    vfprintf(_DEBUG_SERVER,prefix,ap);
    va_end(ap);

#ifdef _WIN32
# ifndef PHARLAP
    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );
    pos = strlen(lpMsgBuf) - 1;
    if (lpMsgBuf[pos]=='\n') lpMsgBuf[pos]=0;
# endif /* !PHARLAP */
    dsDebugServer(priority, "Error %d: %s", dw, lpMsgBuf?lpMsgBuf:""); 
#else /* _WIN32 */
    dsDebugServer(priority, "Error %d: %s", err?err:errno, strerror(err?err:errno));
#endif /* _WIN32 */

}

void dsDebugRequestSystemError(DSDebugPriority priority, int err, const char *prefix, ...) { 
#ifdef _WIN32
# ifndef PHARLAP
    size_t pos;
# endif
    char *lpMsgBuf = NULL;
    DWORD dw = err?err:GetLastError(); 
#endif /* _WIN32 */
    va_list ap;

    if (debug_level<priority) return;

    va_start(ap, prefix);
    vfprintf(_DEBUG_OPC,prefix,ap);
    va_end(ap);

#ifdef _WIN32
# ifndef PHARLAP
    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );
    
    pos = strlen(lpMsgBuf) - 1;
    if (lpMsgBuf[pos]=='\n') lpMsgBuf[pos]=0;
# endif /* ! PHARLAP */
    dsDebugRequest(priority, "Error %d: %s", dw, lpMsgBuf?lpMsgBuf:"");
#else /* _WIN32 */
    dsDebugRequest(priority, "Error %d: %s", err?err:errno, strerror(err?err:errno));
#endif /* _WIN32 */
}

DSDebugPriority dsDebugGetLevelByName(const char *level) {
    ds_assert_err(level, (DSDebugPriority)-1);
    
    if (!strcmp(level, "NOTHING")) return DS_DEBUG_NOTHING;
    if (!strcmp(level, "FATAL")) return DS_DEBUG_FATAL;
    if (!strcmp(level, "EMERGENCY")) return DS_DEBUG_EMERGENCY;
    if (!strcmp(level, "ALERT")) return DS_DEBUG_ALERT;
    if (!strcmp(level, "CRITICAL")) return DS_DEBUG_CRITICAL;
    if (!strcmp(level, "ERROR")) return DS_DEBUG_ERROR;
    if (!strcmp(level, "WARNING")) return DS_DEBUG_WARNING;
    if (!strcmp(level, "NOTICE")) return DS_DEBUG_NOTICE;
    if (!strcmp(level, "INFO")) return DS_DEBUG_INFO;
    if (!strcmp(level, "DEBUG")) return DS_DEBUG_DEBUG;
    
    if (!strcmp(level, "ALL")) return DS_DEBUG_DEBUG;
    return (DSDebugPriority)-1;
}

const char *dsDebugGetName(DSDebugPriority level) {
    if ((level>=0)&&(level<= DS_DEBUG_DEBUG)) return ds_debug_priority[level];
    return NULL;
}
