#define _DS_ERROR_C


#include <dslib.h>
#include <dsassert.h>
#include <dslocale.h>
#include <dshashtable.h>
#include <dsmalloc.h>
#include <dsdebug.h>

#include <dserror.h>


static DSMallocContext mctx = NULL;

static DSHashTable error_messages = NULL;

static unsigned long error_domains_allocated = 64;
static unsigned long error_domains_number = 0;
static char **error_domains = NULL;

static int dsAddErrorsDo(DSError errors, unsigned long errors_number) {
    unsigned long i;
    
    for (i=0;i<errors_number;i++) {
	((DSErrorS*)(errors+i))->domain = error_domains_number - 1;
	if (dsHashTableInsert(error_messages,(void*)(errors+i))) return DS_ERR_ERROR_INIT;
    }
    
    return 0;
}

int dsErrorNewDomain(const char *domain) {
    int err;
    char **new_domains;

    ds_assert_init(error_messages);
    ds_assert_param(domain);

    if (error_domains_number == error_domains_allocated) {
        new_domains = (char**)dsRealloc(mctx, &error_domains, error_domains_allocated*2*sizeof(char*));
        if (!new_domains) return DS_ERR_MEM;
        error_domains_allocated*=2;
	error_domains = new_domains;
    }
    err = dsExchangeString(mctx, error_domains + error_domains_number, domain);
    if (err) return DS_ERR_MEM;
    error_domains_number++;

    return 0;
}

/*
// Register errors in array for future deallocation
int dsErrorAddError(int code, const char *msg) {
}
*/

int dsErrorAddErrors(const char *domain, DSError errors, unsigned long errors_number) {
    int err;
    void *emsg;

    ds_assert_init(error_messages);
    ds_assert_param(errors);
    ds_assert_param(errors_number);

    if (domain) {
	err = dsErrorNewDomain(domain);
	if (err) return err;
    }

    err = dsHashTableGrow(error_messages, errors_number);
    if (err) return err;

    return dsAddErrorsDo(errors, errors_number);
}

int dsErrorInit() {
    int err;
    
    dsDebugServer(DS_DEBUG_NOTICE, "dsErrorInit");    

    if (mctx) return DS_ERR_INITIALIZED;
    mctx = dsMallocGetDefault();

    error_domains = (char**)dsCalloc(mctx, error_domains_allocated, sizeof(char*));
    if (!error_domains) return DS_ERR_MEM;
    
    err = dsExchangeString(mctx, error_domains, DS_GETTEXT_DOMAIN);
    if (err) {
	error_domains_number = 0;
	dsErrorDestroy();
	return DS_ERR_MEM;
    } 
    error_domains_number = 1;

    error_messages = dsHashTableInit(mctx,ds_errors_number,DS_HASH_TYPE_HASH,NULL);
    if (!error_messages) {
	dsErrorDestroy();
	return DS_ERR_HASHTABLE_CREATE;
    }
    
    err = dsAddErrorsDo(ds_errors, ds_errors_number);
    if (err) dsErrorDestroy();
    return err;
}

void dsErrorDestroy() {
    unsigned long i;
    
    dsDebugServer(DS_DEBUG_NOTICE, "dsErrorDestroy");    

    if (error_messages) {
	dsHashTableDestroy(error_messages);
	error_messages = NULL;
    }
    
    if (error_domains) {
	for (i=0;i<error_domains_number;i++)
	    if (error_domains[i]) dsExchangeString(mctx, error_domains+i, NULL);
	dsFree(mctx, error_domains);
	error_domains = NULL;
    }
    
    mctx = NULL;
}

const char *dsErrorName(int err, const char *locale) {
    DSError emsg;
    
    if (!err) {
	if (locale) return  dsLocaleGetDomainText(DS_GETTEXT_DOMAIN,ds_no_error,locale);
	else return ds_no_error;
    }

    if (error_messages) {
        emsg = (DSError)dsHashTableHashFind(error_messages, 0x18);

	emsg = (DSError)dsHashTableHashFind(error_messages, err);
	if (emsg) {
	    if (locale) return dsLocaleGetDomainText(error_domains[emsg->domain],emsg->str,locale);
	    else return emsg->str;
	}
    } 

    if (locale) return dsLocaleGetDomainText(DS_GETTEXT_DOMAIN,ds_unknown_error,locale);
    return ds_unknown_error;
}
