#include <dslib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <dshashtable.h>

#include <dsmodule.h>


static DSMallocContext mctx = NULL;
static DSHashTable ds_module_ht = NULL;

static DSModuleNameS ds_module_names[] = {
	{0x12, "Malloc"},
	{0x13, "Lock"},
	{0x14, "Signal"},
	{0x15, "SystemLock"},
	{0x16, "SystemService"},
	{0x17, "Directory"},
	{0x20, "Condition"},
	{0x24, "BaseThread"},
	{0x25, "Thread"},
	{0x26, "DoubleThread"},
	{0x2A, "Plugin"},
	{0x2B, "PluginSet"},
	{0x30, "Connection"},
	{0x31, "ConnectionServer"},
	{0x32, "DatagramServer"},
	{0xA0, "Locale"},
	{0xA5, "String"},
	{0xA6, "StringList"},
	{0xA7, "HashTable"},
	{0xA8, "DoubleHash"},
	{0xB0, "TreeNode"},
	{0xB1, "TreeComplexNode"},
	{0xB2, "Enumeration"},
	{0xC1, "Config"},
	{0xD0, "DataBase"},
	{0x101, "Buffer"},
	{0x102, "BufferPool"},
	{0x110, "Allocator"},
	{0x111, "AllocatorSimple"},
	{0x112, "AllocatorDataBuffer"},
	{0x120, "BinaryType"},
	{0x121, "OPCBinary"},
	{0x122, "OPCBinaryDictionary"},
	{0x130, "BinaryFormat"},
	{0x131, "BinaryFormatNative"},
	{0x132, "BinaryFormatOPCXML"},
	{0x140, "DataChunk"},
	{0x141, "ChunkPool"},
	{0x150, "DataHolder"},
	{0x151, "DataHolderBucket"},
	{0x160, "DataConverter"},
	{0x161, "Representation"},
	{0x162, "RepresentationBinary"},
	{0x161, "DataManager"},
	{0x162, "DataManagerFast"},
	{0x163, "DataManagerRep"},
	{0x200, "Dispersion"},
	{0x201, "Dime"}
};
#define ds_modules_number 48

static int dsAddModulesDo(DSModuleName names, unsigned long modules_number) {
    int err;
    unsigned long i;
    
    for (i=0;i<modules_number;i++) {
	err = dsHashTableInsert(ds_module_ht,(void*)(names+i));
	if (err) return err;
    }
    
    return 0;
}

int dsModuleAddModules(DSModuleName names, size_t modules_number) {
    int err;

    ds_assert_init(modules_number);
    ds_assert_param(names);

    err = dsHashTableGrow(ds_module_ht, modules_number);
    if (err) return err;
    
    return dsAddModulesDo(names, modules_number);
}

int dsModuleInit() {
    int err;
    
    dsDebugServer(DS_DEBUG_NOTICE, "dsModuleInit");

    if (mctx) return DS_ERR_INITIALIZED;
    mctx = dsMallocGetDefault();

    ds_module_ht = dsHashTableInit(mctx,ds_modules_number,DS_HASH_TYPE_HASH,NULL);
    if (!ds_module_ht) return DS_ERR_HASHTABLE_CREATE;
    
    err = dsAddModulesDo(ds_module_names, ds_modules_number);
    if (err) dsModuleDestroy();

    return err;
}

void dsModuleDestroy() {
    unsigned long i;
    
    dsDebugServer(DS_DEBUG_NOTICE, "dsModuleDestroy");    

    if (ds_module_ht) {
	dsHashTableDestroy(ds_module_ht);
	ds_module_ht = NULL;
    }
    
    mctx = NULL;
}

const char *dsModuleName(int id) {
    DSModuleName name;

    if (ds_module_ht) {
	name = (DSModuleName)dsHashTableHashFind(ds_module_ht, id);
	if (name) return name->name;
    } 

    return NULL;
}

static struct dsModuleRelations {
    int child;
    int parent;
} ds_module_relations[] = {
    { DS_MODULE_Plugin, DS_MODULE_Thread },
    { DS_MODULE_Thread, DS_MODULE_BaseThread },
    { DS_MODULE_DoubleThread, DS_MODULE_BaseThread },
    { DS_MODULE_BaseThread, DS_MODULE_Condition },
    { 0, 0 }
};

int dsModuleCheckType(DSModule ctx, int parent) {
#ifdef DS_TRACK_MODULES
    int i, myid;

    if (ctx->module_id == parent) return 1;

    for (myid = ctx->module_id;
        for (i = 0; ds_module_relations[i].child;i++)
	    if (ds_module_relations[i].child == myid) {
		if (ds_module_relations[i].parent == parent) return 1;
		myid = ds_module_relations[i].child;
		break;
	    }
	}
    }

    return 0;
#else /* DS_TRACK_MODULES */
    return 1;
#endif /* DS_TRACK_MODULES */
}


DSModule dsModuleConstructor(DSMallocContext mctx, int module_id, size_t size) {
    DSModule ctx;
    
    ctx = dsMallocZero(mctx, size);
    if (ctx) {
	ctx->malloc = mctx;
	dsModuleSetup(ctx, module_id, dsModuleName(module_id));
    }
    return ctx;
}

DSModule dsModuleSubConstructor(DSModule module, DSCreateContextFunction creator) {
    DSModule ctx;
    
    ds_assert_null(module);

    ctx = (DSModule)creator(module->malloc);
    if (ctx) {
	ctx->debug_limit = module->debug_limit;
	ctx->owner = module;
    }
    return ctx;
}

void dsModuleDestructor(DSModule ctx, DSFreeContextFunction cleaner) {
    ds_assert_return(ctx);
    cleaner(ctx);
    dsFree(ctx->malloc, ctx);
}


int dsModuleSetup(DSModule ctx, int module_id, const char *name) {
    ds_assert_param(ctx);

    if (ctx->module_id) return 0;
        
    ctx->module_id = module_id;
    ctx->module_name = name;
    
    return 0;
}

void dsModuleClear(DSModule ctx) {
    ctx->flags = 0;
}

int dsModuleSetDebugLimit(DSModule ctx, DSDebugPriority limit) {
    ctx->debug_limit = limit;

    return 0;
}

int dsModuleSetName(DSModule ctx, int module_id, const char *name) {
    ctx->module_id = module_id;
    ctx->module_name = name;

    return 0;
}

int dsModuleSetFlags(DSModule ctx, DSModuleFlags flags) {
    ctx->flags |= flags;
}

int dsModuleResetFlags(DSModule ctx, DSModuleFlags flags) {
    ctx->flags &= ~flags;
}

int dsModuleCheckInitialized(DSModule ctx) {
    return (ctx->flags&DS_MODULE_FLAG_INITIALIZED)?1:0;
}

