#include <dslib.h>

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif /* HAVE_ERRNO_H */

#include <stdarg.h>

#include <dsassert.h>
#include <dsdebug.h>

#include <dsbasethread.h>
#include <dshashtable.h>
#include <dstools.h>

#define DS_WIN32_THREADS_NUMBER 65535

#ifdef HAVE_PTHREAD_H
static unsigned char threads_initialized = 0;
static pthread_key_t threads;
#else /* HAVE_PTHREAD_H */
static DSLockContext threads_lock = NULL;
static DSHashTable threads = NULL;
#endif /* HAVE_PTHREAD_H */

int dsThreadInit() {
    DSMallocContext mctx;
    int err;

#ifdef HAVE_PTHREAD_H
    if (threads_initialized) return 0;
    err = pthread_key_create(&threads, NULL);
    if (err) {
	if (err == EAGAIN) return 0;
	return DS_ERR_SYSCALL_PTHREAD_KEY_CREATE;
    }
    threads_initialized = 1;
#else /* HAVE_PTHREAD_H */
    mctx = dsMallocGetDefault();
    threads = dsHashTableInit(mctx, DS_WIN32_THREADS_NUMBER, DS_HASH_TYPE_ARRAY, NULL);
    if (!threads) return DS_ERR_MEM;
    dsHashTableSetStringSize(threads, sizeof(DSThreadHandleS));

    threads_lock = dsLockCreateContext(mctx);
    if (!threads_lock) {
	dsThreadDestroy();
	return DS_ERR_MEM;
    }
    
    err = dsLockInitContext(threads_lock);
    if (err) {
	dsThreadDestroy();
	return err;
    }
#endif /* HAVE_PTHREAD_H */
    return 0;    
}

void dsThreadDestroy() {
#ifdef HAVE_PTHREAD_H
    if (threads_initialized) {
	pthread_key_delete(threads);
	threads_initialized = 0;
    }
#else /* HAVE_PTHREAD_H */
    if (threads_lock) {
	dsLockDestroyContext(threads_lock);
	threads_lock = NULL;
    }
    
    if (threads) {
	dsHashTableDestroy(threads);
	threads = NULL;
    }
#endif /* HAVE_PTHREAD_H */
}


int dsThreadRegister(DSThreadRealHandle handle) {
    char stmp[DS_BASE_THREAD_ID_SIZE];

#ifdef HAVE_PTHREAD_H
    int err;
#endif /* HAVE_PTHREAD_H */
    
    ds_assert_param(handle);
    
    dsDebugServer(DS_DEBUG_INFO, "dsThread %p: Register: %s", handle->ctx, dsThreadGetID(stmp));
#ifdef HAVE_PTHREAD_H
    if (!threads_initialized) return DS_ERR_NOTINITIALIZED;

    err = pthread_setspecific(threads, handle);
    if (err) return DS_ERR_SYSCALL_PTHREAD_SET_SPECIFIC;
#else /* HAVE_PTHREAD_H */
    if (!threads) return DS_ERR_NOTINITIALIZED;
    dsLockWrite(threads_lock);
    dsHashTableInsert(threads, handle);
    dsLockReturn(threads_lock);
#endif /* HAVE_PTHREAD_H */
    return 0;
}

int dsThreadUnregister(DSThreadRealHandle handle) {
    char stmp[DS_BASE_THREAD_ID_SIZE];

    ds_assert_param(handle);
    
    dsDebugServer(DS_DEBUG_INFO, "dsThread %p: Unregister: %s", handle->ctx, dsThreadGetID(stmp));

#ifdef HAVE_PTHREAD_H
#else /* HAVE_PTHREAD_H */
    if (!threads) return DS_ERR_NOTINITIALIZED;
    dsLockWrite(threads_lock);
    dsHashTableRemove(threads, (void*)handle);
    dsLockReturn(threads_lock);
#endif /* HAVE_PTHREAD_H */
    return 0;
}

#ifdef HAVE_PTHREAD_H
#else /* HAVE_PTHREAD_H */
static DSThreadRealHandle ThreadFindRealHandle(DSThreadHandle handle) {
    DSThreadRealHandle res;

    if (!threads) return NULL;

    dsLockRead(threads_lock);    
    res = (DSThreadRealHandle)dsHashTableFind(threads, (char*)handle);
    dsLockReturn(threads_lock);

    return res;
}
#endif /* HAVE_PTHREAD_H */

int dsThreadGetCurrentHandle(DSThreadHandle handle) {
#if defined(HAVE_PTHREAD_H)
    handle->id = pthread_self();
#else /* defined(HAVE_WIN32_THREADS) */
    handle->id = GetCurrentThreadId();
#endif /* defined(HAVE_PTHREAD_H) */
    return 0;
}


int dsThreadHandleCmp(DSThreadHandle handle1, DSThreadHandle handle2) {
    if (!handle1) {
	if (!handle2) return 0;
	return -1;
    } else if (!handle2) {
	return 1;
    }
    
    return memcmp(handle1, handle2, sizeof(DSThreadHandleS));
}

DSBaseThreadContext dsThreadFindCurrentContext() {
    DSThreadRealHandle handle;
#ifdef HAVE_PTHREAD_H
    if (!threads_initialized) return NULL;
    handle = (DSThreadRealHandle)pthread_getspecific(threads);
#else /* HAVE_PTHREAD_H */
    DSThreadHandleS handle_tmp;

    if (dsThreadGetCurrentHandle(&handle_tmp)) return NULL;
    handle = ThreadFindRealHandle(&handle_tmp);
#endif /* HAVE_PTHREAD_H */

    if (handle) return handle->ctx;
    return NULL;
}

DSThreadRealHandle dsThreadFindCurrentHandle() {
#ifdef HAVE_PTHREAD_H
    if (!threads_initialized) return NULL;
    return (DSThreadRealHandle)pthread_getspecific(threads);
#else /* HAVE_PTHREAD_H */
    DSThreadHandleS handle_tmp;

    if (dsThreadGetCurrentHandle(&handle_tmp)) return NULL;
    return ThreadFindRealHandle(&handle_tmp);
#endif /* HAVE_PTHREAD_H */
}

int dsThreadSignal(DSThreadHandle handle, DSSignal sig) {
#if defined(_WIN32) || (!defined(HAVE_PSELECT_SYSCALL) && !defined(HAVE_SETJMP_H))
    DSThreadRealHandle base;
# if !defined(HAVE_PSELECT_SYSCALL) && !defined(HAVE_SETJMP_H)
    struct timespec ts = {0,1000};
# endif /* !defined(HAVE_PSELECT_SYSCALL) && !defined(HAVE_SETJMP_H) */
#endif /* defined(_WIN32) || (!defined(HAVE_PSELECT_SYSCALL) && !defined(HAVE_SETJMP_H)) */
    
    ds_assert_param(handle);
    ds_assert_param(sig<DS_SIGNAL_NUMBER);
    
#ifdef _WIN32
    base = ThreadFindRealHandle(handle);
    if (!base) return DS_ERR_THREAD_UNREGISTERED;
    SetEvent(base->winsignals[sig]);
#else /* _WIN32 */

# if !defined(HAVE_PSELECT_SYSCALL) && !defined(HAVE_SETJMP_H)
    jmp_status = base->jmp_status;
# endif /* !defined(HAVE_PSELECT_SYSCALL) && !defined(HAVE_SETJMP_H) */

# if defined(HAVE_PTHREAD_H)
    pthread_kill(handle->id, user_sigs[sig]);
# else /* defined(HAVE_PTHREAD_H) */
#  error "Threading library code is missing!"
# endif /* defined(HAVE_PTHREAD_H) */

# if !defined(HAVE_PSELECT_SYSCALL) && !defined(HAVE_SETJMP_H)
    while ((base->jmp_status)&&(base->jmp_status==jmp_status)) {
	nanosleep(&ts, NULL);
#  if defined(HAVE_PTHREAD_H)
	pthread_kill(handle->id, user_sigs[sig]);
#  else /* defined(HAVE_PTHREAD_H) */
#   error "Threading library code is missing!"
#  endif /* defined(HAVE_PTHREAD_H) */
    }
# endif /* !defined(HAVE_PSELECT_SYSCALL) && !defined(HAVE_SETJMP_H) */
#endif /* _WIN32 */


    return 0;

}

char *dsThreadGetID(char *thread_id) {
    ds_assert_null(thread_id);
#if defined(HAVE_PTHREAD_H)
    sprintf(thread_id, "%8.lx", pthread_self());
#else /* defined(HAVE_WIN32_THREADS) */
    sprintf(thread_id, "%8.lx", GetCurrentThreadId());
#endif /* defined(HAVE_PTHREAD_H) */
    return thread_id;
}

int dsThreadRegisterHandles(DSBaseThreadContext ctx, unsigned handles, ...) {
    unsigned i,j;
    DSThreadRealHandle handle;
    va_list ap;

    ds_assert_param(ctx);
    ds_assert_param(!ctx->handles);
    
    if (handles) {
	ctx->handle = (DSThreadRealHandle*)dsMalloc(((DSConditionContext)ctx)->malloc, handles*sizeof(DSThreadRealHandle));
	if (!ctx->handle) return DS_ERR_MEM;
	ctx->handles = handles;

	va_start(ap, handles);
	for (i=0;i<handles;i++) {
	    handle = va_arg(ap, DSThreadRealHandle);
	    handle->ctx = ctx;
	    ctx->handle[i] = handle;
#ifdef _WIN32
	    for (j=0; j<DS_SIGNAL_NUMBER; j++) {
		handle->winsignals[j] = CreateEvent(NULL, FALSE, FALSE, NULL);
	    }
#else /* _WIN32 */
# ifndef HAVE_PSELECT_SYSCALL
	    handle->jmpstatus = 0;
# endif /* HAVE_PSELECT_SYSCALL */
#endif /* _WIN32 */

	}
	va_end(ap);
    }
    
    return 0;
}


int dsThreadUnregisterHandles(DSBaseThreadContext ctx) {
    unsigned i,j;
    DSThreadRealHandle handle;

    ds_assert_param(ctx);

    if (!ctx->handles) return 0;

#ifdef _WIN32
    for (i=0;i<ctx->handles;i++) {
	handle = ctx->handle[i];
	for (j=0; j<DS_SIGNAL_NUMBER; j++)
	    if (handle->winsignals[j]) CloseHandle(handle->winsignals[j]);	
    }
#endif /* _WIN32 */

    dsFree(((DSConditionContext)ctx)->malloc, ctx->handle);
    ctx->handle = NULL;

    return 0;
}


int dsBaseThreadInitContext(DSBaseThreadContext ctx) {
    int err;
    
    ctx->lock = dsLockCreateContext(ctx->cond.malloc);
    if (!ctx->lock) {
	dsBaseThreadFreeContext(ctx);	
	return DS_ERR_LOCK_CREATE;
    }
    err = dsLockInitContext(ctx->lock);
    if (err) {
	dsBaseThreadFreeContext(ctx);
	return err;
    }
    
    err = dsConditionInitContext((DSConditionContext)ctx);
    if (err) {
	dsBaseThreadFreeContext(ctx);
	return err;
    }
    
    ctx->status = 0;
    return 0;
}

void dsBaseThreadFreeContext(DSBaseThreadContext ctx) {
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    dsConditionFreeContext((DSConditionContext)ctx);
    dsThreadUnregisterHandles(ctx);
}

void dsBaseThreadZeroContext(DSBaseThreadContext ctx) {
    ds_assert_return(ctx);
    dsConditionZeroContext((DSConditionContext)ctx);
    ctx->lock = NULL;
    ctx->initialized = 0;
    ctx->handles = 0;
}

int dsThreadSetScheduler(DSBaseThreadContext ctx, DSThreadScheduler sched, DSThreadSchedulerProperties props) {
    ds_assert_param(ctx);
    ds_assert_param(sched<DS_THREAD_SCHEDULER_MAX);
    
    dsLockWrite(ctx->lock);
    ctx->sched = sched;
    dsLockReturn(ctx->lock);

    return 0;
}

static int dsThreadAlterPriorityInternal(DSThreadRealHandle handle, DSThreadScheduler sched, DSThreadPriority prio) {
    int err;

#ifdef HAVE_PTHREAD_H
    struct sched_param ptrd;
#endif /* HAVE_PTHREAD_H */

#ifdef HAVE_PTHREAD_H
    switch(sched) {
	case DS_THREAD_SCHEDULER_REALTIME:
	    ptrd.sched_priority=prio;
	    if (pthread_setschedparam(handle->handle.id, SCHED_FIFO, &ptrd)) err = -1;
	break;
	case DS_THREAD_SCHEDULER_REALTIME_RR:
	    ptrd.sched_priority=prio;
	    if (pthread_setschedparam(handle->handle.id, SCHED_RR, &ptrd)) err = -1;
	break;
	default:
	    err = DS_ERR_NOTSUPPORTED;
    }
    if (err == -1) {
	if (errno == EINVAL) err = DS_ERR_INVALID_ARGUMENT;
	else if (errno == EPERM) err = DS_ERR_PERMISSION;
	else if (errno == ESRCH) err = DS_ERR_THREAD;
	else err = DS_ERR_SYSCALL_PTHREAD_SETSCHEDPARAM;
    }
#else /* WIN32 THREADS */
    if (!SetThreadPriority(handle->thread_handle, prio)) err = DS_ERR_WINAPI;
#endif /* HAVE_PTHREAD_H */
    
    if (!err) handle->priority = prio;

    return err;
}


int dsThreadSetPriority(DSBaseThreadContext ctx, DSThreadPriority prio) {
    int err = 0;
    unsigned long i;

    ds_assert_param(ctx);
    
    dsLockWrite(ctx->lock);
    ctx->priority = prio;
    
    if ((ctx->handle)&&(ctx->handles)) {
	for (i=0;i<ctx->handles;i++)
	    if (ctx->handle[i]->prop.function) dsThreadAlterPriorityInternal(ctx->handle[i], ctx->sched, prio);
    }    
    
    dsLockReturn(ctx->lock);

    return err;
}

int dsThreadAlterPriority(DSThreadRealHandle handle, DSThreadPriority prio) {
    int err = 0;
    DSBaseThreadContext base;

    ds_assert_param(handle);
    
    base = handle->ctx;
    
    dsLockWrite(base->lock);
    if (handle->prop.function) dsThreadAlterPriorityInternal(handle, base->sched, prio);
    dsLockReturn(base->lock);

    return err;
}

int dsThreadIncreasePriority(DSThreadRealHandle handle) {
    int err = 0;
    DSBaseThreadContext base;

    ds_assert_param(handle);
    
    base = handle->ctx;
    
    dsLockWrite(base->lock);
    if (handle->prop.function) dsThreadAlterPriorityInternal(handle, base->sched, handle->priority + 1);
    dsLockReturn(base->lock);

    return err;
}

int dsThreadDecreasePriority(DSThreadRealHandle handle) {
    int err = 0;
    DSBaseThreadContext base;

    ds_assert_param(handle);
    
    base = handle->ctx;
    
    dsLockWrite(base->lock);
    if (handle->prop.function) dsThreadAlterPriorityInternal(handle, base->sched, handle->priority - 1);
    dsLockReturn(base->lock);

    return err;
}

int dsThreadAlterSelfPriority(DSThreadPriority prio) {
    DSThreadRealHandle handle;
    
    handle = dsThreadFindCurrentHandle();
    if (!handle) return DS_ERR_THREAD;
    return dsThreadAlterPriority(handle, prio);
}

int dsThreadIncreaseSelfPriority() {
    DSThreadRealHandle handle;
    
    handle = dsThreadFindCurrentHandle();
    if (!handle) return DS_ERR_THREAD;
    return dsThreadIncreasePriority(handle);
}

int dsThreadDecreaseSelfPriority() {
    DSThreadRealHandle handle;
    
    handle = dsThreadFindCurrentHandle();
    if (!handle) return DS_ERR_THREAD;
    return dsThreadDecreasePriority(handle);
}

DSThreadPriority dsThreadGetMinimalPriority(DSBaseThreadContext ctx) {
    ds_assert_default(ctx, 0);
#ifdef HAVE_PTHREAD_H
    switch (ctx->sched) {
	case DS_THREAD_SCHEDULER_REALTIME:
	    return sched_get_priority_min(SCHED_FIFO);
	case DS_THREAD_SCHEDULER_REALTIME_RR:
	    return sched_get_priority_min(SCHED_RR);
	default:
	    return 0;
    }
#else /* WIN32_THREADS */
    return THREAD_PRIORITY_LOWEST;
#endif /* HAVE_PTHREAD_H */
}

DSThreadPriority dsThreadGetMaximalPriority(DSBaseThreadContext ctx) {
    ds_assert_default(ctx, 0);
#ifdef HAVE_PTHREAD_H
    switch (ctx->sched) {
	case DS_THREAD_SCHEDULER_REALTIME:
	    return sched_get_priority_max(SCHED_FIFO);
	case DS_THREAD_SCHEDULER_REALTIME_RR:
	    return sched_get_priority_max(SCHED_RR);
	default:
	    return 0;
    }
#else /* WIN32_THREADS */
    return THREAD_PRIORITY_HIGHEST;
#endif /* HAVE_PTHREAD_H */
}

DSThreadPriority dsThreadGetKernelPriority(DSBaseThreadContext ctx) {
    return dsThreadGetMaximalPriority(ctx);
}

DSThreadPriority dsThreadGetSystemPriority(DSBaseThreadContext ctx) {
    return dsThreadGetMaximalPriority(ctx);
}

DSThreadPriority dsThreadGetUserPriority(DSBaseThreadContext ctx) {
    return dsThreadGetMinimalPriority(ctx);
}


void dsThreadSetStatus(DSBaseThreadContext ctx, dword bits) {
    DSLockContext lock;

    ds_assert_return(ctx);

    lock=ctx->lock;

    dsLockWrite(lock);
    ctx->status=bits;
    dsLockReturn(lock);
    dsConditionBroadcast((DSConditionContext)ctx);
}

void dsThreadSetStatusBits(DSBaseThreadContext ctx, dword bits) {
    DSLockContext lock;

    ds_assert_return(ctx);

    lock=ctx->lock;

    dsLockWrite(lock);
    ctx->status|=bits;	    
    dsLockReturn(lock);
    dsConditionBroadcast((DSConditionContext)ctx);
}

void dsThreadResetStatusBits(DSBaseThreadContext ctx, dword bits) {
    DSLockContext lock;
    ds_assert_return(ctx);

    lock=ctx->lock;

    dsLockWrite(lock);
    ctx->status&=~bits;	    
    dsLockReturn(lock);

    dsConditionBroadcast((DSConditionContext)ctx);
}

int dsThreadCheckStatusBitsOn(DSBaseThreadContext ctx, dword bits) {
    ds_assert_param(ctx);
    return ((ctx->status&bits)==bits);
}

int dsThreadCheckStatusBitsOff(DSBaseThreadContext ctx, dword bits) {
    ds_assert_param(ctx);
    return ((ctx->status&bits)==0);
}

int dsThreadCheckStatusBits(DSBaseThreadContext ctx, dword onbits, dword offbits) {
    ds_assert_param(ctx);
    return (((ctx->status&offbits)==0)&&((ctx->status&onbits)==onbits));
}

int dsThreadCheckStatusSomeBitsOn(DSBaseThreadContext ctx, dword bits) {
    ds_assert_param(ctx);
    return ((ctx->status&bits)!=0);
}

int dsThreadCheckStatusSomeBitsOff(DSBaseThreadContext ctx, dword bits) {
    ds_assert_param(ctx);
    return ((ctx->status&bits)!=bits);
}

int dsThreadCheckStatusSomeBits(DSBaseThreadContext ctx, dword onbits, dword offbits) {
    return (((ctx->status&offbits)!=offbits)&&((ctx->status&onbits)!=0));
}


void dsThreadWaitStatusBitsOn(DSBaseThreadContext ctx, dword bits) {
    ds_assert_return(ctx);
    dsConditionPreLock((DSConditionContext)ctx);
    while (!dsThreadCheckStatusBitsOn(ctx,bits)) 
	dsConditionWaitLocked((DSConditionContext)ctx);
    dsConditionUnLock((DSConditionContext)ctx);
}

void dsThreadWaitStatusBitsOff(DSBaseThreadContext ctx, dword bits) {
    ds_assert_return(ctx);
    dsConditionPreLock((DSConditionContext)ctx);
    while (!dsThreadCheckStatusBitsOff(ctx,bits)) 
	dsConditionWaitLocked((DSConditionContext)ctx);
    dsConditionUnLock((DSConditionContext)ctx);
}

void dsThreadWaitStatusBits(DSBaseThreadContext ctx, dword onbits, dword offbits) {
    ds_assert_return(ctx);
    dsConditionPreLock((DSConditionContext)ctx);
    while (!dsThreadCheckStatusBits(ctx,onbits,offbits)) 
	dsConditionWaitLocked((DSConditionContext)ctx);
    dsConditionUnLock((DSConditionContext)ctx);
}

void dsThreadWaitStatusSomeBitsOn(DSBaseThreadContext ctx, dword bits) {
    ds_assert_return(ctx);
    dsConditionPreLock((DSConditionContext)ctx);
    while (!dsThreadCheckStatusSomeBitsOn(ctx,bits)) 
	dsConditionWaitLocked((DSConditionContext)ctx);
    dsConditionUnLock((DSConditionContext)ctx);
}

void dsThreadWaitStatusSomeBitsOff(DSBaseThreadContext ctx, dword bits) {
    ds_assert_return(ctx);
    dsConditionPreLock((DSConditionContext)ctx);
    while (!dsThreadCheckStatusSomeBitsOff(ctx,bits)) 
	dsConditionWaitLocked((DSConditionContext)ctx);
    dsConditionUnLock((DSConditionContext)ctx);
}

void dsThreadWaitStatusSomeBits(DSBaseThreadContext ctx, dword onbits, dword offbits) {
    ds_assert_return(ctx);
    dsConditionPreLock((DSConditionContext)ctx);
    while (!dsThreadCheckStatusSomeBits(ctx,onbits,offbits)) 
	dsConditionWaitLocked((DSConditionContext)ctx);
    dsConditionUnLock((DSConditionContext)ctx);
}


void dsThreadTerminate(DSBaseThreadContext ctx) {
    ds_assert_return(ctx);
    dsThreadSetStatusBits(ctx, DS_THREAD_TERMINATE_FLAG);
}

int dsThreadIsTerminated(DSBaseThreadContext ctx) {
    ds_assert_param(ctx);
    return dsThreadCheckStatusBitsOn(ctx, DS_THREAD_TERMINATE_FLAG);
}

int *dsThreadReturnInt(DSThreadRealHandle ctx, int err) {
    int *ret;
    
    ds_assert_null(ctx);

//	Deallocated by pthreads with standard free. So dsMalloc can't be used.
//    ret = (int*)dsMalloc(ctx->prop.malloc, sizeof(int));

    ret = malloc(sizeof(int));
    if (ret) *ret = err;
    return ret;
}

#ifdef _WIN32
DWORD WINAPI dsThreadGo(DSThreadRealHandle ctx) {
#else /* _WIN32 */
int *dsThreadGo(DSThreadRealHandle ctx) {
#endif /* _WIN32 */
    int err;
    
#ifdef _WIN32
    ds_assert_param(ctx);
#else /* _WIN32 */
    ds_assert_null(ctx);
#endif /* _WIN32 */

/*
    Set appropriate signal mask before execution    
    dsSignalINTBlock();
*/

#ifdef _WIN32
    ctx->handle.id = GetCurrentThreadId();

/*
    DS: Changes process wide settings. Think howto handle this. Set maximal and alter priority?
    
    switch (ctx->ctx.sched) {
	case DS_THREAD_SCHEDULER_REALTIME:
	case DS_THREAD_SCHEDULER_REALTIME_RR:
	    SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS);
	break;
	case DS_THREAD_SCHEDULER_IDLE:
	    SetPriorityClass(GetCurrentProcess(), IDLE_PRIORITY_CLASS);
	break;
	default:
	    ;
    }
*/

    if (ctx->priority) SetThreadPriority(ctx->thread_handle, ctx->priority);
#endif /* _WIN32 */
    
    dsThreadRegister(ctx);
    err = ctx->prop.function(ctx->prop.attr);
    dsThreadUnregister(ctx);
    
    dsLockWrite(ctx->ctx->lock);
    ctx->prop.function = NULL;
    dsLockReturn(ctx->ctx->lock);
    
#ifdef _WIN32
    return err;
#else /* _WIN32 */
    return dsThreadReturnInt(ctx, err);
#endif /* _WIN32 */
}

