#include <dslib.h>

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif /* HAVE_ERRNO_H */

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif /* HAVE_SYS_TIME_H */

#include <dsassert.h>
#include <dsmalloc.h>
#include <dscondition.h>

void dsConditionZeroContext(DSConditionContext ctx) {
    ctx->initialized = 0;
#if defined(HAVE_PTHREAD_H)
#else /* defined(HAVE_WIN32_THREADS) */
    ctx->cond[0] = NULL;
    ctx->cond[1] = NULL;
    ctx->okey_cond = NULL;
    ctx->waiting_lock = NULL;
    ctx->waiting = 0;
    ctx->prelocked = 0;
#endif /* defined(HAVE_PTHREAD_H) */
}

DSConditionContext dsConditionCreateContext(DSMallocContext mctx) {
    DSConditionContext ctx;
    ctx = (DSConditionContext)dsMalloc(mctx, sizeof(struct DSConditionContextT));
    if (ctx) {
	ctx->malloc = mctx;
	dsConditionZeroContext(ctx);
    }
    return ctx;
}

int dsConditionInitContext(DSConditionContext ctx) {
    int err;

    ds_assert_param(ctx);

    dsConditionZeroContext(ctx);

#if defined(HAVE_PTHREAD_H)
    err = pthread_mutex_init(&ctx->mutex,NULL);
    if (err) return DS_ERR_SYSCALL_PTHREAD_MUTEX_INIT;

    err = pthread_cond_init(&ctx->cond,NULL);
    if (err) {
	pthread_mutex_destroy(&ctx->mutex);
	return DS_ERR_SYSCALL_PTHREAD_COND_INIT;
    }
#else /* defined(HAVE_WIN32_THREADS) */
    ctx->cond[0] = CreateEvent(NULL,TRUE,FALSE,NULL);
    ctx->cond[1] = CreateEvent(NULL,FALSE,FALSE,NULL);
    ctx->okey_cond = CreateEvent(NULL,FALSE,FALSE,NULL);
    ctx->mutex = dsLockCreateContext(ctx->malloc);
    ctx->waiting_lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->cond[0]||!ctx->cond[1]||!ctx->okey_cond||!ctx->mutex||!ctx->waiting_lock) {
	dsConditionFreeContext(ctx);
    	return DS_ERR_WINAPI_EVENT_CREATE;
    }

    err = dsLockInitContext(ctx->mutex);
    if (!err) err = dsLockInitContext(ctx->waiting_lock);
    if (err) {
	dsConditionFreeContext(ctx);
    	return err;
    }
#endif /* defined(HAVE_PTHREAD_H) */

    ctx->initialized = DS_INITIALIZED;

    return 0;
}

void dsConditionFreeContext(DSConditionContext ctx) {
    ds_assert_return(ctx);

#if defined(HAVE_PTHREAD_H)
    if (ctx->initialized) {
	pthread_cond_destroy(&ctx->cond);
	pthread_mutex_destroy(&ctx->mutex);
    }
#else /* defined(HAVE_WIN32_THREADS) */
    if (ctx->cond[0]) CloseHandle(ctx->cond[0]);
    if (ctx->cond[1]) CloseHandle(ctx->cond[1]);
    if (ctx->okey_cond) CloseHandle(ctx->okey_cond);
    if (ctx->mutex) dsLockDestroyContext(ctx->mutex);
    if (ctx->waiting_lock) dsLockDestroyContext(ctx->waiting_lock);
#endif /* defined(HAVE_PTHREAD_H) */

    dsConditionZeroContext(ctx);
}

void dsConditionDestroyContext(DSConditionContext ctx) {
    ds_assert_return(ctx);
    dsConditionFreeContext(ctx);
    dsFree(ctx->malloc, ctx);
}

void dsConditionBroadcast(DSConditionContext ctx) {
    ds_assert_returnctx(ctx);

#if defined(HAVE_PTHREAD_H)
    pthread_mutex_lock(&ctx->mutex);
    pthread_cond_broadcast(&ctx->cond);
    pthread_mutex_unlock(&ctx->mutex);
#else /* defined(HAVE_WIN32_THREADS) */
    dsLockWrite(ctx->mutex);
    if (ctx->waiting) {
	ResetEvent(ctx->okey_cond);
	SetEvent(ctx->cond[0]);
	WaitForSingleObject(ctx->okey_cond, INFINITE);
	ResetEvent(ctx->cond[0]);
    }
    dsLockReturn(ctx->mutex);
#endif /* defined(HAVE_PTHREAD_H) */
}

void dsConditionSend(DSConditionContext ctx) {
    ds_assert_returnctx(ctx);

#if defined(HAVE_PTHREAD_H)
    pthread_mutex_lock(&ctx->mutex);
    pthread_cond_signal(&ctx->cond);
    pthread_mutex_unlock(&ctx->mutex);
#else /* defined(HAVE_WIN32_THREADS) */
    dsLockWrite(ctx->mutex);
    if (ctx->waiting) SetEvent(ctx->cond[1]);
    dsLockReturn(ctx->mutex);
#endif /* defined(HAVE_PTHREAD_H) */
}

void dsConditionPreLock(DSConditionContext ctx) {
    ds_assert_returnctx(ctx);

#if defined(HAVE_PTHREAD_H)
    pthread_mutex_lock(&ctx->mutex);
#else /* defined(HAVE_WIN32_THREADS) */
    dsLockWrite(ctx->mutex);
    ctx->prelocked = 1;
#endif /* defined(HAVE_PTHREAD_H) */
}

void dsConditionUnLock(DSConditionContext ctx) {
    ds_assert_returnctx(ctx);

#if defined(HAVE_PTHREAD_H)
    pthread_mutex_unlock(&ctx->mutex);
#else /* defined(HAVE_WIN32_THREADS) */
    ctx->prelocked = 0;
    dsLockReturn(ctx->mutex);
#endif /* defined(HAVE_PTHREAD_H) */
}

void dsConditionWait(DSConditionContext ctx) {
    ds_assert_returnctx(ctx);

#if defined(HAVE_PTHREAD_H)
    pthread_cond_wait(&ctx->cond,&ctx->mutex);
    pthread_mutex_unlock(&ctx->mutex);
#else /* defined(HAVE_WIN32_THREADS) */
    if (ctx->prelocked) ctx->prelocked = 0;
    else dsLockWrite(ctx->mutex);
    dsLockWrite(ctx->waiting_lock);
    ctx->waiting++;
    dsLockReturn(ctx->waiting_lock);
    dsLockReturn(ctx->mutex);
    
    WaitForMultipleObjects(2, ctx->cond, FALSE, INFINITE);
    
    dsLockWrite(ctx->waiting_lock);
    ctx->waiting--;
    if (!ctx->waiting) SetEvent(ctx->okey_cond);
    dsLockReturn(ctx->waiting_lock);
#endif /* defined(HAVE_PTHREAD_H) */
}

void dsConditionWaitLocked(DSConditionContext ctx) {
    ds_assert_returnctx(ctx);

#if defined(HAVE_PTHREAD_H)
    pthread_cond_wait(&ctx->cond,&ctx->mutex);
#else /* defined(HAVE_WIN32_THREADS) */
    if (ctx->prelocked) ctx->prelocked = 0;
    else dsLockWrite(ctx->mutex);
    dsLockWrite(ctx->waiting_lock);
    ctx->waiting++;
    dsLockReturn(ctx->waiting_lock);
    dsLockReturn(ctx->mutex);
    
    WaitForMultipleObjects(2, ctx->cond, FALSE, INFINITE);
    
    dsLockWrite(ctx->waiting_lock);
    ctx->waiting--;
    if (!ctx->waiting) SetEvent(ctx->okey_cond);
    dsLockReturn(ctx->waiting_lock);
    
    dsLockWrite(ctx->mutex);
    ctx->prelocked = 1;
#endif /* defined(HAVE_PTHREAD_H) */
}


int dsConditionTimedWait(DSConditionContext ctx, DSDuration timeout) {
    DSTimeS curtime;
    
    dsTimeCurrentTime(&curtime);
    dsTimeAdd(&curtime, timeout);
    return dsConditionTimedWaitUntil(ctx, &curtime);
}

int dsConditionTimedWaitUntil(DSConditionContext ctx, DSTime time) {
    int err;
#if defined(HAVE_PTHREAD_H)
    struct timespec ts;
#else /* defined(HAVE_WIN32_THREADS) */
    DWORD res;
#endif /* defined(HAVE_PTHREAD_H) */

    ds_assert_ctx(ctx);

#if defined(HAVE_PTHREAD_H)
    dsTimeExportTimespec(time, &ts);

    err = pthread_cond_timedwait(&ctx->cond,&ctx->mutex,&ts);
    pthread_mutex_unlock(&ctx->mutex);

    if (err) {
	if (err == ETIMEDOUT) return DS_ERR_TIMEOUT;
	if (err == EINTR) return DS_ERR_SIGNAL_EVENT;
	return DS_ERR_SYSCALL_PTHREAD_COND_WAIT;
    }
#else /* defined(HAVE_WIN32_THREADS) */
    if (ctx->prelocked) ctx->prelocked = 0;
    else dsLockWrite(ctx->mutex);
    dsLockWrite(ctx->waiting_lock);
    ctx->waiting++;
    dsLockReturn(ctx->waiting_lock);
    dsLockReturn(ctx->mutex);
    
    res = WaitForMultipleObjects(2, ctx->cond, FALSE, (DWORD)dsTimeDiffCurrentMilliseconds(time));
    if ((res == WAIT_OBJECT_0)||(res == WAIT_OBJECT_0 + 1)) err = 0;
    else err = -1;
    
    dsLockWrite(ctx->waiting_lock);
    ctx->waiting--;
    if (!ctx->waiting) SetEvent(ctx->okey_cond);
    dsLockReturn(ctx->waiting_lock);

    if (err) return DS_ERR_TIMEOUT;
#endif /* defined(HAVE_PTHREAD_H) */

    return 0;
}

int dsConditionTimedWaitLocked(DSConditionContext ctx, DSDuration timeout) {
    DSTimeS curtime;
    
    dsTimeCurrentTime(&curtime);
    dsTimeAdd(&curtime, timeout);
    return dsConditionTimedWaitLockedUntil(ctx, &curtime);
}

int dsConditionTimedWaitLockedUntil(DSConditionContext ctx, DSTime time) {
    int err;
#if defined(HAVE_PTHREAD_H)
    struct timespec ts;
#else /* defined(HAVE_WIN32_THREADS) */
    DWORD res;
#endif /* defined(HAVE_PTHREAD_H) */

    ds_assert_ctx(ctx);

#if defined(HAVE_PTHREAD_H)
    dsTimeExportTimespec(time, &ts);

    err = pthread_cond_timedwait(&ctx->cond,&ctx->mutex,&ts);

    if (err) {
	if (err == ETIMEDOUT) return DS_ERR_TIMEOUT;
	if (err == EINTR) return DS_ERR_SIGNAL_EVENT;
	return DS_ERR_SYSCALL_PTHREAD_COND_WAIT;
    }
#else /* defined(HAVE_WIN32_THREADS) */
    if (ctx->prelocked) ctx->prelocked = 0;
    else dsLockWrite(ctx->mutex);
    dsLockWrite(ctx->waiting_lock);
    ctx->waiting++;
    dsLockReturn(ctx->waiting_lock);
    dsLockReturn(ctx->mutex);
    
    res = WaitForMultipleObjects(2, ctx->cond, FALSE, (DWORD)dsTimeDiffCurrentMilliseconds(time));
    if ((res == WAIT_OBJECT_0)||(res == WAIT_OBJECT_0 + 1)) err = 0;
    else err = -1;
    
    dsLockWrite(ctx->waiting_lock);
    ctx->waiting--;
    if (!ctx->waiting) SetEvent(ctx->okey_cond);
    dsLockReturn(ctx->waiting_lock);
    
    dsLockWrite(ctx->mutex);
    ctx->prelocked = 1;

    if (err) return DS_ERR_TIMEOUT;
#endif /* defined(HAVE_PTHREAD_H) */

    return 0;
}
