#include <dslib.h>
#include <dsassert.h>

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif /* HAVE_ERRNO_H */

#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif /* HAVE_SYS_SOCKET_H */
#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif /* HAVE_NETINET_IN_H */
#ifdef HAVE_SYS_UN_H
# include <sys/un.h>
#endif /* HAVE_SYS_UN_H */
#ifdef HAVE_NETDB_H
# include <netdb.h>
#endif /* HAVE_NETDB_H */
#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif /* HAVE_FCNTL_H */

#ifdef HAVE_TERMIOS_H
# include <termios.h>
#endif /* HAVE_TERMIOS_H */

#ifdef HAVE_SYS_IOCTL_H
# include <sys/ioctl.h>
#endif /* HAVE_SYS_IOCTL_H */

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif /* HAVE_UNISTD_H */

#include <dsdebug.h>
#include <dslib.h>
#include <dsconnection.h>

#ifndef HAVE_GETADDRINFO
#ifndef _WIN32
static DSLockContext resolverlock = NULL;
extern int h_errno;
#endif /* _WIN32 */
#endif /* HAVE_GETADDRINFO */

int dsConnectionInit() {
    int err = 0;
#ifndef HAVE_GETADDRINFO
#ifndef _WIN32
    resolverlock = dsLockCreateContext(NULL);
    if (!resolverlock) return DS_ERR_MEM;
    
    err = dsLockInitContext(resolverlock);
    if (err) {
	dsLockDestroyContext(resolverlock);
	resolverlock = NULL;
    }
#endif /* _WIN32 */
#endif /* HAVE_GETADDRINFO */
    return err;
}

void dsConnectionDestroy() {
#ifndef HAVE_GETADDRINFO
#ifndef _WIN32
    if (resolverlock) {
	dsLockDestroyContext(resolverlock);
	resolverlock = NULL;
    }
#endif /* _WIN32 */
#endif /* HAVE_GETADDRINFO */
}

int dsConnectionAddressSet(DSConnectionAddress *ctx, DSConnectionAddressType type, size_t size, const char *addr) {
    if (size > DS_CONNECTION_MAX_ADDRESS_SIZE) return DS_ERR_CONNECTION_ADDRESS_TOO_LONG;
    memcpy(ctx->addr, addr, size);
    ctx->size = size;
    ctx->type = type;
    return 0;
}

int dsConnectionAddressCompare(DSConnectionAddress *ctx, DSConnectionAddressType type, size_t size, const char *addr) {
    if ((ctx->size != size)||(ctx->type != type)) return 1;
    return memcmp(ctx->addr, addr, size);
}

int dsConnectionAddressCpy(DSConnectionAddress *ctx1, DSConnectionAddress *ctx2) {
    ds_assert_param(ctx1);
    ds_assert_param(ctx2);
    
    ctx1->size = ctx2->size;
    ctx1->type = ctx2->type;
    memcpy(ctx1->addr, ctx2->addr, ctx2->size);
    
    return 0;
}

int dsConnectionAddressCmp(DSConnectionAddress *ctx1, DSConnectionAddress *ctx2) {
    ds_assert_param(ctx1);
    ds_assert_param(ctx2);
    
    if ((ctx1->size != ctx2->size)||(ctx1->type != ctx2->type)) return 1;
    return memcmp(ctx1->addr, ctx2->addr, ctx2->size);
}

void dsConnectionZeroContext(DSConnectionContext ctx) {
    ds_assert_return(ctx);

    ctx->socket = 0;
    ctx->sigmask = NULL;
    ctx->softsig = NULL;

    ctx->data = NULL;
    
#ifdef _WIN32
    ctx->winsock = INVALID_SOCKET;
    ctx->handle = NULL;
    ctx->hEvent = NULL;
    ctx->hEventRead = NULL;
    ctx->hEventWrite = NULL;
#endif /* _WIN32 */

    ctx->read = NULL;
    ctx->write = NULL;
    ctx->type = DS_CONNECTION_TYPE_UNKNOWN;
    ctx->subtype = DS_CONNECTION_SUBTYPE_DEFAULT;
    ctx->status = DS_CONNECTION_STATUS_UNCONNECTED;
}

DSConnectionContext dsConnectionCreateContext(DSMallocContext mctx) {
    DSConnectionContext ctx;

    dsDebugServer(DS_DEBUG_INFO, "dsConnectionCreateContext");

    ctx = (DSConnectionContext)dsMalloc(mctx, sizeof(DSConnectionContextS));
    if (ctx) {
	ctx->malloc = mctx;
	ctx->timeout = NULL;
	ctx->flags = 0;
	dsConnectionZeroContext(ctx);
    }
    return ctx;
}

int dsConnectionCreateSContext(DSConnectionContext ctx, DSMallocContext mctx) {
    dsDebugServer(DS_DEBUG_INFO, "dsConnectionCreateSContext");
    
    ds_assert_param(ctx);
    
    ctx->malloc = mctx;
    ctx->timeout = NULL;
    ctx->flags = 0;
    dsConnectionZeroContext(ctx);

    return 0;
}

int dsConnectionSetTimeout(DSConnectionContext ctx, DSDuration timeout) {
    ds_assert_param(ctx);
    
    if (timeout) {
	ctx->timeout = &ctx->timeouts;
	dsDurationCopy(ctx->timeout, timeout);
    } else ctx->timeout = timeout;

    return 0;
}

int dsConnectionSetDefaultFlags(DSConnectionContext ctx, DSConnectionFlags flags) {
    ds_assert_param(ctx);
    if (flags&(DS_CONNECTION_FLAG_EXAMINE|DS_CONNECTION_FLAG_OVERRIDEDEFAULT|DS_CONNECTION_FLAG_TRY))
	return DS_ERR_INVALID_ARGUMENT;
    ctx->flags = flags;
    return 0;
}

#ifdef HAVE_SOCKETS
# ifdef _WIN32
int dsConnectionInitSocketContext(DSConnectionContext ctx, DSSignalMask sigmask, SOCKET sock) {
    u_long blocking_mode = 1;	/* non-blocking mode */

    dsDebugServer(DS_DEBUG_INFO, "dsConnectionInitSocketContext");    

    ds_assert_param(ctx);
    
    dsConnectionZeroContext(ctx);

    ctx->winsock = sock;
    if (sigmask) {
	ctx->sigmask = (DSSignalMask)dsMalloc(ctx->malloc, sizeof(DSSignalMaskS));
	if (!ctx->sigmask) {
	    dsConnectionFreeContext(ctx);
	    return DS_ERR_MEM;
	}
	memcpy(ctx->sigmask, sigmask, sizeof(DSSignalMaskS));
    }

    ctx->hEventRead = WSACreateEvent();
    ctx->hEventWrite = WSACreateEvent();
    if ((!ctx->hEventRead)||(!ctx->hEventWrite)) {
	dsConnectionFreeContext(ctx);
	return DS_ERR_WINAPI_EVENT_CREATE;
    }

    if (WSAEventSelect(sock, ctx->hEventRead, FD_READ) == SOCKET_ERROR) {
	    dsConnectionFreeContext(ctx);
	    return DS_ERR_SOCKET_SELECT;
    }

/*
    The last WSAEventSelect is used
    if (WSAEventSelect(sock, ctx->hEventWrite, FD_WRITE) == SOCKET_ERROR) {
	    dsConnectionFreeContext(ctx);
	    return DS_ERR_SOCKET_SELECT;
    }
*/
    
    if (!sigmask) blocking_mode = 0;
    
    if (ioctlsocket(sock, FIONBIO, &blocking_mode) == SOCKET_ERROR) {
	dsConnectionFreeContext(ctx);
	return DS_ERR_SOCKET_IOCTL;
    }

    ctx->type = DS_CONNECTION_TYPE_WINSOCK;
    ctx->status = DS_CONNECTION_STATUS_CONNECTED;
    
    return 0;
}
# else /* _WIN32 */
int dsConnectionInitSocketContext(DSConnectionContext ctx, DSSignalMask sigmask, int sock) {
    int err;
    int flags;
    dsDebugServer(DS_DEBUG_INFO, "dsConnectionInitSocketContext");    

    ds_assert_param(ctx);
    ds_assert_param(sock);
    
    dsConnectionZeroContext(ctx);
    
    if (sigmask) {
	ctx->sigmask = (DSSignalMask)dsMalloc(ctx->malloc, sizeof(DSSignalMaskS));
	if (!ctx->sigmask) {
	    close(sock);
	    return DS_ERR_MEM;
	}
	memcpy(ctx->sigmask, sigmask, sizeof(DSSignalMaskS));
    }

retry_fcntl:	
    flags=fcntl(sock,F_GETFL,0);
    if ((flags<0)||(fcntl(sock,F_SETFL,flags&(~O_NONBLOCK))<0)) {
	err = errno;
	if (err == EAGAIN) goto retry_fcntl;
	close(sock);
	if (ctx->sigmask) dsFree(ctx->malloc, ctx->sigmask);
	return DS_ERR_SOCKET_FCNTL;
    }

    ctx->socket = sock;
    ctx->type = DS_CONNECTION_TYPE_SOCKET;
    ctx->status = DS_CONNECTION_STATUS_CONNECTED;

    return 0;
}
# endif /* _WIN32 */
#endif /* HAVE_SOCKETS */

#ifdef _WIN32
static int dsConnectionReadHandle(DSConnectionContext ctx, void *buf, size_t len, int flags, size_t *rbytes);
static int dsConnectionWriteHandle(DSConnectionContext ctx, const void *buf, size_t len, int flags, size_t *wbytes);
#else
static int dsConnectionReadFile(DSConnectionContext ctx, void *buf, size_t len, DSConnectionFlags flags, size_t *rbytes);
static int dsConnectionWriteFile(DSConnectionContext ctx, const void *buf, size_t len, DSConnectionFlags flags, size_t *wbytes);
#endif /* _WIN32 */
static int dsConnectionReadCFile(DSConnectionContext ctx, void *buf, size_t len, DSConnectionFlags flags, size_t *rbytes);
static int dsConnectionWriteCFile(DSConnectionContext ctx, const void *buf, size_t len, DSConnectionFlags flags, size_t *wbytes);

#ifdef HAVE_SOCKETS
int dsConnectionConnectServer(DSConnectionContext ctx, DSSignalMask sigmask, int protocol, const char *server, in_port_t port) {
    int err;
    
    DSConnectionUDPData udp;

# ifdef _WIN32
    int i;
    
    SOCKET sock;
    SOCKADDR_IN mysock;

    u_long blocking_mode = 1;	/* non-blocking mode */
# else /* _WIN32 */
    int sock;
    int flags;
    struct sockaddr_in mysock;
# endif /* _WIN32 */

    struct hostent *hent;
    
# ifdef HAVE_GETADDRINFO
    struct addrinfo *addr;
    char portstring[16];
# endif /* HAVE_GETADDRINFO */


    ds_assert_param(ctx);
//    if (!server) server = "127.0.0.1";    

    dsConnectionZeroContext(ctx);

# ifdef HAVE_GETADDRINFO
    sprintf(portstring, "%u", port);
# endif /* HAVE_GETADDRINFO */

    if (server) {
retry_gai:

# ifdef HAVE_GETADDRINFO
	err = getaddrinfo(server, portstring, NULL, &addr);
	if (err) {
	    if (err == EAI_AGAIN) goto retry_gai;
	    if ((err == EAI_NONAME)||(err == EAI_SERVICE)||(err == EAI_ADDRFAMILY)) return DS_ERR_CONNECTION_UNKNOWN_HOST;
	    return DS_ERR_SOCKET_GETADDRINFO;
	}
    
	if (sizeof(struct sockaddr_in) < addr->ai_addrlen) return DS_ERR_SOCKET_GETADDRINFO;
	memcpy(&mysock, addr->ai_addr, addr->ai_addrlen);
	freeaddrinfo(addr);
# else /* HAVE_GETADDRINFO */
	mysock.sin_family=AF_INET;
	mysock.sin_port=htons(port);

#  ifdef _WIN32
	for (i=0;server[i];i++) {
	    if (((server[i]>='a')&&(server[i]<='z'))||((server[i]>='A')&&(server[i]<='Z'))) break;
	}
	
	if (server[i]) {
#  else /* _WIN32 */
	if (resolverlock) dsLockWrite(resolverlock);
#  endif /* _WIN32 */

	hent = gethostbyname(server);
	if ((!hent)||(hent->h_length > sizeof(mysock.sin_addr))) {
#  ifdef _WIN32
	    err = WSAGetLastError();
	    if (err == WSATRY_AGAIN) goto retry_gai;
	    return DS_ERR_WINAPI_GETHOSTBYNAME;
#  else 
	    if (resolverlock) dsLockReturn(resolverlock);
	    if (h_errno == TRY_AGAIN) goto retry_gai;
	    if (h_errno == HOST_NOT_FOUND) return DS_ERR_CONNECTION_UNKNOWN_HOST;
	    return DS_ERR_SOCKET_GETHOSTBYNAME;
#  endif /* _WIN32 */
	}
	memcpy(&mysock.sin_addr, hent->h_addr, hent->h_length);
#  ifdef _WIN32
	} else {
	    mysock.sin_addr.s_addr = inet_addr(server);
	}
#  else /* _WIN32 */
	if (resolverlock) dsLockReturn(resolverlock);
#  endif /* _WIN32 */

# endif /* HAVE_GETADDRINFO */
    }
    
    switch (protocol) {
	case IPPROTO_TCP:
	    sock = socket(PF_INET,SOCK_STREAM,protocol);
	break;
	case IPPROTO_UDP:
	    sock = socket(PF_INET,SOCK_DGRAM,protocol);
	break;
	default:
	    return DS_ERR_NOTSUPPORTED;
    };
#ifdef _WIN32
    if (sock==INVALID_SOCKET) return DS_ERR_SOCKET_CREATE;
#else /* _WIN32 */
    if (sock<=0) return DS_ERR_SOCKET_CREATE;
#endif /* _WIN32 */

# ifdef _WIN32
    if ((server)&&(connect(sock, (SOCKADDR*)&mysock, sizeof(SOCKADDR_IN)) == SOCKET_ERROR)) {
	err = WSAGetLastError();
	closesocket(sock);
	if (err == WSAEISCONN) return DS_ERR_CONNECTION_CONNECTED;
	if ((err == WSAENETUNREACH)||(err == WSAEHOSTUNREACH)||(WSAETIMEDOUT)) return DS_ERR_CONNECTION_CONNECT;
	return DS_ERR_SOCKET_CONNECT;
    }
	
    if (sigmask) {
	if (ioctlsocket(sock, FIONBIO, &blocking_mode) == SOCKET_ERROR) {
	    closesocket(sock);    
    	    return DS_ERR_SOCKET_IOCTL;
	}
    }

    ctx->winsock = sock;

    ctx->hEventRead = WSACreateEvent();
    ctx->hEventWrite = WSACreateEvent();
    if ((!ctx->hEventRead)||(!ctx->hEventWrite)) {
	dsConnectionFreeContext(ctx);
	return DS_ERR_WINAPI_EVENT_CREATE;
    }

    if (WSAEventSelect(sock, ctx->hEventRead, FD_READ) == SOCKET_ERROR) {
	dsConnectionFreeContext(ctx);
	return DS_ERR_SOCKET_SELECT;
    }

/*
    The last WSAEventSelect is used
    if (WSAEventSelect(sock, ctx->hEventWrite, FD_WRITE) == SOCKET_ERROR) {
	dsConnectionFreeContext(ctx);
	return DS_ERR_SOCKET_SELECT;
    }
*/
    
    ctx->type = DS_CONNECTION_TYPE_WINSOCK;
    ctx->status = DS_CONNECTION_STATUS_CONNECTED;
# else /* _WIN32 */
retry_connect:
    if ((server)&&(connect(sock,(struct sockaddr*)&mysock,sizeof(mysock))<0)) {
        if (errno == EINPROGRESS) ctx->status = DS_CONNECTION_STATUS_CONNECTING;
	else if (errno == EAGAIN) goto retry_connect;
	else {
	    close(sock);
	    /*if (errno == ECONNREFUSED) return DS_ERR_CONNECTION_REFUSED;
	    else if (errno == ETIMEDOUT) return DS_ERR_CONNECTION_TIMEOUT;
	    else if (errno == ENETUNREACH) return DS_ERR_CONNECTION_NETWORK;
	    else */if ((errno == EACCES)||(errno == EPERM)) return DS_ERR_PERMISSION;
	    return DS_ERR_SOCKET_CONNECT;
	}
    }
    
retry_fcntl:
    if (sigmask) {
	flags=fcntl(sock,F_GETFL,0);
	if ((flags<0)||(fcntl(sock,F_SETFL,flags|O_NONBLOCK)<0)) {
	    err = errno;
	    if (err == EAGAIN) goto retry_fcntl;
	    close(sock);
	    return DS_ERR_SOCKET_FCNTL;
	}
    }

    ctx->socket = sock;
    ctx->type = DS_CONNECTION_TYPE_SOCKET;
    ctx->status = DS_CONNECTION_STATUS_CONNECTED;
# endif /* _WIN32 */

    if (protocol == IPPROTO_UDP) {
	ctx->subtype = DS_CONNECTION_SUBTYPE_UDP;
	udp = (DSConnectionUDPData)dsMalloc(ctx->malloc, sizeof(DSConnectionUDPDataS));
	if (!udp) {
	    dsConnectionFreeContext(ctx);
	    return DS_ERR_MEM;
	}
	ctx->data = (DSConnectionData)udp;
	udp->size = 0;
	udp->fixed = 1;
	
	err = dsConnectionAddressSet(&udp->remote, DS_CONNECTION_ADDRESS_TYPE_SOCKADDR, sizeof(mysock), (char*)&mysock);
	if (err) {
	    dsConnectionFreeContext(ctx);
	    return err;
	}
    }

    if (sigmask) {
	ctx->sigmask = (DSSignalMask)dsMalloc(ctx->malloc, sizeof(DSSignalMaskS));
	if (!ctx->sigmask) {
	    dsConnectionFreeContext(ctx);
	    return DS_ERR_MEM;
	}
	memcpy(ctx->sigmask, sigmask, sizeof(DSSignalMaskS));
    }

    return 0;
}

int dsConnectionConnectUDPServer(DSConnectionContext ctx, DSSignalMask sigmask, const char *server, in_port_t port) {
    return dsConnectionConnectServer(ctx, sigmask, IPPROTO_UDP, server, port);
}

int dsConnectionConnectTCPServer(DSConnectionContext ctx, DSSignalMask sigmask, const char *server, in_port_t port) {
    return dsConnectionConnectServer(ctx, sigmask, IPPROTO_TCP, server, port);
}
#endif /* HAVE_SOCKETS */


int dsConnectionConnectSocket(DSConnectionContext ctx, DSSignalMask sigmask, const char *path, const char *name, const char *extension) {
#if defined(HAVE_SYS_UN_H)
    int err;
    int flags;
    int sock;
    struct sockaddr_un mysock;
#elif defined(_WIN32)
    char *local_name;
#else /* defined(HAVE_SYS_UN_H) */
#endif /* defined(HAVE_SYS_UN_H) */

    dsDebugServer(DS_DEBUG_INFO, "dsConnectionConnectSocket");    

    ds_assert_param(ctx);
    ds_assert_param(name);

    dsConnectionZeroContext(ctx);
    
#if defined(HAVE_SYS_UN_H)
    sock = socket(PF_UNIX, SOCK_STREAM, 0);
    if (sock<=0) return DS_ERR_SOCKET_CREATE;

    mysock.sun_family=AF_UNIX;
    err = dsSetFullName(mysock.sun_path,sizeof(mysock.sun_path), path, name, extension);
    if (err) {
        close(sock);
        return err;
    }	

retry_fcntl:	
    flags=fcntl(sock,F_GETFL,0);
    if ((flags<0)||(fcntl(sock,F_SETFL,flags|O_NONBLOCK)<0)) {
	err = errno;
	if (err == EAGAIN) goto retry_fcntl;
	close(sock);
	return DS_ERR_SOCKET_FCNTL;
    }

retry_connect:
    if (connect(sock,(struct sockaddr*)&mysock,sizeof(mysock))<0) {
        if (errno == EINPROGRESS) ctx->status = DS_CONNECTION_STATUS_CONNECTING;
	else if (errno == EAGAIN) goto retry_connect;
	else {
	    close(sock);
	    /*if (errno == ECONNREFUSED) return DS_ERR_CONNECTION_REFUSED;
	    else if (errno == ETIMEDOUT) return DS_ERR_CONNECTION_TIMEOUT;
	    else if (errno == ENETUNREACH) return DS_ERR_CONNECTION_NETWORK;
	    else */if ((errno == EACCES)||(errno == EPERM)) return DS_ERR_PERMISSION;
	    return DS_ERR_SOCKET_CONNECT;
	}
    } else {
	ctx->status = DS_CONNECTION_STATUS_CONNECTED;
    }
    
    if (!sigmask) {
retry_block_fcntl:
	if (fcntl(sock,F_SETFL,flags&(~O_NONBLOCK))<0) {
	    err = errno;
	    if (err == EAGAIN) goto retry_block_fcntl;
	    close(sock);
	    return DS_ERR_SOCKET_FCNTL;
	}
    }

    ctx->socket = sock;
    ctx->type = DS_CONNECTION_TYPE_SOCKET;
#elif defined(_WIN32)
    if (extension) {
        local_name = (char*)dsMalloca((strlen(extension) + strlen(name)+14)*sizeof(char));
	if (!local_name) return DS_ERR_MEM;
    	sprintf(local_name, "\\\\.\\pipe\\%s%s",name,extension);
    } else {
	local_name = (char*)dsMalloca((strlen(name)+14)*sizeof(char));
	if (!local_name) return DS_ERR_MEM;
    	sprintf(local_name, "\\\\.\\pipe\\%s",name);
    }
    
    ctx->handle = CreateFile(local_name, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);
    if (!ctx->handle) {
	if (WaitNamedPipe(local_name, NMPWAIT_USE_DEFAULT_WAIT)) {
	    ctx->handle = CreateFile(local_name, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL);
	}

	if (!ctx->handle) {
	    dsFreea(local_name);
	    return DS_ERR_CONNECTION_CONNECT;
	}
    }
    dsFreea(local_name);

    ctx->hEventRead = CreateEvent(NULL, TRUE, FALSE, NULL);
    ctx->hEventWrite = CreateEvent(NULL, TRUE, FALSE, NULL);
    if (!ctx->hEventRead||!ctx->hEventWrite) {
	if (ctx->hEventRead) {
	    CloseHandle(ctx->hEventRead);
	    ctx->hEventRead = NULL;
	}
	if (ctx->hEventWrite) {
	    CloseHandle(ctx->hEventWrite);
	    ctx->hEventWrite = NULL;
	}
	CloseHandle(ctx->handle);
	return DS_ERR_WINAPI_EVENT_CREATE;
    }
    
    ctx->read = dsConnectionReadHandle;
    ctx->write = dsConnectionWriteHandle;

    ctx->type = DS_CONNECTION_TYPE_NAMED_PIPE_CLIENT;
    ctx->status = DS_CONNECTION_STATUS_CONNECTED;
#else /* defined(HAVE_SYS_UN_H) */
    return DS_ERR_NOTIMPLEMENTED;
#endif /* defined(HAVE_SYS_UN_H) */
    if (sigmask) {
	ctx->sigmask = (DSSignalMask)dsMalloc(ctx->malloc, sizeof(DSSignalMaskS));
	if (!ctx->sigmask) {
	    dsConnectionFreeContext(ctx);
	    return DS_ERR_MEM;
	}
	memcpy(ctx->sigmask, sigmask, sizeof(DSSignalMaskS));
    }

    return 0;
}

#ifndef _WIN32
int dsConnectionConnectFile(DSConnectionContext ctx, DSSignalMask sigmask, const char *file) {
    int sock;

    dsDebugServer(DS_DEBUG_INFO, "dsConnectionConnectFile");    

    assert(ctx);
    assert(file);

    dsConnectionZeroContext(ctx);

    if (sigmask) sock = open(file, O_RDWR | O_NOCTTY | O_NDELAY);
    else sock = open(file, O_RDWR | O_NOCTTY);
    if (sock<=0) return DS_ERR_FILE_OPEN;
    
    ctx->socket = sock;
    ctx->status = DS_CONNECTION_STATUS_CONNECTED;
    ctx->type = DS_CONNECTION_TYPE_FILE;

    ctx->read = dsConnectionReadFile;
    ctx->write = dsConnectionWriteFile;
    
    if (sigmask) {
	ctx->sigmask = (DSSignalMask)dsMalloc(ctx->malloc, sizeof(DSSignalMaskS));
	if (!ctx->sigmask) {
	    dsConnectionFreeContext(ctx);
	    return DS_ERR_MEM;
	}
	memcpy(ctx->sigmask, sigmask, sizeof(DSSignalMaskS));
    }

    return 0;
}
#endif /* ! _WIN32 */

int dsConnectionConnectCFile(DSConnectionContext ctx, DSSignalMask sigmask, const char *file) {
    return DS_ERR_NOTIMPLEMENTED;
}

DSConnectionCOMOptions dsConnectionCOMGetDefaultOptions() {
    DSConnectionCOMOptions options;
    
    options = (DSConnectionCOMOptions)malloc(sizeof(DSConnectionCOMOptionsS));
    if (!options) return options;
    
    dsConnectionCOMInitDefaultOptions(options);
    return options;
}

int dsConnectionCOMInitDefaultOptions(DSConnectionCOMOptions options) {
    return dsConnectionCOMOptionsSetSpeed(options, DS_DEFAULT_COM_SPEED);
}

void dsConnectionCOMDestroyOptions(DSConnectionCOMOptions options) {
    ds_assert_return(options);
    
    free(options);
}

int dsConnectionCOMOptionsSetSpeed(DSConnectionCOMOptions options, unsigned long speed) {

    ds_assert_param(options);

    if (!speed) speed = DS_DEFAULT_COM_SPEED;

#if defined(_WIN32)
    options->speed = speed;
#elif defined(HAVE_TERMIOS_H)
    switch(speed) {
	case 1200:
	    options->speed = B1200;
	break;
	case 2400:
	    options->speed = B2400;
	break;
	case 4800:
	    options->speed = B4800;
	break;
	case 9600:
	    options->speed = B9600;
	break;
	case 19200:
	    options->speed = B19200;
	break;
	case 38400:
	    options->speed = B38400;
	break;
	case 57600:
	    options->speed = B57600;
	break;
	case 115200:
	    options->speed = B115200;
	break;
	case 230400:
	    options->speed = B230400;
	break;
	case 460800:
	    options->speed = B460800;
	break;
	case 500000:
	    options->speed = B500000;
	break;
	case 576000:
	    options->speed = B576000;
	break;
	case 921600:
	    options->speed = B921600;
	break;
	case 1000000:
	    options->speed = B1000000;
	break;
	case 1152000:
	    options->speed = B1152000;
	break;
	case 2000000:
	    options->speed = B2000000;
	break;
	case 2500000:
	    options->speed = B2500000;
	break;
	case 3000000:
	    options->speed = B3000000;
	break;
	case 3500000:
	    options->speed = B3500000;
	break;
	case 4000000:
	    options->speed = B4000000;
	break;
	default:
	    return DS_ERR_CONNECTION_INVALID_SPEED;
    }
#else
    options->speed = speed;
#endif /* HAVE_TERMIOS_H */

    return 0;
}

#if defined(_WIN32)
int dsConnectionConnectCOM(DSConnectionContext ctx, DSSignalMask sigmask, const char *device, DSConnectionCOMOptions options) {
    return DS_ERR_NOTSUPPORTED;
}
#elif defined(HAVE_TERMIOS_H)
int dsConnectionConfigureCOM(DSConnectionContext ctx, DSConnectionCOMOptions options) {
    DSConnectionCOMOptionsS defoptions;
    struct termios tio;

    if (!options) {
	dsConnectionCOMInitDefaultOptions(&defoptions);
	options = &defoptions;
    }

    tcgetattr(ctx->socket,&tio);
    tio.c_cflag=CLOCAL|CS8|CREAD|options->speed;
    tio.c_iflag=IGNBRK;
    tio.c_oflag=0;
    tio.c_lflag=0;
    tcflush(ctx->socket, TCIFLUSH);
    tcsetattr(ctx->socket, TCSANOW, &tio);
    
    return 0;
}

int dsConnectionConnectCOM(DSConnectionContext ctx, DSSignalMask sigmask, const char *device, DSConnectionCOMOptions options) {
    int err;
    
    ds_assert_param(ctx);
    ds_assert_param(device);
    
    err = dsConnectionConnectFile(ctx, sigmask, device);
    if (err) return err;

    ctx->subtype = DS_CONNECTION_SUBTYPE_UART;
    
	// Do not configure port if options not supplied
    if (options) {    
	err = dsConnectionConfigureCOM(ctx, options);
        if (err) dsConnectionFreeContext(ctx);	
    }
	
    return err;
}
#else
int dsConnectionConnectCOM(DSConnectionContext ctx, DSSignalMask sigmask, const char *device, DSConnectionCOMOptions options) {
    return DS_ERR_NOTSUPPORTED;
}
#endif /* ! HAVE_TERMIOS_H */

#ifdef _WIN32 
int dsConnectionInitHandleContext(DSConnectionContext ctx, DSSignalMask sigmask, HANDLE handle) {
    dsDebugServer(DS_DEBUG_INFO, "dsConnectionInitHandleContext");    

    ds_assert_param(ctx);
    
    dsConnectionZeroContext(ctx);

    if (sigmask) {
	ctx->sigmask = (DSSignalMask)dsMalloc(ctx->malloc, sizeof(DSSignalMaskS));
	if (!ctx->sigmask) return DS_ERR_MEM;
	memcpy(ctx->sigmask, sigmask, sizeof(DSSignalMaskS));
    }

    ctx->hEventRead = CreateEvent(NULL, TRUE, FALSE, NULL);
    ctx->hEventWrite = CreateEvent(NULL, TRUE, FALSE, NULL);

    if (!ctx->hEventRead||!ctx->hEventWrite) {
	if (ctx->sigmask) dsFree(ctx->malloc, ctx->sigmask);
	
	if (ctx->hEventRead) {
	    CloseHandle(ctx->hEventRead);
	    ctx->hEventRead = NULL;
	}
	if (ctx->hEventWrite) {
	    CloseHandle(ctx->hEventWrite);
	    ctx->hEventWrite = NULL;
	}
	CloseHandle(handle);
	return DS_ERR_WINAPI_EVENT_CREATE;
    }

    ctx->handle = handle;
    ctx->read = dsConnectionReadHandle;
    ctx->write = dsConnectionWriteHandle;

    ctx->type = DS_CONNECTION_TYPE_NAMED_PIPE;
    ctx->status = DS_CONNECTION_STATUS_CONNECTED;

    return 0;
}
#endif /* _WIN32 */

void dsConnectionDisconnect(DSConnectionContext ctx) {
    dsConnectionFreeContext(ctx);
}

void dsConnectionFreeContext(DSConnectionContext ctx) {
    dsDebugServer(DS_DEBUG_INFO, "dsConnectionFreeContext");    

    ds_assert_return(ctx);

    if (ctx->sigmask) dsFree(ctx->malloc, ctx->sigmask);
    if (ctx->data) dsFree(ctx->malloc, ctx->data);

#ifdef _WIN32
    if (ctx->hEvent) CloseHandle(ctx->hEvent);
    if (ctx->hEventWrite) CloseHandle(ctx->hEventWrite);
    if (ctx->hEventRead) CloseHandle(ctx->hEventRead);
    if (ctx->winsock != INVALID_SOCKET) closesocket(ctx->winsock);
    if (ctx->handle) {
	FlushFileBuffers(ctx->handle);
	if (ctx->type == DS_CONNECTION_TYPE_NAMED_PIPE) 
	    DisconnectNamedPipe(ctx->handle);
	CloseHandle(ctx->handle);
    }
#else /* _WIN32 */
    if (ctx->socket) close(ctx->socket);
#endif /* _WIN32 */

    dsConnectionZeroContext(ctx);
}

void dsConnectionDestroyContext(DSConnectionContext ctx) {
    dsDebugServer(DS_DEBUG_INFO, "dsConnectionDestroyContext");    

    ds_assert_return(ctx);

    dsConnectionFreeContext(ctx);

//    if (ctx->timeout) dsDurationDestroy(ctx->timeout);
    dsFree(ctx->malloc,ctx);
}



#ifdef HAVE_SOCKETS
static inline int dsConnectionWaitConnection(DSConnectionContext ctx) {
#ifndef _WIN32
    int err = 0;
    socklen_t errsize = sizeof(int);
#endif /* _WIN32 */

    if (!ctx->status) return 0;
#ifdef _WIN32
    ds_assert_err(ctx->status == DS_CONNECTION_STATUS_CONNECTED, DS_ERR_CONNECTION_NOTCONNECTED);
#else /* _WIN32 */
    else if (ctx->status != DS_CONNECTION_STATUS_CONNECTING) return DS_ERR_CONNECTION_NOTCONNECTED;
    
    err = dsSignalWaitWrite(ctx, ctx->sigmask, ctx->softsig, ctx->timeout);
    if (err) return err;
    
    if ((getsockopt(ctx->socket, SOL_SOCKET, SO_ERROR, &err, &errsize))||(err)) {
	dsConnectionFreeContext(ctx);
	if (err) {
	    if (errno == ECONNREFUSED) return DS_ERR_CONNECTION_REFUSED;
	    else if (errno == ETIMEDOUT) return DS_ERR_CONNECTION_TIMEOUT;
	    else if (errno == ENETUNREACH) return DS_ERR_CONNECTION_NETWORK;
	    else if ((errno == EACCES)||(errno == EPERM)) return DS_ERR_PERMISSION;
	    return DS_ERR_SOCKET_CONNECT;
	} 
	return DS_ERR_SOCKET_GETSOCKOPT;
    }
    ctx->status = DS_CONNECTION_STATUS_CONNECTED;
#endif /* _WIN32 */
    
    return 0;
}

static inline int dsConnectionCheckConnection(DSConnectionContext ctx) {
#ifndef _WIN32
    int err = 0;
    socklen_t errsize = sizeof(int);
#endif /* _WIN32 */

    if (ctx->status == DS_CONNECTION_STATUS_CONNECTED) return 0;
#ifdef _WIN32
    ds_assert_err(ctx->status == DS_CONNECTION_STATUS_CONNECTED, DS_ERR_CONNECTION_NOTCONNECTED);
#else /* _WIN32 */
    else if (ctx->status != DS_CONNECTION_STATUS_CONNECTING) return DS_ERR_CONNECTION_NOTCONNECTED;
    
    err = dsSignalCheckWrite(ctx);
    if (err) return DS_ERR_CONNECTION_NOTCONNECTED;
    
    if ((getsockopt(ctx->socket, SOL_SOCKET, SO_ERROR, &err, &errsize))||(err)) {
	dsConnectionFreeContext(ctx);
	if (err) {
	    if (errno == ECONNREFUSED) return DS_ERR_CONNECTION_REFUSED;
	    else if (errno == ETIMEDOUT) return DS_ERR_CONNECTION_TIMEOUT;
	    else if (errno == ENETUNREACH) return DS_ERR_CONNECTION_NETWORK;
	    else if ((errno == EACCES)||(errno == EPERM)) return DS_ERR_PERMISSION;
	    return DS_ERR_SOCKET_CONNECT;
	} 
	return DS_ERR_SOCKET_GETSOCKOPT;
    }
    
    ctx->status = DS_CONNECTION_STATUS_CONNECTED;
#endif /* _WIN32 */

    return 0;
}
#endif /* HAVE_SOCKETS */

DSConnectionStatus dsConnectionCheck(DSConnectionContext ctx) {
    ds_assert_param(ctx);

#ifdef HAVE_SOCKETS
    if (ctx->socket) {
	dsConnectionCheckConnection(ctx);
	/* DS: check for broken pipe */
    }
#endif /* HAVE_SOCKETS */

    return ctx->status;
}


#ifdef _WIN32
static int dsConnectionReadHandle(DSConnectionContext ctx, void *buf, size_t len, DSConnectionFlags flags, size_t *rbytes) {
    int err;
    unsigned long readed = 0;
    HANDLE handle;
    OVERLAPPED oOverlap;
    DWORD rb = 0;
    
    handle = ctx->handle;
    ds_assert_param(handle != INVALID_HANDLE_VALUE);
    ds_assert_err(ctx->status == DS_CONNECTION_STATUS_CONNECTED, DS_ERR_CONNECTION_NOTCONNECTED);

    if (flags&DS_CONNECTION_FLAG_TRY) {
	if (!PeekNamedPipe(handle, NULL, 0, NULL, &rb, NULL)) return DS_ERR_WINAPI;

	if (flags&DS_CONNECTION_FLAG_EXAMINE) {
	    if (rbytes) *rbytes = rb;
	    return rb?0:DS_ERR_TIMEOUT;
	}
	
	if (!rb) return DS_ERR_TIMEOUT;
    }  

    if (flags&DS_CONNECTION_FLAG_EXAMINE) {
	if (rbytes) { 
	    if (!PeekNamedPipe(handle, NULL, 0, NULL, &rb, NULL)) return DS_ERR_WINAPI;
	    *rbytes = rb;
	}
	/* DS: Implement some kind of Waiting? */
	return 0;
    }
    
    memset(&oOverlap, 0, sizeof(OVERLAPPED));
    oOverlap.hEvent = ctx->hEventRead;
    ResetEvent(oOverlap.hEvent);

read_again:
    
    if (!ReadFile(handle, (char*)buf + readed, (DWORD)(len - readed), &rb, &oOverlap)) {
	err = GetLastError();

	if (err == ERROR_IO_PENDING) {
	    if ((flags&DS_CONNECTION_FLAG_TRY) == 0) {
		err = dsSignalWinWait(ctx->sigmask, ctx->hEventRead, ctx->timeout);
		if (err) {
		    while (!CancelIo(handle)) {
			dsDebugRequestSystemError(DS_DEBUG_ERROR, 0, "dsConnection(CancelIO failed): ", ctx); 
		    }
		    if (err != DS_ERR_SIGNAL_EVENT) {
			/* DS: Currently no other errors. Otherwise in some cases
			probably should be checked for partly context */
			if (rbytes) *rbytes = readed;
			return err;
		    }
		}
	    }
	    if (!GetOverlappedResult(handle, &oOverlap, &rb, TRUE)) {
		err = GetLastError();
		if (err == ERROR_OPERATION_ABORTED) {
		    /* DS: Am I sure rb would be correct in this case? */
		    readed += rb;
		    if (rbytes) *rbytes = readed;
		    return DS_ERR_SOCKET_READ;
		}

		while (!CancelIo(handle)) {
		    dsDebugRequestSystemError(DS_DEBUG_ERROR, 0, "dsConnection(CancelIO failed): ", ctx); 
		}

		if (err == ERROR_IO_INCOMPLETE) {
		    /* DS: Am I sure rb will be correct in this case? */
		    GetOverlappedResult(handle, &oOverlap, &rb, TRUE);
		    readed += rb;
		    if (rbytes) *rbytes = readed;
		    
		    return DS_ERR_TIMEOUT;
		} 
		
		if (rbytes) *rbytes = readed;
		return DS_ERR_WINAPI;
	    }
	} else {
	    if (rbytes) *rbytes = readed;
	    return DS_ERR_SOCKET_READ;
	}
    }

    readed += rb;

    if ((flags&DS_CONNECTION_FLAG_EXACTREAD)&&(readed < len)) {
	if (flags&DS_CONNECTION_FLAG_TRY) {
	    if (!PeekNamedPipe(handle, NULL, 0, NULL, &rb, NULL)) {
		if (rbytes) *rbytes = readed;
		return DS_ERR_WINAPI;
	    }

	    if (!rb) {
		if (rbytes) *rbytes = readed;
		return DS_ERR_TIMEOUT;
	    }
	}

	goto read_again;
    }
    
    if (rbytes) *rbytes=readed;
    
    return 0;
}


static int dsConnectionWriteHandle(DSConnectionContext ctx, const void *buf, size_t len, int flags, size_t *wbytes) {
    int err;
    unsigned long writed = 0;
    HANDLE handle;
    OVERLAPPED oOverlap;
    DWORD wr = 0;
    
    handle = ctx->handle;
    ds_assert_param(handle != INVALID_HANDLE_VALUE);
    ds_assert_err(ctx->status == DS_CONNECTION_STATUS_CONNECTED, DS_ERR_CONNECTION_NOTCONNECTED);

    if (flags&DS_CONNECTION_FLAG_EXAMINE) {
	if (wbytes) *wbytes = 0;
	return 0;
    }

    memset(&oOverlap, 0, sizeof(OVERLAPPED));
    oOverlap.hEvent = ctx->hEventWrite;
    ResetEvent(oOverlap.hEvent);

write_again:    
    if (!WriteFile(handle, (char*)buf + writed, (DWORD)(len - writed), &wr, &oOverlap)) {
	err = GetLastError();

	/* Fucking windows don't let to check if there are possibility to 
	send data. Should check and cancel if no. No other ideas. */
	if (err == ERROR_IO_PENDING) {
	    if ((flags&DS_CONNECTION_FLAG_TRY) == 0) {
		err = dsSignalWinWait(ctx->sigmask,ctx->hEventWrite,ctx->timeout);
		if (err) {
		    while (!CancelIo(handle)) {
			dsDebugRequestSystemError(DS_DEBUG_ERROR, 0, "dsConnection(CancelIO failed): ", ctx); 
		    }
		    if (err != DS_ERR_SIGNAL_EVENT) {
			/* DS: Currently no other errors. Otherwise in some cases
			probably should be checked for partly context */
			if (wbytes) *wbytes = writed;
			return err;
		    }
		}
	    }
	    if (!GetOverlappedResult(handle, &oOverlap, &wr, TRUE)) {
		err = GetLastError();
		if (err == ERROR_OPERATION_ABORTED) {
		    /* DS: Am I sure rb would be correct in this case? */
		    writed += wr;
		    if (wbytes) *wbytes = writed;
		    return DS_ERR_SOCKET_READ;
		}

		while (!CancelIo(handle)) {
		    dsDebugRequestSystemError(DS_DEBUG_ERROR, 0, "dsConnection(CancelIO failed): ", ctx); 
		}
		
		if (err == ERROR_IO_INCOMPLETE) {
		    GetOverlappedResult(handle, &oOverlap, &wr, TRUE);
		    /* DS: Am I sure rb will be correct in this case? */
		    if (wbytes) *wbytes = wr + writed;
		    return DS_ERR_TIMEOUT;
		} 

		if (wbytes) *wbytes = writed;
		return DS_ERR_WINAPI;
	    }
	} else {
	    if (wbytes) *wbytes = writed;
	    return DS_ERR_SOCKET_WRITE;
	}
    }
    writed += wr;
    if (writed < len) goto write_again;

    if (wbytes) *wbytes=writed;
    
    return 0;
}
#else /* _WIN32 */

/* If the file was disconnected, then select will return "Ok, ready", and read request will return 0 (bytes read), errno is
also 0. Such situation could happen in the following cases:
    a) USB COM device unplugged from USB port
    b) USB/Network media file persist on is disconnected
We will return DS_ERR_CONNECTION_BROKEN_PIPE in that case.
*/

static int dsConnectionReadFile(DSConnectionContext ctx, void *buf, size_t len, DSConnectionFlags flags, size_t *rbytes) {
    DSConnectionStatus status;
    int err;
    int s;

    unsigned long readed = 0;
    
    ds_assert_param(s = ctx->socket);
    
select_again:    
    if (flags&DS_CONNECTION_FLAG_TRY) err = dsSignalCheckRead(ctx);
    else if ((ctx->sigmask)||(ctx->timeout)) err = dsSignalWaitRead(ctx, ctx->sigmask, ctx->softsig, ctx->timeout);
    else err = 0;
    if (err) {
	if (rbytes) *rbytes = readed;
	return err;
    }
    
    if (flags&DS_CONNECTION_FLAG_EXAMINE) {
	if (rbytes) *rbytes = 0;
	return 0;
    }

    err = read(s, buf + readed, len - readed);
    if (err<=0) {
	if (err) err = DS_ERR_SOCKET_READ;
	else err = DS_ERR_CONNECTION_BROKENPIPE;
	
#if defined(EAGAIN) && defined(EWOULDBLOCK) && EAGAIN != EWOULDBLOCK
	if ((errno==EAGAIN)||(errno==EWOULDBLOCK)) {
#elif defined(EAGAIN)
	if (errno==EAGAIN) {
#elif defined(EWOULDBLOCK)
	if (errno==EWOULDBLOCK) {
#endif /* EAGAIN, EWOULDBLOCK */
	    if (flags&DS_CONNECTION_FLAG_TRY) err = DS_ERR_TIMEOUT;
	    else goto select_again;
	}

	if (rbytes) *rbytes = readed;
	return err;
    } else readed += err;

    if ((readed<len)&&(flags&DS_CONNECTION_FLAG_EXACTREAD)) goto select_again;
    
    if (rbytes) *rbytes = readed;
    
    return 0;
}

static int dsConnectionWriteFile(DSConnectionContext ctx, const void *buf, size_t len, DSConnectionFlags flags, size_t *wbytes) {
    DSConnectionStatus status;
    int err;
    int s;
    unsigned long writed = 0;
    
    ds_assert_param(s = ctx->socket);

select_again:    
    if (flags&DS_CONNECTION_FLAG_TRY) err = dsSignalCheckWrite(ctx);
    else if ((ctx->sigmask)||(ctx->timeout)) err = dsSignalWaitWrite(ctx, ctx->sigmask, ctx->softsig, ctx->timeout);
    else err = 0;
    
    if (err) {
	if (wbytes) *wbytes = writed;
	return err;
    }

    if (flags&DS_CONNECTION_FLAG_EXAMINE) {
	if (wbytes) *wbytes = 0;
	return 0;
    }
    
    err=write(s,(char*)buf+writed,len-writed);
    if (err<0) {
	err = DS_ERR_SOCKET_READ;

#if defined(EAGAIN) && defined(EWOULDBLOCK) && EAGAIN != EWOULDBLOCK
	if ((errno==EAGAIN)||(errno==EWOULDBLOCK)) {
#elif defined(EAGAIN)
	if (errno==EAGAIN) {
#elif defined(EWOULDBLOCK)
	if (errno==EWOULDBLOCK) {
#endif /* EAGAIN, EWOULDBLOCK */
	    if (flags&DS_CONNECTION_FLAG_TRY) err = DS_ERR_TIMEOUT;
	    else goto select_again;
	}

	if (wbytes) *wbytes = writed;
	return err;
    } else writed += err;
    
    if (writed<len) goto select_again;
    
    if (wbytes) *wbytes=writed;

    return 0;
}
#endif /* _WIN32 */

static int dsConnectionReadCFile(DSConnectionContext ctx, void *buf, size_t len, DSConnectionFlags flags, size_t *rbytes) {
    return DS_ERR_NOTIMPLEMENTED;
}

static int dsConnectionWriteCFile(DSConnectionContext ctx, const void *buf, size_t len, DSConnectionFlags flags, size_t *wbytes) {
    return DS_ERR_NOTIMPLEMENTED;
}

int dsConnectionRead(DSConnectionContext ctx, void *buf, size_t len, DSConnectionFlags flags, size_t *rbytes) {
#ifdef HAVE_SOCKETS
    int err;
    unsigned long readed = 0;

    DSConnectionStatus status;
    DSConnectionUDPData udp;
    
# ifdef _WIN32
    SOCKET s;

    SOCKADDR_IN remote_socket;
    socklen_t remote_socket_size = sizeof(SOCKADDR_IN);
# else /* _WIN32 */
    int s;

    struct sockaddr_in remote_socket;
    socklen_t remote_socket_size = sizeof(struct sockaddr_in);
# endif /* _WIN32 */
#endif /* HAVE_SOCKETS */

    ds_assert_param(ctx);

    if ((flags&DS_CONNECTION_FLAG_OVERRIDEDEFAULT)==0) {
	if (flags&DS_CONNECTION_FLAG_TRY) flags |= ctx->flags&(~DS_CONNECTION_FLAG_EXACTREAD);
	else flags |= ctx->flags;
    }
    
    if ((flags&DS_CONNECTION_FLAG_EXAMINE) == 0) {
	ds_assert_param(buf);
    }

    if (ctx->read) return ctx->read(ctx, buf, len, flags, rbytes);
#ifndef HAVE_SOCKETS
    return DS_ERR_NOTIMPLEMENTED;
#else /* ! HAVE_SOCKETS */
# ifdef _WIN32
    ds_assert_param(s = ctx->winsock);
# else /* _WIN32 */
    ds_assert_param(s = ctx->socket);
# endif /* _WIN32 */
    
    status = (flags&DS_CONNECTION_FLAG_TRY)?dsConnectionCheckConnection(ctx):dsConnectionWaitConnection(ctx);
    if (status) {
	if (status == DS_CONNECTION_STATUS_CONNECTING) return DS_ERR_SIGNAL_EVENT;
	return DS_ERR_CONNECTION_NOTCONNECTED;
    }

    if (ctx->subtype == DS_CONNECTION_SUBTYPE_UDP) udp = (DSConnectionUDPData)ctx->data;
    else udp = NULL;
    
select_again:    
    if ((udp)&&(udp->size)){
	if (flags&DS_CONNECTION_FLAG_EXAMINE) {
	    if (rbytes) *rbytes = udp->size;
	    return 0;
	}
    } else {
	if (flags&DS_CONNECTION_FLAG_TRY) err = dsSignalCheckRead(ctx);
	else if ((ctx->sigmask)||(ctx->timeout)) err = dsSignalWaitRead(ctx, ctx->sigmask, ctx->softsig, ctx->timeout);
	else err = 0;

	if (err) {
	    if (rbytes) *rbytes = readed;
	    return err;
	}
    
	if (flags&DS_CONNECTION_FLAG_EXAMINE) {
	    if (rbytes) *rbytes = 0;
	    return 0;
	}
    }
    
    if (udp) {
	if (udp->size) {
	    if (len - readed < udp->size) {
		err = len - readed;
		memcpy((char*)buf + readed, udp->buffer + udp->start, err);
		udp->start += err;
		udp->size -= err;
	    } else {
	        err = udp->size;
		memcpy((char*)buf + readed, udp->buffer + udp->start, err);
		udp->size = 0;
	    }
	} else {
# ifdef _WIN32
	    err = recvfrom(s, udp->buffer, DS_MAX_UDP_MESSAGE_SIZE, 0, (struct sockaddr*)&remote_socket, &remote_socket_size);
# else /* _WIN32 */
	    err = recvfrom(s, udp->buffer, DS_MAX_UDP_MESSAGE_SIZE, (flags&DS_CONNECTION_FLAG_TRY)?MSG_DONTWAIT:0, &remote_socket, &remote_socket_size);
# endif /* _WIN32 */

# ifdef _WIN32
	    if ((err != SOCKET_ERROR)&&(err>0)) {
# else /* _WIN32 */
	    if (err > 0) { 
# endif /* _WIN32 */
		if (udp->fixed) {
		    if (dsConnectionAddressCompare(&udp->remote, DS_CONNECTION_ADDRESS_TYPE_SOCKADDR, remote_socket_size, (char*)&remote_socket)) goto select_again;
		} else {
		    dsConnectionAddressSet(&udp->remote, DS_CONNECTION_ADDRESS_TYPE_SOCKADDR, remote_socket_size, (char*)&remote_socket);
		}
		
	        if ((unsigned int)err > len - readed) {
	    	    udp->size = (ds_udp_size_t)(err - (len - readed));
	    	    err = len - readed;
		    memcpy((char*)buf + readed, udp->buffer, err);
		    udp->start = err;
		} else
		    memcpy((char*)buf + readed, udp->buffer, err);
	    }
	}
    } else {
# ifdef _WIN32
	err = recv(s, (char*)buf + readed, len - readed, 0);
# else /* _WIN32 */
	err = recv(s, (char*)buf + readed, len - readed, (flags&DS_CONNECTION_FLAG_TRY)?MSG_DONTWAIT:0);
# endif /* _WIN32 */
    }
    
# ifdef _WIN32
    if (err == SOCKET_ERROR) {
	err = WSAGetLastError();
	if (err == WSAECONNRESET) {
	    WSAResetEvent(ctx->hEventRead);
	    err = WSAEWOULDBLOCK;
	}
	
	if (flags&DS_CONNECTION_FLAG_TRY) {
	    if (err==WSAEWOULDBLOCK) err = DS_ERR_TIMEOUT;
	    else err = DS_ERR_SOCKET_READ;
	} else {
	    if (err==WSAEWOULDBLOCK) {
		WSAResetEvent(ctx->hEventRead);
		goto select_again;
	    } else err = DS_ERR_SOCKET_READ;
	}
	if (rbytes) *rbytes = readed;
	return err;
# else /* _WIN32 */
    if (err<0) {
	err = DS_ERR_SOCKET_READ;
	
#  if defined(EAGAIN) && defined(EWOULDBLOCK) && EAGAIN != EWOULDBLOCK
	if ((errno==EAGAIN)||(errno==EWOULDBLOCK)) {
#  elif defined(EAGAIN)
	if (errno==EAGAIN) {
#  elif defined(EWOULDBLOCK)
	if (errno==EWOULDBLOCK) {
#  endif /* EAGAIN, EWOULDBLOCK */
	    if (flags&DS_CONNECTION_FLAG_TRY) err = DS_ERR_TIMEOUT;
	    else goto select_again;
	}

	if (rbytes) *rbytes = readed;
	return err;
# endif /* _WIN32 */
    } else readed += err;

    if ((readed<len)&&(flags&DS_CONNECTION_FLAG_EXACTREAD)) goto select_again;
    
    if (rbytes) *rbytes = readed;
    
    return 0;
#endif /* ! HAVE_SOCKETS */
}

int dsConnectionWrite(DSConnectionContext ctx, const void *buf, size_t len, DSConnectionFlags flags, size_t *wbytes) {
    DSConnectionAddress *override;

#ifdef HAVE_SOCKETS
    int err;
    unsigned long writed = 0;

    DSConnectionStatus status;

# ifdef _WIN32
    SOCKET s;
# else /* _WIN32 */
    int s;
# endif /* _WIN32 */
#endif /* HAVE_SOCKETS */

    
    ds_assert_param(ctx);

    if ((flags&DS_CONNECTION_FLAG_OVERRIDEDEFAULT)==0) flags |= ctx->flags; 

    if ((flags&DS_CONNECTION_FLAG_EXAMINE) == 0) {
	ds_assert_param(buf);
    }
    
    if (ctx->write) return ctx->write(ctx, buf, len, flags, wbytes);
#ifndef HAVE_SOCKETS
    return DS_ERR_NOTIMPLEMENTED;
#else /* ! HAVE_SOCKETS */
#  ifdef _WIN32
    ds_assert_param(s = ctx->winsock);
#  else /* _WIN32 */
    ds_assert_param(s = ctx->socket);
#  endif /* _WIN32 */


    status = (flags&DS_CONNECTION_FLAG_TRY)?dsConnectionCheckConnection(ctx):dsConnectionWaitConnection(ctx);
    if (status) {
	if (status == DS_CONNECTION_STATUS_CONNECTING) return DS_ERR_SIGNAL_EVENT;
	return DS_ERR_CONNECTION_NOTCONNECTED;
    }

    if ((ctx->subtype == DS_CONNECTION_SUBTYPE_UDP)&&(ctx->data)&&(!((DSConnectionUDPData)(ctx->data))->fixed)) {
	override = ((DSConnectionUDPData)(ctx->data))->override;
	if ((override)&&(override->type != DS_CONNECTION_ADDRESS_TYPE_SOCKADDR)) return DS_ERR_CONNECTION_ADDRESS_INVALID;
    } else override = NULL;

select_again:    
    if (flags&DS_CONNECTION_FLAG_TRY) err = dsSignalCheckWrite(ctx);
    else if ((ctx->sigmask)||(ctx->timeout)) err = dsSignalWaitWrite(ctx, ctx->sigmask, ctx->softsig, ctx->timeout);
    else err = 0;
    
    if (err) {
	if (wbytes) *wbytes = writed;
	return err;
    }

    if (flags&DS_CONNECTION_FLAG_EXAMINE) {
	if (wbytes) *wbytes = 0;
	return 0;
    }

    if (override) {
	err = sendto(s,(char*)buf+writed,len-writed,0,(struct sockaddr*)&override->addr, override->size);
    } else  {
	err = send(s,(char*)buf+writed,len-writed,0);
    }

# ifdef _WIN32
    if (err == SOCKET_ERROR) {
	err = WSAGetLastError();
	if (flags&DS_CONNECTION_FLAG_TRY) {
	    if (err==WSAEWOULDBLOCK) err = DS_ERR_TIMEOUT;
	    else err = DS_ERR_SOCKET_WRITE;
	} else {
	    if (err==WSAEWOULDBLOCK) goto select_again;
	    else err = DS_ERR_SOCKET_WRITE;
	}
	if (wbytes) *wbytes = writed;
	return err;
# else /* _WIN32 */
    if (err<0) {
	err = DS_ERR_SOCKET_WRITE;

#  if defined(EAGAIN) && defined(EWOULDBLOCK) && EAGAIN != EWOULDBLOCK
	if ((errno==EAGAIN)||(errno==EWOULDBLOCK)) {
#  elif defined(EAGAIN)
	if (errno==EAGAIN) {
#  elif defined(EWOULDBLOCK)
	if (errno==EWOULDBLOCK) {
#  endif /* EAGAIN, EWOULDBLOCK */
	    if (flags&DS_CONNECTION_FLAG_TRY) err = DS_ERR_TIMEOUT;
	    else goto select_again;
	}

	if (wbytes) *wbytes = writed;
	return err;
# endif /* _WIN32 */
    } else writed += err;
    
    if (writed<len) goto select_again;
    
    if (wbytes) *wbytes=writed;

    return 0;
#endif /* ! HAVE_SOCKETS */
}


int dsConnectionSetSignal(DSConnectionContext ctx, DSConnectionSignal sig, DSDuration duration) {
#ifdef HAVE_TERMIOS_H
    int flags;
    
    ds_assert_param(ctx);

    if ((ctx->type == DS_CONNECTION_TYPE_FILE)&&(ctx->subtype == DS_CONNECTION_SUBTYPE_UART)) {
	switch (sig) {
	    case DS_CONNECTION_SIGNAL_RTS:
		if (duration) return DS_ERR_NOTSUPPORTED;

	    	ioctl(ctx->socket, TIOCMGET, &flags);
		if ((flags&TIOCM_RTS)==0) {
		    flags |= TIOCM_RTS;
		    ioctl(ctx->socket, TIOCMSET, &flags);
		}
		return 0;
	    case DS_CONNECTION_SIGNAL_DTR:
		if (duration) return DS_ERR_NOTSUPPORTED;

	    	ioctl(ctx->socket, TIOCMGET, &flags);
		if ((flags&TIOCM_DTR)==0) {
		    flags |= TIOCM_DTR;
		    ioctl(ctx->socket, TIOCMSET, &flags);
		}
		return 0;
	    case DS_CONNECTION_SIGNAL_BREAK:
		if (duration) return DS_ERR_NOTSUPPORTED;
	    
		tcsendbreak(ctx->socket, 0);
		return 0;
	    default:
		return DS_ERR_NOTSUPPORTED;
	}
    }
#endif /* HAVE_TERMIOS_H */

    return DS_ERR_NOTSUPPORTED;
}

int dsConnectionResetSignal(DSConnectionContext ctx, DSConnectionSignal sig, DSDuration duration) {
#ifdef HAVE_TERMIOS_H
    int flags;
    
    ds_assert_param(ctx);
    
    if ((ctx->type == DS_CONNECTION_TYPE_FILE)&&(ctx->subtype == DS_CONNECTION_SUBTYPE_UART)) {
	switch (sig) {
	    case DS_CONNECTION_SIGNAL_RTS:
		if (duration) return DS_ERR_NOTSUPPORTED;

	    	ioctl(ctx->socket, TIOCMGET, &flags);
		if (flags&TIOCM_RTS) {
		    flags &= ~TIOCM_RTS;
		    ioctl(ctx->socket, TIOCMSET, &flags);
		}
		return 0;
	    case DS_CONNECTION_SIGNAL_DTR:
		if (duration) return DS_ERR_NOTSUPPORTED;

	    	ioctl(ctx->socket, TIOCMGET, &flags);
		if (flags&TIOCM_DTR) {
		    flags &= ~TIOCM_DTR;
		    ioctl(ctx->socket, TIOCMSET, &flags);
		}
		return 0;
	    default:
		return DS_ERR_NOTSUPPORTED;
	}
    }
#endif /* HAVE_TERMIOS_H */

    return DS_ERR_NOTSUPPORTED;
}

int dsConnectionCheckSignal(DSConnectionContext ctx, DSConnectionSignal sig) {
#ifdef HAVE_TERMIOS_H
    int flags;
    
    ds_assert_param(ctx);
    
    if ((ctx->type == DS_CONNECTION_TYPE_FILE)&&(ctx->subtype == DS_CONNECTION_SUBTYPE_UART)) {
	switch (sig) {
	    case DS_CONNECTION_SIGNAL_RTS:
	    	ioctl(ctx->socket, TIOCMGET, &flags);
		return (flags&TIOCM_RTS);
	    case DS_CONNECTION_SIGNAL_DTR:
	    	ioctl(ctx->socket, TIOCMGET, &flags);
		return (flags&TIOCM_DTR);
	    break;
	    default:
		return DS_ERR_NOTSUPPORTED;
	}
    }
#endif /* HAVE_TERMIOS_H */

    return -1;
}
