#ifndef _DS_CONNECTION_H
#define _DS_CONNECTION_H

#include <dslib.h>
#include <dsmalloc.h>

#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif /* HAVE_NETINET_IN_H */

#define DS_CONNECTION_MAX_ADDRESS_SIZE 	64
#define DS_MAX_UDP_MESSAGE_SIZE 8192
#define DS_DEFAULT_IO_BUFFER_SIZE 65535
#define DS_DEFAULT_COM_SPEED 38400

typedef struct DSConnectionContextT *DSConnectionContext;

/* TryRead ignores EXACTREAD flag */
typedef int DSConnectionFlags;
#define DS_CONNECTION_FLAG_OVERRIDEDEFAULT 1
#define DS_CONNECTION_FLAG_EXAMINE 2
#define DS_CONNECTION_FLAG_TRY 4
#define DS_CONNECTION_FLAG_EXACTREAD 8
//#define DS_CONNNECTION_FLAG_OOB 16


typedef int (*DSConnectionReadFunction)(DSConnectionContext ctx, void *buf, size_t len, DSConnectionFlags flags, size_t *rbytes);
typedef int (*DSConnectionWriteFunction)(DSConnectionContext ctx, const void *buf, size_t len, DSConnectionFlags flags, size_t *wbytes);

enum DSConnectionStatusT {
    DS_CONNECTION_STATUS_CONNECTED = 0,
    DS_CONNECTION_STATUS_CONNECTING,
    DS_CONNECTION_STATUS_BROKENPIPE,
    DS_CONNECTION_STATUS_UNCONNECTED
};
typedef enum DSConnectionStatusT DSConnectionStatus;

enum DSConnectionTypeT {
    DS_CONNECTION_TYPE_UNKNOWN = 0,
    DS_CONNECTION_TYPE_SOCKET,
    DS_CONNECTION_TYPE_WINSOCK,
    DS_CONNECTION_TYPE_NAMED_PIPE,
    DS_CONNECTION_TYPE_NAMED_PIPE_CLIENT,
    DS_CONNECTION_TYPE_FILE
};
typedef enum DSConnectionTypeT DSConnectionType;

enum DSConnectionSubTypeT {
    DS_CONNECTION_SUBTYPE_DEFAULT = 0,
    DS_CONNECTION_SUBTYPE_UDP,
    DS_CONNECTION_SUBTYPE_UART
};
typedef enum DSConnectionSubTypeT DSConnectionSubType;

#include <dssignal.h>
#include <dssoftsignal.h>

typedef void *DSConnectionData;
    
struct DSConnectionContextT {
    DSMallocContext malloc;
    DSSignalMask sigmask;
    DSSoftSignalList softsig;
    DSDurationS timeouts;

    int socket;
    DSConnectionStatus status;
    DSConnectionType type;
    DSConnectionSubType subtype;
    DSDuration timeout;
    DSConnectionFlags flags;
    
    DSConnectionData data;
//    char *udp;
//    unsigned short int udp_start, udp_size;

#ifdef _WIN32
    HANDLE handle;

    SOCKET winsock;
    HANDLE hEvent;
    HANDLE hEventRead;
    HANDLE hEventWrite;
#endif /* _WIN32 */

    DSConnectionReadFunction read;
    DSConnectionWriteFunction write;
};
typedef struct DSConnectionContextT DSConnectionContextS;

struct DSConnectionOptionsT {
    DSDurationS timeout;
    int flags;
    unsigned char exact_read;
};
typedef struct DSConnectionOptionsT DSConnectionOptionsS;


struct DSConnectionCOMOptionsT {
    unsigned long speed;
};
typedef struct DSConnectionCOMOptionsT DSConnectionCOMOptionsS;
typedef struct DSConnectionCOMOptionsT *DSConnectionCOMOptions;

DSConnectionCOMOptions dsConnectionCOMGetDefaultOptions();
int dsConnectionCOMInitDefaultOptions(DSConnectionCOMOptions options);
void dsConnectionCOMDestroyOptions(DSConnectionCOMOptions options);
int dsConnectionCOMOptionsSetSpeed(DSConnectionCOMOptions options, unsigned long speed);

enum DSConnectionAddressTypeS {
    DS_CONNECTION_ADDRESS_TYPE_SOCKADDR = 0
};
typedef enum DSConnectionAddressTypeS DSConnectionAddressType;

struct DSConnectionAddressT {
    DSConnectionAddressType type;
    size_t size;
    char *addr[DS_CONNECTION_MAX_ADDRESS_SIZE];
};
typedef struct DSConnectionAddressT DSConnectionAddress;
typedef struct DSConnectionAddressT *DSConnectionAddressPtr;

typedef unsigned short int ds_udp_size_t;
struct DSConnectionUDPDataT {
    char buffer[DS_MAX_UDP_MESSAGE_SIZE];
    ds_udp_size_t start, size;

    unsigned char fixed;

    DSConnectionAddress remote;
    DSConnectionAddress *override;
};

typedef struct DSConnectionUDPDataT *DSConnectionUDPData;
typedef struct DSConnectionUDPDataT DSConnectionUDPDataS;

int dsConnectionAddressSet(DSConnectionAddress *ctx, DSConnectionAddressType type, size_t size, const char *addr);
int dsConnectionAddressCompare(DSConnectionAddress *ctx, DSConnectionAddressType type, size_t size, const char *addr);
int dsConnectionAddressCpy(DSConnectionAddress *ctx1, DSConnectionAddress *ctx2);
int dsConnectionAddressCmp(DSConnectionAddress *ctx1, DSConnectionAddress *ctx2);

int dsConnectionInit();
void dsConnectionDestroy();

DSConnectionContext dsConnectionCreateContext(DSMallocContext mctx);
int dsConnectionCreateSContext(DSConnectionContext ctx, DSMallocContext mctx);

#ifdef HAVE_SOCKETS
# ifdef _WIN32 
int dsConnectionInitSocketContext(DSConnectionContext ctx, DSSignalMask sigmask, SOCKET sock);
# else /* _WIN32 */
int dsConnectionInitSocketContext(DSConnectionContext ctx, DSSignalMask sigmask, int sock);
# endif /* _WIN32 */
int dsConnectionConnectServer(DSConnectionContext ctx, DSSignalMask sigmask, int protocol, const char *server, in_port_t port);
int dsConnectionConnectUDPServer(DSConnectionContext ctx, DSSignalMask sigmask, const char *server, in_port_t port);
int dsConnectionConnectTCPServer(DSConnectionContext ctx, DSSignalMask sigmask, const char *server, in_port_t port);
int dsConnectionConnectSocket(DSConnectionContext ctx, DSSignalMask sigmask, const char *path, const char *name, const char *extension);
#endif /* HAVE_SOCKETS */

int dsConnectionConnectCFile(DSConnectionContext ctx, DSSignalMask sigmask, const char *file);
int dsConnectionConnectCOM(DSConnectionContext ctx, DSSignalMask sigmask, const char *device, DSConnectionCOMOptions options);
void dsConnectionDisconnect(DSConnectionContext ctx);

int dsConnectionSetTimeout(DSConnectionContext ctx, DSDuration timeout);
int dsConnectionSetDefaultFlags(DSConnectionContext ctx, DSConnectionFlags flags);
int dsConnectionConfigureCOM(DSConnectionContext ctx, DSConnectionCOMOptions options);

#ifdef _WIN32
int dsConnectionInitHandleContext(DSConnectionContext ctx, DSSignalMask sigmask, HANDLE handle);
#else /* _WIN32 */
int dsConnectionConnectFile(DSConnectionContext ctx, DSSignalMask sigmask, const char *file);
#endif /* _WIN32 */

void dsConnectionFreeContext(DSConnectionContext ctx);
void dsConnectionDestroyContext(DSConnectionContext ctx);

int dsConnectionRead(DSConnectionContext ctx, void *buf, size_t len, DSConnectionFlags flags, size_t *rbytes);
int dsConnectionWrite(DSConnectionContext ctx, const void *buf, size_t len, DSConnectionFlags flags, size_t *wbytes);
#define dsConnectionTryRead(ctx, buf, len, flags, rbytes) dsConnectionRead(ctx, buf, len, flags|DS_CONNECTION_FLAG_TRY, rbytes)
#define dsConnectionTryWrite(ctx, buf, len, flags, wbytes) dsConnectionWrite(ctx, buf, len, flags|DS_CONNECTION_FLAG_TRY, wbytes)
#define dsConnectionWaitRead(ctx, buf, len, flags, rbytes) dsConnectionRead(ctx, buf, len, flags|DS_CONNECTION_FLAG_EXAMINE, rbytes)
#define dsConnectionWaitWrite(ctx, buf, len, flags, wbytes) dsConnectionWrite(ctx, buf, len, flags|DS_CONNECTION_FLAG_EXAMINE, wbytes)
#define dsConnectionCheckRead(ctx, buf, len, flags, rbytes) dsConnectionRead(ctx, buf, len, flags|DS_CONNECTION_FLAG_EXAMINE|DS_CONNECTION_FLAG_TRY, rbytes)
#define dsConnectionCheckWrite(ctx, buf, len, flags, wbytes) dsConnectionWrite(ctx, buf, len, flags|DS_CONNECTION_FLAG_EXAMINE|DS_CONNECTION_FLAG_TRY, wbytes)

DSConnectionStatus dsConnectionCheck(DSConnectionContext ctx);


enum DSConnectionSignalT {
    DS_CONNECTION_SIGNAL_RTS = 1,	/* Ready To Send */
    DS_CONNECTION_SIGNAL_DTR,		/* Data Terminal Ready */
    DS_CONNECTION_SIGNAL_CTS,		/* Clear To Send */
    DS_CONNECTION_SIGNAL_DSR,		/* Data Set Ready */
    DS_CONNECTION_SIGNAL_BREAK,		/* tcsendbreak */	
    DS_CONNECTION_SIGNAL_LAST
};
typedef enum DSConnectionSignalT DSConnectionSignal;

int dsConnectionSetSignal(DSConnectionContext ctx, DSConnectionSignal sig, DSDuration duration);
int dsConnectionResetSignal(DSConnectionContext ctx, DSConnectionSignal sig, DSDuration duration);
int dsConnectionCheckSignal(DSConnectionContext ctx, DSConnectionSignal sig);

#endif /* _DS_CONNECTION_H */
