#include <dslib.h>
#include <dsassert.h>

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif /* HAVE_ERRNO_H */

#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif /* HAVE_SYS_SOCKET_H */
#ifdef HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif /* HAVE_NETINET_IN_H */
#ifdef HAVE_SYS_UN_H
# include <sys/un.h>
#endif /* HAVE_SYS_UN_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */

#include <dsdebug.h>
#include <dserror.h>

#include <dsconnectionserver.h>


#ifdef _WIN32
    unsigned char winsock_initialized = 0;
    static WSADATA WsaDat;
#endif /* _WIN32 */

int dsConnectionServerInit() {
#ifdef _WIN32
    if (WSAStartup(MAKEWORD(2,2), &WsaDat) != 0) return DS_ERR_WINAPI_WINSOCK_INIT;
#endif /* _WIN32 */
    return 0;
}

void dsConnectionServerDestroy() {
#ifdef _WIN32
    if (winsock_initialized) {
	WSACleanup();
	winsock_initialized = 0;
    }
#endif /* _WIN32 */
}


DSConnectionServerOptions dsConnectionServerGetDefaultOptions() {
    int err;
    DSConnectionServerOptions options;
    
    options = (DSConnectionServerOptions)malloc(sizeof(DSConnectionServerOptionsS));
    if (!options) return options;

    err = dsConnectionServerInitDefaultOptions(options);
    if (err) {
	dsConnectionServerDestroyOptions(options);
	return NULL;
    }

    return options;
}

int dsConnectionServerInitDefaultOptions(DSConnectionServerOptions options) {
    memset(options, 0, sizeof(struct DSConnectionServerOptionsT));
    dsSignalGetDefaultMask(&options->sigmask);
    return 0;
}

void dsConnectionServerDestroyOptions(DSConnectionServerOptions options) {
    if (options->local_name) dsExchangeString(NULL,&options->local_name, NULL);
    if (options->local_path) dsExchangeString(NULL,&options->local_path, NULL);
    if (options->local_extension) dsExchangeString(NULL,&options->local_extension, NULL);
    
    free(options);
}

int dsConnectionServerOptionsEnableLocal(DSConnectionServerOptions options, const char *local_name) {
    ds_assert_param(options);
    
    if (dsExchangeString(NULL,&options->local_name, local_name)) return DS_ERR_MEM;
    return 0;
}

int dsConnectionServerOptionsEnableNetwork(DSConnectionServerOptions options, in_port_t network_port) {
    ds_assert_param(options);
    
    options->network_port = network_port;
    return 0;
}

int dsConnectionServerOptionsSetLocalPath(DSConnectionServerOptions options, const char *local_path) {
    ds_assert_param(options);

    if (dsExchangeString(NULL,&options->local_path, local_path)) return DS_ERR_MEM;
    return 0;
}

int dsConnectionServerOptionsSetLocalExtension(DSConnectionServerOptions options, const char *local_extension) {
    ds_assert_param(options);

    if (dsExchangeString(NULL,&options->local_extension, local_extension)) return DS_ERR_MEM;
    return 0;
}


int dsConnectionServerOptionsSetSignalMask(DSConnectionServerOptions options, DSSignalMask sigmask) {
    memcpy(&options->sigmask, sigmask, sizeof(DSSignalMaskS));
    return 0;
}


void dsConnectionServerZeroContext(DSConnectionServerContext ctx) {
    ds_assert_return(ctx);

    ctx->softsig = NULL;
    
#ifdef _WIN32
    ctx->local_handle = INVALID_HANDLE_VALUE;
    ctx->hEventLocal = NULL;
    ctx->oOverlap = NULL;
    ctx->local_ready = 0;
# ifdef HAVE_SOCKETS
    ctx->hEventNetwork = NULL;
    ctx->network_handle = INVALID_SOCKET;
# endif /* HAVE_SOCKETS */
#else /* _WIN32 */
    ctx->local_socket = 0;
    ctx->network_socket = 0;
#endif /* _WIN32 */

    ctx->local_name = NULL;
}

DSConnectionServerContext dsConnectionServerCreateContext(DSMallocContext mctx) {
    DSConnectionServerContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "dsConnectionServerCreateContext");    

    ctx = (DSConnectionServerContext)dsMalloc(mctx, sizeof(DSConnectionServerContextS));
    if (ctx) {
	ctx->malloc = mctx;
	dsConnectionServerZeroContext(ctx);
    }
    return ctx;
}

#ifdef _WIN32
static int ConnectionServerWIN32_LocalReinit(DSConnectionServerContext ctx) {
	int err;
	COMMTIMEOUTS ct;
        HANDLE local_handle;
	LPOVERLAPPED oOverlap;
	HANDLE hEvent;
	
	ct.ReadIntervalTimeout = 1;
	ct.ReadTotalTimeoutMultiplier = 0;
	ct.ReadTotalTimeoutConstant = 0;
	ct.WriteTotalTimeoutMultiplier = 0;
	ct.WriteTotalTimeoutConstant = 0;
	
	ctx->local_handle = INVALID_HANDLE_VALUE;
	
	hEvent = ctx->hEventLocal;
	ds_assert_param(oOverlap = ctx->oOverlap);
	ds_assert_param(hEvent != INVALID_HANDLE_VALUE);
	
	memset(oOverlap, 0, sizeof(OVERLAPPED));
	oOverlap->hEvent = hEvent;
	
	local_handle = CreateNamedPipe(ctx->local_name, 
	    PIPE_ACCESS_DUPLEX|FILE_FLAG_OVERLAPPED,
	    PIPE_TYPE_BYTE|PIPE_READMODE_BYTE|PIPE_WAIT,
	    PIPE_UNLIMITED_INSTANCES,
	    DS_DEFAULT_IO_BUFFER_SIZE,
	    DS_DEFAULT_IO_BUFFER_SIZE,
	    NMPWAIT_USE_DEFAULT_WAIT,
	    NULL);
	
	SetCommTimeouts(local_handle, &ct);
	
	if (local_handle == INVALID_HANDLE_VALUE) return DS_ERR_WINAPI_PIPE_CREATE;
	
	ResetEvent(hEvent);
	if (!ConnectNamedPipe(local_handle, oOverlap)) {
	    err = GetLastError();
	    if (err == ERROR_PIPE_CONNECTED) ctx->local_ready = 1;
	    else if (err != ERROR_IO_PENDING) {
		CloseHandle(local_handle);
	        return DS_ERR_WINAPI_PIPE_CONNECT;
	    }
	} else ctx->local_ready = 1;

	ctx->local_handle = local_handle;

	return 0;
}
#endif /* _WIN32 */

int dsConnectionServerInitContext(DSConnectionServerContext ctx, DSConnectionServerOptions options) {
    int err;
    char *local_name;

#ifdef _WIN32 
# ifdef HAVE_SOCKETS
    SOCKET sock;
    SOCKADDR_IN isock;
    u_long blocking_mode = 1;	/* non-blocking mode */
# endif /* HAVE_SOCKETS */
#else /* _WIN32 */
    int s;
    int flags;

    struct sockaddr_un mysock;
    struct sockaddr_in isock;
#endif /* _WIN32 */

    dsDebugServer(DS_DEBUG_NOTICE, "dsConnectionServerInitContext");    

    ds_assert_param(ctx);
    
    dsConnectionServerZeroContext(ctx);
    
    
    if (options->local_name) {
#ifdef _WIN32 
	if (options->local_extension) {
    	    local_name = (char*)dsMalloc(ctx->malloc, (strlen(options->local_extension) + strlen(options->local_name)+14)*sizeof(char));
	    if (!local_name) return DS_ERR_MEM;
	    ctx->local_name = local_name;
    	    sprintf(local_name, "\\\\.\\pipe\\%s%s",options->local_name,options->local_extension);
	} else {
    	    local_name = (char*)dsMalloc(ctx->malloc, (strlen(options->local_name)+14)*sizeof(char));
	    if (!local_name) return DS_ERR_MEM;
	    ctx->local_name = local_name;
    	    sprintf(local_name, "\\\\.\\pipe\\%s",options->local_name);
	}
	
	ctx->oOverlap = (LPOVERLAPPED)dsMalloc(ctx->malloc, sizeof(OVERLAPPED));
	if (!ctx->oOverlap) {
	    dsConnectionServerFreeContext(ctx);
	    return DS_ERR_MEM;
	}

	ctx->hEventLocal = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (!ctx->hEventLocal) {
	    dsConnectionServerFreeContext(ctx);
	    return DS_ERR_WINAPI_EVENT_CREATE;
	}

	err = ConnectionServerWIN32_LocalReinit(ctx);
	if (err) {
	    dsConnectionServerFreeContext(ctx);
	    return err;
	}
#else /* _WIN32 */
	local_name = dsCreateFullName(ctx->malloc, options->local_path, options->local_name, options->local_extension);
	if (!local_name) return DS_ERR_MEM;
	
	ctx->local_name = local_name;

	unlink(local_name);
	
	err = dsFileCreateDirectory(local_name);
	if (err) {
	    dsConnectionServerFreeContext(ctx);
	    return DS_ERR_MEM;
	}
	
	s=socket(PF_UNIX,SOCK_STREAM,0);
	if (!s) {
	    dsConnectionServerFreeContext(ctx);
	    return DS_ERR_SOCKET_CREATE;
	}

retry_fcntl:	
	flags=fcntl(s,F_GETFL,0);
	if ((flags<0)||(fcntl(s,F_SETFL,flags|O_NONBLOCK)<0)) {
	    err = errno;
	    if (err == EAGAIN) goto retry_fcntl;
	    close(s);
	    dsConnectionServerFreeContext(ctx);
	    return DS_ERR_SOCKET_FCNTL;
	}
    
	memset(&mysock, 0, sizeof(mysock));
	mysock.sun_family=AF_UNIX;
	strncpy(mysock.sun_path,local_name,sizeof(mysock.sun_path));
	mysock.sun_path[sizeof(mysock.sun_path)-1]=0;

	if (bind(s,(struct sockaddr*)&mysock,sizeof(mysock))==-1) {
	    err = errno;
	    close(s);
	    dsConnectionServerFreeContext(ctx);

	    if (err == EINVAL) return DS_ERR_CONNECTIONSERVER_ADDRINUSE;
	    if ((err == EACCES)||(err == EROFS)) return DS_ERR_PERMISSION;
	    if ((err == ENAMETOOLONG)||(err == ENOENT)||(err == ENOTDIR)||(err == ELOOP)) return DS_ERR_INVALID_OPTIONS;
	    return DS_ERR_SOCKET_BIND;
	}

	if (listen(s,1)<0) {
	    err = errno;
	    unlink(local_name);
	    close(s);
	    dsConnectionServerFreeContext(ctx);
	    if (err == EADDRINUSE) return DS_ERR_CONNECTIONSERVER_ADDRINUSE;
	    return DS_ERR_SOCKET_LISTEN;	
	}

	if (chmod(local_name,00777)<0) {
	    err = errno;
	    unlink(local_name);
	    close(s);
	    dsConnectionServerFreeContext(ctx);
	    return DS_ERR_FS_CHMOD;
	}
	
	ctx->local_socket = s;
#endif /* _WIN32 */
    }
    

    if (options->network_port) {
#ifdef _WIN32
# ifdef HAVE_SOCKETS
	if (!winsock_initialized) {
	    err = dsConnectionServerInit();
	    if (err) {
		dsConnectionServerFreeContext(ctx);
		return err;
	    }
	}
	
	ctx->hEventNetwork = WSACreateEvent();
	if (!ctx->hEventNetwork) {
	    dsConnectionServerFreeContext(ctx);
	    return DS_ERR_WINAPI_EVENT_CREATE;
	}
	
	sock = socket(AF_INET, SOCK_STREAM,IPPROTO_TCP);
	if (sock == INVALID_SOCKET) {
	    dsConnectionServerFreeContext(ctx);
	    return DS_ERR_SOCKET_CREATE;
	}
	
	if (ioctlsocket(sock, FIONBIO, &blocking_mode) == SOCKET_ERROR) {
	    dsConnectionServerFreeContext(ctx);
	    return DS_ERR_SOCKET_IOCTL;
	}
	
	if (WSAEventSelect(sock, ctx->hEventNetwork, FD_ACCEPT) == SOCKET_ERROR) {
	    closesocket(sock);
	    dsConnectionServerFreeContext(ctx);
	    return DS_ERR_SOCKET_FCNTL;
	}
	
	isock.sin_family=AF_INET;
	isock.sin_port=htons(options->network_port);
	isock.sin_addr.S_un.S_addr=INADDR_ANY;
	
	if (bind(sock, (SOCKADDR*)&isock, sizeof(SOCKADDR_IN)) == SOCKET_ERROR) {
	    err = WSAGetLastError();
	    closesocket(sock);
	    dsConnectionServerFreeContext(ctx);
	    if (err == WSAEADDRINUSE) return DS_ERR_CONNECTIONSERVER_ADDRINUSE;
	    return DS_ERR_SOCKET_BIND;
	}
	
	if (listen(sock, 1) == SOCKET_ERROR) {
	    err = WSAGetLastError();
	    closesocket(sock);
	    dsConnectionServerFreeContext(ctx);
	    if (err == WSAEADDRINUSE) return DS_ERR_CONNECTIONSERVER_ADDRINUSE;
	    return DS_ERR_SOCKET_LISTEN;
	}
	
	ctx->network_handle = sock;
# endif /* HAVE_SOCKETS */
#else /* _WIN32 */
	s = socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
	if (!s) {
	    err = errno;
	    dsConnectionServerFreeContext(ctx);
	    return DS_ERR_SOCKET_CREATE;
	}

retry_network_fcntl:	
	flags=fcntl(s,F_GETFL,0);
	if ((flags<0)||(fcntl(s,F_SETFL,flags|O_NONBLOCK)<0)) {
	    err = errno;
	    if (err == EAGAIN) goto retry_network_fcntl;
	    close(s);
	    dsConnectionServerFreeContext(ctx);
	    return DS_ERR_SOCKET_FCNTL;
	}
    
	memset(&isock, 0, sizeof(isock));
	isock.sin_family=AF_INET;
	isock.sin_port=htons(options->network_port);
	isock.sin_addr.s_addr=htonl(INADDR_ANY);

	if (bind(s,(struct sockaddr*)&isock,sizeof(struct sockaddr_in))==-1) {
	    err = errno;
	    close(s);
	    dsConnectionServerFreeContext(ctx);
	    if (err == EINVAL) return DS_ERR_CONNECTIONSERVER_ADDRINUSE;
	    return DS_ERR_SOCKET_BIND;
	}
    
	if (listen(s,1)<0) {
	    close(s);
	    dsConnectionServerFreeContext(ctx);
	    if (err == EADDRINUSE) return DS_ERR_CONNECTIONSERVER_ADDRINUSE;
	    return DS_ERR_SOCKET_LISTEN;	
	}
	
	ctx->network_socket = s;
#endif /* _WIN32 */
    }

    memcpy(&ctx->sigmask, &options->sigmask, sizeof(DSSignalMaskS));
    
    return 0;
}


void dsConnectionServerFreeContext(DSConnectionServerContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "dsConnectionServerFreeContext");    

    ds_assert_return(ctx);

#ifdef _WIN32
    if (ctx->local_handle != INVALID_HANDLE_VALUE) {
	if (!ctx->local_ready) {
	    while (!CancelIo(ctx->local_handle)) {
		dsDebugServerSystemError(DS_DEBUG_ERROR, 0, "dsConnectionServer(CancelIO failed): ", ctx); 
	    }
	}
	CloseHandle(ctx->local_handle);
    }
    if (ctx->oOverlap) dsFree(ctx->malloc, ctx->oOverlap);
    if (ctx->hEventLocal) CloseHandle(ctx->hEventLocal);
#else /* _WIN32 */
    if (ctx->local_socket) {
	if (ctx->local_name) unlink(ctx->local_name);
	close(ctx->local_socket);
    }
#endif /* _WIN32 */


#ifdef _WIN32
# ifdef HAVE_SOCKETS
    if (ctx->network_handle != INVALID_SOCKET) {
	/* In case of graceful disconect (SO_DONTLINGER default), data queued for transmittion
	will be sent, if possible, before underlaying socket is closed. */
	WSAEventSelect(ctx->network_handle, NULL, 0);
	closesocket(ctx->network_handle);
    }
    if (ctx->hEventNetwork) WSACloseEvent(ctx->hEventNetwork);
# endif /* HAVE_SOCKETS */
#else /* _WIN32 */
    if (ctx->network_socket) close(ctx->network_socket);
#endif /* _WIN32 */

    if (ctx->local_name) dsFree(ctx->malloc, ctx->local_name);
    
    dsConnectionServerZeroContext(ctx);
}

void dsConnectionServerDestroyContext(DSConnectionServerContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "dsConnectionServerDestroyContext");    

    ds_assert_return(ctx);
    dsConnectionServerFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

DSConnectionContext dsConnectionServerAcceptLocal(DSConnectionServerContext ctx) {
    int err;
    DSConnectionContext c;
    DSConnectionMask mask = DS_CONNECTION_LOCAL;

#ifdef _WIN32
    HANDLE local_handle;
    DWORD bytes;
#else /* _WIN32 */
    int s;
    fd_set rd;
#endif /* _WIN32 */
    
    ds_assert_null(ctx);

#ifdef _WIN32 
    ds_assert_null(ctx->local_handle != INVALID_HANDLE_VALUE);
#else /* _WIN32 */
    ds_assert_null(s = ctx->local_socket);
#endif /* _WIN32 */

select_again:    

#ifdef _WIN32 
    if (ctx->local_ready) {
	ctx->local_ready = 0;
    } else {
	err = dsSignalWinWait(&ctx->sigmask, ctx->hEventLocal, NULL);
	if (err) return NULL;
	    
	if (!GetOverlappedResult(ctx->local_handle, ctx->oOverlap, &bytes, TRUE)) {
	    err = GetLastError();
	    if (err != ERROR_IO_INCOMPLETE) {
		CloseHandle(ctx->local_handle);
		err = ConnectionServerWIN32_LocalReinit(ctx);
		if (err) {
		    ctx->local_handle = INVALID_HANDLE_VALUE;
		    dsDebugServer(DS_DEBUG_ALERT, "ConnectionServer %p: Disabling Local Connector. Reinitialization failed: %s", ctx, dsError(err, NULL));
		    return NULL;		    
		}
	    }
	    goto select_again;
	}	
    }
#else /* _WIN32 */
    err = dsSignalWaitConnection(ctx, &ctx->sigmask, ctx->softsig, &mask, NULL);
#endif /* _WIN32 */
    if (err) return NULL;

#ifdef _WIN32 
    local_handle = ctx->local_handle;
    
    err = ConnectionServerWIN32_LocalReinit(ctx);
    if (err) {
	ctx->local_handle = INVALID_HANDLE_VALUE;
	dsDebugServer(DS_DEBUG_ALERT, "ConnectionServer %p: Disabling Local Connector. Reinitialization failed: %s", ctx, dsError(err, NULL));
	/* return err; Continue in any case. */
    }
#else /* _WIN32 */
    err = accept(s,NULL,NULL);
    if (err<0) {
# if defined(EAGAIN) && defined(EWOULDBLOCK) && EAGAIN != EWOULDBLOCK
	if ((errno==EAGAIN)||(errno==EWOULDBLOCK)) goto select_again;
# elif defined(EAGAIN)
	if (errno==EAGAIN) goto select_again;
# elif defined(EWOULDBLOCK)
	if (errno==EWOULDBLOCK) goto select_again;
# endif /* EAGAIN, EWOULDBLOCK */
	return NULL;
    }
#endif /* _WIN32 */
    
    c = dsConnectionCreateContext(ctx->malloc);
    if (!c) return c;
#ifdef _WIN32 
    err = dsConnectionInitHandleContext(c, &ctx->sigmask, local_handle);
#else /* _WIN32 */
    err = dsConnectionInitSocketContext(c, &ctx->sigmask, err);
#endif /* _WIN32 */

    if (err) {
	dsConnectionDestroyContext(c);
	return NULL;
    }
    return c;


}

DSConnectionContext dsConnectionServerAcceptNetwork(DSConnectionServerContext ctx) {
    int err;

#ifdef _WIN32 
# ifdef HAVE_SOCKETS    
    SOCKET s; 
    SOCKET sock;       
# endif /* HAVE_SOCKETS */
#else /* _WIN32 */
    int s;
#endif /* _WIN32 */

    DSConnectionContext c;
    DSConnectionMask mask = DS_CONNECTION_NETWORK;
    
    ds_assert_null(ctx);

#ifdef _WIN32 
# ifdef HAVE_SOCKETS    
    ds_assert_null(s = ctx->network_handle);
# else /* HAVE_SOCKETS */
    return DS_ERR_NOTSUPPORTED;
# endif /* HAVE_SOCKETS */
#else /* _WIN32 */
    ds_assert_null(s = ctx->network_socket);
#endif /* _WIN32 */

    
select_again:    
#ifdef _WIN32 
    err = dsSignalWinWait(&ctx->sigmask, ctx->hEventNetwork, NULL);
#else /* _WIN32 */
    err = dsSignalWaitConnection(ctx, &ctx->sigmask, ctx->softsig, &mask, NULL);
#endif /* _WIN32 */
    if (err) return NULL;

#ifdef _WIN32 
# ifdef HAVE_SOCKETS    
    sock = accept(s, NULL, NULL);
    if (sock == SOCKET_ERROR) {
	err = WSAGetLastError();
	if ((err == WSAEWOULDBLOCK)||(err == WSAECONNRESET)) goto select_again;
	return NULL;
    }
# endif /* HAVE_SOCKETS */
#else /* _WIN32 */
    err=accept(s,NULL,NULL);
    if (err<0) {
# if defined(EAGAIN) && defined(EWOULDBLOCK) && EAGAIN != EWOULDBLOCK
	if ((errno==EAGAIN)||(errno==EWOULDBLOCK)) goto select_again;
# elif defined(EAGAIN)
	if (errno==EAGAIN) goto select_again;
# elif defined(EWOULDBLOCK)
	if (errno==EWOULDBLOCK) goto select_again;
# endif /* EAGAIN, EWOULDBLOCK */
	return NULL;
    }
#endif /* _WIN32 */

    c = dsConnectionCreateContext(ctx->malloc);
    if (!c) return c;
#ifdef _WIN32 
# ifdef HAVE_SOCKETS    
    err = dsConnectionInitSocketContext(c, &ctx->sigmask, sock);
# endif /* HAVE_SOCKETS */
#else /* _WIN32 */
    err = dsConnectionInitSocketContext(c, &ctx->sigmask, err);
#endif /* _WIN32 */
    if (err) {
	dsConnectionDestroyContext(c);
	return NULL;
    }
    return c;
}

DSConnectionContext dsConnectionServerAccept(DSConnectionServerContext ctx) {
    int err;

# ifdef _WIN32 
    DWORD bytes;
    HANDLE local_handle;
# ifdef HAVE_SOCKETS
    SOCKET sock;       
# endif /* HAVE_SOCKETS */
# else /* _WIN32 */
    int s;
# endif /* _WIN32 */

    DSConnectionContext c;
    DSConnectionMask mask = 0;
    
    ds_assert_null(ctx);
    
    
select_again:    
#ifdef _WIN32
    if (ctx->local_handle != INVALID_HANDLE_VALUE) mask |= DS_CONNECTION_LOCAL;
# ifdef HAVE_SOCKETS
    if (ctx->network_handle != INVALID_SOCKET) mask |= DS_CONNECTION_NETWORK;
# endif /* HAVE_SOCKETS */

    if (ctx->local_ready) {
	mask = DS_CONNECTION_LOCAL;
	ctx->local_ready = 0;
	err = 0;
    } else {
	err = dsSignalWinWaitConnection(ctx, &ctx->sigmask, &mask, NULL);
    }
#else /* _WIN32 */
    if (ctx->local_socket) mask |= DS_CONNECTION_LOCAL;
    if (ctx->network_socket) mask |= DS_CONNECTION_NETWORK;

    err = dsSignalWaitConnection(ctx, &ctx->sigmask, ctx->softsig, &mask, NULL);
#endif /* _WIN32 */
    if (err) return NULL;

#ifdef _WIN32
    if (mask&DS_CONNECTION_LOCAL) {
	if (!GetOverlappedResult(ctx->local_handle, ctx->oOverlap, &bytes, TRUE)) {
    	    err = GetLastError();
	    if (err != ERROR_IO_INCOMPLETE) {
		CloseHandle(ctx->local_handle);
		err = ConnectionServerWIN32_LocalReinit(ctx);
		if (err) {
		    ctx->local_handle = INVALID_HANDLE_VALUE;
		    dsDebugServer(DS_DEBUG_ALERT, "ConnectionServer %p: Disabling Local Connector. Reinitialization failed: %s", ctx, dsError(err, NULL));
		    return NULL;		    
		}
	    }
	    goto select_again;
	}	

	local_handle = ctx->local_handle;
    
	err = ConnectionServerWIN32_LocalReinit(ctx);
	if (err) {
	    dsDebugServer(DS_DEBUG_ALERT, "ConnectionServer %p: Disabling Local Connector. Reinitialization failed: %s", ctx, dsError(err, NULL));
	    /* return err; Continue in any case. */
	}
    } 
# ifdef HAVE_SOCKETS
    else {
	sock = accept(ctx->network_handle, NULL, NULL);
	if (sock == SOCKET_ERROR) {
	    err = WSAGetLastError();
	    if ((err == WSAEWOULDBLOCK)||(err == WSAECONNRESET)) goto select_again;
	    return NULL;
	}
    }
# endif /* HAVE_SOCKETS */
#else /* _WIN32 */
    if (mask&DS_CONNECTION_LOCAL) err = accept(ctx->local_socket, NULL, NULL);
    else err = accept(ctx->network_socket, NULL, NULL);
    if (err<0) {
	if ((errno==EAGAIN)||(errno==EWOULDBLOCK)) goto select_again;
	printf("%lu\n",errno);
	return NULL;
    }
#endif /* _WIN32 */

    c = dsConnectionCreateContext(ctx->malloc);
    if (!c) return c;
#ifdef _WIN32
    if (mask&DS_CONNECTION_LOCAL) err = dsConnectionInitHandleContext(c, &ctx->sigmask, local_handle);
# ifdef HAVE_SOCKETS
    else if (mask&DS_CONNECTION_NETWORK) err = dsConnectionInitSocketContext(c, &ctx->sigmask, sock);
# endif /* HAVE_SOCKETS */
    else err = DS_ERR_NOTSUPPORTED;
#else /* _WIN32 */
    err = dsConnectionInitSocketContext(c, &ctx->sigmask, err);
#endif /* _WIN32 */
    if (err) {
	dsConnectionDestroyContext(c);
	return NULL;
    }
    return c;
}
