#include <stdio.h>
#include <string.h>

#include <dslib.h>

#ifdef HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif /* HAVE_SYS_SOCKET_H */

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif /* HAVE_ERRNO_H */

#include <dsassert.h>
#include <dsdebug.h>
#include <dserror.h>

#include <dstime.h>

#include <dsdatagramserver.h>

#define dsDatagramServerData(ctx) ((DSConnectionUDPData)(ctx->con->data))

static void dsConnectionListItemDestroy(DSConnectionList list, DSConnectionListItem item) {
    ds_assert_return(list);
    ds_assert_return(item);

    if (item->filter) dsFree(list->malloc, item->filter);
    dsFree(list->malloc, item);
}

int dsConnectionListItemSetFilter(DSConnectionList list, DSConnectionListItem item, size_t filter_size, const char *filter) {
    ds_assert_param(list);
    ds_assert_param(item);

    if (item->filter) dsFree(list->malloc, item->filter);

    if ((filter)&&(filter_size)) {
	item->filter = dsMalloc(list->malloc, filter_size);
	if (!item->filter) return DS_ERR_MEM;
	memcpy(item->filter, filter, filter_size);
	item->filter_size = filter_size;
    } else {
	item->filter = NULL;
	item->filter_size = 0;
    }
    return 0;
}

DSConnectionList dsConnectionListCreate(DSMallocContext mctx) {
    DSConnectionList list;
    DSLockContext lock;
    
    
    lock = dsLockCreateContext(mctx);
    if ((lock)&&(dsLockInitContext(lock))) {
	dsLockDestroyContext(lock);
	lock = NULL;
    }

    if (lock) list = (DSConnectionList)dsMalloc(mctx, sizeof(DSConnectionListS));
    else list = NULL;
    
    if (list) {
	list->malloc = mctx;
	list->first = NULL;
	list->lock = lock;
    } else if (lock) dsLockDestroyContext(lock);
    
    return list;
}

void dsConnectionListClear(DSConnectionList list) {
    DSConnectionListItem cur, next;
    
    ds_assert_return(list);
    
    for (cur = list->first; cur; cur = next) {
	next = cur->next;
	dsConnectionListItemDestroy(list, cur);
    }
    list->first = NULL;
}

void dsConnectionListDestroy(DSConnectionList list) {
    ds_assert_return(list);
    dsConnectionListClear(list);
    dsLockDestroyContext(list->lock);
    dsFree(list->malloc, list);
}

int dsConnectionListLock(DSConnectionList list) {
    return dsLockWrite(list->lock);
}

int dsConnectionListUnLock(DSConnectionList list) {
    return dsLockReturn(list->lock);
}

int dsConnectionListIsEmpty(DSConnectionList list) {
    ds_assert_default(list, 1);
    
    if (list->first) return 0;
    return 1;
}

DSConnectionListItem dsConnectionListAdd(DSConnectionList list, DSConnectionAddress *addr, DSDuration livetime) {
    DSConnectionListItem last, cur;
    
    for (cur = list->first; cur; cur = cur->next) {
	if (!dsConnectionAddressCmp(&cur->addr, addr)) {
	    if (livetime) {
		dsTimeCurrentTime(&cur->expire);
		dsTimeAdd(&cur->expire, livetime);
	    } else dsTimeZeroTime(&cur->expire);
	    return cur;
	}
	last = cur;
    }
    
    cur = (DSConnectionListItem)dsMalloc(list->malloc, sizeof(DSConnectionListItemS));
    if (!cur) return NULL;

    cur->next = NULL;
    dsConnectionAddressCpy(&cur->addr, addr);
    if (livetime) {
	dsTimeCurrentTime(&cur->expire);
	dsTimeAdd(&cur->expire, livetime);
    } else dsTimeZeroTime(&cur->expire);
    
    cur->filter_size = 0;
    cur->filter = NULL;

    if (list->first) last->next = cur;
    else list->first = cur;

    return cur;
}

void dsConnectionListManage(DSConnectionList list) {
    DSConnectionListItem last = NULL, cur, next;
    DSTimeS curtime;

    ds_assert_return(list);
        
    dsTimeCurrentTime(&curtime);
    
    for (cur = list->first; cur; cur = next) {
	next = cur->next;

	if (!(dsTimeIsZero(&cur->expire))&&(dsTimeCmp(&curtime, &cur->expire)>0)) {
	    dsConnectionListItemDestroy(list, cur);
	    
	    if (last) last->next = next;
	    else list->first = next;
	} else last = cur;
    }
}


void dsDatagramServerZeroContext(DSDatagramServerContext ctx) {
    ds_assert_return(ctx);
    
    ctx->con = NULL;
    ctx->clients = NULL;
    ctx->master = NULL;
    ctx->refs = 0;
    ctx->buffer = NULL;
    ctx->write_lock = NULL;
}

DSDatagramServerContext dsDatagramServerCreateContext(DSMallocContext mctx) {
    DSDatagramServerContext ctx;
    
    ctx = dsMalloc(mctx, sizeof(DSDatagramServerContextS));
    if (ctx) {
	ctx->malloc = mctx;
	dsDatagramServerZeroContext(ctx);
    }
    return ctx;
}

int dsDatagramServerInitContext(DSDatagramServerContext ctx, DSSignalMask sigmask, in_port_t port) {
    int err;

    DSConnectionContext con;
    DSLockContext write_lock;
    
#ifdef HAVE_SOCKETS    
# ifdef _WIN32
    SOCKADDR_IN mysock;
# else 
    struct sockaddr_in mysock;
# endif /* _WIN32 */
#endif /* HAVE_SOCKETS */
    
    ds_assert_param(ctx);

/*
    if (ctx->refs) return DS_ERR_BUSY;
    dsDatagramServerFreeContext(ctx);
*/
    dsDatagramServerZeroContext(ctx);

    con = dsConnectionCreateContext(ctx->malloc);
    if (con) err = dsConnectionConnectUDPServer(con, sigmask, NULL, 0);
    else err = DS_ERR_MEM;
    
    
    if (err) {
	if (con) dsConnectionDestroyContext(con);
	return err;
    }

#ifdef HAVE_SOCKETS    
    memset(&mysock, 0, sizeof(mysock));
    mysock.sin_family=AF_INET;
    mysock.sin_port=htons(port);

# ifdef _WIN32
    mysock.sin_addr.S_un.S_addr=INADDR_ANY;

    if (bind(con->winsock,(SOCKADDR*)&mysock,sizeof(struct sockaddr_in))==-1) {
	    err = WSAGetLastError();
	    if (err == WSAEADDRINUSE) err = DS_ERR_CONNECTIONSERVER_ADDRINUSE;
	    else err = DS_ERR_SOCKET_BIND;
# else 
    mysock.sin_addr.s_addr=htonl(INADDR_ANY);

    if (bind(con->socket,(struct sockaddr*)&mysock,sizeof(struct sockaddr_in))==-1) {
	    if (errno == EINVAL) err = DS_ERR_CONNECTIONSERVER_ADDRINUSE;
	    else err = DS_ERR_SOCKET_BIND;
# endif /* _WIN32 */
	    dsConnectionDestroyContext(con);
	    return err;
    }
#endif /* HAVE_SOCKETS */

    write_lock = dsLockCreateContext(ctx->malloc);
    if (write_lock) err = dsLockInitContext(write_lock);
    else err = DS_ERR_MEM;
    
    if (err) {
	dsConnectionDestroyContext(con);
	if (write_lock) dsLockDestroyContext(write_lock);
	return err;
    }

    ctx->clients = dsConnectionListCreate(ctx->malloc);
    if (!ctx->clients) {
	dsLockDestroyContext(write_lock);
	dsConnectionDestroyContext(con);
	return DS_ERR_MEM;
    }

    ctx->con = con;    
    ctx->write_lock = write_lock;
    ctx->manager_iterations = (unsigned long)-1;
    dsDatagramServerData(ctx)->fixed = 0;
    dsDurationZeroDuration(&ctx->expiration);
    
    return 0;
}

int dsDatagramServerInitSlaveContext(DSDatagramServerContext ctx, DSDatagramServerContext master) {
    ds_assert_param(ctx);
    ds_assert_param(master);
    ds_assert_param(master->con);

/*
    if (ctx->refs) return DS_ERR_BUSY;
    dsDatagramServerFreeContext(ctx);
*/
    dsDatagramServerZeroContext(ctx);

    ctx->clients = dsConnectionListCreate(ctx->malloc);
    if (!ctx->clients) return DS_ERR_MEM;

    dsDatagramServerRef(master);
    ctx->master = master;
    ctx->con = master->con;    
    ctx->manager_iterations = (unsigned long)-1;
    dsDatagramServerData(ctx)->fixed = 0;
    dsDurationZeroDuration(&ctx->expiration);
	
    return 0;
}


void dsDatagramServerFreeContext(DSDatagramServerContext ctx) {
    ds_assert_return(ctx);
    
/*    
    if (ctx->refs) return;
*/
    
    if (ctx->clients) dsConnectionListDestroy(ctx->clients);

    if (ctx->master) {
	dsDatagramServerUnref(ctx->master);
	if (ctx->buffer) dsFree(ctx->malloc, ctx->buffer);
    } else if (ctx->con) dsConnectionDestroyContext(ctx->con);

    if (ctx->write_lock) dsLockDestroyContext(ctx->write_lock);
    
    dsDatagramServerZeroContext(ctx);
}


void dsDatagramServerDestroyContext(DSDatagramServerContext ctx) {
    ds_assert_return(ctx);

/*
    if (ctx->refs) {
	dsDatagramServerUnref(ctx);
	return;
    }
*/

    dsDatagramServerFreeContext(ctx);
    dsFree(ctx->malloc, ctx);
}

int dsDatagramServerRef(DSDatagramServerContext ctx) {
    ds_assert_param(ctx);
    ctx->refs++;
    return 0;
}

int dsDatagramServerUnref(DSDatagramServerContext ctx) {
    ds_assert_param(ctx);
    
    if (ctx->refs > 0) ctx->refs--;
/*
    else dsDatagramServerDestroyContext(ctx);
*/
    return 0;
}


int dsDatagramServerSetExpirationDuration(DSDatagramServerContext ctx, DSDuration duration) {
    ds_assert_param(ctx);
    
    if (duration) dsDurationCopy(&ctx->expiration, duration);
    else dsDurationZeroDuration(&ctx->expiration);
    return 0;
}

int dsDatagramServerSetMaxManagerIterations(DSDatagramServerContext ctx, unsigned long iterations) {
    ds_assert_param(ctx);
    
    if (iterations) ctx->manager_iterations = iterations;
    else ctx->manager_iterations = (unsigned long)-1;
    
    return 0;
}

int dsDatagramServerDrop(DSDatagramServerContext ctx) {
    ds_assert_param(ctx);
    
//    dsDatagramServerData(ctx)->size = 0;
    ctx->buffer_size = 0;
    return 0;
}


int dsDatagramServerManageClients(DSDatagramServerContext ctx, const char *data, size_t len) {
    int err = 0;
    size_t readed;
    char *rb, *buf;
    unsigned long iter;
    
    DSConnectionListItem item;
    
    
    ds_assert_param(ctx);
    
//    dsDatagramServerDrop(ctx);
    if (!ctx->master) dsDatagramServerData(ctx)->size = 0;

    rb = dsMalloca(len);
    if (!rb) return DS_ERR_MEM;
    
    for (iter=0;(!err)&&(iter<ctx->manager_iterations);iter++) {
	err = dsConnectionTryRead(ctx->con, rb, len, 0, &readed);
	if ((!err)&&(readed == len)&&(!memcmp(data, rb, len))) {
	    item = dsConnectionListAdd(ctx->clients, &dsDatagramServerData(ctx)->remote, dsDurationIsZero(&ctx->expiration)?NULL:&ctx->expiration);
	    if (!item) {
		err = DS_ERR_MEM;
		break;
	    }

	    if (dsDatagramServerData(ctx)->size) {
		buf = dsMalloca(dsDatagramServerData(ctx)->size);
		if (!buf) {
		    err = DS_ERR_MEM;
		    break;
		}
		dsConnectionRead(ctx->con, buf, dsDatagramServerData(ctx)->size, 0, &readed);
		dsConnectionListItemSetFilter(ctx->clients, item, readed, buf);
		dsFreea(buf);
	    }
	    
	} else if (err) dsDatagramServerData(ctx)->size = 0;
	else {
	    dsDatagramServerData(ctx)->size += dsDatagramServerData(ctx)->start;
	    dsDatagramServerData(ctx)->start = 0;
	    err = DS_ERR_NOTSUPPORTED;
	    
	    if (!dsDatagramServerData(ctx)->size) {
		dsDatagramServerData(ctx)->start = 0;
		dsDatagramServerData(ctx)->size = readed;
		memcpy(dsDatagramServerData(ctx)->buffer, rb, readed);
	    }	
	}
    }
    
    dsFreea(rb);

    return (err == DS_ERR_TIMEOUT)?0:err;
}

int dsDatagramServerManage(DSDatagramServerContext ctx) {
    ds_assert_param(ctx);
    
    dsConnectionListManage(ctx->clients);
    return 0;
}

int dsDatagramServerRead(DSDatagramServerContext ctx, DSConnectionAddress *addr, char *data, size_t len, DSConnectionFlags flags, size_t *rbytes) {
    int err;
    
    ds_assert_param(ctx);
    ds_assert_param(data);
    ds_assert_param(len);
    
    if ((ctx->buffer)&&(ctx->buffer_size)) {
	if (ctx->buffer_size > len) {
	    memcpy(data, (char*)ctx->buffer + ctx->buffer_start, len);
	    if (rbytes) *rbytes = len;
	    ctx->buffer_start += len;
	    ctx->buffer_size -= len;
	} else {
	    memcpy(data, (char*)ctx->buffer + ctx->buffer_start, ctx->buffer_size);
	    if (rbytes) *rbytes = ctx->buffer_size;
	    ctx->buffer_size = 0;
	}
	
	return 0;
    }
    
    err = dsConnectionRead(ctx->con, data, len, flags, rbytes);
    if ((!err)&&(addr)) {
	if (ctx->master) {
	    if (dsDatagramServerData(ctx)->size) {
		if (!ctx->buffer) {
		    ctx->buffer = dsMalloc(ctx->malloc, DS_MAX_UDP_MESSAGE_SIZE);
		    if (!ctx->buffer) return DS_ERR_MEM;
		}
           	memcpy(ctx->buffer, dsDatagramServerData(ctx)->buffer + dsDatagramServerData(ctx)->start, dsDatagramServerData(ctx)->size);
		ctx->buffer_start = 0;
		ctx->buffer_size = dsDatagramServerData(ctx)->size;
		
		dsDatagramServerData(ctx)->size = 0;
	    }
	}
	dsConnectionAddressCpy(addr, &dsDatagramServerData(ctx)->remote);
    }
    
    return err;
}

int dsDatagramServerGetWriteLock(DSDatagramServerContext ctx) {
    ds_assert_param(ctx);
    
    if (ctx->master) return dsDatagramServerGetWriteLock(ctx->master);
    
    return dsLockWrite(ctx->write_lock);
}

int dsDatagramServerReturnWriteLock(DSDatagramServerContext ctx) {
    ds_assert_param(ctx);
    
    if (ctx->master) return dsDatagramServerReturnWriteLock(ctx->master);
    
    return dsLockReturn(ctx->write_lock);
}

int dsDatagramServerWrite(DSDatagramServerContext ctx, DSConnectionAddress *addr, const char *data, size_t len, DSConnectionFlags flags, size_t *wbytes) {
    int err;
    
    ds_assert_param(ctx);

    err = dsDatagramServerGetWriteLock(ctx);
    if (!err) {    
	dsDatagramServerData(ctx)->override = addr;
    
	err = dsConnectionWrite(ctx->con, data, len, flags, wbytes);

	dsDatagramServerReturnWriteLock(ctx);
    }

    return err;
}

int dsDatagramServerBroadcastIteration(DSDatagramServerContext ctx, DSConnectionList list, const char *title, const char *data, size_t len, DSConnectionFlags flags, size_t *wbytes) {
    DSConnectionListItem cur;

    int err;
    
    ds_assert_param(ctx);
    
    if (!list) list = ctx->clients;
    ds_assert_param(list);

    for (cur = list->first;cur;cur=cur->next) {
	if ((title)&&(cur->filter)&&(strncmp(title,cur->filter,cur->filter_size))) continue;
	err = dsDatagramServerWrite(ctx, &cur->addr, data, len, flags, wbytes);
	if (err) {
	    dsDebugServer(DS_DEBUG_WARNING, "dsDatagramServer(%p): Error broadcasting data to client(%p): %lx(%s)", ctx, cur->addr, err, dsError(err, NULL));
	}
    }
    
    return 0;
}

int dsDatagramServerBroadcastStart(DSDatagramServerContext ctx, DSConnectionList list) {
    ds_assert_param(ctx);

    if (!list) list = ctx->clients;
    ds_assert_param(list);
    
    return dsConnectionListLock(list);
}

int dsDatagramServerBroadcastDone(DSDatagramServerContext ctx, DSConnectionList list) {
    ds_assert_param(ctx);

    if (!list) list = ctx->clients;
    ds_assert_param(list);

    if (dsDurationIsZero(&ctx->expiration)) dsConnectionListClear(list);

    dsConnectionListUnLock(list);
    
    return 0;
}

int dsDatagramServerBroadcastCancel(DSDatagramServerContext ctx, DSConnectionList list) {
    ds_assert_param(ctx);

    if (!list) list = ctx->clients;
    ds_assert_param(list);

    dsConnectionListUnLock(list);
    
    return 0;
}

int dsDatagramServerBroadcast(DSDatagramServerContext ctx, DSConnectionList list, const char *title, const char *data, size_t len, DSConnectionFlags flags, size_t *wbytes) {
    dsDatagramServerBroadcastStart(ctx,list);
    dsDatagramServerBroadcastIteration(ctx,list,title,data,len,flags,wbytes);
    dsDatagramServerBroadcastDone(ctx,list);
    
    return 0;
}


int dsDatagramServerAddClient(DSDatagramServerContext ctx, DSConnectionAddress *addr) {
    return DS_ERR_NOTIMPLEMENTED;
}

int dsDatagramServerIsEmpty(DSDatagramServerContext ctx) {
    ds_assert_default(ctx, 1);
    return dsConnectionListIsEmpty(ctx->clients);
}
