#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <dslib.h>

#if defined(HAVE_DIRENT_H)
#include <dirent.h>
# if defined(HAVE_FNMATCH_H)
# include <fnmatch.h>
# else /* defined(HAVE_FNMATCH_H) */
# error "Can't find pattern matching routines!"
# endif /* defined(HAVE_FNMATCH_H) */
#else /* defined(_WIN32) */
# include <windows.h>
#endif /* defined(HAVE_DIRENT_H) */

#ifdef HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */

#ifdef HAVE_PWD_H
# include <pwd.h>
#endif /* HAVE_PWD_H */

#include <makepath.h>

#include <dsassert.h>
#include <dstools.h>
#include <dstime.h>
#include <dsdirectory.h>

int dsFileOpen(DSFileContext ctx, const char *path, const char *name) {
    size_t size = 0;
    char *stmp, *spath;
    
    ds_assert_param(ctx);
    ds_assert_init(ctx->initialized != DS_INITIALIZED);

#ifdef _WIN32
#else /* _WIN32 */
    ds_assert_param(name);
#endif /* _WIN32 */

    if (path) {
	size = strlen(path);

#ifdef _WIN32
	stmp = (char*)malloc((strlen(name?name:ctx->FindFileData.cFileName)+size+2)*sizeof(char));
#else /* _WIN32 */
	stmp = (char*)malloc((strlen(name)+size+2)*sizeof(char));
#endif /* _WIN32 */
	if (!stmp) return DS_ERR_MEM;

#ifdef _WIN32
	sprintf(stmp, "%s\\%s",path,name?name:ctx->FindFileData.cFileName);
#else /* _WIN32 */
	sprintf(stmp, "%s/%s",path,name);
#endif /* _WIN32 */

	spath = (char*)malloc((size+1)*sizeof(char));
	if (!spath) {
	    free(stmp);
	    return DS_ERR_MEM;
	}
	strcpy(spath, path);
    } else {
#ifdef _WIN32
	stmp = (char*)malloc((strlen(name?name:ctx->FindFileData.cFileName)+1)*sizeof(char));
#else /* _WIN32 */
	stmp = (char*)malloc((strlen(name)+1)*sizeof(char));
#endif /* _WIN32 */
	if (!stmp) return DS_ERR_MEM;
#ifdef _WIN32
	strcpy(stmp, name?name:ctx->FindFileData.cFileName);
#else /* _WIN32 */
	strcpy(stmp, name);
#endif /* _WIN32 */
    }

#if defined(_WIN32)
    if (name) {
	ctx->hFind = FindFirstFile(stmp, &ctx->FindFileData);
	if (ctx->hFind == INVALID_HANDLE_VALUE) {
	    free(stmp);
	    if (size) free(spath);
	    return DS_ERR_FILE_NOTFOUND;
	}
    }

    ctx->name = ctx->FindFileData.cFileName;
#else /* defined(_WIN32) */
    if (stat(stmp,&ctx->filestat)) {
	free(stmp);
	if (size) free(spath);
	if (errno == EACCES) return DS_ERR_PERMISSION;
	return DS_ERR_FILE_NOTFOUND;
    }
    
    ctx->name = stmp + size + 1;
#endif /* defined(_WIN32) */

    ctx->fullname = stmp;
    if (size) ctx->path = spath;
    else ctx->path = NULL;
    ctx->initialized = DS_INITIALIZED;

    return 0;
}

void dsFileClose(DSFileContext ctx) {
    ds_assert_returnctx(ctx);
#if defined(_WIN32)
    FindClose(ctx->hFind);
#else /* defined(_WIN32) */
#endif /* defined(_WIN32) */
    if (ctx->fullname) free(ctx->fullname);
    if (ctx->path) free(ctx->path);
    ctx->initialized = 0;
}

static void dsFileCloseByDirectory(DSFileContext ctx) {
    ds_assert_returnctx(ctx);
#if defined(_WIN32)
    FindClose(ctx->hFind);
#else /* defined(_WIN32) */
#endif /* defined(_WIN32) */
    if (ctx->fullname) free(ctx->fullname);
    ctx->initialized = 0;
}

static int dsFileOpenByDirectory(DSFileContext ctx, const char *name) {
    size_t size = 0;
    char *stmp, *path;
    
    ds_assert_param(ctx);
    if (ctx->initialized) dsFileCloseByDirectory(ctx);

#ifdef _WIN32
#else /* _WIN32 */
    ds_assert_param(name);
#endif /* _WIN32 */
    
    path = ctx->path;
    
    if (path) {
	size = strlen(path);

#ifdef _WIN32
	stmp = (char*)malloc((strlen(name?name:ctx->FindFileData.cFileName)+size+2)*sizeof(char));
#else /* _WIN32 */
	stmp = (char*)malloc((strlen(name)+size+2)*sizeof(char));
#endif /* _WIN32 */
	if (!stmp) return DS_ERR_MEM;

#ifdef _WIN32
	sprintf(stmp, "%s\\%s",path,name?name:ctx->FindFileData.cFileName);
#else /* _WIN32 */
	sprintf(stmp, "%s/%s",path,name);
#endif /* _WIN32 */
    } else {
#ifdef _WIN32
	stmp = (char*)malloc((strlen(name?name:ctx->FindFileData.cFileName)+1)*sizeof(char));
#else /* _WIN32 */
	stmp = (char*)malloc((strlen(name)+1)*sizeof(char));
#endif /* _WIN32 */
	if (!stmp) return DS_ERR_MEM;
#ifdef _WIN32
	strcpy(stmp, name?name:ctx->FindFileData.cFileName);
#else /* _WIN32 */
	strcpy(stmp, name);
#endif /* _WIN32 */
    }

#if defined(_WIN32)
    if (name) {
	ctx->hFind = FindFirstFile(stmp, &ctx->FindFileData);
	if (ctx->hFind == INVALID_HANDLE_VALUE) {
	    free(stmp);
	    return DS_ERR_FILE_NOTFOUND;
	}
    }

    ctx->name = ctx->FindFileData.cFileName;
#else /* defined(_WIN32) */
    if (stat(stmp,&ctx->filestat)) {
	free(stmp);
	if (errno == EACCES) return DS_ERR_PERMISSION;
	return DS_ERR_FILE_NOTFOUND;
    }
    
    ctx->name = stmp + size + 1;
#endif /* defined(_WIN32) */
    ctx->fullname = stmp;
    ctx->initialized = DS_INITIALIZED;

    return 0;
}


const char *dsFileGetName(const DSFileContext ctx) {
    ds_assert_nullctx(ctx);
    return ctx->name;
}

const char *dsFileGetPath(const DSFileContext ctx) {
    ds_assert_nullctx(ctx);
    return ctx->path;
}

const char *dsFileGetFullName(const DSFileContext ctx) {
    ds_assert_nullctx(ctx);
    return ctx->fullname;
}


time_t dsFileGetModificationTime(const DSFileContext ctx) {
    ds_assert_zeroctx(ctx);

#if defined(_WIN32)
    return dsTimeConvert_WIN32_FILETIME_to_UnixTime(ctx->FindFileData.ftLastWriteTime);
#else /* defined(_WIN32) */
    return ctx->filestat.st_mtime;    
#endif /* defined(_WIN32) */
}


int dsDirectoryOpen(DSDirectoryContext dir, const char *path, const char *mask) {
#if defined(_WIN32)
    size_t len;
    unsigned char slash = 0;
    char *stmp;
#endif /* defined(_WIN32) */

    ds_assert_param(dir);
    ds_assert_param(path);
    

#if defined(_WIN32)
    len = strlen(path) + 1;
    if (path[len-2]=='\\') slash = 1;
    else len++;
    
    if (mask) len += strlen(mask);
    else len += 1;
    
    stmp = (char*)dsMalloca(len*sizeof(char));
    if (!stmp) return DS_ERR_MEM;

    if (mask) {
     	if (slash) sprintf(stmp,"%s%s",path,mask);    
	else sprintf(stmp,"%s\\%s",path,mask);
    } else {
	if (slash) sprintf(stmp,"%s*",path);    
	else sprintf(stmp,"%s\\*",path);
    }
    
    dir->hFind = FindFirstFile(stmp, &dir->current.FindFileData);
    dsFreea(stmp);
    if (dir->hFind == INVALID_HANDLE_VALUE) dir->status = 2;
    else dir->status = 1;
#else /* defined(_WIN32) */
    dir->dir = opendir(path);
    if (!dir->dir) {
	if (errno = EACCES) return DS_ERR_PERMISSION;
	if (errno = ENOTDIR) return DS_ERR_DIRECTORY_NOTADIRECTORY;
	return DS_ERR_DIRECTORY_NOTFOUND;
    }
#endif /* defined(HAVE_DIRENT_H) */
    
    dir->path = (char*)malloc((strlen(path)+1)*sizeof(char));
    if (!dir->path) {
#if defined(_WIN32)
	FindClose(dir->hFind);
#else /* defined(_WIN32) */
	closedir(dir->dir);
#endif /* defined(_WIN32) */
	return DS_ERR_MEM;
    }
    strcpy(dir->path, path);

    if (mask) {
	dir->mask = (char*)malloc((strlen(mask)+1)*sizeof(char));
	if (!dir->mask) {
	    free(dir->path);
#if defined(_WIN32)
	    FindClose(dir->hFind);
#else /* defined(_WIN32) */
	    closedir(dir->dir);
#endif /* defined(_WIN32) */
	    return DS_ERR_MEM;
	}
	strcpy(dir->mask,mask);
    } else dir->mask = NULL;

    dir->current.initialized = 0;
    dir->current.path = dir->path;
    dir->initialized = DS_INITIALIZED;
    
    return 0;    
}

const DSFileContext dsDirectoryRead(DSDirectoryContext dir) {
    int err = 1;
#if defined(_WIN32)
#else /* defined(_WIN32) */
    struct dirent *file;
    char *stmp;
#endif /* defined(HAVE_DIRENT_H) */
    
    ds_assert_nullctx(dir);

#if defined(_WIN32)
    if (dir->status) {
	if (dir->status>1) return NULL;
	dir->status = 0;    

	err = dsFileOpenByDirectory(&dir->current, NULL);
    } else {
	if (!FindNextFile(dir->hFind, &dir->current.FindFileData)) {
	    err = GetLastError();
	    if (err == ERROR_NO_MORE_FILES) return NULL;
	    return NULL;
	}
	err = dsFileOpenByDirectory(&dir->current, NULL);
    }
    if (err) return NULL;
#else /* defined(_WIN32) */
    while (err) {
	do {
	    file = readdir(dir->dir);
	    if (!file) return NULL;
	    if (!dir->mask) break;
	} while (fnmatch(dir->mask, file->d_name, 0));

	err = dsFileOpenByDirectory(&dir->current, file->d_name);
    }
#endif /* defined(HAVE_DIRENT_H) */

    return &dir->current;
}

void dsDirectoryClose(DSDirectoryContext dir) {
    ds_assert_returnctx(dir);

    if (dir->current.initialized) dsFileCloseByDirectory(&dir->current);
    
#if defined(_WIN32)
    FindClose(dir->hFind);
#else /* defined(_WIN32) */
    closedir(dir->dir);
    if (dir->path) {
	free(dir->path);
	dir->path = NULL;
    }
    if (dir->mask) {
	free(dir->mask);
	dir->mask = NULL;
    }
#endif /* defined(HAVE_DIRENT_H) */

    dir->initialized = 0;
}


time_t dsDirectoryGetFileModificationTime(const char *fn) {
    int err;
    DSFileContextS f;
    time_t res;
    
    ds_assert_zero(fn);
    
    err = dsFileOpen(&f, NULL, fn);
    if (err) return 0;
    res = dsFileGetModificationTime(&f);
    dsFileClose(&f);

    return res;    
}

int dsDirectoryCreate(const char *path) {
    int err;
    
    ds_assert_param(path);
    
    err = make_path (path, -1, -1, -1, -1, 1);
    if (err) return DS_ERR_INVALID_PATH;
    return 0;
}

int dsFileCreateDirectory(const char *fn) {
    int err;
    size_t size;
    char *tmp;

    ds_assert_param(fn);
    
    size = strlen(fn);
    for (size--;size;size--) {
#ifdef _WIN32
	if (fn[size]=='\\') {
#else /* _WIN32 */
	if (fn[size]=='/') {
#endif /* _WIN32 */
	    break;
	}
    }
    
    if (size) {
        tmp = (char*)dsMalloca((size+1)*sizeof(char));
	if (!tmp) return DS_ERR_MEM;
	strncpy(tmp,fn,size);
	tmp[size]=0;

	err = dsDirectoryCreate(tmp);
	dsFreea(tmp);
	
	return err;
    }

    return 0;
}

char *dsDirectoryConcatenate(DSMallocContext mctx, const char *dir1, const char *dir2) {
    size_t size;
    char *res;
    
    ds_assert_null(dir1);
    ds_assert_null(dir2);
    
    size = strlen(dir1) + strlen(dir2) + 2;
    res = dsMalloc(mctx, size * sizeof(char));
    if (!res) return res;
#ifdef _WIN32
    sprintf(res, "%s\\%s", dir1, dir2);
#else /* _WIN32 */
    sprintf(res, "%s/%s", dir1, dir2);
#endif /* _WIN32 */
    return res;
}


int dsDirectoryParse(DSMallocContext mctx, char **dir) {
    char *cdir;

#ifdef _WIN32
    int i;
#else /* _WIN32 */
    const char *tmp;
# ifdef HAVE_PWD_H
    struct passwd *pw;
# endif /* HAVE_PWD_H */
#endif /* _WIN32 */
    
    ds_assert_param(dir);
    
    cdir = *dir;
    ds_assert_param(cdir);
    
    if (cdir[0]=='~') {
#ifdef _WIN32
	cdir[0] = '.';
#else
	tmp = getenv ("HOME");
# ifdef HAVE_PWD_H
	if (!tmp) {
	    setpwent ();
	    pw = getpwuid(getuid ());
	    endpwent ();
	    if ((pw)&&(pw->pw_dir)) tmp = pw->pw_dir;
	}
# endif /* HAVE_PWD_H */
	if (tmp) {
	    cdir = dsMalloc(mctx, (strlen(tmp) + strlen(cdir))*sizeof(char));
	    if (!cdir) return DS_ERR_MEM;
	    sprintf(cdir,"%s%s", tmp, (*dir)+1);
	    dsFree(mctx, *dir);
	    *dir = cdir;
	} else cdir[0] = '.';
#endif /* _WIN32 */    
    }
    
#ifdef _WIN32
    for (i=0;cdir[i];i++)
	if (cdir[i]=='/') cdir[i] = DS_DIRECTORY_SEPARATOR;
#endif /* _WIN32 */

    return 0;    
}

int dsDirectoryParseFrom(DSMallocContext mctx, char **dir, const char *orig) {
    int err;
    
    ds_assert_param(dir);
    ds_assert_param(orig);
    
    *dir = NULL;
    
    err = dsExchangeString(mctx, dir, orig);
    if (err) return err;
    
    return dsDirectoryParse(mctx, dir);
}
