#include <stdio.h>
#include <errno.h>

#include <dslib.h>
#include <dsassert.h>
#include <dsdebug.h>
#include <dsmalloc.h>
#include <dslock.h>
#include <dsbasethread.h>

#define DS_LOCK_THREADS_NUMBER	64

/*
Linux Implementation Notes (glibc 2.3.2):
1. EDEADLK isn't returned when trying occure lock from thread alredy have it
2. Lock locked from one thread can be unlocked from another one 
*/

static void dsLockZeroContext(DSLockContext ctx) {
    ctx->initialized = 0;
#if defined(HAVE_PTHREAD_H)
# ifndef DS_SUPPORT_RWLOCK
    ctx->readers = 0;
    ctx->waiting_writers = 0;
    ctx->write_precedence = 1;
# endif /* ! DS_SUPPORT_RWLOCK */
#else /* defined(HAVE_WIN32_THREADS) */
    ctx->readers = 0;
    ctx->waiting_writers = 0;
    ctx->read_mutex = NULL;
    ctx->write_mutex = NULL;
    ctx->lock_mutex = NULL;
    ctx->write_precedence = 1;
#endif /* defined(HAVE_PTHREAD_H) */

#ifdef DS_TRACK_LOCKS
    ctx->omit_tracking = !DS_TRACK_LOCKS_TRACK_BY_DEFAULT;
# ifndef HAVE_PTHREAD_H
    ctx->threads_mutex = NULL;
# endif /* ! HAVE_PTHREAD_H */
    ctx->threads  = NULL;
#endif /* DS_TRACK_LOCKS */

}

DSLockContext dsLockCreateContext(DSMallocContext mctx) {
    DSLockContext ctx;
    ctx = (DSLockContext)dsMalloc(mctx, sizeof(struct DSLockContextT));
    if (ctx) {
	ctx->malloc = mctx;
	dsLockZeroContext(ctx);
    }
    return ctx;
}

int dsLockInitContext(DSLockContext ctx) {
#if defined(HAVE_PTHREAD_H)
    int err;
#endif /* defined(HAVE_PTHREAD_H) */
    
    ds_assert_param(ctx);

    dsLockZeroContext(ctx);

#if defined(HAVE_PTHREAD_H)
# ifdef DS_SUPPORT_RWLOCK
    err = pthread_rwlock_init(&ctx->lock_mutex,NULL);
# else
    err = pthread_mutex_init(&ctx->lock_mutex,NULL);
    if (!err) err = pthread_mutex_init(&ctx->read_mutex,NULL);
    if (!err) err = pthread_mutex_init(&ctx->write_mutex,NULL);
# endif /* DS_SUPPORT_RWLOCK */
    if (err) return DS_ERR_SYSCALL_PTHREAD_MUTEX_INIT;
#else /* defined(HAVE_WIN32_THREADS) */
    ctx->lock_mutex = CreateMutex(NULL,FALSE,NULL);
    ctx->read_mutex = CreateMutex(NULL,FALSE,NULL);
    ctx->write_mutex = CreateMutex(NULL,FALSE,NULL);
    if ((!ctx->lock_mutex)||(!ctx->read_mutex)||(!ctx->write_mutex)) {
	if (ctx->lock_mutex) {
	    CloseHandle(ctx->lock_mutex);
	    ctx->lock_mutex = NULL;
	}
	if (ctx->read_mutex) {
	    CloseHandle(ctx->read_mutex);
	    ctx->read_mutex = NULL;
	}
	if (ctx->write_mutex) {
	    CloseHandle(ctx->read_mutex);
	    ctx->write_mutex = NULL;
	}
	return DS_ERR_WINAPI_MUTEX_CREATE;
    }
#endif /* defined(HAVE_PTHREAD_H) */

#ifdef DS_TRACK_LOCKS
    ctx->threads = dsHashTableInit(ctx->malloc, DS_LOCK_THREADS_NUMBER, DS_HASH_TYPE_ARRAY, NULL);
    if (ctx->threads) {
	dsHashTableSetStringSize(ctx->threads, sizeof(DSThreadHandleS));
	dsHashTableSetDestructor(ctx->threads, NULL);

# ifdef HAVE_PTHREAD_H
	if (pthread_mutex_init(&ctx->threads_mutex, NULL)) err = DS_ERR_SYSCALL_PTHREAD_MUTEX_INIT;
# else /* HAVE_PTHREAD_H */
	ctx->threads_mutex = CreateMutex(NULL,FALSE,NULL);
	if (!ctx->threads_mutex) err = DS_ERR_WINAPI_MUTEX_CREATE;
# endif /* HAVE_PTHREAD_H */
    } else err = DS_ERR_MEM;

    if (err) {
# ifdef HAVE_PTHREAD_H
#  ifdef DS_SUPPORT_RWLOCK
	pthread_rwlock_destroy(&ctx->lock_mutex);
#  else /* DS_SUPPORT_RWLOCK */
	pthread_mutex_destroy(&ctx->lock_mutex);
	pthread_mutex_destroy(&ctx->read_mutex);
	pthread_mutex_destroy(&ctx->write_mutex);
#  endif /* DS_SUPPORT_RWLOCK */
# else /* HAVE_PTHREAD_H */
	CloseHandle(ctx->lock_mutex);
	CloseHandle(ctx->read_mutex);
	CloseHandle(ctx->write_mutex);
	ctx->lock_mutex = NULL;
	ctx->read_mutex = NULL;
	ctx->write_mutex = NULL;
# endif /* HAVE_PTHREAD_H */
	return err;
    }
#endif /* DS_TRACK_LOCKS */

    ctx->initialized = DS_INITIALIZED;
    return 0;
}

void dsLockFreeContext(DSLockContext ctx) {
    ds_assert_return(ctx);
#if defined(HAVE_PTHREAD_H)
# ifdef DS_SUPPORT_RWLOCK
    if (ctx->initialized) pthread_rwlock_destroy(&ctx->lock_mutex);
# else
    if (ctx->initialized) {
	pthread_mutex_destroy(&ctx->lock_mutex);
	pthread_mutex_destroy(&ctx->read_mutex);
	pthread_mutex_destroy(&ctx->write_mutex);
    }
# endif /* DS_SUPPORT_RWLOCK */
#else /* defined(HAVE_WIN32_THREADS) */
    if (ctx->lock_mutex) CloseHandle(ctx->lock_mutex);
    if (ctx->read_mutex) CloseHandle(ctx->read_mutex);
    if (ctx->write_mutex) CloseHandle(ctx->write_mutex);
#endif /* defined(HAVE_PTHREAD_H) */

#ifdef DS_TRACK_LOCKS
    if ((ctx->initialized)&&(!dsHashTableIsEmpty(ctx->threads))) {
	dsDebugServer(DS_DEBUG_ERROR, "dsLock %p: FreeContext. Error: Still locked", ctx);
    }
# ifdef HAVE_PTHREAD_H
    if (ctx->initialized) pthread_mutex_destroy(&ctx->threads_mutex);
# else /* HAVE_PTHREAD_H */
    if (ctx->threads_mutex) CloseHandle(ctx->threads_mutex);
# endif /* HAVE_PTHREAD_H */
    if (ctx->threads) dsHashTableDestroy(ctx->threads);
#endif /* DS_TRACK_LOCKS */

    ctx->initialized = 0;
    dsLockZeroContext(ctx);
}

void dsLockDestroyContext(DSLockContext ctx) {
    ds_assert_return(ctx);
    dsLockFreeContext(ctx);
    dsFree(ctx->malloc, ctx);
}

int dsLockSetWriterPrecedence(DSLockContext ctx, unsigned char precedence) {
#ifndef DS_SUPPORT_RWLOCK
    ctx->write_precedence = precedence;
#endif /* DS_SUPPORT_RWLOCK */
    return 0;
}

int dsLockSetTracking(DSLockContext ctx, unsigned char tracking) {
#ifdef DS_TRACK_LOCKS
    ctx->omit_tracking = !tracking;
#endif /* DS_TRACK_LOCKS */
    return 0;
}

#ifdef DS_TRACK_LOCKS
static int dsLockReadInternal(DSLockContext ctx) {
#else /* DS_TRACK_LOCKS */
int dsLockRead(DSLockContext ctx) {
#endif /* DS_TRACK_LOCKS */
    int err;

    ds_assert_ctx(ctx);

#if defined(HAVE_PTHREAD_H)
# ifdef DS_SUPPORT_RWLOCK
    err = pthread_rwlock_rdlock(&ctx->lock_mutex);
# else /* DS_SUPPORT_RWLOCK */
    err = pthread_mutex_lock(&ctx->read_mutex);
    if (!err) {
	if ((ctx->readers)&&(!ctx->waiting_writers)) ctx->readers++;
	else {
	    err = pthread_mutex_lock(&ctx->lock_mutex);
	    if (!err) ctx->readers = 1;
	}
	pthread_mutex_unlock(&ctx->read_mutex);
    }
# endif /* DS_SUPPORT_RWLOCK */
    if (err) {
	if (err == EDEADLK) err = DS_ERR_LOCK_SELFLOCKED;
	else err = DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK;
    }
#else /* defined(HAVE_WIN32_THREADS) */
    if (WaitForSingleObject(ctx->read_mutex, INFINITE) == WAIT_OBJECT_0) {
	if (ctx->readers) ctx->readers++;
	else {
	    if (WaitForSingleObject(ctx->lock_mutex, INFINITE) == WAIT_OBJECT_0) {
		ctx->readers = 1;
		err = 0;
	    } else err = -1;
	}
	ReleaseMutex(ctx->read_mutex);
    } else err = -1;

    if (err) return DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK;
#endif /* defined(HAVE_PTHREAD_H) */

    return err;
}

#ifdef DS_TRACK_LOCKS
static int dsLockWriteInternal(DSLockContext ctx) {
#else /* DS_TRACK_LOCKS */
int dsLockWrite(DSLockContext ctx) {
#endif /* DS_TRACK_LOCKS */
    int err;

    ds_assert_ctx(ctx);

#if defined(HAVE_PTHREAD_H)
# ifdef DS_SUPPORT_RWLOCK
    err = pthread_rwlock_wrlock(&ctx->lock_mutex);
# else /* DS_SUPPORT_RWLOCK */
    if (ctx->write_precedence) {
	err = pthread_mutex_lock(&ctx->write_mutex);
	if (!err) {
	    ctx->waiting_writers++;
	    pthread_mutex_unlock(&ctx->write_mutex);
	
	    err = pthread_mutex_lock(&ctx->lock_mutex);
	
	    while (pthread_mutex_lock(&ctx->write_mutex));
	    ctx->waiting_writers--;
	    pthread_mutex_unlock(&ctx->write_mutex);
	}
    } else err = pthread_mutex_lock(&ctx->lock_mutex);
# endif /* DS_SUPPORT_RWLOCK */
    if (err) {
	if (err == EDEADLK) err = DS_ERR_LOCK_SELFLOCKED;
	else err = DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK;
    }
#else /* defined(HAVE_WIN32_THREADS) */
    if (ctx->write_precedence) {
	if (WaitForSingleObject(ctx->write_mutex, INFINITE) == WAIT_OBJECT_0) {
	    ctx->waiting_writers++;
	    ReleaseMutex(ctx->write_mutex);
	    
	    if (WaitForSingleObject(ctx->lock_mutex, INFINITE) == WAIT_OBJECT_0) err = 0;
	    else err = -1;
	
	    while (WaitForSingleObject(ctx->write_mutex, INFINITE) != WAIT_OBJECT_0);
	    ctx->waiting_writers--;
	    ReleaseMutex(ctx->write_mutex);
	} else err = -1;
    } else {
	if (WaitForSingleObject(ctx->lock_mutex, INFINITE) == WAIT_OBJECT_0) err = 0;
	else err = -1;
    }

    if (err) return DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK;
#endif /* defined(HAVE_PTHREAD_H) */

    return err;
}

#ifdef DS_TRACK_LOCKS
static int dsLockTryReadInternal(DSLockContext ctx) {
#else /* DS_TRACK_LOCKS */
int dsLockTryRead(DSLockContext ctx) {
#endif /* DS_TRACK_LOCKS */
#if defined(HAVE_PTHREAD_H)
    int err;
#endif /* defined(HAVE_PTHREAD_H) */

    ds_assert_ctx(ctx);
#if defined(HAVE_PTHREAD_H)
# ifdef DS_SUPPORT_RWLOCK
    err = pthread_rwlock_tryrdlock(&ctx->lock_mutex);
# else /* DS_SUPPORT_RWLOCK */
    if (ctx->waiting_writers) return DS_ERR_LOCK_BUSY;
    
    err = pthread_mutex_trylock(&ctx->read_mutex);
    if (!err) {
	if (ctx->readers) ctx->readers++;
	else {
	    err = pthread_mutex_trylock(&ctx->lock_mutex);
	    if (!err) ctx->readers = 1;
	}
	pthread_mutex_unlock(&ctx->read_mutex);
    }
# endif /* DS_SUPPORT_RWLOCK */
    if (err) {
	if (err == EDEADLK) return DS_ERR_LOCK_SELFLOCKED;
	if (err == EBUSY) return DS_ERR_LOCK_BUSY;
	return DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK;
    }
#else /* defined(HAVE_WIN32_THREADS) */
    if (WaitForSingleObject(ctx->lock_mutex, 0) != WAIT_OBJECT_0) return DS_ERR_LOCK_BUSY;
#endif /* defined(HAVE_PTHREAD_H) */
    return 0;
}

#ifdef DS_TRACK_LOCKS
static int dsLockTryWriteInternal(DSLockContext ctx) {
#else /* DS_TRACK_LOCKS */
int dsLockTryWrite(DSLockContext ctx) {
#endif /* DS_TRACK_LOCKS */
#if defined(HAVE_PTHREAD_H)
    int err;
#endif /* defined(HAVE_PTHREAD_H) */

    ds_assert_ctx(ctx);
#if defined(HAVE_PTHREAD_H)
# ifdef DS_SUPPORT_RWLOCK
    err = pthread_rwlock_trywrlock(&ctx->lock_mutex);
# else /* DS_SUPPORT_RWLOCK */
    err = pthread_mutex_trylock(&ctx->lock_mutex);
# endif /* DS_SUPPORT_RWLOCK */
    if (err) {
	if (err == EBUSY) return DS_ERR_LOCK_BUSY;
	if (err == EDEADLK) return DS_ERR_LOCK_SELFLOCKED;
	return DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK;
    }
#else /* defined(HAVE_WIN32_THREADS) */
    if (WaitForSingleObject(ctx->lock_mutex, 0) != WAIT_OBJECT_0) return DS_ERR_LOCK_BUSY;

    if (WaitForSingleObject(ctx->read_mutex, 0) == WAIT_OBJECT_0) {
	if (ctx->readers) ctx->readers++;
	else {
	    if (WaitForSingleObject(ctx->lock_mutex, 0) == WAIT_OBJECT_0) ctx->readers = 1;
	    else {
		ReleaseMutex(ctx->lock_mutex);
		return DS_ERR_LOCK_BUSY;
	    }
	}
	ReleaseMutex(ctx->read_mutex);
    } else return DS_ERR_LOCK_BUSY;
#endif /* defined(HAVE_PTHREAD_H) */

    return 0;
}

#ifdef DS_TRACK_LOCKS
static int dsLockReturnInternal(DSLockContext ctx) {
#else /* DS_TRACK_LOCKS */
int dsLockReturn(DSLockContext ctx) {
#endif /* DS_TRACK_LOCKS */
    int err;

    ds_assert_ctx(ctx);
#if defined(HAVE_PTHREAD_H)
# ifdef DS_SUPPORT_RWLOCK
    err = pthread_rwlock_unlock(&ctx->lock_mutex);
# else /* DS_SUPPORT_RWLOCK */
    if (ctx->readers) { // Releasing read lock
	if (ctx->waiting_writers) { // Pending waiting writers (so, readers is surely not increased, decreases still possible just here)
	    err = pthread_mutex_lock(&ctx->write_mutex);
	    if (!err) {
		if ((--ctx->readers)==0) {
		    err = pthread_mutex_unlock(&ctx->lock_mutex);
	        }
	        pthread_mutex_unlock(&ctx->write_mutex);
	    }
	} else {
	    err = pthread_mutex_lock(&ctx->read_mutex);
	    if (!err) {
		if ((--ctx->readers)==0) {
		    err = pthread_mutex_unlock(&ctx->lock_mutex);
		}
		pthread_mutex_unlock(&ctx->read_mutex);
	    }
	}
    }
    else err = pthread_mutex_unlock(&ctx->lock_mutex);	// Releasing write lock
# endif /* DS_SUPPORT_RWLOCK */
#else /* defined(HAVE_WIN32_THREADS) */
    if (ctx->readers) { // Releasing read lock
	if (ctx->waiting_writers) { // Pending waiting writers (so, readers is surely not increased, decreases still possible just here)
	    if (WaitForSingleObject(ctx->write_mutex, 0) == WAIT_OBJECT_0) {
		if ((--ctx->readers)==0) {
		    if (ReleaseMutex(ctx->lock_mutex)) err = 0;
		    else err = -1;
	        }
		ReleaseMutex(ctx->write_mutex);
	    } else err = -1;
	} else {
	    if (WaitForSingleObject(ctx->read_mutex, 0) == WAIT_OBJECT_0) {
		if ((--ctx->readers) == 0) {
		    if (ReleaseMutex(ctx->lock_mutex)) err = 0;
		    else err = -1;
		} else err = 0;
		ReleaseMutex(ctx->read_mutex);
	    } else err = -1;
	}
    } else { // Releasing write lock
	if (ReleaseMutex(ctx->lock_mutex)) err = 0;
	else err = -1;
    }
#endif /* defined(HAVE_PTHREAD_H) */
    if (err) {
	if (err == EPERM) return DS_ERR_PERMISSION;
	return DS_ERR_SYSCALL_PTHREAD_MUTEX_UNLOCK;
    }

    return 0;
}

#ifdef DS_TRACK_LOCKS
enum DSLockTrackerAction {
    DS_LOCK_TRACKER_READ,
    DS_LOCK_TRACKER_WRITE,
    DS_LOCK_TRACKER_TRYREAD,
    DS_LOCK_TRACKER_TRYWRITE,
};

static int dsLockTracker(DSLockContext ctx, enum DSLockTrackerAction action) {
    int err = 0;
    DSThreadHandle myhandle, handle;
    unsigned long int selflocked = 0;
    
    ds_assert_param(ctx); 
    ds_assert_param(ctx->initialized);
    
    myhandle = (DSThreadHandle)dsMalloc(ctx->malloc, sizeof(DSThreadHandleS) + sizeof(unsigned long int));
    if (myhandle) {
	dsThreadGetCurrentHandle(myhandle);
	if ((action == DS_LOCK_TRACKER_READ)||(action == DS_LOCK_TRACKER_TRYREAD)) {
	    *(unsigned long int*)(((char*)myhandle) + sizeof(DSThreadHandleS)) = 1;
	} else {
	    *(unsigned long int*)(((char*)myhandle) + sizeof(DSThreadHandleS)) = 0;
	}
    } else return DS_ERR_MEM;

# ifdef HAVE_PTHREAD_H
    if (!pthread_mutex_lock(&ctx->threads_mutex)) {
# else /* HAVE_PTHREAD_H */
    if (WaitForSingleObject(ctx->threads_mutex, INFINITE) == WAIT_OBJECT_0) {
# endif /* HAVE_PTHREAD_H */
	handle = dsHashTableFind(ctx->threads, (char*)myhandle);
	if (handle) {
	    if ((action == DS_LOCK_TRACKER_READ)||(action == DS_LOCK_TRACKER_TRYREAD)) {
		selflocked = *(unsigned long int*)(((char*)handle) + sizeof(DSThreadHandleS));
		if (!selflocked) err = DS_ERR_LOCK_SELFLOCKED;
	    } else err = DS_ERR_LOCK_SELFLOCKED;
	}
# ifdef HAVE_PTHREAD_H
	pthread_mutex_unlock(&ctx->threads_mutex);
# else /* HAVE_PTHREAD_H */
	ReleaseMutex(ctx->threads_mutex);
# endif /* HAVE_PTHREAD_H */
    } else err = DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK;
    
    if (err) {
	dsFree(ctx->malloc, myhandle);

	if (err == DS_ERR_LOCK_SELFLOCKED)
	    dsDebugServer(DS_DEBUG_WARNING, "dsLock %p: Lock. Error: Already locked by this thread", ctx);
	else
	    dsDebugServer(DS_DEBUG_WARNING, "dsLock %p: Lock. Error: Request validation failed", ctx);

	return err;
    }
    
    if (selflocked) {
	dsDebugServer(DS_DEBUG_WARNING, "dsLock %p: Lock. WARNING: Already locked by this thread", ctx);
    }

    switch (action) {
	case DS_LOCK_TRACKER_READ:
	    err = dsLockReadInternal(ctx);
	break;
	case DS_LOCK_TRACKER_WRITE:
	    err = dsLockWriteInternal(ctx);
	break;
	case DS_LOCK_TRACKER_TRYREAD:
	    err = dsLockTryReadInternal(ctx);
	break;
	case DS_LOCK_TRACKER_TRYWRITE:
	    err = dsLockTryWriteInternal(ctx);
	break;
	default:
	    err = DS_ERR_INTERNAL;
    }

    if (err) {
	if (err == DS_ERR_LOCK_SELFLOCKED)
	    dsDebugServer(DS_DEBUG_WARNING, "dsLock %p: Lock. Error: Deadlock detected (by system)", ctx);
	else if (err != DS_ERR_LOCK_BUSY) 
	    dsDebugServer(DS_DEBUG_WARNING, "dsLock %p: Lock. Error: Locking error (unknown)", ctx);

	dsFree(ctx->malloc, myhandle);
	return err;
    }

# ifdef HAVE_PTHREAD_H
    if (!pthread_mutex_lock(&ctx->threads_mutex)) {
# else /* HAVE_PTHREAD_H */
    if (WaitForSingleObject(ctx->threads_mutex, INFINITE) == WAIT_OBJECT_0) {
# endif /* HAVE_PTHREAD_H */
	if (selflocked) {
	    *(unsigned long int*)(((char*)handle) + sizeof(DSThreadHandleS)) = selflocked + 1;
	} else err = dsHashTableInsert(ctx->threads, myhandle);
# ifdef HAVE_PTHREAD_H
	pthread_mutex_unlock(&ctx->threads_mutex);
# else /* HAVE_PTHREAD_H */
	ReleaseMutex(ctx->threads_mutex);
# endif /* HAVE_PTHREAD_H */
    } else err = DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK;


    if (err) {
	dsLockReturnInternal(ctx);
	dsFree(ctx->malloc, myhandle);
    }
    
    return err;
}

int dsLockRead(DSLockContext ctx) {
    if (ctx->omit_tracking) return dsLockReadInternal(ctx);
    return dsLockTracker(ctx, DS_LOCK_TRACKER_READ);
}

int dsLockWrite(DSLockContext ctx) {
    if (ctx->omit_tracking) return dsLockWriteInternal(ctx);
    return dsLockTracker(ctx, DS_LOCK_TRACKER_WRITE);
}

int dsLockTryRead(DSLockContext ctx) {
    if (ctx->omit_tracking) return dsLockTryReadInternal(ctx);
    return dsLockTracker(ctx, DS_LOCK_TRACKER_TRYREAD);
}

int dsLockTryWrite(DSLockContext ctx) {
    if (ctx->omit_tracking) return dsLockTryWriteInternal(ctx);
    return dsLockTracker(ctx, DS_LOCK_TRACKER_TRYWRITE);
}

int dsLockReturn(DSLockContext ctx) {
    int err = 0;
    void *handle;
    DSThreadHandleS myhandle;
    unsigned long int selflocked;

    if (ctx->omit_tracking) return dsLockReturnInternal(ctx);
    
    ds_assert_param(ctx); 
    ds_assert_param(ctx->initialized);
    
    dsThreadGetCurrentHandle(&myhandle);

# ifdef HAVE_PTHREAD_H
    if (!pthread_mutex_lock(&ctx->threads_mutex)) {
# else /* HAVE_PTHREAD_H */
    if (WaitForSingleObject(ctx->threads_mutex, INFINITE) == WAIT_OBJECT_0) {
# endif /* HAVE_PTHREAD_H */
	handle = dsHashTableFind(ctx->threads, (char*)&myhandle);
	if (handle) {
	    selflocked = *(unsigned long int*)(((char*)handle) + sizeof(DSThreadHandleS));
	    if (selflocked > 1) {
		*(unsigned long int*)(((char*)handle) + sizeof(DSThreadHandleS))=selflocked - 1;
		err = 0;
	    } else err = dsHashTableRemove(ctx->threads, handle);
	}
	else err = DS_ERR_LOCK_NOTLOCKED;
# ifdef HAVE_PTHREAD_H
	pthread_mutex_unlock(&ctx->threads_mutex);
# else /* HAVE_PTHREAD_H */
	ReleaseMutex(ctx->threads_mutex);
# endif /* HAVE_PTHREAD_H */
    } else err = DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK;
    
    if (err) {
	if (err == DS_ERR_LOCK_NOTLOCKED) 	
	    dsDebugServer(DS_DEBUG_WARNING, "dsLock %p: Return. Error: Not locked by this thread", ctx);
	else 
	    dsDebugServer(DS_DEBUG_WARNING, "dsLock %p: Return. Error: Request validation failed", ctx);
	return err;
    }
    
    return dsLockReturnInternal(ctx);	/* We asume Return always succeeds, otherwise it is too complicated */
}

#endif /* DS_TRACK_LOCKS */
