#ifndef _DS_LOCK_H
#define _DS_LOCK_H

#include <dslib.h>

#if defined(HAVE_PTHREAD_H)
# include <pthread.h>
#elif defined(HAVE_WIN32_THREADS)
# include <windows.h>
#else /* defined(HAVE_PTHREAD_H) */
# error "Missing threading library!"
#endif /* defined(HAVE_PTHREAD_H) */

#include <dsmalloc.h>
#include <dshashtable.h>

/*
OPCItem Locking facility. Should be based on multiple read one write scheme.

Linux Implementation Notes (glibc 2.3.2):
1. EDEADLK isn't returned when trying occure lock from thread alredy have it
2. Lock locked from one thread can be unlocked from another one


ToDo: I need "strong writer synchronization". Check what I'll have here... 

ASSUMPTIONS
-----------
Relocking from some thread policy ( Self locking deadlocks )

Original Lock 		New Lock		Required	Lock(RW)	ILock(RW)
Read			Read			Yes		No(Yes)		Yes(Yes)
Write			Read			No		No(No)		Yes(Yes)
Write			Write			No		No(No)		Yes(Yes)
Read			Write			No		No(No)		No(No)

"Yes" means locking function will correctly lock/unlock second time or return EDEADLK error


Return PROBLEMS
----------------
    There is a problem with Lock/Return single mutex for different threads
    (Locked in one thread, and returned in another). pthread_mutex_return
    in that case returns EPERM error. However, the documentation states
    such behaviour should be only for ErrorChecking mutexes (GNU extension,
    not used in dslib).
    Especially this behaviour dangerous for RWLock emulation. Since it is
    possible what Read lock occured in one thread and returned in another
    one.
*/


struct DSLockContextT {
    DSMallocContext malloc;
#if defined(HAVE_PTHREAD_H)
# ifdef DS_SUPPORT_RWLOCK
    pthread_rwlock_t lock_mutex;
# else
    unsigned long readers;
    unsigned long waiting_writers;
    pthread_mutex_t read_mutex;
    pthread_mutex_t write_mutex;
    pthread_mutex_t lock_mutex;
    unsigned char write_precedence;
# endif /* DS_SUPPORT_RWLOCK */
#else /* HAVE_WIN32_THREADS */
    unsigned long readers;
    unsigned long waiting_writers;
    HANDLE read_mutex;
    HANDLE write_mutex;
    HANDLE lock_mutex;
    unsigned char write_precedence;
#endif /* defined(HAVE_PTHREAD_H) */

#ifdef DS_TRACK_LOCKS
    unsigned char omit_tracking;
# ifdef HAVE_PTHREAD_H
    pthread_mutex_t threads_mutex;
# else /* HAVE_PTHREAD_H */
    HANDLE threads_mutex;
# endif /* HAVE_PTHREAD_H */
    DSHashTable threads;
#endif /* DS_TRACK_LOCKS */

    unsigned char initialized;
};
typedef struct DSLockContextT DSLockContextS;
typedef struct DSLockContextT *DSLockContext;

DSLockContext dsLockCreateContext(DSMallocContext mctx);
int dsLockInitContext(DSLockContext ctx);
void dsLockFreeContext(DSLockContext ctx);
void dsLockDestroyContext(DSLockContext ctx);

int dsLockSetWriterPrecedence(DSLockContext ctx, unsigned char precedence);
int dsLockSetTracking(DSLockContext ctx, unsigned char tracking);

int dsLockRead(DSLockContext ctx);
int dsLockWrite(DSLockContext ctx);
int dsLockTryRead(DSLockContext ctx);
int dsLockTryWrite(DSLockContext ctx);
int dsLockReturn(DSLockContext ctx);

#endif /* _DS_LOCK_H */
