#include <string.h>
#include <stdlib.h>

#include <dslib.h>
#include <dsassert.h>

#include <dsdebug.h>
#include <dshashtable.h>
#include <dsmalloc.h>
#include <dslock.h>

#define DS_MALLOC_ALLOCATIONS_NUMBER 65535

#ifdef DS_TRACK_MEMORY
static DSHashTable allocations = NULL;
static DSLockContext allocations_lock = NULL;
static char tracker_initialized = 0;
#endif /* DS_TRACK_MEMORY */


int dsMallocInit() {
#ifdef DS_TRACK_MEMORY
    int err;
        
    allocations = dsHashTableInit(NULL, DS_MALLOC_ALLOCATIONS_NUMBER, DS_HASH_TYPE_ARRAY, NULL);
    if (allocations) {
	dsHashTableSetStringSize(allocations, sizeof(void*));
	dsHashTableSetDestructor(allocations, NULL);
    } else return DS_ERR_MEM;
    
    allocations_lock = dsLockCreateContext(NULL);
    if (allocations_lock) {
	err = dsLockInitContext(allocations_lock);
	if (!err) err = dsLockSetTracking(allocations_lock, 0);
    } else err = DS_ERR_MEM;
    
    if (err) {
	dsMallocDestroy();
	return err;
    }

    tracker_initialized = 1;
#endif /* DS_TRACK_MEMORY */
    return 0;
}
void dsMallocDestroy() {
#ifdef DS_TRACK_MEMORY
    int check = tracker_initialized;

    tracker_initialized = 0;

    if (allocations_lock) {
	dsLockDestroyContext(allocations_lock);
	allocations_lock = NULL;
    }
    
    if (allocations) { 
	if ((check)&&(!dsHashTableIsEmpty(allocations))) {
	    dsDebugServer(DS_DEBUG_ERROR, "dsMalloc %p: Ghost (unfreed) memory is detected", NULL);
	}

	dsHashTableDestroy(allocations);
	allocations = NULL;
    }
#endif /* DS_TRACK_MEMORY */
}

#ifdef _WIN32
static HANDLE ProcessHeapHandle = NULL;

void *dsMallocWin32Heap(size_t size) {
    if (!ProcessHeapHandle) ProcessHeapHandle = GetProcessHeap();
    return HeapAlloc(ProcessHeapHandle, 0, size);
}

void *dsReallocWin32Heap(void *ptr, size_t size) {
    if (!ProcessHeapHandle) ProcessHeapHandle = GetProcessHeap();
    if (size) {
	if (ptr) return HeapReAlloc(ProcessHeapHandle, 0, ptr, size);
	else return HeapAlloc(ProcessHeapHandle, 0, size);
    } else if (ptr) HeapFree(ProcessHeapHandle, 0, ptr);
    return NULL;
}

void *dsCallocWin32Heap(size_t nmemb, size_t size) {
    return HeapAlloc(ProcessHeapHandle, HEAP_ZERO_MEMORY, nmemb*size);
}

void *dsStrdupWin32Heap(const char *str) {
    char *dup;
    size_t size;
    
    size = strlen(str);
    dup = dsMallocWin32Heap((size+1)*sizeof(char));
    if (!dup) return dup;
    memcpy(dup,str,(size+1)*sizeof(char));
    return dup;
}

void dsFreeWin32Heap(void *ptr) {
    HeapFree(ProcessHeapHandle, 0, ptr);
}
#endif /* _WIN32 */

static DSMallocContext default_ctx = NULL;

DSMallocContext dsMallocGetDefault() {
    return default_ctx;
}

int dsMallocSetDefault(DSMallocContext ctx) {
    default_ctx = ctx;
    return 0;
}

DSMallocContext dsMallocCreateContext(DSMallocContext mctx) {
#ifdef DS_TRACK_MEMORY
    int err;
#endif /* DS_TRACK_MEMORY */

    DSMallocContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "dsMallocCreateContext");

    if (mctx) ctx = (DSMallocContext)dsMalloc(mctx,sizeof(struct DSMallocContextT));
    else {
	    // Should be allocated from shared segment...
	ctx = (DSMallocContext)malloc(sizeof(struct DSMallocContextT));
    }

    if (ctx) {
	ctx->parrent = mctx;

#ifdef DS_TRACK_MEMORY
	ctx->tracker_initialized = 0;
	
	ctx->allocations = dsHashTableInit(ctx, DS_MALLOC_ALLOCATIONS_NUMBER, DS_HASH_TYPE_ARRAY, NULL);
	if (ctx->allocations) {
	    dsHashTableSetStringSize(ctx->allocations, sizeof(void*));
	    dsHashTableSetDestructor(ctx->allocations, NULL);
	} else err = DS_ERR_MEM;
    
	ctx->allocations_lock = dsLockCreateContext(ctx);
	if (ctx->allocations_lock) {
	    err = dsLockInitContext(ctx->allocations_lock);
	    if (!err) err = dsLockSetTracking(ctx->allocations_lock, 0);
	} else err = DS_ERR_MEM;
    
        if (err) {
	    dsMallocDestroyContext(ctx);
	    return NULL;
	}

	ctx->tracker_initialized = 1;
#endif /* DS_TRACK_MEMORY */
    }

    return ctx;
}

void dsMallocDestroyContext(DSMallocContext ctx) {
#ifdef DS_TRACK_MEMORY
    int check = ctx->tracker_initialized;
#endif /* DS_TRACK_MEMORY */

    dsDebugServer(DS_DEBUG_NOTICE, "dsMallocDestroyContext");

    ds_assert_return(ctx);

#ifdef DS_TRACK_MEMORY
    ctx->tracker_initialized = 0;
    if (ctx->allocations_lock) dsLockDestroyContext(ctx->allocations_lock);
    if (ctx->allocations) {
	if ((check)&&(!dsHashTableIsEmpty(ctx->allocations))) {
	    dsDebugServer(DS_DEBUG_ERROR, "dsMalloc %p: DestroyContext. Error: Allocated memory remaining", ctx);
	}
	
	dsHashTableDestroy(ctx->allocations);
    }
#endif /* DS_TRACK_MEMORY */

    if (ctx->parrent) dsFree(ctx->parrent,ctx);
    else free(ctx);
}

#ifdef DS_TRACK_MEMORY
static void *dsMallocInternal(DSMallocContext ctx, size_t size) {
#else /* DS_TRACK_MEMORY */
void *dsMalloc(DSMallocContext ctx, size_t size) {
#endif /* DS_TRACK_MEMORY */

#ifdef _WIN32
    if (!ProcessHeapHandle) ProcessHeapHandle = GetProcessHeap();
#endif /* _WIN32 */

    ds_assert_null(size>0);

/*    if (ctx) {
    } else */

#ifdef _WIN32
    return HeapAlloc(ProcessHeapHandle, 0, size);
#else /* _WIN32 */
    return malloc(size);
#endif /* _WIN32 */
}

#ifdef DS_TRACK_MEMORY
static void *dsReallocInternal(DSMallocContext ctx, void *ptr, size_t size) {
#else /* DS_TRACK_MEMORY */
void *dsRealloc(DSMallocContext ctx, void *ptr, size_t size) {
#endif /* DS_TRACK_MEMORY */
#ifdef _WIN32
    if (!ProcessHeapHandle) ProcessHeapHandle = GetProcessHeap();
#endif /* _WIN32 */

/*  if (ctx) {
    } else */

#ifdef _WIN32
    if (size) {
	if (ptr) return HeapReAlloc(ProcessHeapHandle, 0, ptr, size);
	else return HeapAlloc(ProcessHeapHandle, 0, size);
    } else if (ptr) HeapFree(ProcessHeapHandle, 0, ptr);
#else /* _WIN32 */
    return realloc(ptr,size);
#endif /* _WIN32 */
    return NULL;
}

#ifdef DS_TRACK_MEMORY
static void *dsCallocInternal(DSMallocContext ctx, size_t nmemb, size_t size) {
#else /* DS_TRACK_MEMORY */
void *dsCalloc(DSMallocContext ctx, size_t nmemb, size_t size) {
#endif /* DS_TRACK_MEMORY */
#ifdef _WIN32
    if (!ProcessHeapHandle) ProcessHeapHandle = GetProcessHeap();
#endif /* _WIN32 */

/*    if (ctx) {
    } else */

#ifdef _WIN32
    return HeapAlloc(ProcessHeapHandle, HEAP_ZERO_MEMORY, nmemb*size);
#else /* _WIN32 */
    return calloc(nmemb,size);
#endif /* _WIN32 */
}

#ifdef DS_TRACK_MEMORY
static void dsFreeInternal(DSMallocContext ctx, void *ptr) {
#else /* DS_TRACK_MEMORY */
void dsFree(DSMallocContext ctx, void *ptr) {
#endif /* DS_TRACK_MEMORY */
#ifdef _WIN32
    if (!ProcessHeapHandle) ProcessHeapHandle = GetProcessHeap();
#endif /* _WIN32 */

    ds_assert_return(ptr);

/*    if (ctx) {
    } else */

#ifdef _WIN32
    HeapFree(ProcessHeapHandle, 0, ptr);
#else /* _WIN32 */
    free(ptr);
#endif /* _WIN32 */
}


#ifdef DS_TRACK_MEMORY
void *dsMalloc(DSMallocContext ctx, size_t size) {
    void *res;
    int err = 0;
    void *hash;

    res = dsMallocInternal(ctx, size);

    if ((res)&&(ctx?ctx->tracker_initialized:tracker_initialized)) {
	hash = dsMallocInternal(ctx, sizeof(void*));
	if (hash) {
	    memcpy(hash, &res, sizeof(void*));
	    err = dsLockWrite(ctx?ctx->allocations_lock:allocations_lock);
	} else err = DS_ERR_MEM;
	
	if (!err) {
	    if (ctx) ctx->tracker_initialized = 0;
	    else tracker_initialized = 0;
	    
	    err = dsHashTableInsert(ctx?ctx->allocations:allocations, hash);

	    if (ctx) ctx->tracker_initialized = 1;
	    else tracker_initialized = 1;

	    dsLockReturn(ctx?ctx->allocations_lock:allocations_lock);
	}
	
	if (err) {
	    if (hash) dsFreeInternal(ctx, hash);
	    dsFreeInternal(ctx, res);
	    dsDebugServer(DS_DEBUG_WARNING, "dsMalloc %p: Malloc. Error: Memory chunk registration failed. Freeing", ctx);
	    return NULL;
	}
    }
    
    return res;
}

void *dsRealloc(DSMallocContext ctx, void *ptr, size_t size) {
    void *res;
    int err = 0, err2 = 0;
    void *hash, *handle;
    
    if (!ptr) return dsMalloc(ctx, size);
    if (!size) {
	dsFree(ctx, ptr);
	return;
    }

    if ((res)&&(ctx?ctx->tracker_initialized:tracker_initialized)) {
	err = dsLockWrite(ctx?ctx->allocations_lock:allocations_lock);

	if (!err) {
	    if (ctx) ctx->tracker_initialized = 0;
	    else tracker_initialized = 0;

	    handle = dsHashTableFind(ctx?ctx->allocations:allocations, (char*)&ptr);
	    if (!handle) err = DS_ERR_NOTFOUND;

	    if (ctx) ctx->tracker_initialized = 1;
	    else tracker_initialized = 1;

	    dsLockReturn(ctx?ctx->allocations_lock:allocations_lock);
	}
	
	if (err) {
	    if (err == DS_ERR_NOTFOUND) 	
		dsDebugServer(DS_DEBUG_ERROR, "dsRealloc %p: Free. Error: Invalid pointer", ctx);
	    else 
		dsDebugServer(DS_DEBUG_ERROR, "dsRealloc %p: Free. Error: Request validation failed", ctx);
	    return;
	}
    }
    
    res = dsRealloc(ctx, ptr, size);

    if ((res != ptr)&&(ctx?ctx->tracker_initialized:tracker_initialized)) {

	hash = dsMallocInternal(ctx, sizeof(void*));
	if (hash) {
	    memcpy(hash, &res, sizeof(void*));
	    err = dsLockWrite(ctx?ctx->allocations_lock:allocations_lock);
	} else err = DS_ERR_MEM;
	
	if (!err) {
	    if (ctx) ctx->tracker_initialized = 0;
	    else tracker_initialized = 0;

	    err2 = dsHashTableRemove(ctx?ctx->allocations:allocations, handle);
	    err = dsHashTableInsert(ctx?ctx->allocations:allocations, hash);

	    if (ctx) ctx->tracker_initialized = 1;
	    else tracker_initialized = 1;

	    dsLockReturn(ctx?ctx->allocations_lock:allocations_lock);
	}
	
	if ((err)||(err2)) {
	    if ((err)&&(hash)) dsFreeInternal(ctx, hash);
	    if (err2) dsDebugServer(DS_DEBUG_CRITICAL, "dsMalloc %p: Realloc. Error: Memory chunk deregistration failed", ctx);
	    if (err) dsDebugServer(DS_DEBUG_CRITICAL, "dsMalloc %p: Realloc. Error: Memory chunk registration failed", ctx);
	}
    }
    
    return res;

    
    
}

void *dsCalloc(DSMallocContext ctx, size_t nmemb, size_t size) {
    void *res;
    int err = 0;
    void *hash;
    
    res = dsCallocInternal(ctx, nmemb, size);

    if ((res)&&(ctx?ctx->tracker_initialized:tracker_initialized)) {
	hash = dsMallocInternal(ctx, sizeof(void*));
	if (hash) {
	    memcpy(hash, &res, sizeof(void*));
	    err = dsLockWrite(ctx?ctx->allocations_lock:allocations_lock);
	} else err = DS_ERR_MEM;
	
	if (!err) {
	    if (ctx) ctx->tracker_initialized = 0;
	    else tracker_initialized = 0;

	    err = dsHashTableInsert(ctx?ctx->allocations:allocations, hash);

	    if (ctx) ctx->tracker_initialized = 1;
	    else tracker_initialized = 1;

	    dsLockReturn(ctx?ctx->allocations_lock:allocations_lock);
	}
	
	if (err) {
	    if (hash) dsFreeInternal(ctx, hash);
	    dsFreeInternal(ctx, res);
	    dsDebugServer(DS_DEBUG_WARNING, "dsMalloc %p: Calloc. Error: Memory chunk registration failed. Freeing", ctx);
	    return NULL;
	}
    }
    
    return res;
}


void dsFree(DSMallocContext ctx, void *ptr) {
    int err = 0;
    void *handle;
    

    if ((ptr)&&(ctx?ctx->tracker_initialized:tracker_initialized)) {
	err = dsLockWrite(ctx?ctx->allocations_lock:allocations_lock);

	if (!err) {
	    if (ctx) ctx->tracker_initialized = 0;
	    else tracker_initialized = 0;

	    handle = dsHashTableFind(ctx?ctx->allocations:allocations, (char*)&ptr);
	    if (handle) err = dsHashTableRemove(ctx?ctx->allocations:allocations, handle);
	    else err = DS_ERR_NOTFOUND;

	    if (ctx) ctx->tracker_initialized = 1;
	    else tracker_initialized = 1;

	    dsLockReturn(ctx?ctx->allocations_lock:allocations_lock);
	}
	
	if (err) {
	    if (err == DS_ERR_NOTFOUND) 	
		dsDebugServer(DS_DEBUG_ERROR, "dsMalloc %p: Free. Error: Invalid pointer (double free?)", ctx);
	    else 
		dsDebugServer(DS_DEBUG_ERROR, "dsMalloc %p: Free. Error: Request validation failed", ctx);
	    return;
	}
    }

    dsFreeInternal(ctx, ptr);
}

#endif /* DS_TRACK_MEMORY */


void *dsMallocZero(DSMallocContext ctx, size_t size) {
    void *buf;
    
    buf = dsMalloc(ctx, size);
    if (buf) memset(buf, 0, size);
    return buf;
}

int dsExchangeString(DSMallocContext mctx, char **dst, const char *src) {
    char *tmp;
    if (!src) {
	if (*dst) {
	    dsFree(mctx, *dst);
	    *dst = NULL;
	}
	return 0;
    }
    tmp = (char*)dsMalloc(mctx, (strlen(src)+1)*sizeof(char));
    if (!tmp) return -1;
    strcpy(tmp,src);
    if (*dst) dsFree(mctx, *dst);
    *dst=tmp;
    return 0;
}

int dsSetString(DSMallocContext mctx, char **dst, const char *src) {
    ds_assert_param(dst);
    *dst = NULL;

    return dsExchangeString(mctx, dst, src);
}

int dsExchangeStringChars(DSMallocContext mctx, char **dst, const char *src, unsigned chars) {
    char *tmp;
    if (!src) {
	if (*dst) {
	    dsFree(mctx, *dst);
	    *dst = NULL;
	}
	return 0;
    }
    tmp = (char*)dsMalloc(mctx, (chars+1)*sizeof(char));
    if (!tmp) return -1;
    strncpy(tmp,src,chars);tmp[chars]=0;
    if (*dst) dsFree(mctx, *dst);
    *dst=tmp;
    return 0;
}

int dsSetStringChars(DSMallocContext mctx, char **dst, const char *src, unsigned chars) {
    ds_assert_param(dst);
    *dst = NULL;

    return dsExchangeStringChars(mctx, dst, src, chars);
}


char *dsCreateFullName(DSMallocContext mctx, const char *path, const char *filename, const char *extension) {
    size_t size = 3;
    char *fullname;
    
    if (path) size += strlen(path);
    if (filename) size += strlen(filename);
    if (extension) {
	if (!filename) return NULL;
	size += strlen(extension);
    }
    
    fullname = (char*)dsMalloc(mctx, size*sizeof(char));
    if (!fullname) return fullname;

    if ((path)&&(filename)) {
	if (extension) sprintf(fullname,"%s%c%s%s",path,DS_DIRECTORY_SEPARATOR,filename,extension);
	else sprintf(fullname,"%s%c%s",path,DS_DIRECTORY_SEPARATOR,filename);
    } else if (filename) {
	if (extension) {
/*	    if (filename[0]!=DS_DIRECTORY_SEPARATOR) sprintf(fullname,"%c%s%s",DS_DIRECTORY_SEPARATOR,filename,extension);
	    else */ sprintf(fullname, "%s%s",filename, extension);
	} else {
/*	    if (filename[0]!=DS_DIRECTORY_SEPARATOR) sprintf(fullname,"%c%s",DS_DIRECTORY_SEPARATOR,filename);
	    else */ strcpy(fullname,filename);
	}
    } else if (path) {
	if (!path[0]) sprintf(fullname,"%c",DS_DIRECTORY_SEPARATOR);
	else strcpy(fullname,path);
    } else {
	sprintf(fullname,"%c",DS_DIRECTORY_SEPARATOR);
    }
    return fullname;    
}

int dsSetFullName(char *fullname, size_t buffsize, const char *path, const char *filename, const char *extension) {
    size_t size = 3;
    
    if (path) size += strlen(path);
    if (filename) size += strlen(filename);
    if (extension) {
	if (!filename) return DS_ERR_INVALID_ARGUMENT;
	size += strlen(extension);
    }

    if (size > buffsize) return DS_ERR_TOOLONG;
    
    if ((path)&&(filename)) {
	if (extension) sprintf(fullname,"%s%c%s%s",path,DS_DIRECTORY_SEPARATOR,filename,extension);
	else sprintf(fullname,"%s%c%s",path,DS_DIRECTORY_SEPARATOR,filename);
    } else if (filename) {
	if (extension) {
	    if (filename[0]!=DS_DIRECTORY_SEPARATOR) sprintf(fullname,"%c%s%s",DS_DIRECTORY_SEPARATOR,filename,extension);
	    else sprintf(fullname, "%s%s",filename, extension);
	} else {
	    if (filename[0]!=DS_DIRECTORY_SEPARATOR) sprintf(fullname,"%c%s",DS_DIRECTORY_SEPARATOR,filename);
	    else strcpy(fullname,filename);
	}
    } else if (path) {
	if (!path[0]) sprintf(fullname,"%c",DS_DIRECTORY_SEPARATOR);
	else strcpy(fullname,path);
    } else {
	sprintf(fullname,"%c",DS_DIRECTORY_SEPARATOR);
    }
    return 0;    
}
