#include <stdio.h>
#include <stdlib.h>

#include <dslib.h>

#if defined(HAVE_DLOPEN)
# ifdef HAVE_DLFCN_H
# include <dlfcn.h>
# endif
#elif defined(_WIN32)
# include <windows.h>
# define dlopen(name,attr) (void*)LoadLibrary(name)
# define dlclose(handle)  (FreeLibrary((HMODULE)handle) == 0)
# define dlsym(handle,sym) (void*)GetProcAddress((HMODULE)handle, sym)
#else
# error "Missing plugin support!"
#endif


#include <dsassert.h>

#include <dsdebug.h>

#include <dsdirectory.h>
#include <dsthread.h>
#include <dsplugin.h>

void dsPluginZeroContext(DSPluginContext ctx) {
    ds_assert_return(ctx);
    dsThreadZeroContext((DSThreadContext)ctx);
    ctx->fullname=NULL;
    ctx->next=NULL;
    ctx->modification_time=0;
    ctx->mark=0;
}

DSPluginContext dsPluginCreateContext(DSMallocContext mctx) {
    DSPluginContext ctx;

    ctx = (DSPluginContext)dsMalloc(mctx, sizeof(struct DSPluginContextT));
    dsDebugServer(DS_DEBUG_NOTICE, "dsPlugin %p: CreateContext", ctx);
    if (ctx) {
	ctx->malloc = mctx;
	ctx->thread.base.cond.malloc = mctx;
	dsPluginZeroContext(ctx);
    }
    return ctx;
}

static int dsPluginWaitReload(DSPluginContext ctx) {
    ds_assert_param(ctx);

    dsThreadSetStatusBits((DSBaseThreadContext)ctx, DS_PLUGIN_ERROR_FLAG);
    dsThreadWaitStatusSomeBitsOn((DSBaseThreadContext)ctx, DS_THREAD_TERMINATE_FLAG|DS_PLUGIN_RELOAD_FLAG);
    if (dsThreadCheckStatusBitsOff((DSBaseThreadContext)ctx, DS_THREAD_TERMINATE_FLAG)) return DS_ERR_PLUGIN;
    return 0;
}

static int dsPluginStartCallback(DSPluginContext ctx) {
    int err = 0;
    char stmp[DS_BASE_THREAD_ID_SIZE];
    void *handle;
    DSPluginInfoContext (*info_function)();
    DSPluginInfoContext plugin_info;
    DSPluginStepFunction step;
    void *pctx;
    
    ds_assert_complexctx(ctx, ctx->fullname);
    
    dsDebugServer(DS_DEBUG_NOTICE,"dsPlugin %p: Thread: %s",ctx, dsThreadGetID(stmp));

reload:
    dsDebugServer(DS_DEBUG_INFO,"dsPlugin %p: Reload: %s",ctx, ctx->fullname);
    
    ctx->modification_time = dsDirectoryGetFileModificationTime(ctx->fullname);
    
    handle = dlopen(ctx->fullname, RTLD_LAZY);
    if (!handle) {
	dsDebugServerSystemError(DS_DEBUG_ERROR, 0, "dsPlugin %p: DLOpen(%s) failed: ", ctx, ctx->fullname);
	if (dsPluginWaitReload(ctx)) goto reload;
	dsDebugServer(DS_DEBUG_DEBUG,"dsPlugin %p: Terminating", ctx);
	return DS_ERR_PLUGIN_LOAD;
    }
    
    dsThreadResetStatusBits((DSBaseThreadContext)ctx,DS_PLUGIN_RELOAD_FLAG);
    
    info_function = dlsym(handle,"PluginInfo");
    if (!info_function) {
#ifndef  DS_STATIC_BUILD
	dlclose(handle);
#endif /* DS_STATIC_BUILD */
	dsDebugServerSystemError(DS_DEBUG_ERROR, 0, "dsPlugin %p: DLSym failed: ", ctx);
	if (dsPluginWaitReload(ctx)) goto reload;
	dsDebugServer(DS_DEBUG_DEBUG,"dsPlugin %p: Terminating", ctx);
	return DS_ERR_PLUGIN_INFOFUNCTION;
    }

    plugin_info = info_function();
    if (!plugin_info) {
#ifndef  DS_STATIC_BUILD
	dlclose(handle);
#endif /* DS_STATIC_BUILD */
	dsDebugServerSystemError(DS_DEBUG_ERROR, 0, "dsPlugin %p: Missing InfoFunction: ", ctx);
	if (dsPluginWaitReload(ctx)) goto reload;
	dsDebugServer(DS_DEBUG_DEBUG,"dsPlugin %p: Terminating", ctx);
	return DS_ERR_PLUGIN_INVALIDINFO;
    }
    
    if (plugin_info->create) pctx = plugin_info->create();
    else pctx = NULL;
    if (plugin_info->init) {
	err = plugin_info->init(pctx,ctx,ctx->plugin_attribute);
	if (err) {
	    if (plugin_info->destroy) plugin_info->destroy(pctx);
#ifndef  DS_STATIC_BUILD
	    dlclose(handle);
#endif /* DS_STATIC_BUILD */
	    if (dsPluginWaitReload(ctx)) goto reload;
	    dsDebugServer(DS_DEBUG_DEBUG,"dsPlugin %p: Terminating", ctx);
	    return err;
	}
    }
    
    if (plugin_info->step) {
	step=plugin_info->step;
	while(dsThreadCheckStatusBitsOff((DSBaseThreadContext)ctx, DS_THREAD_TERMINATE_FLAG|DS_PLUGIN_RELOAD_FLAG)) {
	    dsThreadWaitStatusSomeBitsOn((DSBaseThreadContext)ctx, DS_THREAD_TERMINATE_FLAG|DS_PLUGIN_RUN_FLAG|DS_PLUGIN_RELOAD_FLAG);
	    while (dsThreadCheckStatusBits((DSBaseThreadContext)ctx,DS_PLUGIN_RUN_FLAG,DS_THREAD_TERMINATE_FLAG|DS_PLUGIN_RELOAD_FLAG)) {
		dsDebugServer(DS_DEBUG_DEBUG,"dsPlugin %p: Step",ctx);
		err=step(pctx);
		dsDebugServer(err?DS_DEBUG_WARNING:DS_DEBUG_DEBUG,"dsPlugin %p: Step (%lx)",ctx,err);
		if (err) dsPluginWaitReload(ctx);
	    }
	}
    } else {
	dsThreadWaitStatusBitsOn((DSBaseThreadContext)ctx, DS_THREAD_TERMINATE_FLAG);
    }

    if (plugin_info->destroy) plugin_info->destroy(pctx);
#ifndef  DS_STATIC_BUILD
    dlclose(handle);
#endif /* DS_STATIC_BUILD */

    if (dsThreadCheckStatusBits((DSBaseThreadContext)ctx,DS_PLUGIN_RELOAD_FLAG,DS_THREAD_TERMINATE_FLAG)) goto reload;
    
    dsDebugServer(DS_DEBUG_DEBUG,"dsPlugin %p: Terminating", ctx);
    return err;
}

int dsPluginInitContext(DSPluginContext ctx, const char *filename, void *attr) {
    int err;

    ds_assert_param(ctx);
    ds_assert_param(filename);

    dsPluginZeroContext(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "dsPlugin %p: InitContext", ctx);    
    
    err=dsExchangeString(ctx->malloc,&ctx->fullname,filename);
    if (err) {
	dsPluginFreeContext(ctx);
	return err;
    }
    
    err=dsThreadInitContext((DSThreadContext)ctx);
    if (err) {
	dsPluginFreeContext(ctx);
	return err;
    }
    
    ctx->plugin_attribute = attr;

    err= dsThreadRun((DSThreadContext)ctx, (DSThreadFunction)dsPluginStartCallback, (void*)ctx);
    if (err) {
	dsPluginFreeContext(ctx);
	return err;
    }

    return 0;
}

int dsPluginFreeContext(DSPluginContext ctx) {
    int err;
    dsDebugServer(DS_DEBUG_NOTICE, "dsPlugin %p: FreeContext",ctx);    

    ds_assert_param(ctx);

    if (ctx->fullname) dsExchangeString(ctx->malloc,&ctx->fullname,NULL);
    err=dsThreadFreeContext((DSThreadContext)ctx);
    dsPluginZeroContext(ctx);
    return err;
}

int dsPluginDestroyContext(DSPluginContext ctx) {
    int err;
    dsDebugServer(DS_DEBUG_NOTICE, "dsPlugin %p: DestroyContext",ctx);    

    ds_assert_param(ctx);

    err=dsPluginFreeContext(ctx);
    dsFree(ctx->malloc, ctx);
    return err;
}


void dsPluginRun(DSPluginContext ctx) {
    dsThreadSetStatusBits((DSBaseThreadContext)ctx,DS_PLUGIN_RUN_FLAG);
}

void dsPluginStop(DSPluginContext ctx) {
    dsThreadResetStatusBits((DSBaseThreadContext)ctx,DS_PLUGIN_RUN_FLAG);
}

void dsPluginTerminate(DSPluginContext ctx) {
    dsThreadSetStatusBits((DSBaseThreadContext)ctx,DS_THREAD_TERMINATE_FLAG);
}

void dsPluginReload(DSPluginContext ctx) {
    dsThreadSetStatusBits((DSBaseThreadContext)ctx,DS_PLUGIN_RELOAD_FLAG);
}

int dsPluginContinue(DSPluginContext ctx) {
    return dsThreadCheckStatusBitsOff((DSBaseThreadContext)ctx,DS_PLUGIN_RELOAD_FLAG|DS_THREAD_TERMINATE_FLAG|DS_PLUGIN_BREAK_FLAG);
}

void dsPluginMark(DSPluginContext ctx) {
    ds_assert_return(ctx);
    ctx->mark = 1;
}

void dsPluginUnMark(DSPluginContext ctx) {
    ds_assert_return(ctx);
    ctx->mark = 0;
}

int dsPluginMarked(DSPluginContext ctx) {
    ds_assert_param(ctx);
    return ctx->mark;
}
