#include <stdio.h>
#include <stdlib.h>
#include <dsassert.h>

#include <dslib.h>

#include <dsdebug.h>
#include <dserror.h>

#include <dsdirectory.h>
#include <dstools.h>

#include <dspluginset.h>


DSPluginSetContext dsPluginSetCreateContext(DSMallocContext mctx) {
    DSPluginSetContext ctx;

    ctx = (DSPluginSetContext)dsMalloc(mctx, sizeof(struct DSPluginSetContextT));
    dsDebugServer(DS_DEBUG_NOTICE, "dsPluginSet %p: CreateContext", ctx);    
    if (ctx) {
	ctx->malloc = mctx;
	dsPluginSetZeroContext(ctx);
    }
    return ctx;
}

int dsPluginSetInitContext(DSPluginSetContext ctx, const char *path, void *attr) {
    ds_assert_param(ctx);
    ds_assert_param(path);

    dsDebugServer(DS_DEBUG_NOTICE, "dsPluginSet %p: InitContext",ctx);    

    dsPluginSetZeroContext(ctx);
    dsExchangeString(ctx->malloc,&ctx->path,path);
    ctx->plugin_attribute = attr;
    return dsPluginSetRescanDir(ctx);
}

void dsPluginSetZeroContext(DSPluginSetContext ctx) {
    ctx->path = NULL;
    ctx->plugin = NULL;
    ctx->run = 0;
}

void dsPluginSetFreeContext(DSPluginSetContext ctx) {
    int err;
    DSPluginContext pctx, tctx;
    ds_assert_return(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "dsPluginSet %p: FreeContext", ctx);    

    if (ctx->path) dsExchangeString(ctx->malloc,&ctx->path, NULL);

    dsPluginSetTerminate(ctx);
    for (pctx=ctx->plugin;pctx;pctx=tctx) {
	tctx = pctx->next;
	err=dsPluginDestroyContext(pctx);
	dsDebugServer(err?DS_DEBUG_ERROR:DS_DEBUG_INFO,"dsPluginSet %p: opcPlugin %p: Terminating: %s",ctx,pctx,dsError(err,NULL));
    }

    dsPluginSetZeroContext(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "dsPluginSet %p: FreeContext Ok",ctx);    
}

void dsPluginSetDestroyContext(DSPluginSetContext ctx) {
    ds_assert_return(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "dsPluginSet %p: DestroyContext", ctx);    
    
    dsPluginSetFreeContext(ctx);
    dsFree(ctx->malloc, ctx);
}


int dsPluginSetRescanDir(DSPluginSetContext ctx) {
    int err = 0;
    DSPluginContext pctx,last;
    DSDirectoryContextS dir;
    DSFileContext file;
    const char *filename;

    ds_assert_param(ctx);
    ds_assert_param(ctx->path);
    ds_assert_param(ctx->run>=0);

    dsDebugServer(DS_DEBUG_NOTICE, "dsPluginSet %p: RescanDir",ctx);    

    err = dsDirectoryOpen(&dir,ctx->path,"*" DS_PLUGIN_EXTENSION);
    if (err) return DS_ERR_PLUGINSET_PATH;
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next) dsPluginMark(pctx);
    
    while (file = dsDirectoryRead(&dir)) {
	filename = dsFileGetFullName(file);
	
	for (pctx=ctx->plugin;pctx;last=pctx,pctx=pctx->next)
	    if (dsPluginMarked(pctx)) {
		if (strcmp(filename,pctx->filename)==0) {
		    if ((pctx->modification_time)&&(dsFileGetModificationTime(file)))
			dsPluginReload(pctx);
		    dsPluginUnMark(pctx);
		    break;
		}
	    }
	
	if (!pctx) {
	    pctx = dsPluginCreateContext(ctx->malloc);
	    if (!pctx) {
		err++;
		continue;
	    }
	    
	    err = dsPluginInitContext(pctx, filename, ctx->plugin_attribute);
	    if (err) {
		dsPluginDestroyContext(pctx);
		err++;
		continue;
	    }
	    
	    if (ctx->run) dsPluginRun(pctx);
		
	    if (ctx->plugin) last->next = pctx;
	    else ctx->plugin = pctx;
	}
    }

    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (dsPluginMarked(pctx)) dsPluginTerminate(pctx);
    
    for (pctx=ctx->plugin,last=NULL;pctx;last=pctx,pctx=pctx->next)
	if (dsPluginMarked(pctx)) {
	    if (last) last->next=pctx->next;
	    else ctx->plugin=pctx->next;
	    dsPluginDestroyContext(pctx);
	}
    dsDirectoryClose(&dir);

    dsDebugServer(DS_DEBUG_NOTICE, "dsPluginSet %p: RescanDir Ok",ctx);
    return 0;
}

void dsPluginSetSetStatus(DSPluginSetContext ctx, dword bits) {
    DSPluginContext pctx;
    
    ds_assert_return(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	dsThreadSetStatus((DSBaseThreadContext)pctx,bits);
}

void dsPluginSetSetStatusBits(DSPluginSetContext ctx, dword bits) {
    DSPluginContext pctx;
    
    ds_assert_return(ctx);
    for (pctx=ctx->plugin;pctx;pctx=pctx->next) {
	dsThreadSetStatusBits((DSBaseThreadContext)pctx,bits);
    }
}

void dsPluginSetResetStatusBits(DSPluginSetContext ctx, dword bits) {
    DSPluginContext pctx;
    
    ds_assert_return(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	dsThreadResetStatusBits((DSBaseThreadContext)pctx,bits);
}

int dsPluginSetCheckStatusBitsOn(DSPluginSetContext ctx, dword bits) {
    DSPluginContext pctx;
    
    ds_assert_param(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (!dsThreadCheckStatusBitsOn((DSBaseThreadContext)pctx,bits)) return 0;
    return 1;
}

int dsPluginSetCheckStatusBitsOff(DSPluginSetContext ctx, dword bits) {
    DSPluginContext pctx;
    
    ds_assert_param(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (!dsThreadCheckStatusBitsOff((DSBaseThreadContext)pctx,bits)) return 0;
    return 1;
}

int dsPluginSetCheckStatusBits(DSPluginSetContext ctx, dword onbits, dword offbits) {
    DSPluginContext pctx;
    
    ds_assert_param(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (!dsThreadCheckStatusBits((DSBaseThreadContext)pctx,onbits,offbits)) return 0;
    return 1;
}

int dsPluginSetCheckStatusSomeBitsOn(DSPluginSetContext ctx, dword bits) {
    DSPluginContext pctx;
    
    ds_assert_param(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (!dsThreadCheckStatusSomeBitsOn((DSBaseThreadContext)pctx,bits)) return 0;
    return 1;
}

int dsPluginSetCheckStatusSomeBitsOff(DSPluginSetContext ctx, dword bits) {
    DSPluginContext pctx;
    
    ds_assert_param(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (!dsThreadCheckStatusSomeBitsOff((DSBaseThreadContext)pctx,bits)) return 0;
    return 1;
}

int dsPluginSetCheckStatusSomeBits(DSPluginSetContext ctx, dword onbits, dword offbits) {
    DSPluginContext pctx;
    
    ds_assert_param(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	if (!dsThreadCheckStatusSomeBits((DSBaseThreadContext)pctx,onbits,offbits)) return 0;
    return 1;
}

void dsPluginSetWaitStatusBitsOn(DSPluginSetContext ctx, dword bits) {
    DSPluginContext pctx;
    
    ds_assert_return(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	dsThreadWaitStatusBitsOn((DSBaseThreadContext)pctx,bits);
}

void dsPluginSetWaitStatusBitsOff(DSPluginSetContext ctx, dword bits) {
    DSPluginContext pctx;
    
    ds_assert_return(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	dsThreadWaitStatusBitsOff((DSBaseThreadContext)pctx,bits);
}

void dsPluginSetWaitStatusBits(DSPluginSetContext ctx, dword onbits, dword offbits) {
    DSPluginContext pctx;
    
    ds_assert_return(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	dsThreadWaitStatusBits((DSBaseThreadContext)pctx,onbits,offbits);
}

void dsPluginSetWaitStatusSomeBitsOn(DSPluginSetContext ctx, dword bits) {
    DSPluginContext pctx;
    
    ds_assert_return(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	dsThreadWaitStatusSomeBitsOn((DSBaseThreadContext)pctx,bits);
}

void dsPluginSetWaitStatusSomeBitsOff(DSPluginSetContext ctx, dword bits) {
    DSPluginContext pctx;
    
    ds_assert_return(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	dsThreadWaitStatusSomeBitsOff((DSBaseThreadContext)pctx,bits);
}

void dsPluginSetWaitStatusSomeBits(DSPluginSetContext ctx, dword onbits, dword offbits) {
    DSPluginContext pctx;
    
    ds_assert_return(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	dsThreadWaitStatusSomeBits((DSBaseThreadContext)pctx,onbits,offbits);
}

int dsPluginSetStart(DSPluginSetContext ctx) {
    ds_assert_param(ctx);
    ctx->run = 1;
    dsPluginSetSetStatusBits(ctx,DS_PLUGIN_RUN_FLAG);
    return 0;
}

int dsPluginSetStop(DSPluginSetContext ctx) {
    ds_assert_param(ctx);
    ctx->run = 0;
    dsPluginSetSetStatusBits(ctx,DS_PLUGIN_RUN_FLAG);
    return 0;
}

int dsPluginSetBreak(DSPluginSetContext ctx) {
    ds_assert_param(ctx);
    dsPluginSetSetStatusBits(ctx,DS_PLUGIN_BREAK_FLAG);
    return 0;
}

int dsPluginSetTerminate(DSPluginSetContext ctx) {
    ds_assert_param(ctx);
    ctx->run = -1;
    dsPluginSetSetStatusBits(ctx,DS_THREAD_TERMINATE_FLAG);
    return 0;
}

void dsPluginSetWaitTerminate(DSPluginSetContext ctx) {
    DSPluginContext pctx;
    
    ds_assert_return(ctx);
    
    for (pctx=ctx->plugin;pctx;pctx=pctx->next)
	dsThreadWaitTerminate((DSThreadContext)pctx);
}

void dsPluginSetTerminateAndWait(DSPluginSetContext ctx) {
    ds_assert_return(ctx);
    dsPluginSetTerminate(ctx);
    dsPluginSetWaitTerminate(ctx);
}

int dsPluginSetIsTerminated(DSPluginSetContext ctx) {
    ds_assert_param(ctx);
    return dsPluginSetCheckStatusBitsOn(ctx,DS_THREAD_TERMINATE_FLAG);
}
