#define _DS_SIGNAL_C
#include <string.h>

#include <dslib.h>

#include <time.h>

#ifdef HAVE_SIGNAL_H
# include <signal.h>
#else /* HAVE_SIGNAL_H */
# error "singal.h is required!"
#endif /* HAVE_SIGNAL_H */

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif /* HAVE_SYS_SELECT_H */

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif /* HAVE_SYS_TIME_H */
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif /* HAVE_ERRNO_H */

#include <dsassert.h>
#include <dslib.h>
#include <dsdebug.h>

#include <dssignal.h>
#include <dsmalloc.h>
#include <dsstringlist.h>

#ifdef _WIN32
static int posix_signals[DS_POSIX_SIGNAL_NUMBER] = {
    0, SIGINT, SIGTERM, 0, 0
};
#else /* _WIN32 */
static int posix_signals[DS_POSIX_SIGNAL_NUMBER] = {
    0, SIGINT, SIGTERM, SIGTSTP, SIGCONT
};
#endif /* _WIN32 */

static DSSignalMaskS default_signal_mask;

static void signal_nop_handler(int signo) {}
void (*DS_SIGNAL_NOP_HANDLER)(int) = signal_nop_handler;

#ifndef _WIN32
# ifndef HAVE_PSELECT_SYSCALL
#  ifdef HAVE_SETJMP_H
static void signal_jmp_handler(int signo) {
    DSThreadRealHandle handle;
    handle = dsThreadFindCurrentHandle();
    if ((handle)&&(handle->jmpstatus)) {
	handle->jmpstatus = 0;
	siglongjmp(handle->jmpbuf,1);
    }
}
#  endif /* HAVE_SETJMP_H */
# endif /* HAVE_PSELECT_SYSCALL */
#endif /* _WIN32 */

#ifdef _WIN32
static HANDLE DummyHandle = INVALID_HANDLE_VALUE;
#endif /* _WIN32 */

int dsSignalInit() {
    int err = 0;
    int i;
#ifndef _WIN32
    struct sigaction act;
    sigset_t set;
#endif /* _WIN32 */

    dsDebugServer(DS_DEBUG_NOTICE, "dsSignalInit");    

    err = dsSignalMaskEmpty(&default_signal_mask);
    if (err) return err;
    err = dsSignalMaskAddSignal(&default_signal_mask, DS_SIGNAL_USR1);
    if (err) return err;
    err = dsSignalMaskAddSignal(&default_signal_mask, DS_SIGNAL_USR2);
    if (err) return err;

#ifdef _WIN32
    DummyHandle = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (!DummyHandle) return DS_ERR_WINAPI_EVENT_CREATE;
    
    for (i=0; i<DS_SIGNAL_NUMBER; i++) {
	winsignals[i] = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (!winsignals[i]) err = DS_ERR_WINAPI_EVENT_CREATE;
    }
    if (err) {
	dsSignalDestroy();
	return err;
    }
#else /* _WIN32 */
    act.sa_sigaction = NULL;
    act.sa_handler = SIG_IGN;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;

    while (sigaction(SIGPIPE,&act,NULL)==-1) {
	if (errno == EINTR) continue;
	return DS_ERR_SYSCALL_SIGACTION;
    }

# ifndef HAVE_PSELECT_SYSCALL
#  ifdef HAVE_SETJMP_H
    act.sa_handler = signal_jmp_handler;
#  else /* HAVE_SETJMP_H */
    act.sa_handler = signal_nop_handler;
#  endif /* HAVE_SETJMP_H */
# else /* HAVE_PSELECT_SYSCALL */
    act.sa_handler = signal_nop_handler;
# endif /* HAVE_PSELECT_SYSCALL */

    while (sigaction(SIGUSR1,&act,NULL)==-1) {
	if (errno == EINTR) continue;
	return DS_ERR_SYSCALL_SIGACTION;
    }
    
    while (sigaction(SIGUSR2,&act,NULL)==-1) {
	if (errno == EINTR) continue;
	return DS_ERR_SYSCALL_SIGACTION;
    }

    sigemptyset(&set);
    for (i=0;i<DS_SIGNAL_NUMBER;i++)
	sigaddset(&set,user_sigs[i]);

    while (sigprocmask(SIG_BLOCK,&set,NULL)==-1) {
	if (errno == EINTR) continue;
	return DS_ERR_SYSCALL_SIGPROCMASK_BLOCK;
    }
#endif /* _WIN32 */
    
    return 0;
}

void dsSignalDestroy() {
#ifdef _WIN32
    unsigned int i;
#endif /* _WIN32 */

    dsDebugServer(DS_DEBUG_NOTICE, "dsSignalDestroy");    

#ifdef _WIN32
    if (DummyHandle) {
	CloseHandle(DummyHandle);
	DummyHandle = NULL;
    }
    
    for (i = 0; i<DS_SIGNAL_NUMBER; i++) {
	if (winsignals[i]) {
	    CloseHandle(winsignals[i]);
	    winsignals[i] = NULL;
	}
    }
#endif /* _WIN32 */
}

int dsPosixSignalSetAction(DSPosixSignal dsig, DSPosixSignalMask block, void (*action)(int)) {
    int sig = posix_signals[dsig];
    if (!sig) return DS_ERR_NOTSUPPORTED;
    
#ifdef _WIN32
    signal(SIGINT, action?action:signal_nop_handler);
#else /* _WIN32 */
    int i, isig;
    struct sigaction act;

    act.sa_sigaction = NULL;
    act.sa_handler = action?action:signal_nop_handler;
    sigemptyset(&act.sa_mask);
    if (block) {
	for (i=0;block[i];i++) {
	    isig = posix_signals[block[i]];
	    if (isig) sigaddset(&act.sa_mask,isig);
	}
    }
    act.sa_flags = 0;
    
    
    while (sigaction(sig,&act,NULL)==-1) {
	if (errno == EINTR) continue;
	return DS_ERR_SYSCALL_SIGACTION;
    }
#endif /* _WIN32 */
    return 0;
}

int dsPosixSignalBlock(DSPosixSignal dsig) {
#ifndef _WIN32
    int sig = posix_signals[dsig];
    if (!sig) return DS_ERR_NOTSUPPORTED;

    sigset_t set;
    sigemptyset(&set);
    sigaddset(&set,sig);
    if (pthread_sigmask(SIG_BLOCK,&set,NULL)) return DS_ERR_SYSCALL_PTHREAD_SIGMASK_BLOCK;
#endif /* !_WIN32 */
    return 0;
}

int dsPosixSignalUnblock(DSPosixSignal dsig) {
#ifndef _WIN32
    int sig = posix_signals[dsig];
    if (!sig) return DS_ERR_NOTSUPPORTED;

    sigset_t set;
    sigemptyset(&set);
    sigaddset(&set,sig);
    if (pthread_sigmask(SIG_UNBLOCK,&set,NULL)) return DS_ERR_SYSCALL_PTHREAD_SIGMASK_UNBLOCK;
#endif /* !_WIN32 */
    return 0;
}


int dsSignalGetDefaultMask(DSSignalMask sigmask) {
    ds_assert_param(sigmask);
    
    memcpy(sigmask, &default_signal_mask, sizeof(DSSignalMaskS));
    return 0;
}


int dsSignalMaskEmpty(DSSignalMask sigmask) {
    ds_assert_param(sigmask);

#ifdef _WIN32
    sigmask->mask = 0;   
#else /* _WIN32 */
    sigfillset(&sigmask->mask);
#endif /* _WIN32 */

    return 0;
}

int dsSignalMaskFill(DSSignalMask sigmask) {
    ds_assert_param(sigmask);
    
#ifdef _WIN32
    sigmask->mask = 0xFFFFFFFF;
#else /* _WIN32 */
    sigemptyset(&sigmask->mask);
#endif /* _WIN32 */

    return 0;
}

int dsSignalMaskAddSignal(DSSignalMask sigmask, DSSignal sig) {
    ds_assert_param(sigmask);

    ds_assert_param(sig < DS_SIGNAL_NUMBER);
    
#ifdef _WIN32
    sigmask->mask|=(1<<sig);
#else /* _WIN32 */
    sigdelset(&sigmask->mask, user_sigs[sig]);
#endif /* _WIN32 */
    
    return 0;
}

int dsSignalMaskRemoveSignal(DSSignalMask sigmask, DSSignal sig) {
    ds_assert_param(sigmask);

    ds_assert_param(sig >= DS_SIGNAL_NUMBER);

#ifdef _WIN32
    sigmask->mask&=~(1<<sig);
#else /* _WIN32 */
    sigaddset(&sigmask->mask, user_sigs[sig]);
#endif /* _WIN32 */
    
    return 0;
}


void dsSignalZeroContext(DSSignalContext ctx) {
    unsigned i;
    
    ds_assert_return(ctx);
    
    ctx->lock = NULL;
    for (i=0;i<DS_SIGNAL_NUMBER;i++)
	ctx->usr[i] = NULL;
    ctx->ExitFlag = 0;
}

DSSignalContext dsSignalCreateContext(DSMallocContext mctx) {
    DSSignalContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "dsSignalCreateContext");    

    ctx = (DSSignalContext)dsMalloc(mctx, sizeof(struct DSSignalContextT));
    if (ctx) {
	ctx->malloc = mctx;
	dsSignalZeroContext(ctx);
    }
    return ctx;
}

int dsSignalInitContext(DSSignalContext ctx) {
    int err;
    unsigned i;

    dsDebugServer(DS_DEBUG_NOTICE, "dsSignalInitContext");    

    ds_assert_param(ctx);
    dsSignalZeroContext(ctx);

    ctx->lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) return DS_ERR_MEM;
    
    err = dsLockInitContext(ctx->lock);
    if (err) {
	dsSignalFreeContext(ctx);
	return err;
    }
    
    for (i=0;i<DS_SIGNAL_NUMBER;i++) {
        ctx->usr[i] = dsStringListCreate(ctx->malloc);
	if (!ctx->usr[i]) {
	    dsSignalFreeContext(ctx);
	    return DS_ERR_MEM;
	}
    }
    
    return 0;
}

void dsSignalFreeContext(DSSignalContext ctx) {
    unsigned i;
    ds_assert_return(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "dsSignalFreeContext");    
    
    if (ctx->lock) dsLockDestroyContext(ctx->lock);

    for (i=0;i<DS_SIGNAL_NUMBER;i++) {
	if (ctx->usr[i]) dsStringListDestroy(ctx->usr[i]);
    }
        
    dsSignalZeroContext(ctx);
}

void dsSignalDestroyContext(DSSignalContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "dsSignalDestroyContext");    

    ds_assert_return(ctx);
    dsSignalFreeContext(ctx);
    dsFree(ctx->malloc, ctx);
}

int dsSignalRegister(DSSignalContext ctx, DSSignal sig) {
    int err;
    DSThreadHandleS pid;
    
    ds_assert_param(ctx);
    ds_assert_param(sig<DS_SIGNAL_NUMBER);

    dsThreadGetCurrentHandle(&pid);
    
    dsLockWrite(ctx->lock);
    err = dsStringListAdd(ctx->usr[sig],&pid,sizeof(DSThreadHandleS));
    dsLockReturn(ctx->lock);
    
    return err;
}

int dsSignalCancel(DSSignalContext ctx, DSSignal sig) {
    int err;
    DSThreadHandleS pid;
    
    ds_assert_param(ctx);
    ds_assert_param(sig<DS_SIGNAL_NUMBER);
    
    dsThreadGetCurrentHandle(&pid);
    
    dsLockWrite(ctx->lock);
    err=dsStringListRemove(ctx->usr[sig],&pid,sizeof(DSThreadHandleS));
    dsLockReturn(ctx->lock);

    return err;
}

int dsSignalSend(DSSignalContext ctx, DSSignal sig) {
    DSStringList sl;
    DSThreadHandle pid;
    
    ds_assert_param(ctx);
    ds_assert_param(sig<DS_SIGNAL_NUMBER);
    
    dsLockWrite(ctx->lock);
    for (sl=ctx->usr[sig];sl;sl=sl->next) {
        if (sl->item) {
    	    pid = (DSThreadHandle)(sl->item);
	    dsThreadSignal(pid, sig);
	}
    }
    dsLockReturn(ctx->lock);
    
    return 0;
}

int dsSignalSendTo(DSThreadHandle thread, DSSignal sig) {
    return dsThreadSignal(thread, sig);
}

int dsSignalTerminate(DSSignalContext ctx) {
    ds_assert_param(ctx);
    ctx->ExitFlag=1;
    return dsSignalSend(ctx,DS_SIGNAL_NOTIFY);
}

int dsSignalIsTerminated(DSSignalContext ctx) {
    ds_assert_param(ctx);
    return ctx->ExitFlag;
}


#ifndef _WIN32
# ifdef HAVE_PSELECT_SYSCALL
# define ds_pselect pselect
# else  /* HAVE_PSELECT_SYSCALL */
#  ifdef HAVE_SETJMP_H
static inline int ds_pselect(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, const struct timespec *timeout, const sigset_t *sigmask) {
    int err;
    DSThreadRealHandle handle;
#   ifndef HAVE_PSELECT
    struct timeval tvtimeout;
    
    tvtimeout.tv_sec = timeout->tv_sec;
    tvtimeout.tv_usec = timeout->tv_nsec / 1000;
#   endif /* !HAVE_PSELECT */

    handle = dsThreadFindCurrentHandle();
    if (!handle) return DS_ERR_NOTINITIALIZED;


    handle->jmpstatus = 1;
    if (!sigsetjmp(handle->jmpbuf,1)) {
#   ifdef HAVE_PTHREAD_H
	pthread_sigmask(SIG_UNBLOCK,sigmask,NULL);
#   else  /* HAVE_PTHREAD_H */
#    error "Threading library support is missing!"
#   endif /* HAVE_PTHREAD_H */
#   ifdef HAVE_PSELECT
	err = pselect(n,readfds,writefds,exceptfds,timeout,NULL);
#   else /* HAVE_PSELECT */
	err = select(n,readfds,writefds,exceptfds,&tvtimeout);
#   endif /* HAVE_PSELECT */
#   ifdef HAVE_PTHREAD_H
	pthread_sigmask(SIG_BLOCK,sigmask,NULL);
#   else  /* HAVE_PTHREAD_H */
#    error "Threading library support is missing!"
#   endif /* HAVE_PTHREAD_H */
    }
    handle->jmpstatus = 0;
    
    return err;
}
#  else  /* HAVE_SETJMP_H */
#   ifdef HAVE_PSELECT
#    define ds_pselect pselect
#   else /* HAVE_PSELECT */
static inline int ds_pselect(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, const struct timespec *timeout, const sigset_t *sigmask) {
    DSThreadRealHandle handle;
#   ifndef HAVE_PSELECT
    struct timeval tvtimeout;

    tvtimeout.tv_sec = timeout->tv_sec;
    tvtimeout.tv_usec = timeout->tv_nsec / 1000;
#   endif /* HAVE_PSELECT */

    handle = dsThreadFindCurrentHandle();
    if (!handle) return DS_ERR_NOTINITIALIZED;

    handle->jmpstatus = rand() + 1;
    if (!handle->jmpstatus) handle->jmpstatus = 1;
#   ifdef HAVE_PSELECT
    err = pselect(n,readfds,writefds,exceptfds,timeout,sigmask);
#   else /* HAVE_PSELECT */
#    ifdef HAVE_PTHREAD_H
    pthread_sigmask(SIG_UNBLOCK,sigmask,NULL);
#    else  /* HAVE_PTHREAD_H */
#     error "Threading library support is missing!"
#    endif /* HAVE_PTHREAD_H */
    err = select(n,readfds,writefds,exceptfds,&tvtimeout);
#    ifdef HAVE_PTHREAD_H
    pthread_sigmask(SIG_BLOCK,sigmask,NULL);
#    else  /* HAVE_PTHREAD_H */
#     error "Threading library support is missing!"
#    endif /* HAVE_PTHREAD_H */
#   endif /* HAVE_PSELECT */
    handle->jmpstatus = 0;

    return err;
}
#   endif /* HAVE_PSELECT */ 
#  endif /* HAVE_SETJMP_H */
# endif /* HAVE_PSELECT_SYSCALL */
#endif /* !_WIN32 */



#ifdef _WIN32

int dsSignalWinWait(DSSignalMask sigmask, DSSoftSignalList softsig, HANDLE hEvent, const DSDuration timeout) {
    int err = 0;
    DWORD obj;
    unsigned long i, j;

    int softsigs;
    
    DSThreadRealHandle handle;
    HANDLE *hEvents;
    HANDLE hEventsArray[1 + 2*DS_SIGNAL_NUMBER + DS_SOFTSIG_NUMBER];
    HANDLE LocalDummyHandle;

    ds_assert_param((hEvent != INVALID_HANDLE_VALUE)||(sigmask)||(softsig)||(timeout));

    if (softsig) {
	for (softsigs=0;softsig[softsigs];softsigs++);
	if (softsigs > DS_SOFTSIG_NUMBER) {
	    hEvents = (HANDLE*)dsMalloca((1 + 2*DS_SIGNAL_NUMBER + softsigs)*sizeof(HANDLE));
	    if (!hEvents) return DS_ERR_MEM;
	} else hEvents = &hEventsArray;
    } else {
	softsigs = 0;
	hEvents = &hEventsArray;
    }
    

    if (DummyHandle == INVALID_HANDLE_VALUE) {
	LocalDummyHandle = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (LocalDummyHandle == INVALID_HANDLE_VALUE) return DS_ERR_WINAPI_EVENT_CREATE;
    } else LocalDummyHandle = DummyHandle;
    
    if ((!sigmask)&&(!softsigs)) {
	if (hEvent == INVALID_HANDLE_VALUE) {
	    WaitForSingleObject(LocalDummyHandle, timeout?dsDurationGetMilliseconds(timeout):INFINITE);
	} else {
	    if (WaitForSingleObject(hEvent, timeout?dsDurationGetMilliseconds(timeout):INFINITE) != WAIT_OBJECT_0) err = DS_ERR_TIMEOUT;
	}
    } else {
	if (hEvent != INVALID_HANDLE_VALUE) hEvents[0] = hEvent;
	else hEvents[0] = LocalDummyHandle;

	if (sigmask) {
	    handle = dsThreadFindCurrentHandle();

	    for (i=0,j=1;i<DS_SIGNAL_NUMBER; i++,j<<=1) {
		if ((sigmask)&&(sigmask->mask&j)) {
		    hEvents[i+1] = winsignals[i];
		    if (handle == INVALID_HANDLE_VALUE) hEvents[i+DS_SIGNAL_NUMBER+1] = LocalDummyHandle;
		    else hEvents[i+DS_SIGNAL_NUMBER+1] = handle->winsignals[i];
		} else {
		    hEvents[i+1] = LocalDummyHandle;
		    hEvents[i+DS_SIGNAL_NUMBER+1] = LocalDummyHandle;
		}
	    }
	    
	    i = 1 + 2 * DS_SIGNAL_NUMBER;
	} else i = 1;
	
	if (softsigs) {
	    for (j = 0; j < softsigs; j++, i++) {
		hEvents[i] = softsig[j]->h;
	    }
	}
	

	obj = WaitForMultipleObjects(i, hEvents, FALSE, timeout?dsDurationGetMilliseconds(timeout):INFINITE);

	if (obj == WAIT_OBJECT_0) err = 0;
	else if ((obj - WAIT_OBJECT_0)<i) err = DS_ERR_SIGNAL_EVENT; 
	else if (hEvent == INVALID_HANDLE_VALUE) err = 0;
	else err = DS_ERR_TIMEOUT;
    }
    
    if (DummyHandle != LocalDummyHandle) CloseHandle(LocalDummyHandle);
    if (hEvents != &hEventsArray) dsFreea(hEvents);
    
    return err;	
}

int dsSignalWinWaitConnection(DSConnectionServerContext con, DSSignalMask sigmask, DSSoftSignalList softsig, DSConnectionMask *conmask, const DSDuration timeout) {
    unsigned long i,j;
    DWORD obj;
    
    DSConnectionMask mask;
    DSThreadRealHandle handle;
    HANDLE *hEvents;
    HANDLE hEventsArray[2 + 2*DS_SIGNAL_NUMBER + DS_SOFTSIG_NUMBER];
    HANDLE LocalDummyHandle;

    int softsigs;

    ds_assert_param(con);
    ds_assert_param(conmask);

    if (softsig) {
	for (softsigs=0;softsig[softsigs];softsigs++);
	if (softsigs > DS_SOFTSIG_NUMBER) {
	    hEvents = (HANDLE*)dsMalloca((2 + 2*DS_SIGNAL_NUMBER + softsigs)*sizeof(HANDLE));
	    if (!hEvents) return DS_ERR_MEM;
	} else hEvents = &hEventsArray;
    } else {
	softsigs = 0;
	hEvents = &hEventsArray;
    }



    mask = *conmask;

    if (mask&DS_CONNECTION_LOCAL) {
	hEvents[0] = con->hEventLocal;
        ds_assert_init(hEvents[0] != INVALID_HANDLE_VALUE);
    } else {
	hEvents[0] = DummyHandle;
    }
    
    if (mask&DS_CONNECTION_NETWORK) {
        hEvents[1] = con->hEventNetwork;
        ds_assert_init(hEvents[1] != INVALID_HANDLE_VALUE);
    } else {
	hEvents[1] = DummyHandle;
    }

    if (DummyHandle == INVALID_HANDLE_VALUE) {
	LocalDummyHandle = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (LocalDummyHandle == INVALID_HANDLE_VALUE) return DS_ERR_WINAPI_EVENT_CREATE;
	
	if (hEvents[0] == DummyHandle) hEvents[0] = LocalDummyHandle;
	if (hEvents[1] == DummyHandle) hEvents[1] = LocalDummyHandle;
    } else LocalDummyHandle = DummyHandle;

    if ((!sigmask)&&(!softsigs)) {
	obj = WaitForMultipleObjects(2,hEvents,FALSE,timeout?dsDurationGetMilliseconds(timeout):INFINITE);
    } else {
	if (sigmask) {
	    handle = dsThreadFindCurrentHandle();

	    for (i=0,j=1;i<DS_SIGNAL_NUMBER; i++,j<<=1) {
		if ((sigmask)&&(sigmask->mask&j)) {
		    hEvents[i+2] = winsignals[i];
		    if (handle == INVALID_HANDLE_VALUE) hEvents[i+DS_SIGNAL_NUMBER+2] = LocalDummyHandle;
		    else hEvents[i+DS_SIGNAL_NUMBER+2] = handle->winsignals[i];
		} else {
		    hEvents[i+2] = LocalDummyHandle;
		    hEvents[i+DS_SIGNAL_NUMBER+2] = LocalDummyHandle;
		}
	    }
	    i = 2 + 2 * DS_SIGNAL_NUMBER;
	} else i = 2;
	
	if (softsigs) {
	    for (j = 0; j < softsigs; j++, i++) {
		hEvents[i] = softsig[j]->h;
	    }
	}

	obj = WaitForMultipleObjects(i,hEvents,FALSE,timeout?dsDurationGetMilliseconds(timeout):INFINITE);
    }

    if (DummyHandle != LocalDummyHandle) CloseHandle(LocalDummyHandle);
    if (hEvents != &hEventsArray) dsFreea(hEvents);

    if (obj == WAIT_OBJECT_0) *conmask = DS_CONNECTION_LOCAL;
    else if (obj == (WAIT_OBJECT_0 + 1)) *conmask = DS_CONNECTION_NETWORK;
    else if (((sigmask)||(softsig))&&((obj - WAIT_OBJECT_0)<i)) return DS_ERR_SIGNAL_EVENT;
    else return DS_ERR_TIMEOUT;

    return 0;
}
#endif /* _WIN32 */

#ifdef HAVE_SOCKETS
int dsSignalWaitRead(DSConnectionContext con, DSSignalMask sigmask, DSSoftSignalList softsig, const DSDuration timeout) {
#ifdef _WIN32
    return dsSignalWinWait(sigmask, softsig, con->hEventRead, timeout);
/*
    int err = 1;

    TIMEVAL tv = {0,0};
    fd_set readfs;


    do {    
	if (!err) {
	    err = dsSignalWinWait(sigmask,con->hEventRead, timeout);
	    if (err) return err;
	}
	
	FD_ZERO(&readfs);
	FD_SET(con->winsock, &readfs);
	err = select(0, &readfs, NULL, NULL, &tv);
	if (err == SOCKET_ERROR) return DS_ERR_SYSCALL_SELECT;
    } while (!err);

    return 0;
*/
#else /* _WIN32 */
    int err;
    
    struct timespec ts;
    
    int i;
    int s;
    int maxs;
    fd_set rd;


    ds_assert_param(con);
    ds_assert_init(s = con->socket);

    FD_ZERO(&rd);
    FD_SET(s,&rd);
    
    maxs = s;
    
    if (softsig) {
	for (i = 0; softsig[i]; i++) {
	    FD_SET(softsig[i]->r, &rd);
	    if (softsig[i]->r > maxs) maxs = softsig[i]->r;
	}
    }
    

    if (timeout) {
	ts.tv_sec = timeout->s;
	ts.tv_nsec = timeout->us * 1000;
    }

    err = ds_pselect(maxs+1,&rd,NULL,NULL,timeout?&ts:NULL,sigmask?&sigmask->mask:NULL);
    if (err<=0) {
	if (!err) return DS_ERR_TIMEOUT;
	if (errno==EINTR) return DS_ERR_SIGNAL_EVENT;
	return DS_ERR_SYSCALL_PSELECT;
    }

    return FD_ISSET(s, &rd)?0:DS_ERR_SIGNAL_EVENT;
#endif /* _WIN32 */
}

int dsSignalWaitWrite(DSConnectionContext con, DSSignalMask sigmask, DSSoftSignalList softsig, const DSDuration timeout) {
#ifdef _WIN32
    int err = 1;
    int doinit = 1;

    DSDurationS pause;
    DSTimeS endtime, curtime;
    
    TIMEVAL tv = {0,0};
    fd_set writefs;

    do {    
/*
	err = dsSignalWinWait(sigmask,con->hEventWrite,timeout);
	if (err) return err;
*/
	if (!err) {
	    if (doinit) {
		dsDurationImportMilliseconds(&pause, 25);
		if (timeout) {
		    dsTimeCurrentTime(&endtime);
		    dsTimeAdd(&endtime, timeout);
		}
		doinit = 0;
	    }
	    
	    dsTimeCurrentTime(&curtime);
	    if ((timeout)&&(dsTimeCmp(&curtime, &endtime)>0)) return DS_ERR_TIMEOUT;

	    dsSignalWait(sigmask, &pause);
	}
	
	FD_ZERO(&writefs);
	FD_SET(con->winsock, &writefs);
	err = select(0, NULL, &writefs, NULL, &tv);
	if (err == SOCKET_ERROR) return DS_ERR_SYSCALL_SELECT;
    } while (!err);

    return 0;
#else /* _WIN32 */
    int err;
    
    struct timespec ts;
    
    int i;
    int s;
    int maxs;
    fd_set wr;

    ds_assert_param(con);
    ds_assert_init(s = con->socket);

    FD_ZERO(&wr);
    FD_SET(s,&wr);

    maxs = s;
    
    if (softsig) {
	for (i = 0; softsig[i]; i++) {
	    FD_SET(softsig[i]->r, &wr);
	    if (softsig[i]->r > maxs) maxs = softsig[i]->r;
	}
    }

    if (timeout) {
	ts.tv_sec = timeout->s;
	ts.tv_nsec = timeout->us * 1000;
    }

    err=ds_pselect(maxs+1,NULL,&wr,NULL,timeout?&ts:NULL,sigmask?&sigmask->mask:NULL);
    if (err<=0) {
	if (!err) return DS_ERR_TIMEOUT;
	if (errno==EINTR) return DS_ERR_SIGNAL_EVENT;
	return DS_ERR_SYSCALL_PSELECT;
    }

    return FD_ISSET(s, &wr)?0:DS_ERR_SIGNAL_EVENT;
#endif /* _WIN32 */
}

#ifndef _WIN32
int dsSignalWaitConnection(DSConnectionServerContext con, DSSignalMask sigmask, DSSoftSignalList softsig, DSConnectionMask *conmask, const DSDuration timeout) {
    int err;

    struct timespec ts;
    
    DSConnectionMask mask;
    int i;
    int maxs = 0;
    int s1, s2;
    fd_set rd;

    ds_assert_param(con);
    ds_assert_param(conmask);
    
    mask = *conmask;

    FD_ZERO(&rd);

    if (mask&DS_CONNECTION_LOCAL) {
	s1 = con->local_socket;
        FD_SET(s1, &rd);

	maxs = s1;
    }

    if (mask&DS_CONNECTION_NETWORK) {
	s2 = con->network_socket;
	FD_SET(s2, &rd);
	
	if (s2 > maxs) maxs = s2;
    }

    if (!maxs) return DS_ERR_EMPTY;

    if (softsig) {
	for (i = 0; softsig[i]; i++) {
	    FD_SET(softsig[i]->r, &rd);
	    if (softsig[i]->r > maxs) maxs = softsig[i]->r;
	}
    }


    if (timeout) {
	ts.tv_sec = timeout->s;
	ts.tv_nsec = timeout->us * 1000;
    }

    err=ds_pselect(maxs+1,&rd,NULL,NULL,timeout?&ts:NULL,sigmask?&sigmask->mask:NULL);
    if (err<=0) {
	if (!err) return DS_ERR_TIMEOUT;
	if (errno==EINTR) return DS_ERR_SIGNAL_EVENT;
	return DS_ERR_SYSCALL_PSELECT;
    }

    if ((s1)&&(FD_ISSET(s1,&rd))) {
	*conmask = DS_CONNECTION_LOCAL;
	return 0;
    } else if ((s2)&&(FD_ISSET(s2,&rd))) {
	*conmask = DS_CONNECTION_NETWORK;
	return 0;
    }

    return DS_ERR_SIGNAL_EVENT;
}
#endif /* ! _WIN32 */

int dsSignalCheckFDRead(int s) {
#ifdef _WIN32
    return DS_ERR_NOTSUPPORTED;
#else /* _WIN32 */
    int err;

    fd_set rd;
    struct timespec tm = {0,0};

    FD_ZERO(&rd);
    FD_SET(s,&rd);

    err=ds_pselect(s+1,&rd,NULL,NULL,&tm,NULL);
    if (err<=0) {
	if (!err) return DS_ERR_TIMEOUT;
	if (errno==EINTR) return DS_ERR_SIGNAL_EVENT;
	return DS_ERR_SYSCALL_PSELECT;
    }
    
    return 0;
#endif /* _WIN32 */
}

int dsSignalCheckFDWrite(int s) {
#ifdef _WIN32
    return DS_ERR_NOTSUPPORTED;
#else /* _WIN32 */
    int err;

    fd_set wr;
    struct timespec tm = {0,0};

    FD_ZERO(&wr);
    FD_SET(s,&wr);

    err=ds_pselect(s+1,NULL,&wr,NULL,&tm,NULL);
    if (err<=0) {
	if (!err) return DS_ERR_TIMEOUT;
	if (errno==EINTR) return DS_ERR_SIGNAL_EVENT;
	return DS_ERR_SYSCALL_PSELECT;
    }
#endif /* _WIN32 */
}


int dsSignalCheckRead(DSConnectionContext con) {
#ifdef _WIN32
    int err;

    TIMEVAL tv = {0,0};
    fd_set readfs;
	
    FD_ZERO(&readfs);
    FD_SET(con->winsock, &readfs);
    err = select(0, &readfs, NULL, NULL, &tv);
    if (err) {
	if (err == SOCKET_ERROR) return DS_ERR_SYSCALL_SELECT;
	return 0;
    }
    
    return DS_ERR_TIMEOUT;
#else /* _WIN32 */
    int s;
    ds_assert_param(con);
    ds_assert_init(s = con->socket);

    return dsSignalCheckFDRead(s);
#endif /* _WIN32 */
}

int dsSignalCheckWrite(DSConnectionContext con) {
#ifdef _WIN32    
    int err;

    TIMEVAL tv = {0,0};
    fd_set writefs;

	
    FD_ZERO(&writefs);
    FD_SET(con->winsock, &writefs);
    err = select(0, NULL, &writefs, NULL, &tv);
    if (err) {
	if (err == SOCKET_ERROR) return DS_ERR_SYSCALL_SELECT;
	return 0;
    }
    return DS_ERR_TIMEOUT;
#else /* _WIN32 */
    int s;
    ds_assert_param(con);
    ds_assert_init(s = con->socket);

    return dsSignalCheckFDWrite(s);
#endif  /* _WIN32 */
}

#ifndef _WIN32
int dsSignalCheckConnection(DSConnectionServerContext con, DSConnectionMask *conmask) {
    int err;
    
    DSConnectionMask mask;
    int s = 0;
    fd_set rd;
    struct timespec tm = {0,0};

    ds_assert_param(con);
    ds_assert_param(conmask);
    
    mask = *conmask;

    FD_ZERO(&rd);

    if (mask&DS_CONNECTION_LOCAL) {
	s = con->local_socket;
        FD_SET(con->local_socket,&rd);
    }
    if (mask&DS_CONNECTION_NETWORK) {
	if (con->network_socket > s) s = con->network_socket;
	FD_SET(con->network_socket,&rd);
    }

    if (!s) return DS_ERR_EMPTY;


    err=ds_pselect(s+1,&rd,NULL,NULL,&tm,NULL);
    if (err<=0) {
	if (!err) return DS_ERR_TIMEOUT;
	if (errno==EINTR) return DS_ERR_SIGNAL_EVENT;
	return DS_ERR_SYSCALL_PSELECT;
    }

    if (FD_ISSET(con->local_socket,&rd)) *conmask = DS_CONNECTION_LOCAL;
    else *conmask = DS_CONNECTION_NETWORK;

    return 0;
}
#endif /* _WIN32 */
#endif /* HAVE_SOCKETS */

int dsSignalWait(DSSignalMask sigmask, DSSoftSignalList softsig, DSDurationConst duration) {
#ifdef _WIN32
#else  /* _WIN32 */
    int err;
    struct timespec slp;
    int i, maxs = 0;
    fd_set rd;
#endif /* _WIN32 */

    ds_assert_param((sigmask)||(duration));

    if ((duration)&&(duration->s<0)) return 0;

#ifdef _WIN32
    return dsSignalWinWait(sigmask, softsig, INVALID_HANDLE_VALUE, duration);
#else  /* _WIN32 */
    if (duration) {
	slp.tv_sec = duration->s;
	slp.tv_nsec = duration->us * 1000;
    } else {
	slp.tv_sec = 0x7FFFFFFF;
	slp.tv_nsec = 0;
    }
    
    if (softsig) {
	FD_ZERO(&rd);
	for (i = 0; softsig[i]; i++) {
	    FD_SET(softsig[i]->r, &rd);
	    if (softsig[i]->r > maxs) maxs = softsig[i]->r;
	}
    }
    
    if (maxs) {
	 err = ds_pselect(maxs+1,&rd,NULL,NULL,&slp,sigmask?&sigmask->mask:NULL);
	 if (err) {
	    if ((err > 0)||(errno == EAGAIN)||(errno == EINTR)) return DS_ERR_SIGNAL_EVENT;
	    return DS_ERR_SYSCALL_PSELECT;
	 }
    } else {
	if (sigmask) err = sigtimedwait(&sigmask->mask, NULL, &slp);
	else err = nanosleep(&slp, NULL);

        if (err) {
	    if ((errno == EAGAIN)||(errno == EINTR)) return DS_ERR_SIGNAL_EVENT;
	    if (sigmask) return DS_ERR_SYSCALL_SIGTIMEDWAIT;
	    else return DS_ERR_SYSCALL_NANOSLEEP;
	}
    }
    
    return err;
#endif /* _WIN32 */
}

int dsSignalWaitUntil(DSSignalMask sigmask, DSSoftSignalList softsig, DSTimeConst time) {
    DSDurationS duration; 
    DSTimeS cur;
    
    dsTimeCurrentTime(&cur);
    dsTimeCalcDuration(&duration, &cur, time);

    if ((duration.s<0)||((duration.s == 0)&&(duration.us == 0))) return 0;
    return dsSignalWait(sigmask, softsig, &duration);
}

int dsSignalWaitRT(DSSignalMask sigmask, DSSoftSignalList softsig, DSDurationConst duration) {
#ifdef _WIN32
#else  /* _WIN32 */
    int err;
    struct timespec slp;
    unsigned long remaining;
    
    DSTimeS endwait, cur;
    DSDurationS remaining_duration;
#endif /* _WIN32 */

    ds_assert_param((sigmask)||(duration));

    if ((duration)&&(duration->s<0)) return 0;

#ifdef _WIN32
    return dsSignalWinWait(sigmask, softsig, INVALID_HANDLE_VALUE, duration);
#else  /* _WIN32 */

    if (duration) {
	if ((duration->us>=30000)||(duration->s>0)) {
	    dsTimeCurrentTime(&endwait);
	    dsTimeAdd(&endwait, duration);
	    
	    dsDurationCopy(&remaining_duration, duration);
	    dsDurationSubMilliseconds(&remaining_duration, 30);
	    
	    err = dsSignalWait(sigmask, softsig, &remaining_duration);
	    if (err) return err;

	    dsTimeCurrentTime(&cur);
	    dsTimeCalcDuration(&remaining_duration, &cur, &endwait);

	    if (!remaining_duration.s) {
		remaining = remaining_duration.us;
	    } else {
		if (remaining_duration.s < 0) return 0;
		else return DS_ERR_INTERNAL;
	    }
	} else remaining = duration->us;
/*
	if (duration->us>=30000) {
	    dsTimeCurrentTime(&endwait);
	    dsTimeAdd(&endwait, duration);
	    
	    slp.tv_sec = duration->s;
	    slp.tv_nsec = (duration->us - 30000)*1000;

	    if (sigmask) err = sigtimedwait(&sigmask->mask, NULL, &slp);
	    else err = nanosleep(&slp, NULL);
    
	    if (err) {
		if ((errno == EAGAIN)||(errno == EINTR)) return DS_ERR_SIGNAL_EVENT;
		if (sigmask) return DS_ERR_SYSCALL_SIGTIMEDWAIT;
		else return DS_ERR_SYSCALL_NANOSLEEP;
	    }

	    dsTimeCurrentTime(&cur);
	    dsTimeCalcDuration(&remaining_duration, &cur, &endwait);
	    if (!remaining_duration.s) {
		remaining = remaining_duration.us;
	    } else {
		if (remaining_duration.s < 0) return 0;
		else {
		    if (sigmask) return DS_ERR_SYSCALL_SIGTIMEDWAIT;
		    else return DS_ERR_SYSCALL_NANOSLEEP;
		}
	    }
	} else if (duration->s>0) {
	    dsTimeCurrentTime(&endwait);
	    dsTimeAdd(&endwait, duration);

	    slp.tv_sec = duration->s - 1;
	    slp.tv_nsec = (duration->us + 970000)*1000;

	    if (sigmask) err = sigtimedwait(&sigmask->mask, NULL, &slp);
	    else err = nanosleep(&slp, NULL);
    
	    if (err) {
		if ((err == EAGAIN)||(err == EINTR)) return DS_ERR_SIGNAL_EVENT;
		if (sigmask) return DS_ERR_SYSCALL_SIGTIMEDWAIT;
		else return DS_ERR_SYSCALL_NANOSLEEP;
	    }

	    dsTimeCurrentTime(&cur);
	    dsTimeCalcDuration(&remaining_duration, &cur, &endwait);
	    if (!remaining_duration.s) {
		remaining = remaining_duration.us;
	    } else {
		if (remaining_duration.s < 0) return 0;
		else {
		    if (sigmask) return DS_ERR_SYSCALL_SIGTIMEDWAIT;
		    else return DS_ERR_SYSCALL_NANOSLEEP;
		}
	    }
	} else remaining = duration->us;
*/
    }  else return dsSignalWait(sigmask, softsig, duration);
	
	/* DS: We are ignoring signals for short (<30ms) sleep. Since signal
	delivery in standard (non RT) linux have precision around 10ms.
	Special handling should added for RT linux */
	
    slp.tv_sec = 0;
    slp.tv_nsec = 1000000;
    
    while (remaining > 1000) {
//	if (sigmask) pthread_sigmask(SIG_UNBLOCK,sigmask->mask,NULL);
	nanosleep(&slp, NULL);
//	if (sigmask) pthread_sigmask(SIG_BLOCK,sigmask->mask,NULL);
	remaining -= 1000;
    }

    if (remaining>100) {
	slp.tv_nsec = remaining*1000;
//	if (sigmask) pthread_sigmask(SIG_UNBLOCK,sigmask->mask,NULL);
	nanosleep(&slp, NULL);
//	if (sigmask) pthread_sigmask(SIG_BLOCK,sigmask->mask,NULL);
    }

    return 0;

#endif /* _WIN32 */
}


int dsSignalWaitRTUntil(DSSignalMask sigmask, DSSoftSignalList softsig, DSTimeConst time) {
    DSDurationS duration; 
    DSTimeS cur;
    
    dsTimeCurrentTime(&cur);
    dsTimeCalcDuration(&duration, &cur, time);

    if ((duration.s<0)||((duration.s == 0)&&(duration.us == 0))) return 0;
    return dsSignalWaitRT(sigmask, softsig, &duration);
}
