#include <dslib.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */

#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif /* HAVE_FCNTL_H */

#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif /* HAVE_ERRNO_H */

#include <dsassert.h>
#include <dsdebug.h>

#include <dssignal.h>
#include <dsmalloc.h>

void dsSoftSignalZeroContext(DSSoftSignalContext ctx) {
    ds_assert_return(ctx);

#ifdef _WIN32
    ctx->h = INVALID_HANDLE;
#else /* _WIN32 */
    ctx->r = 0;
    ctx->w = 0;
#endif /* _WIN32 */
}

DSSoftSignalContext dsSoftSignalCreateContext(DSMallocContext mctx) {
    DSSoftSignalContext ctx;

    ctx = (DSSoftSignalContext)dsMalloc(mctx, sizeof(struct DSSoftSignalContextT));
    if (ctx) {
	ctx->malloc = mctx;
	dsSignalZeroContext(ctx);
    }
    return ctx;
}

int dsSoftSignalInitContext(DSSoftSignalContext ctx) {
#ifndef _WIN32
    int err;
    
    int flags;
    int b[2];
#endif /* _WIN32 */

    ds_assert_param(ctx);
    
    dsSoftSignalZeroContext(ctx);
    
#ifdef _WIN32
    ctx->h = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (ctx->h == INVALID_HANDLE) return DS_ERR_WINAPI_EVENT_CREATE; 
#else /* _WIN32 */
    
    if (pipe(b)) return DS_ERR_SOCKET_CREATE;

retry_fcntl_read:	
    flags = fcntl(b[0],F_GETFL,0);
    if ((flags<0)||(fcntl(b[0],F_SETFL,flags&(~O_NONBLOCK))<0)) {
	err = errno;
	if (err == EAGAIN) goto retry_fcntl_read;
	close(b[0]);
	close(b[1]);
	return DS_ERR_SOCKET_FCNTL;
    }

retry_fcntl_write:	
    flags = fcntl(b[1],F_GETFL,0);
    if ((flags<0)||(fcntl(b[1],F_SETFL,flags&(~O_NONBLOCK))<0)) {
	err = errno;
	if (err == EAGAIN) goto retry_fcntl_write;
	close(b[0]);
	close(b[1]);
	return DS_ERR_SOCKET_FCNTL;
    }

    ctx->r = b[0];
    ctx->w = b[1];
#endif /* _WIN32 */
    
    return 0;
}

void dsSoftSignalFreeContext(DSSoftSignalContext ctx) {
#ifdef _WIN32
    if (ctx->h != INVALID_HANDLE) CloseHandle(ctx->h);
#else /* _WIN32 */
    if (ctx->w) close(ctx->w);
    if (ctx->r) close(ctx->r);
#endif /* _WIN32 */

    dsSoftSignalZeroContext(ctx);
}

void dsSoftSignalDestroyContext(DSSoftSignalContext ctx) {
    ds_assert_return(ctx);
    dsSoftSignalFreeContext(ctx);
    dsFree(ctx->malloc, ctx);
}


int dsSoftSignalClean(DSSoftSignalContext ctx) {
#ifndef _WIN32
    int err;
    int buf[255];
#endif /* _WIN32 */

    ds_assert_param(ctx);

#ifdef _WIN32
    ds_assert_param(ctx->h);

    ResetEvent(ctx->h);
#else /* _WIN32 */
    ds_assert_param(ctx->r);
    
    while (!dsSignalCheckFDRead(ctx->r)) {
	err = read(ctx->r, buf, 255);
	if (err) {
	    if (errno==EAGAIN) continue;
//	    if (errno==EINTR) continue;
	    return DS_ERR_SOCKET_READ;
	}
    }
#endif /* _WIN32 */

    return 0;
}

int dsSoftSignalSend(DSSoftSignalContext ctx) {
#ifndef _WIN32
    int err;
    char d = 0;
#endif /* _WIN32 */

    ds_assert_param(ctx);

#ifdef _WIN32
    ds_assert_param(ctx->h);
    SetEvent(ctx->h);
#else /* _WIN32 */
    ds_assert_param(ctx->w);
repeat:
    err = write(ctx->w, &d, 1);
    if (err) {
	if ((errno == EAGAIN)||(errno == EINTR)) goto repeat;
	err = DS_ERR_SOCKET_WRITE;
    }
#endif /* _WIN32 */

    return 0;
}

int dsSoftSignalWait(DSSoftSignalContext ctx, DSDurationConst duration) {
    DSSoftSignalContext list[2] = { ctx, NULL };
    return dsSignalWait(NULL, list, duration);
}

int dsSoftSignalWaitRT(DSSoftSignalContext ctx, DSDurationConst duration) {
    DSSoftSignalContext list[2] = { ctx, NULL };
    return dsSignalWaitRT(NULL, list, duration);
}

int dsSoftSignalWaitUntil(DSSoftSignalContext ctx, DSTimeConst time) {
    DSSoftSignalContext list[2] = { ctx, NULL };
    return dsSignalWaitUntil(NULL, list, time);
}

int dsSoftSignalWaitRTUntil(DSSoftSignalContext ctx, DSTimeConst time) {
    DSSoftSignalContext list[2] = { ctx, NULL };
    return dsSignalWaitRTUntil(NULL, list, time);
}


int dsSoftSignalWaitSeconds(DSSoftSignalContext ctx, time_t s) {
    DSDurationS duration;
    dsDurationImportSeconds(&duration, s);
    return dsSoftSignalWait(ctx, &duration);
}

int dsSoftSignalWaitMillseconds(DSSoftSignalContext ctx, unsigned long ms) {
    DSDurationS duration;
    dsDurationImportMilliseconds(&duration, ms);
    return dsSoftSignalWait(ctx, &duration);
}

int dsSoftSignalWaitMicroseconds(DSSoftSignalContext ctx, unsigned long us) {
    DSDurationS duration;
    dsDurationImportMicroseconds(&duration, us);
    return dsSoftSignalWait(ctx, &duration);
}

int dsSoftSignalWaitMicrosecondsRT(DSSoftSignalContext ctx, unsigned long us) {
    DSDurationS duration;
    dsDurationImportMicroseconds(&duration, us);
    return dsSoftSignalWaitRT(ctx, &duration);
}
