#include <dslib.h>
#include <dsassert.h>
#include <dsmalloc.h>
#include <dsdebug.h>
#include <dssignal.h>
#include <dstools.h>

#include <dsthread.h>


void dsThreadZeroContext(DSThreadContext ctx) {
    dsBaseThreadZeroContext((DSBaseThreadContext)ctx);
/*
    ThreadHandleZero(ctx->thread);
*/
    ctx->thread.prop.function = NULL;
}

DSThreadContext dsThreadCreateContext(DSMallocContext mctx) {
    DSThreadContext ctx;

    ctx = (DSThreadContext)dsMalloc(mctx,sizeof(struct DSThreadContextT));
    dsDebugServer(DS_DEBUG_NOTICE, "dsThread %p: CreateContext", ctx);    
    if (ctx) {
	ctx->base.cond.malloc=mctx;
	dsThreadZeroContext(ctx);
    }
    return ctx;
}

int dsThreadInitContext(DSThreadContext ctx) {
    int err;
    ds_assert_param(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "dsThread %p: InitContext", ctx);    
    
    dsThreadZeroContext(ctx);

#if defined(HAVE_WIN32_THREADS)
    /* DS: WINAPI states callback function should return DWORD, - what should happen
    in case of 64bit pointers? */
    if (sizeof(void*) != sizeof(DWORD)) {
	dsDebugServer(DS_DEBUG_CRITICAL, "WIN32 Threads API.  Unsupported Pointer Size: %u", sizeof(void*));
	return DS_ERR_WINAPI;
    }
#endif
    
    err = dsBaseThreadInitContext((DSBaseThreadContext)ctx);
    if (!err) err = dsThreadRegisterHandles((DSBaseThreadContext)ctx, 1, &ctx->thread);
    if (err) {
	dsThreadFreeContext(ctx);
	return err;
    }
    
    ctx->base.initialized = 1;

    return 0;
}

int dsThreadFreeContext(DSThreadContext ctx) {
    int err;
    ds_assert_param(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "dsThread %p: FreeContext",ctx);    

    
    if (ctx->base.initialized) err=dsThreadTerminateAndWait(ctx);
    else err=0;
/*    
    ThreadHandleFree(&ctx->thread);
*/
    dsBaseThreadFreeContext((DSBaseThreadContext)ctx);
    
    dsThreadZeroContext(ctx);
    
    return err;
}

int dsThreadDestroyContext(DSThreadContext ctx) {
    int err;
    ds_assert_param(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "dsThread %p: DestroyContext",ctx);    

    err = dsThreadFreeContext(ctx);
    dsFree(ctx->base.cond.malloc,ctx);
    return err;
}

void dsThreadNotify(DSThreadContext ctx) {
    ds_assert_return(ctx);
    ds_assert_return(ctx->base.initialized);

    dsSignalSendTo(&ctx->thread.handle, DS_SIGNAL_NOTIFY);
}

int dsThreadRun(DSThreadContext ctx, DSThreadFunction thread_function, void *attr) {
    int err;

#ifdef HAVE_PTHREAD_H
    int custom = 0;
    pthread_attr_t atrd;
    struct sched_param ptrd;
#endif /* HAVE_PTHREAD_H */
    
    ds_assert_param(ctx);
    ds_assert_param(thread_function);
    ds_assert_param(ctx->base.initialized);
    ds_assert_param(ctx->thread.prop.function == NULL);
    
    ctx->thread.prop.malloc = ctx->base.cond.malloc;
    ctx->thread.prop.function = thread_function;
    ctx->thread.prop.attr = attr;
    ctx->thread.priority = ctx->base.priority;
    
#if defined(HAVE_PTHREAD_H)
    switch(ctx->base.sched) {
	    case DS_THREAD_SCHEDULER_INHERIT:
		pthread_attr_init(&atrd);
		pthread_attr_setinheritsched(&atrd, PTHREAD_INHERIT_SCHED);
		custom = 1;
	    break;
	    case DS_THREAD_SCHEDULER_REALTIME:
		pthread_attr_init(&atrd);
		pthread_attr_setschedpolicy(&atrd,SCHED_FIFO);
    		ptrd.sched_priority=ctx->base.priority;
    		pthread_attr_setschedparam(&atrd, &ptrd);
		custom = 1;
	    break;
	    case DS_THREAD_SCHEDULER_REALTIME_RR:
		pthread_attr_init(&atrd);
		pthread_attr_setschedpolicy(&atrd,SCHED_RR);
    		ptrd.sched_priority=ctx->base.priority;
    		pthread_attr_setschedparam(&atrd, &ptrd);
		custom = 1;
	    break;
    }

    err = pthread_create(&ctx->thread.handle.id,custom?&atrd:NULL,(void*(*)(void*))dsThreadGo,(void*)&ctx->thread);
    if (err) err = DS_ERR_SYSCALL_PTHREAD_CREATE;
    if (custom) pthread_attr_destroy(&atrd);
#else /* defined(HAVE_WIN32_THREADS) */
    ctx->thread.handle.id = 0;
    ctx->thread.thread_handle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)dsThreadGo, (LPVOID)&ctx->thread, 0, &err);
    if (!ctx->thread.thread_handle) err = DS_ERR_WINAPI;
    else err = 0;
#endif /* defined(HAVE_PTHREAD_H) */
    if (!err) {
#if defined(HAVE_PTHREAD_H)
#else /* defined(HAVE_WIN32_THREADS) */
	while (!ctx->thread.handle.id);
#endif /* defined(HAVE_PTHREAD_H) */
	dsDebugServer(DS_DEBUG_INFO,"dsThread %p: Run %8.lx",ctx, ctx->thread.handle.id);
	dsThreadSetStatusBits((DSBaseThreadContext)ctx,DS_THREAD_RUNNING_FLAG|DS_THREAD_JOINREQUIRED_FLAG);
    }
    return err;
}


int dsThreadWaitTerminate(DSThreadContext ctx) {
    int err;
#if defined(HAVE_PTHREAD_H)
    int *retcode;
#else /* defined(HAVE_WIN32_THREADS) */
    DWORD retcode = STILL_ACTIVE;
#endif /* defined(HAVE_PTHREAD_H) */
    DSLockContext lock;
    
    ds_assert_param(ctx);
    ds_assert_param(ctx->base.initialized);
    
    lock=ctx->base.lock;

    dsLockWrite(lock);
    if ((ctx->base.status&DS_THREAD_JOINREQUIRED_FLAG)==DS_THREAD_JOINREQUIRED_FLAG) {
	ctx->base.status&=~DS_THREAD_JOINREQUIRED_FLAG;
	dsLockReturn(lock);

#if defined(HAVE_PTHREAD_H)
	pthread_join(ctx->thread.handle.id,(void**)&retcode);
#else /* defined(HAVE_WIN32_THREADS) */
	for (err = STILL_ACTIVE;err == STILL_ACTIVE;) {
	    WaitForSingleObject(ctx->thread.thread_handle, INFINITE);
	    retcode = GetExitCodeThread(ctx->thread.thread_handle, &err);
	    if (!retcode) err = 0;
	}
	CloseHandle(ctx->thread.thread_handle);
#endif /* defined(HAVE_PTHREAD_H) */
	dsLockWrite(lock);
	ctx->base.status&=~DS_THREAD_RUNNING_FLAG;
	dsLockReturn(lock);
	dsConditionBroadcast((DSConditionContext)ctx);

	if (retcode) {
#if defined(HAVE_PTHREAD_H)
	    if (*((void**)retcode)==PTHREAD_CANCELED) err=DS_ERR_THREAD_CANCELED;
	    else err=*retcode;
	    free(retcode);
#else /* defined(HAVE_WIN32_THREADS) */
#endif /* defined(HAVE_PTHREAD_H) */
	} else err=DS_ERR_THREAD_RETURNCODE;
	return err;
    } else {
	dsLockReturn(lock);
	dsThreadWaitStatusBitsOff((DSBaseThreadContext)ctx,DS_THREAD_RUNNING_FLAG);
    }
    return 0;
}

int dsThreadTerminateAndWait(DSThreadContext ctx) {
    ds_assert_param(ctx);
    
    dsThreadTerminate((DSBaseThreadContext)ctx);
    dsThreadNotify(ctx);
    return dsThreadWaitTerminate(ctx);
}

int dsThreadIsRunning(DSThreadContext ctx) {
    ds_assert_param(ctx);
    return dsThreadCheckStatusBitsOn((DSBaseThreadContext)ctx, DS_THREAD_RUNNING_FLAG);
}

