#include <stdio.h>
#include <string.h>
#include <time.h>

#include <dsassert.h>
#include <dstime.h>

#ifndef HAVE_LOCALTIME_R
static DSLockContext localtimelock = NULL;
#endif /* HAVE_LOCALTIME_R */

int dsLibTimeInit() {
#ifndef HAVE_LOCALTIME_R
# ifndef _WIN32
    int err = 0;

    localtimelock = dsLockCreateContext(NULL);
    if (!localtimelock) return DS_ERR_MEM;
    
    err = dsLockInitContext(localtimelock);
    if (err) {
	dsLockDestroyContext(localtimelock);
	localtimelock = NULL;
    }
    
    return err;
# endif /* _WIN32 */
#endif /* HAVE_LOCALTIME_R */
    return 0;
}

void dsLibTimeFree() {
#ifndef HAVE_LOCALTIME_R
# ifndef _WIN32
    if (localtimelock) {
	dsLockDestroyContext(localtimelock);
	localtimelock = NULL;
    }
# endif /* _WIN32 */
#endif /* HAVE_LOCALTIME_R */
}

DSTime dsTimeCreate(DSMallocContext mctx) {
    DSTime time;
    
    time = (DSTime)dsMalloc(mctx, sizeof(struct DSTimeT));
    if (!time) return NULL;
    time->malloc=mctx;
    return time;
}

void dsTimeDestroy(DSTime time) {
    assert(time);
    
    dsFree(time->malloc, time);
}


int dsTimeExportLocaltime(DSTimeConst time, struct tm *res) {
    int err = 0;
    struct tm *t;

    ds_assert_param(time);
    ds_assert_param(res);
    
#ifdef HAVE_LOCALTIME_R
    if (!localtime_r(&time->s, res)) return DS_ERR_SYSCALL_LOCALTIME;
#else
    if (localtimelock) dsLockWrite(localtimelock);
    t = localtime(&time->s);
    if (t) memcpy(res, t, sizeof(struct tm));
    else err = DS_ERR_SYSCALL_LOCALTIME;
    if (localtimelock) dsLockReturn(localtimelock);
#endif /* HAVE_LOCALTIME_R */
    return err;
}

int dsTimeExportGmtime(DSTimeConst time, struct tm *res) {
    int err = 0;
    struct tm *t;

    ds_assert_param(time);
    ds_assert_param(res);
    
#ifdef HAVE_GMTIME_R
    if (!gmtime_r(&time->s, res)) return DS_ERR_SYSCALL_GMTIME;
#else
    if (localtimelock) dsLockWrite(localtimelock);
    t = gmtime(&time->s);
    if (t) memcpy(res, t, sizeof(struct tm));
    else err = DS_ERR_SYSCALL_GMTIME;
    if (localtimelock) dsLockReturn(localtimelock);
#endif /* HAVE_GMTIME_R */
    return err;
}

int dsTimeExportOffsetGmtime(DSTimeConst time, struct tm *res) {
    int err = 0;
    struct tm *t;

    time_t offset_time;
    
    ds_assert_param(time);
    ds_assert_param(res);
    
    offset_time = time->s + time->offset;
    
#ifdef HAVE_GMTIME_R
    if (!gmtime_r(&offset_time, res)) return DS_ERR_SYSCALL_GMTIME;
#else
    if (localtimelock) dsLockWrite(localtimelock);
    t = gmtime(&offset_time);
    if (t) memcpy(res, t, sizeof(struct tm));
    else err = DS_ERR_SYSCALL_GMTIME;
    if (localtimelock) dsLockReturn(localtimelock);
#endif /* HAVE_GMTIME_R */
    return err;
}

DSDuration dsDurationCreate(DSMallocContext mctx) {
    DSDuration duration;
    
    duration = (DSDuration)dsMalloc(mctx, sizeof(struct DSDurationT));
    if (!duration) return NULL;
    duration->malloc=mctx;
    return duration;
}

void dsDurationDestroy(DSDuration duration) {
    assert(duration);
    dsFree(duration->malloc, duration);
}

int dsTimeZeroTime(DSTime time) {
    ds_assert_param(time);
    
    time->s = 0;
    time->us = 0;
    return 0;
}

int dsTimeIsZero(DSTimeConst time) {
    ds_assert_param(time);
    
    if ((!time->s)&&(!time->us)) return 1;
    return 0;
}

int dsDurationZeroDuration(DSDuration duration) {
    ds_assert_param(duration);
    
    duration->s = 0;
    duration->us = 0;
    return 0;
}

int dsDurationIsZero(DSDurationConst duration) {
    ds_assert_param(duration);
    
    if ((!duration->s)&&(!duration->us)) return 1;
    return 0;
}


int dsTimeCurrentOffset(DSTime dstime) {
#if defined(HAVE_STRUCT_TM_TM_ZONE) || defined(HAVE_INT_TIMEZONE)
//    DSTimeS fake = { 0 };
    struct tm t;
#endif /* HAVE_STRUCT_TM_TM_ZONE || HAVE_INT_TIMEZONE */

    ds_assert_param(dstime);

#if defined(HAVE_STRUCT_TM_TM_ZONE)
    dsTimeExportLocaltime(dstime, &t);	/* for daylight savings it should be real time */
    dstime->offset=t.tm_gmtoff;
#elif defined(HAVE_INT_TIMEZONE)
    /* DS: timezone variable do not consider daylight savings. We use tm_isdst 
    to adjust timezone, but the daylight savings is not everythere 1h. So. Should
    be fixed. The localtime() function should be called to set exact timezone value */
    dsTimeExportLocaltime(dstime, &t);	/* for daylight savings it should be real time */
# ifndef _WIN32
#  warning "We have problems with daylight savings here!";
# endif /* _WIN32 */
# ifdef HAVE_UNDERSCORE_TIMEZONE
    dstime->offset=-_timezone;
# else /* HAVE_UNDERSCORE_TIMEZONE */
    dstime->offset=-timezone;
# endif /* HAVE_UNDERSCORE_TIMEZONE */
    if (t.tm_isdst>0) dstime->offset+=3600;
#else /* defined(HAVE_STRUCT_TM_TM_ZONE) */
# error "No way to determine timezone!";
#endif /* defined(HAVE_STRUCT_TM_TM_ZONE) */
    return 0;
}

int dsTimeCurrentTime(DSTime time) {
    ds_assert_param(time);
    
    gettimeofday((struct timeval*)time,NULL);
    dsTimeCurrentOffset(time);

    return 0;
}

#ifdef HAVE_TIMESPEC
int dsDurationImportTimespec(DSDuration duration, const struct timespec *ts) {
    ds_assert_param(duration);
    ds_assert_param(ts);

    duration->s = ts->tv_sec;
    duration->us = ts->tv_nsec / 1000;
    if ((!duration->s)&&(!duration->us)&&(ts->tv_nsec))  duration->us = 1;
    return 0;
}

int dsDurationExportTimespec(DSDurationConst duration, struct timespec *ts) {
    ds_assert_param(duration);
    ds_assert_param(ts);
    
    ts->tv_sec = duration->s;
    ts->tv_nsec = duration->us * 1000;
    return 0;
}

int dsTimeImportTimespec(DSTime time, const struct timespec *ts) {
    ds_assert_param(time);
    ds_assert_param(ts);

    time->s = ts->tv_sec;
    time->us = ts->tv_nsec / 1000;
    dsTimeCurrentOffset(time);
    return 0;
}

int dsTimeExportTimespec(DSTimeConst time, struct timespec *ts) {
    ds_assert_param(time);
    ds_assert_param(ts);
    
    ts->tv_sec = time->s;
    ts->tv_nsec = time->us * 1000;
    return 0;
}
#endif /* HAVE_TIMESPEC */

#ifdef HAVE_SYS_TIME_H
int dsTimeImportTimeval(DSTime time, struct timeval *newtime) {
    ds_assert_param(time);
    ds_assert_param(newtime);
    
    memcpy(time, newtime, sizeof(struct timeval));
    dsTimeCurrentOffset(time);

    return 0;
}
#endif /* HAVE_SYS_TIME_H */
int dsDurationImportSeconds(DSDuration duration, time_t s) {
    ds_assert_param(duration);
    duration->s = s;
    duration->us = 0;
    
    return 0;
}

time_t dsTimeGetUnixTime(DSTimeConst time) {
    ds_assert_default(time, 0);
    return time->s;
}


int dsDurationImportMilliseconds(DSDuration duration, unsigned long ms) {
    ds_assert_param(duration);
    
    duration->s = ms/1000;
    duration->us = (unsigned long)((ms-1000*duration->s)*1000);
    return 0;
}

int dsDurationImportMicroseconds(DSDuration duration, unsigned long us) {
    ds_assert_param(duration);
    
    duration->s = us/1000000;
    duration->us = (unsigned long)(us-1000000*duration->s);
    return 0;
}

int dsTimeSetOffset(DSTime time, long offset) {
    time->offset = offset;
    return 0;
}

int dsTimeCopy(DSTime dst, DSTimeConst src) {
    ds_assert_param(dst);
    ds_assert_param(src);
    
    memcpy(dst, src, sizeof(struct DSTimeT));
    return 0;
}

DSTime dsTimeCreateCopy(DSMallocContext mctx, DSTimeConst tm) {
    DSTime res;

    ds_assert_null(tm);
    
    res = dsTimeCreate(mctx);
    if (!res) return res;
    
    dsTimeCopy(res, tm);
    return res;
}

int dsDurationCopy(DSDuration dst, DSDurationConst src) {
    ds_assert_param(dst);
    ds_assert_param(src);
    
    memcpy(dst, src, sizeof(struct DSDurationT));
    return 0;
}

DSDuration dsDurationCreateCopy(DSMallocContext mctx, DSDurationConst dur) {
    DSDuration res;
    
    ds_assert_null(dur);

    res = dsDurationCreate(mctx);
    if (!res) return res;
    
    dsDurationCopy(res, dur);
    return res;
}

char *dsTimeGetTime(DSTimeConst time, char *buffer) {
    struct tm t;
    int hoffset, moffset;
    char *str;
    
    ds_assert_null(time);
    
    if (buffer) str=buffer;
    else {
	str=(char*)malloc(23*sizeof(char));
	if (!str) return NULL;
    }
    
    dsTimeExportOffsetGmtime(time, &t);
    
    sprintf(str,"%02u:%02u:%02u.%07u",
	t.tm_hour,t.tm_min,t.tm_sec,
	time->us*10);

    if (time->offset==0)
	strcpy(str+27,"Z");
    else {
	hoffset=time->offset/3600;
	moffset=abs(time->offset/60-hoffset*60);
	sprintf(str+27,"%+03i:%02u",hoffset,moffset);
    }
    
    return str;
}

char *dsTimeGetDate(DSTimeConst time, char *buffer) {
    struct tm t;
    char *str;
    
    ds_assert_null(time);
    
    if (buffer) str=buffer;
    else {
	str=(char*)malloc(11*sizeof(char));
	if (!str) return NULL;
    }
    
    dsTimeExportOffsetGmtime(time, &t);
    
    sprintf(str,"%04u-%02u-%02u",
	t.tm_year+1900,t.tm_mon+1,t.tm_mday);
    
    return str;

}

char *dsTimeGetLocalDateTime(DSTimeConst time, char *buffer) {
    struct tm t;
    struct tm gm;
    long offset;
    int hoffset, moffset;
    char *str;
    
    ds_assert_null(time);
    
    if (buffer) str=buffer;
    else {
	str=(char*)malloc(34*sizeof(char));
	if (!str) return NULL;
    }

    dsTimeExportLocaltime(time, &t);
    dsTimeExportGmtime(time, &gm);
    
    offset = (t.tm_sec - gm.tm_sec) + 60*((t.tm_min - gm.tm_min) + 60*(t.tm_hour - gm.tm_hour));
    if ((t.tm_mday != gm.tm_mday)||(t.tm_mon != gm.tm_mon)||(t.tm_year != gm.tm_year)) {
	if (t.tm_year > gm.tm_year) offset += 86400;
	else if (t.tm_year < gm.tm_year) offset -= 86400;
	else if (t.tm_mon > gm.tm_mon) offset += 86400;
	else if (t.tm_mon < gm.tm_mon) offset -= 86400;
	else if (t.tm_mday > gm.tm_mday) offset += 86400;
	else offset -= 86400;
    }	
    
    sprintf(str,"%04u-%02u-%02uT%02u:%02u:%02u.%07u",
	t.tm_year+1900,t.tm_mon+1,t.tm_mday,
	t.tm_hour,t.tm_min,t.tm_sec,
	time->us*10);

    if (offset==0)
	strcpy(str+27,"Z");
    else {
	hoffset=offset/3600;
	moffset=abs(offset/60-hoffset*60);
	sprintf(str+27,"%+03i:%02u",hoffset,moffset);
    }
    
    return str;
}


char *dsTimeGetDateTime(DSTimeConst time, char *buffer) {
    struct tm t;
    int hoffset, moffset;
    char *str;
    
    ds_assert_null(time);
    
    if (buffer) str=buffer;
    else {
	str=(char*)malloc(34*sizeof(char));
	if (!str) return NULL;
    }

/*    
    dsTimeExportLocaltime(time, &t);
*/
    dsTimeExportOffsetGmtime(time, &t);
    
    sprintf(str,"%04u-%02u-%02uT%02u:%02u:%02u.%07u",
	t.tm_year+1900,t.tm_mon+1,t.tm_mday,
	t.tm_hour,t.tm_min,t.tm_sec,
	time->us*10);

    if (time->offset==0)
	strcpy(str+27,"Z");
    else {
	hoffset=time->offset/3600;
	moffset=abs(time->offset/60-hoffset*60);
	sprintf(str+27,"%+03i:%02u",hoffset,moffset);
    }
    
    return str;
}


/* ISO 8601 */
int dsTimeSetDateTime(DSTime time, const char *buffer) {
    char *ch, chs;
    char *tmp = (char*)buffer;
    struct tm tms;
    time_t result;
    unsigned long day,month,year;
    unsigned long hour,minute,second;
    unsigned long us;
    long offset;
    size_t uslen;
    
    ds_assert_param(time);
    ds_assert_param(buffer);
    
    ch = strchr(tmp,'-');
    if (!ch) return DS_ERR_INVALID_ARGUMENT;
    *ch=0;year = atol(tmp);*ch='-';
    tmp=ch+1;

    ch = strchr(tmp,'-');
    if (!ch) return DS_ERR_INVALID_ARGUMENT;
    *ch=0;month = atol(tmp);*ch='-';
    tmp=ch+1;

    ch = strchr(tmp,'T');
    if (!ch) return DS_ERR_INVALID_ARGUMENT;
    *ch=0;day = atol(tmp);*ch='T';
    tmp=ch+1;

    ch = strchr(tmp,':');
    if (!ch) return DS_ERR_INVALID_ARGUMENT;
    *ch=0;hour = atol(tmp);*ch=':';
    tmp=ch+1;

    ch = strchr(tmp,':');
    if (!ch) return DS_ERR_INVALID_ARGUMENT;
    *ch=0;minute = atol(tmp);*ch=':';
    tmp=ch+1;


    ch = strchr(tmp,'.');
    if (!ch) ch = strchr(tmp,'Z');
    if (!ch) ch = strchr(tmp,'+');
    if (!ch) ch = strchr(tmp,'-');

    chs=*ch;
    if (chs) *ch=0;
    second = atol(tmp);
    if (chs) *ch=chs;
    tmp=ch+1;
    
    if (chs=='.') {
	ch = strchr(tmp,'Z');
	if (!ch) ch = strchr(tmp,'+');
	if (!ch) ch = strchr(tmp,'-');
	chs=*ch;

	if (chs) *ch=0;
	uslen = strlen(tmp);
	if (uslen>6) {
	    uslen=tmp[6];
	    tmp[6]=0;
	    us = atol(tmp);
	    tmp[6]=(unsigned char)uslen;
	} else {
	    us = atol(tmp);
	    for (uslen=6-uslen;uslen>0;uslen--) us *= 10;
	}
	if (chs) *ch=chs;
	tmp=ch+1;
    } else us=0;
    
    if ((chs)&&(chs!='Z'))  {
	ch = strchr(tmp,':');
	if (!ch) return DS_ERR_INVALID_ARGUMENT;
	*ch=0;offset=atol(tmp);*ch=':';
	tmp=ch+1;

	if (chs=='+') offset = (offset*60+atol(tmp))*60;
	else offset = -(offset*60+atol(tmp))*60;
    } else offset=0;
    
    tms.tm_sec = second;
    tms.tm_min = minute;
    tms.tm_hour = hour;
    tms.tm_mday = day;
    if (month>0) tms.tm_mon = month-1;
    else return DS_ERR_INVALID_ARGUMENT;
    if (year>1899) tms.tm_year = year-1900;
    else tms.tm_year = year;
    tms.tm_isdst = -1;
    
    result=mktime(&tms);
    if (result==(time_t)-1) return DS_ERR_INVALID_ARGUMENT;
    
    time->s = result;
    time->us = us;
    time->offset = offset;

/*
    printf("%u-%u-%u\n",year,month,day);
    printf("%u:%u:%u.%u\n",hour,minute,second,us);
    printf("%u\n",offset);
*/

    return 0;
}

/*
  Format: PxYxMxDTxHxMxS,
    any of YMDHMS can be omited, if HMS specified T must presist,
    x can be any number, i.e. 120S - correct
*/

//ToDo: offsets
char *dsTimeGetDuration(DSTimeConst stime, DSTimeConst etime, char *buffer) {
    size_t s;
    unsigned long us;
    char *str;
    
    ds_assert_null(stime);
    ds_assert_null(etime);
    
    if (buffer) str=buffer;
    else {
	str=(char*)malloc(34*sizeof(char));
	if (!str) return NULL;
    }
    
    if (etime->s>stime->s) {
	s=etime->s-stime->s;
	if (etime->us>stime->us) us=etime->us-stime->us;
	else {
	    s--;
	    us=etime->us-stime->us+1000000;
	}
	sprintf(str,"PT%02u.%07uS",s,us*10);
    } else {
	s=stime->s-etime->s;
	if (etime->us<stime->us) us=stime->us-etime->us;
	else {
	    s--;
	    us=stime->us-etime->us+1000000;
	}
	sprintf(str,"-PT%02u.%07uS",s,us*10);
    }
    
    return str;    
}

char *dsDurationGet(DSDurationConst duration, char *buffer) {
    char *str;

    if (buffer) str=buffer;
    else {
	str=(char*)malloc(34*sizeof(char));
	if (!str) return NULL;
    }
    
    if (duration->s < 0) sprintf(str,"-PT%02u.%07uS",-duration->s,duration->us*10);
    else sprintf(str,"PT%02u.%07uS",duration->s,duration->us*10);
    
    return str;
}

int dsDurationSet(DSDuration duration, const char *buffer) {
    int i,digits;
    char *tmp = (char*)buffer;
    unsigned long value;
    
    ds_assert_param(duration);
    ds_assert_param(buffer);
    
    if ((*tmp) == 'P') {
	if (*(++tmp)=='T') tmp++;

	    // Partial support
	for (i=0,digits=0;tmp[i];i++) {
	    if ((tmp[i]<='9')&&(tmp[i]>='0')) continue;
	    if (tmp[i]=='.') {
		if (digits) return DS_ERR_NOTSUPPORTED;
		else {
		    digits = 1;
		    continue;
		}
	    }
	    if ((tmp[i]=='S')&&(!tmp[i+1])) break;
	    return DS_ERR_NOTSUPPORTED;
	}

	duration->s = atol(tmp);
	tmp = strchr(tmp, '.');
	if (tmp) {
	    value = atol(tmp + 1);
	    if (value) {
//		for (digits = strlen(tmp+1);((digits>0)&&((tmp[digits]<'0')||(tmp[digits]>'9')));digits--);
		digits = strlen(tmp+1);
		if ((digits)&&((tmp[digits]=='S')||(tmp[digits]=='s'))) digits--;
		if (digits<6) {
		    for (;digits<6;digits++) value*=10;
		} else {
		    for (;digits>6;digits--) value/=10;
		}
	    }
	    duration->us = value;
	} else duration->us = 0;
    } else {
	value = atol(tmp);
	if (strstr(tmp, "ms")) {
	    duration->s = value / 1000;
	    duration->us = (value - 1000*duration->s)*1000;
	} else if (strstr(tmp, "us")) {
	    duration->s = value / 1000000;
	    duration->us = value - 1000000 * duration->s;
	} else if (strstr(tmp, "s")) {
	    duration->s = value;
	    duration->us = 0;
	} else if (strstr(tmp, "m")) {
	    duration->s = value * 60;
	    duration->us = 0;
	} else if (strstr(tmp, "h")) {
	    duration->s = value * 3600;
	    duration->us = 0;
	} else { // Assuming "us"
	    duration->s = value / 1000000;
	    duration->us = value - 1000000 * duration->s;
	}
    }
    
    return 0;
}

int dsTimeCalcDuration(DSDuration duration, DSTimeConst stime, DSTimeConst etime) {
    long diff;
    
    ds_assert_param(duration);
    ds_assert_param(stime);
    ds_assert_param(etime);
    
    diff = stime->offset - etime->offset;
    
    if ((etime->s+diff)<stime->s) {
	duration->s=0;
	duration->us=0;
    } else {
	duration->s=diff+etime->s-stime->s;
	if (etime->us>stime->us) duration->us=etime->us-stime->us;
	else {
	    duration->s--;
	    duration->us=etime->us-stime->us+1000000;
	}
    }
    return 0;
}

int dsTimeCalcCurrentDuration(DSDuration duration, DSTimeConst time) {
    DSTimeS curtime;
    
    dsTimeCurrentTime(&curtime);
    return dsTimeCalcDuration(duration, &curtime, time);
}

int dsTimeCalcPassedDuration(DSDuration duration, DSTimeConst time) {
    DSTimeS curtime;
    
    dsTimeCurrentTime(&curtime);
    return dsTimeCalcDuration(duration, time, &curtime);
}

int dsTimeDiffSeconds(DSTimeConst stime, DSTimeConst etime) {
    int err;
    DSDurationS duration;
    
    ds_assert_param(stime);
    ds_assert_param(etime);
    
    err = dsTimeCmp(stime, etime);

    if (err < 0) {
	dsTimeCalcDuration(&duration, stime, etime);
	return dsDurationGetSeconds(&duration);
    } else if (err > 0) {
	dsTimeCalcDuration(&duration, etime, stime);
	return -(int)dsDurationGetSeconds(&duration);
    }
    return 0;
}

int dsTimeDiffCurrentSeconds(DSTimeConst time) {
    DSTimeS curtime;
    
    dsTimeCurrentTime(&curtime);
    return dsTimeDiffSeconds(&curtime, time);
}

int dsTimeDiffMilliseconds(DSTimeConst stime, DSTimeConst etime) {
    int err;
    DSDurationS duration;
    
    ds_assert_param(stime);
    ds_assert_param(etime);    

    err = dsTimeCmp(stime, etime);
    
    if (err < 0) {
        dsTimeCalcDuration(&duration, stime, etime);
	return dsDurationGetMilliseconds(&duration);
    } else if (err > 0) {
	dsTimeCalcDuration(&duration, etime, stime);
	return -(int)dsDurationGetMilliseconds(&duration);
    }
    return 0;
}

int dsTimeDiffCurrentMilliseconds(DSTimeConst time) {
    DSTimeS curtime;
    
    dsTimeCurrentTime(&curtime);
    return dsTimeDiffMilliseconds(&curtime, time);
}

int dsTimeDiffMicroseconds(DSTimeConst stime, DSTimeConst etime) {
    int err;
    DSDurationS duration;
    
    ds_assert_param(stime);
    ds_assert_param(etime);    

    err = dsTimeCmp(stime, etime);
    
    if (err < 0) {
	dsTimeCalcDuration(&duration, stime, etime);
        return dsDurationGetMicroseconds(&duration);
    } else if (err > 0) {
	dsTimeCalcDuration(&duration, etime, stime);
        return -(int)dsDurationGetMicroseconds(&duration);
    }
    return 0;
}

int dsTimeDiffCurrentMicroseconds(DSTimeConst time) {
    DSTimeS curtime;
    
    dsTimeCurrentTime(&curtime);
    return dsTimeDiffMicroseconds(&curtime, time);
}

int dsTimeAdd(DSTime time, DSDurationConst duration) {
    ds_assert_param(time);
    ds_assert_param(duration);
    
    time->s+=duration->s;
    time->us+=duration->us;
    if (time->us>999999) {
	time->s++;
	time->us-=1000000;
    }
    return 0;
}

int dsTimeAddSeconds(DSTime time, time_t sec) {
    ds_assert_param(time);
    
    time->s+=sec;
    return 0;
}

int dsTimeAddMilliseconds(DSTime time, unsigned long ms) {
    ds_assert_param(time);
    
    time->us+=(ms%1000)*1000;
    if (time->us>999999) {
	time->us -= 1000000;
	time->s+=ms/1000 + 1;
    } else time->s += ms/1000;

    return 0;
}

int dsTimeAddMicroseconds(DSTime time, unsigned long us) {
    ds_assert_param(time);
    
    time->us+=us%1000000;
    if (time->us>999999) {
	time->us -= 1000000;
	time->s+=us/1000000 + 1;
    } else time->s+=us/1000000;
    
    return 0;
}

int dsTimeSub(DSTime time, DSDuration duration) {
    ds_assert_param(time);
    ds_assert_param(duration);
    
    time->s-=duration->s;
    
    if (time->us<duration->us) {
	time->s--;
	time->us+=1000000-duration->us;
    } else {
	time->us-=duration->us;
    }
	
    return 0;
}

int dsTimeSubSeconds(DSTime time, time_t s) {
    ds_assert_param(time);
    
    time->s -= s;
    
    return 0;
}

int dsTimeSubMilliseconds(DSTime time, unsigned long ms) {
    time_t s;
    
    ds_assert_param(time);

    s = ms / 1000;
    ms = (ms - s*1000)*1000;

    if ((unsigned int)time->us < ms) {
	time->us += 1000000 - ms;
	time->s -= s + 1;
    } else {
	time->us -= ms;
	time->s -= s;
    }

    return 0;
}

int dsTimeSubMicroseconds(DSTime time, unsigned long us) {
    time_t s;
    
    ds_assert_param(time);

    s = us / 1000000;
    us -= s*1000000;
    
    if ((unsigned int)time->us < us) {
	time->us += 1000000 - us;
	time->s -= s + 1;
    } else {
	time->us -= us;
	time->s -= s;
    }
    
    return 0;

}

int dsDurationAdd(DSDuration duration, DSDurationConst addition) {
    ds_assert_param(duration);
    
    if (!addition) return 0;
    
    duration->us += addition->us;
    if (duration->us > 999999) {
	duration->us -= 1000000;
	duration->s += addition->s + 1;
    } else duration->s += addition->s;
    
    return 0;
}

int dsDurationSub(DSDuration duration, DSDurationConst substraction) {
    ds_assert_param(duration);
    
    if (!substraction) return 0;
    
    if (duration->us < substraction->us) {
	duration->us += 1000000 - substraction->us;
	duration->s -= substraction->s + 1;
    } else {
	duration->us -= substraction->us;
	duration->s -= substraction->s;
    }
    
    return 0;
}

int dsDurationSubSeconds(DSDuration duration, time_t s) {
    ds_assert_param(duration);
    
    duration->s -= s;
    return 0;
}

int dsDurationSubMilliseconds(DSDuration duration, unsigned long ms) {
    time_t s;
    
    ds_assert_param(duration);

    s = ms / 1000;
    ms = (ms - s*1000)*1000;

    if ((unsigned int)duration->us < ms) {
	duration->us += 1000000 - ms;
	duration->s -= s + 1;
    } else {
	duration->us -= ms;
	duration->s -= s;
    }
    
    return 0;
}

int dsDurationSubMicroseconds(DSDuration duration, unsigned long us) {
    time_t s;
    
    ds_assert_param(duration);
    
    s = us / 1000000;
    us -= s*1000000;
    
    if ((unsigned int)duration->us < us) {
	duration->us += 1000000 - us;
	duration->s -= s + 1;
    } else {
	duration->us -= us;
	duration->s -= s;
    }
    
    return 0;
}


int dsDurationAddSeconds(DSDuration duration, time_t s) {
    ds_assert_param(duration);
    
    duration->s += s;
    return 0;
}

int dsDurationAddMilliseconds(DSDuration duration, unsigned long ms) {
    ds_assert_param(duration);
    
    duration->us += (ms % 1000)*1000;
    if (duration->us > 999999) {
	duration->us -= 1000000;
	duration->s += ms / 1000 + 1;
    } else duration->s += ms / 1000;
    
    return 0;
}

int dsDurationAddMicroseconds(DSDuration duration, unsigned long us) {
    ds_assert_param(duration);
    
    duration->us += us%1000000;
    if (duration->us > 999999) {
	duration->us -= 1000000;
	duration->s += us / 1000000 + 1;
    } else duration->s = us / 1000000;
    
    return 0;
}

int dsDurationMultiply(DSDuration duration, double multiplier) {
    double s, us;
    time_t sec, usec;
    
    ds_assert_param(duration);
    
    s = duration->s * multiplier;
    sec = (time_t)s;
    
    us = duration->us * multiplier + (s - sec)*1000000;
    usec =(time_t)(us / 1000000);
    
    duration->s = sec + usec;
    duration->us = (unsigned long)(us - usec*1000000);

    return 0;    
}

int dsDurationIntegerMultiply(DSDuration duration, unsigned long multiplier) {
    ds_assert_param(duration);

    if (duration->us%1000) {
	if (multiplier > 4000) return DS_ERR_INVALID_ARGUMENT;
	
	duration->us *= multiplier;
	duration->s = duration->s*multiplier + duration->us / 1000000;
	duration->us = duration->us % 1000000;
    } else if (duration->us) {
	if (multiplier > 4000000) return DS_ERR_INVALID_ARGUMENT;
	
	duration->us = (duration->us / 1000) * multiplier; 
	duration->s = duration->s * multiplier + duration->us / 1000;
	duration->us = (duration->us % 1000) * 1000;
    } else {
	duration->s *= multiplier;
    }
    
    return 0;
}

int dsDurationIntegerDivide(DSDuration duration, unsigned long divider) {
    time_t s;
    unsigned long us;
    
    ds_assert_param(duration);
    
    s = duration->s / divider;
    if (divider < 4000) us = ((duration->s % divider)*1000000 + duration->us) / divider;
    else us = (suseconds_t)((duration->s % divider)*1000000.0 + duration->us) / divider;
    
    duration->s = s +  us / 1000000;
    duration->us = us % 1000000;
    
    return 0;
}

int dsTimeCmp(DSTimeConst time1, DSTimeConst time2) {
    long diff;
    
    ds_assert_default(time1, time2?-1:0);
    ds_assert_default(time2, 1);
    
    diff = time1->offset - time2->offset;
    
    if (time1->s>(time2->s+diff)) return 1;
    else if (time1->s<(time2->s+diff)) return -1;
    else {
	if (time1->us>time2->us) return 1;
	else if (time1->us<time2->us) return -1;
    }
    return 0;
}

int dsTimeCurrentCmp(DSTimeConst time) {
    DSTimeS curtime;
    
    ds_assert_default(time, -1);
    
    dsTimeCurrentTime(&curtime);
    return dsTimeCmp(time, &curtime);
}

int dsDurationCmp(DSDurationConst duration1, DSDurationConst duration2) {
    ds_assert_default(duration1, duration2?-1:0);
    ds_assert_default(duration2, 1);
    
    if (duration1->s > duration2->s) return 1;
    else if (duration1->s < duration2->s) return -1;
    else {
	if (duration1->us > duration2->us) return 1;
	else if (duration2->us < duration2->us) return -1;
    }
    return 0;
}


unsigned long dsDurationGetSeconds(DSDurationConst duration) {
    ds_assert_default(duration, 0);
    if (duration->s < 0) return 0;

    return (duration->us)?duration->s+1:duration->s;
}

unsigned long dsDurationGetMilliseconds(DSDurationConst duration) {
    unsigned long res;
    
    ds_assert_default(duration, 0);
    if (duration->s < 0) return 0;

    res = duration->s*1000 + duration->us / 1000;
    return (duration->us%1000)?res+1:res;
}


unsigned long dsDurationGetMicroseconds(DSDurationConst duration) {
    ds_assert_default(duration, 0);
    if (duration->s < 0) return 0;

    return duration->s*1000000 + duration->us;
}


int dsDurationCeilMicroseconds(DSDuration duration, unsigned long us) {
    unsigned long dur, m;
    
    ds_assert_param(duration);
    ds_assert_param(us);
    
    dur = duration->s*1000000 + duration->us;

    m = dur / us;
    if (m * us < dur) m++;
    
    dur = m*us;
    duration->s = dur / 1000000;
    duration->us = dur - (duration->s*1000000);

    return 0;
}

int dsDurationCeilMilliseconds(DSDuration duration, unsigned long ms) {
    unsigned long dur, m;
    
    ds_assert_param(duration);
    ds_assert_param(ms);
    
    if (duration->us%1000) dur = duration->s*1000 + duration->us/1000 + 1; 
    else dur = duration->s*1000 + duration->us/1000; 
    
    m = dur / ms;
    if (m * ms < dur) m++;
    
    dur = m * ms;
    duration->s = dur / 1000;
    duration->us = (dur - duration->s*1000)*1000;

    return 0;
}

int dsDurationCeilSeconds(DSDuration duration, unsigned long s) {
    unsigned long dur, m;
    
    ds_assert_param(duration);
    ds_assert_param(s);
    
    if (duration->us) {
	dur = duration->s + 1;
	duration->us = 0;
    } else dur = duration->s;

    m = dur / s;
    if (m * s < dur) m++;

    duration->s = m * s;
    return 0;
}

#define DAYSECONDS 86400
#define HOURSECONDS 3600
int dsTimeCeil(DSTime time, DSDuration duration) {
    unsigned long m;
    unsigned long us;
    unsigned long ts, tus;
    
    ds_assert_param(time);
    ds_assert_param(duration);
    
    
    
    if (duration->us%1000) {
	if (duration->s >= HOURSECONDS) return DS_ERR_INVALID_ARGUMENT;
	
	ts = time->s % HOURSECONDS;

	us = duration->s * 1000000 + duration->us;
	tus = ts * 1000000 + time->us;
	
	m = tus / us;
	if  (m * us < tus) m++;
	
	tus = m * us;
	
	time->s = (time->s - ts) + tus / 1000000;
	time->us = tus % 1000000;
    } else if (duration->us) {
	if (duration->s >= DAYSECONDS) return DS_ERR_INVALID_ARGUMENT;
	
	ts = time->s % DAYSECONDS;
	
	us = duration->s * 1000 + duration->us / 1000;
	if (time->us%1000) tus = ts * 1000 + time->us / 1000 + 1;
	else tus = ts * 1000 + time->us / 1000;
	
	m = tus / us;
	if (m * us  < tus) m++;
	
	tus = m * us;
	
	time->s = (time->s - ts) + tus / 1000;
	time->us = (tus % 1000)*1000;
    } else {
	if (time->us) {
	    ts = time->s + 1;
	    time->us = 0;
	}  else ts = time->s;
	
	m = ts / duration->s;
	if ( m * duration->s < ts) m++;
	time->s = m * duration->s;
    }
    
    return 0;
}

#ifdef _WIN32
#define SECS_BETWEEN_EPOCHS 11644473600LL
#define SECS_TO_100NS 10000000LL

time_t dsTimeConvert_WIN32_FILETIME_to_UnixTime(FILETIME FileTime) {
    unsigned long long tmp;
    
    tmp = ((unsigned long long)FileTime.dwHighDateTime << 32) + FileTime.dwLowDateTime;
    tmp -= SECS_BETWEEN_EPOCHS * SECS_TO_100NS;
    tmp /= SECS_TO_100NS;
    
    return (time_t)tmp;
}

# ifndef HAVE_GETTIMEOFDAY
#  define TIME_FACTOR (0x19db1ded53e8000LL)
static double timer_freq = 0;
static LARGE_INTEGER primed_pc;
static LARGE_INTEGER primed_ft;

static int dsTimeWinInit() {
    LARGE_INTEGER ifreq;
    FILETIME f;
    int priority;

    if (!QueryPerformanceFrequency (&ifreq)) return DS_ERR_WINAPI;

    priority = GetThreadPriority (GetCurrentThread ());
    SetThreadPriority (GetCurrentThread (), THREAD_PRIORITY_TIME_CRITICAL);
    
    if (!QueryPerformanceCounter (&primed_pc)) {
      SetThreadPriority (GetCurrentThread (), priority);
      return DS_ERR_WINAPI;
    }
    GetSystemTimeAsFileTime (&f);
    SetThreadPriority (GetCurrentThread (), priority);

    primed_ft.HighPart = f.dwHighDateTime;
    primed_ft.LowPart = f.dwLowDateTime;
    primed_ft.QuadPart -= TIME_FACTOR;
    primed_ft.QuadPart /= 10;
    timer_freq = (double) ((double) 1000000. / (double) ifreq.QuadPart);

    return 0;
}

int gettimeofday(struct timeval *tv, void *fake) {
    unsigned long long tmp;
    FILETIME FileTime;
#  ifdef USE_WIN32_HIGH_PRECISION_TIME
    LARGE_INTEGER now;
    LONGLONG res;
    
	/* DS: How will affect NTP correction? Perediocally Reinit */
    if ((!timer_freq)&&(dsTimeWinInit())) goto low_precision;
    if (!QueryPerformanceCounter (&now)) goto low_precision;
    res = (LONGLONG) (timer_freq * (double) (now.QuadPart - primed_pc.QuadPart)) + primed_ft.QuadPart;

    tv->tv_sec = (time_t)(res / 1000000);
    tv->tv_usec = (unsigned long)(res % 1000000);

    return 0;
low_precision:
#  endif /* USE_WIN32_HIGH_PRECISION_TIME */
    GetSystemTimeAsFileTime (&FileTime);

    tmp = ((unsigned long long)FileTime.dwHighDateTime << 32) + FileTime.dwLowDateTime;
    tv->tv_usec = (unsigned long)((tmp / 10)%1000000);
    tv->tv_sec = (unsigned long)((tmp - SECS_BETWEEN_EPOCHS * SECS_TO_100NS) / SECS_TO_100NS);

    return 0;
}
# endif /* HAVE_GETTIMEOFDAY */
#endif /* _WIN32 */

int dsDurationAddSeconds(DSDuration duration, time_t s);
int dsDurationAddMilliseconds(DSDuration duration, unsigned long ms);

int dsWaitSeconds(time_t s) {
    DSDurationS duration;
    dsDurationImportSeconds(&duration, s);
    return dsWait(&duration);
}

int dsWaitMilliseconds(unsigned long ms) {
    DSDurationS duration;
    dsDurationImportMilliseconds(&duration, ms);
    return dsWait(&duration);
}

int dsWaitMicroseconds(unsigned long us) {
    DSDurationS duration;
    dsDurationImportMicroseconds(&duration, us);
    return dsWait(&duration);
}

int dsWaitMicrosecondsRT(unsigned long us) {
    DSDurationS duration;
    dsDurationImportMicroseconds(&duration, us);
    return dsWaitRT(&duration);
}
