/* Huge speed optimization posible:
	Replace corelation matrix on square of correlation matrix, it will
	remove need of cpu-consuming sqrt function
*/

#include <stdio.h>
#include <math.h>

#include <dslib.h>
#include <dsassert.h>

#include <dsdispersion.h>

DSDispersionContext dsDispersionCreateContext(DSMallocContext mctx) {
    DSDispersionContext disp;
    
    disp = (DSDispersionContext)dsMalloc(mctx, sizeof(DSDispersionContextS));
    if (disp) {
	disp->malloc = mctx;
    }
    return disp;
}

int dsDispersionInitContext(DSDispersionContext disp) {
    ds_assert_param(disp);
    
    dsDispersionClear(disp);
    return 0;
}

void dsDispersionFreeContext(DSDispersionContext disp) {
}

void dsDispersionDestroyContext(DSDispersionContext disp) {
    ds_assert_return(disp);
    
    dsFree(disp->malloc, disp);
}

int dsDispersionEvent(DSDispersionContext disp, double value) {
    unsigned long n;
    
    ds_assert_param(disp);
    
    n = (++disp->events);
    disp->m += value;

    if (n>1)
	disp->d += (disp->m - value*n)*(disp->m - value*n) / (n * (n - 1));

    return 0;    
}

int dsDispersionFinalyze(DSDispersionContext disp) {
    unsigned long n;

    ds_assert_param(disp);
    
    n = disp->events;
    if (n > 0) {
	disp->m /= n;
	if (n > 1) {
	    disp->d = sqrt(disp->d / (n - 1));
	}
    }
    disp->events = 1;
    
    return 0;
}

int dsDispersionClear(DSDispersionContext disp) {
    ds_assert_param(disp);
    
    disp->events = 0;
    disp->m = 0; 
    disp->d = 0;

    return 0;
}


double dsDispersionGetMedian(DSDispersionContext disp) {
    ds_assert_param(disp);
    
    if (disp->events > 1) return disp->m / disp->events;
    return disp->m;
}

double dsDispersionGetDispersion(DSDispersionContext disp) {
    ds_assert_param(disp);
    
    if (disp->events > 1) return sqrt(disp->d / (disp->events - 1));
    return disp->d;
}


void dsCorrelationZeroContext(DSCorrelationContext corr) {
    corr->corr = NULL;
    corr->s = NULL;
}

DSCorrelationContext dsCorrelationCreateContext(DSMallocContext mctx) {
    DSCorrelationContext corr;
    
    corr = (DSCorrelationContext)dsMalloc(mctx, sizeof(DSCorrelationContextS));
    if (corr) {
	corr->malloc = mctx;
	dsCorrelationZeroContext(corr);
    }
    return corr;
}

int dsCorrelationInitContext(DSCorrelationContext corr, unsigned int elements) {
    ds_assert_param(corr);
    
    corr->n = elements;
    corr->corr = (double*)dsMalloc(corr->malloc, sizeof(double)*(corr->n*(corr->n+1)/2));
    corr->s = (double*)dsMalloc(corr->malloc, sizeof(double)*corr->n);
    if ((!corr->corr)||(!corr->s)) {
	dsCorrelationFreeContext(corr);
	return DS_ERR_MEM;
    }

    dsCorrelationClear(corr);

    return 0;
}

void dsCorrelationFreeContext(DSCorrelationContext corr) {
    ds_assert_return(corr);
    
    if (corr->corr) dsFree(corr->malloc, corr->corr);
    if (corr->s) dsFree(corr->malloc, corr->s);
    dsCorrelationZeroContext(corr);
}

void dsCorrelationDestroyContext(DSCorrelationContext corr) {
    ds_assert_return(corr);
    
    dsCorrelationFreeContext(corr);
    dsFree(corr->malloc, corr);
}


int dsCorrelationComplexEvent(DSCorrelationContext corr, DSCorrelationGetValueFunction get, void *data) {
	unsigned int n,i,j,k;
	unsigned long t;
	unsigned long nd;
	double *v;
	double *d, *s;
	
	ds_assert_param(corr);
	ds_assert_param(get);
	
	n = (++corr->events);
	t = n * (n - 1);

	nd = corr->n;

	s = corr->s;
	d = corr->corr;

	v = (double*)dsMalloca(corr->n*sizeof(double));
	if (!v) return DS_ERR_MEM;

	for (i=0;i<nd;i++) {
	    v[i] = get(data, i);
	    s[i] += v[i];
	    v[i] *= n;
	}
	
	if (!t) {
	    dsFreea(v);
	    return 0;
	}

	for (i=0;i<nd;i++) {
	    d[i] += (double)(((s[i]-v[i])*(s[i]-v[i])))/t;
	}
	
	for (i=0,k=nd;i<nd;i++) {
		for (j=i+1;j<nd;j++,k++) {
		    d[k]+=(double)(int)(((s[i]-v[i])*(s[j]-v[j])))/t;
		}
		v[i]=0;
	}
	
	dsFreea(v);

	return 0;
}

int dsCorrelationEvent(DSCorrelationContext corr, DSCorrelationDataType type, void *data) {
	unsigned int n,i,j,k;
	unsigned long t;
	unsigned long nd;
	double *v;
	double *d, *s;
	
	ds_assert_param(corr);
	ds_assert_param(data);
	
	
	n = (++corr->events);
	t = n * (n - 1);

	nd = corr->n;

	s = corr->s;
	d = corr->corr;

	v = (double*)dsMalloca(corr->n*sizeof(double));
	if (!v) return DS_ERR_MEM;
	
	switch (type) {
	    case DS_CORRELATION_INT8:
		for (i=0;i<nd;i++) {
		    s[i] += ((int8_t*)data)[i];
	    	    v[i] = ((int8_t*)data)[i] * n;
		}
	    break;
	    case DS_CORRELATION_UINT8:
		for (i=0;i<nd;i++) {
		    s[i] += ((uint8_t*)data)[i];
	    	    v[i] = ((uint8_t*)data)[i] * n;
		}
	    break;
	    case DS_CORRELATION_INT16:
		for (i=0;i<nd;i++) {
		    s[i] += ((int16_t*)data)[i];
	    	    v[i] = ((int16_t*)data)[i] * n;
		}
	    break;
	    case DS_CORRELATION_UINT16:
		for (i=0;i<nd;i++) {
		    s[i] += ((uint16_t*)data)[i];
	    	    v[i] = ((uint16_t*)data)[i] * n;
		}
	    break;
	    case DS_CORRELATION_INT32:
		for (i=0;i<nd;i++) {
		    s[i] += ((int32_t*)data)[i];
	    	    v[i] = ((int32_t*)data)[i] * n;
		}
	    break;
	    case DS_CORRELATION_UINT32:
		for (i=0;i<nd;i++) {
		    s[i] += ((uint32_t*)data)[i];
	    	    v[i] = ((uint32_t*)data)[i] * n;
		}
	    break;
	    case DS_CORRELATION_INT64:
		for (i=0;i<nd;i++) {
		    s[i] += (double)((int64_t*)data)[i];
	    	    v[i] = (double)((int64_t*)data)[i] * n;
		}
	    break;
	    case DS_CORRELATION_UINT64:
		for (i=0;i<nd;i++) {
		    s[i] += (double)((uint64_t*)data)[i];
	    	    v[i] = (double)((uint64_t*)data)[i] * n;
		}
	    break;
	    case DS_CORRELATION_FLOAT:
		for (i=0;i<nd;i++) {
		    s[i] += ((float*)data)[i];
	    	    v[i] = ((float*)data)[i] * n;
		}
	    break;
	    case DS_CORRELATION_DOUBLE:
		for (i=0;i<nd;i++) {
		    s[i] += ((double*)data)[i];
	    	    v[i] = ((double*)data)[i] * n;
		}
	    break;
	    case DS_CORRELATION_LONGDOUBLE:
		for (i=0;i<nd;i++) {
		    s[i] += ((long double*)data)[i];
	    	    v[i] = ((long double*)data)[i] * n;
		}
	    break;
	    default:
		dsFreea(v);
		return DS_ERR_INVALID_OPTIONS;
	}

	if (!t) {
	    dsFreea(v);
	    return 0;
	}

	for (i=0;i<nd;i++) {
	    d[i] += (double)(((s[i]-v[i])*(s[i]-v[i])))/t;
	}
	
	for (i=0,k=nd;i<nd;i++) {
		for (j=i+1;j<nd;j++,k++) {
		    d[k]+=(double)(int)(((s[i]-v[i])*(s[j]-v[j])))/t;
		}
		v[i]=0;
	}
	
	dsFreea(v);

	return 0;
}

int dsCorrelationFinalyze(DSCorrelationContext corr) {
    unsigned long n, nd;
    unsigned int i,j,k;
    double *d, *s;

    ds_assert_param(corr);

    n = corr->events;
    d = corr->corr;
    s = corr->s;
    nd = corr->n;
    
    if (n > 1) {
	for (i=0;i<nd;i++) {
	    s[i]/=n;
	    d[i]/=(n-1);
	    d[i]=sqrt(d[i]);
	}
    
	for (i=0,k=nd;i<nd;i++) {
	    for (j=i+1;j<nd;j++,k++) {
		if (d[i]&&d[j]) d[k]/=((n-1)*d[i]*d[j]);
	    }
	}

	corr->events = 1;
    }
    
    
    return 0;
}

int dsCorrelationClear(DSCorrelationContext corr) {
    int i;
    double *d, *s;

    ds_assert_param(corr);    

    corr->events = 0;

    s = corr->s;
    for (i=corr->n-1;i>=0;i--) s[i]=0;
    
    d = corr->corr;
    for (i=corr->n*(corr->n+1)/2-1;i>=0;i--) d[i]=0;

    return 0;
}


double dsCorrelationGetMedian(DSCorrelationContext corr, unsigned int i) {
    ds_assert_default(corr, 0);
    ds_assert_default(i<corr->n, 0);
    
    if (corr->events > 1) return corr->s[i] / corr->events;
    return corr->s[i];
}

double dsCorrelationGetDispersion(DSCorrelationContext corr, unsigned int i) {
    ds_assert_default(corr, 0);
    ds_assert_default(i<corr->n, 0);
    
    if (corr->events > 1) return sqrt(corr->corr[i] / (corr->events - 1));
    return corr->corr[i];
}

double dsCorrelationGetCorrelation(DSCorrelationContext corr, unsigned int i, unsigned int j) {
    double di, dj;
    unsigned int x, y, pos;
    
    ds_assert_default(corr, 0);
    ds_assert_default(i<corr->n, 0);
    ds_assert_default(j<corr->n, 0);
    
    if (i > j) {
	x = i;
	y = j;
    } else if (i < j) {
	x = j;
	y = i;
    } else return dsCorrelationGetDispersion(corr, i);
    
    pos = corr->n + (y*corr->n - y*(y+1)/2) + (x - y - 1);
    
    if (corr->events > 1) {
	di = dsCorrelationGetDispersion(corr, i);
	dj = dsCorrelationGetDispersion(corr, j);
	if (di&&dj) return corr->corr[pos]/((corr->events-1)*di*dj);
    }
    return corr->corr[pos];
}


double *dsCorrelationGetMedianArray(DSCorrelationContext corr) {
    ds_assert_null(corr);
    
    return corr->s;
}

double *dsCorrelationGetDispersionArray(DSCorrelationContext corr) {
    ds_assert_null(corr);
    
    return corr->corr;
}

double *dsCorrelationGetCorrelationArray(DSCorrelationContext corr) {
    ds_assert_null(corr);
    
    return corr->corr + corr->n;
}

