#include <dsassert.h>

#include <dslib.h>

#include <hash-string.h>

#include <dsmalloc.h>

#include "../3rd_party/prime.h"

#include <dsdoublehash.h>

static inline_function unsigned long DoubleHash1(DSDoubleHashContext ctx, unsigned long k) {
	return k%ctx->size;
}

static inline_function unsigned long DoubleHash2(DSDoubleHashContext ctx, unsigned long k) {
	return 1+k%(ctx->size-1);
}

#define find_prime findUpperPrime

static void dsDoubleHashZeroContext(DSDoubleHashContext ctx) {
    ds_assert_return(ctx);
    ctx->buffer = NULL;
    ctx->max_size = 0;
}

DSDoubleHashContext dsDoubleHashCreateContext(DSMallocContext mctx) {
    DSDoubleHashContext ctx;
    ctx = (DSDoubleHashContext)dsMalloc(mctx, sizeof(struct DSDoubleHashContextT));
    if (ctx) {
	ctx->mctx = mctx;
	dsDoubleHashZeroContext(ctx);
    }
    return ctx;
}

int dsDoubleHashInitContext(DSDoubleHashContext ctx, unsigned long size, unsigned long step_size, unsigned char free_on_clear, void (*cfunc)(void *data)) {
    unsigned long i;
    
    ds_assert_param(ctx);
    
    dsDoubleHashZeroContext(ctx);

    ctx->gift = free_on_clear;
    ctx->cfunc = cfunc;

    ctx->size = find_prime(size);
    ctx->step_size = step_size;
    
    ctx->buffer = (void**)dsMalloc(ctx->mctx,ctx->size*sizeof(void*));
    if (!ctx->buffer) return 1;
    for (i=0;i<ctx->size;i++) ctx->buffer[i]=NULL;
    
    return 0;
}

void dsDoubleHashFreeContext(DSDoubleHashContext ctx) {
    unsigned long i;
    
    ds_assert_return(ctx);
    
    if (ctx->buffer) {
	for (i=0;i<ctx->size;i++)
	    if (ctx->buffer[i]) dsDoubleHashRemove(ctx, i);
	dsFree(ctx->mctx, ctx->buffer);
    }

    dsDoubleHashZeroContext(ctx);
}

void dsDoubleHashDestroyContext(DSDoubleHashContext ctx) {
    ds_assert_return(ctx);
    dsDoubleHashFreeContext(ctx);
    dsFree(ctx->mctx, ctx);
}

unsigned long dsDoubleHashInsert(DSDoubleHashContext ctx, void *data, unsigned long hash) {
    unsigned long i,pos;
    unsigned long size;
    void **buffer;
    
    ds_assert_err(ctx, (unsigned long)-1);
    ds_assert_err(ctx->buffer, (unsigned long)-1);
    
    for (i=0;i<ctx->size;i++) {
	pos=(DoubleHash1(ctx,hash)+i*DoubleHash2(ctx,hash))%ctx->size;
	if (!ctx->buffer[pos]) break;
    }
    
    if (i==ctx->size) {
	size = find_prime(ctx->size + ctx->step_size);
	if ((ctx->max_size)&&(size > ctx->max_size)) return (unsigned long)-2;
	
	buffer=(void**)dsRealloc(ctx->mctx,ctx->buffer,size*sizeof(void*));
	if (!buffer) {
	    ctx->buffer=buffer;
	    return (unsigned long)-1;
	}
	for (i=ctx->size+1;i<size;i++) ctx->buffer[i]=NULL;

	pos = ctx->size;

	ctx->buffer = buffer;
	ctx->size = size;
    }
    
    ctx->buffer[pos]=data;
    return pos;
}

unsigned long dsDoubleHashInsertS(DSDoubleHashContext ctx, void *data, const char *str) {
    unsigned long hash;
    
    ds_assert_err(str, (unsigned long)-1);

    hash = hash_string(str);
    return dsDoubleHashInsert(ctx,data,hash);
}

void dsDoubleHashRemove(DSDoubleHashContext ctx, unsigned long pos) {
    ds_assert_return(ctx);
    ds_assert_return(ctx->buffer);
    ds_assert_return(pos<ctx->size);
    
    if (ctx->gift) {
	if (ctx->cfunc) ctx->cfunc(ctx->buffer[pos]);
	else dsFree(ctx->mctx, ctx->buffer[pos]);
    } 
    ctx->buffer[pos]=NULL;
}

void dsDoubleHashUnlink(DSDoubleHashContext ctx, unsigned long pos) {
    ds_assert_return(ctx);
    ds_assert_return(ctx->buffer);
    ds_assert_return(pos<ctx->size);
    
    ctx->buffer[pos]=NULL;    
}

const void *dsDoubleHashGetBuffer(DSDoubleHashContext ctx) {
    ds_assert_null(ctx);

    return ctx->buffer;
}

const void *dsDoubleHashGetLink(DSDoubleHashContext ctx, unsigned long pos) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->buffer);
    ds_assert_null(pos<ctx->size);

    return ctx->buffer[pos];
}

void *dsDoubleHashGet(DSDoubleHashContext ctx, unsigned long pos) {
    void *item;
    
    ds_assert_null(ctx);
    ds_assert_null(ctx->buffer);
    ds_assert_null(pos<ctx->size);

    item = ctx->buffer[pos];
    ctx->buffer[pos]=NULL;
    
    return item;
}

int dsDoubleHashSetMaximumSize(DSDoubleHashContext ctx, unsigned long max_size) {
    ds_assert_param(ctx);
    
    ctx->max_size=max_size;
    return 0;
}

unsigned long dsDoubleHashClean(DSDoubleHashContext ctx, DSDoubleHashCleanFunction clean_func, void *attr) {
    int res;
    unsigned long pos;
    unsigned long size;
    unsigned long cleaned=0;

    ds_assert_err(ctx, (unsigned long)-1);
    
    size = ctx->size;
    for (pos=0;pos<size;pos++) {
	if (ctx->buffer[pos]) {
	    res=clean_func(ctx->buffer[pos], attr);
	    if (res) {
		if ((res>1)&&(ctx->gift)) {
		    if (ctx->cfunc) ctx->cfunc(ctx->buffer[pos]);
		    else dsFree(ctx->mctx, ctx->buffer[pos]);
		} 
		ctx->buffer[pos]=NULL;
		cleaned++;
	    }
	}
    }
    return cleaned;
}
