#include <dsassert.h>

#include <dslib.h>
#include <dsenumeration.h>

void dsEnumerationZeroContext(DSEnumerationContext ctx) {
    ctx->hash = NULL;
    ctx->destroy_bucket = 0;
    ctx->string = NULL;
}

DSEnumerationContext dsEnumerationCreateContext(DSMallocContext mctx) {
    DSEnumerationContext ctx;

    ctx = (DSEnumerationContext)dsMalloc(mctx, sizeof(struct DSEnumerationContextT));
    if (ctx) {
	ctx->malloc = mctx;
	dsEnumerationZeroContext(ctx);
    }
    return ctx;
}

int dsEnumerationInitContext(DSEnumerationContext ctx, DSEnumerationBucket bucket) {
    int err = 0;
    unsigned long i;

    ds_assert_param(ctx);
    ds_assert_param(bucket);

    dsEnumerationZeroContext(ctx);
    
    for (i=0;bucket[i].string;i++);
    ctx->maxid=i;
    
    ctx->hash = dsHashTableInit(ctx->malloc, (unsigned long)(1.5*i), DS_HASH_TYPE_POINTER, NULL);
    if (!ctx->hash) return DS_ERR_MEM;
    
    for (i=0;bucket[i].string;i++)
	err+=dsHashTableInsert(ctx->hash, &bucket[i]);

    if (err) {
	dsEnumerationFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    return 0;
}

int dsEnumerationInitContextFromArray(DSEnumerationContext ctx, const char **title_array, unsigned char copy_strings) {
    int err = 0;
    DSEnumerationBucket bucket;
    unsigned long i;
    unsigned long size;

    ds_assert_param(ctx);
    ds_assert_param(title_array);

    dsEnumerationZeroContext(ctx);
    
    for (size=0;title_array[size];size++);
    ctx->maxid=size;
    
    if (copy_strings) 
	ctx->hash = dsHashTableInit(ctx->malloc, (unsigned long)(1.5*size), DS_HASH_TYPE_POINTER, NULL);
    else
	ctx->hash = dsHashTableInit(ctx->malloc, (unsigned long)(1.5*size), DS_HASH_TYPE_STATICPOINTER, NULL);

    if (!ctx->hash) return DS_ERR_MEM;
    
    bucket = (DSEnumerationBucket)dsMalloc(ctx->malloc, size*sizeof(struct DSEnumerationBucketT));
    if (!bucket) {
	dsEnumerationFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    ctx->bucket = bucket;
    ctx->destroy_bucket = 1;
    
    for (i=0;i<size;i++) {
	bucket[i].id = i;
	if (copy_strings) {
	    bucket[i].string = NULL;
	    dsExchangeString(ctx->malloc, &bucket[i].string, title_array[i]);
	} else bucket[i].string = (char*)title_array[i];
	err+=dsHashTableInsert(ctx->hash, &bucket[i]);
    }

    if (err) {
	dsEnumerationFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    return 0;    
}

int dsEnumerationInitContextFromString(DSEnumerationContext ctx, const char *titles, unsigned char copy_string) {
    int err;
    DSEnumerationBucket bucket;
    unsigned long i,size;
    char *str = NULL;
    
    ds_assert_param(ctx);
    ds_assert_param(titles);

    dsEnumerationZeroContext(ctx);
    
    if (copy_string) {
	err = dsExchangeString(ctx->malloc,&str,titles);
	if (err) return DS_ERR_MEM;
	ctx->string = str; 
    } else {
	str = (char*)titles;
    }
    
    if ((str[0]==',')||(str[0]==0)) {
	dsEnumerationFreeContext(ctx);
	return DS_ERR_ENUMERATION_INVALID_DESCRIPTION;
    }
    
    for(size=1,i=1;str[i];i++) 
	if (str[i]==',') {
	    if (str[i-1]==',') break;
	    size++;
	} else if ((str[i]=='<')||(str[i]=='&')||(str[i]=='"')) break;
    
    if ((str[i])||(str[i-1]==',')) {
	dsEnumerationFreeContext(ctx);
	return DS_ERR_ENUMERATION_INVALID_DESCRIPTION;
    }
    
    ctx->maxid = size;


    ctx->hash = dsHashTableInit(ctx->malloc, (unsigned long)(1.5*size), DS_HASH_TYPE_STATICPOINTER, NULL);
    if (!ctx->hash) return DS_ERR_MEM;
    dsHashTableSetDestructor(ctx->hash, NULL);

    bucket = (DSEnumerationBucket)dsMalloc(ctx->malloc, size*sizeof(struct DSEnumerationBucketT));
    if (!bucket) {
	dsEnumerationFreeContext(ctx);
	return DS_ERR_MEM;
    }

    ctx->bucket = bucket;
    ctx->destroy_bucket = 1;
    
    bucket[0].string = str;
    bucket[0].id = 0;
    
    for (size=1,i=1;str[i];i++) {
	if (str[i]==',') {
	    bucket[size].id=size;
	    bucket[size++].string=str+i+1;
	    str[i]=0;
	}
    }

    return 0;
}

void dsEnumerationFreeContext(DSEnumerationContext ctx) {
    ds_assert_return(ctx);
    if (ctx->hash) dsHashTableDestroy(ctx->hash);
    if ((ctx->destroy_bucket)&&(ctx->bucket)) dsFree(ctx->malloc, ctx->bucket);
    if (ctx->string) dsExchangeString(ctx->malloc, &ctx->string, NULL);
    dsEnumerationZeroContext(ctx);
}

void dsEnumerationDestroyContext(DSEnumerationContext ctx) {
    dsEnumerationFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

unsigned long dsEnumerationGetID(DSEnumerationContext ctx, const char *str) {
    DSEnumerationBucket bucket;
    
    ds_assert_err(ctx, (unsigned long)-1);
    ds_assert_err(str, (unsigned long)-1);
    ds_assert_err(ctx->hash, (unsigned long)-1);
    
    bucket = dsHashTableFind(ctx->hash,str);
    if (!bucket) return (unsigned long)-1;
    return bucket->id;
    
}

const char *dsEnumerationGetString(DSEnumerationContext ctx, unsigned long number) {
    ds_assert_null(ctx);
    ds_assert_null(number<ctx->maxid);
    ds_assert_null(ctx->bucket);
    
    return ctx->bucket[number].string;
}

DSEnumerationBucket dsEnumerationGetBucketByString(DSEnumerationContext ctx, const char *str) {
    ds_assert_null(ctx);
    ds_assert_null(str);
    ds_assert_null(ctx->hash);
    
    return (DSEnumerationBucket)dsHashTableFind(ctx->hash,str);
}

DSEnumerationBucket dsEnumerationGetBucketByID(DSEnumerationContext ctx, unsigned long number) {
    ds_assert_null(ctx);
    ds_assert_null(number<ctx->maxid);
    ds_assert_null(ctx->bucket);
    
    return ctx->bucket+number;
}
