#include <stdio.h>
#include <string.h>
#include <dsassert.h>

#include <dslib.h>
#include <dsmalloc.h>
#include <dshashtable.h>
#include <hash-string.h>


#define mod1(x) ((x) - (long)(x))
#define ROUND(x) (unsigned long)(x)

static char *fake="Fake";

static const char *HashTableDefaultPointer(const DSHashStringPtr ctx) {
    return ctx->str;
}

static const char *HashTableDefaultArray(const DSHashStringArray ctx) {
    return ctx->str;
}

static unsigned long HashTableDefaultHash(const DSHashStringHash ctx) {
    return ctx->hash;
}

char *dsHashTableFake(const void *ctx) {
    return fake;
}

DSHashTable dsHashTableInit(DSMallocContext mctx, unsigned long size, DSHashType hash_type, DSHashTableStringFunction string_func) {
    DSHashTable ctx;
    if (hash_type>=DS_HASH_TYPE_LAST) return NULL;
    if (((hash_type==DS_HASH_TYPE_FUNCTION)||(hash_type==DS_HASH_TYPE_HASH_FUNCTION))&&(!string_func)) return NULL;
    
    ctx = (DSHashTable)dsMalloc(mctx, sizeof(struct DSHashTableT));
    if (ctx) {
	ctx->malloc=mctx;
	ctx->size=size;
	ctx->mode=hash_type;
	ctx->gift=0;
	ctx->cfunc=NULL;
	ctx->prefunc=NULL;
	ctx->postfunc=NULL;
	ctx->ppfuncattr=NULL;
	ctx->string_size=0;
	switch(hash_type) {
	    case DS_HASH_TYPE_HASH:
		ctx->string_func=(DSHashTableStringFunction)&HashTableDefaultHash;
	    break;
	    case DS_HASH_TYPE_ARRAY:
		ctx->string_func=(DSHashTableStringFunction)&HashTableDefaultArray;
	    break;
	    case DS_HASH_TYPE_POINTER:
	    case DS_HASH_TYPE_STATICPOINTER:
	    	ctx->string_func=(DSHashTableStringFunction)&HashTableDefaultPointer;
	    break;
	    default:
		ctx->string_func=string_func;
	}
	ctx->bucket=(DSHashTableBucket*)dsCalloc(mctx,size,sizeof(DSHashTableBucket));
	if (!ctx->bucket) {
	    dsFree(mctx,ctx);
	    return NULL;
	}
    }
    return ctx;
}

DSHashTable dsHashTableHashInit(DSMallocContext mctx, unsigned long size, DSHashType hash_type, DSHashTableHashFunction hash_func) {
    if (hash_type<DS_HASH_TYPE_HASH) return NULL;
    return dsHashTableInit(mctx,size,hash_type, (DSHashTableStringFunction)hash_func);
}


int dsHashTableSetDestructor(DSHashTable ctx, DSHashTableDestructorFunction cfunc) {
    ds_assert_param(ctx);
    ctx->gift = 1;
    ctx->cfunc = cfunc;
    return 0;
}

int dsHashTableSetPPDFunctions(DSHashTable ctx, DSHashTablePPDFunction prefunc, DSHashTablePPDFunction postfunc, void *attr) {
    ds_assert_param(ctx);
    
    ctx->prefunc=prefunc;
    ctx->postfunc=postfunc;
    ctx->ppfuncattr=attr;
    
    return 0;
}

int dsHashTableSetStringSize(DSHashTable ctx, unsigned long size) {
    ds_assert_param(ctx);
    
    ctx->string_size = size;
    return 0;
}

void dsHashTableFree(DSHashTable ctx) {
    DSHashTableBucket bucket;
    DSHashTableBucket tmpbucket;
    unsigned long i;

    ds_assert_return(ctx);

    if (ctx->prefunc) ctx->prefunc(ctx,DS_HASHTABLE_DESTRUCTIONMODE_ALL,ctx->ppfuncattr);

    for (i=0;i<ctx->size;i++) {
	for (bucket=ctx->bucket[i],ctx->bucket[i]=NULL;bucket;) {
	    if (ctx->gift) {
		if (ctx->cfunc) {
		    ctx->cfunc(bucket->data);
		} else {
		    if (ctx->mode==DS_HASH_TYPE_POINTER)
			dsFree(ctx->malloc, ((DSHashStringPtr)(bucket->data))->str);
		    dsFree(ctx->malloc, bucket->data);
		}
	    }    
	    tmpbucket=bucket;
	    bucket=bucket->next;
	    dsFree(ctx->malloc, tmpbucket);
	}
    }
    dsFree(ctx->malloc, ctx->bucket);

    if (ctx->postfunc) ctx->postfunc(ctx,DS_HASHTABLE_DESTRUCTIONMODE_ALL,ctx->ppfuncattr);
}

void dsHashTableDestroy(DSHashTable ctx) {
    dsHashTableFree(ctx);
    dsFree(ctx->malloc, ctx);
}

int dsHashTableInsert(DSHashTable ctx, void *data) {
    unsigned long pos;
    DSHashTableBucket bucket;
    
    ds_assert_param(ctx);
    ds_assert_param(data);
    
    bucket = (DSHashTableBucket)dsMalloc(ctx->malloc, sizeof(struct DSHashTableBucketT));
    if (!bucket) return DS_ERR_MEM;

    if (ctx->mode<DS_HASH_TYPE_HASH) {
	bucket->string = ctx->string_func(data);
	bucket->hash=hash_nstring(bucket->string, ctx->string_size);
    } else bucket->hash=((DSHashTableHashFunction)(ctx->string_func))(data);

    pos = ROUND(ctx->size * mod1( hash_table_coef * bucket->hash ));
    bucket->next = ctx->bucket[pos];
    bucket->data = data;
    ctx->bucket[pos]=bucket;
    return 0;
}

int dsHashTableHashInsert(DSHashTable ctx, unsigned long hash, void *data) {
    unsigned long pos;
    DSHashTableBucket bucket;
    
    ds_assert_param(ctx);
    ds_assert_param(data);
    
    bucket = (DSHashTableBucket)dsMalloc(ctx->malloc, sizeof(struct DSHashTableBucketT));
    if (!bucket) return DS_ERR_MEM;

    if (ctx->mode<DS_HASH_TYPE_HASH) {
	bucket->string = ctx->string_func(data);
    }
    bucket->hash = hash;
    
    pos = ROUND(ctx->size * mod1( hash_table_coef * bucket->hash ));
    bucket->next = ctx->bucket[pos];
    bucket->data = data;
    ctx->bucket[pos]=bucket;
    return 0;
}

int dsHashTableHashStringInsert(DSHashTable ctx, unsigned long hash, char *str, void *data) {
    unsigned long pos;
    DSHashTableBucket bucket;
    
    ds_assert_param(ctx);
    ds_assert_param(data);
    
    if (!str) return dsHashTableHashInsert(ctx,hash,data);
    ds_assert_param(ctx->mode<DS_HASH_TYPE_ARRAY);
    

    if (ctx->mode == DS_HASH_TYPE_POINTER) {
	ds_assert_param(ctx->cfunc);

	bucket = (DSHashTableBucket)dsMalloc(ctx->malloc, sizeof(struct DSHashTableBucketT));
	if (!bucket) return DS_ERR_MEM;
	
	if (ctx->string_size) {
	    bucket->string = (char*)dsMalloc(ctx->malloc, ctx->string_size);
	    if (!bucket->string) return DS_ERR_MEM;
	    memcpy(bucket->string, str, ctx->string_size);
	} else {
	    bucket->string = (char*)dsMalloc(ctx->malloc, (strlen(str)+1)*sizeof(char));
	    if (!bucket->string) return DS_ERR_MEM;
	    strcpy(bucket->string,str);
	}
	
	dsFree(ctx->malloc, ((DSHashStringPtr)(data))->str);
    } else {
	bucket = (DSHashTableBucket)dsMalloc(ctx->malloc, sizeof(struct DSHashTableBucketT));
	if (!bucket) return DS_ERR_MEM;

	bucket->string = str;
    }
    bucket->hash = hash;

    pos = ROUND(ctx->size * mod1( hash_table_coef * bucket->hash ));
    bucket->next = ctx->bucket[pos];
    bucket->data = data;
    ctx->bucket[pos]=bucket;
    return 0;
}

/* Inserts new item into the hash. If already exists replaces that or reports
failure depending on the report parameter */
int dsHashTableAdd(DSHashTable ctx, void *data, unsigned char report) {
    int err;
    unsigned long hash;
    unsigned long pos;
    DSHashTableBucket bucket;
    char *str;
    
    ds_assert_param(ctx);
    ds_assert_param(data);

    if (ctx->mode<DS_HASH_TYPE_HASH) {
	str = ctx->string_func(data);
	hash = hash_nstring(str, ctx->string_size);
    } else {
	str = NULL;
	hash = ((DSHashTableHashFunction)(ctx->string_func))(data);
    }
    
    pos = ROUND(ctx->size * mod1( hash_table_coef * hash ));
    for (bucket = ctx->bucket[pos];bucket;bucket=bucket->next) {
	if (bucket->hash==hash) {
	    if (ctx->mode<DS_HASH_TYPE_HASH) {
		if (ctx->string_size) {
		    if (!memcmp(str,bucket->string,ctx->string_size)) err = 1;
		    else err = 0;
		} else {
		    if (!strcmp(str,bucket->string)) err = 1;
		    else err = 0;
		}
	    } else err = 1;
		
	    if (err) {
		if (report) return DS_ERR_EXIST;
	        if (ctx->gift) {
	    	    if (ctx->cfunc) ctx->cfunc(data);
		    else {
		        if (ctx->mode==DS_HASH_TYPE_POINTER) dsFree(ctx->malloc, (char*)str);
		        dsFree(ctx->malloc, data);
		    }
		}
		break; 	/* DS: Instead of "return 0", July 2006 */
	    }
	}
    }

    bucket = (DSHashTableBucket)dsMalloc(ctx->malloc, sizeof(struct DSHashTableBucketT));
    if (!bucket) return DS_ERR_MEM;

    bucket->hash = hash;
    bucket->string = str;
    bucket->next = ctx->bucket[pos];
    bucket->data = data;
    ctx->bucket[pos]=bucket;
    
    return 0;
}

int dsHashTableHashAdd(DSHashTable ctx, unsigned long hash, void *data, unsigned char report) {
    int err;
    unsigned long pos;
    DSHashTableBucket bucket;
    char *str;
    
    ds_assert_param(ctx);
    ds_assert_param(data);

    if (ctx->mode<DS_HASH_TYPE_HASH) str = ctx->string_func(data);
    else str = NULL;

    pos = ROUND(ctx->size * mod1( hash_table_coef * hash ));
    for (bucket = ctx->bucket[pos];bucket;bucket=bucket->next) {
	if (bucket->hash==hash) {
	    if (ctx->mode<DS_HASH_TYPE_HASH) {
		if (ctx->string_size) {
		    if (!memcmp(str,bucket->string,ctx->string_size)) err = 1;
		    else err = 0;
		} else {
		    if (!strcmp(str,bucket->string)) err = 1;
		    else err = 0;
		}
	    } else err = 1;
		
	    if (err) {
		if (report) return DS_ERR_EXIST;
	        if (ctx->gift) {
	    	    if (ctx->cfunc) ctx->cfunc(data);
		    else {
		        if (ctx->mode==DS_HASH_TYPE_POINTER) dsFree(ctx->malloc, (char*)str);
		        dsFree(ctx->malloc, data);
		    }
		}
		break; 	/* DS: Instead of "return 0", July 2006 */
	    }
	}
    }

    bucket = (DSHashTableBucket)dsMalloc(ctx->malloc, sizeof(struct DSHashTableBucketT));
    if (!bucket) return 1;

    bucket->hash = hash;
    bucket->string = str;
    bucket->next = ctx->bucket[pos];
    bucket->data = data;
    ctx->bucket[pos]=bucket;
    
    return 0;
}

int dsHashTableHashStringAdd(DSHashTable ctx, unsigned long hash, char *str, void *data, unsigned char report) {
    int err;
    unsigned long pos;
    DSHashTableBucket bucket;
    
    ds_assert_param(ctx);
    ds_assert_param(data);
    
    if (!str) return dsHashTableHashAdd(ctx,hash,data,report);
    ds_assert_param(ctx->mode<DS_HASH_TYPE_ARRAY);
    ds_assert_param((!ctx->cfunc)||(ctx->mode==DS_HASH_TYPE_STATICPOINTER));

    pos = ROUND(ctx->size * mod1( hash_table_coef * hash ));
    for (bucket = ctx->bucket[pos];bucket;bucket=bucket->next) {
	if (bucket->hash==hash) {
	    if (ctx->mode<DS_HASH_TYPE_HASH) {
		if (ctx->string_size) {
		    if (!memcmp(str,bucket->string,ctx->string_size)) err = 1;
		    else err = 0;
		} else {
		    if (!strcmp(str,bucket->string)) err = 1;
		    else err = 0;
		}
	    } else err = 1;
		
	    if (err) {
		if (report) return DS_ERR_EXIST;
	        if (ctx->gift) {
	    	    if (ctx->cfunc) ctx->cfunc(data);
		    else {
		        if (ctx->mode==DS_HASH_TYPE_POINTER) dsFree(ctx->malloc, (char*)str);
		        dsFree(ctx->malloc, data);
		    }
		}
		break; 	/* DS: Instead of "return 0", July 2006 */
	    }
	}
    }

    bucket = (DSHashTableBucket)dsMalloc(ctx->malloc, sizeof(struct DSHashTableBucketT));
    if (!bucket) return DS_ERR_MEM;

    bucket->hash = hash;
    if (ctx->mode == DS_HASH_TYPE_POINTER) {
	if (ctx->string_size) {
	    bucket->string = (char*)malloc(ctx->string_size);
	    if (!bucket->string) return DS_ERR_MEM;
	    memcpy(bucket->string, str, ctx->string_size);
	} else {
	    bucket->string = (char*)malloc((strlen(str)+1)*sizeof(char));
	    if (!bucket->string) return DS_ERR_MEM;
	    strcpy(bucket->string,str);
	}

	dsFree(ctx->malloc, ((DSHashStringPtr)(data))->str);
    } else bucket->string = str;
    bucket->next = ctx->bucket[pos];
    bucket->data = data;
    ctx->bucket[pos]=bucket;
    
    return 0;
}



static int HashTableDeleteMS(DSHashTable ctx, const char *string, unsigned long justhash, unsigned char multi) {
    int err,res=0;
    unsigned long hash;
    unsigned long pos;
    DSHashTableBucket bucket;
    DSHashTableBucket prebucket = NULL;
    
    ds_assert_param(ctx);
    
    if (ctx->mode<DS_HASH_TYPE_HASH) {
	ds_assert_param(string);
	hash = hash_nstring(string, ctx->string_size);
    } else {
	if (string) hash = hash_nstring(string, ctx->string_size);
	else hash = justhash;
    }
    
    pos = ROUND(ctx->size * mod1( hash_table_coef * hash ));
    for (bucket = ctx->bucket[pos];bucket;) {
	if (bucket->hash==hash) {
	    if (ctx->mode<DS_HASH_TYPE_HASH) {
		if (ctx->string_size) {
		    if (!memcmp(string,bucket->string,ctx->string_size)) err = 1;
		    else err = 0;
		} else {
		    if (!strcmp(string,bucket->string)) err = 1;
		    else err = 0;
		}
	    } else err = 1;

	    if (err) {
		if (prebucket) prebucket->next = bucket->next;
		else ctx->bucket[pos]=bucket->next;
		if (ctx->gift) {
		    if (ctx->cfunc) ctx->cfunc(bucket->data);
		    else {
			if (ctx->mode==DS_HASH_TYPE_POINTER) dsFree(ctx->malloc, bucket->string);
			dsFree(ctx->malloc, bucket->data);
		    }
		}
		dsFree(ctx->malloc, bucket);
		if (multi) {
		    if (prebucket) bucket=prebucket->next;
		    else bucket=ctx->bucket[pos];
		    res++;
		    continue;
		} else return DS_ERR_EXIST;
	    }
	}
	prebucket=bucket;
	bucket=bucket->next;
    }
    return res;
}

int dsHashTableDeleteMulti(DSHashTable ctx, const char *string) {
    int res;
    
    ds_assert_param(ctx);
    
    if (ctx->prefunc) ctx->prefunc(ctx,DS_HASHTABLE_DESTRUCTIONMODE_MULTI,ctx->ppfuncattr);
    res=HashTableDeleteMS(ctx,string,0,1);
    if (ctx->postfunc) ctx->postfunc(ctx,DS_HASHTABLE_DESTRUCTIONMODE_MULTI,ctx->ppfuncattr);
    return res;
}

int dsHashTableHashDeleteMulti(DSHashTable ctx, unsigned long hash) {
    int res;
    
    ds_assert_param(ctx);
    
    if (ctx->prefunc) ctx->prefunc(ctx,DS_HASHTABLE_DESTRUCTIONMODE_MULTI,ctx->ppfuncattr);
    res=HashTableDeleteMS(ctx,NULL,hash,1);
    if (ctx->postfunc) ctx->postfunc(ctx,DS_HASHTABLE_DESTRUCTIONMODE_MULTI,ctx->ppfuncattr);
    return res;
}

int dsHashTableDelete(DSHashTable ctx, const char *string) {
    int res;
    
    ds_assert_param(ctx);
    
    if (ctx->prefunc) ctx->prefunc(ctx,DS_HASHTABLE_DESTRUCTIONMODE_SINGLE,ctx->ppfuncattr);
    res=HashTableDeleteMS(ctx,string,0,0);
    if (ctx->postfunc) ctx->postfunc(ctx,DS_HASHTABLE_DESTRUCTIONMODE_SINGLE,ctx->ppfuncattr);
    return res;
}

int dsHashTableHashDelete(DSHashTable ctx, unsigned long hash) {
    int res;
    
    ds_assert_param(ctx);
    
    if (ctx->prefunc) ctx->prefunc(ctx,DS_HASHTABLE_DESTRUCTIONMODE_SINGLE,ctx->ppfuncattr);
    res=HashTableDeleteMS(ctx,NULL,hash,0);
    if (ctx->postfunc) ctx->postfunc(ctx,DS_HASHTABLE_DESTRUCTIONMODE_SINGLE,ctx->ppfuncattr);
    return res;
}



static int HashTableRemoveOrUnlink(DSHashTable ctx, const void *data, unsigned char  remove) {
    int err = 0;
    unsigned long hash;
    unsigned long pos;
    DSHashTableBucket bucket;
    DSHashTableBucket prebucket = NULL;
    const char *str;

    ds_assert_param(ctx);
    ds_assert_param(data);
    
    if (ctx->mode<DS_HASH_TYPE_HASH) {
	str = ctx->string_func(data);
	hash = hash_nstring(str, ctx->string_size);
    } else hash=((DSHashTableHashFunction)(ctx->string_func))(data);

    pos = ROUND(ctx->size * mod1( hash_table_coef * hash ));

    for (bucket = ctx->bucket[pos];bucket;bucket = bucket?bucket->next:NULL) {
	if (bucket->data==data) {
	    if (prebucket) prebucket->next = bucket->next;
	    else ctx->bucket[pos]=bucket->next;
	    if ((ctx->gift)&&(remove)) {
		if (ctx->cfunc) {
		    ctx->cfunc(bucket->data);
		} else {
		    if (ctx->mode==DS_HASH_TYPE_POINTER)
			dsFree(ctx->malloc, bucket->string);
		    dsFree(ctx->malloc, bucket->data);
		}
	    }    
	    dsFree(ctx->malloc, bucket);
	    bucket = prebucket;
	} else prebucket = bucket;
    }
    return err;
}

int dsHashTableUnlink(DSHashTable ctx, const void *data) {
    return HashTableRemoveOrUnlink(ctx,data,0);
}

int dsHashTableRemove(DSHashTable ctx, const void *data) {
    int res;
    
    ds_assert_param(ctx);
    
    if (ctx->prefunc) ctx->prefunc(ctx,DS_HASHTABLE_DESTRUCTIONMODE_SINGLE,ctx->ppfuncattr);
    res=HashTableRemoveOrUnlink(ctx,data,1);
    if (ctx->postfunc) ctx->postfunc(ctx,DS_HASHTABLE_DESTRUCTIONMODE_SINGLE,ctx->ppfuncattr);
    return res;
}

void *dsHashTableFindNext(DSHashTable ctx, const char *string, const void *start) {
    unsigned long hash;
    unsigned long pos;
    DSHashTableBucket bucket;
    
    ds_assert_null(ctx);
    ds_assert_null(string);
    ds_assert_null(ctx->mode<DS_HASH_TYPE_HASH);
    
    hash = hash_nstring(string,ctx->string_size);
    pos = ROUND(ctx->size * mod1( hash_table_coef * hash ));
    for (bucket = ctx->bucket[pos];bucket;bucket=bucket->next) {
	if (start) {
	    if (bucket->data!=start) continue;
	    else start=NULL;
	}
	if (bucket->hash==hash) {
	    if (ctx->string_size) {
		if (!memcmp(string,bucket->string,ctx->string_size)) return bucket->data;
	    } else {
		if (!strcmp(string,bucket->string)) return bucket->data;
	    }
	}
    }
    return NULL;
}

void *dsHashTableFind(DSHashTable ctx, const char *string) {
    return dsHashTableFindNext(ctx,string,NULL);
}

void *dsHashTableHashStringFindNext(DSHashTable ctx, unsigned long hash, const char *string, const void *start) {
    unsigned long pos;
    DSHashTableBucket bucket;
    
    ds_assert_null(ctx);
    
    pos = ROUND(ctx->size * mod1( hash_table_coef * hash ));
    for (bucket = ctx->bucket[pos];bucket;bucket=bucket->next) {
	if (start) {
	    if (bucket->data!=start) continue;
	    else start=NULL;
	}
	if (bucket->hash==hash) {
	    if (string) {
		if (ctx->string_size) {
		    if (!memcmp(string,bucket->string,ctx->string_size)) return bucket->data;
		} else {
		    if (!strcmp(string,bucket->string)) return bucket->data;
		}
	    } else return bucket->data;
	}
    }
    return NULL;
}

void *dsHashTableHashStringFind(DSHashTable ctx, unsigned long hash, const char *string) {
    return dsHashTableHashStringFindNext(ctx,hash,string,NULL);
}

void *dsHashTableHashFindNext(DSHashTable ctx, unsigned long hash, const void *start) {
    return dsHashTableHashStringFindNext(ctx,hash,NULL,start);    
}

void *dsHashTableHashFind(DSHashTable ctx, unsigned long hash) {
    return dsHashTableHashStringFindNext(ctx,hash,NULL,NULL);
}

int dsHashTableWalk(DSHashTable ctx, DSHashTableWalkFunction func, void *attr) {
    DSHashTableBucket bucket;
    DSHashTableBucket next;
    unsigned long i;
    int err;

    for (i=0;i<ctx->size;i++) {
	for (bucket=ctx->bucket[i];bucket;) {
	    next=bucket->next;
	    err=func(bucket->data, attr); /* Can remove current element */
	    if (err) return err;
	    bucket=next;
	}
    }
    return 0;
}

int dsHashTableIsEmpty(DSHashTable ctx) {
    unsigned long i;

    for (i=0;i<ctx->size;i++) {
	if (ctx->bucket[i]) return 0;
    }
    
    return 1;
}

int dsHashTableClean(DSHashTable ctx, DSHashTableCleanFunction func) {
    DSHashTableBucket prebucket;
    DSHashTableBucket bucket;
    DSHashTableBucket next;
    unsigned long i;

    for (i=0;i<ctx->size;i++) {
	prebucket = NULL;
	for (bucket=ctx->bucket[i];bucket;) {
	    next=bucket->next;
	    if (func(bucket->data)) {
		if (ctx->gift) {
		    if (ctx->cfunc) ctx->cfunc(bucket->data);
		    else {
			if (ctx->mode==DS_HASH_TYPE_POINTER)
			    dsFree(ctx->malloc, ((DSHashStringPtr)(bucket->data))->str);
			dsFree(ctx->malloc, bucket->data);
		    }
		}
		dsFree(ctx->malloc, bucket);
		
		if (prebucket) prebucket->next = next;
		else ctx->bucket[i] = next;
	    } else prebucket = bucket;
	    bucket=next;
	}
    }
    return 0;
}

unsigned long dsHashTableGetHash(DSHashTable ctx, const char *string) {
    ds_assert_err(ctx, (unsigned long)-1);
    ds_assert_err(string, (unsigned long)-1);
    
    return hash_nstring(string,ctx->string_size);
}


unsigned long dsHashTableGetPos(DSHashTable ctx, const char *string) {
    ds_assert_err(ctx, (unsigned long)-1);
    ds_assert_err(string, (unsigned long)-1);

    return ROUND(ctx->size * mod1( hash_table_coef *  hash_nstring(string,ctx->string_size)));
}

unsigned long dsHashTableHashGetPos(DSHashTable ctx, unsigned long hash) {
    ds_assert_err(ctx, (unsigned long)-1);

    return ROUND(ctx->size * mod1( hash_table_coef *  hash));
}

int dsHashTableGrow(DSHashTable ctx, unsigned long size) {
    DSHashTableBucket *newbucket;
    DSHashTableBucket bucket, nbucket, prebucket;
    unsigned long i, pos;
    unsigned long old_size, new_size;
    
    ds_assert_param(ctx);
    ds_assert_param(size);
    
    old_size = ctx->size;
    new_size = old_size + size;
    
    newbucket=(DSHashTableBucket*)dsCalloc(ctx->malloc,new_size,sizeof(DSHashTableBucket));
    if (!newbucket) return DS_ERR_MEM;

    for (i=0;i<old_size;i++) {
	for (bucket=ctx->bucket[i];bucket;bucket=nbucket) {
	    pos = ROUND(new_size * mod1( hash_table_coef * bucket->hash ));
	    for (prebucket = NULL, nbucket = newbucket[pos]; nbucket; nbucket = nbucket->next) prebucket = nbucket;
	    
	    nbucket = bucket->next;

	    bucket->next = NULL;
	    if (prebucket) {
		prebucket->next = bucket;
	    } else { 
		newbucket[pos] = bucket;
	    }
	}
    }
    
    dsFree(ctx->malloc, ctx->bucket);
    ctx->bucket = newbucket;
    ctx->size = new_size;
    
    return 0;
}
