#define _DS_INCLUDE_XML_HEADERS
#include <string.h>

#include <dslib.h>

#include <dsassert.h>
#include <dsdirectory.h>
#include <dsxml.h>
#include <dsstring.h>
#include <dsconfig.h>

#define DS_CONFIG_SUBCONFIGS_NUMBER 	128
#define DS_CONFIG_REFPROPS_NUMBERS	32

static DSSubConfigContext dsSubConfigCreateContext(DSConfigContext ctx, DSConfigReference reference, unsigned char copy);
static void dsSubConfigDestroyContext(DSSubConfigContext sub);
static int dsSubConfigLock(DSSubConfigContext sub);
static void dsSubConfigReturn(DSSubConfigContext sub);

void dsConfigZeroContext(DSConfigContext ctx) {
    ctx->lock = NULL;

    ctx->name = NULL;
    ctx->sysdir = NULL;
    ctx->userdir = NULL;
    ctx->dir = NULL;
    
    ctx->sysconfig = NULL;
    ctx->userconfig = NULL;
    ctx->config = NULL;    
    ctx->defconfig = NULL;
    
    ctx->xp = NULL;
    ctx->userxp = NULL;
    ctx->sysxp = NULL;
    
    ctx->nodes = 0;
    ctx->nodesets = 0;
    
    ctx->refprops = NULL;
    ctx->refprops_lock = NULL;
    
    ctx->subconfigs = NULL;

    ctx->autoset = 0;
    ctx->autoset_refprops = 1;
    ctx->changed = 0;

    ctx->fake_references = 0;
    ctx->masterref = NULL;
}


DSConfigContext dsConfigCreateContext(DSMallocContext mctx) {
    DSConfigContext ctx;

    dsDebugServer(DS_DEBUG_INFO, "dsConfigCreateContext");
    
    ctx = (DSConfigContext)dsMalloc(mctx, sizeof(DSConfigContextS));
    if (ctx) {
	ctx->malloc = mctx;
	dsConfigZeroContext(ctx);
    }
    
    return ctx;
}


int dsConfigInitContext(DSConfigContext ctx, const char *name) {
    int err;

    ds_assert_param(ctx);
    ds_assert_param(name);

    dsDebugServer(DS_DEBUG_INFO, "dsConfig %s: InitContext", name);
    
    dsConfigZeroContext(ctx);

    ctx->lock = dsLockCreateContext(ctx->malloc);
    ctx->refprops_lock = dsLockCreateContext(ctx->malloc);
    
    if (ctx->lock) err = dsLockInitContext(ctx->lock);
    else err = DS_ERR_MEM;
    
    if (!err) {
	if (ctx->refprops_lock) err = dsLockInitContext(ctx->refprops_lock);
	else err = DS_ERR_MEM;
    }

    if (err) {
	dsConfigFreeContext(ctx);
	return err;
    }
    
    dsConfigSetSystemConfigDirectory(ctx, DS_CONFIGDIR);
    dsConfigSetUserConfigDirectory(ctx, "~");
    
    err = dsExchangeString(ctx->malloc, &ctx->name, name);
    if (err) {
	dsConfigFreeContext(ctx);
	return err;
    }
    
    ctx->masterref = dsConfigCreateFakeReference(ctx);
    if (!ctx->masterref) {
	dsConfigFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    return 0;
}

void dsConfigFreeContext(DSConfigContext ctx) {
    ds_assert_return(ctx);

    dsDebugServer(DS_DEBUG_INFO, "dsConfig %s: FreeContext", ctx->name?ctx->name:"");
    
    dsConfigUnload(ctx); 
    
    if (ctx->masterref) dsConfigDestroyReference(ctx, ctx->masterref);
    
    dsExchangeString(ctx->malloc, &ctx->name, NULL);
    dsExchangeString(ctx->malloc, &ctx->sysdir, NULL);
    dsExchangeString(ctx->malloc, &ctx->userdir, NULL);
    dsExchangeString(ctx->malloc, &ctx->dir, NULL);

    if (ctx->refprops_lock) dsLockDestroyContext(ctx->refprops_lock);
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    
    dsConfigZeroContext(ctx);
}

void dsConfigDestroyContext(DSConfigContext ctx) {
    dsDebugServer(DS_DEBUG_INFO, "dsConfig %s: DestroyContext", ctx->name?ctx->name:"");
    dsConfigFreeContext(ctx);
    dsFree(ctx->malloc, ctx);
}

int dsConfigSetSystemConfigDirectory(DSConfigContext ctx, const char *sysdir) {
    int err;
    ds_assert_param(ctx);
    err = dsExchangeString(ctx->malloc, &ctx->sysdir, sysdir);
    if (!err) err = dsDirectoryParse(ctx->malloc, &ctx->sysdir);
    return err;
}

int dsConfigSetUserConfigDirectory(DSConfigContext ctx, const char *userdir) {
    int err;
    ds_assert_param(ctx);
    err = dsExchangeString(ctx->malloc, &ctx->userdir, userdir);
    if (!err) err = dsDirectoryParse(ctx->malloc, &ctx->userdir);
    return err;
}

int dsConfigSetConfigDirectory(DSConfigContext ctx, const char *dir) {
    int err;
    ds_assert_param(ctx);
    err = dsExchangeString(ctx->malloc, &ctx->dir, dir);
    if (!err) err = dsDirectoryParse(ctx->malloc, &ctx->dir);
    return err;
}

int dsConfigEnableAutosetDefaults(DSConfigContext ctx) {
    ds_assert_param(ctx);
    ctx->autoset = 1;
    return 0;
}

int dsConfigDisableAutosetDefaults(DSConfigContext ctx) {
    ds_assert_param(ctx);
    ctx->autoset = 0;
    return 0;
}

int dsConfigEnableIDAttributeDetection(DSConfigContext ctx) {
    ds_assert_param(ctx);
    ctx->autoset_refprops = 1;
    return 0;
}
int dsConfigDisableIDAttributeDetection(DSConfigContext ctx) {
    ds_assert_param(ctx);
    ctx->autoset_refprops = 0;
    return 0;
}

int dsConfigCustomLoad(DSConfigContext ctx, int flags) {
    int err;
    DSFileContextS file;
    
    ds_assert_param(ctx);
    ds_assert_param(ctx->name);
    ds_assert_param(ctx->config==NULL&&ctx->userconfig==NULL&&ctx->sysconfig==NULL);
    
    dsDebugServer(DS_DEBUG_INFO, "dsConfig %s: Load", ctx->name);
    

    if (flags&DS_CONFIG_SKIP_LOCAL) err = DS_ERR;
    else err = dsFileOpen(&file, ctx->dir, ctx->name);
    if (!err) {
	ctx->config = xmlReadFile(dsFileGetFullName(&file), NULL, 0);
	dsFileClose(&file);

	if (ctx->config) {
	    ctx->xp = xmlXPathNewContext(ctx->config);
	    if (ctx->xp) ctx->defconfig = ctx->config;
	    else {
		xmlFreeDoc(ctx->config);
		ctx->config = NULL;
	    }
	}
    }

    if (flags&DS_CONFIG_SKIP_USER) err = DS_ERR;
    err = dsFileOpen(&file, ctx->userdir, ctx->name);
    if (!err) {
	ctx->userconfig = xmlReadFile(dsFileGetFullName(&file), NULL, 0);
	dsFileClose(&file);
	
	if (ctx->userconfig) {
	    ctx->userxp = xmlXPathNewContext(ctx->userconfig);
	    if (ctx->userxp) {
		if (!ctx->defconfig) ctx->defconfig = ctx->userconfig;
	    } else {
		xmlFreeDoc(ctx->userconfig);
		ctx->userconfig = NULL;
	    }
	}
    }

    if (flags&DS_CONFIG_SKIP_SYSTEM) err = DS_ERR;
    else err = dsFileOpen(&file, ctx->sysdir, ctx->name);
    if (!err) {
	ctx->sysconfig = xmlReadFile(dsFileGetFullName(&file), NULL, 0);
	dsFileClose(&file);
	
	if (ctx->sysconfig) {
	    ctx->sysxp = xmlXPathNewContext(ctx->sysconfig);
	    if (ctx->sysxp) {
		if (!ctx->defconfig) ctx->defconfig = ctx->sysconfig;
	    } else {
		xmlFreeDoc(ctx->sysconfig);
		ctx->sysconfig = NULL;
	    }
	}
    }

	/* Creating empty config */
    if ((!ctx->config)&&(!ctx->userconfig)&&(!ctx->sysconfig)) {
	ctx->config = xmlNewDoc("1.0");
	if (ctx->config) {
	    ctx->xp = xmlXPathNewContext(ctx->config);
	    if (ctx->xp) ctx->defconfig = ctx->config;
	    else {
		xmlFreeDoc(ctx->config);
		ctx->config = NULL;
		return DS_ERR_MEM;
	    }
	} else return DS_ERR_MEM;
    }

	/* DS: Adjust hash size, basing on the config size */
    ctx->subconfigs = dsHashTableInit(NULL, DS_CONFIG_SUBCONFIGS_NUMBER, DS_HASH_TYPE_POINTER, NULL);
    ctx->refprops = dsHashTableInit(NULL, DS_CONFIG_REFPROPS_NUMBERS, DS_HASH_TYPE_ARRAY, NULL);
    
    if ((!ctx->subconfigs)||(!ctx->refprops)) {
	dsConfigUnload(ctx);
	return DS_ERR_MEM;
    }

    dsHashTableSetDestructor(ctx->subconfigs, (DSHashTableDestructorFunction)dsSubConfigDestroyContext);
    dsHashTableSetDestructor(ctx->refprops, NULL);
    dsConfigRegisterIDAttribute(ctx, "id");

    return 0;
}

int dsConfigLoad(DSConfigContext ctx) {
    return dsConfigCustomLoad(ctx, 0);
}

void dsConfigUnload(DSConfigContext ctx) {
    ds_assert_return(ctx);

    dsDebugServer(DS_DEBUG_INFO, "dsConfig %s: UnLoad", ctx->name);
    
    if (ctx->refprops) {
	dsHashTableDestroy(ctx->refprops);
	ctx->refprops = NULL;
    }
    
    if (ctx->subconfigs) {
	dsHashTableDestroy(ctx->subconfigs);
	ctx->subconfigs = NULL;
    }
    
    if (ctx->sysconfig) {
	xmlXPathFreeContext(ctx->sysxp);
	xmlFreeDoc(ctx->sysconfig);
	
	ctx->sysconfig = NULL;
	ctx->sysxp = NULL;
    }
    if (ctx->userconfig) {
	xmlXPathFreeContext(ctx->userxp);
	xmlFreeDoc(ctx->userconfig);
	
	ctx->userconfig = NULL;
	ctx->userxp = NULL;
    }
    if (ctx->config) {
	xmlXPathFreeContext(ctx->xp);
	xmlFreeDoc(ctx->config);
	
	ctx->config = NULL;
	ctx->xp = NULL;
    }

    ctx->defconfig = NULL;
}

DSConfigReference dsConfigCopyReference(DSConfigContext ctx, DSConfigReference reference) {
    DSConfigReference res = NULL;

    ds_assert_null(ctx);
    ds_assert_null(reference);
        
    if (dsExchangeString(ctx->malloc, &res, reference)) return NULL;
     
    return res;
}


DSConfigReference dsConfigCreateFakeReference(DSConfigContext ctx) {
    unsigned long refnum;
    
    ds_assert_null(ctx);
    
    dsLockWrite(ctx->lock);
    refnum = (++ctx->fake_references);
    dsLockReturn(ctx->lock);
    
    return dsStringCreate(ctx->malloc, "/DS_CONFIG_FAKEREF_%lu", refnum);
}

static int dsConfigReferenceSetMasterReferenceInternal(DSConfigContext ctx, DSSubConfigContext sub, DSConfigReference master) {
    int err = 0;
    DSSubConfigContext mastersub;

    dsDebugServer(DS_DEBUG_DEBUG, "dsConfig %s: SetMasterReference: %s->%s", ctx->name, sub->ref, master);

    mastersub = dsHashTableFind(ctx->subconfigs, master);
    if (!mastersub) {
	mastersub = dsSubConfigCreateContext(ctx, master, 1);
	if (!mastersub) {
	    return DS_ERR_MEM;
	}

	if (master != ctx->masterref) {
	    err = dsConfigReferenceSetMasterReferenceInternal(ctx, mastersub, ctx->masterref);
	    if (err) {
		dsSubConfigDestroyContext(mastersub);
		return err;
	    }
	}

	err = dsHashTableInsert(ctx->subconfigs, mastersub);
	if (err) {
	    dsSubConfigDestroyContext(mastersub);
	    return err;
	}
    }

    sub->master = mastersub;
    
    return 0;
}

int dsConfigReferenceSetMasterReference(DSConfigContext ctx, DSConfigReference reference, DSConfigReference master) {
    int err = 0;
    DSSubConfigContext sub;
    
    ds_assert_param(ctx);
    ds_assert_param(reference);
    ds_assert_param(master);

    err = dsLockWrite(ctx->lock);
    if (err) return err;

    sub = dsHashTableFind(ctx->subconfigs, reference);
    if (sub) {
        err = dsConfigReferenceSetMasterReferenceInternal(ctx, sub, master);
    } else {
	sub = dsSubConfigCreateContext(ctx, reference, 1);
	if (!sub) err = DS_ERR_MEM;

        if (!err) err = dsConfigReferenceSetMasterReferenceInternal(ctx, sub, master);
	
	if (!err) err = dsHashTableInsert(ctx->subconfigs, sub);

	if ((err)&&(sub)) dsSubConfigDestroyContext(sub);
    } 
    
    dsLockReturn(ctx->lock);
    
    return err;
}

static int dsConfigLockReference(DSConfigContext ctx, xmlNodePtr master, DSConfigReference reference) {
    int err;
    DSConfigReference mref;
    DSSubConfigContext sub;
    
    ds_assert_param(ctx);
    ds_assert_param(reference);
    
    sub = dsHashTableFind(ctx->subconfigs, reference);
    if (!sub) {
	sub = dsSubConfigCreateContext(ctx, reference, 1);
	if (!sub) return DS_ERR_MEM;

	if (master) {
	    mref = dsConfigGetConfigReference(ctx, master);
	    if (mref) {
		err = dsConfigReferenceSetMasterReferenceInternal(ctx, sub, mref);
		dsConfigDestroyReference(ctx, mref);
	    } else err = DS_ERR_MEM;
	} else
	    err = dsConfigReferenceSetMasterReferenceInternal(ctx, sub, ctx->masterref);

	if (err) {
	    dsSubConfigDestroyContext(sub);
	    return err;
	}
	
	err = dsHashTableInsert(ctx->subconfigs, sub);
	if (err) {
	    dsSubConfigDestroyContext(sub);
	    return err;
	}
    }
    
    return dsSubConfigLock(sub);
}

static void dsConfigUnLockReference(DSConfigContext ctx, DSConfigReference ref) {
    DSSubConfigContext sub;

    ds_assert_return(ctx);
    ds_assert_return(ref);

    sub = dsHashTableFind(ctx->subconfigs, ref);
    ds_assert_return(sub);
    
    dsSubConfigReturn(sub);
}

static int dsConfigRegisterNodeReference(DSConfigContext ctx, xmlNodePtr master, xmlNodePtr node) {
    int err;
    DSConfigReference ref;

    ds_assert_param(ctx);
    ds_assert_param(node);
    
    ref = dsConfigGetConfigReference(ctx, node);
    if (!ref) return DS_ERR_MEM;
    
    dsDebugServer(DS_DEBUG_DEBUG, "dsConfig %s: RegisterNodeReference: %s", ctx->name, ref);

    err = dsConfigLockReference(ctx, master, ref);

    dsConfigDestroyReference(ctx, ref);
    
    return err;
    
}


static void dsConfigReturnNodeReference(DSConfigContext ctx, xmlNodePtr node) {
    DSConfigReference ref;

    ds_assert_return(ctx);
    ds_assert_return(node);
    
    ref = dsConfigGetConfigReference(ctx, node);
    ds_assert_return(ref);

    dsDebugServer(DS_DEBUG_DEBUG, "dsConfig %s: ReturnNodeReference: %s", ctx->name, ref);

    dsConfigUnLockReference(ctx, ref);

    dsConfigDestroyReference(ctx, ref);
}


static int dsConfigRegisterNodesetReference(DSConfigContext ctx, xmlNodePtr master, xmlNodePtr *nodeset) {
    int err;
    int i, j;

    ds_assert_param(ctx);
    ds_assert_param(nodeset);

    dsDebugServer(DS_DEBUG_DEBUG, "dsConfig %s: RegisterNodesetReference", ctx->name);
    
    for (i=0;nodeset[i];i++) {
	err = dsConfigRegisterNodeReference(ctx, master, nodeset[i]);
	if (err) {
	    for (j=0;j<i;j++) {
		dsConfigReturnNodeReference(ctx, nodeset[j]);
	    }
	    return err;
	}
    }
    return 0;
}


static void dsConfigReturnNodesetReference(DSConfigContext ctx, xmlNodePtr *nodeset) {
    int i;

    ds_assert_return(ctx);
    ds_assert_return(nodeset);
    
    dsDebugServer(DS_DEBUG_DEBUG, "dsConfig %s: ReturnNodesetReference", ctx->name);

    for (i=0;nodeset[i];i++) {
	dsConfigReturnNodeReference(ctx, nodeset[i]);
    }
}

static xmlNodePtr dsConfigGetConfigInternal(DSConfigContext ctx, xmlNodePtr node, const char *request);
static xmlNodePtr dsConfigCreateCurrentConfig(DSConfigContext ctx, xmlNodePtr res);

static DSConfigReference dsConfigGetParentReference(DSConfigContext ctx, xmlNodePtr node) {
    size_t len;
    const char *ch;
    DSConfigReference pxpath, xpath;

    xpath = dsConfigGetConfigReference(ctx, node);
    if (!xpath) return NULL;
    
    ch = dsStringStrrchr(xpath, '/');
    if (!ch) return xpath;

    len = (ch - (const char*)xpath);

    if (len) {
	pxpath = (xmlChar*)dsMalloc(ctx->malloc, (len + 1)*sizeof(char));
	if (pxpath) {
    	    memcpy(pxpath, xpath, len);
	    pxpath[len] = 0;
	}
    } else {
	pxpath = (xmlChar*)dsMalloc(ctx->malloc, 2*sizeof(char));
	if (pxpath) {
	    pxpath[0] = '/';
	    pxpath[1] = 0;
	}
    }
    
    dsConfigDestroyReference(ctx, xpath);
    
    return pxpath;
}    


static xmlNodePtr dsConfigCreateCurrentConfig(DSConfigContext ctx, xmlNodePtr res) {
    int err = 0;
    DSConfigReference ref;
    xmlNodePtr root;    
    xmlNodePtr newres, newtree = NULL;
    xmlNodePtr cnode, mynode;
    
    if (res->doc == ctx->defconfig) return res;

    root = xmlDocGetRootElement(ctx->defconfig);
    if (root) {
    	for (err=DS_ERR_NOTFOUND,cnode = res;(cnode)&&(cnode->name)&&(cnode->type = XML_ELEMENT_NODE);cnode = cnode->parent) {
	    if (cnode == res) {
		newres = xmlDocCopyNode(res, ctx->defconfig, 1);
		if (newres) newtree = newres;
		else {
		    err = DS_ERR_MEM;
		    break;
		}
	    } else {
		mynode = xmlDocCopyNode(cnode, ctx->defconfig, 0);
		if (mynode) {
		    xmlAddChild(mynode, newtree);
		    newtree = mynode;
		} else {
		    err = DS_ERR_MEM;
		    break;
		}
	    }
		    
	    ref = dsConfigGetParentReference(ctx, cnode);
	    mynode = dsConfigGetConfigInternal(ctx, root, ref);
	    dsConfigDestroyReference(ctx, ref);
	    if (mynode) {
		xmlAddChild(mynode, newtree);
		err = 0;
		break;	
	    } 
	}
	
	if (err) {
	    if (newtree) xmlFreeNode(newtree);
	    return res;
	} 

    } else {
    	for (cnode = res;(cnode)&&(cnode->name)&&(cnode->type = XML_ELEMENT_NODE);cnode = cnode->parent) {
	    if (cnode == res) {
		newres = xmlDocCopyNode(res, ctx->defconfig, 1);
		if (newres) newtree = newres;
		else {
		    err = DS_ERR_MEM;
		    break;
		}
	    } else {
		mynode = xmlDocCopyNode(cnode, ctx->defconfig, 0);
		if (mynode) {
		    xmlAddChild(mynode, newtree);
		    newtree = mynode;
		} else {
		    err = DS_ERR_MEM;
		    break;
		}
	    }
	}
	if (err) {
	    if (newtree) xmlFreeNode(newtree);
	    return res;
	} 
	xmlDocSetRootElement(ctx->defconfig, newtree);
    }

    return newres;
}	    


static xmlNodePtr dsConfigGetConfigInternal(DSConfigContext ctx, xmlNodePtr node, const char *req) {
    int err = 0;
    
    xmlNodeSetPtr node_set;
    xmlNodePtr res;
    xmlXPathObjectPtr obj;
    xmlDocPtr nodedoc;

    if (node) nodedoc = node->doc;
    else nodedoc = NULL;

    if ((ctx->xp)&&((!nodedoc)||(ctx->config==nodedoc))) {
	if (node) ctx->xp->node = node;
	obj = xmlXPathEvalExpression((xmlChar*)req, ctx->xp);
	if (obj) {
	    node_set = obj->nodesetval;
	    if ((!node_set)||(!node_set->nodeNr)) {
		xmlXPathFreeObject(obj);
		obj = NULL;
	    } else if (node_set->nodeNr>1) {
		err = DS_ERR_INVALID_REQUEST;

		xmlXPathFreeObject(obj);
		obj = NULL;
	    }
	}
    } else obj = NULL;

    if ((!err)&&(!obj)&&(ctx->userxp)&&((!nodedoc)||(ctx->userconfig==nodedoc))) {
	if (node) ctx->userxp->node = node;
	obj = xmlXPathEvalExpression((xmlChar*)req, ctx->userxp);
	if (obj) {
	    node_set = obj->nodesetval;
	    if ((!node_set)||(!node_set->nodeNr)) {
		xmlXPathFreeObject(obj);
		obj = NULL;
	    } else if (node_set->nodeNr>1) {
		err = DS_ERR_INVALID_REQUEST;
		xmlXPathFreeObject(obj);
		obj = NULL;
	    }
	}
    }

    if ((!err)&&(!obj)&&(ctx->sysxp)&&((!nodedoc)||(ctx->sysconfig==nodedoc))) {
	if (node) ctx->sysxp->node = node;
	obj = xmlXPathEvalExpression((xmlChar*)req, ctx->sysxp);
	if (obj) {
	    node_set = obj->nodesetval;
	    if ((!node_set)||(node_set->nodeNr != 1)) {
		xmlXPathFreeObject(obj);
		obj = NULL;
	    } 
	}
    }

    if (obj) ctx->nodes++;

    if (obj) {
	res = node_set->nodeTab[0];
	xmlXPathFreeObject(obj);
    } else res = NULL;
    
    if (res) {
	if (res->doc != ctx->defconfig)
	    res = dsConfigCreateCurrentConfig(ctx, res);
    }

    return res;
}

int dsConfigRegisterIDAttribute(DSConfigContext ctx, const char *attr) {
    int err;
    char *data;
    
    ds_assert_param(ctx);
    ds_assert_param(attr);

    ds_assert_param(ctx->refprops);
    

    data = dsStringCopy(ctx->malloc, attr);
    if (data) {    
	err = dsLockWrite(ctx->refprops_lock);
	if (!err) {
	    err = dsHashTableAdd(ctx->refprops, data, 1);
	    dsLockReturn(ctx->refprops_lock);
	}
	if (err) dsStringDestroy(ctx->malloc, data);
	else dsDebugServer(DS_DEBUG_INFO, "dsConfig %s: RegisterIDAttribute: %s", ctx->name, attr);
    } else err = DS_ERR_MEM;

    return err;
}

static int dsConfigExtractRefProps(DSConfigContext ctx, const xmlChar *req) {
    int err = 0;
    size_t size;
    DSStringConstant pos, attr;
    DSString str, end, attrend;

    if (!ctx->autoset_refprops) return 0;
    
    size = strlen(req) + 1;

    str = dsMalloca(size*sizeof(char));
    if (!str) return DS_ERR_MEM;
    memcpy(str, req, size);

    for (pos = dsStringStrchr(str, '['); pos; pos = dsStringStrchr(end+1, '[')) {
    	end = (DSString)dsStringStrchr(pos+1, ']');
	if (!end) {
	    err = DS_ERR_INVALID_REQUEST;
	    break;
	}
	*end = 0;
	
	for (attr = dsStringStrchr(pos+1, '@');((attr)&&(attr<end));attr = dsStringStrchr(pos+1, '@')) {
	    attrend = (DSString)xmlStrchr(attr, '=');
	    
	    if (attrend) {
		pos = (DSString) dsStringStrchr(attr+1, '@');
		if ((pos)&&(attrend>pos)) attrend = NULL;
	    }

	    pos = (DSString)xmlStrchr(attrend?(attrend+1):(attr+1), ' ');
	    if (!pos) pos = (DSString)xmlStrchr(attrend?(attrend+1):(attr+1), '\t');
	    if (!pos) pos = (DSString)xmlStrchr(attrend?(attrend+1):(attr+1), ')');
	    if (!pos) pos = (DSString)xmlStrchr(attrend?(attrend+1):(attr+1), '(');

	    if (!attrend) attrend = (DSString)pos;
	    if (attrend) *attrend = 0;
	    
	    attrend = (DSString)dsStringStrchr(attr+1, ' ');
	    if (!attrend) attrend = (DSString)dsStringStrchr(attr+1, '\t');
	    if (attrend) *attrend = 0;
	    
	    dsConfigRegisterIDAttribute(ctx, attr+1);

	    if (!pos) break;
	}
    }

    dsFreea(str);
    
    return 0;
}

xmlNodePtr dsConfigGetConfig(DSConfigContext ctx, xmlNodePtr node, const char *request, ...) {
    int err;
    xmlNodePtr res;
    va_list ap;
    char *req;

    ds_assert_null(ctx);
    ds_assert_null(ctx->lock);
    ds_assert_null(request);
    
    va_start(ap, request);
    req = dsStringVCreate(NULL, request, ap);
    va_end(ap);

    if (!req) return NULL;

    dsConfigExtractRefProps(ctx, req);

    if (dsLockWrite(ctx->lock)) {
	dsStringDestroy(NULL, req);
	return NULL;
    }
    
    res = dsConfigGetConfigInternal(ctx, node, req);

    if (res) {
	err = dsConfigRegisterNodeReference(ctx, node, res);
	if (err) res = NULL;
    }

    dsLockReturn(ctx->lock);
    
    dsStringDestroy(NULL, req);
    
    return res;
}

xmlNodePtr *dsConfigGetConfigs(DSConfigContext ctx, xmlNodePtr node, const char *request, ...) {
    int err;
    
    xmlDocPtr nodedoc;
    xmlNodeSetPtr node_set;
    xmlNodePtr *res;
    xmlXPathObjectPtr obj;

    unsigned int i, args = 0;
    va_list ap;
    char *req;

    ds_assert_null(ctx);
    ds_assert_null(ctx->lock);
    ds_assert_null(request);

    va_start(ap, request);
    req = dsStringVCreate(NULL, request, ap);
    va_end(ap);

    if (!req) return NULL;
    
    dsConfigExtractRefProps(ctx, req);

    if (node) nodedoc = node->doc;
    else nodedoc = NULL;

    dsLockWrite(ctx->lock);
    if ((ctx->xp)&&((!nodedoc)||(ctx->config==nodedoc))) {
	if (node) ctx->xp->node = node;
	obj = xmlXPathEvalExpression((xmlChar*)req, ctx->xp);
	if (obj) {
	    node_set = obj->nodesetval;
	    if ((!node_set)||(!node_set->nodeNr)) {
		xmlXPathFreeObject(obj);
		obj = NULL;
	    }
	}
    } else obj = NULL;
    
    if ((!obj)&&(ctx->userxp)&&((!nodedoc)||(ctx->userconfig==nodedoc))) {
	if (node) ctx->userxp->node = node;
	obj = xmlXPathEvalExpression((xmlChar*)req, ctx->userxp);
	if (obj) {
	    node_set = obj->nodesetval;
	    if ((!node_set)||(!node_set->nodeNr)) {
		xmlXPathFreeObject(obj);
		obj = NULL;
	    }
	}
    }

    if ((!obj)&&(ctx->sysxp)&&((!nodedoc)||(ctx->sysconfig==nodedoc))) {
	if (node) ctx->sysxp->node = node;
	obj = xmlXPathEvalExpression((xmlChar*)req, ctx->sysxp);
	if (obj) {
	    node_set = obj->nodesetval;
	    if ((!node_set)||(!node_set->nodeNr)) {
		xmlXPathFreeObject(obj);
		obj = NULL;
	    }
	}
    }
    
    if (obj) ctx->nodesets++;

    if (args) dsFreea(req);

    if (obj) {
	
	res = (xmlNodePtr*)dsMalloc(ctx->malloc, (node_set->nodeNr+1)*sizeof(xmlNodePtr));
	if (res) {
	    for (i=0;(int)i<node_set->nodeNr;i++)
		res[i] = dsConfigCreateCurrentConfig(ctx, node_set->nodeTab[i]);
	    res[i] = NULL;
	} else ctx->nodesets--;

/*
	    // Aligning 32 bytes
	size = (sizeof(xmlNodeSet)&~(32-1))+32;
	res = (xmlNodeSetPtr)dsMalloc(ctx->malloc, size + node_set->nodeNr*sizeof(xmlNodePtr));
	if (res) {
	    res->nodeNr = node_set->nodeNr;
	    res->nodeMax = node_set->nodeNr;
	    res->nodeTab = (xmlNodePtr*)(((char*)res)+32);
	    memcpy(res->nodeTab,node_set->nodeTab,node_set->nodeNr*sizeof(xmlNodePtr));
	} else ctx->nodesets--;
*/
	xmlXPathFreeObject(obj);
    } else res = NULL;

    if (res) {
	err = dsConfigRegisterNodesetReference(ctx, node, res);
	if (err) {
	    dsFree(ctx->malloc, res);
	    res = NULL;
	}
    }

    dsLockReturn(ctx->lock);

    dsStringDestroy(NULL, req);
    
    return res;
}


void dsConfigReturnConfig(DSConfigContext ctx, xmlNodePtr node) {
    ds_assert_return(ctx);
    ds_assert_return(ctx->lock);
    ds_assert_return(node);

    dsLockWrite(ctx->lock);
    ds_assert_info(ctx->nodes > 0);
    if (ctx->nodes > 0) ctx->nodes--;
    dsConfigReturnNodeReference(ctx, node);
    dsLockReturn(ctx->lock);
}

void dsConfigReturnConfigs(DSConfigContext ctx, xmlNodePtr *nodeset) {
    ds_assert_return(ctx);
    ds_assert_return(ctx->lock);
    ds_assert_return(nodeset);
    
    dsLockWrite(ctx->lock);
    ds_assert_info(ctx->nodesets > 0);
    if (ctx->nodesets > 0) ctx->nodesets--;
    dsConfigReturnNodesetReference(ctx, nodeset);
    dsLockReturn(ctx->lock);

    dsFree(ctx->malloc, nodeset);
}

DSConfigReference dsConfigGetConfigReference(DSConfigContext ctx, xmlNodePtr node) {
    int err;
    const char *id;
    const char *content;
    DSConfigReference ref;
    size_t pos, oldsize, size = 0;
    xmlNodePtr cnode;
    xmlAttrPtr props;
    int first;
    char ch;

    ds_assert_null(ctx);
    ds_assert_null(node);
    
    err = dsLockRead(ctx->refprops_lock);
    if (err) return NULL;
    
    for (cnode = node;(cnode)&&(cnode->name)&&(cnode->type = XML_ELEMENT_NODE);cnode = cnode->parent) {
	size += strlen(cnode->name) + 1;
	for (first=1, props = cnode->properties;props;props=props->next) {
	    if (dsHashTableFind(ctx->refprops, props->name)) {
		content = props->children->content;
		if ((content)&&(*content)) {
		    size += strlen(props->name) + strlen(content) + (first?6:9);/* "[@='']":" and @=''" */
		} else {
		    size += strlen(props->name) + (first?3:6);/* "[@]":" and @" */
		}
		first = 0;
	    }
	}
    }

    id = /*dsXmlGetPropString(node, "id", NULL)*/ NULL;
    
    if (id) {
	ref = (DSConfigReference)dsMalloc(ctx->malloc, (size+strlen(id)+9)*sizeof(char));
	if (!ref) {
	    dsLockReturn(ctx->refprops_lock);
	    return ref;
	}
	sprintf(ref + size - 1,"[@id=\"%s\"]", id);
    } else {
	ref = (DSConfigReference)dsMalloc(ctx->malloc, (size+1)*sizeof(char));
	if (!ref) {
	    dsLockReturn(ctx->refprops_lock);
	    return ref;
	}
	ref[size] = 0;
    }

    ref[0] = '/';
    for (cnode = node;(cnode)&&(cnode->name)&&(cnode->name)&&(cnode->type = XML_ELEMENT_NODE);cnode = cnode->parent) {
	for (first=1, props = cnode->properties;props;props=props->next) {
	    if (dsHashTableFind(ctx->refprops, props->name)) {
		ch = ref[size]; oldsize = size;
		content = props->children->content;
		if ((content)&&(*content)) {
		    if (first) {
		        size -= strlen(props->name) + strlen(content) + 5;
			sprintf(ref + size, "@%s='%s']", props->name, content);
		    } else {
		        size -= strlen(props->name) + strlen(content) + 9;
			sprintf(ref + size, "@%s='%s' and ", props->name, content);
		    }
		} else {
		    if (first) {
		        size -= strlen(props->name) + 2;
			sprintf(ref + size, "@%s]", props->name);
		    } else {
		        size -= strlen(props->name) + 6;
			sprintf(ref + size, "@%s and ", props->name);
		    }
		}
		ref[oldsize] = ch;
		first = 0;
	    }
	}
	if (!first) ref[(--size)]='[';

	pos = strlen(cnode->name);
	size -=  pos + 1;
	
	ref[size] = '/';
	memcpy(ref + size + 1, cnode->name, pos);
    }

    dsLockReturn(ctx->refprops_lock);
    
    return ref;
}

void dsConfigDestroyReference(DSConfigContext ctx, DSConfigReference ref) {
    ds_assert_return(ctx);
    ds_assert_return(ref);

    dsFree(ctx->malloc, ref);
}

DSConfigReference *dsConfigGetConfigsReferences(DSConfigContext ctx, xmlNodePtr *nodeset) {
    int i;

    DSConfigReference ref;
    DSConfigReferenceSet refset;
    
    ds_assert_null(ctx);
    ds_assert_null(nodeset);

    for (i=0;nodeset[i];i++);

    refset = (DSConfigReferenceSet)dsMalloc(ctx->malloc, (i + 1)*sizeof(DSConfigReference));
    if (!refset) return refset;    
    
    for (i=0;nodeset[i];i++) {
	ref = dsConfigGetConfigReference(ctx, nodeset[i]);
	refset[i] = ref;

	if (!ref) {
	    dsConfigDestroyReferences(ctx, refset);
	    return NULL;
	}
    }

    refset[i] = NULL;
    
    return refset;
}

void dsConfigDestroyReferences(DSConfigContext ctx, DSConfigReference *refset) {
    int i;

    ds_assert_return(ctx);
    ds_assert_return(refset);
    
    for (i=0;refset[i];i++)
	dsConfigDestroyReference(ctx, refset[i]);
    
    dsFree(ctx->malloc, refset);
}

xmlNodePtr dsConfigGetConfigByReference(DSConfigContext ctx, DSConfigReference ref) {
    ds_assert_null(ctx);
    ds_assert_null(ref);
    
    return dsConfigGetConfig(ctx, NULL, ref);
}

int dsConfigDumpMemory(DSConfigContext ctx, char **buf, size_t *size) {
    int err;
    int bufsize;
    xmlDocPtr doc;
    
    ds_assert_param(ctx);
    ds_assert_param(buf);

/*
    doc = ctx->config;
    if (!doc) doc = ctx->userconfig;
    if (!doc) doc = ctx->sysconfig;
*/
    doc = ctx->defconfig;

    err = dsConfigLockReference(ctx, NULL, ctx->masterref);
    if (!err) {
	xmlDocDumpFormatMemoryEnc(doc, (xmlChar**)buf, &bufsize, "UTF-8", 1);
	dsConfigUnLockReference(ctx, ctx->masterref);
	if (!*buf) return DS_ERR_XML;
    
        if (size) *size = bufsize;
    }

    return err;
}

int dsConfigSave(DSConfigContext ctx) {
    int err = DS_ERR_XML_WRITER;
    char *fn = NULL;
    
    ds_assert_param(ctx);

    if (ctx->defconfig == ctx->config) {
	err = dsConfigSaveLocal(ctx);
    } else if (ctx->defconfig == ctx->userconfig) {
	err = dsConfigSaveUser(ctx);
	if (err) err = dsConfigSaveLocal(ctx);
    } else if (ctx->defconfig == ctx->sysconfig) {
	err = dsConfigSaveSystem(ctx);
	if (err) err = dsConfigSaveUser(ctx);
	if (err) err = dsConfigSaveLocal(ctx);
    }

    return err;
}

int dsConfigSaveTo(DSConfigContext ctx, const char *fn) {
    int err;
    xmlDocPtr doc;
        
    ds_assert_param(ctx);
    ds_assert_param(fn);

/*
    doc = ctx->config;
    if (!doc) doc = ctx->userconfig;
    if (!doc) doc = ctx->sysconfig;
*/
    doc = ctx->defconfig;
    
    err = dsConfigLockReference(ctx, NULL, ctx->masterref);
    if (!err) {
	if (xmlSaveFormatFileEnc(fn, doc, "UTF-8", 1)<0) err = DS_ERR_XML_WRITER;
	dsConfigUnLockReference(ctx, ctx->masterref);
    }

    return err;
}

int dsConfigSaveLocal(DSConfigContext ctx) {
    int err;
    char *fn;
    
    fn = dsCreateFullName(NULL, ctx->dir, ctx->name, NULL);
    err = dsConfigSaveTo(ctx, fn);
    dsFree(NULL, fn);
    
    return err;
}

int dsConfigSaveUser(DSConfigContext ctx)  {
    int err;
    char *fn;
    
    fn = dsCreateFullName(NULL, ctx->userdir, ctx->name, NULL);
    err = dsConfigSaveTo(ctx, fn);
    dsFree(NULL, fn);
    
    return err;
}

int dsConfigSaveSystem(DSConfigContext ctx) {
    int err;
    char *fn;
    
    fn = dsCreateFullName(NULL, ctx->sysdir, ctx->name, NULL);
    err = dsConfigSaveTo(ctx, fn);
    dsFree(NULL, fn);
    
    return err;
}

static DSSubConfigContext dsSubConfigCreateContext(DSConfigContext ctx, DSConfigReference reference, unsigned char copy) {
    int err = 0;
    
    DSConfigReference ref;
    DSSubConfigContext sub;
    DSLockContext lock;
    
    ds_assert_null(ctx);
    ds_assert_null(reference);

    sub = (DSSubConfigContext)dsMalloc(ctx->malloc, sizeof(DSSubConfigContextS));
    lock = dsLockCreateContext(NULL);
    if (lock) err = dsLockInitContext(lock);
    if (copy) ref = dsConfigCopyReference(ctx, reference);
    else ref = reference; /* On failure destroying in any case */

    if ((err)||(!sub)||(!lock)||(!ref)) {
	if (ref) dsConfigDestroyReference(ctx, ref);
	if (sub) dsFree(NULL, sub);
	if (lock) dsLockDestroyContext(lock);
	return NULL;
    }
    
    sub->config = ctx;
    sub->ref = ref;
    sub->lock = lock;
    sub->master = NULL;

//    printf("Sub %s: %p\n", ref, lock);    
    return sub;
}

static void dsSubConfigDestroyContext(DSSubConfigContext sub) {
    ds_assert_return(sub);
    
    dsLockDestroyContext(sub->lock);
    dsConfigDestroyReference(sub->config, sub->ref);
    dsFree(sub->config->malloc, sub);
}

static int dsSubConfigMasterLock(DSSubConfigContext master) {
    int err;

    ds_assert_param(master);

    if (master->master) {
	err = dsSubConfigMasterLock(master->master);
	if (err) return err;
    }

//    printf("Master Lock: %p\n", master->lock);
    err = dsLockRead(master->lock);
    if ((err)&&(master->master)) dsSubConfigReturn(master->master);
    return err;
}

static int dsSubConfigLock(DSSubConfigContext sub) {
    int err;

    ds_assert_param(sub);

    if (sub->master) {
	err = dsSubConfigMasterLock(sub->master);
	if (err) return err;
    }

//    printf("Lock: %p\n", sub->lock);
    err = dsLockWrite(sub->lock);
    if ((err)&&(sub->master)) dsSubConfigReturn(sub->master);
    return err;
}

static void dsSubConfigReturn(DSSubConfigContext sub) {
    ds_assert_return(sub);

    dsLockReturn(sub->lock);
    if (sub->master) dsSubConfigReturn(sub->master);
}

static xmlXPathContextPtr dsConfigGetXPathContext(DSConfigContext ctx, xmlNodePtr config) {
    xmlDocPtr nodedoc = config->doc;
    
    if ((ctx->xp)&&(ctx->config==nodedoc)) return ctx->xp;
    if ((ctx->userxp)&&(ctx->userconfig==nodedoc)) return ctx->userxp;
    if ((ctx->sysxp)&&(ctx->sysconfig==nodedoc)) return ctx->sysxp;
    
    return NULL;
}


static xmlElementType dsConfigGetXPathType(DSConfigContext ctx, const char *xpath) {
    const char *ch;

    ch = dsStringStrrchr(xpath, '/');
    if (ch) ch++;
    else ch = xpath;

    if (*ch=='@') return XML_ATTRIBUTE_NODE;
    return XML_ELEMENT_NODE;
}

static xmlNodePtr dsConfigGetXPathParent(DSConfigContext ctx, xmlNodePtr config, const char *xpath) {
    size_t len;
    const char *ch;
    xmlChar *pxpath;
    xmlNodePtr parent;

    ch = dsStringStrrchr(xpath, '/');
    if (!ch) return config;

    len = (ch - xpath);

    pxpath = (xmlChar*)dsMalloca((len + 1)*sizeof(char));
    if (!pxpath) return NULL;
    
    memcpy(pxpath, xpath, len);
    pxpath[len] = 0;
    
    parent = dsConfigFindNode(ctx, config, pxpath);
    dsFreea(pxpath);
    
    return (parent->type == XML_ELEMENT_NODE)?parent:NULL;
}



static xmlNodePtr dsConfigFindNodeInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath) {
    xmlXPathContextPtr xp;
    xmlXPathObjectPtr obj;
    xmlNodeSetPtr node_set;
    xmlNodePtr node = NULL;

    xp = dsConfigGetXPathContext(ctx, config);
    if (!xp) return NULL;
    
    dsLockWrite(ctx->lock);
    xp->node = config;
    obj = xmlXPathEvalExpression((xmlChar*)xpath, xp);
    if (obj) {
	node_set = obj->nodesetval;
	if ((node_set)&&(node_set->nodeNr==1)) node = node_set->nodeTab[0];
	xmlXPathFreeObject(obj);
    }
    dsLockReturn(ctx->lock);

    
    return node;
}

static xmlNodePtr *dsConfigFindNodesInternal(DSConfigContext ctx, xmlNodePtr config,  const char *xpath) {
    int i;
    xmlXPathContextPtr xp;
    xmlXPathObjectPtr obj;
    xmlNodeSetPtr node_set;
    xmlNodePtr *nodes = NULL;

    xp = dsConfigGetXPathContext(ctx, config);
    if (!xp) return NULL;
    
    dsLockWrite(ctx->lock);
    xp->node = config;
    obj = xmlXPathEvalExpression((xmlChar*)xpath, xp);
    if (obj) {
	node_set = obj->nodesetval;
	if (node_set) nodes = dsMalloc(NULL, (node_set->nodeNr+1)*sizeof(xmlNodePtr));
	else nodes = dsMalloc(NULL, sizeof(xmlNodePtr));
	
	if (nodes) {
	    if (node_set) {
		for (i = 0; i < node_set->nodeNr; i++)
		    nodes[i] = node_set->nodeTab[i];
		nodes[i] = NULL;
	    } else nodes[0] = NULL;
	}
	xmlXPathFreeObject(obj);
    }
    dsLockReturn(ctx->lock);
    
    return nodes;
}

static int dsConfigSetNodeInternal2(DSConfigContext ctx, xmlNodePtr config, const char *xpath, xmlNodePtr value, unsigned char copy) {
    xmlNodePtr node = NULL;
    xmlNodePtr parent;
    
    
    node = dsConfigFindNodeInternal(ctx, config, xpath);
    if (node) {
	if (node->type != XML_ELEMENT_NODE) return DS_ERR_INVALID_ARGUMENT;
	if (value) {
	    if (dsConfigGetXPathType(ctx, xpath) != XML_ELEMENT_NODE) return DS_ERR_INVALID_ARGUMENT;
	    
	    if (copy) {
	        value = xmlDocCopyNode(value, node->doc, 1);
		if (!value) return DS_ERR_MEM;
	    } else if ((value->doc)&&(value->doc!=node->doc)) return DS_ERR_INVALID_ARGUMENT;

	    if (!xmlReplaceNode(node, value)) {
		if (copy) xmlFreeNode(value);
		return DS_ERR_XML;
	    } 
	} else xmlUnlinkNode(node);
	xmlFreeNode(node);
    } else if (value) {
	if (dsConfigGetXPathType(ctx, xpath) != XML_ELEMENT_NODE) return DS_ERR_INVALID_ARGUMENT;
	parent = dsConfigGetXPathParent(ctx, config, xpath);
	if ((!parent)||(parent->type != XML_ELEMENT_NODE)) return DS_ERR_INVALID_ARGUMENT;

	if (copy) {
	    value = xmlDocCopyNode(value, parent->doc, 1);
	    if (!value) return DS_ERR_MEM;
	} else if ((value->doc)&&(value->doc!=parent->doc)) return DS_ERR_INVALID_ARGUMENT;

	if (!xmlAddChild(parent, value)) {
	    if (copy) xmlFreeNode(value);
	    return DS_ERR_XML;
	}
	
	node = dsConfigFindNodeInternal(ctx, config, xpath);
	if (value != node) {
	    xmlUnlinkNode(value);
	    if (copy) xmlFreeNode(value);
	    return DS_ERR_INVALID_ARGUMENT;
	}
    }

    ctx->changed = 1;
    
    return 0;
}
#define dsConfigSetNodeInternal(ctx, config, xpath, value) dsConfigSetNodeInternal2(ctx, config, xpath, value, 1)


static int dsConfigSetStringInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, const char *value) {
    int err = 0;
    
    xmlNodePtr parent, node, test;
    const char *start, *attrs;
    const char *attrnameptr, *attrvalptr;
    size_t len, attrnamelen, attrvallen;
    xmlChar *name, *attrname, *attrval;
    const xmlChar *attrname_real;
    char quotes, *pos;

    node = dsConfigFindNodeInternal(ctx, config, xpath);
    if (node) {
	if (value) xmlNodeSetContent(node, value);
	else {
	    xmlUnlinkNode(node);
	    xmlFreeNode(node);
	}
	return 0;
    } else if (!value) return 0;
    
    parent = dsConfigGetXPathParent(ctx, config, xpath);
    if ((!parent)||(parent->type != XML_ELEMENT_NODE)) return DS_ERR_INVALID_ARGUMENT;
    
    start = dsStringStrrchr(xpath, '/');
    if (start) start++;
    else start = xpath;

    if (*start == '@') {	
	    // Setting attribute of the parent node
	xmlSetProp(parent, start+1, value);
    } else {
	// Creating new node and adding to the xpath parent
	attrs = xmlStrchr(start,'[');
	if (attrs) {
	    len = attrs - start;
	    name = (xmlChar*)dsMalloca((len+1)*sizeof(xmlChar));
	    if (!name) return DS_ERR_MEM;
	    memcpy(name, start, len); name[len] = 0;
	    node = xmlNewNode(NULL, name);
	    dsFreea(name);
	    
	    if (!node) return DS_ERR_MEM;
	    
	    while (!err) {
		start = xmlStrchr(attrs, '@');
		if (start) attrnameptr = start + 1;
		else break;
		
		start = xmlStrchr(attrnameptr, '=');
		if (start) attrvalptr = start + 1;
		else break;
		
		attrnamelen = (start - attrnameptr);
	    
		while ((*attrvalptr==' ')||(*attrvalptr=='\t')) attrvalptr++;

		quotes = *attrvalptr;
		if ((quotes=='"')||(quotes=='\'')) {
			// DS: check for escaped symbols
		    start = xmlStrchr(++attrvalptr, quotes);
		} else {
		    start = xmlStrchr(attrvalptr, ' ');
		    if (!start) start = xmlStrchr(attrs, ']');
		    if (!start) start = xmlStrchr(attrs, '\t');
		    if (!start) start = xmlStrchr(attrs, ')');
		    if (!start) start = xmlStrchr(attrs, '(');
		}
		
		if (start) attrs = start + 1;
		else break;
		
		attrvallen = (start - attrvalptr);
		
		attrname = (xmlChar*)dsMalloca((attrnamelen+1)*sizeof(xmlChar));
		attrval = (xmlChar*)dsMalloca((attrvallen+1)*sizeof(xmlChar));
		if ((!attrname)&&(!attrval)) {
		    err = DS_ERR_MEM;
		    if (attrname) { dsFreea(attrname); }
		    if (attrval) { dsFreea(attrval); }
		    break;
		}
		memcpy(attrname, attrnameptr, attrnamelen); attrname[attrnamelen] = 0;
		memcpy(attrval, attrvalptr, attrvallen); attrval[attrvallen] = 0;
		
		
		attrname_real = dsStringStrrchr(attrname, '@');
		if (attrname_real) attrname_real++;
		else attrname_real = attrname;
		
		pos = strchr(attrname_real, ' ');
		if (!pos) pos = strchr(attrname_real, '\t');
		if (pos) *pos = 0;

		xmlSetProp(node, attrname_real, attrval);
		
		dsFreea(attrname);
		dsFreea(attrval);
	    }
	} else
	    node = xmlNewNode(NULL, start);
	
	if (!node) return DS_ERR_MEM;
        xmlNodeSetContent(node, value);
	xmlAddChild(parent, node);

	    // Checking if everything is right
	test = dsConfigFindNodeInternal(ctx, config, xpath);
	if (test != node) {
	    xmlUnlinkNode(node);
	    xmlFreeNode(node);
	    return DS_ERR_INVALID_ARGUMENT;
	}
	
    }

    ctx->changed = 1;
    
    return 0;
}

static int dsConfigSetIntInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, int value) {
    char str[16];
    
    sprintf(str, "%i", value);
    return dsConfigSetStringInternal(ctx, config, xpath, str);
}

static int dsConfigSetLongInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, long value) {
    char str[16];
    
    sprintf(str, "%li", value);
    return dsConfigSetStringInternal(ctx, config, xpath, str);
}

static int dsConfigSetDoubleInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, double value) {
    char str[32];
    
    sprintf(str, "%lf", value);
    return dsConfigSetStringInternal(ctx, config, xpath, str);
}

static int dsConfigSetBooleanInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, unsigned char value) {
    if (value) return dsConfigSetStringInternal(ctx, config, xpath, "true");
    else return dsConfigSetStringInternal(ctx, config, xpath, "false");
}

static int dsConfigSetDateTimeInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, DSTime value) {
    char str[64];

    ds_assert_param(value);
    
    if (!dsTimeGetDateTime(value, str)) return DS_ERR_MEM;
    
    return dsConfigSetStringInternal(ctx, config, xpath, str);
}

static int dsConfigSetDurationInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, DSDuration value) {
    char str[64];

    ds_assert_param(value);
    
    if (!dsDurationGet(value, str)) return DS_ERR_MEM;
    
    return dsConfigSetStringInternal(ctx, config, xpath, str);
}

static int dsConfigSetDebugLevelInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, DSDebugPriority value) {
    const char *str;

    str = dsDebugGetName(value);
    if (!str) return DS_ERR_INVALID_ARGUMENT;

    return dsConfigSetStringInternal(ctx, config, xpath, str);
}

static xmlNodePtr dsConfigGetNodeInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, xmlNodePtr default_value) {
    xmlNodePtr node = NULL;

    node = dsConfigFindNodeInternal(ctx, config, xpath);
    if (node) return node;
    
    if ((ctx->autoset)&&(default_value)) {
	if (dsConfigSetNodeInternal2(ctx, config, xpath, default_value, 0)) return NULL;
	/* DS: Check. Do not copy, should be not connected to any document,
	otherwise NULL will be returned. (In the case then we are copying 
	default_value, the search should be performed again).
	return dsConfigFindNodeInternal(ctx, config, xpath); */
    }

    return default_value;
}

xmlNodePtr dsConfigGetSubNode(DSConfigContext ctx, xmlNodePtr config, const char *name, unsigned char create) {
    xmlNodePtr res;
    xmlNodePtr node;

    if (!create) return dsConfigGetNode(ctx, NULL, config, name);
        
    node = xmlNewNode(NULL, name);
    res = dsConfigGetNode(ctx, node, config, name);
    if (res != node) xmlFreeNode(node);

    return res;
}

static const char *dsConfigGetStringInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, const char *default_value) {
    xmlNodePtr node;

    node = dsConfigFindNodeInternal(ctx, config, xpath);
    if (node) {
	// xmlNodeGetContent - allocates new string
	switch (node->type) {
	    case XML_ELEMENT_NODE:
	    case XML_ATTRIBUTE_NODE:
	        if ((node->children)&&(node->children->type == XML_TEXT_NODE)&&(node->children->content)) return (const char*)node->children->content;
//	        if ((node->children)&&(node->children->type == XML_TEXT_NODE)&&(node->children->content)&&(*node->children->content)) return (const char*)node->children->content;
	    break;
	    case XML_TEXT_NODE:
	        if (node->content) return (const char*)node->content;
//	        if ((node->content)&&(*node->content)) return (const char*)node->content;
	    break;			
	    default:
	    break;
	}
    } 

    if (ctx->autoset) dsConfigSetStringInternal(ctx, config, xpath, default_value);
    return default_value;
}

static int dsConfigGetIntInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, int default_value) {
    const char *str;

    str = dsConfigGetStringInternal(ctx, config, xpath, NULL);
    if (str) return strtol(str, NULL, 0);

    if (ctx->autoset) dsConfigSetIntInternal(ctx, config, xpath, default_value);

    return default_value;
}

static long dsConfigGetLongInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, long default_value) {
    const char *str;

    str = dsConfigGetStringInternal(ctx, config, xpath, NULL);
    if (str) return strtol(str,NULL,0);

    if (ctx->autoset) dsConfigSetLongInternal(ctx, config, xpath, default_value);

    return default_value;
}

static double dsConfigGetDoubleInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, double default_value) {
    const char *str;

    str = dsConfigGetStringInternal(ctx, config, xpath, NULL);
    if (str) return atof(str);

    if (ctx->autoset) dsConfigSetDoubleInternal(ctx, config, xpath, default_value);

    return default_value;
}

static unsigned char dsConfigGetBooleanInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, unsigned char default_value) {
    const char *str;

    str = dsConfigGetStringInternal(ctx, config, xpath, NULL);
    if (str) {
	if ((str[0]=='T')||(str[0]=='t')||((str[0]>='1')&&(str[0]<='9'))) return 1;
	else return 0;
    }

    if (ctx->autoset) dsConfigSetBooleanInternal(ctx, config, xpath, default_value);

    return default_value;
}

static DSTime dsConfigGetDateTimeInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, DSTime default_value) {
    const char *str;

    ds_assert_default(default_value, default_value);
    
    str = dsConfigGetStringInternal(ctx, config, xpath, NULL);
    if ((str)&&(dsTimeSetDateTime(default_value, str))) return NULL;
    
    if (ctx->autoset) dsConfigSetDateTimeInternal(ctx, config, xpath, default_value);

    return default_value;
}

static DSDuration dsConfigGetDurationInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, DSDuration default_value) {
    const char *str;

    ds_assert_default(default_value, default_value);
    
    str = dsConfigGetStringInternal(ctx, config, xpath, NULL);
    if ((str)&&(dsDurationSet(default_value, str))) return NULL;
    
    if (ctx->autoset) dsConfigSetDurationInternal(ctx, config, xpath, default_value);

    return default_value;
}

static DSDebugPriority dsConfigGetDebugLevelInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, DSDebugPriority default_value) {
    DSDebugPriority level;
    const char *str;

    str = dsConfigGetStringInternal(ctx, config, xpath, NULL);
    if (str) {
	level = dsDebugGetLevelByName(str);
	if (level != (DSDebugPriority)-1) return level;
    }

    if (ctx->autoset) dsConfigSetDebugLevelInternal(ctx, config, xpath, default_value);

    return default_value;
}


static int dsConfigAddNodeInternal(DSConfigContext ctx, xmlNodePtr config, const char *xpath, xmlNodePtr value) {
    xmlNodePtr node = NULL;
    xmlNodePtr parent;

    ds_assert_param(value);

    if ((node->type != XML_ELEMENT_NODE)||(dsConfigGetXPathType(ctx, xpath) != XML_ELEMENT_NODE)) return DS_ERR_INVALID_ARGUMENT;

    parent = dsConfigGetXPathParent(ctx, config, xpath);
    if ((!parent)||(parent->type != XML_ELEMENT_NODE)) return DS_ERR_INVALID_ARGUMENT;
	
    value = xmlDocCopyNode(value, parent->doc, 1);
    if (!value) return DS_ERR_MEM;
    
    if (!xmlAddChild(parent, value)) {
	xmlFreeNode(value);
	return DS_ERR_XML;
    }

    /* DS: Check if really added. Get Nodes with XPath and check if value is in */

    ctx->changed = 1;

    return 0;
}

#define dsConfigFunctionVars() \
    DSString req; \
    va_list ap;

#define dsConfigFunctionStart() \
    if (strchr(xpath, '%')) { \
	va_start(ap, xpath); \
	req = dsStringVCreate(NULL, xpath, ap); \
	va_end(ap); \
    } else req = (DSString)xpath;


#define dsConfigFunctionEnd() \
    if (req != xpath) dsStringDestroy(NULL, req); \
\
    return res; \
}


#define dsConfigFunctionFind(name,type) \
type dsConfigFind##name(DSConfigContext ctx, xmlNodePtr config, const char *xpath, ...) { \
    type res; \
    dsConfigFunctionVars() \
    ds_assert_null(ctx); \
    ds_assert_null(config); \
    ds_assert_null(xpath); \
    dsConfigFunctionStart() \
    res = dsConfigFind##name##Internal(ctx, config, req); \
    if (!req) return NULL; \
    dsConfigFunctionEnd()

#define dsConfigFunctionGet(name,type) \
type dsConfigGet##name(DSConfigContext ctx, type default_value, xmlNodePtr config, const char *xpath, ...) { \
    type res; \
    dsConfigFunctionVars() \
    ds_assert_default(ctx, default_value); \
    ds_assert_default(config, default_value); \
    ds_assert_default(xpath, default_value); \
    dsConfigFunctionStart() \
    if (!req) return default_value; \
    res = dsConfigGet##name##Internal(ctx, config, req, default_value); \
    dsConfigFunctionEnd()

#define dsConfigFunctionSet(name,type) \
int dsConfigSet##name(DSConfigContext ctx, type default_value, xmlNodePtr config, const char *xpath, ...) { \
    int res; \
    dsConfigFunctionVars() \
    ds_assert_param(ctx); \
    ds_assert_param(config); \
    ds_assert_param(xpath); \
    dsConfigFunctionStart() \
    if (!req) return DS_ERR_MEM; \
    res = dsConfigSet##name##Internal(ctx, config, req, default_value); \
    dsConfigFunctionEnd()

#define dsConfigFunctionAdd(name,type) \
int dsConfigAdd##name(DSConfigContext ctx, type default_value, xmlNodePtr config, const char *xpath, ...) { \
    int res; \
    dsConfigFunctionVars() \
    ds_assert_param(ctx); \
    ds_assert_param(config); \
    ds_assert_param(xpath); \
    dsConfigFunctionStart() \
    if (!req) return DS_ERR_MEM; \
    res = dsConfigAdd##name##Internal(ctx, config, req, default_value); \
    dsConfigFunctionEnd()



dsConfigFunctionFind(Node, xmlNodePtr)
dsConfigFunctionFind(Nodes, xmlNodePtr*)

dsConfigFunctionGet(Node, xmlNodePtr)
dsConfigFunctionGet(String, const char*)
dsConfigFunctionGet(Int, int)
dsConfigFunctionGet(Long, long)
dsConfigFunctionGet(Double, double)
dsConfigFunctionGet(Boolean, unsigned char)
dsConfigFunctionGet(DateTime, DSTime)
dsConfigFunctionGet(Duration, DSDuration)
dsConfigFunctionGet(DebugLevel, DSDebugPriority)

dsConfigFunctionSet(Node, xmlNodePtr)
dsConfigFunctionSet(String, const char*)
dsConfigFunctionSet(Int, int)
dsConfigFunctionSet(Long, long)
dsConfigFunctionSet(Double, double)
dsConfigFunctionSet(Boolean, unsigned char)
dsConfigFunctionSet(DateTime, DSTime)
dsConfigFunctionSet(Duration, DSDuration)
dsConfigFunctionSet(DebugLevel, DSDebugPriority)

dsConfigFunctionAdd(Node, xmlNodePtr)



