#ifndef _DS_CONFIG_H
#define _DS_CONFIG_H

#include <stdarg.h>

#include <dsmalloc.h>
#include <dslock.h>
#include <dsxml.h>
#include <dshashtable.h>

#define DS_CONFIG_SKIP_LOCAL	1
#define DS_CONFIG_SKIP_USER	2
#define DS_CONFIG_SKIP_SYSTEM	4

typedef struct DSConfigContextT *DSConfigContext;
typedef struct DSSubConfigContextT *DSSubConfigContext;
typedef char *DSConfigReference; /* DS: Nodes with indentical path are identified by @id attribute */
typedef DSConfigReference *DSConfigReferenceSet;

struct DSConfigContextT {
    DSMallocContext malloc;
    DSLockContext lock;
    
    char *name;
    char *sysdir;
    char *userdir;
    char *dir;
    
    xmlDocPtr sysconfig;
    xmlDocPtr userconfig;
    xmlDocPtr config;
    
    xmlDocPtr defconfig;

    xmlXPathContextPtr sysxp;
    xmlXPathContextPtr userxp;
    xmlXPathContextPtr xp;

    unsigned long nodes;
    unsigned long nodesets;
    unsigned long fake_references;
    
    DSHashTable refprops;
    DSLockContext refprops_lock;
    
    DSHashTable subconfigs;
    DSConfigReference masterref;

    unsigned char autoset;
    unsigned char autoset_refprops;
    unsigned char changed;
};

typedef struct DSConfigContextT DSConfigContextS;

struct DSSubConfigContextT {
    DSConfigReference ref; /* DS: Assumes it is C string */
    DSLockContext lock;
    DSSubConfigContext master;
    DSConfigContext config;
};
typedef struct DSSubConfigContextT DSSubConfigContextS;

DSConfigContext dsConfigCreateContext(DSMallocContext mctx);
int dsConfigInitContext(DSConfigContext ctx, const char *name);
void dsConfigFreeContext(DSConfigContext ctx);
void dsConfigDestroyContext(DSConfigContext ctx);

int dsConfigSetSystemConfigDirectory(DSConfigContext ctx, const char *sysdir);
int dsConfigSetUserConfigDirectory(DSConfigContext ctx, const char *userdir);
int dsConfigSetConfigDirectory(DSConfigContext ctx, const char *dir);

int dsConfigEnableAutosetDefaults(DSConfigContext ctx);
int dsConfigDisableAutosetDefaults(DSConfigContext ctx);
int dsConfigEnableIDAttributeDetection(DSConfigContext ctx);
int dsConfigDisableIDAttributeDetection(DSConfigContext ctx);
int dsConfigRegisterIDAttribute(DSConfigContext ctx, const char *attr);


int dsConfigLoad(DSConfigContext ctx);
int dsConfigCustomLoad(DSConfigContext ctx, int flags);
void dsConfigUnload(DSConfigContext ctx);

xmlNodePtr dsConfigGetConfig(DSConfigContext ctx, xmlNodePtr node, const char *request, ...);
xmlNodePtr *dsConfigGetConfigs(DSConfigContext ctx, xmlNodePtr node, const char *request, ...);

void dsConfigReturnConfig(DSConfigContext ctx, xmlNodePtr node);
void dsConfigReturnConfigs(DSConfigContext ctx, xmlNodePtr *nodeset);

DSConfigReference dsConfigGetConfigReference(DSConfigContext ctx, xmlNodePtr node);
void dsConfigDestroyReference(DSConfigContext ctx, DSConfigReference ref);
DSConfigReference *dsConfigGetConfigsReferences(DSConfigContext ctx, xmlNodePtr *node);
void dsConfigDestroyReferences(DSConfigContext ctx, DSConfigReference *refset);

DSConfigReference dsConfigCopyReference(DSConfigContext ctx, DSConfigReference reference);
DSConfigReference dsConfigCreateFakeReference(DSConfigContext ctx);
int dsConfigReferenceSetMasterReference(DSConfigContext ctx, DSConfigReference reference, DSConfigReference master);
xmlNodePtr dsConfigGetConfigByReference(DSConfigContext ctx, DSConfigReference ref);

int dsConfigDumpMemory(DSConfigContext ctx, char **buf, size_t *size);
int dsConfigSave(DSConfigContext ctx);
int dsConfigSaveLocal(DSConfigContext ctx);
int dsConfigSaveUser(DSConfigContext ctx);
int dsConfigSaveSystem(DSConfigContext ctx);
int dsConfigSaveTo(DSConfigContext ctx, const char *fn);

//int dsConfigSetConfig(DSConfigContext ctx, xmlNodePtr currentconfig or null, parrent_node_xpath or null)

/*
 We are assuming following XPath expressions:
    <just anything specifying parrent node>/@attr
    <just anything specifying parrent node>/node[@attr1=value1 and @attr2=value2 and anything]
    
 
 dsConfigFindNode: 	NULL is returned if more than one node is found.
 dsConfigFindNodes: 	Returns NULL terminated array. Should be freed.
 dsConfigSetNode:	Copies value.
 dsConfigGetNode	Eats default value. Should not be from the other XML
			document, nodes not associated with any document are OK.
 dsConfigSetString:	NULL destroys attribute and node.
*/
xmlNodePtr dsConfigFindNode(DSConfigContext ctx, xmlNodePtr config, const char *xpath, ...);
xmlNodePtr *dsConfigFindNodes(DSConfigContext ctx, xmlNodePtr config,  const char *xpath, ...);

xmlNodePtr dsConfigGetNode(DSConfigContext ctx, xmlNodePtr default_value, xmlNodePtr config, const char *xpath, ...);
const char *dsConfigGetString(DSConfigContext ctx, const char *default_value, xmlNodePtr config, const char *xpath, ...);
int dsConfigGetInt(DSConfigContext ctx, int default_value, xmlNodePtr config, const char *xpath, ...);
long dsConfigGetLong(DSConfigContext ctx, long default_value, xmlNodePtr config, const char *xpath, ...);
double dsConfigGetDouble(DSConfigContext ctx, double default_value, xmlNodePtr config, const char *xpath, ...);
unsigned char dsConfigGetBoolean(DSConfigContext ctx, unsigned char default_value, xmlNodePtr config, const char *xpath, ...);
DSTime dsConfigGetDateTime(DSConfigContext ctx, DSTime default_value, xmlNodePtr config, const char *xpath, ...);
DSDuration dsConfigGetDuration(DSConfigContext ctx, DSDuration default_value, xmlNodePtr config, const char *xpath, ...);
DSDebugPriority dsConfigGetDebugLevel(DSConfigContext ctx, DSDebugPriority default_value, xmlNodePtr config, const char *xpath, ...);

int dsConfigSetNode(DSConfigContext ctx, xmlNodePtr value, xmlNodePtr config, const char *xpath, ...);
int dsConfigSetString(DSConfigContext ctx, const char *value, xmlNodePtr config, const char *xpath, ...);
int dsConfigSetInt(DSConfigContext ctx, int value, xmlNodePtr config, const char *xpath, ...);
int dsConfigSetLong(DSConfigContext ctx, long value, xmlNodePtr config, const char *xpath, ...);
int dsConfigSetDouble(DSConfigContext ctx, double value, xmlNodePtr config, const char *xpath, ...);
int dsConfigSetBoolean(DSConfigContext ctx, unsigned char value, xmlNodePtr config, const char *xpath, ...);
int dsConfigSetDateTime(DSConfigContext ctx, DSTime value, xmlNodePtr config, const char *xpath, ...);
int dsConfigSetDuration(DSConfigContext ctx, DSDuration value, xmlNodePtr config, const char *xpath, ...);
int dsConfigSetDebugLevel(DSConfigContext ctx, DSDebugPriority value, xmlNodePtr config, const char *xpath, ...);

int dsConfigAddNode(DSConfigContext ctx, xmlNodePtr value, xmlNodePtr config, const char *xpath, ...);

xmlNodePtr dsConfigGetSubNode(DSConfigContext ctx, xmlNodePtr config, const char *name, unsigned char create);
#define dsConfigGetNodeString(ctx, node, name, default_value) dsConfigGetString(ctx, default_value, node, name)
#define dsConfigGetPropString(ctx, node, name, default_value) dsConfigGetString(ctx, default_value, node, "@" name)
#define dsConfigGetNodeInt(ctx, node, name, default_value) dsConfigGetInt(ctx, default_value, node, name)
#define dsConfigGetPropInt(ctx, node, name, default_value) dsConfigGetInt(ctx, default_value, node, "@" name)
#define dsConfigGetNodeLong(ctx, node, name, default_value) dsConfigGetLong(ctx, default_value, node, name)
#define dsConfigGetPropLong(ctx, node, name, default_value) dsConfigGetLong(ctx, default_value, node, "@" name)
#define dsConfigGetNodeDouble(ctx, node, name, default_value) dsConfigGetDouble(ctx, default_value, node, name)
#define dsConfigGetPropDouble(ctx, node, name, default_value) dsConfigGetDouble(ctx, default_value, node, "@" name)
#define dsConfigGetNodeBoolean(ctx, node, name, default_value) dsConfigGetBoolean(ctx, default_value, node, name)
#define dsConfigGetPropBoolean(ctx, node, name, default_value) dsConfigGetBoolean(ctx, default_value, node, "@" name)
#define dsConfigGetNodeDateTime(ctx, node, name, default_value) dsConfigGetDateTime(ctx, default_value, node, name)
#define dsConfigGetPropDateTime(ctx, node, name, default_value) dsConfigGetDateTime(ctx, default_value, node, "@" name)
#define dsConfigGetNodeDuration(ctx, node, name, default_value) dsConfigGetDuration(ctx, default_value, node, name)
#define dsConfigGetPropDuration(ctx, node, name, default_value) dsConfigGetDuration(ctx, default_value, node, "@" name)
#define dsConfigGetNodeDebugLevel(ctx, node, name, default_value) dsConfigGetDebugLevel(ctx, default_value, node, name)
#define dsConfigGetPropDebugLevel(ctx, node, name, default_value) dsConfigGetDebugLevel(ctx, default_value, node, "@" name)

#define dsConfigSetNodeString(ctx, node, name, default_value) dsConfigSetString(ctx, default_value, node, name)
#define dsConfigSetPropString(ctx, node, name, default_value) dsConfigSetString(ctx, default_value, node, "@" name)
#define dsConfigSetNodeInt(ctx, node, name, default_value) dsConfigSetInt(ctx, default_value, node, name)
#define dsConfigSetPropInt(ctx, node, name, default_value) dsConfigSetInt(ctx, default_value, node, "@" name)
#define dsConfigSetNodeLong(ctx, node, name, default_value) dsConfigSetLong(ctx, default_value, node, name)
#define dsConfigSetPropLong(ctx, node, name, default_value) dsConfigSetLong(ctx, default_value, node, "@" name)
#define dsConfigSetNodeDouble(ctx, node, name, default_value) dsConfigSetDouble(ctx, default_value, node, name)
#define dsConfigSetPropDouble(ctx, node, name, default_value) dsConfigSetDouble(ctx, default_value, node, "@" name)
#define dsConfigSetNodeBoolean(ctx, node, name, default_value) dsConfigSetBoolean(ctx, default_value, node, name)
#define dsConfigSetPropBoolean(ctx, node, name, default_value) dsConfigSetBoolean(ctx, default_value, node, "@" name)
#define dsConfigSetNodeDateTime(ctx, node, name, default_value) dsConfigSetDateTime(ctx, default_value, node, name)
#define dsConfigSetPropDateTime(ctx, node, name, default_value) dsConfigSetDateTime(ctx, default_value, node, "@" name)
#define dsConfigSetNodeDuration(ctx, node, name, default_value) dsConfigSetDuration(ctx, default_value, node, name)
#define dsConfigSetPropDuration(ctx, node, name, default_value) dsConfigSetDuration(ctx, default_value, node, "@" name)
#define dsConfigSetNodeDebugLevel(ctx, node, name, default_value) dsConfigSetDebugLevel(ctx, default_value, node, name)
#define dsConfigSetPropDebugLevel(ctx, node, name, default_value) dsConfigSetDebugLevel(ctx, default_value, node, "@" name)

#endif /* _DS_CONFIG_H */
