#define _OPC_C

#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dsmalloc.h>
#include <dsdebug.h>
#include <dsxml.h>

#include <dssignal.h>
#include <dsbinaryconfiguration.h>
#include <dsbufferpoolsimplebuffer.h>

#include <opcerror.h>
#include <opcmodule.h>

#include <opcserverinterface.h>
#include <opc.h>

int opcInit() {
    int err;

	// XML Libraries Initialisation
    
    err = dsLibInit();
    if (!err) err=dsLocaleRegisterDomain(OPC_GETTEXT_DOMAIN,OPC_LOCALEDIR);
    if (!err) err=dsLocaleSetDomain(OPC_GETTEXT_DOMAIN);
    if (!err) err=opcErrorInit();
    if (!err) err=opcModuleInit();
    if (!err) err=opcPropertiesInit();
    return err;
}

void opcDestroy() {
    opcPropertiesDestroy();
    opcModuleDestroy();
    opcErrorDestroy();
    dsLibFree();
}

static void opcZeroContext(OPCContext ctx) {
    ds_assert_return(ctx);

    ctx->version=NULL;
    ctx->state=NULL;
    ctx->start_time=NULL;
    ctx->locales=NULL;
    ctx->tree=NULL;
    ctx->holder=NULL;
    ctx->config=NULL;
    
    ctx->fd_opcbinary=NULL;
    ctx->chunks=NULL;
    ctx->buffers=NULL;
    ctx->properties=NULL;
}

OPCContext opcCreateContext(DSMallocContext mctx) {
    OPCContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "opcCreateContext");

    ctx = (OPCContext)dsMalloc(mctx,sizeof(struct OPCContextT));
    if (ctx) {
	ctx->malloc = mctx;
	opcZeroContext(ctx);
    }
    return ctx;
}

void opcFreeContext(OPCContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcFreeContext");

    ds_assert_return(ctx);
    if (ctx->tree) opcTreeDestroyContext(ctx->tree);
    if (ctx->holder) dsHolderDestroyContext(ctx->holder);
    if (ctx->fd_opcbinary) dsOPCBinaryDestroyContext(ctx->fd_opcbinary);
    if (ctx->chunks) dsChunkPoolDestroyContext(ctx->chunks);
    if (ctx->buffers) dsBufferPoolDestroyContext(ctx->buffers);
    if (ctx->properties) opcPropertiesDestroyContext(ctx->properties);
    if (ctx->version) opcVersionDestroy(ctx->version);
    if (ctx->start_time) dsTimeDestroy(ctx->start_time);
    if (ctx->state) opcStateDestroyContext(ctx->state);
    if (ctx->locales) opcLocalesDestroyContext(ctx->locales);
    if (ctx->config) opcConfigurationDestroyContext(ctx->config);
    opcZeroContext(ctx);
}

void opcDestroyContext(OPCContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcDestroyContext");
    
    ds_assert_return(ctx);
    opcFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

int opcInitContext(OPCContext ctx, OPCConfigurationContext cfg, unsigned long scalability) {
    int err=0;

    dsDebugServer(DS_DEBUG_NOTICE, "opcInitContext");    

    ds_assert_param(ctx);
    
    opcZeroContext(ctx);
	// GetText-DS Library Initialisation
    
	// Initialising server structures
    if (cfg) ctx->config = cfg;
    else {
	ctx->config = opcConfigurationCreateContext(ctx->malloc);
	if (ctx->config) err=opcConfigurationInitContext(ctx->config, scalability);
	else err=DS_ERR_MEM;
    }
    if (!err) {
	ctx->holder = dsHolderCreateContext(ctx->malloc);
	if (ctx->holder) err=dsHolderInitContext(ctx->holder, ctx->config->holder_size);
	else err=DS_ERR_MEM;
	if (!err) {
	    dsHolderSetMaxSize(ctx->holder, ctx->config->holder_maximal_size);
	    dsHolderSetTimeout(ctx->holder, &ctx->config->resource_wait_timeout);
	}
    }
    if (!err) err=opcServerVersionSet(ctx,OPC_VERSION_MAJOR,OPC_VERSION_MINOR,OPC_VERSION_BUILD);
    if (!err) err=opcServerLocalesAdd(ctx,opc_supported_locales);    
    if (!err) err=opcServerVendorInfoSet(ctx,OPC_VENDOR_INFO);
    if (!err) err=opcServerStateSet(ctx,OPC_STATE_RUNNING,opc_string("All OK"));
    if (!err) err=opcServerStartTimeSet(ctx);
    if (!err) err=opcServerInitTree(ctx);

	// Initialising properties
    if (!err) {
	ctx->buffers = dsBufferPoolCreateContext(ctx->malloc);
	if (ctx->buffers) {
	    err = dsBufferPoolInitContext(ctx->buffers, ctx->config->buffer_pool_smallest, ctx->config->buffer_pool_bins);
	    if (!err) err = dsBufferPoolSetSimpleBufferFunctions(ctx->buffers);
	} else err = DS_ERR_BUFFERPOOL_CREATE;
    }
    if (!err) {
	ctx->fd_opcbinary = dsOPCBinaryCreateContext(ctx->malloc);
	if (ctx->fd_opcbinary) err = dsOPCBinaryInitContext(ctx->fd_opcbinary,ctx->config->opcbinary_schema,ctx->config->opcbinary_dictionaries);
	else err=dsErrorModule(OPC_ERR_CREATE, DS_MODULE_OPCBinary);
    }
    if (!err) {
	ctx->chunks = dsChunkPoolCreateContext(ctx->malloc);
	if (ctx->chunks) err = dsChunkPoolInitContext(ctx->chunks, ctx->config->chunk_pool_bins);
	else err=dsErrorModule(OPC_ERR_CREATE, DS_MODULE_ChunkPool);
    }
    if (!err) {
	ctx->properties = opcPropertiesCreateContext(ctx->malloc);
	if (ctx->properties) err=opcPropertiesInitContext(ctx->properties, ctx->config->properties_hash_size, ctx->config->property_maxid, ctx->config->property_userid);
	else err=OPC_ERR_PROPERTIES_CREATE;
    }

    return err;
}

void opcMantain(OPCContext ctx) {
    ds_assert_return(ctx);
    dsHolderMantain(ctx->holder);
    dsBufferPoolMantain(ctx->buffers);
}
