#include <string.h>

#include <dslib.h>
#include <dsdebug.h>
#include <dsassert.h>

#include <dsmalloc.h>
#include <dstools.h>
#include <dsdirectory.h>

#include <opcconfiguration.h>

void opcConfigurationZeroContext(OPCConfigurationContext ctx) {
    memset(&ctx->namespaces, 0, sizeof(OPCNamespacesS));

    ctx->data_directory = NULL;
    ctx->schema_directory = NULL;
    ctx->opc_schema = NULL;
    ctx->opcbinary_schema = NULL;
    ctx->dictionary_directory = NULL;
    ctx->opcbinary_dictionaries = NULL;
}

OPCConfigurationContext opcConfigurationCreateContext(DSMallocContext mctx) {
    OPCConfigurationContext ctx;
    ctx = (OPCConfigurationContext)dsMalloc(mctx, sizeof(struct OPCConfigurationContextT));
    if (ctx) {
	ctx->malloc = mctx;
	ctx->destructor = NULL;
	opcConfigurationZeroContext(ctx);
    }

    dsDebugServer(DS_DEBUG_NOTICE, "opcConfigurationCreateContext");

    return ctx;
}

int opcConfigurationInitContext(OPCConfigurationContext ctx, unsigned long scalability) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcConfigurationInitContext");
    opcConfigurationZeroContext(ctx);
    opcConfigurationDefault(ctx, scalability);
    dsDebugServer(DS_DEBUG_NOTICE, "opcConfigurationInitContext Done");
    return 0;
}

void opcConfigurationFreeContext(OPCConfigurationContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcConfigurationFreeContext");

    ds_assert_return(ctx);

    if (ctx->data_directory) dsExchangeString(ctx->malloc, &ctx->data_directory, NULL);
    if (ctx->schema_directory) dsFree(ctx->malloc, ctx->schema_directory);
    if (ctx->opcbinary_schema) dsFree(ctx->malloc, ctx->opcbinary_schema);
    if (ctx->dictionary_directory) dsFree(ctx->malloc, ctx->dictionary_directory);
    if (ctx->opcbinary_dictionaries) dsFree(ctx->malloc, ctx->opcbinary_dictionaries);

    if (ctx->namespaces.opcbinary_namespace) dsFree(ctx->malloc, ctx->namespaces.opcbinary_namespace);

    opcConfigurationZeroContext(ctx);
}

void opcConfigurationDestroyContext(OPCConfigurationContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcConfigurationDestroyContext");
    
    ds_assert_return(ctx);

    if (ctx->destructor) ctx->destructor(ctx);
    else {
        opcConfigurationFreeContext(ctx);
	dsFree(ctx->malloc, ctx);
    }
}


void opcConfigurationDefault(OPCConfigurationContext ctx, unsigned long scalability) {
    ds_assert_return(ctx);

    opcConfigurationSetBrowseTimeout(ctx, OPC_CONFIGURATION_BROWSE_TIMEOUT);
    opcConfigurationSetResourceWaitTimeout(ctx, OPC_CONFIGURATION_RESOURCE_TIMEOUT);
    opcConfigurationSetRequestTimeout(ctx, OPC_CONFIGURATION_REQUEST_TIMEOUT);
    opcConfigurationSetHolderSize(ctx, OPC_CONFIGURATION_HOLDER_SIZE, OPC_CONFIGURATION_HOLDER_MAXIMAL_SIZE);
    opcConfigurationSetFormatsHashSize(ctx, OPC_CONFIGURATION_FORMATS_HASH_SIZE);
    opcConfigurationSetPropertiesHashSize(ctx, OPC_CONFIGURATION_PROPERTIES_HASH_SIZE);
    opcConfigurationSetPropertiesIDs(ctx, OPC_CONFIGURATION_PROPERTY_MAXID, OPC_CONFIGURATION_PROPERTY_USERID);
    opcConfigurationSetChunkPoolConfig(ctx, OPC_CONFIGURATION_CHUNK_POOL_BINS);
    opcConfigurationSetBufferPoolConfig(ctx, OPC_CONFIGURATION_BUFFER_POOL_SMALLEST, OPC_CONFIGURATION_BUFFER_POOL_BINS);
    dsDebugServer(DS_DEBUG_NOTICE, "opcConfigurationInitContext: %s", OPC_CONFIGURATION_DATA_DIRECTORY);
    opcConfigurationSetDataDirectory(ctx, OPC_CONFIGURATION_DATA_DIRECTORY);
    dsDebugServer(DS_DEBUG_NOTICE, "opcConfigurationInitContext: %u", scalability);
}

int opcConfigurationSetBrowseTimeout(OPCConfigurationContext ctx, time_t browse_timeout) {
    ds_assert_param(ctx);
    dsDurationImportSeconds(&ctx->browse_timeout, browse_timeout);
    return 0;
}

int opcConfigurationSetResourceWaitTimeout(OPCConfigurationContext ctx, time_t resource_timeout) {
    ds_assert_param(ctx);
    dsDurationImportSeconds(&ctx->resource_wait_timeout, resource_timeout);
    return 0;
}

int opcConfigurationSetRequestTimeout(OPCConfigurationContext ctx, time_t request_timeout) {
    ds_assert_param(ctx);
    dsDurationImportSeconds(&ctx->request_timeout, request_timeout);
    return 0;
}

int opcConfigurationSetHolderSize(OPCConfigurationContext ctx, unsigned long holder_size, unsigned long holder_maximal_size) {
    ds_assert_param(ctx);
    ctx->holder_size = holder_size;
    ctx->holder_maximal_size = holder_maximal_size;
    return 0;
}

int opcConfigurationSetFormatsHashSize(OPCConfigurationContext ctx, unsigned long formats_hash_size) {
    ds_assert_param(ctx);
    ctx->formats_hash_size = formats_hash_size;
    return 0;
}

int opcConfigurationSetPropertiesHashSize(OPCConfigurationContext ctx, unsigned long properties_hash_size) {
    ds_assert_param(ctx);
    ctx->properties_hash_size = properties_hash_size;
    return 0;
}

int opcConfigurationSetPropertiesIDs(OPCConfigurationContext ctx, unsigned short maxid, unsigned short userid) {
    ds_assert_param(ctx);
    ctx->property_maxid = maxid;
    ctx->property_userid = userid;
    return 0;
}

int opcConfigurationSetChunkPoolConfig(OPCConfigurationContext ctx, unsigned char bins) {
    ds_assert_param(ctx);
    ds_assert_param(bins);
    
    ctx->chunk_pool_bins = bins;
    return 0;
}

int opcConfigurationSetBufferPoolConfig(OPCConfigurationContext ctx, unsigned long smallest, unsigned char bins) {
    ds_assert_param(ctx);
    ds_assert_param(smallest>15);
    ds_assert_param(bins);
    
    ctx->buffer_pool_smallest = smallest;
    ctx->buffer_pool_bins = bins;
    return 0;
}

int opcConfigurationSetDataDirectory(OPCConfigurationContext ctx, const char *data_directory) {
    int err;
    char *tmp;
    DSMallocContext mctx;

    ds_assert_param(ctx);
    ds_assert_param(data_directory);
    
    mctx = ctx->malloc;
    err = dsExchangeString(mctx, &ctx->data_directory, data_directory);
    if (err) return err;

    if (ctx->schema_directory) dsExchangeString(mctx, &ctx->schema_directory, NULL);
    ctx->schema_directory = dsDirectoryConcatenate(mctx, data_directory, OPC_CONFIGURATION_SCHEMA_DIRECTORY);
    if (!ctx->schema_directory) return DS_ERR_MEM;
    if (ctx->opc_schema) dsExchangeString(mctx, &ctx->opc_schema, NULL);
    ctx->opc_schema = dsDirectoryConcatenate(mctx, ctx->schema_directory, OPC_CONFIGURATION_OPC_SCHEMA);
    if (!ctx->opc_schema) return DS_ERR_MEM;
    if (ctx->opcbinary_schema) dsExchangeString(mctx, &ctx->opcbinary_schema, NULL);
    ctx->opcbinary_schema = dsDirectoryConcatenate(mctx, ctx->schema_directory, OPC_CONFIGURATION_OPCBINARY_SCHEMA);
    if (!ctx->opcbinary_schema) return DS_ERR_MEM;
    if (ctx->dictionary_directory) dsExchangeString(mctx, &ctx->dictionary_directory, NULL);
    ctx->dictionary_directory = dsDirectoryConcatenate(mctx, data_directory,OPC_CONFIGURATION_DICTIONARY_DIRECTORY);
    if (!ctx->dictionary_directory) return DS_ERR_MEM;
    if (ctx->opcbinary_dictionaries) dsExchangeString(mctx, &ctx->opcbinary_dictionaries, NULL);
    ctx->opcbinary_dictionaries = dsDirectoryConcatenate(mctx, ctx->dictionary_directory, OPC_CONFIGURATION_OPCBINARY_DICTIONARIES_DIRECTORY);
    if (!ctx->opcbinary_dictionaries) return DS_ERR_MEM;

    return 0;
}

int opcNamespacesSetOPCBinaryNamespace(OPCNamespaces ctx, const char *opcbinary_namespace) {
    ds_assert_param(ctx);
    ds_assert_param(opcbinary_namespace);
    if (dsExchangeString(((OPCConfigurationContext)ctx)->malloc, &ctx->opcbinary_namespace, opcbinary_namespace)) return DS_ERR_MEM;
    return 0;
}
