#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dsdebug.h>
#include <dserror.h>

#include <dsmalloc.h>
#include <dslocale.h>
#include <dshashtable.h>
#include <dslock.h>
#include <dsstringlist.h>
#include <dsallocatorsimple.h>

#include <opctree.h>
//#include <opcproperty.h>
#include <opcitem.h>
#include <opcdatainfo.h>

static int opcItemInitDataContext(OPCItemContext ctx);
static void opcItemFreeDataContext(OPCItemContext ctx);

void opcItemZeroContext(OPCItemContext ctx) {
    ctx->manual = 2;
}

OPCItemContext opcItemCreateContext(DSMallocContext mctx) {
    DSMallocContext mymalloc;
    OPCItemContext ctx;

    if (mctx) {
	mymalloc = dsMallocCreateContext(mctx);
	if (!mymalloc) return NULL;
    } else mymalloc = NULL;
    
    ctx = opcNewModule(Item, mymalloc);
    if (ctx) opcItemZeroContext(ctx);

    return ctx;
}

int opcItemInitContext(OPCItemContext ctx, const char *fullname, const char *title) {
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "opcItem(%p): InitContext: \"%s\".", ctx, fullname);    
    
    ds_assert_param(ctx);
    ds_assert_param(fullname);
    
    opcItemZeroContext(ctx);
    
    err = dsExchangeString(DS_MODULE(ctx)->malloc, &ctx->fullname, fullname);
    if (err) return DS_ERR_MEM;

/*
    ctx->name = strrchr(ctx->fullname, '/');
    if (ctx->name) ctx->name++;
    else ctx->name = ctx->fullname;
*/
    
    if (title) err=dsLocaleExchangeString(DS_MODULE(ctx)->malloc,&ctx->title, title);
    else err=dsLocaleExchangeString(DS_MODULE(ctx)->malloc,&ctx->title, fullname);
    
    if (err) {
	opcItemFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    ctx->lock = dsLockCreateContext(DS_MODULE(ctx)->malloc);
    if (!ctx->lock) {
	opcItemFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    err = dsLockInitContext(ctx->lock);
    if (err) {
	opcItemFreeContext(ctx);
	return err;
    }
    
    ctx->client_cond = dsConditionCreateContext(DS_MODULE(ctx)->malloc);
    if (!ctx->client_cond) {
	opcItemFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    err = dsConditionInitContext(ctx->client_cond);
    if (err) {
	opcItemFreeContext(ctx);
	return err;
    }
    
    ctx->subitems = dsStringListCreate(DS_MODULE(ctx)->malloc);
    if (!ctx->subitems) {
	opcItemFreeContext(ctx);
	return err;
    }

/*  This must be in over place and full destruction should occure only if haven't manual 
    sub(sub?)childs 
    
    ctx->pctx = opcPropertyGetInstance(sctx->pctx);
    if (!ctx->pctx) {
	opcItemFreeContext(ctx);
	return OPC_ERR_PROPERTIES_INSTANCE_INIT;
    }
*/	
    dsDebugServer(DS_DEBUG_INFO, "opcItem(%p): InitContext Ok", ctx);

    dsModuleMarkInitialized(ctx);
    
    return 0;
}

int opcItemFreeContext(OPCItemContext ctx) {
    int err = 0;
    
    dsDebugServer(DS_DEBUG_NOTICE, "opcItem(%p): FreeContext %s", ctx, (ctx->fullname)?ctx->fullname:NULL);

    ds_assert_param(ctx);

    
    if ((ctx->tree)&&(opcTreeGetStatus(ctx->tree)!=OPC_TREE_DESTROYING_FLAG)) {
	if ((ctx->tree)&&(ctx->status&OPC_ITEM_LINKED_FLAG)&&(!opcItemIsMarked(ctx))) {
	    err=opcTreeUnregisterItem(ctx->tree, ctx);
	    if (err==OPC_ERR_SERVERTREE_ITEM_HAVE_MANUAL_CHILDS) {
		opcItemResetManualMode(ctx);
	    } else {
		dsDebugServer(DS_DEBUG_CRITICAL, "opcItem(%p): Unregister failed: %s", ctx, dsError(err, NULL));
		err = 0;
	    }
	}
    }

    if (ctx->clients) {
	dsDebugServer(DS_DEBUG_DEBUG, "opcItem(%p): Waiting %lu clients to finish!", ctx, ctx->clients);
	
	dsConditionPreLock(ctx->client_cond);
	while (ctx->clients)
	    dsConditionWaitLocked(ctx->client_cond);
	dsConditionUnLock(ctx->client_cond);
	
	dsDebugServer(DS_DEBUG_DEBUG, "opcItem(%p): No more clients!", ctx);
    }

    ctx->status&=~OPC_ITEM_CONNECTED_FLAG;
    opcItemFreeDataContext(ctx);
	// ToDo: Destroy properties

    if (err==OPC_ERR_SERVERTREE_ITEM_HAVE_MANUAL_CHILDS) return err;


    if (ctx->subitems) dsStringListDestroy(ctx->subitems);

    // Unmark item
    if (ctx->client_cond) dsConditionDestroyContext(ctx->client_cond);
    if (ctx->lock) dsLockDestroyContext(ctx->lock);

    if (ctx->fullname) dsExchangeString(DS_MODULE(ctx)->malloc,&ctx->fullname,NULL);
    if (ctx->title) dsLocaleExchangeString(DS_MODULE(ctx)->malloc,&ctx->title,NULL);

    opcItemZeroContext(ctx);

    dsDebugServer(DS_DEBUG_INFO, "opcItem(%p): FreeContext Ok", ctx);    
    return 0;
}

void opcItemDestroyContext(OPCItemContext ctx) {
    DSMallocContext mymalloc;
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "opcItem(%p): DestroyContext", ctx);    

    err=opcItemFreeContext(ctx);
    if (err!=OPC_ERR_SERVERTREE_ITEM_HAVE_MANUAL_CHILDS) {
	mymalloc = DS_MODULE(ctx)->malloc;
	dsFree(mymalloc,ctx);
	if (mymalloc) dsMallocDestroyContext(mymalloc);
    }
}

int opcItemLock(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);

    dsDebugServer(DS_DEBUG_INFO, "opcItem %p: Lock", ctx);

    return dsLockWrite(ctx->lock);
}

int opcItemUnlock(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);

    dsDebugServer(DS_DEBUG_INFO, "opcItem %p: Unlock", ctx);
    
    return dsLockReturn(ctx->lock);
}

int opcItemGet(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);

    dsDebugServer(DS_DEBUG_INFO, "opcItem %p: Get", ctx);
    
    dsLockWrite(ctx->lock);
    ctx->clients++;
    dsLockReturn(ctx->lock);

    dsConditionBroadcast(ctx->client_cond);

    return 0;
}

int opcItemReturn(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);

    dsDebugServer(DS_DEBUG_INFO, "opcItem %p: Return", ctx);
    
    dsLockWrite(ctx->lock);
    ctx->clients--;
    dsLockReturn(ctx->lock);

    dsConditionBroadcast(ctx->client_cond);

    return 0;
}


int opcItemSetAssociation(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ctx->status|=OPC_ITEM_CONNECTED_FLAG;
    return 0;
}

int opcItemSetManualMode(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    
    dsLockWrite(ctx->lock);
    ctx->manual=1;
    dsLockReturn(ctx->lock);

    return 0;
}

int opcItemResetManualMode(OPCItemContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    
    dsLockWrite(ctx->lock);
    ctx->manual=0;
    dsLockReturn(ctx->lock);

    return 0;
}

int opcItemSetVendorFilter(OPCItemContext ctx, OPCVendorFilter vendor_filter) {
    ds_assert_param(ctx);
    
    dsLockWrite(ctx->lock);
    ctx->vendor_filter = vendor_filter;
    dsLockReturn(ctx->lock);
    return 0;
}

int opcItemSetTitle(OPCItemContext ctx, const char *title) {
    ds_assert_param(ctx);
    ds_assert_param(title);
    dsLockWrite(ctx->lock);
    dsLocaleExchangeString(DS_MODULE(ctx)->malloc,&ctx->title,title);
    dsLockReturn(ctx->lock);

    return 0;
}

dschar opcItemGetTitle(OPCItemContext ctx, const char *locale) {
    return dsLocaleGetText(ctx->title, locale);
}

static char *opcItemGetFullName_internal(OPCItemContext ctx) {
    char *res = NULL;
    ds_assert_null(ctx->fullname);
    dsExchangeString(DS_MODULE(ctx)->malloc, &res, ctx->fullname);
    return res;
}

char *opcItemGetFullName(OPCItemContext ctx) {
    char *res;
    ds_assert_null(ctx);
    ds_assert_null(ctx->lock);

    dsLockRead(ctx->lock);
    res = opcItemGetFullName_internal(ctx);
    dsLockReturn(ctx->lock);
    
    return res;
}

const char *opcItemGetFullNamePtr(OPCItemContext ctx) {
    ds_assert_null(ctx);
    return (const char*)ctx->fullname;
}

int opcItemBrowse(OPCItemContext ctx, OPCItemBrowseFilter browse_filter, const char *name_filter, const char *vendor_filter, DSStringList *result) {
    OPCItemContext ictx;
    DSStringList sl, res;
    char *fullname;
    int err;

    ds_assert_param(ctx);
    ds_assert_param(result);
    ds_assert_param(ctx->lock);
    
    *result = NULL;

    dsLockRead(ctx->lock);
    
    if ((ctx->subitems)&&(ctx->status&OPC_ITEM_LINKED_FLAG)) {
	res = dsStringListCreate(DS_MODULE(ctx)->malloc?DS_MODULE(ctx)->malloc->parrent:NULL);
	if (!res) err = DS_ERR_MEM;
	else err = 0;
    } else  err = OPC_ERR_SERVERTREE_ITEM_NOTLINKED;
    
    if (!err) {
	for (sl=ctx->subitems;sl;sl=sl->next) {
	    if (sl->item) {
		ictx = *(OPCItemContext*)sl->item;
		if (ictx) {
		    dsLockRead(ictx->lock);
		    if (((!browse_filter)||(!opcItemBrowseFilter(ictx, browse_filter)))&&
			((!name_filter)||(!opcItemNameFilter(ictx, name_filter)))&&
			((!vendor_filter)||((ctx->vendor_filter)&&(!ctx->vendor_filter(ictx, vendor_filter))))) {
			    fullname = ictx->fullname;
			    if (fullname) dsStringListAdd(res, fullname, (unsigned long)(strlen(fullname) + 1));
		    }
		    dsLockReturn(ictx->lock);
		}
	    }
	}
    }
    
    dsLockReturn(ctx->lock);
    
    if (err) *result=NULL;
    else *result=res;
    
    return err;
}

/*******************************************************************************
**************************** Non-Locking Functions  ****************************
*******************************************************************************/

int opcItemHaveChilds(OPCItemContext ctx) {
    ds_assert_param(ctx);
    
    if (ctx->status&OPC_ITEM_LINKED_FLAG) {
	ds_assert_param(ctx->subitems);
	return (!dsStringListIsEmpty(ctx->subitems));
    } 

    return 0;
}

int opcItemIsConnected(OPCItemContext ctx) {
    ds_assert_param(ctx);
    
    return ctx->status&OPC_ITEM_CONNECTED_FLAG;
}

/*******************************************************************************
*********************************** Filters ************************************
*******************************************************************************/
int opcItemBrowseFilter(OPCItemContext ctx, OPCItemBrowseFilter filter) {
    switch(filter) {
	case BRANCH:
	    if ((ctx->status&OPC_ITEM_CONNECTED_FLAG)&&(opcItemHaveChilds(ctx))) 
		return OPC_ERR_ITEM_FILTERED;
	case ITEM:
	    if ((ctx->status&OPC_ITEM_CONNECTED_FLAG)==0) 
		return OPC_ERR_ITEM_FILTERED;
    }
    return 0;
}


int opcItemNameFilter(OPCItemContext ctx, const char *filter) {
    return 0;
}


/*******************************************************************************
************************* Item SubItems Managment ******************************
*******************************************************************************/

int opcItemLink(OPCItemContext ctx, OPCItemContext ictx) {
    int err;
    ds_assert_param(ctx);
    ds_assert_param(ctx->subitems);
    ds_assert_param(ictx);
    ds_assert_param(ctx->lock);
    ds_assert_param(ictx->lock);
    
    dsLockWrite(ctx->lock);
    err=dsStringListAdd(ctx->subitems, (void*)&ictx, sizeof(OPCItemContext));
    dsLockReturn(ctx->lock);
    return err;
}

int opcItemUnlink(OPCItemContext ctx, OPCItemContext ictx) {
    ds_assert_param(ctx);
    ds_assert_param(ictx);
    ds_assert_param(ctx->lock);
    ds_assert_param(ictx->lock);

    dsLockWrite(ctx->lock);
    opcItemMarkParrentsCallback(ctx, ictx);
    opcTreeCleanMarked(ctx->tree);
    dsLockReturn(ctx->lock);
    return 0;
}
/*******************************************************************************
******************************* Tree Analysis **********************************
*******************************************************************************/

int opcItemHaveManualSubChilds(OPCItemContext ctx) {
    OPCItemContext ictx;
    DSStringList sl;
    
    ds_assert_param(ctx);
    ds_assert_param(ctx->subitems);
    
    for (sl=ctx->subitems;sl;sl=sl->next) {
	ictx = *(OPCItemContext*)sl->item;
	if (ictx) {
	    if (ictx->manual) return 1;
	    if (opcItemHaveManualSubChilds(ictx)) return 1;
	}
    }
    return 0;
}

int opcItemSetTree(OPCItemContext ctx, OPCTreeContext tree) {
    ds_assert_param(ctx);
    ds_assert_param(tree);
    ds_assert_param(ctx->tree==NULL);

    ctx->tree = tree;
    return 0;
}

/*******************************************************************************
************************* Deletion Marks Managment *****************************
*******************************************************************************/

void opcItemMark(OPCItemContext ctx) {
    ds_assert_return(ctx);
    ds_assert_return(ctx->lock);
    
    dsDebugServer(DS_DEBUG_INFO, "opcItem %p: Mark", ctx);
    
    if (!ctx->mark) {
	dsLockWrite(ctx->lock);
	ctx->mark=1;
    }
}

void opcItemMarkSubTree(OPCItemContext ctx) {
    DSStringList sl;
    OPCItemContext ictx;
    
    ds_assert_return(ctx);
    
    opcItemMark(ctx);
    for (sl=ctx->subitems;sl;sl=sl->next) {
    	ictx = *(OPCItemContext*)sl->item;
	if (ictx) opcItemMarkSubTree(ictx);
    }
}

void opcItemUnMark(OPCItemContext ctx) {
    ds_assert_return(ctx);
    ds_assert_return(ctx->lock);

    dsDebugServer(DS_DEBUG_INFO, "opcItem %p: UnMark", ctx);
    
    if (ctx->mark) {
	ctx->mark=0;
	dsLockReturn(ctx->lock);
    }
}

void opcItemUnMarkSubTree(OPCItemContext ctx) {
    DSStringList sl;
    OPCItemContext ictx;

    ds_assert_return(ctx);
    
    opcItemUnMark(ctx);
    for (sl=ctx->subitems;sl;sl=sl->next) {
    	ictx = *(OPCItemContext*)sl->item;
	if (ictx) opcItemUnMarkSubTree(ictx);
    }
}

int opcItemIsMarked(OPCItemContext ctx) {
    ds_assert_param(ctx);
    return ctx->mark;
}

/*******************************************************************************
************************* Tree managment callbacks *****************************
*******************************************************************************/


int opcItemMarkChildsCallback(OPCItemContext ctx, OPCItemContext ictx) {
    OPCItemContext ictx2;
    DSStringList sl;
    
    if ((ctx==ictx)||((opcItemIsMarked(ctx))&&(!ctx->manual))) return 0;
    
    for (sl=ctx->subitems;sl;sl=sl->next) {
	ictx2 = *(OPCItemContext*)sl->item;
        if ((ictx2)&&(opcItemIsMarked(ictx2))) opcItemUnMarkSubTree(ictx2);
    }

    return 0;
}

int opcItemMarkParrentsCallback(OPCItemContext ctx, OPCItemContext ictx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->subitems);
    ds_assert_param(ctx->tree);
    ds_assert_param(ictx);

    if (dsStringListRemove(ctx->subitems, (void*)&ictx, sizeof(OPCItemContext*)))
	opcTreeMarkUnused(ctx->tree, ctx);
    return 0;
}

/*******************************************************************************
*************************** Properties and Data ********************************
*******************************************************************************/
static int opcItemInitDataContext(OPCItemContext ctx) {
    int err;
    
    return 0;
}

static void opcItemFreeDataContext(OPCItemContext ctx) {
    ds_assert_return(ctx);
    if (ctx->dm) dsDataManagerUnref(ctx->dm);
}

DSDataManagerContext opcItemRegisterDataManager(OPCItemContext ctx, DSDataManagerContext dm) {
    int err;

    ds_assert_null(ctx);
    
    dsLockWrite(ctx->lock);
    if (ctx->dm) {
	dsLockReturn(ctx->lock);
	return NULL;
    }
/*
    err = opcItemInitDataContext(ctx);
    if (err) {
	opcItemFreeDataContext(ctx);
	dsLockReturn(ctx->lock);
	return NULL;
    } else */ ctx->dm = (DSDataManagerContext)dm;
    
    dsLockReturn(ctx->lock);
    
    ctx->native_representation = dsDataManagerGetFormByName(dm, DS_DATA_MANAGER_FORM_NATIVE);
    ctx->opcxml_representation = dsDataManagerGetFormByName(dm, DS_DATA_MANAGER_FORM_OPCXML);
	
    if (ctx->opcxml_representation 	!= DS_DATA_CHUNK_FORM_INVALID) {
	ctx->representation = ctx->opcxml_representation;
    } else if (ctx->native_representation 	!= DS_DATA_CHUNK_FORM_INVALID) {
	ctx->representation = ctx->native_representation;
    } else {
	ctx->representation = 0;
    }
    
    return dm;
}


DSDataManagerFastContext opcItemRegisterFastDataManager(OPCItemContext ctx, DSBinaryType type, DSDataManagerFastOptions options, unsigned long flags) {
    int err;
    DSDataManagerFastContext dm;
    struct DSDataManagerFastOptionsT opts;
    OPCContext sctx = NULL;

    ds_assert_null(ctx);
    ds_assert_null(type);

    ds_assert_null(ctx->tree);
    sctx = ctx->tree->sctx;
    ds_assert_null(sctx);
    

    if (!options) {
	dsDataManagerFastInitDefaultOptions(&opts);
	options = &opts;
    }
    
    dsDataManagerOptionsSetMissingChunkPool((DSDataManagerOptions)options, sctx->chunks);
    dsDataManagerFastOptionsSetMissingBufferPool(options, sctx->buffers);
    dsDataManagerFastOptionsSetFixedAllocationSize(options, type->size);
    
    dm = dsDataManagerFastCreateContext(DS_MODULE(ctx)->malloc);
    if (dm) {
	err = dsDataManagerFastInitContext(dm, type, options);
	if (err) {
	    dsDataManagerUnref((DSDataManagerContext)dm);
	    return NULL;
	} 
    } return NULL;

    if (!opcItemRegisterDataManager(ctx, (DSDataManagerContext)dm)) {
	dsDataManagerUnref((DSDataManagerContext)dm);
	return NULL;
    }
    
    return dm;
}

DSDataManagerRepContext opcItemRegisterLegacyDataManager(OPCItemContext ctx, DSBinaryType type, DSDataManagerRepOptions options, unsigned long flags) {
    int err;
    DSDataManagerRepContext dm;
    struct DSDataManagerRepOptionsT opts;
    OPCContext sctx = NULL;

    ds_assert_null(ctx);
    ds_assert_null(type);

    ds_assert_null(ctx->tree);
    sctx = ctx->tree->sctx;
    ds_assert_null(sctx);
    

    if (!options) {
	dsDataManagerRepInitDefaultOptions(&opts);
	options = &opts;
    }

    if (flags) dsDataManagerOptionsSetFlags((DSDataManagerOptions)options, flags);
    dsDataManagerOptionsSetMissingChunkPool((DSDataManagerOptions)options, sctx->chunks);
    dsDataManagerRepOptionsSetMissingBufferPool(options, sctx->buffers);
    dsDataManagerRepOptionsSetFixedAllocationSize(options, type->size);
    
    dm = dsDataManagerRepCreateOPCXML(DS_MODULE(ctx)->malloc, type, options);
    if (!dm) return NULL;

    if (!opcItemRegisterDataManager(ctx, (DSDataManagerContext)dm)) {
	dsDataManagerUnref((DSDataManagerContext)dm);
	return NULL;
    }

    return dm;
}


void opcItemRemoveDataManager(OPCItemContext ctx) {
    DSDataManagerContext dm;
    ds_assert_return(ctx);

    opcItemDisableDataManager(ctx);

    dsLockWrite(ctx->lock);    
    if (ctx->dm) {
	dm = ctx->dm;
	ctx->dm = NULL;
	dsDataManagerUnref(dm);
    }
    dsLockReturn(ctx->lock);
    
}

int opcItemEnableDataManager(OPCItemContext ctx) {
    ds_assert_param(ctx);
    
    dsLockWrite(ctx->lock);
    if (ctx->dm) ctx->status|=OPC_ITEM_CONNECTED_FLAG;
    dsLockReturn(ctx->lock);
    
    return 0;
}

int  opcItemDisableDataManager(OPCItemContext ctx) {
    ds_assert_param(ctx);
    
    dsLockWrite(ctx->lock);
    ctx->status&=~OPC_ITEM_CONNECTED_FLAG;
    dsLockReturn(ctx->lock);
    
    return 0;
}


DSDataManagerContext opcItemGetDataManager(OPCItemContext ctx) {
    return NULL;
}

void opcItemReturnDataManager(OPCItemContext ctx) {
}

int opcItemCheckConnection(OPCItemContext ctx) {
    if (!(ctx->status&OPC_ITEM_CONNECTED_FLAG)) return OPC_ERR_ITEM_NOTCONNECTED;
    ds_assert_param(ctx->dm);

    return 0;
}

DSTimeConst opcItemGetTimestamp(OPCItemContext ctx) {
    ds_assert_null(ctx);
    
    return opcDataInfoGetTime(&ctx->info);
}

int opcItemCheckAge(OPCItemContext ctx, const DSTime current_time, const DSDuration maxage) {
    DSTimeS time;

    ds_assert_param(ctx);
    ds_assert_param(maxage);
    
//    if (!(ctx->status&OPC_ITEM_CONNECTED_FLAG)) return OPC_ERR_ITEM_NOTCONNECTED;
    
    return 0;
//    if (NOT_APPLICABLE) return 0;
/*    
    if (!current_time) dsTimeCurrentTime(&time);
    else dsTimeCopy(&time,current_time);
    dsTimeSub(&time,maxage);
    if (dsTimeCmp(&time, &ctx->timestamp)>0) {
//	if (RELOADING_APPLICABLE) OPC_ERR_ITEM_VALUE_OUTOFDATE_REACQUIRING;
	// Ask item to reload if possible otherwise
	
	return OPC_ERR_ITEM_OUTOFDATE;
    }
    return 0;
*/
}


int opcItemGetData(OPCItemContext ctx, DSData *info, DSData *data) {
    int err;
    ds_assert_param(ctx);
    
    dsLockRead(ctx->lock);
    if (ctx->status&OPC_ITEM_CONNECTED_FLAG) {
	dsDataManagerRef(ctx->dm);
	dsLockReturn(ctx->lock);

	err = dsDataManagerGetData(ctx->dm, ctx->representation, info, data);
	dsDataManagerUnref(ctx->dm);
	
	return err;
    }
    dsLockReturn(ctx->lock);
    
    return OPC_ERR_ITEM_NOTCONNECTED;
}

void opcItemReturnData(OPCItemContext ctx, DSData info) {
    dsLockRead(ctx->lock);
    
    if (ctx->status|OPC_ITEM_CONNECTED_FLAG) {
	dsDataManagerRef(ctx->dm);
	dsLockReturn(ctx->lock);
	dsDataManagerReturnData(ctx->dm, info);
	dsDataManagerUnref(ctx->dm);
    } else dsLockReturn(ctx->lock);
}

int opcItemPushData(OPCItemContext ctx, DSDataChunkForm repid, DSData info, DSData data) {
    int err;
    
    OPCDataInfoS default_info;
    DSDataS info_holder;

    ds_module_assert(ctx);
    ds_assert_param(data);
    
    if (!info) {
	opcDataInfoSetup(&default_info);
	opcDataInfoEnclosureSetup(&info_holder, &default_info);

	info = &info_holder;
    }
    
    dsLockRead(ctx->lock);
    if (ctx->status&OPC_ITEM_CONNECTED_FLAG) {
	dsDataManagerRef(ctx->dm);
	dsLockReturn(ctx->lock);
	
        err = dsDataManagerPushData(ctx->dm, repid, info, data);
	dsDataManagerUnref(ctx->dm);
	
	if (err) return err;
	
	opcDataInfoCopy(&ctx->info, opcDataInfoEnclosureExtract(info));
	return 0;
    }
    dsLockReturn(ctx->lock);

    return OPC_ERR_ITEM_NOTCONNECTED;
}

int opcItemPushNativeData(OPCItemContext ctx, DSData info, DSData data) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->native_representation != DS_DATA_CHUNK_FORM_INVALID);

    return opcItemPushData(ctx, ctx->native_representation, info, data);
}

int opcItemPushNativePointer(OPCItemContext ctx, DSData info, void *ptr, size_t size) {
    DSDataS data;
    
    ds_assert_param(ctx);
    ds_assert_param(ptr);
    
    dsDataSetPointer(&data, NULL, ptr, size);
    
    return opcItemPushNativeData(ctx, info, &data);
}

int opcItemPushNativeAllocation(OPCItemContext ctx, DSData info, DSAllocatorContext allocator, DSAllocation allocation) {
    DSDataS data;
    
    ds_assert_param(ctx);
    ds_assert_param(allocator);
    ds_assert_param(allocation);
    
    dsDataSetAllocation(&data, allocator, allocation, 0);
    
    return opcItemPushNativeData(ctx, info, &data);
}


int opcItemPushOPCXMLData(OPCItemContext ctx, DSData info, DSData data) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->opcxml_representation != DS_DATA_CHUNK_FORM_INVALID);

    return opcItemPushData(ctx, ctx->opcxml_representation, info, data);
}

int opcItemPushOPCXMLNode(OPCItemContext ctx, DSData info, xmlNodePtr node) {
    DSDataS data;

    ds_assert_param(ctx);
    ds_assert_param(node);
    
    return DS_ERR_NOTIMPLEMENTED;

//    return opcItemPushOPCXMLData(ctx, info, data);
}

int opcItemPushOPCXMLString(OPCItemContext ctx, DSData info, xmlChar *string) {
    DSDataS data;

    ds_assert_param(ctx);
    ds_assert_param(string);
    
    return DS_ERR_NOTIMPLEMENTED;

//    return opcItemPushOPCXMLData(ctx, info, data);
}


/*
int opcItemPush(OPCItemContext ctx, DSData info, void *data, unsigned long size) {
    DSDataS data;
    
    dsDataSetupPointer(&data, NULL, data, size);
    return opcItemPushData(ctx, info, data);
}

int opcItemPushAllocation(OPCItemContext
*/
// ToDo: GetFirst, GetNext
