#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <dsmalloc.h>

#include <opc.h>
#include <opclocales.h>

static void opcLocalesZeroContext(OPCLocalesContext ctx) {
    ds_assert_return(ctx);
    ctx->locales = NULL;
    ctx->lock = NULL;
}

OPCLocalesContext opcLocalesCreateContext(DSMallocContext mctx) {
    OPCLocalesContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "opcLocalesCreateContext");    

    ctx = (OPCLocalesContext)dsMalloc(mctx, sizeof(struct OPCLocalesContextT));
    if (ctx) {
	ctx->malloc = mctx;
	opcLocalesZeroContext(ctx);
    }
    return ctx;
}

int opcLocalesInitContext(OPCLocalesContext ctx) {
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "opcLocalesInitContext");    
    ds_assert_param(ctx);

    opcLocalesZeroContext(ctx);
    ctx->lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) return DS_ERR_MEM;
    err = dsLockInitContext(ctx->lock);
    if (err) {
	opcLocalesFreeContext(ctx);
	return err;
    }
    ctx->locales = dsHashTableInit(ctx->malloc,opc_supported_locales_number,DS_HASH_TYPE_ARRAY,NULL);
    if (!ctx->locales) {
	opcLocalesFreeContext(ctx);
	return DS_ERR_MEM;
    }
    err=dsHashTableAdd(ctx->locales,opc_default_locale, 0);
    if (err) {
	opcLocalesFreeContext(ctx);
	return err;
    }
    return 0;
}

void opcLocalesFreeContext(OPCLocalesContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcLocalesFreeContext");    

    ds_assert_return(ctx);
    if (ctx->locales) dsHashTableDestroy(ctx->locales);
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    opcLocalesZeroContext(ctx);
}

void opcLocalesDestroyContext(OPCLocalesContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcLocalesDestroyContext");    

    ds_assert_return(ctx);
    opcLocalesFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}


int opcLocalesAddLocale(OPCLocalesContext ctx, const char *newlocale) {
    int err;
    DSHashTable locales;
    DSLockContext lock;

    ds_assert_param(ctx);
    locales = ctx->locales;
    ds_assert_param(locales);
    lock = ctx->lock;
    ds_assert_param(lock);
    
    dsLockWrite(lock);
    err=dsHashTableAdd(ctx->locales,(void*)newlocale, 0);
    dsLockReturn(lock);
    return err;
}

int opcLocalesAddLocales(OPCLocalesContext ctx, const char **newlocales) {
    unsigned long i;
    int err = 0;
    
    ds_assert_param(ctx);
    
    if (newlocales) {
	for (i=0;newlocales[i];i++) {
	    if (!err) opcLocalesAddLocale(ctx,newlocales[i]);
	}
    }
    return err;
}

OPCLocale opcLocalesGetLocale(OPCLocalesContext ctx, char *attr) {
    DSHashTable locales;
    OPCLocale lctx;
    char *locale = NULL;
    size_t i;
    
    ds_assert_null(ctx);
    locales = ctx->locales;
    ds_assert_null(locales);
    
    lctx = opcLocaleCreate();
    if (!lctx) return NULL;

    if (attr) {
	i=strlen(attr);
	if (i>5) attr[5]=0;
	if ((i>2)&&(attr[2]=='-')) attr[2]='_';
	locale=(char*)dsHashTableFind(locales,attr);
	if ((!locale)&&(i>2)) {
	    attr[2]=0;
	    locale=(char*)dsHashTableFind(locales,attr);
	}
    }

    if (locale) {
	opcLocaleSetLocale(lctx,locale);
	if ((locale)&&(i>5)) opcLocaleSetCharset(lctx,attr+6);
	else opcLocaleSetCharset(lctx,opc_default_charset);
    } else {
	opcLocaleSetLocale(lctx,opc_default_locale);
	opcLocaleSetCharset(lctx,opc_default_charset);
    }
    return lctx;
}

int opcLocalesList(OPCLocalesContext ctx, OPCLocalesListFunctionT list_function, void *attr) {
    int err;
    DSLockContext lock;
    ds_assert_param(ctx);
    ds_assert_param(list_function);
    lock=ctx->lock;
    ds_assert_param(lock);
    
    dsLockRead(lock);        
    err=dsHashTableWalk(ctx->locales, (DSHashTableWalkFunction)list_function, attr);
    dsLockReturn(lock);
    return err;
}
