#define _OPC_PROPERTIES_C

#include <opclib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <opcproperty.h>
#include <opcstandardproperties.h>
#include <opcproperties.h>

int opcPropertiesInit() {
    return opcPropertiesCreateTypes();
}

void opcPropertiesDestroy() {
    opcPropertiesDestroyTypes();
}

void opcPropertiesZeroContext(OPCPropertiesContext ctx) {
	ctx->id=NULL;
	ctx->lock=NULL;
	ctx->properties=NULL;
}

OPCPropertiesContext opcPropertiesCreateContext(DSMallocContext mctx) {
    OPCPropertiesContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "opcPropertiesCreateContext");    

    ctx = (OPCPropertiesContext)dsMalloc(mctx,sizeof(struct OPCPropertiesContextT));
    if (ctx) {
	ctx->malloc=mctx;
	opcPropertiesZeroContext(ctx);
    }
    return ctx;
}

int opcPropertiesInitContext(OPCPropertiesContext ctx, unsigned long properties_hash_size, OPCPropertyID properties_maxid, OPCPropertyID properties_userid) {
    unsigned short i;
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "opcPropertiesInitContext");    

    ds_assert_param(ctx);

    opcPropertiesZeroContext(ctx);

    ctx->id=dsHashTableInit(NULL,properties_hash_size,DS_HASH_TYPE_POINTER,NULL);
    if (!ctx->id) {
	opcPropertiesFreeContext(ctx);
	return DS_ERR_HASHTABLE_CREATE;
    }

    ctx->lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcPropertiesFreeContext(ctx);
	return DS_ERR_LOCK_CREATE;
    }
    err = dsLockInitContext(ctx->lock);
    if (err) {
	opcPropertiesFreeContext(ctx);
	return err;
    }

    ctx->properties = (OPCPropertyContext*)dsMalloc(ctx->malloc, properties_maxid*sizeof(OPCPropertyContext));
    if (!ctx->properties) {
	opcPropertiesFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    for (i=0;i<properties_maxid;i++)
	ctx->properties[i]=NULL;
        
    ctx->maxid = properties_maxid;

/*    if (properties_userid > 0) {
	ctx->userid = properties_userid;
	ctx->curid = properties_userid;
    } else { */
	ctx->userid = 0;
	ctx->curid = 0;
/*    } */
        
    return opcPropertiesRegisterStandard(ctx);
}

void opcPropertiesFreeContext(OPCPropertiesContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcPropertiesFreeContext");    

    ds_assert_return(ctx);

    if (ctx->properties) dsFree(ctx->malloc, ctx->properties);
    if (ctx->id) dsHashTableDestroy(ctx->id);
    if (ctx->lock) dsLockDestroyContext(ctx->lock);

    opcPropertiesZeroContext(ctx);
}

void opcPropertiesDestroyContext(OPCPropertiesContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcPropertiesDestroyContext");    

    opcPropertiesFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}



int opcPropertiesRegisterProperty(OPCPropertiesContext ctx, OPCPropertyContext property) {
    int err=0;
    char *name;
    unsigned short id;
    
    ds_assert_param(ctx);
    ds_assert_param(property);
    
    name = property->name;
    if (!name) return OPC_ERR_PROPERTIES_INVALID_PROPERTY;
    if (opcPropertiesFindProperty(ctx,name)) return OPC_ERR_PROPERTIES_ALREADY_REGISTERED;
/*  if (property->id) {
	id = property->id;
	if (ctx->properties[id]) return OPC_ERR_PROPERTIES_ID_USED;
    } else*/ 
    if (ctx->curid<ctx->maxid) id = 0;
    else return OPC_ERR_PROPERTIES_LIMIT_EXCEEDED;


    dsLockWrite(ctx->lock);
    err = dsHashTableInsert(ctx->id, property);
    if (!err) {
        if (!id) {
	    id=ctx->curid++;
	    opcPropertySetId(property, id);
	}
        ctx->properties[id]=property;
    }
    dsLockReturn(ctx->lock);
    
    return err;
}


int opcPropertiesRegisterPropertyByType(OPCPropertiesContext ctx, const char *name, OPCPropertyType ptype, DSBinaryType type) {
    int err;
    OPCPropertyContext property;
    
    ds_assert_param(ctx);
    ds_assert_param(name);
    ds_assert_param(type);
    
    if (opcPropertiesFindProperty(ctx,name)) return OPC_ERR_PROPERTIES_ALREADY_REGISTERED;
    
    property = opcPropertyCreateContext(ctx->malloc);
    if (!property) return OPC_ERR_PROPERTY_CREATE;
    err = opcPropertyInitContext(property, name, ptype, type);
    if (err) {
	opcPropertyDestroyContext(property);
	return err;
    }

    err = opcPropertiesRegisterProperty(ctx, property);
    if (err) opcPropertyDestroyContext(property);
    return err;
}

int opcPropertiesRegisterPropertyByStandardType(OPCPropertiesContext ctx, const char *name, OPCPropertyType ptype, DSBinaryTypeStandard type, const void *attr) {
    int err;
    OPCPropertyContext property;
    
    ds_assert_param(ctx);
    ds_assert_param(name);
    ds_assert_param(type);
    
    if (opcPropertiesFindProperty(ctx,name)) return OPC_ERR_PROPERTIES_ALREADY_REGISTERED;
    
    property = opcPropertyCreateContext(ctx->malloc);
    if (!property) return OPC_ERR_PROPERTY_CREATE;
    err = opcPropertyInitStandardContext(property, name, ptype, type, attr);
    if (err) {
	opcPropertyDestroyContext(property);
	return err;
    }

    err = opcPropertiesRegisterProperty(ctx, property);
    if (err) opcPropertyDestroyContext(property);
    return err;
}

OPCPropertyContext opcPropertiesFindProperty(OPCPropertiesContext ctx, const char *property) {
    OPCPropertyContext pctx;
    
    ds_assert_null(ctx);
    
    dsLockRead(ctx->lock);
    pctx=(OPCPropertyContext)dsHashTableFind(ctx->id, property);
    dsLockReturn(ctx->lock);
    
    return pctx;
}

OPCPropertyID opcPropertiesFindPropertyId(OPCPropertiesContext ctx, const char *property) {
    OPCPropertyContext pctx;
    ds_assert_err(ctx, (unsigned short)-1);

    dsLockRead(ctx->lock);
    pctx=(OPCPropertyContext)dsHashTableFind(ctx->id, property);
    dsLockReturn(ctx->lock);
    
    if (pctx) return pctx->id;
    return OPC_PROPERTY_ID_INVALID;
}

OPCPropertyContext opcPropertiesGetProperty(OPCPropertiesContext ctx, OPCPropertyID id) {
    ds_assert_null(ctx);
    ds_assert_null(ctx->properties);
    
    return ctx->properties[id];
}

OPCPropertyID opcPropertiesGetLastID(OPCPropertiesContext ctx) {
    ds_assert_default(ctx, OPC_PROPERTY_ID_INVALID);
    return ctx->curid;
}
