#include <stdio.h>
#include <errno.h>

#include <opclib.h>
#include <dsassert.h>
#include <dsdebug.h>

#include <dsmalloc.h>
#include <dslocale.h>
#include <dshashtable.h>
#include <dslock.h>

#include <opcitem.h>

#include <opctree.h>

void opcTreeZeroContext(OPCTreeContext ctx) {
    ds_assert_return(ctx);
    ctx->status = 0;    
    ctx->item = NULL;
    ctx->lock = NULL;
}

OPCTreeContext opcTreeCreateContext(DSMallocContext mctx) {
    OPCTreeContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "opcTreeCreateContext");    

    ctx = (OPCTreeContext)dsMalloc(mctx, sizeof(struct OPCTreeContextT));
    if (ctx) {
	ctx->malloc = mctx;
	opcTreeZeroContext(ctx);
    }
    return ctx;
}

int opcTreeInitContext(OPCTreeContext ctx, OPCContext sctx) {
    int err;
    OPCItemContext ictx;
    
    dsDebugServer(DS_DEBUG_NOTICE, "opcTreeInitContext");    

    ds_assert_param(ctx);

    opcTreeZeroContext(ctx);

//    ctx->item=dsHashTableInit(ctx->malloc, opc_max_items, DS_HASH_TYPE_POINTER, NULL);
    ctx->item=dsHashTableInit(ctx->malloc, opc_max_items, DS_HASH_TYPE_FUNCTION, (DSHashTableStringFunction)opcItemGetFullNamePtr);
    dsHashTableSetDestructor(ctx->item,(void*)&opcItemDestroyContext);
    if (!ctx->item) return DS_ERR_MEM;
    
    ctx->lock=dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcTreeFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    err = dsLockInitContext(ctx->lock);
    if (err) {
	opcTreeFreeContext(ctx);
	return err;
    }
    
    ictx = opcItemCreateContext(ctx->malloc);
    if (!ictx) {
	opcTreeFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    err = opcItemInitContext(ictx,"","");
    
    if (!err) err = opcTreeRegisterItem(ctx,ictx,NULL);
    
    if (err) {
	opcTreeFreeContext(ctx);
	return err;
    }
    
    opcItemSetManualMode(ictx);

    ctx->sctx = sctx;

    dsDebugServer(DS_DEBUG_INFO, "opcTreeInitContext Ok");    

    return 0;
}

void opcTreeFreeContext(OPCTreeContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcTreeFreeContext");    

    ds_assert_return(ctx);
    
    ctx->status|=OPC_TREE_DESTROYING_FLAG;
    
    if (ctx->item) {
	/* I don't need to lock here because all threads before this point should be terminated */
	dsHashTableDestroy(ctx->item);
    }
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    
    opcTreeZeroContext(ctx);

    dsDebugServer(DS_DEBUG_INFO, "opcTreeFreeContext Ok");    
}

void opcTreeDestroyContext(OPCTreeContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcTreeDestroyContext");    

    opcTreeFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

int opcTreeGetStatus(OPCTreeContext ctx) {
    ds_assert_param(ctx);
    return ctx->status;
}

OPCItemContext opcTreeFindItem(OPCTreeContext ctx, const char *fullname) {
    OPCItemContext ictx;

    ds_assert_null(ctx);
    ds_assert_null(fullname);
    ds_assert_null(ctx->lock);
    ds_assert_null(ctx->item);
    
    dsLockRead(ctx->lock);
    ictx = dsHashTableFind(ctx->item,fullname);
    dsLockReturn(ctx->lock);

    return ictx;
}

OPCItemContext opcTreeGetItem(OPCTreeContext ctx, const char *fullname) {
    int err;
    OPCItemContext ictx;

    ds_assert_null(ctx);
    ds_assert_null(fullname);
    ds_assert_null(ctx->lock);
    ds_assert_null(ctx->item);
    
    dsLockRead(ctx->lock);
    ictx = dsHashTableFind(ctx->item,fullname);
    if (ictx) err = opcItemGet(ictx);
    else err = 0;
    dsLockReturn(ctx->lock);

    if (err) return NULL;
    return ictx;
}

int opcTreeWalk(OPCTreeContext ctx, OPCTreeWalkT func, void *attr) {
    int err;
    ds_assert_param(ctx);
    ds_assert_param(ctx->item);
    ds_assert_param(ctx->lock);
    
    dsLockRead(ctx->lock);
    err=dsHashTableWalk(ctx->item,(DSHashTableWalkFunction)func,attr);
    dsLockReturn(ctx->lock);
    return err;
}

int opcTreeLock(OPCTreeContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    
    return dsLockWrite(ctx->lock);
}

int opcTreeUnlock(OPCTreeContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    
    return dsLockReturn(ctx->lock);
}

int opcTreeLinkItems(OPCTreeContext ctx, OPCItemContext ictx, OPCItemContext child) {
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "opcTreeLinkItems %p,%p.", ictx, child);    
    
    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    ds_assert_param(ictx);
    ds_assert_param(child);
    
    dsLockWrite(ctx->lock);
    err = opcItemLink(ictx, child);
    dsLockReturn(ctx->lock);

    dsDebugServer(DS_DEBUG_INFO, "opcTreeLinkItems %p,%p. Ok", ictx, child);    

    return err;
}

int opcTreeUnlinkItems(OPCTreeContext ctx, OPCItemContext ictx, OPCItemContext child) {
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "opcTreeUnlinkItems %p,%p.", ictx, child);    

    ds_assert_param(ctx);
    ds_assert_param(ctx->lock);
    ds_assert_param(ictx);
    ds_assert_param(child);

    dsLockWrite(ctx->lock);
    err = opcItemUnlink(ictx, child);        
    dsLockWrite(ctx->lock);

    dsDebugServer(DS_DEBUG_INFO, "opcTreeUnlinkItems %p,%p. Ok", ictx, child);    
    return err;
}

OPCItemContext opcTreeCreatePath(OPCTreeContext ctx, const char *path) {
    OPCItemContext ictx;
    char *ch, *ppath;
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "opcTreeCreatePath \"%s\"", path);    
    
    ds_assert_null(ctx);
    ds_assert_null(path);
    
    ictx = opcTreeFindItem(ctx,path);
    if (ictx) return ictx;
    
    ictx=opcItemCreateContext(ctx->malloc);
    if (!ictx) return NULL;
	
    ch=strrchr(path,'/');

    if (ch) {
        err = opcItemInitContext(ictx,path,ch+1); 
	if (!err) {
	    err = (unsigned long)(ch - path);
	    ppath=(char*)malloc((err+1)*sizeof(char));
	    if (ppath) {
		if (err) strncpy(ppath,path,err); 
		ppath[err]=0;
		err=opcTreeRegisterItem(ctx,ictx,ppath);
		free(ppath);
	    } else err=DS_ERR_MEM;
	}
    } else {
	err = opcItemInitContext(ictx,"","");
	if (!err) err=opcTreeRegisterItem(ctx,ictx,"");
    }
    
    if (err) {
	opcItemDestroyContext(ictx);
	if (err!=OPC_ERR_SERVERTREE_ALREADY_REGISTERED) return NULL;
	ictx = opcTreeFindItem(ctx,path);
    }
    
    dsDebugServer(DS_DEBUG_INFO, "opcTreeCreatePath \"%s\" Ok", path);    

    return ictx;
}

int opcTreeRegisterItem(OPCTreeContext ctx, OPCItemContext ictx, const char *path) {
    OPCItemContext parrent;
    int err;

    if (path)
	dsDebugServer(DS_DEBUG_NOTICE, "opcTreeRegisterItem %p under \"%s\"", ictx, path);    
    else 
	dsDebugServer(DS_DEBUG_NOTICE, "opcTreeRegisterItem %p", ictx);    

    ds_assert_param(ctx);
    ds_assert_param(ictx);
    ds_assert_param(ctx->lock);
    ds_assert_param(ictx->fullname);

    opcItemLock(ictx);

    dsLockWrite(ctx->lock);
    if (dsHashTableFind(ctx->item,ictx->fullname)) return OPC_ERR_SERVERTREE_ALREADY_REGISTERED;
    if (dsHashTableInsert(ctx->item, (void*)ictx)) return DS_ERR_MEM;
    opcItemSetTree(ictx,ctx);
    dsLockReturn(ctx->lock);
    
    if (path) {
	parrent = opcTreeCreatePath(ctx,path);
	if (parrent) {
	    err = opcTreeLinkItems(ctx, parrent, ictx);
	    if (parrent->manual==2) parrent->manual = 0;
	} else err = OPC_ERR_SERVERTREE_CREATE_PATH;
    } else err = 0;

    if (err) {
	dsLockWrite(ctx->lock);
	if (opcTreeMarkUnused(ctx,parrent)!=DS_ERR_NOTFOUND) opcTreeCleanMarked(ctx);
	dsLockReturn(ctx->lock);
    } else ictx->status |= OPC_ITEM_LINKED_FLAG;
    
    opcItemUnlock(ictx);

    if (path)
	dsDebugServer(DS_DEBUG_INFO, "opcTreeRegisterItem %p under \"%s\". Linked to %p. Ok", ictx, path, parrent);    
    else
	dsDebugServer(DS_DEBUG_INFO, "opcTreeRegisterItem %p Ok", ictx);    

    return err;
}

int opcTreeUnregisterItem(OPCTreeContext ctx, OPCItemContext ictx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcTreeUnregisterItem %p", ictx);    
    
    ds_assert_param(ctx);
    ds_assert_param(ictx);
    ds_assert_param(ctx->item);
    
    dsLockWrite(ctx->lock);
    
    if (opcItemHaveManualSubChilds(ictx)) return OPC_ERR_SERVERTREE_ITEM_HAVE_MANUAL_CHILDS;
    
    opcItemMarkSubTree(ictx);
    dsHashTableWalk(ctx->item, (DSHashTableWalkFunction)opcItemMarkParrentsCallback, (void*)ictx);
    dsHashTableWalk(ctx->item, (DSHashTableWalkFunction)opcItemMarkChildsCallback, (void*)ictx);
    ictx->mark = 0;
    opcTreeCleanMarked(ctx);
    dsHashTableUnlink(ctx->item,ictx);
    ictx->status &= ~OPC_ITEM_LINKED_FLAG;    

    opcItemUnlock(ictx);
    dsLockReturn(ctx->lock);

    dsDebugServer(DS_DEBUG_INFO, "opcTreeUnregisterItem %p Ok", ictx);
    return 0;
}

int opcTreeBrowse(OPCTreeContext ctx, const char *path, OPCItemBrowseFilter browse_filter, const char *name_filter, const char *vendor_filter, DSStringList *result) {
    int err;
    OPCItemContext ictx;
    
    ds_assert_param(ctx);
    ds_assert_param(path);
    ds_assert_param(result);
    ds_assert_param(ctx->lock);
    
    dsLockRead(ctx->lock);
    ictx = dsHashTableFind(ctx->item,path);
    if (ictx) err = opcItemBrowse(ictx,browse_filter,name_filter,vendor_filter,result);
    else {
	*result = NULL;
	err = OPC_ERR_SERVERTREE_ITEM_NOTFOUND;
    }
    dsLockReturn(ctx->lock);
    
    return err;
}

/*******************************************************************************
**************************** Internal Functions ********************************
*******************************************************************************/

int opcTreeMarkUnused(OPCTreeContext tree, OPCItemContext ctx) {
    ds_assert_param(ctx);
    ds_assert_param(tree);
    ds_assert_param(tree->item);
    
    if ((opcItemHaveChilds(ctx))||(ctx->manual)||(ctx->status&OPC_ITEM_CONNECTED_FLAG)) return DS_ERR_NOTFOUND;
    opcItemMark(ctx);
    return dsHashTableWalk(tree->item, (DSHashTableWalkFunction)opcItemMarkParrentsCallback, (void*)ctx);
}

int opcTreeCleanMarked(OPCTreeContext tree) {
    ds_assert_param(tree);
    ds_assert_param(tree->item);
    return dsHashTableClean(tree->item, (DSHashTableCleanFunction)opcItemIsMarked);
}

