#include "httpd.h"
#include "http_main.h"
#include "http_request.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"

#ifdef AP_SERVER_MAJORVERSION
# undef  DS_APCHE
# define DS_APACHE2
module AP_MODULE_DECLARE_DATA opc_module;
#else
# define DS_APACHE
# undef  DS_APACHE2
# include "multithread.h"
#endif

#include <dslib.h>
#include <dsdebug.h>
#include <dsconnection.h>

#include <opccommands.h>
#include <opcdefines.h>
#include "mod_opc.h"

// There are several servers operating on a machine, therefore reconnecting
// on each client request (if defined)
#define opc_multi_server 1
//#undef opc_multi_server
//#define DEBUG_INFO APLOG_INFO
#define DEBUG_INFO APLOG_NOTICE
#define MIME_BOUNDARY "qazwsxedcrfvtgbyhnujm"
#define ATACHMENT_MIME_TYPE "application/binary"
#define MAIN_CONTENT_ID "SOAP_message"



struct DSApacheThreadDataT {
    int initialized;
    DSConnectionContextS conn;
    char buffer[OPC_BUFFER_SIZE + 4];
#ifndef opc_multi_server
    char last_server[MAX_SERVER_NAME + 1];
#endif
};
typedef struct DSApacheThreadDataT DSApacheThreadDataS;
typedef struct DSApacheThreadDataT *DSApacheThreadData;


#if defined(WIN32)
static DWORD tls_index;
#else
# ifdef DS_APACHE2
#  ifdef HAVE_PTHREAD_H
#   include <pthread.h>
    static pthread_key_t tls_keys;
#   else
#    error "No supported TLS(Thread Local Storage) found!"
#  endif /* HAVE_PTHREAD_H */
# else /* DS_APACHE2 */
    static APACHE_TLS DSApacheThreadDataS sptd = {0};
# endif /* DS_APACHE2 */
#endif /* WIN32 */

static DSApacheThreadData dsApacheThreadDataGet(void)
{
#if defined(WIN32)
    return (DSApacheThreadData)TlsGetValue(tls_index);
#else
# ifdef DS_APACHE2
    void *tls_key;

    tls_key = pthread_getspecific(tls_keys);
    if (!tls_key) {
        tls_key = malloc(sizeof(DSApacheThreadDataS));
	if ((tls_key)&&(pthread_setspecific(tls_keys, tls_key))) {
	    free(tls_key);
	    tls_key = NULL;
	}
    }
    return (DSApacheThreadData)tls_key;
# else
    return &sptd;
# endif /* DS_APACHE2 */
#endif
}

#ifdef DS_APACHE2
# define ap_hard_timeout(r, timeout) ;
# define ap_reset_timeout(r) ;
# define ap_kill_timeout(r) ;
# define ap_send_http_header(r) ;
# define ap_palloc(a,b) apr_palloc(a,b)
# define ds_ap_log_error(a,b,...) ap_log_error(a,b,0,__VA_ARGS__)
# define ds_ap_get_remote_host(con, dir) ap_get_remote_host(con, dir, REMOTE_NOLOOKUP, NULL)
#else
# define ds_ap_log_error(a,b,...) ap_log_error(a,b,__VA_ARGS__)
# define ds_ap_get_remote_host(con, dir) ap_get_remote_host(con, dir, REMOTE_NAME)
#endif /* DS_APACHE2 */

static void opc_log(request_rec *r, char *str) {
    ds_ap_log_error(APLOG_MARK,APLOG_INFO,r->server,str);
}

static void opc_log_invalid_srvresponse(request_rec *r, uint32_t *resp, uint32_t resp_size) {
    if (resp_size<sizeof(uint32_t))
	ds_ap_log_error(APLOG_MARK,APLOG_NOTICE,r->server,"Incorrect server response, length: %lu",resp_size);
    else
	ds_ap_log_error(APLOG_MARK,APLOG_NOTICE,r->server,"Incorrect server response: %lX, length: %lu",ntohl(*resp),resp_size);
}


static DSApacheThreadData dsApacheThreadDataInit(server_rec *sr, const char *srv) {
    int err;
    DSApacheThreadData data;
    
    data = dsApacheThreadDataGet();
    if (!data) return data;
    
    
    if (!data->initialized) {
	ds_ap_log_error(APLOG_MARK, APLOG_NOTICE, sr, "Initializing Per Thread Context");
	err = dsConnectionCreateSContext(&data->conn, NULL);
	if (err) {
	    free(data->buffer);
	    ds_ap_log_error(APLOG_MARK, APLOG_WARNING, sr, "ConnectionContext Initialization Failed: %lx", err);
	    return NULL;
	}

	data->initialized = 1;
    }

    if (!srv) return NULL;

    if (strlen(srv)>MAX_SERVER_NAME) {
	    ds_ap_log_error(APLOG_MARK, APLOG_WARNING, sr, "Server Name to Long: %s", srv);
	    return NULL;
    }
    

#ifndef opc_multi_server
    if (dsConnectionCheck(&data->conn)==DS_CONNECTION_STATUS_CONNECTED) {
	if (!strcmp(data->last_server, srv)) {
	    ds_ap_log_error(APLOG_MARK, APLOG_DEBUG, sr, "Using existing connection to the server: %s", srv);
	    return data;
	}
    } 
    strcpy(data->last_server,srv);    
#endif

    if (dsConnectionCheck(&data->conn)==DS_CONNECTION_STATUS_CONNECTED) {
        dsConnectionFreeContext(&data->conn);
	ds_ap_log_error(APLOG_MARK, APLOG_INFO, sr, "ReConnecting to the server: %s", srv);
    } else {
        dsConnectionFreeContext(&data->conn);
	ds_ap_log_error(APLOG_MARK, APLOG_INFO, sr, "Connecting to the server: %s", srv);
    }
    err = dsConnectionConnectSocket(&data->conn, NULL, OPC_TEMP_DIR, srv, OPC_FRONTEND_EXTENSION);
    if (err) {
	ds_ap_log_error(APLOG_MARK, APLOG_WARNING, sr, "Connection Initialization Failed: %lx", err);
	return NULL;
    }
    
    return data;
}

static DSApacheThreadData dsApacheThreadDataCreate(server_rec *sr, const char *srv) {
    return dsApacheThreadDataInit(sr, srv);
}

static void dsApacheThreadDataDestroy() {
    DSApacheThreadData data;

    data = dsApacheThreadDataGet();
    if ((data)&&(data->initialized)) {
	dsConnectionFreeContext(&data->conn);
	data->initialized = 0;
    }
}

static void dsApacheThreadDataFree(DSApacheThreadData data, int force) {
#ifndef opc_multi_server
    if (force) {
#endif /* ! opc_multi_server */
	dsConnectionFreeContext(&data->conn);
#ifndef opc_multi_server
    }
#endif /* ! opc_multi_server */
}

static int opc_return_error(request_rec *r, DSApacheThreadData data, int err, int aerr, int fatal) {
//    ds_ap_log_error(APLOG_MARK,APLOG_NOTICE,r->server,"opc_return_error: %lX %lX %lX", fatal, err, aerr);

	// Closing socket if required
    if (data) {
	dsApacheThreadDataFree(data, fatal);
    }

	// Logging OPC Error code and Errno
    if (err) ds_ap_log_error(APLOG_MARK,APLOG_NOTICE,r->server,"OPC Error: %lX", err);

	// Returning apropriate error to client
    if (aerr) return aerr;

    switch (err) {
	case 0:
	    return OK;
	case DS_ERR_SOCKET_CONNECT:
	case DS_ERR_CONNECTION_REFUSED:
	case DS_ERR_CONNECTION_TIMEOUT:
	case DS_ERR_CONNECTION_NETWORK:
//	    return HTTP_NOT_FOUND;
	    return HTTP_SERVICE_UNAVAILABLE;
	case OPC_ERR_CLI_XMLREQUEST:
	    return HTTP_BAD_REQUEST;
	case DS_ERR_BUSY:
	    return HTTP_SERVICE_UNAVAILABLE;
	default:
	    return HTTP_INTERNAL_SERVER_ERROR;
    }
}



static int opc_new_request(request_rec *r, DSApacheThreadData data, char *req_string, uint32_t req_size, uint32_t req_response) {
    int err;
    unsigned long readed, writed;
    void *opc_buffer = data->buffer;
    
    err = dsConnectionWrite(&data->conn, req_string, req_size, 0, &writed);
    if (err||(req_size!=writed)) return err;
    err = dsConnectionRead(&data->conn, data->buffer, OPC_BUFFER_SIZE, 0, &readed);
    if (err||(readed==0)) return err;
    if (readed!=sizeof(uint32_t)) {
	opc_log_invalid_srvresponse(r,(uint32_t*)opc_buffer,readed);
	return OPC_ERR_CLI_SRVRESPONSE;
    }
    
    if (ntohl(*((uint32_t*)opc_buffer))!=req_response) {
	switch (ntohl(*(uint32_t*)opc_buffer)) {
	    case OPC_FECMD_REPEAT:
		ds_ap_log_error(APLOG_MARK, APLOG_NOTICE,r->server, "It's seems previous request isn't correctly closed! Repeating initialization request!");
		return opc_new_request(r, data, req_string, req_size, req_response);
	    case OPC_FECMD_INVALIDREQ:
		ds_ap_log_error(APLOG_MARK, APLOG_NOTICE,r->server,"It's seems server doesn't understand us!");
	        return OPC_ERR_CLI_UNDERSTAND;
	    case OPC_FECMD_BUSY:
		ds_ap_log_error(APLOG_MARK, APLOG_NOTICE,r->server,"It's seems server is busy!");
		return DS_ERR_BUSY;
	    break;
	    case OPC_FECMD_OPC_INVALIDREQ:
		ds_ap_log_error(APLOG_MARK, APLOG_NOTICE,r->server,"It's seems server doesn't understand us! Responded what invalid OPC request submited!");
	        return OPC_ERR_CLI_UNDERSTAND;
	    default:
		opc_log_invalid_srvresponse(r,(uint32_t*)opc_buffer,readed);
	        return OPC_ERR_CLI_SRVRESPONSE;
	}
    }
    return 0;
}

static inline int opc_handler_dime(request_rec *r, DSApacheThreadData data, unsigned long readed, uint32_t pos) { 
    int err;
    DSConnectionContext conn = &data->conn;
    char *opc_buffer = data->buffer;
    char *buffer;
    unsigned long bytes;
    uint32_t size;
    unsigned char last = 0;

    r->content_type = "multipart/dime";
    ap_send_http_header(r); // send headers 

    while(!last) {
	if (readed<12) {
	    if (readed) {
		ds_ap_log_error(APLOG_MARK, APLOG_NOTICE,r->server,"Moving %lu %lu %u", readed, pos, *(uint8_t*)(opc_buffer+pos));
		memmove(opc_buffer,opc_buffer+pos,readed);
/*
		*((uint32_t*)opc_buffer)=*((uint32_t*)(opc_buffer+pos));
		if (readed>4) {
	    	    *((uint32_t*)(opc_buffer+4))=*((uint32_t*)(opc_buffer+pos+4));
		    if (readed>8) {
			*((uint32_t*)(opc_buffer+8))=*((uint32_t*)(opc_buffer+pos+8));
		    }
		}
*/
	    }
	    pos = 0;	

	    err = dsConnectionRead(conn, opc_buffer+readed, OPC_BUFFER_SIZE-readed, 0, &bytes);
	    if (err||(bytes==0)) return opc_return_error(r, data, err?err:OPC_ERR_CLI_SRVINTERNAL, 0, 1);
	    readed += bytes;
	    if (readed<12) return opc_return_error(r, data, OPC_ERR_CLI_SRVINTERNAL, 0, 1);
	}
	
	buffer = opc_buffer + pos;
	size = 12 + ntohs(*(uint16_t*)(buffer+2));
	if (size&3) size = (size&~3)+4;
	size += ntohs(*(uint16_t*)(buffer+4));
	if (size&3) size = (size&~3)+4;
	size += ntohs(*(uint16_t*)(buffer+6));
	if (size&3) size = (size&~3)+4;
	size += ntohl(*(uint32_t*)(buffer+8));
	if (size&3) size = (size&~3)+4;

	last = (*(uint8_t*)(buffer))&64;
	
	ds_ap_log_error(APLOG_MARK, APLOG_NOTICE,r->server,"New DIME part %lu %u\n",size,(*(uint8_t*)(buffer)));

	while (size>readed) {
	    ap_rwrite(opc_buffer+pos,readed,r);
	    size-=readed;
	    err = dsConnectionRead(conn, opc_buffer, OPC_BUFFER_SIZE, 0, &readed);
	    if (err||(readed==0)) return opc_return_error(r,data,err?err:OPC_ERR_CLI_SRVINTERNAL,0, 1);
	    pos = 0;
	}
    
	if (size) {
	    ap_rwrite(opc_buffer+pos,size,r);
	    readed-=size;
	    pos+=size;
	}
    }

    return opc_return_error(r,data,0,0,0);
}

static inline int opc_handler_attachment(request_rec *r, DSApacheThreadData data, unsigned long *readedm, unsigned long *posm) {
    int err;
    DSConnectionContext conn = &data->conn;
    char *opc_buffer = data->buffer;
    uint32_t size;
    uint16_t tsize;
    unsigned long bytes;
    unsigned long readed, pos;
    
    readed = *readedm;
    pos = *posm;
    
    if (readed<sizeof(uint32_t)+sizeof(uint16_t)) {
	memmove(opc_buffer, opc_buffer+pos, readed);
/*	if (readed) {
	    *((uint32_t*)opc_buffer)=*((uint32_t*)(opc_buffer+pos));
	    if (readed>sizeof(uint32_t)) 
		*((uint32_t*)(opc_buffer+sizeof(uint32_t)))=*((uint32_t*)(opc_buffer+pos+sizeof(uint32_t)));
	}*/
	err = dsConnectionRead(conn, opc_buffer+readed, OPC_BUFFER_SIZE-readed, 0, &bytes);
	if (err||(bytes<sizeof(uint32_t)+sizeof(uint16_t))) return err?err:OPC_ERR_CLI_SRVINTERNAL;
	readed+=bytes;
	pos = 0;
    }
    
    size = ntohl(*(uint32_t*)(opc_buffer+pos));
    tsize = ntohs(*(uint16_t*)(opc_buffer+pos+sizeof(uint32_t)));
    readed -= sizeof(uint32_t) + sizeof(uint16_t);
    pos += sizeof(uint32_t) + sizeof(uint16_t);
    
    ap_rprintf(r, "--%s\r\n", MIME_BOUNDARY);
    ap_rprintf(r, "Content-Type: %s\r\n", ATACHMENT_MIME_TYPE);
    ap_rprintf(r, "Content-Lengh: %lu\r\n", size);
    ap_rprintf(r, "Content-Transfer-Encoding: 8bit\r\n");
    ap_rprintf(r, "Content-ID: ");

    
    while (tsize>readed) {
	ap_rwrite(opc_buffer+pos,readed,r);
	tsize-=(uint16_t)readed;
	
	err = dsConnectionRead(conn, opc_buffer, OPC_BUFFER_SIZE, 0, &readed);
	if (err||(readed==0)) return err?err:OPC_ERR_CLI_SRVINTERNAL;
	pos = 0;
    }
    
    if (tsize) {
	ap_rwrite(opc_buffer+pos,tsize,r);
	readed-=tsize;
	pos+=tsize;
    }
    
    ap_rprintf(r, "\r\n\r\n");
	
    while (size>readed) {
	ap_rwrite(opc_buffer+pos,readed,r);
	size-=readed;

	err = dsConnectionRead(conn, opc_buffer, OPC_BUFFER_SIZE, 0, &readed);
	if (err||(readed==0)) return err?err:OPC_ERR_CLI_SRVINTERNAL;
	pos = 0;
    }
    
    if (size) {
	ap_rwrite(opc_buffer+pos,size,r);
	readed-=size;
	pos+=size;
    }
    
    ap_rprintf(r, "\r\n");
    
    *readedm = readed;
    *posm = pos;

    return 0;
}

/* here's the content handler */
static int opc_handler(request_rec *r) {
    DSApacheThreadData data;
    DSConnectionContext conn;
    char *opc_buffer;
    const char *opc_file;			// OPC Server Name
    char *opc_dir;				// Directory on server
    int err,rc;					// Error code
    const char *hostname;			// Client host name
    const char username[16]="";			// Client user name
    char opc_request[OPC_REQRESP_BUFFER_SIZE];	// OPC Request
    uint32_t i,j;				// Just counters
    unsigned long readed,writed;		// Bytes readed from socket
    uint32_t len;				// Response size in bytes
    unsigned long pos;				// Current position in buffer
    char attachments = 0;			// Response with attachments
    char dime = 0;				// DIME transfer

#ifdef DS_APACHE2
    if (strcmp(r->handler, "opc-handler")) return DECLINED;
#endif /* DS_APACHE2 */

	// Requesting message body, accepting chunks (not tested - DS)
    if ((rc=ap_setup_client_block(r,REQUEST_CHUNKED_DECHUNK/*PASS*/))!=OK) 
	return opc_return_error(r,NULL,OPC_ERR_CLI_APACHE,rc,0);
    
	// Getting server name and path
    for (i=0;r->uri[i]!=0;i++)
	if (r->uri[i]=='/') opc_file=r->uri+i+1;

    if (strlen(opc_file)>MAX_SERVER_NAME) return opc_return_error(r,NULL,DS_ERR_TOOLONG,0,0);
    opc_dir = (char*)ap_palloc(r->pool,i+2+MAX_SERVER_NAME);
    if (!opc_dir) return opc_return_error(r,NULL, DS_ERR_MEM, 0, 0);
    strncpy(opc_dir,r->uri,(opc_file-r->uri));
    opc_dir[opc_file-r->uri]=0;

	// Getting information about clients
    hostname = ds_ap_get_remote_host(r->connection,r->per_dir_config);

	// hostname,username assumed to have less then 255 chars
    *((uint32_t*)opc_request)=htonl(OPC_FECMD_NEWREQ);
    opc_request[sizeof(uint32_t)]=strlen(hostname);
    for (i=0,j=sizeof(uint32_t)+1;hostname[i]!=0;i++,j++) 
	opc_request[j]=hostname[i];
    opc_request[j]=strlen(username);
    for (i=0,j++;username[i]!=0;i++,j++)
	opc_request[j]=username[i];


	// Initialising connection to the server
    ap_hard_timeout("OPC XML-DA Server",r);
    data = dsApacheThreadDataInit(r->server,opc_file);
    if (!data) return opc_return_error(r, NULL, OPC_ERR_CLI_INITIALIZE, 0, 0);
    conn = &data->conn;
    opc_buffer = data->buffer;
    ap_reset_timeout(r);

    err = opc_new_request(r,data,opc_request,j,OPC_FECMD_NEWREQ);
	// Trying to reinitialise socket
    if (dsErrorIs(err,DS_ERR_SOCKET_WRITE)) {
#ifndef opc_multi_server
	data->last_server[0]=0;
#endif
	data = dsApacheThreadDataInit(r->server,opc_file);
	if (!data) return opc_return_error(r, NULL, OPC_ERR_CLI_INITIALIZE, 0, 0);

	ap_reset_timeout(r);
	err=opc_new_request(r,data,opc_request,j,OPC_FECMD_NEWREQ);
    }
    ap_kill_timeout(r);
    if (err) return opc_return_error(r,data,err,0,1);

    
	// Processing Request
    if (!ap_should_client_block(r)) {
	    // Sending to server request termination command
	*((uint32_t*)opc_buffer)=htonl(OPC_FECMD_CANCEL);

	err = dsConnectionWrite(conn, opc_buffer, sizeof(uint32_t), 0, &writed);
	if (err||(writed!=sizeof(uint32_t))) return opc_return_error(r,data,err?err:DS_ERR_SOCKET_WRITE,0,1);

	strcat(opc_dir,"error.html");
	ap_internal_redirect(opc_dir,r);
	return opc_return_error(r,data,0,0,0);
    } else {
	// Initialising (if required) client socket connection
	ap_hard_timeout("OPC XML-DA Request",r);
	while ((len=ap_get_client_block(r,opc_buffer+sizeof(uint32_t),OPC_BUFFER_SIZE-sizeof(uint32_t)))>0) {
	    ap_reset_timeout(r);
	    *((uint32_t*)opc_buffer)=htonl(len);

	    err = dsConnectionWrite(conn, opc_buffer, len + sizeof(uint32_t), 0, &writed);
	    if (err||(writed!=len+sizeof(uint32_t))) return opc_return_error(r,data,err?err:DS_ERR_SOCKET_WRITE,0,1);
	    ap_reset_timeout(r);
	}
/******** Feel full buffer
	opc_buffer[0]=0;
	while ((len=ap_get_client_block(r,opc_buffer+slen,OPC_BUFFER_SIZE-slen))>0) {
	    opc_buffer[slen+len]=0;
	    slen+=len;
	    ap_reset_timeout(r);
	}*/	

	*((uint32_t*)opc_buffer)=htonl(OPC_FECMD_XMLEND);
	err = dsConnectionWrite(conn, opc_buffer, sizeof(uint32_t), 0, &writed);
	if (err||(writed!=sizeof(uint32_t))) return opc_return_error(r,data,err?err:DS_ERR_SOCKET_WRITE,0,1);
	ap_kill_timeout(r);
    } 

    ds_ap_log_error(APLOG_MARK, APLOG_DEBUG, r->server, "Request is sent. Waiting for answer!");

	
	// Reading answer
    pos = 0;
    while (1) {
	if (!pos) {
	    err = dsConnectionRead(conn, opc_buffer, OPC_BUFFER_SIZE, 0, &readed);
	    if (err||(readed<sizeof(uint32_t))) return opc_return_error(r,data,err?err:DS_ERR_SOCKET_READ,0,1);
	}
	else if (readed<sizeof(uint32_t)) {
	    if (readed) memmove(opc_buffer, opc_buffer+pos, readed);
	    
	    err = dsConnectionRead(conn, opc_buffer + readed, OPC_BUFFER_SIZE - readed, 0, &pos);
	    if (err||(pos==0)) return opc_return_error(r,data,err?err:DS_ERR_SOCKET_READ,0,1);
	    readed+=pos;
	    if (readed<sizeof(uint32_t)) return opc_return_error(r,data,OPC_ERR_CLI_SRVINTERNAL,0,1);
	    pos = 0;
	}
    
	len=ntohl(*(uint32_t*)(opc_buffer+pos));
	if (len>OPC_FECMD) {
	    switch (len) {
		case OPC_FECMD_DIME_RESPONSE:
		    readed-=sizeof(uint32_t);
		    pos+=sizeof(uint32_t);
		    return opc_handler_dime(r, data, readed, pos);
		case OPC_FECMD_XMLEND:
		    if (attachments==-2) ap_rprintf(r, "\r\n--%s--", MIME_BOUNDARY);
		    return opc_return_error(r,data,0,0,0);
		case OPC_FECMD_INTERNAL_ERROR:
		    return opc_return_error(r,data,OPC_ERR_CLI_SRVINTERNAL,0,0);
		case OPC_FECMD_RCERROR:
		    return opc_return_error(r,data,OPC_ERR_CLI_SRVINITRC,0,0);
		case OPC_FECMD_OPC_INVALIDREQ:
		    return opc_return_error(r,data,OPC_ERR_CLI_XMLREQUEST,0,0);
		case OPC_FECMD_ATTACHMENT:
		    if (attachments<=0) {
			if (attachments<0) return opc_return_error(r, data, OPC_ERR_CLI_SRVINTERNAL,0,0);
			r->content_type = "multipart/related; boundary=" MIME_BOUNDARY "; type=text/xml; start=" MAIN_CONTENT_ID "\r\n";
			ap_send_http_header(r);
			attachments = 1;
		    }
		    readed-=sizeof(uint32_t);
		    pos+=sizeof(uint32_t);
		    err = opc_handler_attachment(r, data, &readed, &pos);
		    if (err) opc_return_error(r, data, err?err:OPC_ERR_CLI_SRVINTERNAL,0,1);
		    continue;
		default:
		    opc_log_invalid_srvresponse(r,(uint32_t*)opc_buffer,readed);
		    return opc_return_error(r,data,OPC_ERR_CLI_SRVRESPONSE,0,1);
	    }
	} else if (attachments>=0) {
	    if (attachments) {
		ap_rprintf(r, "--%s\r\n", MIME_BOUNDARY);
		ap_rprintf(r, "Content-Type: %s\r\n", "text/xml");
		ap_rprintf(r, "Content-ID: %s\r\n\r\n", MAIN_CONTENT_ID);
		attachments = -2;
	    } else {
	        r->content_type = "text/xml";
		ap_send_http_header(r); // send headers 
		attachments = -1;
	    }
	}
	readed-=sizeof(uint32_t);
	pos+=sizeof(uint32_t);
	

	if (readed<=len) {
	    if (readed) {
		ap_rwrite(opc_buffer+pos,readed,r);
		len-=readed;
	    }
	    pos = 0;
	    
	    if (len) {
		err = dsConnectionRead(conn, opc_buffer, OPC_BUFFER_SIZE, 0, &readed);
		if (err||(readed==0)) return opc_return_error(r,data, err?err:DS_ERR_SOCKET_READ,0,1);
	    
		while (readed<=len) {
		    ap_rwrite(opc_buffer,readed,r);
		    len-=readed;
		    if (len) {
			err = dsConnectionRead(conn, opc_buffer, OPC_BUFFER_SIZE, 0, &readed);
			if (err||(readed==0)) return opc_return_error(r,data, err?err:DS_ERR_SOCKET_READ,0,1);
		    }
		}
	    }
	}
	
	if (len) {
	    ap_rwrite(opc_buffer+pos,len,r);
	    pos+=len;
	    readed-=len;
	}
    }
    
    return opc_return_error(r,data,0,0,0);
}

/* DS: From apache articles:
    Called on process startup, not thread.
    For Linux only processes, for Windows only threads

    If it is true: In case of multiple threads non-windows environment
    we will have problem with DSConnectionDestructor. Close connection
    on the end of request!
*/


#ifdef DS_APACHE2
int opc_child_exit (void *sr) {
/* apache 2.2.x */
# ifndef WIN32
    pthread_key_delete(tls_keys);
#  endif
#else
void opc_child_exit (server_rec *sr, pool *sp) {
#endif /* DS_APACHE2 */
    dsApacheThreadDataDestroy();
}

#ifdef DS_APACHE2
static void keyfree(void *ptr) {
    if (ptr) free(ptr);
}

void opc_child_init(apr_pool_t *p, server_rec *sr) {
    apr_pool_cleanup_register(p, sr, opc_child_exit, NULL);
# ifndef WIN32
    pthread_key_create(&tls_keys, keyfree);
# endif
#else
void opc_child_init (server_rec *sr, pool *sp) {
#endif /* DS_APACHE2 */
    dsApacheThreadDataCreate(sr, NULL);
}

#ifdef DS_APACHE2
static int opc_init(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp) {
/* apache 2.2.x
# ifndef WIN32
    pthread_key_delete(tls_keys);
#  endif
*/
#else
void opc_init(server_rec *s, pool *p) {
#endif /* DS_APACHE2 */
    dsDebugSetLevel(DS_DEBUG_NOTHING);
}

#if defined WIN32
BOOL WINAPI DllMain(HINSTANCE dllhandle, DWORD reason, LPVOID reserved)
{
    LPVOID memptr;

    switch (reason) {
	case DLL_PROCESS_ATTACH:
    	    tls_index = TlsAlloc();

	case DLL_THREAD_ATTACH:     /* intentional no break */
	    memptr = malloc(sizeof(DSApacheThreadDataS));
	    ((DSApacheThreadData)memptr)->initialized = 0;
    	    TlsSetValue(tls_index, memptr);
        break;

	case DLL_THREAD_DETACH:
    	    memptr = TlsGetValue(tls_index);
	    if (memptr) {
		if (((DSApacheThreadData)memptr)->initialized)
	    	    dsConnectionFreeContext(&((DSApacheThreadData)memptr)->conn);
        	free(memptr);
        	TlsSetValue(tls_index, 0);
    	    }
    	break;
    }

    return TRUE;
}
#endif



#ifdef DS_APACHE2
static void register_hooks(apr_pool_t *p) {
    ap_hook_pre_config(opc_init, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_child_init(opc_child_init, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_handler(opc_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA mod_opc_module = {
    STANDARD20_MODULE_STUFF,
    NULL,     			/* dir config creater */
    NULL,                       /* dir merger --- default is to override */
    NULL,                       /* server config */
    NULL,                       /* merge server config */
    NULL,	                /* command apr_table_t */
    register_hooks              /* register hooks */
};

#else
/* Make the name of the content handler known to Apache */
static handler_rec opc_handlers[] =
{
    {"opc-handler", opc_handler},
    {NULL}
};

/* Tell Apache what phases of the transaction we handle */
module MODULE_VAR_EXPORT opc_module =
{
   STANDARD_MODULE_STUFF,
   opc_init,           /* module initializer                 */
   NULL,               /* per-directory config creator       */
   NULL,               /* dir config merger                  */
   NULL,               /* server config creator              */
   NULL,               /* server config merger               */
   NULL,               /* command table                      */
   opc_handlers,     /* [9]  content handlers              */
   NULL,               /* [2]  URI-to-filename translation   */
   NULL,               /* [5]  check/validate user_id        */
   NULL,               /* [6]  check user_id is valid *here* */
   NULL,               /* [4]  check access by host address  */
   NULL,               /* [7]  MIME type checker/setter      */
   NULL,               /* [8]  fixups                        */
   NULL,               /* [10] logger                        */
   NULL,               /* [3]  header parser                 */
   opc_child_init,     /* process initialization             */
   opc_child_exit,     /* process exit/cleanup               */
   NULL                /* [1]  post read_request handling    */
};
#endif /* DS_APACHE2 */
