#define _FRONTEND_C

#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dsdebug.h>
#include <dserror.h>
#include <dslock.h>

#include <opcxmldarequestpool.h>
#include <opccommands.h>

#include <opcerrors.h>
#include <frontend.h>
#include <frontendchild.h>

void opcFrontendZeroContext(OPCFrontendContext ctx) {
    ds_assert_return(ctx);

    ctx->consrv=NULL;
    ctx->buffpool=NULL;
    ctx->rpctx=NULL;
    ctx->sigctx=NULL;
    ctx->lock=NULL;
    ctx->child=NULL;
    ctx->config=NULL;
}

OPCFrontendContext opcFrontendCreateContext(DSMallocContext mctx) {
    OPCFrontendContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "opcFrontend: CreateContext");    

    ctx = (OPCFrontendContext)dsMalloc(mctx, sizeof(struct OPCFrontendContextT));
    if (ctx) {
	ctx->malloc = mctx;
	opcFrontendZeroContext(ctx);
    }
    return ctx;    
}

void opcFrontendFreeContext(OPCFrontendContext ctx) {
    ds_assert_return(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "opcFrontend: FreeContext");    

    if (ctx->consrv) dsConnectionServerDestroyContext(ctx->consrv);
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    if (ctx->child) dsFree(ctx->malloc, ctx->child);
    
    opcFrontendZeroContext(ctx);
    
}

void opcFrontendDestroyContext(OPCFrontendContext ctx) {
    ds_assert_return(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "opcFrontend: DestroyContext");    

    opcFrontendFreeContext(ctx);
    dsFree(ctx->malloc, ctx);
}

int opcFrontendInitContext(OPCFrontendContext ctx, OPCConfigContext config, OPCXmldaRequestPoolContext rpctx, DSBufferPoolContext bp, DSSignalContext sigctx) {
    int err;
    unsigned long i;
    DSConnectionServerOptionsS csopts;


    ds_assert_param(ctx);
    ds_assert_param(config);
    ds_assert_param(rpctx);    
    ds_assert_param(bp);
    ds_assert_param(sigctx);

    dsDebugServer(DS_DEBUG_NOTICE, "opcFrontend: InitContext");    
    opcFrontendZeroContext(ctx);

    ds_assert_param(config->max_frontend_threads);
    
    ctx->child = (OPCFrontendChildContext*)dsMalloc(ctx->malloc, config->max_frontend_threads*sizeof(OPCFrontendChildContext));
    if (!ctx->child) return DS_ERR_MEM;
    for (i=0;i<config->max_frontend_threads;i++) ctx->child[i] = NULL;
    
    ctx->consrv = dsConnectionServerCreateContext(ctx->malloc);
    if (!ctx->consrv) {
	opcFrontendFreeContext(ctx);
	return DS_ERR_CONNECTIONSERVER_CREATE;
    }
    
    err = dsConnectionServerInitDefaultOptions(&csopts);
    if (!err) err = dsConnectionServerOptionsSetLocalPath(&csopts, OPC_TEMP_DIR);
    if (!err) err = dsConnectionServerOptionsSetLocalExtension(&csopts, OPC_FRONTEND_EXTENSION);
    if ((!err)&&(config->srvname)) err = dsConnectionServerOptionsEnableLocal(&csopts, config->srvname);
    if ((!err)&&(config->srvport)) err = dsConnectionServerOptionsEnableNetwork(&csopts, config->srvport);
    if (!err) err = dsConnectionServerInitContext(ctx->consrv, &csopts);
    if (err) {
	opcFrontendFreeContext(ctx);
	return err;
    }


    ctx->lock = dsLockCreateContext(NULL);
    if (!ctx->lock) {
	opcFrontendFreeContext(ctx);
	return DS_ERR_LOCK_CREATE;
    }
    err = dsLockInitContext(ctx->lock);
    if (err) {
	opcFrontendFreeContext(ctx);
	return err;
    }

    ctx->buffpool = bp;
    ctx->rpctx = rpctx;
    ctx->sigctx = sigctx;
    ctx->config = config;

    return 0;
}


int opcFrontendListen(OPCFrontendContext ctx) {
    int err = 0;
    DSConnectionContext con;
    OPCFrontendChildContext child_ctx;
    char tmp_buffer[OPC_REQRESP_BUFFER_SIZE];
    unsigned long max_frontend_childs;
    unsigned long i = 0, j, flags;
    uint32_t answer;

    ds_assert_param(ctx);
    ds_assert_param(ctx->sigctx);
    ds_assert_param(ctx->rpctx);
    ds_assert_param(ctx->buffpool);
    ds_assert_param(ctx->config);
    ds_assert_param(ctx->consrv);
    
    max_frontend_childs = ctx->config->max_frontend_threads;

    dsSignalRegister(ctx->sigctx, DS_SIGNAL_NOTIFY);
    
    while (!dsSignalIsTerminated(ctx->sigctx)) {
	    // Clearing errors
	err = 0;

	    // Accepting connection
	con = dsConnectionServerAccept(ctx->consrv);
	if (!con) {
	    dsDebugServer(DS_DEBUG_ERROR, "opcFrontend: Connection failed");    
	    continue;
	}
	
	dsLockWrite(ctx->lock);
	
	    // Finding first free child slot
	if (i<max_frontend_childs) {
	    for (j=++i,flags=0;i<max_frontend_childs;i++) {
		if (!ctx->child[i]) { flags = 0; break; }
		else if (dsThreadCheckStatusBitsOn((DSBaseThreadContext)ctx->child[i],DS_THREAD_TERMINATE_FLAG)) { 
		    if (!flags) flags = i + 1;
		}
	    }
	}
	
	if (i==max_frontend_childs) {
	    for (i=0;i<j;i++) {
		if (!ctx->child[i]) { flags = 0; break; }
		else if (dsThreadCheckStatusBitsOn((DSBaseThreadContext)ctx->child[i],DS_THREAD_TERMINATE_FLAG)) { 
		    if (!flags) flags = i + 1;
		}
	    }

	    if (i==j) {
		if (flags) {
		    i = flags - 1;
		} else {
		    dsLockReturn(ctx->lock);

		    /* Wait a bit here */
		    err = dsConnectionTryRead(con,&tmp_buffer,OPC_REQRESP_BUFFER_SIZE,0,&j);
		    if ((!err)&&(j>=sizeof(uint32_t))&&((*(uint32_t*)tmp_buffer)==OPC_FECMD_NEWREQ)) {
			answer = htonl(OPC_FECMD_BUSY);
			err = dsConnectionTryWrite(con,&answer,sizeof(uint32_t),0,&j);
		    	dsDebugServer(err?DS_DEBUG_ERROR:DS_DEBUG_WARNING, "opcFrontend: Busy. Skiping Request");
		    } else {
		    	dsDebugServer(DS_DEBUG_ERROR, "opcFrontend: Busy. Skiping Request");    
		    }
		    dsConnectionDestroyContext(con);
		    continue;
		}
	    }
	}
	
	switch(flags) {
	    case 0:
		// Creating child ctx
		child_ctx = opcFrontendChildCreateContext(ctx->malloc);
		if (!child_ctx) err = OPC_ERR_SERVER_FRONTEND_CHILD_CREATE;
	    break;
	    default:
		child_ctx = ctx->child[i];
		opcFrontendChildFreeContext(child_ctx);
	}

	if (err) {
		// Not possible if flags = 0
	    dsLockReturn(ctx->lock);
	    dsConnectionDestroyContext(con);
	    dsDebugServer(DS_DEBUG_ERROR, "opcFrontend: Accept Error: %s", dsError(err, NULL));
	    continue;
	}
	
	err = opcFrontendChildInitContext(child_ctx, ctx, con);
	if (err) {
	    dsConnectionDestroyContext(con);
	    opcFrontendChildDestroyContext(child_ctx);
	    ctx->child[i] = NULL;
	    dsLockReturn(ctx->lock);
	    dsDebugServer(DS_DEBUG_ERROR, "opcFrontend: Accept Error: %s", dsError(err, NULL));    
	    continue;
	}
	
	err = dsDoubleThreadRun((DSDoubleThreadContext)child_ctx, (DSThreadFunction)opcFrontendChildReader, (DSThreadFunction)opcFrontendChildParser, child_ctx);
	if (err) {
	    opcFrontendChildDestroyContext(child_ctx);
	    ctx->child[i] = NULL;
	    dsLockReturn(ctx->lock);
	    dsDebugServer(DS_DEBUG_ERROR, "opcFrontend: Accept Error: %s", dsError(err, NULL));    
	    continue;
	}
	ctx->child[i] = child_ctx;
	dsLockReturn(ctx->lock);
    }

	// Terminating and cleaning childs
    for (i=0;i<max_frontend_childs;i++) {
	if (ctx->child[i]) 
	    dsDoubleThreadTerminate((DSDoubleThreadContext)ctx->child[i]);
    }    


    for (i=0;i<max_frontend_childs;i++) {
	if (ctx->child[i]) {
	    opcFrontendChildDestroyContext(ctx->child[i]);
	    ctx->child[i]=NULL;
	}
    }    

    return 0;
}


void opcFrontendCleanupChilds(OPCFrontendContext ctx) {
    unsigned long i;
    unsigned long max_frontend_childs;

    ds_assert_return(ctx);
    ds_assert_return(ctx->config);

    max_frontend_childs = ctx->config->max_frontend_threads;
    
    dsLockWrite(ctx->lock);
    for (i=0;i<max_frontend_childs;i++) {
	if ((ctx->child[i])&&(dsThreadCheckStatusBits((DSBaseThreadContext)ctx->child[i],DS_THREAD_TERMINATE_FLAG,DS_DOUBLETHREAD_RUNNING1_FLAG|DS_DOUBLETHREAD_RUNNING2_FLAG))) {
	    opcFrontendChildDestroyContext(ctx->child[i]);
	    ctx->child[i]=NULL;
	}
    }    
    dsLockReturn(ctx->lock);
}
