#include <stdio.h>

#include <opclib.h>

#include <dsassert.h>
#include <dsdebug.h>
#include <dserror.h>
#include <dsdime.h>
#include <dsdoublethread.h>
#include <dsconnection.h>

#include <opcxmldarequestpool.h>
#include <opcxmldarequest.h>


#include <opccommands.h>

#include <frontendchild.h>


void opcFrontendChildZeroContext(OPCFrontendChildContext ctx) {
    ds_assert_return(ctx);

    dsDoubleThreadZeroContext((DSDoubleThreadContext)ctx);
    
    ctx->con=NULL;
    ctx->hostname=0;
    ctx->username=0;
    ctx->fctx=NULL;
    ctx->buffpool=NULL;
    ctx->buffer=NULL;
    ctx->rpctx=NULL;
    ctx->rctx=NULL;
    ctx->lbp=NULL;
    ctx->err=0;

    ctx->lbp_lock = NULL;
    ctx->error_lock = NULL;
}

OPCFrontendChildContext opcFrontendChildCreateContext(DSMallocContext mctx) {
    OPCFrontendChildContext ctx;

    ctx = (OPCFrontendChildContext)dsMalloc(mctx, sizeof(struct OPCFrontendChildContextT));
    dsDebugServer(DS_DEBUG_NOTICE, "FrontendChild %p: CREATE", ctx);    
    if (ctx) {
	((DSConditionContext)ctx)->malloc = mctx;
	opcFrontendChildZeroContext(ctx);
    }
    return ctx;
}

int opcFrontendChildInitContext(OPCFrontendChildContext ctx, OPCFrontendContext fctx, DSConnectionContext con) {
    int err;
    
    dsDebugServer(DS_DEBUG_NOTICE, "FrontendChild %p: INIT", ctx);    

    ds_assert_param(ctx);
    ds_assert_param(fctx);
    ds_assert_param(fctx->rpctx);
    ds_assert_param(fctx->buffpool);
    ds_assert_param(con);

    opcFrontendChildZeroContext(ctx);

    err=dsDoubleThreadInitContext((DSDoubleThreadContext)ctx);
    if (err) return err;
    
    ctx->con=con;
    ctx->fctx=fctx;
    ctx->buffpool=fctx->buffpool;
    ctx->rpctx=fctx->rpctx;
    ctx->rctx=opcXmldaRequestPoolGetRequest(ctx->rpctx);
    
    ctx->buffer = dsBufferPoolGetBuffer(ctx->buffpool, 0);
    if (!ctx->buffer) {
	opcFrontendChildFreeContext(ctx);
	return DS_ERR_BUFFERPOOL_GET;
    }
    
    ctx->error_lock = dsLockCreateContext(((DSConditionContext)ctx)->malloc);
    ctx->lbp_lock = dsLockCreateContext(((DSConditionContext)ctx)->malloc);
    if ((!ctx->error_lock)||(!ctx->lbp_lock)) {
	opcFrontendChildFreeContext(ctx);
	return DS_ERR_MEM;
    }
    
    err = dsLockInitContext(ctx->error_lock);
    if (err) {
	opcFrontendChildFreeContext(ctx);
	return err;
    }

    err = dsLockInitContext(ctx->lbp_lock);
    if (err) {
	opcFrontendChildFreeContext(ctx);
	return err;
    }

    return 0;
}

void opcFrontendChildReturnBuffers(OPCFrontendChildContext ctx) {
    OPCFrontendChildBP clbp;	// Local buffer pool

    ds_assert_return(ctx);
    
    dsLockWrite(ctx->lbp_lock);
    while (ctx->lbp) {
	clbp=ctx->lbp;
	dsBufferPoolReturnBuffer(ctx->buffpool, clbp->buffer);
	ctx->lbp=clbp->next;
	dsFree(((DSConditionContext)ctx)->malloc, clbp);
    }
    dsLockReturn(ctx->lbp_lock);
}

void opcFrontendChildFreeContext(OPCFrontendChildContext ctx) {
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "FrontendChild %p: FREE", ctx);    

    ds_assert_return(ctx);

    dsDoubleThreadTerminateAndWait((DSDoubleThreadContext)ctx);
    opcFrontendChildReturnBuffers(ctx);

    if (ctx->hostname) dsFree(((DSConditionContext)ctx)->malloc, ctx->hostname);
    if (ctx->username) dsFree(((DSConditionContext)ctx)->malloc, ctx->username);
    if (ctx->con) dsConnectionDestroyContext(ctx->con);
    if (ctx->rctx) {
	err=opcXmldaRequestPoolReturnRequest(ctx->rpctx,ctx->rctx);
	ds_assert(err==0);
    }
    if (ctx->buffer) dsBufferPoolReturnBuffer(ctx->buffpool, ctx->buffer);
    
    if (ctx->error_lock) dsLockDestroyContext(ctx->error_lock);
    if (ctx->lbp_lock) dsLockDestroyContext(ctx->lbp_lock);
    
    dsDoubleThreadFreeContext((DSDoubleThreadContext)ctx);
    
    opcFrontendChildZeroContext(ctx);
}

void opcFrontendChildDestroyContext(OPCFrontendChildContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "FrontendChild %p: DESTROY", ctx);    
    opcFrontendChildFreeContext(ctx);
    dsFree(((DSConditionContext)ctx)->malloc, ctx);
}

int opcFrontendChildSendRespCode(OPCFrontendChildContext ctx, unsigned long code) {
    int err;
    uint32_t rcode;
    unsigned long writed;

    ds_assert_param(ctx);
    
    rcode = htonl(code);
    err = dsConnectionWrite(ctx->con,&rcode,sizeof(uint32_t),0,&writed);
    if (err) {
	if (err==DS_ERR_SIGNAL_EVENT) {
	    dsConnectionTryWrite(ctx->con,&rcode,sizeof(uint32_t),0,&writed);
/*
	    err = dsConnectionTryWrite(ctx->con,&rcode,sizeof(uint32_t),0,&writed);
	    if (err == DS_ERR_TIMEOUT) do something
*/
	}
    }
    return err;    
}

int opcFrontendChildReader(OPCFrontendChildContext ctx) {
    int err = 0;
    unsigned long readed;
    unsigned char hnl,unl;
    unsigned char UseBufferFlag=0;
    OPCFrontendChildBP clbp;
    DSBufferPoolBuffer current_buffer;
    unsigned char *buffer;
    char stmp[DS_BASE_THREAD_ID_SIZE];

    dsDebugServer(DS_DEBUG_NOTICE, "FrontendChild %p: READER (thread %s)", ctx, dsThreadGetID(stmp));

    ds_assert_param(ctx);

    while (dsThreadCheckStatusBitsOff((DSBaseThreadContext)ctx,DS_THREAD_TERMINATE_FLAG)) {
	if ((!UseBufferFlag)||(!readed)) {
	    if (!UseBufferFlag) {
		current_buffer=ctx->buffer;
		buffer = current_buffer->buffer;
	    }

	    err = dsConnectionRead(ctx->con, current_buffer->buffer, current_buffer->buffer_size, 0, &readed);
	    if (err) {
		if (err==DS_ERR_SIGNAL_EVENT) continue;
		dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
		break;
	    }
	} 

	if ((readed<sizeof(unsigned long))||((*(uint32_t*)buffer)!=OPC_FECMD_NEWREQ)) {
	    if (opcFrontendChildSendRespCode(ctx,OPC_FECMD_INVALIDREQ)) {
		dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
		break;
	    }
	    if (UseBufferFlag) readed=0;
	    continue;
	}
	
	if (ctx->hostname) dsFree(((DSConditionContext)ctx)->malloc, ctx->hostname);

	hnl=buffer[sizeof(uint32_t)];
	if (hnl) {
	    if ((hnl+2+sizeof(uint32_t))>readed) { /* checking for unl size also */
		ctx->hostname=NULL;
		if (opcFrontendChildSendRespCode(ctx,OPC_FECMD_INVALIDREQ)) {
		    dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
		    break;
		}
		if (UseBufferFlag) readed=0;
		continue;
	    }
	    ctx->hostname=(char*)dsMalloc(((DSConditionContext)ctx)->malloc, hnl+1);
	    if (ctx->hostname) {
		strncpy(ctx->hostname,buffer+sizeof(uint32_t)+1,hnl);
	        ctx->hostname[hnl]=0;
	    } else {
		dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
		break;	    
	    }
	} else ctx->hostname=NULL;
	
	if (ctx->username) dsFree(((DSConditionContext)ctx)->malloc, ctx->username);
	
	unl=buffer[sizeof(uint32_t)+hnl+1];
	if (unl) {
	    if ((unl+hnl+2+sizeof(uint32_t))>readed) {
		ctx->username=NULL;
		if (opcFrontendChildSendRespCode(ctx,OPC_FECMD_INVALIDREQ)) {
		    dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
		    break;
		}
		if (UseBufferFlag) readed=0;
		continue;
	    }
	    ctx->username=(char*)dsMalloc(((DSConditionContext)ctx)->malloc, unl+1);
	    if (ctx->username) {
		strncpy(ctx->username,buffer+hnl+sizeof(uint32_t)+2,unl);
	        ctx->hostname[unl]=0;
	    } else {
		dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
		break;	    
	    }
	} else ctx->username=NULL;

	dsDebugRequest(DS_DEBUG_NOTICE,"Frontend event (Connection %p)",ctx->con);
	if (hnl) dsDebugRequest(DS_DEBUG_INFO," client host: %s",ctx->hostname);
	if (unl) dsDebugRequest(DS_DEBUG_INFO," client user: %s",ctx->username);
	
	    // If RequestContext still uninitialized trying to initialize it
	if (!ctx->rctx) {
	    ctx->rctx=opcXmldaRequestPoolGetRequest(ctx->rpctx);
	    if (!ctx->rctx) {
		dsDebugRequest(DS_DEBUG_CRITICAL, "Internal Error. Request Aborted.");
		opcFrontendChildSendRespCode(ctx,OPC_FECMD_RCERROR);
		dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
		err = OPC_ERR_XLIB_REQUEST_CREATE;
		break;
	    }
	}

	    // Sending ACK
	if (opcFrontendChildSendRespCode(ctx,OPC_FECMD_NEWREQ)) {
	    dsDebugRequest(DS_DEBUG_CRITICAL, " Error Sending ACK response to client");
	    dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
	    break;
	}

	dsThreadResetStatusBits((DSBaseThreadContext)ctx,OPC_FRONTEND_CHILD_ENDEVENT_FLAG);
	while (dsThreadCheckStatusBitsOff((DSBaseThreadContext)ctx,DS_THREAD_TERMINATE_FLAG|OPC_FRONTEND_CHILD_ENDEVENT_FLAG)) {
	    if (UseBufferFlag) {
		UseBufferFlag=0;
	    } else {
		current_buffer=dsBufferPoolGetBuffer(ctx->buffpool, 0);
		if (current_buffer) buffer = current_buffer->buffer;
		else {
		    dsDebugRequest(DS_DEBUG_CRITICAL, "Internal Error. Request Aborted.");
		    dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
		    err=DS_ERR_BUFFERPOOL_GET;
		    break;
		} 
	    }
	    
	    err = dsConnectionRead(ctx->con, buffer, current_buffer->buffer_size, 0, &readed);
	    if (err) {
		if (err==DS_ERR_SIGNAL_EVENT) {
		    readed=0;
		    UseBufferFlag=1;
		    continue;
		}
		dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
		dsDebugRequest(DS_DEBUG_CRITICAL, " Error reading xml request from apache module (Error: %lu)\n", err);
		break;
	    }

	    dsLockWrite(ctx->error_lock);
	    if (dsThreadCheckStatusBitsOn((DSBaseThreadContext)ctx,OPC_FRONTEND_CHILD_ENDEVENT_FLAG)) {
		    UseBufferFlag=1;
		    dsLockReturn(ctx->error_lock);
		    break;
	    }

	    dsLockWrite(ctx->lbp_lock);
	    if (ctx->lbp) {
		clbp=ctx->lbp;
		while(clbp->next) clbp=clbp->next;
		clbp->next=(OPCFrontendChildBP)dsMalloc(((DSConditionContext)ctx)->malloc, sizeof(struct OPCFrontendChildBPT));
		clbp=clbp->next;
	    } else clbp=(OPCFrontendChildBP)dsMalloc(((DSConditionContext)ctx)->malloc, sizeof(struct OPCFrontendChildBPT));

	    if (!clbp) {
		dsDebugRequest(DS_DEBUG_CRITICAL, "Internal Error. Request Aborted.");
		dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
		dsLockReturn(ctx->lbp_lock);
		dsLockReturn(ctx->error_lock);
		err = DS_ERR_MEM;
		break;
	    }

	    clbp->buffer=current_buffer;
	    clbp->size=readed;
	    clbp->next=NULL;
	    if (!ctx->lbp) ctx->lbp=clbp;
	    
	    dsLockReturn(ctx->lbp_lock);
	    dsDoubleThreadSendCondition((DSBaseThreadContext)ctx);
	    dsLockReturn(ctx->error_lock);
	}
    }

    if (UseBufferFlag) {
	dsBufferPoolReturnBuffer(ctx->buffpool,current_buffer);
    }

    dsThreadResetStatusBits((DSBaseThreadContext)ctx, DS_DOUBLETHREAD_RUNNING1_FLAG);

    dsDebugServer(DS_DEBUG_INFO, "FrontendChild %p: READER Finished",ctx);

    if (err!=DS_ERR_SIGNAL_EVENT) return err;
    return 0;
}

void opcFrontendChildReturnBuffer(OPCFrontendChildContext ctx) {
    OPCFrontendChildBP clbp;	// Local buffer pool

    dsLockWrite(ctx->lbp_lock);
    clbp=ctx->lbp;
    dsBufferPoolReturnBuffer(ctx->buffpool,clbp->buffer);
    ctx->lbp=clbp->next;
    dsFree(((DSConditionContext)ctx)->malloc, clbp);
    dsLockReturn(ctx->lbp_lock);
}

int opcFrontendChildParserError(OPCFrontendChildContext ctx, int error, int code) {
    int err;
    
    ds_assert_param(ctx);
    
    dsLockWrite(ctx->error_lock);
    
	// Setting EndEvent Flag
    dsThreadSetStatusBits((DSBaseThreadContext)ctx,OPC_FRONTEND_CHILD_ENDEVENT_FLAG);

	// Free Buffers
    opcFrontendChildReturnBuffers(ctx);
    
	// ReInitParser
    err=opcXmldaRequestPoolReturnRequest(ctx->rpctx,ctx->rctx); ds_assert_param(err==0);
    ctx->rctx=opcXmldaRequestPoolGetRequest(ctx->rpctx);
    
    dsLockReturn(ctx->error_lock);
    
	// Send message back
    if (opcFrontendChildSendRespCode(ctx,code)) {
	dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
	dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason=Response Failed");
	return 1;
    }

    dsDebugRequest(DS_DEBUG_ERROR,"Request served, status=FAILED, reason = %s", dsError(error,NULL));

    return 0;
}

static int opcFrontendSendAttachmentCallback(OPCFrontendChildContext ctx, const char *buffer, unsigned long len) {
    int err;
    unsigned long writed;
    
    err = dsConnectionWrite(ctx->con,buffer,len,0,&writed);
    if (err) {
	ctx->err = err;
	if (err==DS_ERR_SIGNAL_EVENT) {
	    dsDebugRequest(DS_DEBUG_WARNING, "Request failed, status=Connection Closed, reason=Server Shutdown");
	} else {
	    dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason=Data Send Failed");
	}
	return -1;
    }

    return 0;
}

static int opcFrontendSendCallback(OPCFrontendChildContext ctx, const char *buffer, int len) {
    int err;
    int size;
    unsigned long writed;
    
    if ((!ctx)||(!buffer)||(len<0)) return -1;
    if (!len) return 0;

    if (ctx->format) {
	if (ctx->format==1) {
	    err = dsDimeSendHeader(opc_soap_message_id, len, DS_DIME_TYPE_XML, NULL, 0, 0, 1, (DSSendFunction)opcFrontendSendAttachmentCallback, ctx);
	    ctx->format=2;	    
	} else {
	    err = dsDimeSendChunk(len, 0, (DSSendFunction)opcFrontendSendAttachmentCallback, ctx);
	}
	if (err) {
	    ctx->err = err;
	    dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = Response Failed");
	    return -1;
	}

	if (len&3) size = (len&~3)+4; /* Buffer size should be multiple of 4 */
	else size = len;
    } else {
	err = opcFrontendChildSendRespCode(ctx,len);
	if (err) {
	    ctx->err = err;
	    dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = Response Failed");
	    return -1;
	}
	size = len;
    }

    err = dsConnectionWrite(ctx->con,buffer,size,0,&writed);
    if (err) {
	ctx->err = err;
	if (err==DS_ERR_SIGNAL_EVENT) {
	    dsDebugRequest(DS_DEBUG_WARNING, "Request failed, status=Connection Closed, reason=Server Shutdown");
	} else {
	    dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason=Data Send Failed");
	}
	return -1;
    }
    
    return len;
}

static void *opcFrontendOpenMIMEAttachmentCallback(OPCFrontendChildContext ctx, const char *path, const char *name, DSData info, DSData data) {
    int err;
    char *buffer;
    size_t pos;
    size_t len, pathlen, namelen;
    unsigned char c;

/*    if ((data->type!=OPC_DATA_TYPE_DATABUFFER)&&(data->type!=OPC_DATA_TYPE_ALLOCATION)) return NULL; */
    if (!data->size) return NULL;
    
    
    if (path) pathlen = strlen(path);
    if (name) namelen = strlen(name);
    
    if (pathlen) {
	if ((namelen)&&(*name!='/')) c=1;
	else c=0;
    } else {
	if ((namelen)&&(*name=='/')) c=0;
	else c=1;
    }

    len = c + namelen + pathlen;
    ds_assert_null(len < 0xFFFF);
    
    buffer = dsMalloca(len+2*sizeof(uint32_t)+sizeof(uint16_t)+1);
    if (!buffer) {
	ctx->err = DS_ERR_MEM;
	return NULL;
    }
    
    *(uint32_t*)buffer = htonl(OPC_FECMD_ATTACHMENT);
    *(uint32_t*)(buffer+sizeof(uint32_t)) = htonl(data->size);
    *(uint16_t*)(buffer+2*sizeof(uint32_t)) = htons((unsigned short int)len);
    pos = 2*sizeof(uint32_t)+sizeof(uint16_t);
    
    if (pathlen) {
	strcpy(buffer+pos, path);
	pos+=pathlen;
    }
    if (c) buffer[pos++]='/';
    if (namelen) strcpy(buffer+pos, name);

    len += 2*sizeof(uint32_t)+sizeof(uint16_t);

    err = dsConnectionWrite(ctx->con,buffer,len,0,NULL);
    if (err) {
	ctx->err = err;
	if (err==DS_ERR_SIGNAL_EVENT) {
	    dsDebugRequest(DS_DEBUG_WARNING, "Request failed, status=Connection Closed, reason=Server Shutdown");
	} else {
	    dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason=Data Send Failed");
	}
	dsFreea(buffer);
	return NULL;
    }

    dsFreea(buffer);
    return (void*)ctx;
}

static void *opcFrontendOpenDIMEAttachmentCallback(OPCFrontendChildContext ctx, const char *item, DSData info, DSData data) {
    int err;

/*    if ((data->type!=OPC_DATA_TYPE_DATABUFFER)&&(data->type!=OPC_DATA_TYPE_ALLOCATION)) return NULL; */
    if (!data->size) return NULL;
    
    if (ctx->format) {
	err = dsDimeSendHeader(item,data->size,DS_DIME_TYPE_BINARY,NULL,0,0,0,(DSSendFunction)opcFrontendSendAttachmentCallback, ctx);
	if (err) {
	    dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = DIME send failed");
	    ctx->err = err;
	    return NULL;
	}
    } else {
	err = opcFrontendChildSendRespCode(ctx,OPC_FECMD_DIME_RESPONSE);
	if (err) {
	    dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = Start DIME transfer failed");
	    ctx->err = err;
	    return NULL;
	}
	err = dsDimeSendHeader(item,data->size,DS_DIME_TYPE_BINARY,NULL,1,0,0,(DSSendFunction)opcFrontendSendAttachmentCallback, ctx);
	if (err) {
	    dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = DIME send failed");
	    ctx->err = err;
	    return NULL;
	}
	ctx->format = 1;
    }
	
    return (void*)ctx;
}


int opcFrontendChildParser(OPCFrontendChildContext ctx) {
    int err;
    char stmp[DS_BASE_THREAD_ID_SIZE];
    uint32_t len=0;			// Length of following message block
    unsigned long size;			// Byte length of buffer
    DSBufferPoolBuffer current_buffer;
    char *buffer;			// Position in buffer
    unsigned char ExitFlag;
//    struct timespec abstime;
    
    dsDebugServer(DS_DEBUG_NOTICE, "FrontendChild %p: PARSER (thread %s)",ctx, dsThreadGetID(stmp));

    ds_assert_param(ctx);

    while (dsThreadCheckStatusBitsOff((DSBaseThreadContext)ctx,DS_THREAD_TERMINATE_FLAG)) {
	dsConditionPreLock((DSConditionContext)ctx);
	while ((!ctx->lbp)&&(dsThreadCheckStatusBitsOff((DSBaseThreadContext)ctx,DS_THREAD_TERMINATE_FLAG))) {
//	    abstime.tv_sec=time(NULL)+opc_frontend_child_waittime;
//	    err=pthread_cond_timedwait(&ctx->lbp_cond,&ctx->lbp_cond_mutex,&abstime);
//	    if (err==ETIMEDOUT) continue;
	    dsConditionWaitLocked((DSConditionContext)ctx);
	}
	dsConditionUnLock((DSConditionContext)ctx);
	if (!ctx->lbp) break;
	
	dsDebugRequest(DS_DEBUG_INFO," Got a buffer, length=%lu",ctx->lbp->size);

	current_buffer=ctx->lbp->buffer;
	buffer = current_buffer->buffer;
	size=ctx->lbp->size;//-sizeof(unsigned long);
	
	for (ExitFlag=0;!ExitFlag;) {
	    dsDebugRequest(DS_DEBUG_DEBUG," %i %u",size,len);
	    if (!len) {
		if ((signed long)(size-len)<0) {
		    dsDebugRequest(DS_DEBUG_DEBUG,"%i %u...\n",size,len);
		    opcFrontendChildParserError(ctx,OPC_ERR_FRRONTED_REQUEST_INVALID,OPC_FECMD_INVALIDREQ);
		    len=0;
		    ExitFlag=1;
		    break;
		}
		len=ntohl(*(uint32_t*)buffer);
		buffer+=sizeof(uint32_t);
		size-=sizeof(uint32_t); // Error
	    } 
	    dsDebugRequest(DS_DEBUG_DEBUG,"r: %i %i %lx\n",size,len,len);

	    switch(len) {
		case OPC_FECMD_NEWREQ:
			// Sending Retransmit Response to Server
		    if (opcFrontendChildSendRespCode(ctx,OPC_FECMD_REPEAT)) {
			dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
			dsDebugRequest(DS_DEBUG_ERROR, " ERROR Sending REPEAT response to client");
		    }
		case OPC_FECMD_CANCEL:
		    dsDebugRequest(DS_DEBUG_NOTICE," Request canceled");

		    // Interrupt old request, start new one
		    opcFrontendChildReturnBuffer(ctx);
		    dsThreadSetStatusBits((DSBaseThreadContext)ctx,OPC_FRONTEND_CHILD_ENDEVENT_FLAG);

			// Free Buffers (all unused buffers)
		    dsLockWrite(ctx->error_lock);
		    opcFrontendChildReturnBuffers(ctx);
		    dsLockReturn(ctx->error_lock);
		    
			// Reinitialising Request Context
		    opcXmldaRequestPoolReturnRequest(ctx->rpctx,ctx->rctx);
		    ctx->rctx=opcXmldaRequestPoolGetRequest(ctx->rpctx);
		    ExitFlag=1;
		    len=0;
		break;
		case OPC_FECMD_XMLEND: 
		    // Check if no parts left (Multipart MIME)
		    dsDebugRequest(DS_DEBUG_NOTICE," END of XML Request");
		    opcFrontendChildReturnBuffer(ctx);

		    ctx->format = 0;
//		    err=opcXmldaRequestProcess(ctx->rctx, NULL, NULL, NULL, NULL);
		    err=opcXmldaRequestProcess(ctx->rctx, (OPCXmldaRequestOpenAttachmentFunction)&opcFrontendOpenDIMEAttachmentCallback, (DSSendFunction)&opcFrontendSendAttachmentCallback, NULL, ctx);
		    if (err) {
		    	opcFrontendChildParserError(ctx,err,OPC_FECMD_OPC_INVALIDREQ);
			len=0;
			ExitFlag=1;
			break;
		    }

		    dsThreadSetStatusBits((DSBaseThreadContext)ctx,OPC_FRONTEND_CHILD_ENDEVENT_FLAG);

		    err=opcXmldaRequestSend(ctx->rctx, (xmlOutputWriteCallback)opcFrontendSendCallback, NULL, ctx);
		    if (err) {
			if ((ctx->err)||(ctx->format)) {
			    dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = Response Serialization Failed");
			    dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
			} else {
		    	    opcFrontendChildParserError(ctx,err,OPC_FECMD_INTERNAL_ERROR);
			    len=0;
			}
			ExitFlag=1;
			break;
		    }
		    
		    if (ctx->format) {
			if (dsDimeSendCloseRecord((DSSendFunction)opcFrontendSendAttachmentCallback, ctx)) {
			    dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = Final DIME Record Send Failed");
			    dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
			    ExitFlag = 1;
			}
		    } else {
			if (opcFrontendChildSendRespCode(ctx,OPC_FECMD_XMLEND)) {
			    dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = EOF send failed");
			    dsDoubleThreadTerminate((DSDoubleThreadContext)ctx);
			    ExitFlag=1;
			    break;
			}			    
		    }

/*		    

		    if (opcFrontendChildSendRespCode(ctx,ctx->rctx->answer_size)) {
			opcFrontendChildTerminate(ctx);
			dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason = Response Failed");
			ExitFlag=1;
			break;
		    }
		    
		    err = opcSignalWrite(ctx->socket,ctx->rctx->answer,ctx->rctx->answer_size,0,&writed);
		    if ((err)||(writed!=ctx->rctx->answer_size)) {
			if (err==ERR_OPC_SIGNAL_EVENT) {
			    // Trying send it fast ones again (ToDo: check if something alredy sent)
			    writed += send(ctx->socket,ctx->rctx->answer+writed,ctx->rctx->answer_size-writed,0);
			    if (writed!=ctx->rctx->answer_size) {
				dsDebugRequest(DS_DEBUG_WARNING, "Request failed, status=Connection Closed, reason=Server Shutdown");
				ExitFlag=1;
				break;
			    }
			} else {
			    printf("Errno: %lx %lx (%lu %lu)\n",err, errno, ctx->rctx->answer_size, writed);
			    opcFrontendChildTerminate(ctx);
			    dsDebugRequest(DS_DEBUG_CRITICAL, "Request failed, status=Connection Closed, reason=Data Send Failed");
			    ExitFlag=1;
			    break;
			}
			
		    }
*/		    
		    dsDebugRequest(DS_DEBUG_NOTICE,"Request served, status=OK");

			// Free Buffers (all unused buffers)
/* DS: Error?
		    pthread_mutex_lock(&ctx->error_mutex);
		    opcFrontendChildReturnBuffers(ctx);
		    pthread_mutex_unlock(&ctx->error_mutex);
*/		    
			// Reinitialising Request Context
		    opcXmldaRequestPoolReturnRequest(ctx->rpctx,ctx->rctx);
		    ctx->rctx=opcXmldaRequestPoolGetRequest(ctx->rpctx);
		    ExitFlag=1;
		    len=0;
		break;
		default:
		    if (size<=len) { /* We have one or even part of message in buffer */
			if (!size) {
			    ExitFlag=1;
			    break;
			}
			dsDebugRequest(DS_DEBUG_INFO," Message, length=%lu",len);

			err=opcXmldaRequestXMLData(ctx->rctx,buffer,size);
			if (err) {
		    	    opcFrontendChildParserError(ctx,err,OPC_FECMD_OPC_INVALIDREQ);
			    len=0;
			    ExitFlag=1;
			    break;
			}
			opcFrontendChildReturnBuffer(ctx);
			ExitFlag=1;
			len-=size;
		    } else { /* We have several messages in buffer */
			dsDebugRequest(DS_DEBUG_INFO," Multipart Message, length=%lu",len);
			err=opcXmldaRequestXMLData(ctx->rctx,buffer,len);
			if (err) {
			    opcFrontendChildParserError(ctx,err,OPC_FECMD_OPC_INVALIDREQ);
			    len=0;
			    ExitFlag=1;
			    break;
			}	
			dsDebugRequest(DS_DEBUG_DEBUG, "%i %u!",size,len);		    
			buffer+=len;
			size-=len;
			len=0;
		    }
	    }
	    dsDebugRequest(DS_DEBUG_DEBUG, "EF: %i",ExitFlag);
	}
    }
    
    dsThreadResetStatusBits((DSBaseThreadContext)ctx, DS_DOUBLETHREAD_RUNNING2_FLAG);

    dsDebugServer(DS_DEBUG_INFO, "FrontendChild %p: PARSER Finished",ctx);
    return 0;
}

void opcFrontendChildClear(OPCFrontendChildContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "FrontendChild %p: CLEAR", ctx);    

    ds_assert_return(ctx);

    dsDoubleThreadTerminateAndWait((DSDoubleThreadContext)ctx);
    opcFrontendChildReturnBuffers(ctx);

    if (ctx->hostname) dsFree(((DSConditionContext)ctx)->malloc, ctx->hostname);
    if (ctx->username) dsFree(((DSConditionContext)ctx)->malloc, ctx->username);
    if (ctx->con) dsConnectionDestroyContext(ctx->con);
    if (ctx->rctx) opcXmldaRequestPoolReturnRequest(ctx->rpctx,ctx->rctx);

    ctx->con = NULL;
    ctx->hostname=0;
    ctx->username=0;

    ctx->rctx=NULL;
    ctx->err=0;
}

