#include <stdio.h>

#include <dslock.h>
#include <dserror.h>
#include <opcserver.h>
#include <opcserverinterface.h>

#ifdef _WIN32
# define OPC_LABVIEW_DATA_DIRECTORY "C:\\OPC"
# define export __declspec(dllexport)
#else
# define export
# define OPC_LABVIEW_DATA_DIRECTORY NULL
#endif /* _WIN32 */

static OPCServerContext ctx = NULL;
static OPCInterface iface = NULL;
static DSOPCBinaryDictionaryContext dict;
static DSLockContextS lock; 
static int status = 0;


static void opcLabviewStopInternal(int dolock) {
    if (!status) return;
    
    if (dolock) {
	dsLockWrite(&lock);
	if (!ctx) {
	    dsLockReturn(&lock);
	    return;
	}
    }
    
    dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: Stopping Server");
    
    if (iface) opcDestroyInterface(iface);
    opcServerDestroyContext(ctx);
    ctx = NULL;
    
    if (dolock) dsLockReturn(&lock);

    dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: Server Stopped");
}


static int opcLabviewStartInternal(unsigned char dolock) {
    int err;
    
    if (!status) return -1;
    
    if (dolock) {
        dsLockWrite(&lock);
	if (ctx) {
	    dsLockReturn(&lock);
	    return 0;
	}
    }

    dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: Starting Server");

    ctx = opcServerCreateContext();

    if (ctx) err = opcServerInitContext(ctx, OPC_LABVIEW_DATA_DIRECTORY, OPC_SERVER_NO_BACKEND);
    else err = DS_ERR_MEM;

    if (!err) err = opcServerRun(ctx);
    
    if (!err) {
	iface = opcCreateInterface(ctx->ctx);
	dict = dsOPCBinaryCreateDictionary(ctx->ctx->fd_opcbinary, "labview.xml");
	if (dict) opcFDOPCBinaryDictionaryLocalise(dict);
    }

/* DS: start mantainer thread */

    if (err) opcLabviewStopInternal(0);
    else dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: Server Started");

    if (dolock) dsLockReturn(&lock);


    return err;
}

int export opcLabviewStart() {
    return opcLabviewStartInternal(1);
}

void export opcLabviewStop() {
    opcLabviewStopInternal(1);
}

#ifdef _WIN32
BOOL WINAPI DllMain(HINSTANCE dllhandle, DWORD reason, LPVOID reserved)
{
    int err;
    switch (reason) {
	case DLL_PROCESS_ATTACH:
	    ctx = NULL;
	    dsDebugSetMethod(DS_DEBUG_METHOD_FILE, OPC_LABVIEW_DATA_DIRECTORY "\\opc.log");
	    dsDebugSetLevel(DS_DEBUG_ERROR);
	    dsDebugServer(DS_DEBUG_NOTICE, "opcLabview Loading");
	    err = opcServerInit();
	    if (!err) err = dsLockInitContext(&lock);
	    if (!err) {
		status = 1;
		break;
	    }
	case DLL_PROCESS_DETACH:
	    if (status) dsLockFreeContext(&lock);
	    opcServerDestroy();
	    dsDebugServer(DS_DEBUG_NOTICE, "opcLabview Unloading");
	    if (err) return FALSE;
	break;

	case DLL_THREAD_ATTACH:
        break;

	case DLL_THREAD_DETACH:
    	break;
    }

    return TRUE;
}
#else
void _init() {
    int err;
    ctx = NULL;
    dsDebugSetMethod(DS_DEBUG_METHOD_FILE, OPC_DIR "/opc.log");
    dsDebugSetLevel(DS_DEBUG_ERROR);
    dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: Loading");

    err = opcServerInit();
    if (!err) err = dsLockInitContext(&lock);
    if (!err) status = 1;

    dsDebugServer(DS_DEBUG_WARNING, "opcLabview: Loaded");
}

void _fini() {
    if (!status) return;
    dsDebugServer(DS_DEBUG_WARNING, "opcLabview: Unloading");
    opcLabviewStop();
    dsLockFreeContext(&lock);
    opcServerDestroy();
    dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: Unloaded");
}
#endif /* _WIN32 */



int export opcLabviewRegisterItem() {
    return 0;    
}

int export opcLabviewPublish(const char *path, const char *name, void *value) {
    int err;
    OPCItemContext ictx;
    DSBinaryType type;
    DSDataManagerContext dm;
    
    if (!status) return -1;
    
    dsLockWrite(&lock);
    if (!ctx) opcLabviewStartInternal(0);
    
    if (ctx) {
//	dsDebugServer(DS_DEBUG_NOTICE, "opcLabview: %s/%s = %lf", path, name, *value);
	
	ictx = opcInterfaceGetItem(iface, path, name);
	if (ictx) {
	    dm = ictx->dm;
	    err = 0;
	} else {

	    err = opcServerRegisterItem(ctx->ctx, &ictx, path, name);
	    if (err) ictx = NULL;
	    else {
		type = dsOPCBinaryDictionaryCreateType(dict, ((path)&&(*path))?path:"/");
		if (!type) 
		    type = dsOPCBinaryDictionaryCreateType(dict, ictx->fullname);
		if (type) {
		    dm = (DSDataManagerContext)opcItemRegisterLegacyDataManager(ictx, type, NULL, 0);
		    opcItemEnableDataManager(ictx);
		}
		opcItemResetManualMode(ictx);
	    }
	    
	    err = 1;
	}


	if (ictx) {
	    opcDataPushData(dm, 0/*NativeRepresentation*/, NULL, value, 0);
	    if (!err) opcItemReturn(ictx);
	}
	
    }
    dsLockReturn(&lock);
    return 0;
}



/*
int __declspec(dllexport) opcLabviewPushData(const char *ItemName, const char *DataType, const void *data, unsigned long options, unsigned long timeout) {
}

/*
opcLabviewPushData
opcLabviewWaitRequest

opcLabviewReadData
//opcLabviewWaitData

Properties

Labview Client
*/
