#include <stdio.h>

#include <dserror.h>
#include <dsdebug.h>
#include <dssignal.h>
#include <dssoftsignal.h>

#include <opcserver.h>

DSSoftSignalContext exitsig = NULL;
static int RunFlag = 1;

static void signal_exit_handler(int signo) { 
    RunFlag = 0; 
    dsSoftSignalSend(exitsig);
}

int InitSignals() {
    int err;
    
    exitsig = dsSoftSignalCreateContext(NULL);
    if (!exitsig) err = DS_ERR_MEM;
    else err = dsSoftSignalInitContext(exitsig);

    if (!err)
	err = dsPosixSignalSetAction(DS_POSIX_SIGNAL_SIGTERM, NULL, signal_exit_handler);
    
    if (!err)
	err = dsPosixSignalBlock(DS_POSIX_SIGNAL_SIGTERM);

    if (!err) {
#ifdef DEBUG_GDB
	dsPosixSignalSetAction(DS_POSIX_SIGNAL_SIGINT, NULL, DS_SIGNAL_NOP_HANDLER);
#else /* DEBUG_GDB */
	dsPosixSignalSetAction(DS_POSIX_SIGNAL_SIGINT, NULL, signal_exit_handler);
	dsPosixSignalBlock(DS_POSIX_SIGNAL_SIGINT);
#endif /* DEBUG_GDB */

	dsPosixSignalSetAction(DS_POSIX_SIGNAL_SIGTSTP, NULL, DS_SIGNAL_NOP_HANDLER);
    }
}

void FreeSignals() {
    dsPosixSignalSetAction(DS_POSIX_SIGNAL_SIGTERM, NULL, DS_SIGNAL_NOP_HANDLER);
#ifndef DEBUG_GDB
    dsPosixSignalSetAction(DS_POSIX_SIGNAL_SIGINT, NULL, DS_SIGNAL_NOP_HANDLER);
#endif /* !DEBUG_GDB */
    dsSoftSignalDestroyContext(exitsig);
}

int UnblockSignals() {
    int err;
    err = dsPosixSignalUnblock(DS_POSIX_SIGNAL_SIGTERM);
#ifndef DEBUG_GDB
    dsPosixSignalUnblock(DS_POSIX_SIGNAL_SIGINT);
#endif /* !DEBUG_GDB */
    return err;
}

void BlockSignals() {
    dsPosixSignalBlock(DS_POSIX_SIGNAL_SIGTERM);
#ifndef DEBUG_GDB
    dsPosixSignalBlock(DS_POSIX_SIGNAL_SIGINT);
#endif /* !DEBUG_GDB */
}


main() {
    int err;
    const char *serr;
    OPCServerContext ctx;
    DSTime next;

    dsDebugSetLevel(DS_DEBUG_INFO);

    err = InitSignals();
    if (err) {
	serr = dsError(err,NULL);

	printf("Initalization Error: Signal Handling initialization is failed with Error: %lx\n", err);
	printf("%s\n", serr);
    }

    err = opcServerInit();
    if (err) {
	FreeSignals();
	printf("Initialization Error: Server Global Initialization Error: %lx\n", err);
	return 0;
    }
    
    ctx=opcServerCreateContext();
    if (!ctx) {
	serr = dsError(err,NULL);
	opcServerDestroy();
	FreeSignals();

	printf("Initialization Error: Cann't allocate memory for OPCServerContext\n");
	printf("%s\n", serr);
	return 0;
    }
    
    err=opcServerInitContext(ctx, NULL, 0);
    if (err) {
	serr = dsError(err,NULL);
	opcServerDestroyContext(ctx);
	opcServerDestroy();
	FreeSignals();

	printf("Initalization Error: Server Initialization Error: %lx\n", err);
	printf("%s\n", serr);
	return 0;
    }
    
    err=opcServerRun(ctx);
    if (err) {
	serr = dsError(err,NULL);
	opcServerDestroyContext(ctx);
	opcServerDestroy();
	FreeSignals();

	printf("Execution Error: Server Execution Error: %lx\n", err);
	printf("%s\n",serr);
	return 0;
    }
        
    next = dsTimeCreate(NULL);
    if (next) {
	err = dsTimeCurrentTime(next);
	if (!err) dsTimeAddSeconds(next, OPC_MANTAIN_DELAY);
    } else err = DS_ERR_MEM;

    err = err?err:UnblockSignals();
    
    while ((RunFlag)&&(!err)) {
	puts("Iteration");

	err = dsSoftSignalWaitUntil(exitsig, next);
	if (err) {
	    if (err == DS_ERR_SIGNAL_EVENT) err = 0;
	    continue;
	}

	opcServerMantain(ctx);

	err = dsTimeCurrentTime(next);
	if (!err) dsTimeAddSeconds(next, OPC_MANTAIN_DELAY);
    }

    BlockSignals();

    if (next) dsTimeDestroy(next);

    serr = dsError(err,NULL);
    
    opcServerDestroyContext(ctx);
    opcServerDestroy();
    
    FreeSignals();

    printf("\nFinished: %s\n\n",serr);
    return 0;
}
