/* RCS_Id: formvars.c 10.2 1999/12/10 16:47:09 clark Exp $ */
//
// formvars.c - The function(s) in this module are used to convert a query
//	string (the pQuery field in the REQ_INPUTS structure) into a linked list
//	of form variable pairs.  Each form variable pair is made up of the
//	variable's name and its setting.
//
//	The function GetFormVariableList() is used to build the linked list of
//	pairs.  The function DeleteFormVariableList() is used to delete the
//	memory allocated for the list.
//

  //////////////
 // Includes //
//////////////

#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <windows.h>

#include <stdlib.h>
#include <time.h>
#include <httpserv.h>

#include "httputil.h"


  /////////////
 // Defines //
/////////////

// See if the character is a hex digit.
#define IsHex(ch) ( ((ch >= 'a') && (ch <= 'f')) \
		  || ((ch >= 'A') && (ch <= 'F')) \
		  || ((ch >= '0') && (ch <= '9')) )


  //////////////////////
 // Type Definitions //
//////////////////////

//
// This structure points to the first and last member of a singly linked list
// of FORM_DATA structures.
//
typedef struct {

    FORM_VARIABLE *pFirst;	// Pointer to first form variable in list.
    FORM_VARIABLE *pLast;	// Pointer to last form variable in list.

} FORM_VAR_LIST;


  ///////////////////////////////////
 // Private Function Declarations //
///////////////////////////////////

static short ConvertNonAlphanumerics (char *pStr);

static FORM_VAR_LIST *FormVarListNew        (void);
static FORM_VARIABLE *FormVarNew            (void);
static char          *FormVarSetStringField (char *pStr,
					     DWORD StrLen,
					     short *pErrVal);
static short          FormVarSet            (FORM_VARIABLE *pFormVar,
					     char *pStr,
					     DWORD StrLen);
static void           FormVarListAdd        (FORM_VAR_LIST *pVarList,
					     FORM_VARIABLE *pNewVar);
static void           FormVarDelete         (FORM_VARIABLE *pFormVar);
static void           FormVarListDelete     (FORM_VAR_LIST *pVarList);


  /////////////////////////////////
 // Public Function Definitions //
/////////////////////////////////

//
// GetFirstFormVariable() - This function returns a list of the form variables
//	and their settings.
//
// Input:
//	pQuery		The query string to be broken into variable pairs.
//
// Outpput:
//	pErrVal 	An error value describing the reason the function
//			failed.
//
//			Success		FORM_NO_ERR		(0)
//			Fail (syntax)	FORM_SYNTAX_ERR		(-1)
//			Fail (malloc)	FORM_ALLOC_ERR		(-2)
//
// Return:
//	Success		A pointer to a variable list.
//	Failure		NULL
//
// Warning:
//	This function may also return NULL if pQuery is the NULL string.  Make
//	sure to check *pErrVal.  This parameter will be set to FORM_NO_ERR if
//	pQuery was empty.
//
FORM_VARIABLE * __cdecl GetFormVariableList (char *pQuery, short *pErrVal)
{
    char  *pCur;
    char  *pTemp;
    short  ErrVal;

    FORM_VAR_LIST *pVarList;
    FORM_VARIABLE *pNewVar;
    FORM_VARIABLE *pFirstVar;


    // An empty form was passed to the server.  Return NULL but set *pErrVal to
    // no error.
    if ( pQuery[0] == '\0' ) {

	if ( pErrVal != NULL ) *pErrVal = FORM_NO_ERR;

	return NULL;
    }

    //
    // Allocate a FORM_VAR_LIST structure.  this structure will point to the
    // first and last element in the list of variables.
    //
    pVarList = FormVarListNew();
    if (! pVarList)
    {
	if (pErrVal != NULL)
	    *pErrVal = FORM_ALLOC_ERR;
	return NULL;
    }

    ErrVal = FORM_NO_ERR;

    pCur = pQuery;

    //
    // Cycle through the query string extracting each variable and value pair.
    //
    while ( *pCur != '\0' ) {

	//
	// An amphersand or NULL character marks the end of a variable setting.
	//
	if ( ((pTemp = strchr(pCur, '&')) == NULL)
	  && ((pTemp = strchr(pCur, '\0')) == NULL) ) {

	    ErrVal = FORM_SYNTAX_ERR;

	    break;
	}

	//
	// If we found an amphersand or NULL character, allocate a FORM_VARIABLE
	// structure.  This structure will hold the variable name and its
	// setting.
	//
	if ( (pNewVar = FormVarNew()) == NULL ) {

	    ErrVal = FORM_ALLOC_ERR;

	    break;
	}

	//
	// Set the FORM_VARIABLE to the values in the query string.  If this
	// function fails, we need to delete pNewVar before we break out of the
	// loop.
	//
	if ( (ErrVal = FormVarSet(pNewVar,
				  pCur,
				  pTemp - pCur)) != FORM_NO_ERR ) {

	    FormVarDelete(pNewVar);

	    break;
	}

	//
	// Add the new FORM_VARIABLE to the list.
	//
	FormVarListAdd(pVarList, pNewVar);

	pCur = pTemp;

	if ( *pTemp != '\0' )

	    pCur++;
    }

    //
    // If we had an error, we want to delete the list of form variables.
    //
    if (ErrVal == FORM_NO_ERR)
    {
	pFirstVar = pVarList->pFirst;
	// if pVarList, not needed, handled above
	free(pVarList);
    }
    else
    {
	pFirstVar = NULL;
	FormVarListDelete(pVarList);
    }

    if (pErrVal != NULL)
	*pErrVal = ErrVal;

    return (pFirstVar);
}


//
// DeleteFormVariableList() - This function deletes a list of form variables.
//	The memory for these variables was allocated in the
//	GetFormVariableList() function.  This function should always be called
//	to free the list allocated by GetFormVariableList().
//
// Input:
//	pVarList	A pointer to the variable list to be deleted.
//
void __cdecl DeleteFormVariableList (FORM_VARIABLE *pVarList)
{
    FORM_VARIABLE *pCur;
    FORM_VARIABLE *pNext;


    pCur = pVarList;

    while ( pCur != NULL ) {

	pNext = pCur->pNext;

	FormVarDelete(pCur);

	pCur = pNext;
    }
}


  //////////////////////////////////
 // Private Function Definitions //
//////////////////////////////////

//
// ConvertNonAlphanumerics() - When a query is sent, all non-alphanumeric
//	characters are convert to a percent sign followed by their equivalent
//	hex values.  This function converts them back to characters.
//
// Output:
//	pStr	This string will be returned with all of the non-alphanumeric
//		characters converted from %xx to the actual character.
//
// Return:
//	Success		FORM_NO_ERR      (0)
//	Fail (syntax)	FORM_SYNTAX_ERR (-1)
//	Fail (malloc)	FORM_ALLOC_ERR  (-2)
//
static short ConvertNonAlphanumerics (char *pStr)
{
    char  *pCurParse;
    char  *pCurAdd;
    char   SaveCh;
    short  RetVal;


    pCurParse = pCurAdd = pStr;

    RetVal = FORM_NO_ERR;

    //
    // Loop throught the string replacing all %xx strings with the
    // corresponding character.
    //
    while ( (*pCurParse != '\0') && (RetVal == FORM_NO_ERR) ) {

	switch (*pCurParse) {

	    case '%':

		pCurParse++;

		//
		// Make sure the next two characters are actually hex digits.
		//
		if ( !IsHex(*pCurParse) || !IsHex(*(pCurParse + 1)) )

		    RetVal = FORM_SYNTAX_ERR;

		else {

		    //
		    // Advance pCur to the end of the unprintable character.
		    //
		    SaveCh = *(pCurParse + 2);

		    // Set pCur to the NULL character so we can use strtoul().
		    *(pCurParse + 2) = '\0';

		    *pCurAdd = (char) strtoul(pCurParse,
					      NULL,
					      NONALPHA_CHAR_BASE);

		    *(pCurParse + 2) = SaveCh;

		    pCurAdd++;

		    pCurParse += 2;
		}

		break;

	    case '+':

		*pCurAdd = ' ';

		pCurParse++;

		pCurAdd++;

		break;

	    default:

		*pCurAdd = *pCurParse;

		pCurAdd++;

		pCurParse++;

		break;
	}
    }

    //
    // If everything worked, add a '\0' to the end of the string.
    //
    if ( RetVal == FORM_NO_ERR )

	*pCurAdd = '\0';


    return(RetVal);
}


//
// The following functions operate on the FORM_VAR_LIST and FORM_VARIABLE
// structures.
//

//
// FormVarListNew() - Allocates a new FORM_VAR_LIST structure.
//
static FORM_VAR_LIST *FormVarListNew (void)
{
    FORM_VAR_LIST *pVarList;

    if ( (pVarList = (FORM_VAR_LIST *) malloc(sizeof(FORM_VAR_LIST))) != NULL )

	memset(pVarList, 0, sizeof(*pVarList));

		;
    return(pVarList);
}


//
// FormVarNew() - Allocates a new FORM_VAR structure.
//
static FORM_VARIABLE *FormVarNew (void)
{
    FORM_VARIABLE *pVar;


    if ( (pVar = (FORM_VARIABLE *) malloc(sizeof(FORM_VARIABLE))) != NULL )

	memset(pVar, 0, sizeof(*pVar));


    return(pVar);
}


//
// FormVarSetStringField() - This function allocates memory for the string
//	passed to it, sets the memory to the string passed in, and converts
//	all escape sequences (%xx) to their equivalent characters
//
// Inputs:
//	pStr	A portion of the query string containing a variable and its
//		setting.
//	StrLen	The length of pStr.
//
// Outputs:
//	pErrVal		A value that equates to why the function failed.
//
// Return:
//	Success		A pointer to the new string.
//	Failure		NULL
//
static char *FormVarSetStringField(char *pStr, DWORD StrLen, short *pErrVal)
{
    char  *pNewStr;
    short  ErrVal;

    pNewStr = SetNewString(pStr, StrLen);
    if (! pNewStr)
    {
	ErrVal = FORM_ALLOC_ERR;
    }
    else
    {
	ErrVal = ConvertNonAlphanumerics(pNewStr);
	if (ErrVal != FORM_NO_ERR)
	{
	    // if (pNewStr) not needed
	    free(pNewStr);
	    pNewStr = NULL;
	}
    }

    if (pErrVal)
	*pErrVal = ErrVal;

    return (pNewStr);
}


//
// FormVarSet() - This function breaks the string passed in to the form variable
//	and value.
//
// Inputs:
//	pStr	A portion of the query string containing a variable and its
//		setting.
//	StrLen	The length of pStr.
//
// Outputs:
//	pFormVar	This structure will have its pVar and pVal fields set
//			to hold a form variable's name and its value.
//
// Return:
//	Success			FORM_NO_ERR		(0)
//	Fail (syntax)		FORM_SYNTAX_ERR		(-1)
//	Fail (malloc)		FORM_ALLOC_ERR		(-2)
//
static short FormVarSet (FORM_VARIABLE *pFormVar, char *pStr, DWORD StrLen)
{
    char  *pCur;
    char  *pTemp;
    char   SaveCh;
    short  RetVal;


    // Since strchr() needs a null-terminated string, we will temporarily
    // terminate the string passed ninto us.
    SaveCh = *(pStr + StrLen);

    *(pStr + StrLen) = '\0';

    pFormVar->pVar = pFormVar->pVal = NULL;

    pCur = pStr;

    if ( (pTemp = strchr(pStr, '=')) == NULL ) {

	RetVal = FORM_SYNTAX_ERR;

	goto Err;
    }

    if ( (pFormVar->pVar = FormVarSetStringField(pCur,
						 pTemp - pCur,
						 &RetVal)) == NULL )

	goto Err;

    pCur = pTemp + 1;

    if ( ((pTemp = strchr(pCur, '&')) == NULL)
	 && ((pTemp = strchr(pCur, '\0')) == NULL) ) {

	RetVal = FORM_SYNTAX_ERR;

	goto Err;
    }

    if ( (pFormVar->pVal = FormVarSetStringField(pCur,
						 pTemp - pCur,
						 &RetVal)) == NULL )

	goto Err;

    *(pStr + StrLen) = SaveCh;

    return (FORM_NO_ERR);

Err:

    *(pStr + StrLen) = SaveCh;

    if (pFormVar->pVar)
    {
	free(pFormVar->pVar);
	pFormVar->pVar = NULL;
    }
    if (pFormVar->pVal)
    {
	free(pFormVar->pVal);
	pFormVar->pVal = NULL;
    }
    return (RetVal);
}


//
// FormVarListAdd() - This function adds a form variable to the list of form
//	variables.
//
// Inputs:
//	pNewVar		The form variable to be added to the form variable list.
//
// Outputs:
//	pVarList	The list of form variables that the new variable will
//			be added to.
//
static void FormVarListAdd (FORM_VAR_LIST *pVarList, FORM_VARIABLE *pNewVar)
{
    if ( pVarList->pFirst == NULL )

	pVarList->pFirst = pVarList->pLast = pNewVar;

    else {

	pVarList->pLast->pNext = pNewVar;

	pVarList->pLast = pNewVar;
    }

}


//
// FormVarDelete() - This function deletes a FORM_VARIABLE structure and all
//	memory associated with that structure (i.e. the memory used for some
//	of the fields).
//
// Inputs:
//	pFormVar	The formvariable to be deleted.
//
static void FormVarDelete(FORM_VARIABLE *pFormVar)
{
    if (pFormVar)
    {
	if (pFormVar->pVar)
	{
	    free(pFormVar->pVar);
	}
	if (pFormVar->pVal)
	{
	    free(pFormVar->pVal);
	}
	free(pFormVar);
    }
}


//
// FormVarListDelete() - This function deletes all of the members of the list
//	of FORM_VARIABLES associated with the FORM_VAR_LIST structure and then
//	deletes the FORM_VAR_LIST structure itself.
//
// Inputs:
//	pVarList	The form variable list to be deleted.
//
static void FormVarListDelete(FORM_VAR_LIST *pVarList)
{
    if (pVarList)
    {
	DeleteFormVariableList(pVarList->pFirst);
	free(pVarList);
    }
}
