// RCS_Id: httpserv.c 10.3 2000/05/05 10:22:20 clark Exp $
//
// httpserv
// This library implements an HTTP/1.0 server that complies with RFC 1945.
// The interface to the server is defined by HTTPSERV.H.
//
// This is the source code to the HTTPSERV.LIB file in \PHAREMB\LIB\*CLIB

#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <winsock.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <process.h>
#include <time.h>
#include <sys\types.h>
#include <httpserv.h>
#include <embkern.h>

#include "httputil.h"

#if defined(__BORLANDC__)
#define _strnicmp strnicmp
#endif

#define STACK_SIZE 0x4000	// Size of per-thread stack
#define HTTP_TIMEOUT 60		// HTTP client inactive timeout, in seconds
#define EOS '\0'		// end-of-string character
#define isdig(a) ((a) >= '0' && (a) <= '9')
#define isxdig(a) (((a) >= '0' && (a) <= '9') || \
			((a) >= 'a' && (a) <= 'f') || \
			((a) >= 'A' && (a) <= 'F'))

//
// Struct used to save one line of text of an HTTP request
//
typedef struct request_line {
	struct request_line *pNext;	// next in list
	char	*pLineText;		// EOS-terminated text
} REQUEST_LINE;

//
// Per-thread connection struct allocated for each thread created to service
// an HTTP request.
//
typedef struct {
	SOCKADDR_IN RemoteAddr;		// remote TCP port and IP addr
	SOCKET	hSock;			// socket for accepted connection
	time_t	CliRequestTime;		// time client request was accepted
	time_t	ResponseTime;		// time of last response send
	char	Buf[1024];		// receive data buffer
	char	*pBuf;			// ptr to 1st empty byte in buffer
} HTTP_CONNECT;

//
// Function prototypes
//
static void	DoShutdown(BOOL fSignalServerThread);
static BOOL	StartThread(void (*pFunc)(void *), void *arglist);
static void	HttpThread(short Port);
static void	ServiceRequestThread(HTTP_CONNECT *pConnect);
static REQUEST_LINE *ReadRequestLine(HTTP_CONNECT *pConn, REQ_OUTPUTS *pOutp,
				     BOOL *pfSendResponse);
static BOOL	ReadEntity(HTTP_CONNECT *pConn, REQ_INPUTS *pInp,
			   REQ_OUTPUTS *pOutp, BOOL *pfSendResponse);
static BOOL	SendResponse(HTTP_CONNECT *pConnect, char *pBuf, int Len);
static void	MakeHttpTime(time_t Time, char *pBuf, int gmt);
static time_t	ReadHttpTime(char *pStr);
static BOOL	ParseURI(HTTP_CONNECT *pConn, char *pURI, REQ_INPUTS *pInp,
			 REQ_OUTPUTS *pOutp);
static void	SendRequestLog(HTTP_CONNECT *pConnect, REQUEST_LINE *pLine1,
			       char *pStatusCode);
static void	SendError(BOOL fFatal, SOCKADDR_IN *pRemoteAddr,
			  const char *pFormat, ...);
static void	HtmlErrorPage(REQ_OUTPUTS *pOutp, const char *pFormat, ...);

static void	HandlePostError(short RetVal, SOCKADDR_IN *pRemoteAddr,
				REQ_INPUTS *pInp, REQ_OUTPUTS *pOutp);
static char	HexToBin(char c);
static char	*GetHeaderField(REQUEST_LINE *pLine, char *pFieldName);
static void	slurpExtraBytes(HTTP_CONNECT *pConnect);

//
// Global variables.
//
static BOOL fRunning=FALSE;	// T ==> web server active, F ==> inactive
static WSADATA	wsaData;	// WinSock data
static HANDLE	hTerminateEvent;// event used to signal program termination
static HANDLE	hKilledEvent;	// event used to signal main thread dead
static CRITICAL_SECTION csLog; // critical section to serialize request logging
static CRITICAL_SECTION csCleanup; // critical section for shutting down

//
// Status codes
//
static char *pStat200 = "200 OK";
static char *pStat304 = "304 Not Modified";
static char *pStat400 = "400 Bad Request";
static char *pStat404 = "404 Not Found";
static char *pStat500 = "500 Internal Server Error";
static char *pStat501 = "501 Not Implemented";

//
// Starts the HTTP server
//	Port	0 ==> use well-known HTTP port (80), otherwise use
//		specified port (passed in host byte order)
// Returns:
//	TRUE	if success
//	FALSE	if error (error message passed to ServerError())
//
BOOL __cdecl StartWebServer(short Port)
{
	int	ErrVal;
	time_t	now;
	char	timeStr[128];

	if (fRunning)
	{
		SendError(FALSE, NULL, "StartWebServer called when HTTP "
			"server is already running");
		return FALSE;
	}

//
// init WinSock interface
//
	ErrVal = WSAStartup(0x0101, &wsaData);
	if (ErrVal != 0)
	{
		SendError(TRUE, NULL, "WSAStartup() returned error %d",ErrVal);
		return FALSE;
	}
	if (wsaData.wVersion != 0x0101)
	{
		SendError(TRUE, NULL, "WinSock 1.1 not supported");
		goto ERR1;
	}

//
// create synchronization objects we need
//
	hTerminateEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (! hTerminateEvent)
	{
		SendError(TRUE, NULL, "CreateEvent() for termination event "
			"returned error %d", GetLastError());
		goto ERR1;
	}
//DS	EtsSetDebugName(hTerminateEvent, "HTTP Server Terminate Event");

	hKilledEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
	if (! hKilledEvent)
	{
	    SendError(TRUE, NULL, "CreateEvent() for termination event "
		      "returned error %d", GetLastError());
	    goto ERR1;
	}
//DS	EtsSetDebugName(hKilledEvent, "HTTP Server Terminated Event");

//
// Print out GMT and Local time.
//
	time(&now);
	MakeHttpTime(now, timeStr, 1);
	strcat(timeStr, "\n");
	printf(timeStr);
	MakeHttpTime(now, timeStr, 0);
	strcat(timeStr, "\n");
	printf(timeStr);

//
// create HTTP thread
//
	if (! StartThread((void (*)(void *))HttpThread, (void *)Port))
	{
		SendError(TRUE, NULL, "Can't create HTTP server thread");
		goto ERR2;
	}

//
// flag server running and return success
//
	fRunning = TRUE;
	return TRUE;

ERR2:
	CloseHandle(hTerminateEvent);

ERR1:
	WSACleanup();
	return FALSE;
}

//
// Stops operation of the HTTP server
//
void __cdecl StopWebServer(void)
{
	DoShutdown(TRUE);
}

//
// Shut down the web server.  Can be called from several locations in
// different threads, so it uses a critical section to protect against
// simultaneous calls.
//
static void DoShutdown(BOOL fSignalServerThread)
{
	EnterCriticalSection(&csCleanup);
	if (fRunning)
	{
		// tell HTTP thread to exit, if requested
		if (fSignalServerThread)
		    SetEvent(hTerminateEvent);

		// 2 * timeout, -> millisecs
		WaitForSingleObject(hKilledEvent, HTTP_TIMEOUT * 2000);

		CloseHandle(hTerminateEvent);
		CloseHandle(hKilledEvent);

		WSACleanup();
		fRunning = FALSE;
	}
	LeaveCriticalSection(&csCleanup);
}

//
// StartThread - Start up a thread.
//
// Returns: TRUE if success
//	    FALSE if error
//
static BOOL StartThread(void (*pFunc)(void *), void *arglist)
{
	if ((int) _beginthread(pFunc, STACK_SIZE, arglist) != -1)
		return TRUE;
	return FALSE;
}

//
// HTTP thread -- sits in an infinite loop responding to HTTP queries.
// It puts the socket in polling mode rather than blocking mode so it can
// also check for an event signaled by the main thread telling it to exit.
//
static void HttpThread(short Port)
{
	u_long	fPolling;		// 1 ==> polling, 0 ==> blocking
	SOCKET	hListenSock;		// listen socket handle
	SERVENT	*pServEnt;		// server database entry for HTTP
	short	LocalPort;		// local TCP port for HTTP
	SOCKADDR_IN LocalAddr;		// local TCP port and IP addr
	SOCKADDR_IN RemoteAddr;		// remote client TCP port and IP addr
	SOCKET	hSock;			// socket for remote connection
	HTTP_CONNECT *pConnect;		// info on remote connection
	int	RemoteLen;		// sizeof RemoteAddr
	int	ErrVal;			// error number
	BOOL	fErrExit = TRUE;	// if T, shut down when exiting

	// set our name
//DS	EtsSetDebugName(GetCurrentThread(), "HTTP Server Thread");

//
// create a TCP socket, put it in non-blocking mode, bind it to the
// HTTP port, and set up to listen for incoming connections
//
	hListenSock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (hListenSock == INVALID_SOCKET)
	{
		SendError(TRUE, NULL, "Error %d creating socket() to listen "
				"for connections", WSAGetLastError());
		goto EXIT;
	}
	fPolling = 1;
	if (ioctlsocket(hListenSock, FIONBIO, &fPolling) == SOCKET_ERROR)
	{
		SendError(TRUE, NULL, "Error %d from ioctlsocket() to make "
				"listening socket non-blocking",
				WSAGetLastError());
		goto ERR;
	}
	if (Port != 0)
		LocalPort = htons(Port);
	else
	{
		pServEnt = getservbyname("http", "tcp");
		if (pServEnt == NULL)
			LocalPort = htons(80);	// 80 is well-known port
		else
			LocalPort = pServEnt->s_port;
	}
	LocalAddr.sin_family = AF_INET;
	LocalAddr.sin_port = LocalPort;
	LocalAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	if (bind(hListenSock, (PSOCKADDR) &LocalAddr, sizeof(LocalAddr)) ==
							SOCKET_ERROR)
	{
		SendError(TRUE, NULL, "Error %d bind()ing to HTTP port %d",
					WSAGetLastError(), ntohs(LocalPort));
		goto ERR;
	}
	if (listen(hListenSock, SOMAXCONN) == SOCKET_ERROR)
	{
		SendError(TRUE, NULL, "Error %d listen()ing on HTTP port",
							WSAGetLastError());
		goto ERR;
	}

//
// Just sit in a loop servicing HTTP queries, until the main thread
// signals us to terminate.  Because the socket is in non-blocking mode
// we can just poll for remote connections inside the loop.
//
	while (TRUE)
	{
//
// Check if the main thread told us to terminate
//
		if (WaitForSingleObject(hTerminateEvent, 0) == WAIT_OBJECT_0)
		{
		    SetEvent(hKilledEvent);
		    break;
		}

//
// Poll for a connection from a remote HTTP client.  If there is one,
// create a thread to service the request.  We create a thread for each
// request to improve response times and minimize the chance we would
// overflow the pending connection queue in the stack and have connection
// requests refused.
//
		RemoteLen = sizeof(RemoteAddr);
		hSock = accept(hListenSock, (PSOCKADDR) &RemoteAddr,
			       &RemoteLen);
		if (hSock == INVALID_SOCKET)
		{
		    ErrVal = WSAGetLastError();
		    if (ErrVal != WSAEWOULDBLOCK)
			SendError(FALSE, NULL, "Error %d accept()ing "
				  "connection", ErrVal);
		    // Relinquish time slice to any other runnable
		    // threads, so we don't hog the CPU just checking for
		    // connections.  Sleep(0) doesn't allow lower
		    // priority threads to run.
		    Sleep(10);
		    continue;
		}
		else
		{
		    pConnect = malloc(sizeof(HTTP_CONNECT));	// m1
		    if (! pConnect)
		    {
			SendError(FALSE, &RemoteAddr, "Can't malloc() "
				  "memory for thread connect struct");
			closesocket(hSock);
			continue;
		    }
		    pConnect->hSock = hSock;
		    memcpy(&pConnect->RemoteAddr, &RemoteAddr,
			   sizeof(RemoteAddr));
		    pConnect->pBuf = pConnect->Buf;
		    time(&pConnect->CliRequestTime);
		    if (! StartThread((void(*)(void *))ServiceRequestThread,
				      pConnect))
		    {
			SendError(FALSE, &RemoteAddr, "Unable to "
				  "create thread to service connection");
			closesocket(hSock);
			free(pConnect);		// f1
#if ! defined(__BORLANDC__)
			pConnect = NULL;
#endif
			continue;
		    }
		}
	}
	fErrExit = FALSE;

//
// cleanup and exit
//
ERR:
	closesocket(hListenSock);
EXIT:
	// this needs to be before DoShutdown(), as that checks the semaphore
	if (fErrExit)
	{
		SetEvent(hKilledEvent);
		// wait for threads to die, then cleanup before exit
		DoShutdown(FALSE);
	}
}


//
// Thread created to service each HTTP connection
//
static void ServiceRequestThread(HTTP_CONNECT *pConnect)
{
    REQUEST_LINE *pLine1;		// head of linked list of request lines
    REQUEST_LINE *pLine;		// current request line
    REQUEST_LINE *pNextLine;		// next request line
    char	*pURI;			// ptr to URI in request
    char	*pHttpVer;		// ptr to HTTP-version in request
    BOOL	fHttp10;		// T ==> HTTP 1.0 request, F ==> 0.9
    REQ_INPUTS ReqInp;			// request inputs
    REQ_OUTPUTS ReqOutp;		// request outputs
    BOOL	fSendResponse;		// T ==> send a response
    BOOL	fSentResponse;		// T ==> a response was sent
    time_t	ModSinceTime;		// if-modified-since time in request
    int	ModSinceLen;			// len in if-mod-since line, 0 if none
    char	*pFieldData;		// pointer to header field data string
#define BUFLEN 512
    char	Buf[BUFLEN];		// send buffer
    char	*pIpAddr;		// IP addr of remote host
    SOCKADDR_IN LocalAddr;		// local IP addr
    int	AddrSize;			// size of LocalAddr
    enum {METHOD_GET, METHOD_HEAD, METHOD_POST} Method;

    pIpAddr = inet_ntoa(pConnect->RemoteAddr.sin_addr);
    if (pIpAddr == NULL)
	strcpy(Buf, "HTTP Servicing Request");
    else
	sprintf(Buf, "HTTP Connection to %s", pIpAddr);
//DS    EtsSetDebugName(GetCurrentThread(), Buf);

    //
    // init variables that control response
    //
    fHttp10 = FALSE;
    ModSinceTime = 0;
    ModSinceLen = 0;
    fSentResponse = FALSE;
    ReqInp.pFrom = NULL;
    ReqInp.ContentLength = 0;
    ReqInp.pEntityBody = NULL;
    ReqInp.pContentType = NULL;
    ReqInp.pContentEncoding = NULL;
    ReqInp.fHeadMethod = FALSE;
    ReqOutp.LastModTime = 0;
    ReqOutp.ContentLength = 0;
    ReqOutp.pEntityBody = NULL;
    ReqOutp.Expires = 0;
    ReqOutp.ContentType[0] = EOS;
    ReqOutp.ContentEncoding[0] = EOS;
    ReqOutp.StatusCode[0] = EOS;
    ReqOutp.HeaderFields[0] = EOS;

    //
    // Read the first request line and parse it.  We support the following methods:
    //	GET - retrieve data from URI
    //	HEAD - get info about URI
    //	POST - accept subordinate of URI
    //
    pLine1 = ReadRequestLine(pConnect, &ReqOutp, &fSendResponse);
    if (pLine1 == NULL)
    {
	if (fSendResponse)
	    goto RESPOND;
	goto DONE;
    }
    if (strncmp(pLine1->pLineText, "GET ", 4) == 0)
    {
	Method = METHOD_GET;
	pURI = pLine1->pLineText + 4;
    }
    else if (strncmp(pLine1->pLineText, "HEAD ", 5) == 0)
    {
	Method = METHOD_HEAD;
	ReqInp.fHeadMethod = TRUE;
	pURI = pLine1->pLineText + 5;
    }
    else if (strncmp(pLine1->pLineText, "POST ", 5) == 0)
    {
	Method = METHOD_POST;
	pURI = pLine1->pLineText + 5;
    }
    else
    {
	strcpy(ReqOutp.StatusCode, pStat501);
	HtmlErrorPage(&ReqOutp, "Unrecognized method: %s",
		      pLine1->pLineText);
	goto RESPOND;
    }

    pHttpVer = strstr(pLine1->pLineText, "HTTP/");
    if (pHttpVer != NULL)
	fHttp10 = *(pHttpVer + 5) > '0';

    //
    // Read the rest of the header lines in the request, if HTTP 1.0 client.
    // As we read each line, check for header fields we process.
    //
    if (fHttp10)
    {
	for (pLine = pLine1; TRUE; pLine = pLine->pNext)
	{
	    // Read in the next line, and check if end of
	    // header lines (marked by an empty line)
	    pLine->pNext = ReadRequestLine(pConnect, &ReqOutp,
					   &fSendResponse);
	    if (pLine->pNext == NULL)
	    {
		if (fSendResponse)
		    goto RESPOND;
		goto DONE;
	    }
	    if (*pLine->pNext->pLineText == EOS)
	    {
		// if (pLine->pNext handled above
		free(pLine->pNext);	// f2
		pLine->pNext = NULL;
		break;
	    }

	    // Check for header fields we process
	    pFieldData = GetHeaderField(pLine->pNext, "If-Modified-Since");
	    if (pFieldData)
	    {
		// ReadHttpTime returns 0 if error
		ModSinceTime = ReadHttpTime(pFieldData);
		ModSinceLen = 0;
		pFieldData = strchr(pFieldData, ';');
		if (pFieldData != NULL)
		{
		    ++ pFieldData;
		    pFieldData += strspn(pFieldData," \t");
		    if (_strnicmp(pFieldData, "length=", 7) == 0)
		    {
			pFieldData += 7;
			// atoi returns 0 if error
			ModSinceLen = atoi(pFieldData);
		    }
		}
	    }
	    else if ((pFieldData = GetHeaderField(pLine->pNext,
		"Content-Encoding")) != NULL)
	    {
		ReqInp.pContentEncoding = pFieldData;
	    }
	    else if ((pFieldData = GetHeaderField(pLine->pNext,
		"Content-Type")) != NULL)
	    {
		ReqInp.pContentType = pFieldData;
	    }
	    else if ((pFieldData = GetHeaderField(pLine->pNext,
		"Content-Length")) != NULL)
	    {
		// atoi returns 0 if error
		ReqInp.ContentLength = atoi(pFieldData);
	    }
	    else if ((pFieldData = GetHeaderField(pLine->pNext, "From"))
		     != NULL)
	    {
		ReqInp.pFrom = pFieldData;
	    }
	}
    }

    //
    // If POST method, read the entity-body, if any, sent by the client
    //
    if (Method == METHOD_POST && ReqInp.ContentLength > 0)
    {
	ReqInp.pEntityBody = malloc(ReqInp.ContentLength);	// m3
	if (! ReqInp.pEntityBody)
	{
	    SendError(FALSE, &pConnect->RemoteAddr, "Can't malloc "
		      "memory for POSTed entity-body");
	    strcpy(ReqOutp.StatusCode, pStat500);
	    HtmlErrorPage(&ReqOutp, "Server error, unable to "
			  "allocate memory");
	    goto RESPOND;
	}
	if (! ReadEntity(pConnect, &ReqInp, &ReqOutp, &fSendResponse))
	{
	    if (fSendResponse)
		goto RESPOND;
	    goto DONE;
	}
    }

    //
    // Parse the URI into its pieces and process the method
    // Note we do NOT serialize calls to PostURI() and GetURI() (in the interests
    // of server responsiveness), so they must be written to be reentrant.
    //
    if (! ParseURI(pConnect, pURI, &ReqInp, &ReqOutp))
	goto RESPOND;
    AddrSize = sizeof(LocalAddr);
    if (getsockname(pConnect->hSock, (LPSOCKADDR) &LocalAddr, &AddrSize) ==
								SOCKET_ERROR)
	ReqInp.LocalIpAddr.s_addr = htonl(INADDR_NONE);
    else
	ReqInp.LocalIpAddr = LocalAddr.sin_addr;

    if (Method == METHOD_POST)
    {
/*DS	short RetVal;

	//
	// Find out what type of POST we are dealing with.
	//
	if (ReqInp.pContentType && strnicmp(ReqInp.pContentType,
					    FORM_URL_ENCODED,
					    strlen(FORM_URL_ENCODED)) == 0)
	{
	    //
	    // Rebuild pInp to hold a standard POST.
	    //
	    RetVal = GetURLEncodedPostInfo(&ReqInp);
	}
	else if (ReqInp.pContentType && strnicmp(ReqInp.pContentType,
					    MULTIPART_FORM_DATA,
					    strlen(MULTIPART_FORM_DATA)) == 0 )
	{
	    //
	    // Rebuild pInp to hold a multipart POST.
	    //
	    RetVal = GetMultiPartPostInfo(&ReqInp);
	}
	else
	{
	    // It is possible for a POST to have no content-type.
	    // If this is the case, we will not process it.  We will
	    // let the application parse the fields of ReqInp.
	}

	//
	// If there was a problem handling the POST, we should call
	// SendError() to let the application know there was a problem.  We
	// should also build an HTML page describing the error,
	//
	if (RetVal != FORM_NO_ERR)
	{
	    HandlePostError(RetVal, &pConnect->RemoteAddr, &ReqInp, &ReqOutp);
	    goto RESPOND;
	}*/
	PostURI(&ReqInp, &ReqOutp);
    }
    else
	GetURI(&ReqInp, &ReqOutp);

    if (ReqInp.pURI)
    {
	free(ReqInp.pURI);
	ReqInp.pURI = NULL;
    }
    if (ReqInp.pPath)
    {
	free(ReqInp.pPath);
	ReqInp.pPath = NULL;
    }
    if (ReqInp.pParams)
    {
	free(ReqInp.pParams);
	ReqInp.pParams = NULL;
    }
    if (ReqInp.pQuery)
    {
	free(ReqInp.pQuery);
	ReqInp.pQuery = NULL;
    }
    if (ReqInp.pEntityBody)
    {
	free(ReqInp.pEntityBody);
	ReqInp.pEntityBody = NULL;
    }
    //
    // Method was processed successfully.  Now decide whether we actually include
    // the entity-body (if any) in our response, based on conditions which vary
    // for each method.
    //
    if (ReqOutp.LastModTime > time(NULL))
	// future times for current content are not permitted.
	time(&ReqOutp.LastModTime);
    if (ReqOutp.StatusCode[0] == EOS)
	strcpy(ReqOutp.StatusCode, pStat200);
    switch(Method)
    {
	case METHOD_HEAD:
	    // Send all the info about the entity, but not the entity-body
	    if (ReqOutp.pEntityBody)
	    {
		free(ReqOutp.pEntityBody);
		ReqOutp.pEntityBody = NULL;
	    }
	    break;
	case METHOD_GET:
	    // Don't send the entity-body, or any info about it, if
	    // the last-modified date is earlier than the if-modified-since
	    // date (which is set to 0 if none) AND the length, if
	    // specified (0 if unspecified), is unchanged.
/*DS	    if ((ReqOutp.LastModTime <= ModSinceTime) &&
		ReqOutp.pEntityBody &&
		(ModSinceLen == 0 || ModSinceLen == ReqOutp.ContentLength))
	    {
		free(ReqOutp.pEntityBody);
		ReqOutp.pEntityBody = NULL;
		ReqOutp.ContentLength = 0;
		ReqOutp.ContentType[0] = EOS;
		ReqOutp.ContentEncoding[0] = EOS;
		strcpy(ReqOutp.StatusCode, pStat304);
	    }*/
	    break;
	case METHOD_POST:
	    // although POST is mostly used to send data to the server,
	    // it may also return an entity-body (preferably of type
	    // "text/html") describing how to access a resource that
	    // has been created as a result of the POST.  If there
	    // is such an entity-body, we always send it.
	    break;
    }

RESPOND:
    //
    // If at least HTTP/1.0 client, send the status line and headers.
    //
    time(&pConnect->ResponseTime);
    if (fHttp10)
    {
	// The header lines are batched and sent all at once, for
	// improved performance.

	// send Status-Line
	sprintf(Buf, "HTTP/1.0 %s\r\n", ReqOutp.StatusCode);
	// send Date general-header
	strcat(Buf, "Date: ");
	MakeHttpTime(time(NULL), &Buf[strlen(Buf)], 1);
	strcat(Buf, "\r\n");
	// send MIME-Version general-header
	strcat(Buf, "MIME-Version: 1.0\r\n");
	// send Server response-header
	strcat(Buf, "Server: MicroWeb/1.0\r\n");
	// send Content-Type entity-header, if any
	if (ReqOutp.ContentType[0] != EOS)
	{
	    if ((strlen(Buf)+strlen(ReqOutp.ContentType)+17) > BUFLEN)
	    {
		if (!SendResponse(pConnect, Buf, 0))
		    goto DONE;
		Buf[0] = EOS;
	    }
	    strcat(Buf, "Content-Type: ");
	    strcat(Buf, ReqOutp.ContentType);
	    strcat(Buf, "\r\n");
	}
	// send Content-Encoding entity-header, if any
	if (ReqOutp.ContentEncoding[0] != EOS)
	{
	    if ((strlen(Buf)+strlen(ReqOutp.ContentEncoding)+21) > BUFLEN)
	    {
		if (!SendResponse(pConnect, Buf, 0))
		    goto DONE;
		Buf[0] = EOS;
	    }
	    strcat(Buf, "Content-Encoding: ");
	    strcat(Buf, ReqOutp.ContentEncoding);
	    strcat(Buf, "\r\n");
	}
	// send Content-Length entity-header
	if ((strlen(Buf)+26) > BUFLEN) // generously allow 7 digits for length
	{
	    if (!SendResponse(pConnect, Buf, 0))
		goto DONE;
	    Buf[0] = EOS;
	}
	strcat(Buf, "Content-Length: ");
	sprintf(&Buf[strlen(Buf)], "%d\r\n", ReqOutp.ContentLength);
	// send Last-Modified entity-header, if any
	if (ReqOutp.ContentLength > 0 && ReqOutp.LastModTime != 0)
	{
	    if ((strlen(Buf)+29+18) > BUFLEN) // 29 is MakeHttpTime string
	    {
		if (! SendResponse(pConnect, Buf, 0))
		    goto DONE;
		Buf[0] = EOS;
	    }
	    strcat(Buf, "Last-Modified: ");
	    MakeHttpTime(ReqOutp.LastModTime, &Buf[strlen(Buf)], 1);
	    strcat(Buf, "\r\n");
	}
	// send Expires entity-header, if any
	if (ReqOutp.ContentLength > 0 && ReqOutp.Expires != 0)
	{
	    if ((strlen(Buf)+29+12) > BUFLEN) // 29 is MakeHttpTime string
	    {
		if (! SendResponse(pConnect, Buf, 0))
		    goto DONE;
		Buf[0] = EOS;
	    }
	    strcat(Buf, "Expires: ");
	    MakeHttpTime(ReqOutp.Expires, &Buf[strlen(Buf)], 1);
	    strcat(Buf, "\r\n");
	}
	// send any additional header fields
	if (ReqOutp.HeaderFields[0] != EOS)
	{
	    if ((strlen(Buf)+strlen(ReqOutp.HeaderFields)) > BUFLEN)
	    {
		if (!SendResponse(pConnect, Buf, 0))
		    goto DONE;
		Buf[0] = EOS;
	    }
	    strcat(Buf, ReqOutp.HeaderFields);
	}
	// send CRLF to terminate headers, and finally really send!
	strcat(Buf, "\r\n");
	if (! SendResponse(pConnect, Buf, 0))
	    goto DONE;
    }
    else if (ReqOutp.pEntityBody == NULL)
    {
	// HTTP 0.9 and no response to send.  Send status line as
	// response, which will just be interpreted as HTML by the
	// HTTP 0.9 client
	sprintf(Buf, "<HTML>\n<BODY>\n%s\n</BODY>\n</HTML>\n",
		ReqOutp.StatusCode);
	if (! SendResponse(pConnect, Buf, 0))
	    goto DONE;
    }

    //
    // Send the entity body (if any).
    //
    if (ReqOutp.pEntityBody != NULL)
    {
	if (!SendResponse(pConnect, ReqOutp.pEntityBody,
			  ReqOutp.ContentLength))
	    goto DONE;
    }
    fSentResponse = TRUE;

DONE:
    // since we don't support linger (yet), we have to suck any extra
    // data off of the recv() socket before we close it.  This _is_
    // required, or we send may send a reset to the remote host, as
    // POSTs from browsers tend to have extra CRLFs after the entity
    // body
    slurpExtraBytes(pConnect);

    //
    // Close the connection
    //
    closesocket(pConnect->hSock);

    //
    // Log the request if it was successfully received and replied to.  We
    // serialize calls to this function to make the application easier to write,
    // since now that the response has been sent it won't impact performance
    // to wait here.
    //
    if (fSentResponse)
    {
	EnterCriticalSection(&csLog);
	SendRequestLog(pConnect, pLine1, ReqOutp.StatusCode);
	LeaveCriticalSection(&csLog);
    }

    //
    // Then free allocated memory and exit.
    //
    if (ReqOutp.pEntityBody)
    {
	free(ReqOutp.pEntityBody);
	ReqOutp.pEntityBody = NULL;
    }

    for (pLine = pLine1; pLine != NULL; pLine = pNextLine)
    {
	pNextLine = pLine->pNext;
	free(pLine);
    }

    free(pConnect);
}


//
// Returns ptr to malloc'd REQUEST_LINE struct containing the next line
// (terminated by <LF> or <CRLF>, which are stripped) of the request.
//
// If success:
//	ptr to malloc'd REQUEST_LINE struct returned
//
// If error:
//	NULL returned
//	*pfSendResponse == T if a response should be sent to the client
//			== F if the connection should be closed
//
static REQUEST_LINE *ReadRequestLine(HTTP_CONNECT *pConn, REQ_OUTPUTS *pOutp,
				     BOOL *pfSendResponse)
{
    char	*pScan;		// scan ptr within receive buffer
    int		Len;		// size of data in bytes
    REQUEST_LINE *pLine;	// malloc'd line struct
    int		nRecv;		// number of bytes received from client
    int		ErrVal;

    pLine = NULL;
    //
    // Loop, checking buffer and reading chars until
    //   (1) we get a complete line,
    //   (2) our buffer overflows, or
    //   (3) we timeout.
    //
    // Remember buffer may already have data in it (including one or more
    // complete lines) when this routine is called.
    //
    for (pScan = pConn->Buf; TRUE; )
    {
	// scan buffer for terminating <LF> or <CRLF>
	for ( ; pScan < pConn->pBuf; ++ pScan)
	{
	    if (*pScan == '\r' || *pScan == '\n')
	    {
		// We have a complete line in the buffer
		// Copy line to struct
		Len = pScan - pConn->Buf;
		pLine = malloc(sizeof(REQUEST_LINE) + Len + 1);	// m2
		if (pLine == NULL)
		{
		    SendError(FALSE, &pConn->RemoteAddr,
			      "Can't malloc() memory for REQUEST_LINE struct");
		    HtmlErrorPage(pOutp, "Server error, unable to allocate "
				  "memory");
		    goto ERR_INTERNAL;
		}
		pLine->pNext = NULL;
		pLine->pLineText = (char *) (pLine + 1);

		if (Len > 0)
		    memcpy(pLine->pLineText, pConn->Buf, Len);

		*(pLine->pLineText + Len) = EOS;

		// Now remove the copied line from the buffer
		if (*pScan == '\r')
		{
		    // At <CR>, consume <LF> also if there
		    if (pScan < pConn->pBuf - 1)
		    {
			if (*(pScan + 1) == '\n')
			    ++pScan;
		    }
		    else
		    {
			// end of buf.  Try to read 1 char, keep it unless
			// it's a <LF>.  Ignore errors incl. closed connection.
			nRecv = recv(pConn->hSock, pScan, 1, 0);
			if (nRecv == 1 && *pScan != '\n')
			{
			    --pScan;
			}
		    }
		}
		++pScan;
		Len = pConn->pBuf - pScan;
		if (Len > 0)
		    memcpy(pConn->Buf, pScan, Len);
		pConn->pBuf = pConn->Buf + Len;
		return pLine;
	    }
	}

	// No complete line in buffer.  Check for buffer overflow.
	if (pConn->pBuf - pConn->Buf >= sizeof(pConn->Buf))
	{
	    SendError(FALSE, &pConn->RemoteAddr, "HTTP request "
		      "overflowed receive buffer");
	    HtmlErrorPage(pOutp, "HTTP request overflowed "
			  "server's receive buffer");
	    goto ERR_INTERNAL;
	}

	// Check for timeout before complete request received
	if (time(NULL) - pConn->CliRequestTime > HTTP_TIMEOUT)
	{
	    SendError(FALSE, &pConn->RemoteAddr, "Timed out "
		      "before complete HTTP request received");
	    HtmlErrorPage(pOutp, "Server timed out waiting for "
			  "complete request");
	    goto ERR_TIMEOUT;
	}

	// Read as much data as we can into buffer
	nRecv = recv(pConn->hSock, pConn->pBuf,
		     sizeof(pConn->Buf) - (pConn->pBuf - pConn->Buf), 0);
	if (nRecv == SOCKET_ERROR)
	{
	    ErrVal = WSAGetLastError();
	    if (ErrVal == WSAEWOULDBLOCK)
		nRecv = 0;
	    else
	    {
		SendError(FALSE, &pConn->RemoteAddr, "Error "
			  "%d recv()ing HTTP request", ErrVal);
		goto ERR_NO_REPLY;
	    }
	}
	else if (nRecv == 0)
	{
	    SendError(FALSE, &pConn->RemoteAddr, "HTTP client "
		      "closed connection before sending complete "
		      "request");
	    goto ERR_NO_REPLY;
	}
	pConn->pBuf += nRecv;
    }

ERR_NO_REPLY:
    if (pLine)
	free(pLine);		// f2
    *pfSendResponse = FALSE;
    return NULL;

ERR_TIMEOUT:
    if (pLine)
	free(pLine);		// f2
    // return bad request status code (they took too long to send it)
    *pfSendResponse = TRUE;
    strcpy(pOutp->StatusCode, pStat400);
    return NULL;

ERR_INTERNAL:
    if (pLine)
	free(pLine);		// f2
    *pfSendResponse = TRUE;
    strcpy(pOutp->StatusCode, pStat500);
    return NULL;
}

//
// Returns a pointer to the field data if the specified field name matches,
// NULL otherwise.
//
static char *GetHeaderField(REQUEST_LINE *pLine, char *pFieldName)
{
    int	Len;
    char	*pFieldData;

    Len = strlen(pFieldName);
    if (_strnicmp(pLine->pLineText, pFieldName, Len) != 0)
	return NULL;

    pFieldData = pLine->pLineText + Len;
    pFieldData += strspn(pFieldData, " \t");

    if (*pFieldData != ':')
	return NULL;

    ++pFieldData;
    pFieldData += strspn(pFieldData, " \t");

    if (pFieldData == EOS)
	return NULL;

    return pFieldData;
}

//
// Returns a POSTed entity-body into already malloc'd input buffer.
//
// Returns:  TRUE	if success
//	     FALSE	if error
//
// If success:
//	TRU returned
//
// If error:
//	FALSE returned
//	*pfSendResponse == T if a response should be sent to the client
//			== F if the connection should be closed
//
static BOOL ReadEntity(HTTP_CONNECT *pConn, REQ_INPUTS *pInp,
		       REQ_OUTPUTS *pOutp, BOOL *pfSendResponse)
{
	char	*pBuf;			// scan ptr thru buffer
	int	nRead;			// number of bytes read so far
	int	nLeft;			// number of bytes left in rcv buf
	int	nRecv;			// number of bytes received from client
	int	ErrVal;

//
// Read any data currently in our receive buffer
//
	if (pInp->ContentLength <= 0)
		return TRUE;
	nRead = pConn->pBuf - pConn->Buf;
	if (nRead > 0)
	{
		if (nRead > pInp->ContentLength)
		{
			nLeft = pInp->ContentLength - nRead;
			nRead = pInp->ContentLength;
		}
		else
			nLeft = 0;
		memcpy(pInp->pEntityBody, pConn->Buf, nRead);
		pConn->pBuf = pConn->Buf + nLeft;
		if (nLeft > 0)
		{
			memcpy(pConn->Buf, pConn->Buf + nRead, nLeft);
			return TRUE;
		}
	}

//
// Now loop, reading data from the client until we get our total or
// we timeout.
//
	for (pBuf = pInp->pEntityBody + nRead; nRead < pInp->ContentLength;
						nRead += nRecv, pBuf += nRecv)
	{
		// Check for timeout before complete request received
		if (time(NULL) - pConn->CliRequestTime > HTTP_TIMEOUT)
		{
			SendError(FALSE, &pConn->RemoteAddr, "Timed out "
				"before complete POSTed entity-body received");
			HtmlErrorPage(pOutp, "Server timed out waiting for "
					"POSTed entity-body");
			goto ERR_TIMEOUT;
		}

		// Read as much data as we can into buffer
		nRecv = recv(pConn->hSock, pBuf, pInp->ContentLength-nRead, 0);
		if (nRecv == SOCKET_ERROR)
		{
			ErrVal = WSAGetLastError();
			if (ErrVal == WSAEWOULDBLOCK)
				nRecv = 0;
			else
			{
				SendError(FALSE, &pConn->RemoteAddr, "Error "
					"%d recv()ing POSTed entity-body",
					ErrVal);
				goto ERR_NO_REPLY;
			}
		}
		else if (nRecv == 0)
		{
			SendError(FALSE, &pConn->RemoteAddr, "HTTP client "
				"closed connection before sending complete "
				"POSTed entity-body");
			goto ERR_NO_REPLY;
		}
	}
	return TRUE;

ERR_NO_REPLY:
	*pfSendResponse = FALSE;
	return FALSE;
ERR_TIMEOUT:
	// Timed out waiting for complete request, return bad request err.
	*pfSendResponse = TRUE;
	strcpy(pOutp->StatusCode, pStat400);
	return FALSE;
}

//
// Sends data to the client.  If Len is 0, uses strlen(pBuf) to determine
// how much to send.
//
// Returns: TRUE if success
//	    FALSE if error
//
static BOOL SendResponse(HTTP_CONNECT *pConnect, char *pBuf, int Len)
{
	char	*pData;			// ptr to current data to send
	int	nSend;			// number of bytes sent to client
	fd_set	SendSet;		// hSock, used in in select() call
	struct timeval BlockTime;	// max time to block
	int	ErrVal;

//
// Send the response string to the HTTP client.  Because this is a
// non-blocking socket, we may have to call send() multiple times to
// transmit all the data.  We use select() to wait until more data can
// be accepted, to avoid calling send() repeatedly and consuming CPU cycles.
//
	if (Len == 0)
		Len = strlen(pBuf);
	FD_ZERO(&SendSet);
	FD_SET(pConnect->hSock, &SendSet);
	BlockTime.tv_usec = 0;
	for (pData = pBuf; pData - pBuf < Len; pData += nSend)
	{
		BlockTime.tv_sec = HTTP_TIMEOUT;
		ErrVal = select(0, NULL, &SendSet, NULL, &BlockTime);
		if (ErrVal == 0)
		{
			SendError(FALSE, &pConnect->RemoteAddr, "Timeout "
				"before complete HTTP response sent");
			return FALSE;
		}
		else if (ErrVal == SOCKET_ERROR)
		{
			SendError(FALSE, &pConnect->RemoteAddr, "Error %d "
				"from select() waiting to send response",
							WSAGetLastError());
			return FALSE;
		}
		nSend = send(pConnect->hSock, pData, Len - (pData-pBuf), 0);
		if (nSend == SOCKET_ERROR)
		{
			ErrVal = WSAGetLastError();
			if (ErrVal == WSAEWOULDBLOCK)
				nSend = 0;
			else
			{
				SendError(FALSE, &pConnect->RemoteAddr,
					"Error %d send()ing HTTP response",
								ErrVal);
				return FALSE;
			}
		}
	}
	return TRUE;
}

//
// Constructs a 0-terminated HTTP-format date string in Buf, which must
// be at least 30 bytes in size.  Time is in units of seconds since
// 00:00:00 GMT 1-Jan-1970.
//
static void MakeHttpTime(time_t Time, char *pBuf, int gmt)
{
	struct tm *pTime;
	static char *pDay[7] = {"Sun","Mon","Tue","Wed","Thu","Fri","Sat"};
	static char *pMonth[12] = {"Jan","Feb","Mar","Apr","May","Jun","Jul",
				    "Aug","Sep","Oct","Nov","Dec"};

	if (gmt)
	    pTime = gmtime(&Time);
	else
	    pTime = localtime(&Time);

	if (pTime == NULL)
	{
		strcpy(pBuf, "Thu, 01 Jan 1970 00:00:01 GMT");
		return;
	}
	if (pTime->tm_wday > 6)
		pTime->tm_wday = 0;
	if (pTime->tm_mday > 31 || pTime->tm_mday < 1)
		pTime->tm_mday = 1;
	if (pTime->tm_mon > 11)
		pTime->tm_mon = 0;
	if (pTime->tm_year > 9999 - 1900)
		pTime->tm_year = 0;
	if (pTime->tm_hour > 23)
		pTime->tm_hour = 0;
	if (pTime->tm_min > 59)
		pTime->tm_min = 0;
	if (pTime->tm_sec > 59)
		pTime->tm_sec = 0;
	sprintf(pBuf,"%s, %02d %s %4d %02d:%02d:%02d %s",pDay[pTime->tm_wday],
		pTime->tm_mday, pMonth[pTime->tm_mon], pTime->tm_year + 1900,
		pTime->tm_hour, pTime->tm_min, pTime->tm_sec,
		gmt ? "GMT" : "Local Time");
}

//
// Converts an HTTP date string into seconds since 00:00:00 GMT 1-Jan-1970.
// If an invalid time (before 1-Jan-1970, or after the current
// time, or unrecognized string format) the value 0 is returned.
//
typedef struct {
	char	*pStr;		// ptr to string
	int	Val;		// corresponding integer value
} DATE_CONV;

static time_t ReadHttpTime(char *pStr)
{
	struct tm tmTime;
	time_t	Time;
	char	*pComma;
	char	*pToken;
	char	Token[20];
	DATE_CONV *pDateStr;
	char *NextToken(char *pCurToken, char *pTokenBuf);
	DATE_CONV *FindDateStr(char *pStr, DATE_CONV *pDateConv);
	BOOL ConvTime(char *pStr, struct tm *pTime);
	static DATE_CONV Month[] = {{"Jan",0},{"Feb",1},{"Mar",2},{"Apr",3},
		{"May",4},{"Jun",5},{"Jul",6},{"Aug",7},{"Sep",8},{"Oct",9},
		{"Nov",10},{"Dec",11},{NULL,0}};

//
// Parse the string (which is always GMT) to construct a struct tm
// We don't need to calculate tm_wday or tm_yday, which are not needed as
// input to mktime().
//
// The HTTP spec permits three different formats for a date/time string.
//
	tmTime.tm_isdst = 0;
	tmTime.tm_wday = 0;
	tmTime.tm_yday = 0;
	if ((pToken = NextToken(pStr, Token)) == NULL)
		goto BAD_TIME;
	pComma = strchr(pStr, ',');
	if (pComma == NULL)
	{
		// asctime() format string
		// wkday SP mon SP [SP] d[d] SP hh:mm:ss SP yyyy
		if ((pDateStr = FindDateStr(Token, Month)) == NULL)
			goto BAD_TIME;
		tmTime.tm_mon = pDateStr->Val;
		if ((pToken = NextToken(pToken, Token)) == NULL)
			goto BAD_TIME;
		if (!isdig(Token[0]) || (Token[1] != EOS &&
				(!isdig(Token[1]) || Token[2] != EOS)))
			goto BAD_TIME;
		tmTime.tm_mday = atoi(Token);
		if ((pToken = NextToken(pToken, Token)) == NULL)
			goto BAD_TIME;
		if (!ConvTime(Token, &tmTime))
			goto BAD_TIME;
		if ((pToken = NextToken(pToken, Token)) == NULL)
			goto BAD_TIME;
		tmTime.tm_year = atoi(Token);
	}
	else if (pComma - pStr == 3)
	{
		// RFC 1123 format string
		// wkday, SP dd SP mon SP yyyy SP hh:mm:ss SP "GMT"
		if (!isdig(Token[0]) || !isdig(Token[1]) || Token[2]!=EOS)
			goto BAD_TIME;
		tmTime.tm_mday = atoi(Token);
		if ((pToken = NextToken(pToken, Token)) == NULL)
			goto BAD_TIME;
		if ((pDateStr = FindDateStr(Token, Month)) == NULL)
			goto BAD_TIME;
		tmTime.tm_mon = pDateStr->Val;
		if ((pToken = NextToken(pToken, Token)) == NULL)
			goto BAD_TIME;
		tmTime.tm_year = atoi(Token);
		if ((pToken = NextToken(pToken, Token)) == NULL)
			goto BAD_TIME;
		if (!ConvTime(Token, &tmTime))
			goto BAD_TIME;
		if ((pToken = NextToken(pToken, Token)) == NULL)
			goto BAD_TIME;
		if (memcmp(Token, "GMT", 3) != 0)
			goto BAD_TIME;
	}
	else
	{
		// RFC 850 format string
		// weekday, SP dd-mon-yy SP hh:mm:ss SP "GMT"
		if (!isdig(Token[0]) || !isdig(Token[1]) || Token[2]!='-')
			goto BAD_TIME;
		Token[2] = EOS;
		tmTime.tm_mday = atoi(Token);
		if (Token[6] != '-')
			goto BAD_TIME;
		Token[6] = EOS;
		if ((pDateStr = FindDateStr(&Token[3], Month)) == NULL)
			goto BAD_TIME;
		tmTime.tm_mon = pDateStr->Val;
		if (!isdig(Token[7]) || !isdig(Token[8]) || Token[9]!=EOS)
			goto BAD_TIME;
		tmTime.tm_year = atoi(&Token[7]) + 1900;
		if ((pToken = NextToken(pToken, Token)) == NULL)
			goto BAD_TIME;
		if (!ConvTime(Token, &tmTime))
			goto BAD_TIME;
		if ((pToken = NextToken(pToken, Token)) == NULL)
			goto BAD_TIME;
		if (memcmp(Token, "GMT", 3) != 0)
			goto BAD_TIME;
	}
	if (tmTime.tm_mday > 31 || tmTime.tm_year < 1970)
		goto BAD_TIME;
	tmTime.tm_year -= 1900;

//
// OK, we have the time in GMT.  mktime() will convert it into a time_t, but
// will do so assuming we are giving it the time in local time.  So we have
// to correct the result using the _timezone global in the C run-time library,
// which gives the difference in seconds between GMT and local time.
//
	Time = mktime(&tmTime);
	if (Time == -1)
		goto BAD_TIME;
	Time -= _timezone;
	if (Time > time(NULL))
		goto BAD_TIME;
	return Time;

BAD_TIME:
	return 0;
}

//
// Returns ptr to next space-delimited token, or NULL if none.  EOS-terminated
// token is copied into pTokenBuf, which is assumed at least 20 bytes in len.
//
static char *NextToken(char *pCurToken, char *pTokenBuf)
{
	char	*pToken;
	char	*pSpace;

	if ((pSpace = strchr(pCurToken, ' ')) == NULL)
		return NULL;
	pToken = pSpace + strspn(pSpace, " ");
	if (pToken == EOS)
		return NULL;
	pSpace = strchr(pToken, ' ');
	if (pSpace == NULL)
		pSpace = pToken + strlen(pToken);
	if (pSpace - pToken > 19)
		return NULL;
	memcpy(pTokenBuf, pToken, pSpace - pToken);
	*(pTokenBuf + (pSpace - pToken)) = EOS;
	return pToken;
}

//
// finds a matching DATE_CONV struct in an array of structs terminated
// by a struct with a NULL pStr.  Returns NULL if not found.
//
static DATE_CONV *FindDateStr(char *pStr, DATE_CONV *pDateConv)
{
	for ( ; pDateConv->pStr != NULL; ++ pDateConv)
		if (strcmp(pStr, pDateConv->pStr) == 0)
			return pDateConv;
	return NULL;
}

//
// Converts hh:mm:ss in ASCII to numbers in pTime.
// returns TRUE if success, FALSE if error
//
// NOTE:  *pStr is modified by this function
//
static BOOL ConvTime(char *pStr, struct tm *pTime)
{
	if (!isdig(*pStr) || !isdig(*(pStr+1)) || *(pStr+2) != ':')
		return FALSE;
	*(pStr+2) = EOS;
	pTime->tm_hour = atoi(pStr);
	pStr += 3;
	if (!isdig(*pStr) || !isdig(*(pStr+1)) || *(pStr+2) != ':')
		return FALSE;
	*(pStr+2) = EOS;
	pTime->tm_min = atoi(pStr);
	pStr += 3;
	if (!isdig(*pStr) || !isdig(*(pStr+1)) || *(pStr+2) != EOS)
		return FALSE;
	pTime->tm_sec = atoi(pStr);
	if (pTime->tm_hour > 23 || pTime->tm_min > 59 || pTime->tm_sec > 59)
		return FALSE;
	return TRUE;
}

//
// Parses a URI.  If success, sets pURI and pPath (required) and
// pParams and pQuery (optional, NULL if not present) in the REQ_INPUTS struct.
// All are in malloc()'d memory which must be freed by the caller.
//
// Returns:	TRUE if success
//		FALSE if error, StatusCode and pEntityBody set in pOutp
//
static BOOL ParseURI(HTTP_CONNECT *pConn, char *pURI, REQ_INPUTS *pInp,
		     REQ_OUTPUTS *pOutp)
{
	char	*pDest;		// ptr within pPath, pParams, or pQuery
	int	Len;
	enum {PIECE_PATH, PIECE_PARAMS, PIECE_QUERY} Piece;

//
// Alloc memory for the URI and path strings, copy the URI string over.
//
	pInp->pParams = pInp->pQuery = NULL;
	pURI += strspn(pURI, " \t");
	Len = strcspn(pURI, " \t");
	if (Len == 0)
	{
		SendError(FALSE, &pConn->RemoteAddr, "Empty URI in HTTP "
			"request");
		HtmlErrorPage(pOutp, "Empty URI in HTTP request");
		strcpy(pOutp->StatusCode, pStat400);
		return FALSE;
	}
	pInp->pURI = malloc(Len + 1);	// m4
	if (! pInp->pURI)
		goto NO_MEM;
	pInp->pPath = malloc(Len + 2);	// m5
	if (! pInp->pPath)
	{
		// if (pInp->pURI) handled above
		free(pInp->pURI);	// f4
		pInp->pURI = NULL;
NO_MEM:
		SendError(FALSE, &pConn->RemoteAddr, "Can't malloc() memory "
			"for parsing URI");
		HtmlErrorPage(pOutp, "Server error, can't allocate memory");
		strcpy(pOutp->StatusCode, pStat500);
		return FALSE;
	}
	memcpy(pInp->pURI, pURI, Len);
	*(pInp->pURI + Len) = EOS;

//
// Scan the URI to extract the DOS file system path, and if present the
// parameters (following a ';') and query (following a '?').  We always
// reduce escape sequences, but for the path part only we also turn
// '/' into '\'.
// See the HTTP 1.0 specification and RFC 1738 (Uniform Resource Locators)
// for documenation on URI syntax.
//
	// URI should ALWAYS start with '/', if not put one in
	pURI = pInp->pURI;
	pDest = pInp->pPath;
	if (*pURI != '/' && *pURI != '\\')
	    *pDest++ = '\\';
	for (Piece = PIECE_PATH; *pURI != EOS; ++ pDest, ++ pURI)
	{
	    switch (*pURI)
	    {
		case '/':
		    if (Piece == PIECE_PATH)
			// for path only, change '/' to DOS-compat '\'
			*pDest = '\\';
		    else
			*pDest = *pURI;
		    break;
		case ';':
		    if (Piece == PIECE_PATH)
		    {
			// switch from path to parameters
			*pDest = EOS;
			pInp->pParams = malloc(strlen(pURI));	// m6
			if (! pInp->pParams)
			{
			    SendError(FALSE, &pConn->RemoteAddr,
				      "Can't malloc() memory for "
				      "parameters in URI");
			    HtmlErrorPage(pOutp, "Server error, "
					  "can't allocate memory");
			    strcpy(pOutp->StatusCode, pStat500);
			    goto ERR;
			}
			pDest = pInp->pParams - 1;
			Piece = PIECE_PARAMS;
		    }
		    else
			*pDest = *pURI;
		    break;
		case '?':
		    if (Piece != PIECE_QUERY)
		    {
			// switch from path or params to query
			*pDest = EOS;
			pInp->pQuery = malloc(strlen(pURI));	// m7
			if (! pInp->pQuery)
			{
			    SendError(FALSE, &pConn->RemoteAddr,
				      "Can't malloc() memory for "
				      "query in URI");
			    HtmlErrorPage(pOutp, "Server error, "
					  "can't allocate memory");
			    strcpy(pOutp->StatusCode, pStat500);
			    goto ERR;
			}
			pDest = pInp->pQuery - 1;
			Piece = PIECE_QUERY;
		    }
		    else
			*pDest = *pURI;
		    break;
		case '%':
		    if (Piece == PIECE_QUERY)
			*pDest = *pURI;
		    else
		    {
			++pURI;
			if ((! isxdig(*pURI)) || (! isxdig(*(pURI + 1))))
			{
			    SendError(FALSE, &pConn->RemoteAddr, "Invalid "
				      "hex escape sequence in URI");
			    HtmlErrorPage(pOutp, "Invalid hex escape "
					  "sequence in URI");
			    strcpy(pOutp->StatusCode, pStat400);
			    goto ERR;
			}
			// put two hex characters (a nibble each) into one
			// byte
			*pDest = (HexToBin(*pURI++) << 4);
			*pDest |= HexToBin(*pURI);
		    }
		    break;
		default:
		    *pDest = *pURI;
		    break;
	    }
	}
	*pDest = EOS;

	return TRUE;

ERR:
	if (pInp->pURI)
	{
	    free(pInp->pURI);		// f4
	    pInp->pURI = NULL;
	}
	if (pInp->pPath)
	{
	    free(pInp->pPath);		// f5
	    pInp->pPath = NULL;
	}
	if (pInp->pParams)
	{
	    free(pInp->pParams);	// f6
	    pInp->pParams = NULL;
	}
	if (pInp->pQuery)
	{
	    free(pInp->pQuery);		// f7
	    pInp->pQuery = NULL;
	}
	return FALSE;
}

//
// Converts a hex digit to binary.  Returns 0 if bad input
//
static char HexToBin(char c)
{
	if (isdig(c))
		return c - '0';
	c = toupper(c);
	if (c >= 'A' && c <= 'F')
		return 10 + c - 'A';
	return 0;
}

//
// Notify application of request so it can keep a log if desired
//
#define RESP_INDENT	8	// # chars to indent when printing response
static void SendRequestLog(HTTP_CONNECT *pConnect, REQUEST_LINE *pLine1,
							char *pStatusCode)
{
	REQUEST_RECORD ReqRec;
	char	*pDotIP;
	HOSTENT	*pHostEnt;
	REQUEST_LINE *pLine;
	char	*pBuf;
	char	RespIndent[RESP_INDENT+1];

	strcpy(ReqRec.RequestIPAddr, "unknown");
	strcpy(ReqRec.RequestHostName, "unknown");
	pDotIP = inet_ntoa(pConnect->RemoteAddr.sin_addr);
	if (pDotIP != NULL)
		strcpy(ReqRec.RequestIPAddr, pDotIP);
	pHostEnt=gethostbyaddr((char *)&pConnect->RemoteAddr.sin_addr.s_addr,
						sizeof(u_long), AF_INET);
	if (pHostEnt != NULL && pHostEnt->h_name != NULL)
	{
		// just truncate name if too long
		strncpy(ReqRec.RequestHostName, pHostEnt->h_name,
					sizeof(ReqRec.RequestHostName));
		ReqRec.RequestHostName
			[sizeof(ReqRec.RequestHostName) - 1] = EOS;
	}
	MakeHttpTime(pConnect->CliRequestTime, ReqRec.RequestTime, 1);
	strncpy(ReqRec.ResponseStatus, pStatusCode,
						sizeof(ReqRec.ResponseStatus));
	ReqRec.ResponseStatus[sizeof(ReqRec.ResponseStatus) - 1] = EOS;
	for (ReqRec.RequestLen = 0, pLine = pLine1; pLine != NULL;
							pLine = pLine->pNext)
		ReqRec.RequestLen += strlen(pLine->pLineText) + RESP_INDENT+1;
	++ ReqRec.RequestLen;	// for EOS
	ReqRec.pRequestText = malloc(ReqRec.RequestLen);
	if (ReqRec.pRequestText == NULL)
		ReqRec.RequestLen = 0;
	else
	{
		memset(RespIndent, ' ', sizeof(RespIndent));
		RespIndent[RESP_INDENT] = EOS;
		for (pBuf = ReqRec.pRequestText, pLine = pLine1;
					pLine != NULL; pLine = pLine->pNext)
		{
			pBuf += sprintf(pBuf, "%s%s\n", RespIndent,
							pLine->pLineText);
		}
		*pBuf = EOS;
	}
	LogRequest(&ReqRec);
	if (ReqRec.pRequestText != NULL)
	{
	    free(ReqRec.pRequestText);
	    ReqRec.pRequestText = NULL;
	}
}

//
// Notify application when an error occurs
// The string should have no newlines in it, this routine will put in
// any necessary line breaks.
//
static void SendError(BOOL fFatal, SOCKADDR_IN *pRemoteAddr,
						const char *pFormat, ...)
{
	char	ErrBuf[1024];		// buffers error message
	char	*pIpAddr;		// IP addr of remote client
	int	CurIdx;			// current index in ErrBuf
	int	StIdx;			// start index of next word
	int	EndIdx;			// end index of next word
	int	Col;			// current output column
	BOOL	fNewl;			// T ==> just inserted newline
	va_list	ArgList;

	// error message preamble
	if (fFatal)
		strcpy(ErrBuf, "Fatal HTTP error");
	else
		strcpy(ErrBuf, "Non-fatal HTTP error");
	if (pRemoteAddr != NULL &&
			(pIpAddr = inet_ntoa(pRemoteAddr->sin_addr)) != NULL)
		sprintf(&ErrBuf[strlen(ErrBuf)], " while servicing request "
						"from %s:\n", pIpAddr);
	else
		strcat(ErrBuf, ":\n");


	// error message
	StIdx = strlen(ErrBuf);
	va_start(ArgList, pFormat);
	vsprintf(&ErrBuf[StIdx], pFormat, ArgList);
	va_end(ArgList);

	// do line breaks and append newline
	CurIdx = StIdx;
	fNewl = TRUE;
	while (TRUE)
	{
		// get start and end of next word in buffer, append newline
		// and exit loop if no more words.
		StIdx = CurIdx + strspn(&ErrBuf[CurIdx], " ");
		EndIdx = StIdx + strcspn(&ErrBuf[StIdx], " ");
		if (StIdx == EndIdx)
		{
			if (!fNewl)
			{
				ErrBuf[CurIdx] = '\n';
				ErrBuf[CurIdx + 1] = EOS;
			}
			break;
		}

		// if we just did a newline, do the indent at start of line
		if (fNewl)
		{
			Col = 3;	// indent spacing count
			memmove(&ErrBuf[CurIdx + Col], &ErrBuf[StIdx],
						strlen(&ErrBuf[StIdx]) + 1);
			memset(&ErrBuf[CurIdx], ' ', Col);
			CurIdx += Col;
			fNewl = FALSE;
			continue;
		}

		// If we don't need a line break, continue loop
		if (Col + EndIdx - CurIdx <= 80)
		{
			Col += EndIdx - CurIdx;
			CurIdx = EndIdx;
			continue;
		}

		// we need a line break
		memmove(&ErrBuf[CurIdx + 1], &ErrBuf[CurIdx],
						strlen(&ErrBuf[CurIdx]) + 1);
		ErrBuf[CurIdx++] = '\n';
		fNewl = TRUE;
	}

	// call application's error routine
	WebServerError(fFatal, ErrBuf);
}

//
// Construct an HTML error page to return to the client.
// For most HTTP errors, the HTTP spec recommends that a brief HTML entity-body
// describing the error be returned along with the error status.
//
static void HtmlErrorPage(REQ_OUTPUTS *pOutp, const char *pFormat, ...)
{
	char	ErrPage[1024];
	va_list	ArgList;

	// construct the error page in our buffer
	strcpy(ErrPage, "<HTML>\n<BODY>\n");
	va_start(ArgList, pFormat);
	vsprintf(&ErrPage[strlen(ErrPage)], pFormat, ArgList);
	va_end(ArgList);
	strcat(ErrPage, "\n</BODY>\n</HTML>\n");

	// return it in the output structure
	pOutp->ContentLength = strlen(ErrPage);
	pOutp->pEntityBody = malloc(pOutp->ContentLength);
	if (pOutp->pEntityBody == NULL)
	{
		pOutp->ContentLength = 0;
		return;
	}
	memcpy(pOutp->pEntityBody, ErrPage, pOutp->ContentLength);
	strcpy(pOutp->ContentType, "text/html");
	pOutp->LastModTime = time(NULL);
}

//
// HandlePostError() - This function takes care of errors that may occur while
//	processing a POST method.  This function builds an HTML error page to
//	notify the client of the error.  It calls the function SendError() to
//	notify the application of the error, and it frees all of the allocated
//	memory.
//
// Inputs:
//	RetVal		A short describing the error that occured.  The
//			following values are accepted.
//				Success			FORM_NO_ERR         (0)
//				Fail (syntax)		FORM_SYNTAX_ERR    (-1)
//				Fail (malloc)		FORM_ALLOC_ERR     (-2)
//				Fail (multiple files)	FORM_MULT_FILE_ERR (-3)
//	pRemoteAddr	IP related information about the HTTP client.
//	pInp		Stores information relating to the POST received.
//
// Output:
//	pOutp		Stores the HTML page that will be sent back to the
//			client.
//
static void HandlePostError(short RetVal, SOCKADDR_IN *pRemoteAddr,
			    REQ_INPUTS *pInp, REQ_OUTPUTS *pOutp)
{
    strcpy(pOutp->StatusCode, pStat500);

    switch (RetVal)
    {
	case FORM_SYNTAX_ERR:
	    SendError(FALSE, pRemoteAddr,
		      "Syntax error, unable to parse the form");
	    HtmlErrorPage(pOutp, "Server error, could not interpret the form");
	    break;

	case FORM_ALLOC_ERR:
	    SendError(FALSE, pRemoteAddr,
		      "Could not malloc memory while processing POST");
	    HtmlErrorPage(pOutp, "Server error, unable to allocate memory");
	    break;

	case FORM_MULT_FILE_ERR:
	    SendError(FALSE, pRemoteAddr,
		      "The form sent contains more than one file");
	    HtmlErrorPage(pOutp, "Server error, cannot process forms "
			  "containing more than one file");
	    break;

	default:
	    SendError(FALSE, pRemoteAddr,
		      "Undexpected error while processing the POST");
	    HtmlErrorPage(pOutp, "Server error, unexpected error occurred");
	    break;
    }

    // Free the memory that was allocated for pInp.
    if (pInp->pURI)
    {
	free(pInp->pURI);
	pInp->pURI = NULL;
    }
    if (pInp->pPath)
    {
	free(pInp->pPath);
	pInp->pPath = NULL;
    }
    if (pInp->pParams)
    {
	free(pInp->pParams);
	pInp->pParams = NULL;
    }
    if (pInp->pQuery)
    {
	free(pInp->pQuery);
	pInp->pQuery = NULL;
    }
    if (pInp->pEntityBody)
    {
	free(pInp->pEntityBody);
	pInp->pEntityBody = NULL;
    }
}

static void slurpExtraBytes(HTTP_CONNECT *pConn)
{
    int nRecv, total;

    // Read as much data as we can into buffer. toss it
    // if we don't slurp these bytes, we'll send a reset to the remote
    // host during closesocket(), due to unreceived bytes
    //
    // after 10k extra or so, just give up and let the reset happen
    //
    for (total = 0; total < 10000; total += nRecv)
    {
	nRecv = recv(pConn->hSock, pConn->Buf, sizeof(pConn->Buf), 0);
	if (nRecv == SOCKET_ERROR || nRecv == 0)
	{
	    // we could care less what error we get here<g>
	    break;
	}
	total += nRecv;
    }
}

#ifdef DBGMEM
int _http_malloc_cnt = 0;
int _http_free_cnt = 0;
int _http_mem_cnt = 0;
unsigned _http_malloc_amt = 0;
unsigned _http_free_amt = 0;
unsigned _http_mem_amt = 0;
unsigned _http_free_app = 0;

int	_http_mem_verbose = 1;

#undef malloc
#undef free

void _p_xprintf(UCHAR *strp, ...);
#define xprintf _p_xprintf

void *	__cdecl	_http_malloc(size_t len)
{
    unsigned *ptr;

    ptr = LocalAlloc(LMEM_ZEROINIT, len + 12);
    if (! ptr)
    {
	xprintf("_http_malloc: real malloc() failed for %d\n", len);
    }
    else
    {
	_http_malloc_cnt++;
	_http_malloc_amt += len;
	_http_mem_cnt++;
	_http_mem_amt += len;

	*ptr++ = 0xDEADBEBE;
	*ptr++ = _http_malloc_cnt;
	*ptr++ = len;
	if (_http_mem_verbose)
	{
	    xprintf("%08x: alloc: %08x (len %d), ret %08x\n", *(ptr - 2),
		    ptr - 3, len, ptr);
	}
    }
    return (ptr);
}

void __cdecl _http_free(void *ptr)
{
    unsigned len, which;
    unsigned *uptr = ptr;

    if (uptr)
    {
	len = *--uptr;
	which = *--uptr;

	if (*--uptr != 0xDEADBEBE)
	{
	    free(ptr);
	    _http_free_app++;
	    if (_http_mem_verbose)
	    {
		xprintf("_http_free: app malloc %08x\n", ptr);
	    }
	}
	else
	{
	    LocalFree(uptr);
	    if (_http_mem_verbose)
	    {
		xprintf("%08x: free: %08x (len %d) actual %08x\n", which,
			ptr, len, uptr);
	    }
	    _http_free_cnt++;
	    _http_free_amt += len;
	    _http_mem_cnt--;
	    _http_mem_amt -= len;
	}
    }
}
#endif

//
// don't put code here, put it above the #ifdef DBGMEM line, as the
// memory debugging code will not work right after this point, due to
// the #undefs above.
//
// You've been warned
//
