// RCS_Id: httpserv.h 10.1 1999/12/09 14:52:28 clark Exp $

// HTTPSERV.H -- Interface to HTTPSERV.LIB

#ifndef ETS_HTTPSERV_INCLUDED
#define ETS_HTTPSERV_INCLUDED

#ifdef __cplusplus
extern "C" {
#endif

// Make sure all data structures are packed for compatibility between
// compilers.
#pragma pack(4)

////////////////////////////////////////////////////////////////////////
// Functions exported by HTTPSERV.LIB
////////////////////////////////////////////////////////////////////////

//
// Starts the HTTP server
//	Port	0 ==> use well-known HTTP port (80), otherwise use
//		specified port (passed in host byte order)
// Returns:
//	TRUE	if success
//	FALSE	if error (error message passed to ServerError())
//
BOOL __cdecl StartWebServer(short Port);

//
// Stops operation of the HTTP server
//
void __cdecl StopWebServer(void);

////////////////////////////////////////////////////////////////////////
// Functions imported by HTTPSERV.LIB
//
// NOTE: These functions are called in the context of a thread created by
//	HTTPSERV.LIB to service an HTTP request.  THIS MEANS THEY WILL 
//	EXECUTE CONCURRENTLY WITH THE REST OF YOUR PROGRAM.  In particular,
//	if these functions access global data which is also used by
//	other parts of your program, you must gate access to the global
//	data with a critical section or some other synchronization object.
//
//	When these functions are called, there is at least 14K of space
//	available on the thread stack (with the exception of WebServerError(),
//	which has at least 12K of stack space).  If you need more stack,
//	you can rebuild HTTPSERV.LIB, changing the stack size parameter
//	passed to beginthread().
////////////////////////////////////////////////////////////////////////

// only if winsock.h is included
#ifdef _WINSOCKAPI_
//
// Input parameters to GetURI() and PostURI()
// All members are optional except pURI and pPath.
//
typedef struct {
	char	*pURI;		// original URI in the request
	char	*pPath;		// DOS-compatible file path (the URI up to any
				// ';' or '?' character, with escapes replaced 
				// and '/' changed to '\'
	char	*pParams;	// start of params, or NULL if none (anything
				// past an initial ';' and up to any '?', with
				// escapes replaced)
	char	*pQuery;	// the query, or NULL if none (anything past
				// the '?' character in the URI, with escapes
				// replaced)
	char	*pFrom;		// text from "From:" header field, or NULL
// all fields below this are used only by POST method (PostURI())
	int	ContentLength;	// length of Entity-Body (could be 0)
	char	*pEntityBody;	// Entity-Body, or NULL if ContentLength == 0
	char	*pContentType;	// text from "Content-Type:" field, or NULL
	char	*pContentEncoding; // text from "Content-Encoding:", or NULL
	IN_ADDR	LocalIpAddr;	// local IP address for this connection
// except the following which is used only by GetURI()
	BOOL	fHeadMethod;	// T ==> HEAD request.  The GetURI() routine
				// only needs to fill in the information
				// about the entity (particularly length and
				// last modified time), but can leave 
				// pEntityBody set to NULL.  Reading in the
				// entity body is allowed, and the HTTP server
				// will correctly free it if present.
} REQ_INPUTS;

//
// Output parameters from GetURI() and PostURI()
// If the URI is successfully read:
//      All members are optional except LastModTime and ContentLength.  If
//	ContentLength is nonzero, pEntityBody should be non-NULL.  If
//	StatusCode is not set, the standard "200 OK" status is sent to client.
// If an error occurs:
//	All members are optional except StatusCode.  For most errors, the 
//	HTTP spec recommends also sending a short HTML entity-body describing 
//	the error (check the HTTP spec for details).  If a
//	non-NULL entity-body is returned for an error, set the LastModTime
//	to the current time.
//
typedef struct {
	time_t	LastModTime;	// the last modified time for the content,
				// or the current time() if unknown, in
				// time_t format (seconds since 00:00:00 GMT
				// 1-Jan 1970)
	int	ContentLength;	// length of Entity-Body (could be 0)
	char	*pEntityBody;	// ptr to Entity-Body, NULL if ContentLength==0
				// The memory should be malloc()'d, and it is
				// the caller's responsibility to free() it
	time_t	Expires;	// "Expires:" date, in time_t format, or 0
	char	ContentType[64]; // "Content-Type:" text, or empty string
	char	ContentEncoding[64]; // "Content-Encoding:" text, or empty str
	char	StatusCode[64]; // Status Code string to send, or empty
				// string to send default "200 OK".  If error,
				// this MUST be set to an appropriate string.
	char	HeaderFields[256]; // extra header fields, each terminated
				// by \r\n (e.g. "Pragma: no-cache\r\n")
} REQ_OUTPUTS;

//
// This function is called for a GET or HEAD method, and reads the 
// content ("entity-body") for a requested URI into memory.
//
// NOTE: This function MUST be reentrant.  In the interests of good response 
//	times, HTTPSERV.LIB does NOT serialize calls to this function.  If
//	multiple HTTP requests arrive close together, two (or more) calls
//	to this function may execute concurrently.
//
void __cdecl GetURI(REQ_INPUTS *pInp, REQ_OUTPUTS *pOutp);

//
// This function is called for a POST method.  It processes the posted
// entity-body and optionally returns an entity-body to the client.
//
// NOTE: This function MUST be reentrant.  In the interests of good response 
//	times, HTTPSERV.LIB does NOT serialize calls to this function.  If
//	multiple HTTP requests arrive close together, two (or more) calls
//	to this function may execute concurrently.
//
void __cdecl PostURI(REQ_INPUTS *pInp, REQ_OUTPUTS *pOutp);

#endif

//
// Struct used to save record of each HTTP request in the log file
//
typedef struct request_record {
	char	RequestIPAddr[20]; // IP addr, in dotted ASCII
	char	RequestHostName[128]; // host name
	char	RequestTime[32]; // time, in ASCII, when request was received
	int	RequestLen;	// the length, including EOS, of the request
				// sent by the client (could be 0 if unable 
				// to allocate memory).  Does not include
				// entity-body for POST method
	char	*pRequestText;	// the request sent by the client (NULL if
				// RequestLen == 0)
	char	ResponseStatus[128]; // response status string
} REQUEST_RECORD;

//
// This function is called with a record of each processed HTTP request.
// The application can use this information to keep a log of server
// activity, if desired.
//
// NOTE: This function does not need to be reentrant.  Because it is called
//	after the HTTP request has been serviced, HTTPSERV.LIB is able to
//	serialize calls to this function without impacting response times.
//	However, as noted above, you still need to serialize access to
//	any global data also used by the main thread of program execution.
//
void __cdecl LogRequest(REQUEST_RECORD *pReq);

//
// This function is called if a server error occurs.
//	fFatal		T ==> fatal server error, the server DLL has done
//				the equivalent of a call to StopWebServer()
//				and is no longer responding to HTTP requests
//			F ==> non-fatal error (usually an error servicing
//				a specific HTTP request), the server is
//				still running and responding to HTTP requests
//	pErrMsg		a (possibly multi-line) error message
//
void __cdecl WebServerError(BOOL fFatal, char *pErrMsg);

//
// This function is called when the HTTP server library wants to warn the user
//	about the occurrence of a specific event.
//
//	pWarnMsg	a (possibly multi-line) warning message
//
void __cdecl WebServerWarning(char *pWarnMsg) ;

//
// This function is called when the HTTP server wants to inform the user about
//	the occurrence of a specific event.
//
//	pInfoMsg	a (possibly multi-line) informational message
//
void __cdecl WebServerInfo(char *pInfoMsg);

//
// The following structures, defines, and functions are used when converting
// query strings (the field pQuery in REQ_INPUTS) into a linked list of form
// variable pairs.
//

//
// Error codes past back from the GetFormVariableList() function.
//
#define FORM_NO_ERR		 0	// Success.
#define FORM_SYNTAX_ERR		-1	// Syntax error.
#define FORM_ALLOC_ERR		-2	// Unable to allocate memory.
#define FORM_MULT_FILE_ERR	-3	// More than one file body in the POST.


//
// This structure contains fields which hold a variable name and its setting, as
// well as a pointer to the next FORM_VARIABLE structure in the list.
//
typedef struct _FORM_VARIABLE_ {

   char *pVar ;		// Variable name
   char *pVal ;		// Variable setting

   struct _FORM_VARIABLE_ *pNext ;	// Pointer to next item in list.

} FORM_VARIABLE ;


//
// GetFormVariableList() takes a query string and returns back the first
// structure in a singly list of form variable names and their settings.  If
// their is a problem interpreting pQuery, NULL will be returned.  Optionally,
// the application writer can specify the address of a short which will be set
// to a value that translates to the reason the function failed.
//
// The order of the form variables in the linked list is defined by the order
// that the variables appear on the form (e.g.  The first field on the form
// will be the first item in the linked list).
//
// Warning:
//	This function may also return NULL if pQuery is the NULL string.  Make
//	sure to check *pErrVal.  This parameter will be set to FORM_NO_ERR if
//	pQuery was empty.
//
FORM_VARIABLE * __cdecl GetFormVariableList (char *pQuery, short *pErrVal) ;

//
// This function frees all of the memory used to build the linked list returned
// by GetFormVarList().
//
void __cdecl DeleteFormVariableList (FORM_VARIABLE *pVarList) ;


// restore default packing
#pragma pack()

#ifdef __cplusplus
}
#endif

#endif	// ETS_HTTPSERV_INCLUDED
