/* RCS_Id: multpost.c 10.2 1999/12/10 16:45:41 clark Exp $ */
//
// multpost.c - This module contains functions that can be used to break a
//	multipart form-data style POST into the query string (form field
//	settings) and the file body.
//

  //////////////
 // Includes //
//////////////
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#include <winsock.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <httpserv.h>

#include "httputil.h"


  /////////////
 // Defines //
/////////////

//
// The following strings are all used when parsing a form contained in the
// pEntityBody field of the REQ_INPUTS structure.
//
#define CR_LF			"\x0D\x0A"
#define QUOTE_CR_LF		"\"\x0D\x0A"
#define CR_LF_CR_LF		"\x0D\x0A\x0D\x0A"
#define COLON_SPACE		": "
#define SEMI_SPACE		"; "
#define QUOTE_SEMI_SPACE	"\"; "
#define CONTENT_DISPOSITION	"Content-Disposition: form-data; name=\""
#define CONTENT_TYPE		"Content-Type: "
#define FILENAME		"FILENAME=\""
#define END_OF_FORM		"--"
#define BOUNDARY_VAR		"boundary="


//
// These are the states used when parsing the entity body.
//
#define FIND_BOUNDARY		0
#define FIND_CONTENT_DISP	1
#define FIND_FILE_OR_VAR	2
#define GET_VARIABLE_NAME	3
#define GET_VARIABLE_VAL	4
#define GET_FILE_VAR		5
#define GET_FILE_VAL		6
#define GET_FILE_CONTENT_TYPE	7
#define GET_FILE_CONTENT	8
#define ADD_FIELD		9


// See if the character is a hex digit.
#define IsAlphaNumeric(ch) ( ((ch >= 'a') && (ch <= 'z')) \
			   || ((ch >= 'A') && (ch <= 'Z')) \
			   || ((ch >= '0') && (ch <= '9')) )

// Macro that converts a hex value between 0x0 and 0xF into a character between
// '0' and '9' or 'A' and 'F'.
#define HexToCh(Num) ((Num >= 0x00) && (Num <= 0x09)) ? \
		     '0' + Num : 'A' + (Num - 0x0A)


  //////////////////////
 // Type Definitions //
//////////////////////

//
// This structure contains the boundary string and the length of the
// boundary string used to seperated the fields in a form of type
// multipart/form-data.
//
typedef struct {

    char *pString;	// The string used as the boundary.

    DWORD Length;	// The length of the boundary string.

} BOUNDARY_INFO;


//
// This structure stores the information contained in a field of a form of type
// multipart/form-data.
//
typedef struct _FORM_FIELD_ {

    char *pName;			// The name associated with a particular
					// a particular field of a form.
    char *pValue;			// The value entered in a form field.

    struct _FORM_FIELD_ *pNext;	// A pointer to the next field structure.

    //
    // The following three fields deal with form fields that represent a file.
    // These will be NULL for fields that deal with text values and not files.
    //
    char *pContentType;		// The content type of the file.

    char *pFileBody;			// The contents of the file.

    DWORD FileLength;			// The length of the contents of the
					// file.

} FORM_FIELD;


//
// This structure contains the length of the query string that needs to be built
// from the form that we were sent.  This structure also points to a list of
// fields that represent the information sent in the form.
//
typedef struct {

    DWORD QueryLen;    // This is the combined length of all of the field names,
			// field values, and amphersands and equal signs that
			// will seperate them.  This length lets us know how big
			// the query string is.  The reason amphersands and
			// equal signs are added into the length is due to how
			// the query string is built.  A sample query string
			// would look like this:
			//
			//	field_1=BLUE&field_2=RED&field_3=YELLOW

    FORM_FIELD *pFirstField;	// A pointer to the first entry in the linked
				// list of fields that make up the form.

    FORM_FIELD *pLastField;	// A pointer to the last entry in the linked
				// list of fields that make up the form.

} FORM_DATA;


  ///////////////////////////////////
 // Private Function Declarations //
///////////////////////////////////

static short GetBoundary (char *pContentType, BOUNDARY_INFO *pBoundInfo);

static char *FindNextBoundary (char *pStartCh,
			       char *pEndCh,
			       BOUNDARY_INFO *pBoundInfo);

static short ParseEntityBody (REQ_INPUTS *pInp, FORM_DATA **pForm);

static short GetFormDataPostInfo (FORM_DATA *pForm,
				  REQ_INPUTS *pInp);

static FORM_FIELD *FormFieldNew         (void);
static char       *FormFieldSetString   (char *pStr, DWORD StrLen);
static BOOL        FormFieldSetFileBody (FORM_FIELD *pField,
				         char *pFileBody,
				         DWORD FileLen);
static void        FormFieldDelete      (FORM_FIELD *pField);
static void        FormDelete           (FORM_DATA *pForm);
static void        FormAddField         (FORM_DATA *pForm, FORM_FIELD *pField);


  /////////////////////////////////
 // Public Function Definitions //
/////////////////////////////////

//
// GetMultiPartPostInfo() - This function is used when setting pInp to a POST
//	of the form multipart form-data.  This function takes the query string
//	out of the entity body and moves it into the query string (pQuery).
//	If the POST sent a file, this file will be copied to the pEntityBody
//	field.  The ContentLength field will be set to the length of the file.
//
// Outputs:
//	pInp	If this function is successful, this structure will have a query
//		string containing the fields past in with the POST method,
//		the entity body will be set to the file passed in, and
//		ContentLength will be set to the length of that file.  If no
//		file was passed in, pEntityBody will be freed and set to NULL.
//		ContentLength will be set to zero.
//
// Return:
//	Success			FORM_NO_ERR		(0)
//	Fail (syntax)		FORM_SYNTAX_ERR		(-1)
//	Fail (malloc)		FORM_ALLOC_ERR		(-2)
//	Fail (multiple files)	FORM_MULT_FILE_ERR	(-3)
//
short GetMultiPartPostInfo (REQ_INPUTS *pInp)
{
    short RetVal;

    FORM_DATA *pForm;


    //
    // Parse the entity body to get all of the fields that make up the form
    // that was passed to us by the POST.
    //
    if ( (RetVal = ParseEntityBody(pInp, &pForm)) == FORM_NO_ERR ) {

	//
	// Take the form data we found and rebuild pInp with it.
	//
	RetVal = GetFormDataPostInfo(pForm, pInp);

	//
	// Delete the linked list used in rebuild pInp.
	//
	FormDelete(pForm);
    }

    return (RetVal);
}


  //////////////////////////////////
 // Private Function Definitions //
//////////////////////////////////

//
// GetBoundary() - Every multipart POST contains a boundary that is used to
//	divide up the fields in the form.  This function attempts to find that
//	boundary.
//
// Input:
//	pContentType	This field contains the boundary string.
//
// Output:
//	pBoundInfo	This structure holds the information about the boundary.
//			Upon success this structure will be filled in.
//
// Return:
//	Success		FORM_NO_ERR      (0)
//	Fail (syntax)	FORM_SYNTAX_ERR (-1)
//	Fail (malloc)	FORM_ALLOC_ERR  (-2)
//
static short GetBoundary (char *pContentType,
			  BOUNDARY_INFO *pBoundInfo)
{
    char  *pCurCh;
    char  *pTemp;


    pBoundInfo->pString = NULL;

    if ( strnicmp(pContentType,
		  MULTIPART_FORM_DATA,
		  strlen(MULTIPART_FORM_DATA)) )

	return FORM_SYNTAX_ERR;


    //
    // Advance past the multipart form-data string and look for the boundary
    // variable
    //
    pCurCh = pContentType + strlen(MULTIPART_FORM_DATA);

    if ( (pTemp = strstr(pCurCh, BOUNDARY_VAR)) == NULL )

	return FORM_SYNTAX_ERR;


    pCurCh = pTemp + strlen(BOUNDARY_VAR);

    //
    // The boundary should be right after the multipart form-data string.
    // We try to find the end of the string so that we can get the length of
    // the boundary.
    //
    if ( (pTemp = strchr(pCurCh, '\0')) == NULL )

	return FORM_SYNTAX_ERR;


    //
    // When the boundary is used inside of the actual POST, two hyphens are
    // added to it.
    //
    pBoundInfo->Length = (pTemp - pCurCh) + 2;

    pBoundInfo->pString = (char *) malloc((pBoundInfo->Length + 1)
					  * sizeof(*(pBoundInfo->pString)));

    if ( pBoundInfo->pString == NULL )

	return FORM_ALLOC_ERR;


    //
    // Set the boundary to the string we found along with two hyphens appended
    // to the front of it.
    //
    memset(pBoundInfo->pString,
	   (int) '\0',
	   sizeof(*(pBoundInfo->pString)) * (pBoundInfo->Length + 1));

    strcpy(pBoundInfo->pString, "--");

    strncat(pBoundInfo->pString, pCurCh, pBoundInfo->Length);

    pBoundInfo->pString[pBoundInfo->Length] = '\0';


    return (FORM_NO_ERR);
}


//
// FindNextBoundary() - This function scans from pStartCh to pEndCh looking for
//	the next occurence of the boundary string.
//
// Input:
//	pStartCh	A pointer to the character where the boundary search
//			will begin.
//	pEndCh		A pointer to the character where the search will end.
//	pBOundInfo	A pointer to the boundary related information.
//
// Return:
//	Success		A pointer to the boundary string.
//	Failure		NULL
//
static char *FindNextBoundary (char *pStartCh,
			       char *pEndCh,
			       BOUNDARY_INFO *pBoundInfo)
{
    char *pCur;
    char *pTemp;
    BOOL  fError;


    pCur = pStartCh;

    fError = TRUE;

    while ( pCur < pEndCh ) {

	// If there is not enough characters left between where we currently
	// are and the end of the entity body, we should break.  We were unable
	// to find the next boundary.
	if ( (pCur + pBoundInfo->Length) >= pEndCh )

	    break;

	// If the current character is the NULL character, move to the next one.
	if ( *pCur == '\0' )

	    pCur++;

	else {

	    // If we did not find the boundary string, move to the next NULL
	    // character.
	    if ( (pTemp = strstr(pCur, pBoundInfo->pString)) == NULL )

		pCur += strlen(pCur);

	    else {

		fError = FALSE;

		break;
	    }
	}
    }


    // If there were no errors, pTemp should point to the next boundary.
    return(fError ? NULL : pTemp);
}


//
// ParseEntityBody() - This function parses the entity body and extracts the
//	different pieces that compose the form passed to us by the POST.  This
//	information is all stored in a linked list.  Each element of the linked
//	list contains the name of a field and its settings.
//
// Input:
//	pInp		We are going to parse pEntityBody.
//	pBoundInfo	The boundary used to divide the form into fields.
//
// Output:
//	pForm		This parameter upon success will contain the linked list
//			of fields.
//
// Return:
//	Success			FORM_NO_ERR		(0)
//	Fail (syntax)		FORM_SYNTAX_ERR		(-1)
//	Fail (malloc)		FORM_ALLOC_ERR		(-2)
//	Fail (multiple files)	FORM_MULT_FILE_ERR	(-3)
//
static short ParseEntityBody (REQ_INPUTS *pInp,
			      FORM_DATA **pForm)
{
    char       *pCurCh;
    char       *pTemp;
    BOOL        fNotDone;
    BOOL        fNoFilesYet;
    short       ErrorType;
    short	ParseState;

    BOUNDARY_INFO  BoundInfo;
    FORM_FIELD    *pTempField;
    FORM_DATA     *pNewForm;


    if ( (ErrorType = GetBoundary(pInp->pContentType,
				  &BoundInfo)) != FORM_NO_ERR )

	return ErrorType;


    // First allocate a form
    pNewForm = (FORM_DATA *) malloc(sizeof(*pNewForm));

    if ( pNewForm == NULL )

	return FORM_ALLOC_ERR;


    memset(pNewForm, 0, sizeof(*pNewForm));


    pCurCh = pInp->pEntityBody;

    ErrorType = FORM_NO_ERR;

    fNotDone = fNoFilesYet = TRUE;

    ParseState = FIND_BOUNDARY;

    while ( fNotDone && (ErrorType == FORM_NO_ERR) ) {

	switch ( ParseState ) {

	    case FIND_BOUNDARY:

		// Initialize pTempField to NULL at the beginning of each parse.
		// We do this so that we can tell if we have allocated memory
		// for pTempField each time through the loop.
		pTempField = NULL;

		// The thing that proceeds any field assignment is a boundary.
		if ( strncmp(pCurCh,
			     BoundInfo.pString,
			     BoundInfo.Length) )

		    ErrorType = FORM_SYNTAX_ERR;

		else {

		    pCurCh += BoundInfo.Length;

		    // If we are at the end of the form, the boundary will be
		    // followed by two dashes.
		    if ( strnicmp(pCurCh,
				  END_OF_FORM,
				  strlen(END_OF_FORM)) == 0 ) {

			fNotDone = FALSE;
		    }

		    else

			ParseState = FIND_CONTENT_DISP;
		}

		break;

	    case FIND_CONTENT_DISP:

		// If we are not at the end of the form, the boundary should be
		// followed by a carriage return and a line-feed.
		if ( strnicmp(pCurCh, CR_LF, strlen(CR_LF)) )

		    ErrorType = FORM_SYNTAX_ERR;

		else {

		    pCurCh += strlen(CR_LF);

		    // The next thing we should see is the content disposition
		    // string.
		    if ( strnicmp(pCurCh,
				  CONTENT_DISPOSITION,
				  strlen(CONTENT_DISPOSITION)) )

			ErrorType = FORM_SYNTAX_ERR;

		    else {

			pCurCh += strlen(CONTENT_DISPOSITION);

			ParseState = FIND_FILE_OR_VAR;
		    }
		}

		break;

	    case FIND_FILE_OR_VAR:

		// The value for the field ends with a quote.  If the field
		// contains a file, the next thing should be a semi-colon
		// followed by a space.  If the field just contains a value,
		// the next thing is <CR><LF><CR><LF>.
		if ( (pTemp = strchr(pCurCh, '\"')) == NULL )

		    ErrorType = FORM_SYNTAX_ERR;

		else {

		    // If we do not find a semi-colon followed by a space, that
		    // means we have a field that contains a variable.
		    if ( strnicmp(pTemp + 1,
				  SEMI_SPACE,
				  strlen(SEMI_SPACE)) )

			ParseState = GET_VARIABLE_NAME;

		    else

			ParseState = GET_FILE_VAR;
		}

		break;

	    case GET_VARIABLE_NAME:

		// It appears we have a variable. Check for <CR><LF><CR><LF>.
		if ( strnicmp(pTemp + 1,
			      CR_LF_CR_LF,
			      strlen(CR_LF_CR_LF)) )

		    ErrorType = FORM_SYNTAX_ERR;

		else {

		    // If we get here, we have to allocate a FORM_FIELD
		    // structure to store the info contained in the field.
		    if ( (pTempField = FormFieldNew()) == NULL )

			ErrorType = FORM_ALLOC_ERR;

		    else {

			pTempField->pName = FormFieldSetString(pCurCh,
							       pTemp - pCurCh);

			if ( pTempField->pName == NULL )

			    ErrorType = FORM_ALLOC_ERR;

			// We add one for the equal sign that will follow the
			// name when we build the query string.
			pNewForm->QueryLen += (strlen(pTempField->pName) + 1);

			// Add an extra one for the quote character.
			pCurCh = pTemp + strlen(CR_LF_CR_LF) + 1 ;

			ParseState = GET_VARIABLE_VAL;
		    }
		}

		break;

	    case GET_VARIABLE_VAL:

		// After reading in the field name and value, we should be at
		// the boundary.
		if ( (pTemp = strstr(pCurCh, BoundInfo.pString)) == NULL )

		    ErrorType = FORM_SYNTAX_ERR;

		else {

		    // Back pTemp up two and we should see <CR><LF>.
		    pTemp -= 2;

		    if ( strnicmp(pTemp, CR_LF, strlen(CR_LF)) )

			ErrorType = FORM_SYNTAX_ERR;

		    else {

			// Set the field name to its value (the value specified
			// in the form).
			pTempField->pValue = FormFieldSetString(pCurCh,
								pTemp - pCurCh);

			if ( pTempField->pValue == NULL )

			    ErrorType = FORM_SYNTAX_ERR;

			else {

			    // We add one for the amphersand that will follow
			    // the name when we build the query string.
			    pNewForm->QueryLen += (strlen(pTempField->pValue)
						   + 1);

			    pCurCh = pTemp + strlen(CR_LF);

			    ParseState = ADD_FIELD;
			}
		    }
		}

		break;

	    case GET_FILE_VAR:

		// We already have read in a file.  We do not allow forms
		// containing more than one file.
		if ( fNoFilesYet )

		    fNoFilesYet = FALSE;

		else {

		    ErrorType = FORM_MULT_FILE_ERR;

		    break;
		}

		// If we get here, we have to allocate a FORM_FIELD structure
		// to store the info contained in the field.
		if ( (pTempField = FormFieldNew()) == NULL )

		    ErrorType = FORM_ALLOC_ERR;

		else {

		    pTempField->pName = FormFieldSetString(pCurCh,
							   pTemp - pCurCh);

		    if ( pTempField->pName == NULL )

			ErrorType = FORM_ALLOC_ERR;

		    else {

			// We add one for the equal sign that will follow the
			// name when we build the query string.
			pNewForm->QueryLen += (strlen(pTempField->pName) + 1);

			ParseState = GET_FILE_VAL;
		    }
		}

		break;

	    case GET_FILE_VAL:

		pCurCh = pTemp + strlen(QUOTE_SEMI_SPACE);

		// We should see the filename string now.
		if ( strnicmp(pCurCh, FILENAME, strlen(FILENAME)) )

		    ErrorType = FORM_SYNTAX_ERR;

		else {

		    pCurCh += strlen(FILENAME);

		    // The filename string should be followed by a quote, a
		    // <CR>, and an <LF>.
		    if ( (pTemp = strstr(pCurCh, QUOTE_CR_LF)) == NULL )

			ErrorType = FORM_SYNTAX_ERR;

		    else {

			pTempField->pValue = FormFieldSetString(pCurCh,
								pTemp - pCurCh);

			if ( pTempField->pValue == NULL )

			    ErrorType = FORM_ALLOC_ERR;

			else {

			    // We add one for the amphersand that will follow
			    // the name when we build the query string.
			    pNewForm->QueryLen += (strlen(pTempField->pValue)
				                  + 1);

			    pCurCh = pTemp + strlen(QUOTE_CR_LF);

			    ParseState = GET_FILE_CONTENT_TYPE;
			}
		    }
		}

		break;

	    case GET_FILE_CONTENT_TYPE:

		// We should now be at the string that describes the
		// content-type of the file.
		if ( strnicmp(pCurCh, CONTENT_TYPE, strlen(CONTENT_TYPE)) ) {

		    // If we do not find the content-type string, that means
		    // that the field where the file name is specified was left
		    // blank or the content-type could not be determined.
		    // Therefore, there should be no content-type string.
		    // There will be <CR><LF><CR><LF> if the file name field
		    // was not filled in.  There will be <CR><LF> if the file
		    // was filled in, but the content-type could not be
		    // determined.
		    if ( strnicmp(pCurCh, CR_LF, strlen(CR_LF)) )

			ErrorType = FORM_SYNTAX_ERR;

		    else {

			pCurCh += strlen(CR_LF);

			// Check to see if there is another <CR><LF>.  If there
			// is, the file name was not specified in the form.
			if ( strnicmp(pCurCh, CR_LF, strlen(CR_LF)) )

			    ParseState = GET_FILE_CONTENT;

			else {

			    pCurCh += strlen(CR_LF);

			    ParseState = ADD_FIELD;
			}
		    }
		}

		else {

		    pCurCh += strlen(CONTENT_TYPE);

		    // Following the content-type string, there should be a
		    // string that describes the file's content (text, html,
		    // etc).  This string should be followed by
		    // <CR><LF><CR><LF>.
		    if ( (pTemp = strstr(pCurCh, CR_LF_CR_LF)) == NULL )

			ErrorType = FORM_SYNTAX_ERR;

		    else {

			pTempField->pContentType = SetNewString(pCurCh,
								pTemp - pCurCh);

			if ( pTempField->pContentType == NULL )

			    ErrorType = FORM_ALLOC_ERR;

			else {

			    pCurCh = pTemp + strlen(CR_LF_CR_LF);

			    ParseState = GET_FILE_CONTENT;
			}
		    }
		}

		break;

	    case GET_FILE_CONTENT:

		// The boundary preceded by <CR><LF> marks the end of the body
		// of the file.
		if ( (pTemp = FindNextBoundary(pCurCh,
					       pInp->pEntityBody
					       + (pInp->ContentLength - 1),
					       &BoundInfo)) == NULL )

		    ErrorType = FORM_SYNTAX_ERR;

		else {

		    // Back pTemp up two and we should see <CR><LF>.
		    pTemp -= 2;

		    if ( strnicmp(pTemp, CR_LF, strlen(CR_LF)) )

			ErrorType = FORM_SYNTAX_ERR;

		    else {

			if ( !FormFieldSetFileBody(pTempField,
						   pCurCh,
						   pTemp - pCurCh) )

			    ErrorType = FORM_ALLOC_ERR;

			else {

			    pCurCh = pTemp + strlen(CR_LF);

			    ParseState = ADD_FIELD;
			}
		    }
		}

		break;

	    case ADD_FIELD:

		FormAddField(pNewForm, pTempField);

		ParseState = FIND_BOUNDARY;

		break;
	}
    }

    if (BoundInfo.pString)
	free(BoundInfo.pString);

    // If we ended the while loop without setting fNotDone to FALSE, there was
    // an error.
    if ( fNotDone ) {

	FormFieldDelete(pTempField);

	FormDelete(pNewForm);

	return ErrorType;
    }

    *pForm = pNewForm;

    return FORM_NO_ERR;
}


//
// GetFormDataPostInfo() - This function is used when setting pInp to a POST of
//	the form multipart form-data.  This function fills in the REQ_INPUTS
//	structure with the data that we obtained after parsing the entity body.
//
// Inputs:
//	pForm	This structure contains the information obtained while parsing
//		the entity body.
//
// Outputs:
//	pInp	If this function is successful, this structure will have a query
//		string containing the fields past in with the POST method, and
//		the entity body will be set to the contents of the file sent, if
//		there was a file sent.
//
// Return:
//	Success			FORM_NO_ERR	(0)
//	Fail (malloc)		FORM_ALLOC_ERR	(-2)
//
static short GetFormDataPostInfo(FORM_DATA *pForm, REQ_INPUTS *pInp)
{
    FORM_FIELD *pCurField;
    DWORD       NewLen;
    char       *pNewQuery;
    short       RetVal;


    //
    // If the query string already has something in it, we have to append to
    // the query string.  If the query string is NULL, we have to allocate
    // and build the query string.
    //
    if ( pInp->pQuery == NULL ) {

	//
	// Allocate the memory for the new query string.
	//
	pInp->pQuery = (char *) malloc(sizeof(*pInp->pQuery)
				       * (pForm->QueryLen + 1));

	if ( pInp->pQuery != NULL )

	    memset(pInp->pQuery,
		   (int) '\0',
		   sizeof(*pInp->pQuery) * (pForm->QueryLen + 1) );
    }

    else {

	// We add two to account for a NULL character and an ampersand.
	NewLen = pInp->ContentLength + strlen(pInp->pQuery) + 2;

	pNewQuery = BuildNewQueryString(pInp->pQuery, &NewLen,
					pForm->QueryLen != 0);

//	if (pInp->pQuery)	// not necesary, see if above
	free(pInp->pQuery);

	pInp->pQuery = pNewQuery;
    }

    if ( pInp->pQuery == NULL )

	return FORM_ALLOC_ERR;


    //
    // Free the entity body and set it to NULL.  Set the content length to 0.
    // If we have a file body, these fields will be set to new values.
    //
    if (pInp->pEntityBody)
    {
	free(pInp->pEntityBody);
	pInp->pEntityBody = NULL;
    }

    pInp->ContentLength = 0;

    RetVal = FORM_NO_ERR;

    pCurField = pForm->pFirstField;

    while ( (pCurField != NULL) && (RetVal == FORM_NO_ERR) ) {

	strcat(pInp->pQuery, pCurField->pName);

	strcat(pInp->pQuery, "=");

	strcat(pInp->pQuery, pCurField->pValue);

	if ( pCurField->FileLength > 0 ) {

	    pInp->pEntityBody = (char *) malloc(sizeof(*pCurField->pFileBody)
						* pCurField->FileLength);

	    if ( (pInp->pEntityBody == NULL) )

		RetVal = FORM_ALLOC_ERR;

	    else {

		pInp->ContentLength = pCurField->FileLength;

		memcpy(pInp->pEntityBody,
		       pCurField->pFileBody,
		       pInp->ContentLength);
	    }
	}

	pCurField = pCurField->pNext;

	if ( pCurField != NULL )

	    strcat(pInp->pQuery, "&");
    }

    // If we had an error, we do not want to return a partially build query
    // string.
    if (RetVal != FORM_NO_ERR)
    {
	if (pInp->pQuery)
	{
	    free(pInp->pQuery);
	    pInp->pQuery = NULL;
	}
    }
    return (RetVal);
}


//
// The following functions all perform operations on the FORM_FIELD structure
// and the FORM_DATA structure. These structures represent pieces of the form
// that was passed in via the POST method.
//

//
// FormFieldNew() - Allocates a new field.
//
// Results:
//	Success		A pointer to a FORM_FIELD structure.
//	Failure		NULL
//
static FORM_FIELD *FormFieldNew (void)
{
    FORM_FIELD *pFormField;

    pFormField = (FORM_FIELD *) malloc (sizeof(*pFormField));

    if ( pFormField != NULL )

	memset(pFormField, 0, sizeof(*pFormField));


    return(pFormField);
}


//
// FormFieldSetString() - This function makes builds a string and converts all
//	non-alphanumeric characters to escape sequences.
//
// Inputs:
//	pStr	The source string.
//	StrLen	The source string length.
//
// Return:
//	Success		A pointer to the new string
//	Failure		NULL (Could not allocate memory).
//
static char *FormFieldSetString (char *pStr, DWORD StrLen)
{
    char  *pCur;
    char  *pAdd;
    char  *pNewStr;
    DWORD  i;
    DWORD  NewLen;


    // If the string is empty pass back an empty string.
    if ( StrLen == 0 )

	return (SetNewString(pStr, StrLen));


    NewLen = 0;

    // Get the length of the new string.  The old string will not have enough
    // memory if some characters need to be converted to escape sequences.
    for ( i = 0, pCur = pStr; i < StrLen; i++, pCur++ ) {

	if ( IsAlphaNumeric(*pCur) || (*pCur == ' ') )

	    NewLen++;

	else

	    NewLen += 3;
    }

    pNewStr = (char *) malloc(sizeof(*pNewStr) * (NewLen + 1));

    if ( pNewStr == NULL )

	return (NULL);


    // If the new length does not equal the previous string length, we found
    // some non-alphanumeric characters.
    if ( NewLen == StrLen )

	strncpy(pNewStr, pStr, StrLen);

    else {

	for ( i = 0, pAdd = pNewStr, pCur = pStr;
	      i < StrLen;
	      i++, pAdd++, pCur++ ) {

	    if ( IsAlphaNumeric(*pCur) || (*pCur == ' ') )

		*pAdd = *pCur;

	    else {

		*pAdd = '%';

		pAdd++;

		*pAdd = HexToCh((char)(*pCur / NONALPHA_CHAR_BASE));

		pAdd++;

		*pAdd = HexToCh((char)(*pCur % NONALPHA_CHAR_BASE));
	    }
	}
    }

    pNewStr[NewLen] = '\0';

    return (pNewStr);
}


//
// FormFieldSetFileBody() - This field allocates memory for the pFileBody field
//	of the FORM_DATA structure and then sets the pFileBody field to the
//	contents of the file passed in as the second parameter.
//
// Input:
//	pFileBody	The file to copy.
//	FileLen		The length of the file to copy.
//
// Output:
//	pField		The pFielBody and FileLength fields will be filled in if
//			this file is successful.
// Results:
//	Success		TRUE
//	Failure		FALSE
//
static BOOL FormFieldSetFileBody (FORM_FIELD *pField,
				  char *pFileBody,
				  DWORD FileLen)
{
    pField->FileLength = FileLen;

    if ( pField->FileLength == 0 ) {

	pField->pFileBody = NULL;

	return TRUE;
    }


    pField->pFileBody = (char *) malloc(sizeof(*(pField->pFileBody))
					* pField->FileLength);

    if ( pField->pFileBody == NULL )

	return FALSE;


    memcpy(pField->pFileBody, pFileBody, pField->FileLength);

    return TRUE;
}


//
// FormFieldDelete() - This function deletes the FORM_FIELD passed into it
//
// Input:
//	pField	The form field to delete.
//
static void FormFieldDelete (FORM_FIELD *pField)
{
    if (pField)
    {
	if (pField->pName)
	    free(pField->pName);
	if (pField->pValue)
	    free(pField->pValue);
	if (pField->pContentType)
	    free(pField->pContentType);
	if (pField->pFileBody)
	    free(pField->pFileBody);

	free(pField);
    }
}


//
// FormDelete() - This function deletes a FORM_DATA structure as well as all of
//	its contents.
//
// Input:
//	pForm	A pointer to the form to be deleted.
//
static void FormDelete(FORM_DATA *pForm)
{
    FORM_FIELD *pNextField;
    FORM_FIELD *pCurField;

    if (pForm)
    {
	pCurField = pForm->pFirstField;

	while (pCurField)
	{
	    pNextField = pCurField->pNext;
	    FormFieldDelete(pCurField);
	    pCurField = pNextField;
	}
	free(pForm);
    }
}


//
// FormAddField() - This function adds a field to the list of fields in the
//	FORM_DATA structure.
//
// Input:
//	pField	The field to be added.
//
// Output:
//	pForm	A pointer to the form where the field will be added.
//
static void FormAddField (FORM_DATA *pForm, FORM_FIELD *pField)
{
    if ( pForm->pLastField == NULL )

	pForm->pFirstField = pForm->pLastField = pField;

    else {

	pForm->pLastField->pNext = pField;

	pForm->pLastField = pField;
    }
}
