#include <stdio.h>

#include <opclib.h>
#include <dsassert.h>
#include <dserror.h>
#include <dsdebug.h>

#include <opc.h>
#include <dssignal.h>
#include <opcxmldarequestpool.h>

#include <opcdefines.h>
#include <opcerrors.h>

#include <opcconfig.h>
#include <backend.h>
#include <opcserver.h>

#ifdef OPC_EMBEDED_HTTP
# include <microweb.h>
#endif /* OPC_EMBEDED_HTTP */

int opcServerInit() {
    return opcInit();
}

void opcServerDestroy() {
    opcDestroy();
}

void opcServerZeroContext(OPCServerContext ctx) {
    ds_assert_return(ctx);

    ctx->ctx=NULL;
    ctx->sigctx=NULL;
    ctx->rpctx=NULL;
#ifdef OPC_EMBEDED_HTTP
#else
    ctx->fctx=NULL;
#endif /* ! OPC_EMBEDED_HTTP */
    ctx->bctx=NULL;
    ctx->buffpool=NULL;
    ctx->status=0;
    ctx->backend=NULL;
    ctx->frontend=NULL;
    ctx->config=NULL;
}


OPCServerContext opcServerCreateContext() {
    DSMallocContext mctx;
    OPCServerContext ctx;
    
    dsDebugServer(DS_DEBUG_NOTICE,"opcServerCreateContext");
    
    mctx = dsMallocGetDefault();
    ctx = (OPCServerContext)dsMalloc(mctx, sizeof(struct OPCServerContextT));
    if (ctx) {
	ctx->malloc = mctx;
	opcServerZeroContext(ctx);
    }
    return ctx;
}

int opcServerInitContext(OPCServerContext ctx, const char *opc_directory, unsigned char setup) {
    int err;
    
    dsDebugServer(DS_DEBUG_NOTICE,"opcServerInitContext");
    
    ds_assert_param(ctx);

    ctx->config=opcConfigCreateContext(ctx->malloc);
    if (!ctx->config) {
	opcServerFreeContext(ctx);
	return OPC_ERR_SERVER_CONFIG_CREATE;
    } else ctx->config->opc.destructor = &dsDestroyContextSkip;
    
    err = opcConfigInitContext(ctx->config, opc_directory, OPC_DEFAULT_CONFIG);
    if (err) {
	opcServerFreeContext(ctx);
	return err;
    }
    
    ctx->sigctx=dsSignalCreateContext(ctx->malloc);
    if (!ctx->sigctx) {
	opcServerFreeContext(ctx);
	return DS_ERR_SIGNAL_CREATE;
    }

    err=dsSignalInitContext(ctx->sigctx);
    if (err) {
	opcServerFreeContext(ctx);
	return err;
    }

    ctx->buffpool=dsBufferPoolCreateContext(ctx->malloc);
    if (!ctx->buffpool) {
	opcServerFreeContext(ctx);
	return DS_ERR_BUFFERPOOL_CREATE;
    }
    
    err = dsBufferPoolInitContext(ctx->buffpool, OPC_BUFFER_SIZE, 1);
    if (!err) err = dsBufferPoolSetDefaultSize(ctx->buffpool, OPC_BUFFER_SIZE);
    if (err) {
	opcServerFreeContext(ctx);
	return err;
    }

    ctx->ctx=opcCreateContext(ctx->malloc);
    if (!ctx->ctx) {
	opcServerFreeContext(ctx);
	return OPC_ERR_CREATE;
    }

    err=opcInitContext(ctx->ctx, (OPCConfigurationContext)ctx->config, 0);
    if (err) {
	opcServerFreeContext(ctx);
	return err;
    }
    
    ctx->status|=OPC_SERVER_CONTEXT_INITIALIZED;

    ctx->rpctx=opcXmldaRequestPoolCreateContext(ctx->malloc);
    if (!ctx->rpctx) {
	opcServerFreeContext(ctx);
	return OPC_ERR_XLIB_CREATE;
    }

    err=opcXmldaRequestPoolInitContext(ctx->rpctx,ctx->ctx);
    if (err) {
	opcServerFreeContext(ctx);
	return err;
    }
    
    if ((setup&OPC_SERVER_NO_BACKEND)==0) {
	ctx->bctx=opcBackendCreateContext(ctx->malloc);
	if (!ctx->bctx) {
	    opcServerFreeContext(ctx);
	    return OPC_ERR_SERVER_BACKEND_CREATE;
	}
    
	err = opcBackendInitContext(ctx->bctx, ctx->config, ctx->ctx, ctx->sigctx);
	if (err) {
	    opcServerFreeContext(ctx);
	    return err;
	}
    }
    
    if ((setup&OPC_SERVER_NO_FRONTEND)==0) {
#ifdef OPC_EMBEDED_HTTP
	MicrowebConfig(ctx->config, ctx->rpctx, ctx->buffpool, ctx->sigctx);	
#else
	ctx->fctx=opcFrontendCreateContext(ctx->malloc);
	if (!ctx->fctx) {
	    opcServerFreeContext(ctx);
	    return OPC_ERR_SERVER_FRONTEND_CREATE;
	}
    
	err=opcFrontendInitContext(ctx->fctx, ctx->config, ctx->rpctx, ctx->buffpool, ctx->sigctx);
        if (err) {
	    opcServerFreeContext(ctx);
	    return err;
	}
#endif /* OPC_EMBEDED_HTTP */
    }

    if ((setup&OPC_SERVER_NO_BACKEND)==0) {
	ctx->backend = dsThreadCreateContext(ctx->malloc);
	if (!ctx->backend) {
	    opcServerFreeContext(ctx);
	    return DS_ERR_THREAD_CREATE;
	}
	err = dsThreadInitContext(ctx->backend);
        if (err) {
    	    opcServerFreeContext(ctx);
	    return err;
	}
    }

    if ((setup&OPC_SERVER_NO_FRONTEND)==0) {
	ctx->frontend = dsThreadCreateContext(ctx->malloc);
	if (!ctx->frontend) {
	    opcServerFreeContext(ctx);
	    return DS_ERR_THREAD_CREATE;
	}
	err = dsThreadInitContext(ctx->frontend);
        if (err) {
	    opcServerFreeContext(ctx);
	    return err;
	}
    }

    
    return 0;
}

    // All spawned childs must been destroyed prior to this call
void opcServerFreeContext(OPCServerContext ctx) {
    int err=0;

    ds_assert_return(ctx);

    dsDebugServer(DS_DEBUG_NOTICE, "opcServerFreeContext");

    if (ctx->status&OPC_SERVER_CONTEXT_INITIALIZED) opcServerTerminate(ctx);

    if (ctx->frontend) {
	err = dsThreadDestroyContext(ctx->frontend);
	dsDebugServer(err?DS_DEBUG_ERROR:DS_DEBUG_INFO,"Terminating Frontend Thread: %s", dsError(err,NULL));
    }
    if (ctx->backend) {
	err = dsThreadDestroyContext(ctx->backend);
	dsDebugServer(err?DS_DEBUG_ERROR:DS_DEBUG_INFO,"Terminating Backend Thread: %s", dsError(err,NULL));
    }
    
#ifdef OPC_EMBEDED_HTTP
#else
    if (ctx->fctx) opcFrontendDestroyContext(ctx->fctx);
#endif /* OPC_EMBEDED_HTTP */
    if (ctx->bctx) opcBackendDestroyContext(ctx->bctx);
    if (ctx->rpctx) opcXmldaRequestPoolDestroyContext(ctx->rpctx);
    if (ctx->ctx) opcDestroyContext(ctx->ctx);
    if (ctx->buffpool) dsBufferPoolDestroyContext(ctx->buffpool);
    if (ctx->sigctx) dsSignalDestroyContext(ctx->sigctx);
    if (ctx->config) opcConfigDestroyContext(ctx->config);
    opcServerZeroContext(ctx);
}

void opcServerDestroyContext(OPCServerContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcServerDestroyContext");
    
    ds_assert_return(ctx);
    opcServerFreeContext(ctx);
    dsFree(ctx->malloc, ctx);
}

int opcServerRun(OPCServerContext ctx) {
    int err;

    dsDebugServer(DS_DEBUG_NOTICE, "opcServerRun");
    
    ds_assert_param(ctx);
    
    if (ctx->backend) {
	err = dsThreadRun(ctx->backend, (DSThreadFunction)opcBackendListen, (void*)ctx->bctx);
	if (err) return err;
	else ctx->status|=OPC_SERVER_BACKEND_INITIALIZED;
    }
    
#ifdef OPC_EMBEDED_HTTP
    err = MicrowebStart();
    if (err) {
	opcServerTerminate(ctx);
	return OPC_ERR_SERVER;
    }
#else
    if (ctx->frontend) {
	err = dsThreadRun(ctx->frontend, (DSThreadFunction)opcFrontendListen, (void*)ctx->fctx);
	if (err) {
	    opcServerTerminate(ctx);
	    return err;
	} else ctx->status|=OPC_SERVER_FRONTEND_INITIALIZED;
    }
#endif /* OPC_EMBEDED_HTTP */

    return 0;
}

int opcServerTerminate(OPCServerContext ctx) {
    ds_assert_param(ctx);

#ifdef OPC_EMBEDED_HTTP
    MicrowebStop();
#endif /* OPC_EMBEDED_HTTP */
    
    ctx->status|=OPC_SERVER_TERMINATE_FLAG;
    
    if (ctx->sigctx) dsSignalTerminate(ctx->sigctx);

    return 0;
}

void opcServerMantain(OPCServerContext ctx) {
    dsDebugServer(DS_DEBUG_INFO, "opcServerMantain");
    ds_assert_return(ctx);

#ifdef OPC_EMBEDED_HTTP
#else
    if (ctx->fctx) opcFrontendCleanupChilds(ctx->fctx);
#endif /* OPC_EMBEDED_HTTP */
    opcMantain(ctx->ctx);
}
